# Copyright (C) 2019-2021:
#   Max Planck Institut fuer extraterrestrische Physik
#
# This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
# redistribute it and/or modify it under the terms of the GNU General
# Public License as published by the Free Software Foundation; either
# version 2 of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this library; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
# USA, or see https://www.gnu.org/licenses/.
#

#!/bin/tcsh -f

# -----------------------------------------------------------------------------
#  Author: T. Dwelly (dwelly@mpe.mpg.de)
# 
#  First version: 2017-06-14
#  Current version controlled by CVS:
set CVS_REVISION = '$Revision$'
#  CVS Date: $Date$
#
# -----------------------------------------------------------------------------
# 
# -----------------------------------------------------------------------------
#  Summary:
#   This is a demonstration script that builds FITS format Analytic Vignetting 
#   description files for insertion into the eROSITA CALDB 
#   see https://wiki.mpe.mpg.de/eRosita/EroCat/AnalyticVignetting
#  -----------------------------------------------------------------------------

set BASE = ${HOME}/erosita/avign_caldb
mkdir -p $BASE
cd $BASE

#set INST_STEM = "fm"
set INST_STEM = "tm"

if ( $#argv == 1 ) then
  set INST_STEM = "$argv[1]"
  echo "Using suser-supplied INST_STEM=$INST_STEM" 
endif

# list of eROSITA telescopes (${INST_STEM}i)
set INST_LIST = ( 1 2 3 4 5 6 7 ) 

#  this is the list of monochromatic energies at which the vignetting is 
# measured at PANTER
#set ENERGY_LIST = ( 0.277 0.930 1.486 4.508 6.398 8.040 ) 
## copy these from swift-XRT : /home/tdwelly/swift/caldb/data/swift/xrt/cpf/vign/swxvign20010101v001.fits
#set NOMINAL_COEFFS = (  1.0  0.000124799 1.55219  0.00185628 0.0 ) 
#set COEFFS_LIST = ( "$NOMINAL_COEFFS" "$NOMINAL_COEFFS" "$NOMINAL_COEFFS" "$NOMINAL_COEFFS" "$NOMINAL_COEFFS" "$NOMINAL_COEFFS" ) 

#
# New coeffs derived from fit to SIXTE vignetting model: 
#   share/sixte/instruments/srg/erosita_vignetting_v2.1.fits
#   supplied in the SIXTE instrument description package v1.4.0 
#   http://www.sternwarte.uni-erlangen.de/research/sixte/downloads/sixte/instruments/instruments_srg-1.4.0.tar.gz
# See /home/tdwelly/erosita/eSASSvsSIXTE/derive_analytic_vignetting_coeffs.plot for details
set ENERGY_LIST = ( 1.25 4.0 7.5 ) 
set COEFFS_LIST = ( "0.44966  -1.02111  1.71288"   "0.22575 -0.83151 1.97408"   "0.10547 -0.50753 2.52732" ) 
set NUM_COEFFS = 3



set NOMINAL_OA_X = "192.5"
set NOMINAL_OA_Y = "192.5"
set NOMINAL_PIXSCALE_X = "2.6666666667e-3"
set NOMINAL_PIXSCALE_Y = "2.6666666667e-3"

set OA_X_LIST = ( $NOMINAL_OA_X $NOMINAL_OA_X $NOMINAL_OA_X $NOMINAL_OA_X $NOMINAL_OA_X $NOMINAL_OA_X ) 
set OA_Y_LIST = ( $NOMINAL_OA_Y $NOMINAL_OA_Y $NOMINAL_OA_Y $NOMINAL_OA_Y $NOMINAL_OA_Y $NOMINAL_OA_Y ) 
set PIXSCALE_X_LIST = ( $NOMINAL_PIXSCALE_X $NOMINAL_PIXSCALE_X $NOMINAL_PIXSCALE_X $NOMINAL_PIXSCALE_X $NOMINAL_PIXSCALE_X $NOMINAL_PIXSCALE_X ) 
set PIXSCALE_Y_LIST = ( $NOMINAL_PIXSCALE_Y $NOMINAL_PIXSCALE_Y $NOMINAL_PIXSCALE_Y $NOMINAL_PIXSCALE_Y $NOMINAL_PIXSCALE_Y $NOMINAL_PIXSCALE_Y ) 


set CALDB_DATE_VERSION = "010101v01"

# write the structure of the binary table
echo "\
ENERGY     1D             keV       \
COEFFS     ${NUM_COEFFS}D           \
OA_X       1D             pix       \
OA_Y       1D             pix       \
PIXSCALE_X 1D             deg/pix   \
PIXSCALE_Y 1D             deg/pix   " > cdfile


# note that these caldb files should really be in the CPF directory, sance they contain computed parameters.
# However, the policy for eROSITA is to put everything ibn the bcf/ directory, and so we use BCF in the header too (CCLS0001)
echo "HDUCLASS= 'OGIP    ' \
HDUCLAS1= 'RESPONSE' \
HDUVERS1= '1.0.0   ' \
HDUCLAS2= 'VIGNET  ' \
HDUVERS2= '1.1.0   ' \
MISSION = 'SRG     ' \
TELESCOP= 'eROSITA ' \
DETNAM  = 'NONE    ' \
INSTRUME= 'XXX     ' \
CDES0001= 'eROSITA Vignetting Coefficients' \
CCLS0001= 'BCF     ' \
CDTP0001= 'TASK    ' \
CCNM0001= 'AVIGNET ' \
CBD10001= 'ENERG(100.0-15000.0)eV' \
CBD20001= 'THETA(0.0-60.0)arcmin' \
CBD30001= 'OBSMODE(POINTING)' \
CVSD0001= '2001-01-01' \
CVST0001= '00:00:00' \
CREATOR = ${0:t} / (CVS:${CVS_REVISION}) - Software used \
DATE    = `date +'%Y-%m-%dT%H:%M:%S.%3N'` / Date file created yyyy-mm-ddTHH:MM:SS.sss" > ${BASE}/headfile
# add any other required header keywords

# now create the files
foreach INST ( $INST_LIST ) 
  set INST_NAME = "${INST_STEM}${INST}"

  set OUTDIR   = "${BASE}/${INST_NAME}/bcf"
  set OUTFILE  =  ${OUTDIR}/${INST_NAME}_avignet_${CALDB_DATE_VERSION}.fits
  set HEADFILE =  ${OUTDIR}/headfile   
  set DATAFILE =  ${OUTDIR}/datafile
  set CDFILE   =  ${BASE}/cdfile
  mkdir -p ${OUTDIR}
  if ( -e $OUTFILE ) mv $OUTFILE $OUTFILE~
  echo "Working on telescope: $INST_NAME => $OUTFILE"

  # replace the INSTRUME keyword with appropriate value
  gawk -v instname="${INST_NAME}" '//{gsub("XXX",toupper(instname),$0);print $0}' ${BASE}/headfile > $HEADFILE

  # now build the data file for each instrument, this could be done manually
  if ( -e $DATAFILE ) rm $DATAFILE
  set E = 1
  while ( $E <= $#ENERGY_LIST ) 
    set COEFFS_E = ( $COEFFS_LIST[$E] )
 #   set COEFFS_E[1] = $ENERGY_LIST[$E]
    echo "$ENERGY_LIST[$E] $COEFFS_E $OA_X_LIST[$E] $OA_Y_LIST[$E] $PIXSCALE_X_LIST[$E] $PIXSCALE_Y_LIST[$E]" >> $DATAFILE
    @ E ++
  end

  # now convert to FITS
  ftcreate cdfile=$CDFILE datafile=$DATAFILE outfile=$OUTFILE headfile=$HEADFILE tabtyp=BINARY extname="VIGN_COEFF" clobber=yes mode=q history=no

  rm ${OUTDIR}/headfile ${OUTDIR}/datafile

  # fstruct $OUTFILE mode=q
end


rm ${BASE}/headfile ${BASE}/cdfile






exit



# To install
foreach INST ( 1 2 3 4 5 6 7 )
  cp -pv /home/tdwelly/erosita/avign_caldb/tm${INST}/bcf/tm${INST}_avignet_010101v01.fits /home/erosita/caldb/data/erosita/tm${INST}/bcf/
  cp -pv /home/tdwelly/erosita/avign_caldb/fm${INST}/bcf/fm${INST}_avignet_010101v01.fits /home/erosita/caldb/data/erosita/fm${INST}/bcf/
end

##########



# to add to cif
setenv CALDB /data28s/erosita/caldb
foreach INST ( 1 2 3 4 5 6 7 )
  foreach TMFM ( tm fm )
    cd ${CALDB}/data/erosita/${TMFM}${INST}/bcf
    udcif ${TMFM}${INST}_avignet_010101v01.fits ../caldb.indx quality=0 editc=yes
  end
end






