! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_arf

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-25
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling ARF data/files
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_ARF_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_arf
  public  :: srctool_arf_empty_arf
  public  :: srctool_arf_empty_arflist
  public  :: srctool_arf_read_arf_from_chdu
  public  :: srctool_arf_setup_dummy_arf
  private :: srctool_arf_setup_arf
  public  :: srctool_arf_setup_arflist
  public  :: srctool_arf_coadd_arflist
  private :: srctool_arf_write_arf_keys
  private :: srctool_arf_write_arf
  public  :: srctool_arf_write_arflist


contains






  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of ARF handling -----------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_arf_empty_arf(arf)
    type(arf_type), intent(inout)      :: arf
!    integer(kind=int32) :: i
!    character(len=STR_MAX) :: str_buffer
    if (allocated(arf%pelo))      deallocate (arf%pelo)
    if (allocated(arf%pehi))      deallocate (arf%pehi)
    if (allocated(arf%presp))     deallocate (arf%presp)
    if (allocated(arf%pcorrcomb)) deallocate (arf%pcorrcomb)
    if (allocated(arf%pcorrpsf))  deallocate (arf%pcorrpsf)
    if (allocated(arf%pcorrvign)) deallocate (arf%pcorrvign)

    arf%num_rows = 0
    arf%is_valid = .false.
    arf%to_consider = .false.

    arf%backscale_src  = 0.0d0
    arf%backscale_back = 0.0d0
    arf%regarea_src = 0.0d0
    arf%regarea_back = 0.0d0
    arf%regmodarea_src = 0.0d0
    arf%regmodarea_back = 0.0d0

    call fitswrap_init_hduinfo(arf%hduinfo)
    call fitswrap_init_key(arf%key_phafile,  "PHAFILE",  "Spectral file to which this ARF applies")
    call fitswrap_init_key(arf%key_telapse,  "TELAPSE",  "[s] Total observation duration" )
    call fitswrap_init_key(arf%key_ontime,   "ONTIME",   "[s] On-target exposure time (pre-deadtime corr)")
    call fitswrap_init_key(arf%key_livetime, "LIVETIME", "[s] On-target active-detector exposure time")
    call fitswrap_init_key(arf%key_exposure, "EXPOSURE", "[s] Fully corrected effective exposure time")

!    call fitswrap_init_key(arf%key_ninst, "NINST", "Number of instruments which contributed")
!    do i=1,MAX_INST
!!      write(str_buffer, '(A,I0)') "INSTRUM", i
!      call fitswrap_init_key(arf%pkey_instrum(i), "INSTRUM" // trim(i2s(i)), "Instruments which contributed")
!    end do

  end subroutine srctool_arf_empty_arf
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_arf_empty_arflist(arflist)
    type(arflist_type), intent(inout)   :: arflist
    integer(kind=int32) :: i
    do i=0,MAX_INST
      call srctool_arf_empty_arf(arflist%arf(i))
    end do
    arflist%num_arfs = 0
  end subroutine srctool_arf_empty_arflist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the arf from the current hdu of the fits file identified by unit
  subroutine srctool_arf_read_arf_from_chdu(unit, arf, ierr)
    character(len=*), parameter :: subroutinename='srctool_arf_read_arf_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(arf_type), intent(inout) :: arf
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
    character(len=STR_MAX) :: str_buffer

    integer(kind=int32) :: colnum_elo
    integer(kind=int32) :: colnum_ehi
    integer(kind=int32) :: colnum_resp
    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    real(kind=real32)  , parameter :: nullval_real32  = 0.0
    real(kind=real64)  , parameter :: nullval_real64  = 0.0d0

    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in ARF extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, arf%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if
    !- check the following keader keywords are present
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1
    !- HDUCLAS2
    ! DEBUG to get around bad cal input
!    if ( trim(arf%hduinfo%key_telescop%value) .ne. "eROSITA" ) then
!      call fitswrap_set_key(arf%hduinfo%key_telescop, FITSWRAP_SET_KEY_BLANK, "eROSITA", FITSWRAP_SET_KEY_BLANK)
!    endif

    !get the number of rows in the extension
    call FTGNRW(unit, arf%num_rows, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    !check that have at least some entries and then read in the column data
    if ( arf%num_rows .gt. 0 ) then
      write (str_buffer, '(A,I0,A,A,A,A,A)') &
          "Reading in: ", arf%num_rows, " rows of ARF info from extname='", &
          trim(arf%hduinfo%key_extname%value), "' instrument='", &
          trim(arf%hduinfo%key_instrume%value), "'"
      call wrschatty (subroutinename, str_buffer)
!      call fitswrap_proc_col(unit, "ENERG_LO", arf%num_rows, arf%pelo,  ierr)
!      call fitswrap_proc_col(unit, "ENERG_HI", arf%num_rows, arf%pehi,  ierr)
!      call fitswrap_proc_col(unit, "SPECRESP", arf%num_rows, arf%presp, ierr)
!
!      if (ierr .ne. ERROR_LEVEL_OK ) then
!        call wrserr (subroutinename, ierr, "Problem reading column(s) from HDU")
!        return
!      end if

      call fitswrap_prep_col(unit, "ENERG_LO", arf%num_rows, colnum_elo,  arf%pelo,  ierr)
      call fitswrap_prep_col(unit, "ENERG_HI", arf%num_rows, colnum_ehi,  arf%pehi,  ierr)
      call fitswrap_prep_col(unit, "SPECRESP", arf%num_rows, colnum_resp, arf%presp, ierr)
      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if

      ! get the chunksize
      call FTGRSZ(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      do row = 1, arf%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,arf%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVE(unit, colnum_elo,   row,1,nrows_to_read,nullval_real32,arf%pelo(row:row_end),  anyf,status)
        call FTGCVE(unit, colnum_ehi,   row,1,nrows_to_read,nullval_real32,arf%pehi(row:row_end),  anyf,status)
        call FTGCVD(unit, colnum_resp,  row,1,nrows_to_read,nullval_real64,arf%presp(row:row_end), anyf,status)

        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do

    end if

    arf%is_valid = .true.
  end subroutine srctool_arf_read_arf_from_chdu
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_arf_setup_dummy_arf(arf, ierr)
    character(len=*), parameter :: subroutinename='srctool_arf_setup_dummy_arf'
    type(arf_type), intent(inout)      :: arf
    integer(kind=int32), intent(out)  :: ierr
    integer(kind=int32) :: i
    integer(kind=int32), parameter :: dummy_arf_num_rows = 1000
    real(kind=real64), parameter   :: dummy_arf_elo = 0.1d0  ! keV
    real(kind=real64), parameter   :: dummy_arf_ehi = 10.0d0 ! keV
    real(kind=real64) :: multiplier
    real(kind=real64) :: ee
    integer(kind=int32) :: status
    status = 0
    ierr = ERROR_LEVEL_OK

    call srctool_arf_empty_arf(arf)
    call fitswrap_set_key(arf%hduinfo%key_extname,  "SPECRESP" )
    call fitswrap_set_key(arf%hduinfo%key_hduclass, "OGIP"     )
    call fitswrap_set_key(arf%hduinfo%key_hduclas1, "RESPONSE" )
    call fitswrap_set_key(arf%hduinfo%key_hduclas2, "SPECRESP" )
    call fitswrap_set_key(arf%hduinfo%key_telescop, "DUMMY"    )
    call fitswrap_set_key(arf%hduinfo%key_instrume, "DUMMY"    )
    call fitswrap_set_key(arf%hduinfo%key_filter,   "INDEF"    )
    call fitswrap_set_key(arf%hduinfo%key_hduvers,  "1.1.0"    )
    call fitswrap_set_key(arf%hduinfo%key_object,   "NONE"     )
!    call fitswrap_set_key(arf%hduinfo%key_creator,  FITSWRAP_SET_KEY_BLANK, CREATOR_STRING, FITSWRAP_SET_KEY_BLANK)

    arf%is_valid = .true.
    arf%to_consider = .true.
    arf%num_rows = dummy_arf_num_rows
    allocate(arf%pelo(arf%num_rows),&
        arf%pehi(arf%num_rows),&
        arf%presp(arf%num_rows),&
        arf%pcorrcomb(arf%num_rows),&
        arf%pcorrpsf(arf%num_rows),&
        arf%pcorrvign(arf%num_rows),&
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return


    multiplier = (dummy_arf_ehi/dummy_arf_elo)**(1.0d0/real(dummy_arf_num_rows,real64))
!    arf%pelo(1) = dummy_arf_elo
!    arf%pehi(1) = arf%pelo(1) * multiplier
!    do i=2,arf%num_rows
!      arf%pelo(i) = arf%pehi(i-1)
!      arf%pehi(i) = arf%pelo(i) * multiplier
!    end do

    ee = dummy_arf_elo
    do i=1,arf%num_rows
      arf%pelo(i) = REAL(ee,real32)
      ee = ee * multiplier
      arf%pehi(i) = REAL(ee,real32)
    end do


    arf%presp(:)  = 1.0d0

  end subroutine srctool_arf_setup_dummy_arf
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for an output arf file
  subroutine srctool_arf_setup_arf(arf, arf_template, inst, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_arf_setup_arf'
    type(arf_type),      intent(inout) :: arf
    type(arf_type),      intent(in)    :: arf_template
    type(inst_type),     intent(in),target :: inst
    type(src_type),      intent(in)    :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)                :: status
    character(len=STR_MAX)             :: str_buffer
    character(len=STR_MAX)             :: str_buffer2

    ierr = ERROR_LEVEL_OK
    status = 0

    arf%to_consider = .true.

    ! base the arf structure on the arf template
    arf%num_rows   = arf_template%num_rows
    allocate(arf%pelo(arf%num_rows),&
        arf%pehi(arf%num_rows),&
        arf%presp(arf%num_rows),&
        arf%pcorrcomb(arf%num_rows),&
        arf%pcorrpsf(arf%num_rows),&
        arf%pcorrvign(arf%num_rows),&
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! initialise the values in the arrays
    arf%pelo(:)            = 0.0
    arf%pehi(:)            = 0.0
    arf%presp(:)           = 0.0d0
    arf%pcorrcomb(:)       = 0.0d0
    arf%pcorrpsf(:)        = 0.0d0
    arf%pcorrvign(:)       = 0.0d0

    ! copy the energy bins over from the arf template
    arf%pelo(:) = arf_template%pelo(:)
    arf%pehi(:) = arf_template%pehi(:)

    ! now copy the hduinfo
    arf%hduinfo         = arf_template%hduinfo

    ! set the output filename + DSTYPE
    call srctool_misc_set_filename(arf%hduinfo%filename, inst, src, SRCTOOL_FILETYPE_CODE_ARF )
    call fitswrap_set_key(arf%hduinfo%key_dstype, ESASS_DSTYPE_ARF )
    call fitswrap_set_key(arf%hduinfo%key_dscode, &
        srctool_misc_dscode(inst%index,ESASS_FILTER_STR(inst%filter_code),0) )
    call fitswrap_set_key(arf%hduinfo%key_filter, ESASS_FILTER_STR(inst%filter_code))

!    ! now set some appropriate values for an OGIP compliant spectrum file
    call fitswrap_set_key(arf%hduinfo%key_object,  src%name      )
    call fitswrap_set_key(arf%hduinfo%key_ra_obj,  src%coord%long) ! TODO always convert source coords to J2000
    call fitswrap_set_key(arf%hduinfo%key_dec_obj, src%coord%lat ) ! TODO
    call fitswrap_set_key(arf%hduinfo%key_radecsys,COORDSYS_STR(src%coord%coordsys_code)) ! TODO
    call fitswrap_set_key(arf%hduinfo%key_equinox, 2000.0d0)

    call fitswrap_set_key(arf%hduinfo%key_date_end, inst%evtlistlist%hduinfo%key_date_end%value)
    call fitswrap_set_key(arf%hduinfo%key_date_obs, inst%evtlistlist%hduinfo%key_date_obs%value)
    call fitswrap_set_key(arf%hduinfo%key_tstart,   inst%evtlistlist%hduinfo%key_tstart%value)
    call fitswrap_set_key(arf%hduinfo%key_tstop,    inst%evtlistlist%hduinfo%key_tstop%value)

    call srctool_misc_set_filename(str_buffer, inst, src, SRCTOOL_FILETYPE_CODE_SRC_SPEC )
    call pathtools_extract_tail(trim(str_buffer), str_buffer2)
    call fitswrap_set_key(arf%key_phafile , trim(str_buffer2))

    call fitswrap_set_key(arf%key_telapse,  0.0d0)
    call fitswrap_set_key(arf%key_ontime,   0.0d0)
    call fitswrap_set_key(arf%key_livetime, 0.0d0)
    call fitswrap_set_key(arf%key_exposure, 0.0d0)



  end subroutine srctool_arf_setup_arf
  !-------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for an arflist
  subroutine srctool_arf_setup_arflist(arflist, instlist, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_arf_setup_arflist'
    type(arflist_type),    intent(inout) :: arflist
    type(instlist_type),  intent(in), target :: instlist
    type(src_type),       intent(in)    :: src
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32)                 :: i
    type(arf_type),    pointer          :: prefarf
!    character(len=STR_MAX)              :: str_buffer
    integer(kind=int32)                 :: nok


    ierr = ERROR_LEVEL_OK
    do i=0,MAX_INST
      if ( (i .eq. 0) .or. instlist%inst(i)%to_consider ) then

        call srctool_misc_select_refarf_pointer(instlist%inst(i),prefarf,ierr)
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Could not find suitable reference ARF")
          return
        end if

        ! TODO - this will not work correctly for TM0 if no TM0 refarf has been supplied
        call srctool_arf_setup_arf(arflist%arf(i), prefarf, instlist%inst(i), src, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error setting up ARF data structure")
          return
        end if
      end if
    end do

!    ! now set up the hduinfo values in the merged instrument arf
!    ! derive them from the first valid arf - with INSTRUME and filename adapted accordingly
!    do i=1,MAX_INST
!      if ( instlist%inst(i)%to_consider ) then
!        arflist%arf(0)%hduinfo = arflist%arf(i)%hduinfo
!        call fitswrap_set_key(arflist%arf(0)%hduinfo%key_instrume, &
!            "INSTRUME",trim(instlist%inst(0)%name),"Instrument name")
!        ! sort out the output filename
!        call srctool_misc_set_filename(arflist%arf(0)%hduinfo%filename, instlist%inst(0), &
!            src, SRCTOOL_FILETYPE_CODE_ARF )
!        exit
!      end if
!    end do


    ! copy the basic keywords from the first valid instrument for the merged instrument files
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        arflist%arf(0)%hduinfo  = arflist%arf(i)%hduinfo
        exit
      end if
    end do

    ! now set up the hduinfo values in the "MERGED" arf files
    ! derive them from the first valid arf file - with INSTRUME and filename adapted accordingly
    call fitswrap_set_key(arflist%arf(0)%hduinfo%key_instrume, INSTRUME_IS_MERGED)
    call srctool_misc_set_filename(arflist%arf(0)%hduinfo%filename, &
          instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_ARF )
    call fitswrap_set_key(arflist%arf(0)%hduinfo%key_dscode, &
        srctool_misc_dscode(instlist%inst(0)%index,ESASS_FILTER_STR(instlist%inst(0)%filter_code),0) )

    ! now deal with the NINST and INSTRUMn keywords
    nok = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        nok = nok + 1
        call fitswrap_set_key(arflist%arf(0)%hduinfo%pkey_instrum(nok), trim(instlist%inst(i)%name))
      end if
    end do
    call fitswrap_set_key(arflist%arf(0)%hduinfo%key_ninst, nok)



  end subroutine srctool_arf_setup_arflist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! coadd the arf data across all valid telescopes
  subroutine srctool_arf_coadd_arflist(arflist, instlist )
    character(len=*), parameter :: subroutinename='srctool_arf_coadd_arflist'
    type(instlist_type),  intent(in) :: instlist
    type(arflist_type),   intent(inout), target :: arflist
    type(arf_type),   pointer :: arf0
    integer(kind=int32) :: i
    integer(kind=int32) :: nok
    real(kind=real64) :: arffrac
    integer(kind=int32) :: nconsidered

    call wrschatty(subroutinename, "Coadding ARF for merged instrument")
    !-----------------------------------------------
    ! average the exposure times from the individual instruments
    nok = 0
    arf0 => arflist%arf(0)
    arf0%key_ontime%value   = 0.0d0
    arf0%key_livetime%value = 0.0d0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        if ( arflist%arf(i)%to_consider ) then
          nok = nok + 1
          arf0%key_ontime%value = &
              arf0%key_ontime%value + arflist%arf(i)%key_ontime%value
          arf0%key_livetime%value = &
              arf0%key_livetime%value + arflist%arf(i)%key_livetime%value
        end if
      end if
    end do
    if ( nok .gt. 0 ) then
      arf0%key_ontime%value   = arf0%key_ontime%value / REAL(nok,real64)
      arf0%key_livetime%value = arf0%key_livetime%value / REAL(nok,real64)
    end if
    arf0%key_exposure%value = arf0%key_livetime%value
    arf0%key_telapse%value  = arf0%key_ontime%value ! TODO do this properly
    !-----------------------------------------------

    arf0%presp(:) = 0d0
    arf0%pcorrcomb(:) = 0d0
    arf0%pcorrpsf(:) = 0d0
    arf0%pcorrvign(:) = 0d0

    !-----------------------------------------------
    ! weight the arfs by effective exposure livetime in each telescope
    if ( arf0%key_livetime%value .gt. 0.0d0 ) then
      nconsidered = 0
      do i=1,MAX_INST
        if (instlist%inst(i)%to_consider ) then
          if ( arflist%arf(i)%to_consider ) then
            !          arf0%presp(:) = arf0%presp(:) + arflist%arf(i)%presp(:)
            arffrac = arflist%arf(i)%key_livetime%value / arf0%key_livetime%value
            arf0%presp(:) = arf0%presp(:) + arflist%arf(i)%presp(:) * arffrac
            arf0%pcorrcomb(:) = arf0%pcorrcomb(:) + arflist%arf(i)%pcorrcomb(:) * arffrac
            arf0%pcorrpsf(:) = arf0%pcorrpsf(:) + arflist%arf(i)%pcorrpsf(:) * arffrac
            arf0%pcorrvign(:) = arf0%pcorrvign(:) + arflist%arf(i)%pcorrvign(:) * arffrac
            nconsidered = nconsidered + 1
          end if
        end if
      end do

      ! make an average correction
      if( nconsidered > 0 ) then
         arf0%pcorrcomb(:) = arf0%pcorrcomb(:) / nconsidered
         arf0%pcorrpsf(:) = arf0%pcorrpsf(:) / nconsidered
         arf0%pcorrvign(:) = arf0%pcorrvign(:) / nconsidered
      end if
    endif
    !-----------------------------------------------

    nullify(arf0)
  end subroutine srctool_arf_coadd_arflist
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! write arf fits keys info to the chdu
  subroutine srctool_arf_write_arf_keys(unit, arf, ierr)
    character(len=*), parameter :: subroutinename='srctool_arf_write_arf_keys'
    integer(kind=int32), intent(in)  :: unit
    type(arf_type),      intent(in)  :: arf
    integer(kind=int32), intent(out)   :: ierr
!    integer(kind=int32)    :: i
    ierr = ERROR_LEVEL_OK

    !-----------------------------------
    ! write the standard keywords to the binary table HDU:
    call fitswrap_write_hduinfo_quiet(unit, arf%hduinfo, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      return
    end if



!    call fitswrap_write_key (unit, arf%hduinfo%key_telescop, ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_instrume, ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_filter,   ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_hduclass, ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_hduclas1, ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_hduclas2, ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_hduvers,  ierr)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!     end if
!
!    ! now the optional keys
!    call fitswrap_write_key (unit, arf%hduinfo%key_creator,  ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_object,   ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_ra_obj,   ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_dec_obj,  ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_equinox,  ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_radecsys, ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_tstart,   ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_tstop,    ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_date_obs, ierr)
!    call fitswrap_write_key (unit, arf%hduinfo%key_date_end, ierr)

!    ! XFLTXXXX keywords
!    call fitswrap_write_key (unit, arf%hduinfo%key_xflt(i), ierr)


    ! now the ARF-specific keys
    call fitswrap_write_key (unit, arf%key_phafile, ierr)
    call fitswrap_write_key (unit, arf%key_telapse,  ierr)
    call fitswrap_write_key (unit, arf%key_ontime,   ierr)
    call fitswrap_write_key (unit, arf%key_livetime, ierr)
    call fitswrap_write_key (unit, arf%key_exposure, ierr)

!    if ( trim(arf%hduinfo%key_instrume%value) .eq. "MERGED" ) then
!      call fitswrap_write_key (unit, arf%key_ninst,    ierr)
!      do i=1,arf%key_ninst%value
!        call fitswrap_write_key (unit, arf%pkey_instrum(i), ierr)
!      end do
!    end if

    if (ierr .ne. ERROR_LEVEL_OK) then
      return
    end if
    !-----------------------------------
    ! TODO - add any more keywords?
    !-----------------------------------

  end subroutine srctool_arf_write_arf_keys
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !write an arf to file
  subroutine srctool_arf_write_arf(arf, gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_arf_write_arf'

    type(arf_type), intent(inout) :: arf
    type(gtilist_type), intent(in) :: gtilist
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32), parameter :: blank_bitpix = 16
    integer(kind=int32), parameter :: blank_naxis = 0
    integer(kind=int32), parameter :: blank_naxes(1) = (/ 0 /)
!    logical(kind=log32), parameter :: simple = .true.
!    logical,             parameter :: simple = .true.
    integer(kind=int32), parameter :: ncols=6
    character(len=*),    parameter :: ttype(ncols) = (/ "ENERG_LO", "ENERG_HI", "SPECRESP", "CORRCOMB", "CORRPSF ", "CORRVIGN" /)
    character(len=*),    parameter :: tunit(ncols) = (/ "keV "    , "keV "     ,"cm^2"    , "    "    , "    "    , "    "     /)
    character(len=*),    parameter :: tform(ncols) = (/ "E"       , "E"        ,"D"       , "E"       , "E"       , "E"        /)
    integer(kind=int32), parameter :: COLNUM_ENERG_LO  =  1
    integer(kind=int32), parameter :: COLNUM_ENERG_HI  =  2
    integer(kind=int32), parameter :: COLNUM_SPECRESP  =  3
    integer(kind=int32), parameter :: COLNUM_CORRCOMB  =  4
    integer(kind=int32), parameter :: COLNUM_CORRPSF   =  5
    integer(kind=int32), parameter :: COLNUM_CORRVIGN  =  6

!    character(len=*),    parameter :: extname = "SPECRESP"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int64), parameter :: frow=1, felem=1

!    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)    :: unit
    integer(kind=int32)    :: status

    ierr = ERROR_LEVEL_OK
    status = 0
    unit = 0


    if ( len_trim(arf%hduinfo%filename) .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Illegal filename for arf file: " // trim(arf%hduinfo%filename))
      return
    end if

    call wrschatty (subroutinename,  "Writing ARF file, filename: " //  trim(arf%hduinfo%filename))


    !-----------------------------------
    ! get a fortran unit
    call FTGIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------


    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT(unit,arf%hduinfo%filename,FITSIO_BLOCKSIZE, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

    ! append a new empty primary image hdu and write date keyword to the HDU
    call FTIIMG(unit,blank_bitpix,blank_naxis,blank_naxes, status)
    call FTPDAT(unit, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
!
!    !-----------------------------------
!    ! write the standard FITS header:
!    call FTPHPS(unit,bitpix,naxis,naxes,status)
!    if (status .ne. 0) then
!      ierr = ERROR_LEVEL_SEVERE
!      call fitswrap_write_fits_error(subroutinename, status)
!      return
!    end if
!    !-----------------------------------
!
!    !-----------------------------------
!    ! write the compulsory keywords to the primary HDU
!    ! write date keyword
!    call  FTPDAT(unit, status)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------

    !-----------------------------------
    ! append a new binary table HDU
    !FTIBIN(unit,nrows,tfields,ttype,tform,tunit,extname,varidat > status)
    call FTIBIN(unit,arf%num_rows,ncols,ttype,tform,tunit,trim(arf%hduinfo%key_extname%value),varidat,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

!    !-----------------------------------
!    ! append a new table HDU
!    call FTCRHD(unit, status)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------
!
!    !-----------------------------------
!    ! write header info for the binary table:
!    call FTPHBN(unit,arf%num_rows,ncols,ttype,tform,tunit,extname,varidat,status)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------


    ! update TSTART/TSTOP using GTI
    if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
       call fitswrap_set_key(arf%hduinfo%key_tstart, gtilist%pstart(1))
       call fitswrap_set_key(arf%hduinfo%key_tstop, gtilist%pstop(gtilist%num_rows))
    end if

    !-----------------------------------
    call srctool_arf_write_arf_keys(unit, arf, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing ARF fits header keys")
      return
    end if
    !-----------------------------------

    !-----------------------------------
    ! now write the column data to the file
    ! format: FTPCL[SLBIJKEDCM](unit,colnum,frow,felem,nelements,values, > status)
    ! TODO do this in chunks for better efficiency

    ! write ENERG_LO column
    call  FTPCLE(unit,COLNUM_ENERG_LO,frow,felem,arf%num_rows,arf%pelo,status)
    ! write ENERG_HI column
    call  FTPCLE(unit,COLNUM_ENERG_HI,frow,felem,arf%num_rows,arf%pehi,status)
    ! write SPECRESP column
    call  FTPCLD(unit,COLNUM_SPECRESP,frow,felem,arf%num_rows,arf%presp,status)
    ! write CORR column
    call  FTPCLD(unit,COLNUM_CORRCOMB,frow,felem,arf%num_rows,arf%pcorrcomb,status)
    ! write CORRPSF column
    call  FTPCLD(unit,COLNUM_CORRPSF,frow,felem,arf%num_rows,arf%pcorrpsf,status)
    ! write CORRVIGN column
    call  FTPCLD(unit,COLNUM_CORRVIGN,frow,felem,arf%num_rows,arf%pcorrvign,status)

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    !-----------------------------------

    !-----------------------------------
    ! now close the fits file
    call FTCLOS(unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

  end subroutine srctool_arf_write_arf
  !---------------------------------------------------------------------------------------------

 !-------------------------------------------------------------------------------
  !write the arflist to file(s)
  subroutine srctool_arf_write_arflist(arflist, instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_arf_write_arflist'

    type(arflist_type), intent(inout) :: arflist
    type(instlist_type),  intent(in) :: instlist
    integer(kind=int32), intent(out)  :: ierr
    integer(kind=int32) :: i
    ierr = ERROR_LEVEL_OK

    do i=0,MAX_INST
      if ((i .eq. 0 ) .or. instlist%inst(i)%to_consider ) then
        call srctool_arf_write_arf(arflist%arf(i), instlist%inst(i)%src_gtilist, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem writing arf file")
          return
        end if
      end if
    end do


  end subroutine srctool_arf_write_arflist
 !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of ARF handling -------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------











end module srctool_arf
