! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_att


  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-24
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of attitude routines
  !
  !-----------------------------------------------------------------------------

  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
!  use srctool_generic_fitswrap  ! generic fits file handling routines
!  use srctool_generic_regions   ! generic region handling routines
  use srctool_datatypes         ! module with some srctool-specific constants and datatypes
  use srctool_misc              ! module with some srctool-specific constants and datatypes

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_ATT_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_att
  private :: srctool_att_radec2rawxy_as
  private :: srctool_att_radec2rawxy_b
  private :: srctool_att_radec2rawxy_bs
  private :: srctool_att_radec2rawxy_c
  private :: srctool_att_radec2rawxy_cs
  private :: srctool_att_rawxy2radec_as

  public ::  srctool_att_radec2rawxy_a
  public ::  srctool_att_rawxy2radec_a
  public ::  srctool_att_radec2rawxy_a2
  public ::  srctool_att_radec2rawxy_a2_fast
  public ::  srctool_att_radec2rawxy_a2_faster
  public ::  srctool_att_radec2rawxy_as2_fast
  public ::  srctool_att_empty_attlist
  public ::  srctool_att_merge_attlists_for_inst
  public ::  srctool_att_get_current_att
  public ::  srctool_att_radec2rawxy_as2
  public ::  srctool_att_read_attlist_from_chdu
  public ::  srctool_att_test_harness
  public ::  srctool_att_calc_stats


  interface srctool_att_radec2rawxy
    module procedure srctool_att_radec2rawxy_a
    module procedure srctool_att_radec2rawxy_as
    module procedure srctool_att_radec2rawxy_a2
    module procedure srctool_att_radec2rawxy_as2
    module procedure srctool_att_radec2rawxy_b
    module procedure srctool_att_radec2rawxy_bs
    module procedure srctool_att_radec2rawxy_c
    module procedure srctool_att_radec2rawxy_cs
  end interface

  interface srctool_att_rawxy2radec
    module procedure srctool_att_rawxy2radec_a
    module procedure srctool_att_rawxy2radec_as
  end interface


contains



  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of ATTLIST handling -------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_att_empty_attlist(attlist)
    type(attlist_type), intent(inout)      :: attlist
    if ( allocated(attlist%ptime))       deallocate (attlist%ptime)
    if ( allocated(attlist%pra))         deallocate (attlist%pra)
    if ( allocated(attlist%pdec))        deallocate (attlist%pdec)
    if ( allocated(attlist%proll))       deallocate (attlist%proll)
    if ( allocated(attlist%pdist))       deallocate (attlist%pdist)
    if ( allocated(attlist%pcosdist))    deallocate (attlist%pcosdist)
    if ( allocated(attlist%psindist))    deallocate (attlist%psindist)
    if ( allocated(attlist%pbearing))    deallocate (attlist%pbearing)
    if ( allocated(attlist%pcosbearing)) deallocate (attlist%pcosbearing)
    if ( allocated(attlist%psinbearing)) deallocate (attlist%psinbearing)
    if ( allocated(attlist%pcosdec))     deallocate (attlist%pcosdec)
    if ( allocated(attlist%psindec))     deallocate (attlist%psindec)
    attlist%num_rows = 0
    attlist%is_valid = .false.

    attlist%mean_ra    = huge(0.0d0)
    attlist%mean_dec   = huge(0.0d0)
    attlist%mean_roll  = huge(0.0d0)
!    attlist%stdev_ra   = huge(0.0d0)
!    attlist%stdev_dec  = huge(0.0d0)
!    attlist%stdev_roll = huge(0.0d0)
    attlist%stdev      = huge(0.0d0)

    call fitswrap_init_hduinfo(attlist%hduinfo)
  end subroutine srctool_att_empty_attlist
  !-------------------------------------------------------------------------------

  ! THIS DOES NOT YET WORK!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!11

  ! project points onto a tangent plane centered on first point, and then average in x,y coords
  ! project tthis centroid back to ra,dec
  ! then repeat using the new centroid as the tangent point
  subroutine srctool_att_calc_stats(attlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_att_calc_stats'
    type(attlist_type), intent(inout)      :: attlist
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32) :: i, counter
    integer(kind=int32) :: status

    real(kind=real64) :: long0,lat0, long1, lat1,dist, temp_roll
    real(kind=real64), allocatable, dimension(:) :: x,y
    real(kind=real64) :: x_mean,y_mean
    real(kind=real64), parameter :: convergence_criterion = 0.1d0 / 3600.0d0  ! 0.1arcsec in degrees
    real(kind=real64), parameter  :: oa_x = 0.0d0
    real(kind=real64), parameter  :: oa_y = 0.0d0
    real(kind=real64), parameter  :: pixscale_x = 1.0d0
    real(kind=real64), parameter  :: pixscale_y = 1.0d0
    real(kind=real64), parameter  :: inv_pixscale_x = 1.0d0/pixscale_x
    real(kind=real64), parameter  :: inv_pixscale_y = 1.0d0/pixscale_y
    real(kind=real64), parameter  :: roll0 = 0.0d0
    ierr = ERROR_LEVEL_OK

    long1 = attlist%pra(1)
    lat1  = attlist%pdec(1)

    allocate(x(attlist%num_rows), y(attlist%num_rows), STAT=status)

    if ( check_allocate_status (subroutinename, status, ierr)) return
    counter = 0
    dist = huge(dist)
    do while ( (dist .gt. convergence_criterion) .and. (counter .lt. 10) )
      long0 = long1
      lat0  = lat1
      call longlats2xys(oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, long0, lat0, roll0, &
          attlist%pra(1:attlist%num_rows), attlist%pdec(1:attlist%num_rows), x(:), y(:))

      x_mean = sum(x(:))/ MAX(1.0d0,REAL(attlist%num_rows,kind=real64))
      y_mean = sum(y(:))/ MAX(1.0d0,REAL(attlist%num_rows,kind=real64))
      call xy2longlat(oa_x, oa_y, pixscale_x, pixscale_y, long0, lat0, roll0, &
          x_mean, y_mean, long1, lat1)
      dist = calc_ang_sep(long0,lat0,long1,lat1)

      counter = counter + 1
!      write(*,'(A,x,I4,7(x,F14.7))'), "ATTLIST_TEST1",  counter, long0, lat0, long1, lat1, x_mean, y_mean, dist
    end do
    attlist%mean_ra = long1
    attlist%mean_dec = lat1
    attlist%mean_roll = -1.0d0

    ! range of roll values should be within range attlist%proll(1)-180.0<=temp_roll<attlist%proll(1)+180.0
    do i=1,attlist%num_rows
      if ( attlist%proll(i) >= attlist%proll(i) + 180.0d0 ) then
        temp_roll = attlist%proll(i) - 360.0d0
      else if  ( attlist%proll(i) < attlist%proll(i) - 180.0d0 ) then
        temp_roll = attlist%proll(i) + 360.0d0
      else
        temp_roll = attlist%proll(i)
      endif
      attlist%mean_roll = attlist%mean_roll + temp_roll
    end do
    attlist%mean_roll = attlist%mean_roll / MAX(1.0d0,REAL(attlist%num_rows,kind=real64))
    if ( attlist%mean_roll .lt.   0.0d0 ) attlist%mean_roll = attlist%mean_roll + 360.0d0
    if ( attlist%mean_roll .ge. 360.0d0 ) attlist%mean_roll = attlist%mean_roll - 360.0d0

    if (allocated(x)) deallocate(x)
    if (allocated(y)) deallocate(y)

    ! compute the stdev of the great circle distances to the mean position
    attlist%stdev = 0.0d0
    do i=1,attlist%num_rows
      attlist%stdev = attlist%stdev + (calc_ang_sep(attlist%mean_ra,attlist%mean_dec,attlist%pra(i),attlist%pdec(i)))**2
    end do
    attlist%stdev = sqrt(attlist%stdev / MAX(1.0d0,REAL(attlist%num_rows,kind=real64)))

!    write(*,'(A,4(x,F14.7))'), "ATTLIST_TEST2",  attlist%mean_ra, attlist%mean_dec, attlist%mean_roll, attlist%stdev

  end subroutine srctool_att_calc_stats

  !-------------------------------------------------------------------------------
  !read the attlist from the current hdu of the fits file identified by unit
  subroutine srctool_att_read_attlist_from_chdu(unit, attlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_att_read_attlist_from_chdu'

    integer(kind=int32), intent(in)    :: unit
    type(attlist_type), intent(inout)  :: attlist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32) :: status
    character(len=STR_MAX) :: str_buffer

    integer(kind=int32) :: colnum_time
    integer(kind=int32) :: colnum_ra
    integer(kind=int32) :: colnum_dec
    integer(kind=int32) :: colnum_roll
    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    real(kind=real64)  , parameter :: nullval_real64  = 0.0d0


    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in CORRATT extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, attlist%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    !- check the following keader keywords are present
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1       'TEMPORALDATA'
    !- HDUCLAS2       'ASPECT'


    !get the number of rows in the extension
    call FTGNRW(unit, attlist%num_rows, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    !check that have at least some entries and then read in the column data
    if ( attlist%num_rows .gt. 0 ) then
      write (str_buffer, '(A,I0,A,A,A,A,A)') &
          "Reading in: ", attlist%num_rows, " rows of attitude info from extname='", &
          trim(attlist%hduinfo%key_extname%value), "' instrument='", &
          trim(attlist%hduinfo%key_instrume%value), "'"
      call wrschatty (subroutinename, str_buffer)
!      call fitswrap_proc_col(unit, "TIME", attlist%num_rows, attlist%ptime, ierr)
!      call fitswrap_proc_col(unit, "RA"  , attlist%num_rows, attlist%pra,   ierr)
!      call fitswrap_proc_col(unit, "DEC" , attlist%num_rows, attlist%pdec,  ierr)
!      call fitswrap_proc_col(unit, "ROLL", attlist%num_rows, attlist%proll, ierr)
!
!      if (ierr .ne. ERROR_LEVEL_OK ) then
!        call wrserr (subroutinename, ierr, "Problem reading column(s) from HDU")
!        return
!      end if

      call fitswrap_prep_col(unit, "TIME", attlist%num_rows, colnum_time, attlist%ptime, ierr)
      call fitswrap_prep_col(unit, "RA"  , attlist%num_rows, colnum_ra,   attlist%pra,   ierr)
      call fitswrap_prep_col(unit, "DEC" , attlist%num_rows, colnum_dec,  attlist%pdec,  ierr)
      call fitswrap_prep_col(unit, "ROLL", attlist%num_rows, colnum_roll, attlist%proll, ierr)
      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if
      ! get the chunksize
      call FTGRSZ(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      do row = 1, attlist%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,attlist%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVD(unit, colnum_time,   row,1,nrows_to_read,nullval_real64,attlist%ptime(row:row_end), anyf,status)
        call FTGCVD(unit, colnum_ra,     row,1,nrows_to_read,nullval_real64,attlist%pra(row:row_end),   anyf,status)
        call FTGCVD(unit, colnum_dec,    row,1,nrows_to_read,nullval_real64,attlist%pdec(row:row_end),  anyf,status)
        call FTGCVD(unit, colnum_roll,   row,1,nrows_to_read,nullval_real64,attlist%proll(row:row_end), anyf,status)

        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do

    end if
    attlist%is_valid = .true.

    call  srctool_att_calc_stats(attlist, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem calculating ATTLIST stats")
      return
    endif
  end subroutine srctool_att_read_attlist_from_chdu
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! merge all of the attlists for a given instrument
  subroutine srctool_att_merge_attlists_for_inst(inst, ierr)
    character(len=*), parameter :: subroutinename='srctool_att_merge_attlists_for_inst'
    type(inst_type), intent(inout)  :: inst
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)   :: i, j1,j2,n
    integer(kind=int32) :: status
 !   character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK
    status = 0

    ! copy the header from the first evtlist
    if ( inst%evtlistlist%num_evtlists .gt. 0 ) then
      inst%attlist%hduinfo = inst%evtlistlist%evtlist(1)%attlist%hduinfo
      ! TODO update the correct values of the TSTART, TSTOP, DATE-OBS, DATE-END etc keywords
    else
      ! exit quietly
      return
    end if

    inst%attlist%num_rows = 0
    do i=1,inst%evtlistlist%num_evtlists
      if ( inst%evtlistlist%evtlist(i)%attlist%is_valid ) then
        inst%attlist%num_rows = inst%attlist%num_rows + inst%evtlistlist%evtlist(i)%attlist%num_rows
      end if
    end do
    if ( inst%attlist%num_rows .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not find any valid entries in attlist")
      return
    end if

    allocate(inst%attlist%ptime(inst%attlist%num_rows), &
        inst%attlist%pra(inst%attlist%num_rows), &
        inst%attlist%pdec(inst%attlist%num_rows), &
        inst%attlist%proll(inst%attlist%num_rows), &
        inst%attlist%pdist(inst%attlist%num_rows), &
        inst%attlist%pcosdist(inst%attlist%num_rows), &
        inst%attlist%psindist(inst%attlist%num_rows), &
        inst%attlist%pbearing(inst%attlist%num_rows), &
        inst%attlist%pcosbearing(inst%attlist%num_rows), &
        inst%attlist%psinbearing(inst%attlist%num_rows), &
        inst%attlist%pcosdec(inst%attlist%num_rows), &
        inst%attlist%psindec(inst%attlist%num_rows), &
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

!    write (str_buffer, '(A,5(1x,A8))'), "TEST", "num_rows", "i", "n", "j1", "j2"
!    call wrschatty (subroutinename, str_buffer)
    j1 = 1
    do i=1,inst%evtlistlist%num_evtlists
      if ( inst%evtlistlist%evtlist(i)%attlist%is_valid ) then
        n = inst%evtlistlist%evtlist(i)%attlist%num_rows
        j2 = j1 + n - 1
!        write (str_buffer, '(A,5(1x,I8))'), "TEST", inst%attlist%num_rows, i, n, j1, j2
!        call wrschatty (subroutinename, str_buffer)

        inst%attlist%ptime(j1:j2) = inst%evtlistlist%evtlist(i)%attlist%ptime(1:n)
        inst%attlist%pra(j1:j2)   = inst%evtlistlist%evtlist(i)%attlist%pra(1:n)
        inst%attlist%pdec(j1:j2)  = inst%evtlistlist%evtlist(i)%attlist%pdec(1:n)
        inst%attlist%proll(j1:j2) = inst%evtlistlist%evtlist(i)%attlist%proll(1:n)
        j1 = j2+1
      end if
    end do
    inst%attlist%is_valid = .true.

    ! compute the distance travelled per step
    inst%attlist%pdist(1) = 0.0d0
    inst%attlist%pbearing(1) = 0.0d0
    do i=2,inst%attlist%num_rows
      inst%attlist%pdist(i) = calc_ang_sep(inst%attlist%pra(i-1),inst%attlist%pdec(i-1),&
          inst%attlist%pra(i),inst%attlist%pdec(i))
      inst%attlist%pbearing(i) = calc_bearing(inst%attlist%pra(i-1),inst%attlist%pdec(i-1),&
          inst%attlist%pra(i),inst%attlist%pdec(i))
    end do
    inst%attlist%pcosdist(:)    = cos(DEG2RAD*inst%attlist%pdist(:))
    inst%attlist%psindist(:)    = sin(DEG2RAD*inst%attlist%pdist(:))
    inst%attlist%pcosbearing(:) = cos(DEG2RAD*inst%attlist%pbearing(:))
    inst%attlist%psinbearing(:) = sin(DEG2RAD*inst%attlist%pbearing(:))
    inst%attlist%pcosdec(:)     = cos(DEG2RAD*inst%attlist%pdec(:))
    inst%attlist%psindec(:)     = sin(DEG2RAD*inst%attlist%pdec(:))


    call  srctool_att_calc_stats(inst%attlist, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem calculating ATTLIST stats")
      return
    endif

  end subroutine srctool_att_merge_attlists_for_inst
  !---------------------------------------------------------------------------------------------

!  recursive subroutine srctool_att_get_current_att_withhint(attlist, time, ra, dec, roll, hint, ierr)
!    character(len=*), parameter :: subroutinename='srctool_att_get_current_att'
!    type(attlist_type), intent(in)     :: attlist
!    real(kind=real64),  intent(in)     :: time
!    real(kind=real64),  intent(out)    :: ra, dec, roll
!    integer(kind=int32), intent(inout) :: hint
!    integer(kind=int32), intent(out)   :: ierr
!    character(len=STR_MAX) :: str_buffer
!    integer(kind=int32)        :: i
!    integer(kind=int32)        :: status
!
!    ! if just a constant attitude:
!    if ( attlist%num_rows .eq. 1 ) then
!      ra   = attlist%pra(1)
!      dec  = attlist%pdec(1)
!      roll = attlist%proll(1)
!      hint = 1
!      return
!    endif
!
!    if ( (hint .ge. 1) .and. (hint .lt. attlist%num_rows) ) then
!      if ( time .lt. attlist%ptime(hint+1) ) then
!
!      roll = interpolate_periodic_v(attlist%ptime(i), attlist%proll(i), &
!          attlist%ptime(i+1), attlist%proll(i+1), 360.0d0, time)
!
!      call interpolate_long_lat_fast ( &
!          attlist%ptime(i),   attlist%pra(i),   attlist%pdec(i),   &
!          attlist%pcosdec(i), attlist%psindec(i),  &
!          attlist%ptime(i+1), attlist%pra(i+1), attlist%pdec(i+1), &
!          attlist%pdist(i+1), attlist%pcosbearing(i+1), attlist%psinbearing(i+1),&
!          time, ra, dec, ierr )
!
!      endif
!    endif
!
!    ! if we get to here then just revert to the old way
!    call srctool_att_get_current_att(attlist, time, ra, dec, roll, ierr, hint)
!    return
!
!
!  end subroutine srctool_att_get_current_att_withhint


  !-------------------------------------------------------------------------------
  ! get the current attitude by interpolating the attitude list entries
  ! assume that attlist is already sorted in time
  recursive subroutine srctool_att_get_current_att(attlist, time, ra, dec, roll, ierr)

    character(len=*), parameter :: subroutinename='srctool_att_get_current_att'

    type(attlist_type),  intent(in)   :: attlist
    real(kind=real64),   intent(in)   :: time
    real(kind=real64),   intent(out)  :: ra, dec, roll
    integer(kind=int32), intent(out)  :: ierr

    character(len=STR_MAX)     :: str_buffer
    integer(kind=int32)        :: i
    integer(kind=int32)        :: status
    real(kind=real64)          :: deltat, tbin

    status = 0
    ierr = ERROR_LEVEL_OK

    ra   = huge(ra)
    dec  = huge(dec)
    roll = 0.0d0

    if ( attlist%num_rows == 0 ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Empty attitude list")
       return
    end if

    if ( attlist%num_rows == 1 ) then
       ! if just a constant attitude:
       ra   = attlist%pra(1)
       dec  = attlist%pdec(1)
       roll = attlist%proll(1)
       return
    end if

    if ( (time < attlist%ptime(1)) .or. &
         (time > attlist%ptime(attlist%num_rows)) ) then

       ! we should interpolate outside time range
       if ( time < attlist%ptime(1) ) then
          deltat = attlist%ptime(1)-time  ! difference from time bin
          i = 1
       else
          deltat = time-attlist%ptime(attlist%num_rows)
          i = attlist%num_rows-1
       end if

       tbin = attlist%ptime(i+1) - attlist%ptime(i)  ! bin width
       if( deltat > 0.5d0*tbin ) then
          ! only warn if we interpolate outside time bin
          write (str_buffer, '(A,F14.3,A,F14.3,A,F14.3,A)') &
               "Interpolating attitude time (", time, &
               ") outside valid range: [", attlist%ptime(1), &
               ":", attlist%ptime(attlist%num_rows), "]"
          call wrserr(subroutinename, ERROR_LEVEL_WARNING, str_buffer)
       end if

    else
       ! time within attitude range
       ! do a binary search here to get the two attitudes which bracket the desired time
       call binary_search_real64(attlist%ptime, time, i, status)
       if (status /= 0) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Could not interpolate attitude list")
          return
       end if
    end if

    roll = interpolate_periodic_v(attlist%ptime(i), attlist%proll(i), &
         attlist%ptime(i+1), attlist%proll(i+1), 360.0d0, time)

    call interpolate_long_lat_fast ( &
         attlist%ptime(i),   attlist%pra(i),   attlist%pdec(i),   &
         attlist%pcosdec(i), attlist%psindec(i),  &
         attlist%ptime(i+1), attlist%pra(i+1), attlist%pdec(i+1), &
         attlist%pdist(i+1), attlist%pcosbearing(i+1), attlist%psinbearing(i+1), &
         time, ra, dec, ierr )

  end subroutine srctool_att_get_current_att

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of ATTLIST handling -------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------- START of RADEC2RAWXY handling ----------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! on_chip_flag is set to true if position falls within chip bounds 0.5<=rawx|rawy<NPIX+0.5
  recursive subroutine srctool_att_radec2rawxy_a (instpar, ra0, dec0, roll, ra, dec, rawx, rawy, on_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_a'
    type(instpar_type), intent(in)  :: instpar
    real(kind=real64),  intent(in)   :: ra0, dec0, roll, ra, dec
    real(kind=real64),  intent(out)  :: rawx, rawy
    logical,            intent(out)  :: on_chip_flag
    integer(kind=int32) :: rawx_int, rawy_int

    call longlat2xy (instpar%oa_rawx, instpar%oa_rawy, instpar%inv_pixscale_x, instpar%inv_pixscale_y, &
        ra0, dec0, roll, ra, dec, rawx, rawy )

    rawx_int = int(rawx+0.5d0)
    rawy_int = int(rawy+0.5d0)
    if ((rawx_int .ge. 1) .and. (rawx_int .le. instpar%naxis1) .and. &
        (rawy_int .ge. 1) .and. (rawy_int .le. instpar%naxis2))   then
      on_chip_flag = .true.
    else
      on_chip_flag = .false.
    end if
  end subroutine srctool_att_radec2rawxy_a
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! on_chip_flag is set to true if position falls within chip bounds 0.5<=rawx|rawy<NPIX+0.5
  ! this performs the calculation for an array of input ra,dec pairs
  recursive subroutine srctool_att_radec2rawxy_as (instpar, ra0, dec0, roll, pra, pdec, prawx, prawy, pon_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_as'
    type(instpar_type), intent(in)  :: instpar
    real(kind=real64),  intent(in)   :: ra0, dec0, roll
    real(kind=real64),  intent(in),  dimension(:)  :: pra, pdec
    real(kind=real64),  intent(out), dimension(:)  :: prawx, prawy
    logical,            intent(out), dimension(:)  :: pon_chip_flag
    integer(kind=int32) :: rawx_int, rawy_int
    integer(kind=int32)  :: i,dim

    call longlats2xys (instpar%oa_rawx, instpar%oa_rawy, instpar%inv_pixscale_x, instpar%inv_pixscale_y, &
        ra0, dec0, roll, pra, pdec, prawx, prawy )

    dim = min(size(pra),size(pdec),size(prawx),size(prawy),size(pon_chip_flag))
    do i=1,dim
      rawx_int = int(prawx(i)+0.5d0)
      rawy_int = int(prawy(i)+0.5d0)
      if ((rawx_int .ge. 1) .and. (rawx_int .le. instpar%naxis1) .and. &
          (rawy_int .ge. 1) .and. (rawy_int .le. instpar%naxis2))   then
        pon_chip_flag(i) = .true.
      else
        pon_chip_flag(i) = .false.
      end if
    end do
  end subroutine srctool_att_radec2rawxy_as
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! on_chip_flag is set to true if position falls within chip bounds 0.5<=rawx|rawy<NPIX+0.5
  ! this performs the calculation for a single input ra,dec pair
  ! also return the integer version of the rawx,rawy coords
  recursive subroutine srctool_att_radec2rawxy_a2 (instpar, ra0, dec0, roll, ra, dec, rawx, rawy, &
      rawx_int, rawy_int, on_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_a2'
    type(instpar_type), intent(in)   :: instpar
    real(kind=real64),  intent(in)   :: ra0, dec0, roll
    real(kind=real64),  intent(in)   :: ra, dec
    real(kind=real64),  intent(out)  :: rawx, rawy
    integer(kind=int32), intent(out) :: rawx_int, rawy_int
    logical,            intent(out)  :: on_chip_flag

    call longlat2xy (instpar%oa_rawx, instpar%oa_rawy, instpar%inv_pixscale_x, instpar%inv_pixscale_y, &
        ra0, dec0, roll, ra, dec, rawx, rawy )

    rawx_int = INT(rawx+0.5d0)
    rawy_int = INT(rawy+0.5d0)
    on_chip_flag = .false.

    if ( rawx_int .ge. 1 ) then
      if ( rawx_int .le. instpar%naxis1 ) then
        if ( rawy_int .ge. 1 ) then
          if ( rawy_int .le. instpar%naxis2)  then
            on_chip_flag = .true.
          end if
        end if
      end if
    end if
  end subroutine srctool_att_radec2rawxy_a2
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! on_chip_flag is set to true if position falls within chip bounds 0.5<=rawx|rawy<NPIX+0.5
  ! this performs the calculation for a single input ra,dec pair
  ! also return the integer version of the rawx,rawy coords
  recursive subroutine srctool_att_radec2rawxy_a2_fast (instpar, ra0, dec0, roll, ra, cosdec, sindec, rawx, rawy, &
      rawx_int, rawy_int, on_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_a2_fast'
    type(instpar_type), intent(in)   :: instpar
    real(kind=real64),  intent(in)   :: ra0, dec0, roll
    real(kind=real64),  intent(in)   :: ra, sindec, cosdec
    real(kind=real64),  intent(out)  :: rawx, rawy
    integer(kind=int32),intent(out)  :: rawx_int, rawy_int
    logical,            intent(out)  :: on_chip_flag

    call longlat2xy_fast (instpar%oa_rawx, instpar%oa_rawy, instpar%inv_pixscale_x, instpar%inv_pixscale_y, &
        ra0, dec0, roll, ra, cosdec, sindec, rawx, rawy )

    rawx_int = INT(rawx+0.5d0)
    rawy_int = INT(rawy+0.5d0)
    on_chip_flag = .false.

    if ( rawx_int .ge. 1 ) then
      if ( rawx_int .le. instpar%naxis1 ) then
        if ( rawy_int .ge. 1 ) then
          if ( rawy_int .le. instpar%naxis2)  then
            on_chip_flag = .true.
          end if
        end if
      end if
    end if
  end subroutine srctool_att_radec2rawxy_a2_fast
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! on_chip_flag is set to true if position falls within chip bounds 0.5<=rawx|rawy<NPIX+0.5
  ! this performs the calculation for a single input ra,dec pair
  ! also return the integer version of the rawx,rawy coords
  recursive subroutine srctool_att_radec2rawxy_a2_faster (&
                   instpar, ra0, cosdec0, sindec0, cosroll, sinroll, &
                   ra, cosdec, sindec, rawx, rawy, &
                   rawx_int, rawy_int, on_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_a2_fast'
    type(instpar_type), intent(in)   :: instpar
    real(kind=real64),  intent(in)   :: ra0, cosdec0, sindec0, cosroll, sinroll
    real(kind=real64),  intent(in)   :: ra, sindec, cosdec
    real(kind=real64),  intent(out)  :: rawx, rawy
    integer(kind=int32),intent(out)  :: rawx_int, rawy_int
    logical,            intent(out)  :: on_chip_flag

    call longlat2xy_faster (instpar%oa_rawx, instpar%oa_rawy, instpar%inv_pixscale_x, instpar%inv_pixscale_y, &
        ra0, cosdec0, sindec0, cosroll, sinroll, ra, cosdec, sindec, rawx, rawy )

    rawx_int = INT(rawx+0.5d0)
    rawy_int = INT(rawy+0.5d0)
    on_chip_flag = .false.

    if ( rawx_int .ge. 1 ) then
      if ( rawx_int .le. instpar%naxis1 ) then
        if ( rawy_int .ge. 1 ) then
          if ( rawy_int .le. instpar%naxis2)  then
            on_chip_flag = .true.
          end if
        end if
      end if
    end if
  end subroutine srctool_att_radec2rawxy_a2_faster
  !-------------------------------------------------------------------------------













  !-------------------------------------------------------------------------------
  ! on_chip_flag is set to true if position falls within chip bounds 0.5<=rawx|rawy<NPIX+0.5
  ! this performs the calculation for an array of input ra,dec pairs
  ! also return the integer versions of teh rawx,rawy coords
  recursive subroutine srctool_att_radec2rawxy_as2 (instpar, ra0, dec0, roll, pra, pdec, prawx, prawy, &
      prawx_int, prawy_int, pon_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_as2'
    type(instpar_type), intent(in)  :: instpar
    real(kind=real64),  intent(in)   :: ra0, dec0, roll
    real(kind=real64),  intent(in), dimension(:)   :: pra, pdec
    real(kind=real64),  intent(out), dimension(:)  :: prawx, prawy
    integer(kind=int32), intent(out), dimension(:)  :: prawx_int, prawy_int
    logical,            intent(out), dimension(:)  :: pon_chip_flag
    integer(kind=int32)  :: i,dim

    call longlats2xys (instpar%oa_rawx, instpar%oa_rawy, instpar%inv_pixscale_x, instpar%inv_pixscale_y, &
        ra0, dec0, roll, pra, pdec, prawx, prawy )

    prawx_int(:) = int(prawx(:)+0.5d0)
    prawy_int(:) = int(prawy(:)+0.5d0)
    pon_chip_flag(:) = .false.

    dim = min(size(pra),size(pdec),size(prawx),size(prawy),size(pon_chip_flag))
    do i=1,dim
      if ( prawx_int(i) .ge. 1 ) then
        if ( prawx_int(i) .le. instpar%naxis1 ) then
          if ( prawy_int(i) .ge. 1 ) then
            if ( prawy_int(i) .le. instpar%naxis2)  then
              pon_chip_flag(i) = .true.
            end if
          end if
        end if
      end if
    end do
  end subroutine srctool_att_radec2rawxy_as2
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! same as above but using precomputed coslat,sinlat
  recursive subroutine srctool_att_radec2rawxy_as2_fast (instpar, ra0, dec0, roll, &
      pra, pcosdec, psindec, prawx, prawy, prawx_int, prawy_int, pon_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_as2_fast'
    type(instpar_type), intent(in)  :: instpar
    real(kind=real64),  intent(in)   :: ra0, dec0, roll
    real(kind=real64),  intent(in), dimension(:)   :: pra
    real(kind=real64),  intent(in), dimension(:)   :: pcosdec, psindec
    real(kind=real64),  intent(out), dimension(:)  :: prawx, prawy
    integer(kind=int32), intent(out), dimension(:)  :: prawx_int, prawy_int
    logical,            intent(out), dimension(:)  :: pon_chip_flag
    integer(kind=int32)  :: i,dim

    call longlats2xys_fast (instpar%oa_rawx, instpar%oa_rawy, instpar%inv_pixscale_x, instpar%inv_pixscale_y, &
        ra0, dec0, roll, pra, pcosdec, psindec, prawx, prawy )

    prawx_int(:) = int(prawx(:)+0.5d0)
    prawy_int(:) = int(prawy(:)+0.5d0)
    pon_chip_flag(:) = .false.

    dim = min(size(pra),size(pcosdec),size(psindec),size(prawx),size(prawy),size(pon_chip_flag))
    do i=1,dim
      if ( prawx_int(i) .ge. 1 ) then
        if ( prawx_int(i) .le. instpar%naxis1 ) then
          if ( prawy_int(i) .ge. 1 ) then
            if ( prawy_int(i) .le. instpar%naxis2)  then
              pon_chip_flag(i) = .true.
            end if
          end if
        end if
      end if
    end do
  end subroutine srctool_att_radec2rawxy_as2_fast
  !-------------------------------------------------------------------------------

!  !-------------------------------------------------------------------------------
!  ! rotate the xx,yy into x,y and
!  recursive subroutine srctool_att_xxyy2rawxy_as (instpar, roll, &
!      pxx, pyy, prawx, prawy, prawx_int, prawy_int, pon_chip_flag )
!    character(len=*), parameter :: subroutinename='srctool_att_xxyy2rawxy_as'
!    type(instpar_type), intent(in)  :: instpar
!    real(kind=real64),  intent(in)   :: roll
!    real(kind=real64),  intent(in), dimension(:)   :: pxx, pyy
!    real(kind=real64),  intent(out), dimension(:)  :: prawx, prawy
!    integer(kind=int32), intent(out), dimension(:)  :: prawx_int, prawy_int
!    logical            , intent(out), dimension(:)  :: pon_chip_flag
!    integer(kind=int32)  :: i,dim
!
!    call xxyys2xys (instpar%oa_rawx, instpar%oa_rawy, &
!        roll, pxx(:), pyy(:), prawx(:), prawy(:) )
!
!    prawx_int(:) = int(prawx(:)+0.5d0)
!    prawy_int(:) = int(prawy(:)+0.5d0)
!    pon_chip_flag(:) = .false.
!
!    dim = min(size(pxx),size(pyy),size(prawx),size(prawy),size(pon_chip_flag))
!    do i=1,dim
!      if ( prawx_int(i) .ge. 1 ) then
!        if ( prawx_int(i) .le. instpar%naxis1 ) then
!          if ( prawy_int(i) .ge. 1 ) then
!            if ( prawy_int(i) .le. instpar%naxis2)  then
!              pon_chip_flag(i) = .true.
!            end if
!          end if
!        end if
!      end if
!    end do
!  end subroutine srctool_att_xxyy2rawxy_as
!  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! wrapper for srctool_radec2rawxy_a() - first calculates the ra0,dec0,roll at time
  recursive subroutine srctool_att_radec2rawxy_b (inst, time, ra, dec, rawx, rawy, on_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_b'
    type(inst_type),    intent(in)   :: inst
    real(kind=real64),  intent(in)   :: time, ra, dec
    real(kind=real64),  intent(out)  :: rawx, rawy
    logical,            intent(out)  :: on_chip_flag
    type(instpar_type), pointer :: pinstpar
    real(kind=real64)   :: ra0, dec0, roll
    integer(kind=int32)  :: ierr

    ierr = ERROR_LEVEL_OK

    call srctool_att_get_current_att(inst%attlist, time, ra0, dec0, roll, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error getting current attitude")
      return
    end if

    nullify(pinstpar)
    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable instpar")
      return
    end if

    call srctool_att_radec2rawxy(pinstpar, ra0, dec0, roll, ra, dec, rawx, rawy, on_chip_flag )

  end subroutine srctool_att_radec2rawxy_b
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! wrapper for srctool_radec2rawxy_a() - first calculates the ra0,dec0,roll at time
  ! this performs the calculation for an array of input ra,dec pairs
  ! assume that all dimensions are sensible
  recursive subroutine srctool_att_radec2rawxy_bs (inst, time, pra, pdec, prawx, prawy, pon_chip_flag )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_bs'
    type(inst_type),    intent(in)   :: inst
    real(kind=real64),  intent(in)   :: time
    real(kind=real64),  intent(in), dimension(:)   :: pra, pdec
    real(kind=real64),  intent(out), dimension(:)  :: prawx, prawy
    logical,            intent(out), dimension(:)  :: pon_chip_flag
    real(kind=real64)   :: ra0, dec0, roll
    integer(kind=int32)  :: ierr
    type(instpar_type), pointer :: pinstpar

    ierr = ERROR_LEVEL_OK
    call srctool_att_get_current_att(inst%attlist, time, ra0, dec0, roll, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error getting current attitude")
      return
    end if

    nullify(pinstpar)
    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable instpar")
      return
    end if

    if (size(pra) .ne. size(pdec)  .or. &
        size(pra) .gt. size(prawx) .or. &
        size(pra) .gt. size(prawy) .or. &
        size(pra) .gt. size(pon_chip_flag) ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Array dimensions are mismatched")
    end if

!    dim = min(size(pra),size(pdec),size(prawx),size(prawy),size(pon_chip_flag))
!    do i=1,dim
!      call srctool_radec2rawxy(pinstpar, ra0, dec0, roll, &
!          pra(i), pdec(i), prawx(i), prawy(i), pon_chip_flag(i) )
!    end do

    call srctool_att_radec2rawxy(pinstpar, ra0, dec0, roll, &
        pra, pdec, prawx, prawy, pon_chip_flag )

  end subroutine srctool_att_radec2rawxy_bs
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! wrapper for srctool_radec2rawxy_a() - first calculates the ra0,dec0,roll at time
  ! this version also supplies the current ra0,dec0,roll back to the calling routine.
  recursive subroutine srctool_att_radec2rawxy_c (inst, time, ra, dec, rawx, rawy, on_chip_flag, ra0, dec0, roll )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_c'
    type(inst_type),    intent(in)   :: inst
    real(kind=real64),  intent(in)   :: time, ra, dec
    real(kind=real64),  intent(out)  :: rawx, rawy
    logical,            intent(out)  :: on_chip_flag
    real(kind=real64),  intent(out)  :: ra0, dec0, roll
    integer(kind=int32)              :: ierr
    type(instpar_type), pointer      :: pinstpar

    ierr = ERROR_LEVEL_OK
    call srctool_att_get_current_att(inst%attlist, time, ra0, dec0, roll, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error getting current attitude")
      return
    end if

    nullify(pinstpar)
    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable instpar")
      return
    end if

    call srctool_att_radec2rawxy(pinstpar, ra0, dec0, roll, ra, dec, rawx, rawy, on_chip_flag )

  end subroutine srctool_att_radec2rawxy_c


  !-------------------------------------------------------------------------------
  ! wrapper for srctool_radec2rawxy_a() - first calculates the ra0,dec0,roll at time
  ! this version also supplies the current ra0,dec0,roll back to the calling routine.
  ! this performs the calculation for an array of input ra,dec pairs
  ! assume that all dimensions are sensible
  recursive subroutine srctool_att_radec2rawxy_cs (inst, time, pra, pdec, prawx, prawy, pon_chip_flag, ra0, dec0, roll )
    character(len=*), parameter :: subroutinename='srctool_att_radec2rawxy_cs'
    type(inst_type),    intent(in)   :: inst
    real(kind=real64),  intent(in)   :: time
    real(kind=real64),  intent(in), dimension(:)   :: pra, pdec
    real(kind=real64),  intent(out), dimension(:)  :: prawx, prawy
    logical,            intent(out), dimension(:)  :: pon_chip_flag
    real(kind=real64),  intent(out)                :: ra0,dec0,roll
    integer(kind=int32)  :: ierr
    integer(kind=int32)  :: i,dim
    type(instpar_type), pointer :: pinstpar

    ierr = ERROR_LEVEL_OK
    call srctool_att_get_current_att(inst%attlist, time, ra0, dec0, roll, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error getting current attitude")
      return
    end if

    nullify(pinstpar)
    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable instpar")
      return
    end if

    if (size(pra) .ne. size(pdec)  .or. &
        size(pra) .gt. size(prawx) .or. &
        size(pra) .gt. size(prawy) .or. &
        size(pra) .gt. size(pon_chip_flag) ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Array dimensions are mismatched")
    end if
    dim = min(size(pra),size(pdec),size(prawx),size(prawy),size(pon_chip_flag))

    do i=1,dim
      call srctool_att_radec2rawxy(pinstpar, ra0, dec0, roll, &
          pra(i), pdec(i), prawx(i), prawy(i), pon_chip_flag(i))
    end do

  end subroutine srctool_att_radec2rawxy_cs
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_att_rawxy2radec_a (instpar, ra0, dec0, roll, rawx, rawy, ra, dec )
    character(len=*), parameter :: subroutinename='srctool_att_rawxy2radec_a'
    type(instpar_type), intent(in)  :: instpar
    real(kind=real64),  intent(in)   :: ra0, dec0, roll, rawx, rawy
    real(kind=real64),  intent(out)  :: ra, dec

    call xy2longlat (instpar%oa_rawx, instpar%oa_rawy, instpar%pixscale_x, instpar%pixscale_y, &
        ra0, dec0, roll, rawx, rawy, ra, dec )

  end subroutine srctool_att_rawxy2radec_a
  !-------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------
  ! array version of the above
  recursive subroutine srctool_att_rawxy2radec_as (instpar, ra0, dec0, roll, rawx, rawy, ra, dec )
    character(len=*), parameter :: subroutinename='srctool_att_rawxy2radec_as'
    type(instpar_type), intent(in)  :: instpar
    real(kind=real64),  intent(in) :: ra0, dec0, roll
    real(kind=real64),  intent(in),dimension(:)   :: rawx, rawy
    real(kind=real64),  intent(out),dimension(:)  :: ra, dec

    call xys2longlats (instpar%oa_rawx, instpar%oa_rawy, instpar%pixscale_x, instpar%pixscale_y, &
        ra0, dec0, roll, rawx, rawy, ra, dec )


  end subroutine srctool_att_rawxy2radec_as
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------- END of RADEC2RAWXY handling ------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------





  !-----------------------------------------------------------------------------------
  subroutine srctool_att_test_harness(ierr)
    character(len=*), parameter :: subroutinename='srctool_att_test_harness'
    integer(kind=int32), intent(inout) :: ierr
    integer(kind=int32) :: i
    real(kind=real64)   :: ra1, dec1, time1
    real(kind=real64)   :: ra2, dec2, time2
    real(kind=real64)   :: time
    real(kind=real64)   :: ra_a, dec_a
    real(kind=real64)   :: ra_b, dec_b
    real(kind=real64)   :: dist12, bearing12, distab, mean_distab
    real(kind=real64)   :: pool(1:7)
    real(kind=real64), parameter :: test_radius     = 0.05d0 ! degrees
    real(kind=real64), parameter :: tolerance1       = 0.10d0 ! arcsec, tolerance on individual errors
    real(kind=real64), parameter :: mean_tolerance1  = 0.01d0 ! arcsec, tolerance on the mean error
    real(kind=real64), parameter :: tolerance2       = 1.0d-3 ! arcsec, tolerance for test 2
    integer(kind=int32), parameter :: ntrials = 100000
    integer(kind=int32) :: nfailures
    character(len=STR_MAX) :: str_buffer

    real(kind=real64) :: long0,lat0, long1, lat1, dist01
    real(kind=real64) :: long2,lat2
    real(kind=real64), parameter  :: oa_x = 0.0d0
    real(kind=real64), parameter  :: oa_y = 0.0d0
    real(kind=real64), parameter  :: pixscale_x = 1.0d0
    real(kind=real64), parameter  :: pixscale_y = 1.0d0
    real(kind=real64), parameter  :: inv_pixscale_x = 1.0d0/pixscale_x
    real(kind=real64), parameter  :: inv_pixscale_y = 1.0d0/pixscale_y
    real(kind=real64), parameter  :: roll0 = 0.0d0
    real(kind=real64) :: x1,y1

    ! skip processing if have already found errors
    if (ierr .ne. ERROR_LEVEL_OK) return

    call wrschatty(subroutinename, "Testing the srctool attitude module")

    !------------------------------------------------------------
    ! Test 1: test accuracy of coordinate interpolation
    !------------------------------------------------------------
!    call random_seed()
    nfailures = 0
    mean_distab = 0.0d0
    do i=1,ntrials
      call random_number(pool)
      ra1  = pool(1)*360.0d0
      dec1 = pool(2)*180.0d0 - 90.0d0
      dist12    = pool(3)*test_radius
      bearing12 = pool(4)*360.0d0
      call calc_destination (ra1, dec1, dist12, bearing12, ra2, dec2)
      time1 = pool(5)*10.0d8
      time2 = time1 + pool(6)*10.0d2
      time =  time1 + pool(7) * (time2-time1)

      call interpolate_long_lat ( &
          time1, ra1,  dec1,   &
          time2, ra2, dec2, &
          time, ra_a, dec_a, ierr )

      if (ierr .ne. ERROR_LEVEL_OK) then
        call wrserr (subroutinename, ierr, "Errors found during test")
      endif

      call interpolate_long_lat_fast ( &
          time1, ra1,  dec1,   &
          cosd(dec1), sind(dec1),  &
          time2, ra2, dec2, &
          dist12, cosd(bearing12), sind(bearing12),&
          time, ra_b, dec_b, ierr )

      if (ierr .ne. ERROR_LEVEL_OK) then
        call wrserr (subroutinename, ierr, "Errors found during test")
      endif

      distab = calc_ang_sep(ra_a,dec_a,ra_b,dec_b)*DEG2ARCSEC
      mean_distab = mean_distab + distab
      if ( distab .gt. tolerance1 ) then
        nfailures = nfailures + 1
!        if ( modulo(nfailures,50) .eq. 1 ) then
!          write (*, '(A,x,A7,3(3(x,A14),x,A),2(x,A14,x,A14,x,A),x,A12,x,A)'), &
!             "#TEST_ATT", "trial", &
!             "time1",   "ra1",   "dec1",  "|", &
!             "time2",   "ra2",   "dec2",  "|", &
!             "time", "dist12",   "bearing12",  "|", &
!             "ra_a", "dec_a", "|", &
!             "ra_b", "dec_b", "|", &
!             'distab(")', "|"
!        end if
!        write (*, '(A,x,I7,3(x,F14.4,x,2(F14.8,x),A),2(x,F14.8,x,F14.8,x,A),x,F12.8,x,A)'), &
!            " TEST_ATT",  i, &
!            time1,   ra1,   dec1,  "|", &
!            time2,   ra2,   dec2,  "|", &
!            time, dist12,  bearing12,   "|", &
!            ra_a, dec_a, "|", &
!            ra_b, dec_b, "|", &
!            distab, "|"

      end if
    end do

    if ( nfailures .gt. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Individual errors on fast coordinate interpolation exceeded tolerance")
      return
    endif

    mean_distab = mean_distab / REAL(ntrials,kind=real64)
    if ( mean_distab .gt. mean_tolerance1 ) then
      write(str_buffer, '(F12.7)') mean_distab
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Mean error on fast coordinate interpolation exceeded tolerance")
      call wrserr(subroutinename,  ierr, &
        "Mean difference between fast and normally computed coordinates = " // trim(str_buffer) // " arcsec")
      return
    endif
    ! Test 1: end
    !------------------------------------------------------------

    !------------------------------------------------------------
    ! Test 2: test accuracy of coordinate projection onto plane and back again
    !------------------------------------------------------------

!    write(*,'(A,x,A4,10(x,A14))'), "ATT_TEST2",  &
!        "i", "long0", "lat0", "long1", "lat1",  "dist01", "x1", "y1", "long2",  "lat2", "dist12"
    nfailures = 0
    do i=1,100000
      call random_number(long0)
      call random_number(lat0)
      long0 = 360.0d0 * long0
      lat0  = -90.0d0 + 180.0d0 * lat0    ! this will produce many more points near the poles than is
      dist01 = huge(dist01)
      do while ( dist01 .gt. 60.0d0 ) ! only test for separations smaller than 60deg
        call random_number(long1)
        call random_number(lat1)
        long1 = 360.0d0 * long1
        lat1  = -90.0d0 + 180.0d0 * lat1
        dist01 = calc_ang_sep(long0,lat0,long1,lat1)
      end do

      call longlat2xy(oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, long0, lat0, roll0, &
          long1, lat1, x1, y1)
      call xy2longlat(oa_x, oa_y, pixscale_x, pixscale_y, long0, lat0, roll0, &
          x1, y1, long2, lat2)
      dist12 = calc_ang_sep(long1,lat1,long2,lat2)

      if ( dist12 .gt. tolerance2 ) then
        nfailures = nfailures + 1

        !write(*,'(A,x,I4,10(x,F14.7))'), "ATT_TEST2",  i, long0, lat0, long1, lat1, dist01, x1, y1, long2, lat2, dist12
      endif
    end do

    if ( nfailures .gt. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Accuracy of projection onto plane and back failed accuracy requirement")
      return
    endif




    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Errors found during test")
    endif

  end subroutine srctool_att_test_harness
  !-----------------------------------------------------------------------------------










end module srctool_att



!
!!    ! will need to use this routine to
!!    ! convert the attitude info to pixel locations and vice versa
!!    subroutine radec2pix (ra, de, n, ra0, de0, x0, y0, xd, yd, x, y)
!!      !-----------------------------------------------------------------------
!!      ! HB 20100326
!!      !-----------------------------------------------------------------------
!!      ! This routine transforms a vector of ra[deg], dec[deg] sky locations
!!      ! into x, y pixel coordinates (parallel projection) for a set of WCS
!!      ! keywords. We use parallel projection because the transformation
!!      ! requires less computation than other tangential projections
!!      ! (formulae taken from Konrad Dennerl's "Survey II" document).
!!      !-----------------------------------------------------------------------
!!
!!
!  ! input:
!  !    ra      r*8 1d-array  deg
!  !    de      r*8 1d-array  deg
!  !    n       i*4                  number of positions to transform
!  !    ra0     r*8           deg    sky position of reference pixel (WCS)
!  !    dec0    r*8           deg    sky position of reference pixel (WCS)
!  !    x0      r*8                  reference pixel from WCS keywords
!  !    y0      r*8                  reference pixel from WCS keywords
!  !    xd, yd  r*8           deg    pixel size (increment) from WCS
!  ! output:
!  !    x       r*8 1d-array
!  !    y       r*8 1d-array
!
!  rad2xpix = 1d0/(xd*deg2rad)
!  rad2ypix = 1d0/(yd*deg2rad)
!
!  ra0r = ra0*deg2rad
!  de0r = de0*deg2rad
!
!  sr0 = dsin(ra0r)
!  cr0 = dcos(ra0r)
!
!  sd0 = dsin(de0r)
!  cd0 = dcos(de0r)
!
!! vector operations start here; could be changed into a
!! traditional do loop if this turns out to be faster
!
!  rar = ra(1:n)*deg2rad
!  der = de(1:n)*deg2rad
!
!  sr = dsin(rar)
!  cr = dcos(rar)
!
!  sd = dsin(der)
!  cd = dcos(der)
!
!  x1 = cd*cr
!  y1 = cd*sr
!
!  y2 = -cr0*x1-sr0*y1
!
!  x3 = -sr0*x1+cr0*y1
!  y3 =  sd0*y2+cd0*sd
!
!  x(1:n) = x0 + x3*rad2xpix
!  y(1:n) = y0 + y3*rad2ypix
!
!!
!!
!!
!!      ! the expanded formula is as follows
!!
!  x = x0 + (-sin(ra0) *cos(dec)*cos(ra)                                         + cos(ra0) *cos(dec)*sin(ra))/(xd*deg2rad)
!  y = y0 + ( sin(dec0)*(-cos(ra0)*cos(dec)*cos(ra) - sin(ra0)*cos(dec)*sin(ra)) + cos(dec0)*sin(dec)        )/(yd*deg2rad)
!
!
!!
!!
!!
