! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_deadcor

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-25
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling DEADCOR data/files
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_DEADCOR_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_deadcor
  public :: srctool_deadcor_merge_deadcors_for_inst
  public :: srctool_deadcor_empty_deadcor
  public :: srctool_deadcor_read_deadcor_from_chdu

contains





  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of DEADCOR extension handling ------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_deadcor_empty_deadcor(deadcor)
    type(deadcor_type), intent(inout)      :: deadcor
    if ( allocated(deadcor%ptime))   deallocate (deadcor%ptime)
    if ( allocated(deadcor%pdeadc))  deallocate (deadcor%pdeadc)
    if ( allocated(deadcor%pstart))  deallocate (deadcor%pstart)
!    if ( allocated(deadcor%pstop))   deallocate (deadcor%pstop)
    deadcor%num_rows  = 0
    deadcor%is_valid = .false.
    call fitswrap_init_hduinfo(deadcor%hduinfo)
  end subroutine srctool_deadcor_empty_deadcor
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the deadcor info from the current hdu of the fits file identified by unit
  subroutine srctool_deadcor_read_deadcor_from_chdu(unit, deadcor, ierr)
    character(len=*), parameter :: subroutinename='srctool_deadcor_read_deadcor_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(deadcor_type), intent(inout) :: deadcor
    integer(kind=int32), intent(out)  :: ierr

    real(kind=real64)  , parameter :: nullval_real64  = 0.0d0
    real(kind=real32)  , parameter :: nullval_real32  = 0.0e0

    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer

    integer(kind=int32) :: colnum_time
    integer(kind=int32) :: colnum_deadc

    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf

    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in DEADCOR extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, deadcor%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    !get the number of rows in the extension
    call FTGNRW(unit, deadcor%num_rows, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    !check that have at least some entries and then read in the column data
    if ( deadcor%num_rows .gt. 0 ) then
      call wrschatty (subroutinename,  &
          "Reading in: " // trim(i2s(deadcor%num_rows)) // " rows of DEADCOR info from extname='" // &
          trim(deadcor%hduinfo%key_extname%value) // "' instrument='" // &
          trim(deadcor%hduinfo%key_instrume%value) // "'" )

      call fitswrap_prep_col(unit, "TIME",  deadcor%num_rows, colnum_time,  deadcor%ptime,   ierr)
      call fitswrap_prep_col(unit, "DEADC", deadcor%num_rows, colnum_deadc, deadcor%pdeadc,  ierr)

      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if

      ! get the chunksize
      call ftgrsz(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      do row = 1, deadcor%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,deadcor%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVD(unit,colnum_time, row,1,nrows_to_read,nullval_real64,deadcor%ptime(row:row_end), anyf,status)
        call FTGCVE(unit,colnum_deadc,row,1,nrows_to_read,nullval_real32,deadcor%pdeadc(row:row_end),anyf,status)

        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do

    end if

    !  Leave this until we merge the deadcor arrays
    !    !----------------------------------------
    !    ! now process the ptime array
    !    ! ptime(i) gives time at centre of bin i
    !    allocate(deadcor%pstart(1:deadcor%num_rows),&
    !!        deadcor%pstop(1:deadcor%num_rows), &
    !        STAT = status)
    !
    !    if ( check_allocate_status(subroutinename, status, ierr) ) return
    !
    !      deadcor%pstart(1)                = VERY_NEGATIVE_DOUBLE
    !  !    deadcor%pstop(deadcor%num_rows)  = VERY_POSITIVE_DOUBLE
    !  !    tstop = VERY_POSITIVE_DOUBLE
    !      if ( deadcor%num_rows .gt. 1 ) then
    !  !      deadcor%pstop(1)  = 0.5d0*(deadcor%ptime(1) + deadcor%ptime(2))
    !        tstop  = 0.5d0*(deadcor%ptime(1) + deadcor%ptime(2))
    !        do i=2,deadcor%num_rows-1
    !  !        deadcor%pstart(i) = deadcor%pstop(i-1)
    !          deadcor%pstart(i) = tstop
    !  !        deadcor%pstop(i) = 0.5d0*(deadcor%ptime(i) + deadcor%ptime(i+1))
    !          tstop = 0.5d0*(deadcor%ptime(i) + deadcor%ptime(i+1))
    !        end do
    !  !      deadcor%pstart(deadcor%num_rows) = deadcor%pstop(deadcor%num_rows-1)
    !        deadcor%pstart(deadcor%num_rows) = tstop
    !      end if
  !----------------------------------------


    deadcor%is_valid = .true.
  end subroutine srctool_deadcor_read_deadcor_from_chdu
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! merge all of the deadcors for a given instrument
  subroutine srctool_deadcor_merge_deadcors_for_inst(inst, ierr)
    character(len=*), parameter :: subroutinename='srctool_deadcor_merge_deadcors_for_inst'
    type(inst_type), intent(inout)  :: inst
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)   :: i, j1,j2,n
    integer(kind=int32) :: status
    real(kind=real64) :: tstop
    ierr = ERROR_LEVEL_OK
    status = 0

    ! copy the header from the first evtlist
    if ( inst%evtlistlist%num_evtlists .gt. 0 ) then
      inst%deadcor%hduinfo = inst%evtlistlist%evtlist(1)%deadcor%hduinfo
      ! TODO update the correct values of the TSTART, TSTOP, DATE-OBS, DATE-END etc keywords
    else
      ! exit quietly
      return
    end if

    inst%deadcor%num_rows = 0
    do i=1,inst%evtlistlist%num_evtlists
      if ( inst%evtlistlist%evtlist(i)%deadcor%is_valid ) then
        inst%deadcor%num_rows = inst%deadcor%num_rows + inst%evtlistlist%evtlist(i)%deadcor%num_rows
      end if
    end do
    if ( inst%deadcor%num_rows .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not find any valid deadcor entries ")
      return
    end if

    allocate(inst%deadcor%ptime(inst%deadcor%num_rows+1), &
        inst%deadcor%pstart(inst%deadcor%num_rows+1), &
        inst%deadcor%pdeadc(inst%deadcor%num_rows+1), &
        STAT = status)

    inst%deadcor%ptime(:)  = -1.0d0
    inst%deadcor%pstart(:) = -1.0d0
    inst%deadcor%pdeadc(:) = -1.0d0

    if ( check_allocate_status(subroutinename, status, ierr) ) return

    j1 = 1
    do i=1,inst%evtlistlist%num_evtlists
      if ( inst%evtlistlist%evtlist(i)%deadcor%is_valid ) then
        n = inst%evtlistlist%evtlist(i)%deadcor%num_rows
        j2 = j1 + n - 1
        inst%deadcor%ptime(j1:j2)  = inst%evtlistlist%evtlist(i)%deadcor%ptime(1:n)
        inst%deadcor%pdeadc(j1:j2) = inst%evtlistlist%evtlist(i)%deadcor%pdeadc(1:n)
        j1 = j2+1
      end if
    end do

    ! sort the deadcors then recalc the pstart() array
    call insertion_cosort(inst%deadcor%pstart(1:inst%deadcor%num_rows), &
        inst%deadcor%pdeadc(1:inst%deadcor%num_rows))

! debug info
!    write (*,'(A,x,A6,3(x,A12))'), "DEBUG_DEADCOR1", "i", "pstart(i)", "ptime(i)", "pdeadc(i)"
!    do i=1,size(inst%deadcor%pstart)
!      write (*,'(A,x,I6,3(x,E12.4))'), "DEBUG_DEADCOR1", i, inst%deadcor%pstart(i), inst%deadcor%ptime(i), inst%deadcor%pdeadc(i)
!    end do

    ! TODO deal with duplicates/overlaps?

    inst%deadcor%pstart(1)  = -1d0 * huge(inst%deadcor%pstart(1)) ! VERY_NEGATIVE_DOUBLE
    if ( inst%deadcor%num_rows .gt. 1 ) then
      tstop  = 0.5d0*(inst%deadcor%ptime(1) + inst%deadcor%ptime(2))
      do i=2,inst%deadcor%num_rows-1
        inst%deadcor%pstart(i) = tstop
        tstop = 0.5d0*(inst%deadcor%ptime(i) + inst%deadcor%ptime(i+1))
      end do
      inst%deadcor%pstart(inst%deadcor%num_rows) = tstop
    end if
    !----------------------------------------
    inst%deadcor%ptime(inst%deadcor%num_rows+1) = huge(inst%deadcor%ptime(1)) ! VERY_POSITIVE_DOUBLE
    inst%deadcor%pstart(inst%deadcor%num_rows+1) = huge(inst%deadcor%pstart(1)) ! VERY_POSITIVE_DOUBLE
    inst%deadcor%pdeadc(inst%deadcor%num_rows+1) = inst%deadcor%pdeadc(inst%deadcor%num_rows)

    inst%deadcor%is_valid = .true.


! debug info
!    write (*,'(A,x,A6,3(x,A12))'), "DEBUG_DEADCOR", "i", "pstart(i)", "ptime(i)", "pdeadc(i)"
!    do i=1,size(inst%deadcor%pstart)
!      write (*,'(A,x,I6,3(x,E12.4))'), "DEBUG_DEADCOR", i, inst%deadcor%pstart(i), inst%deadcor%ptime(i), inst%deadcor%pdeadc(i)
!    end do
!    write (*,*) "DEBUG_DEADCOR2"
  end subroutine srctool_deadcor_merge_deadcors_for_inst
  !---------------------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------------
  !-------------------------END of DEADCOR extension handling  -------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------





end module srctool_deadcor
