! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_ebounds


  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-24
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling RMF EBOUNDS extensions
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_EBOUNDS_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_ebounds
!unused!  private :: srctool_ebounds_setup_dummy_ebounds
  private :: srctool_ebounds_write_ebounds_keys

  public ::  srctool_ebounds_append_ebounds_to_file
  public ::  srctool_ebounds_empty_ebounds
  public ::  srctool_ebounds_read_ebounds_from_chdu
  public ::  srctool_ebounds_copy_ebounds
  public ::  srctool_ebounds_report_ebounds_info


contains

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of EBOUNDS file handling------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_ebounds_empty_ebounds(ebounds)
    type(ebounds_type), intent(inout)      :: ebounds
    if ( allocated(ebounds%pchannel))  deallocate (ebounds%pchannel)
    if ( allocated(ebounds%pemin)) deallocate (ebounds%pemin)
    if ( allocated(ebounds%pemax)) deallocate (ebounds%pemax)
    ebounds%num_rows = 0
    ebounds%is_valid = .false.
    call fitswrap_init_hduinfo(ebounds%hduinfo)
!    call fitswrap_init_key(ebounds%key_phafile,  "PHAFILE")
!    call fitswrap_init_key(ebounds%key_detchans, "DETCHANS")
!    call fitswrap_init_key(ebounds%key_tlmin,   "TLMIN")
  end subroutine srctool_ebounds_empty_ebounds
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! take a new copy of the ebounds info
  subroutine srctool_ebounds_copy_ebounds(ebounds_in, ebounds_out, ierr)
    character(len=*), parameter :: subroutinename='srctool_ebounds_copy_ebounds'
    type(ebounds_type), intent(in)      :: ebounds_in
    type(ebounds_type), intent(inout)   :: ebounds_out
    integer(kind=int32), intent(out)  :: ierr
    integer(kind=int32) :: status
    ierr = ERROR_LEVEL_OK
    status = 0
    call wrschatty (subroutinename, "Duplicating EBOUNDS for merged instrument (from " &
        // trim(ebounds_in%hduinfo%key_instrume%value) // ")")

    if ( .not. ebounds_in%is_valid .or. (ebounds_in%num_rows .lt. 1 )) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Invalid EBOUNDS supplied")
      return
    endif

    call srctool_ebounds_empty_ebounds(ebounds_out)

    ! make space for the allocated arrays
    allocate(ebounds_out%pemin(ebounds_in%num_rows),&
        ebounds_out%pemax(ebounds_in%num_rows),&
        ebounds_out%pchannel(ebounds_in%num_rows),&
        STAT = status)

    if ( status .ne. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Problem duplicating EBOUNDS for merged instrument")
      return
    end if

    ! now copy the data
    ebounds_out = ebounds_in
    call fitswrap_set_key(ebounds_out%hduinfo%key_instrume, INSTRUME_IS_MERGED)
    call fitswrap_set_key(ebounds_out%hduinfo%key_filter,   trim(ESASS_FILTER_STR(ESASS_FILTER_CODE_INDEF)))

  end subroutine srctool_ebounds_copy_ebounds
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !read the ebounds from the current hdu of the fits file identified by unit
  subroutine srctool_ebounds_read_ebounds_from_chdu(unit, ebounds, ierr)
    character(len=*), parameter :: subroutinename='srctool_ebounds_read_ebounds_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(ebounds_type), intent(inout) :: ebounds
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer

    integer(kind=int32) :: colnum_emin
    integer(kind=int32) :: colnum_emax
    integer(kind=int32) :: colnum_channel

    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    integer(kind=int32), parameter :: nullval_int32  = 0
    real(kind=real32)  , parameter :: nullval_real32  = 0.0


    ierr = ERROR_LEVEL_OK
    status = 0
    call wrschatty (subroutinename, "Reading in EBOUNDS extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, ebounds%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    !- check the following keader keywords are present
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1       'TEMPORALDATA'
    !- HDUCLAS2       'ASPECT'

    !get the number of rows in the extension
    call FTGNRW(unit, ebounds%num_rows, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    !check that have at least some entries and then read in the column data
    if ( ebounds%num_rows .gt. 0 ) then
      call wrschatty (subroutinename, &
          "Reading in: " // trim(i2s(ebounds%num_rows)) // " rows of EBOUNDS info from extname='" // &
          trim(ebounds%hduinfo%key_extname%value) // "' instrument='" // &
          trim(ebounds%hduinfo%key_instrume%value) // "'" )


!      call fitswrap_proc_col(unit, "CHANNEL", ebounds%num_rows, ebounds%pchannel, ierr)
!      call fitswrap_proc_col(unit, "E_MIN",   ebounds%num_rows, ebounds%pemin,    ierr)
!      call fitswrap_proc_col(unit, "E_MAX",   ebounds%num_rows, ebounds%pemax,    ierr)
!
!      if (ierr .ne. ERROR_LEVEL_OK ) then
!        call wrserr (subroutinename, ierr, "Problem reading column(s) from HDU")
!        return
!      end if
      call fitswrap_prep_col(unit, "CHANNEL", ebounds%num_rows, colnum_channel, ebounds%pchannel, ierr)
      call fitswrap_prep_col(unit, "E_MIN",   ebounds%num_rows, colnum_emin,    ebounds%pemin,    ierr)
      call fitswrap_prep_col(unit, "E_MAX",   ebounds%num_rows, colnum_emax,    ebounds%pemax,    ierr)

      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if

      ! get the chunksize
      call ftgrsz(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      do row = 1, ebounds%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,ebounds%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVJ(unit, colnum_channel,  row,1,nrows_to_read,nullval_int32, ebounds%pchannel(row:row_end), anyf,status)
        call FTGCVE(unit, colnum_emin,     row,1,nrows_to_read,nullval_real32,ebounds%pemin(row:row_end),    anyf,status)
        call FTGCVE(unit, colnum_emax,     row,1,nrows_to_read,nullval_real32,ebounds%pemax(row:row_end),    anyf,status)

        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do
    end if
    ebounds%is_valid = .true.

  end subroutine srctool_ebounds_read_ebounds_from_chdu
  !-------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! report the ebounds info for all instruments
  subroutine srctool_ebounds_report_ebounds_info(instlist)
    character(len=*), parameter :: subroutinename='srctool_ebounds_report_ebounds_info'
    type(instlist_type), intent(in)  :: instlist
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i
    write(str_buffer, '(A12,1x,A6,3(1x,A10))') &
        "#EBOUNDSInfo", &
        "Inst", "NumChans", "Emin", "Emax"
    call wrschatty (subroutinename, str_buffer)
    write(str_buffer, '(A12,1x,A6,3(1x,A10))') &
        "#EBOUNDSInfo", &
        "-", "-", "(keV)", "(keV)"
    call wrschatty (subroutinename, str_buffer)
    do i=0,MAX_INST
      if ( instlist%inst(i)%ebounds%is_valid ) then
        if ( instlist%inst(i)%ebounds%num_rows .gt. 0 ) then
          write(str_buffer, '(A12,1x,A6,1x,I10,2(1x,F10.3))') &
              "EBOUNDSInfo", &
              trim(instlist%inst(i)%name), &
              instlist%inst(i)%ebounds%num_rows, &
              instlist%inst(i)%ebounds%pemin(1), &
              instlist%inst(i)%ebounds%pemax(instlist%inst(i)%ebounds%num_rows)
          call wrschatty (subroutinename, str_buffer)
        end if
      end if
    end do
  end subroutine srctool_ebounds_report_ebounds_info
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! write ebounds fits keys info to the chdu
  subroutine srctool_ebounds_write_ebounds_keys(unit, ebounds, ierr)
    character(len=*), parameter :: subroutinename='srctool_ebounds_write_ebounds_keys'
    integer(kind=int32), intent(in)  :: unit
    type(ebounds_type),      intent(in)  :: ebounds
    integer(kind=int32), intent(out)   :: ierr
!    type(fitskey_string_type) :: key_creator ! use a local version of key here so we don't need to
!                                             ! write to the input ebounds structure

    ierr = ERROR_LEVEL_OK

    ! write the standard keywords to the binary table HDU:
    call fitswrap_write_hduinfo_quiet(unit, ebounds%hduinfo, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      return
    end if

!    !-----------------------------------
!    ! write the compulsory keywords to the binary table HDU:
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_telescop, ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_instrume, ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_filter,   ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_hduclass, ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_hduclas1, ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_hduclas2, ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_hduvers,  ierr)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!     end if
!
!    call fitswrap_init_key(key_creator,  "CREATOR")
!    call fitswrap_set_key(key_creator,  "CREATOR", CREATOR_STRING, "Software which created this file")

!    ! now the optional keys
!    call fitswrap_write_key (unit, key_creator,  ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_object,   ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_ra_obj,   ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_dec_obj,  ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_equinox,  ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_radecsys, ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_tstart,   ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_tstop,    ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_date_obs, ierr)
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_date_end, ierr)

!    ! XFLTXXXX keywords
!    call fitswrap_write_key (unit, ebounds%hduinfo%key_xflt(i), ierr)


!    ! now the EBOUNDS-specific keys
!    call fitswrap_write_key (unit, ebounds%key_phafile, ierr)
!    call fitswrap_write_key (unit, ebounds%key_detchans, ierr)
!    call fitswrap_write_key (unit, ebounds%key_tlmin, ierr)

!    ! do this quietly
!    if (ierr .ne. ERROR_LEVEL_OK) then
!      ierr = ERROR_LEVEL_OK
!      return
!    end if
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
    ! TODO - any other keywords
    !-----------------------------------

  end subroutine srctool_ebounds_write_ebounds_keys
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !append an ebounds extension to the end of the (open) file identified by unit
  subroutine srctool_ebounds_append_ebounds_to_file(unit, ebounds, ierr)
    character(len=*), parameter :: subroutinename='srctool_ebounds_append_ebounds_to_file'

    integer(kind=int32), intent(in)   :: unit
    type(ebounds_type), intent(in)    :: ebounds
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32), parameter :: ncols=3
    character(len=*),    parameter :: ttype(ncols) = (/ "CHANNEL", "E_MIN  ", "E_MAX  " /)
    character(len=*),    parameter :: tunit(ncols) = (/ "   "    , "keV"    , "keV"     /)
    character(len=*),    parameter :: tform(ncols) = (/ "J"      , "E"      , "E"       /)
    integer(kind=int32), parameter :: COLNUM_CHANNEL = 1
    integer(kind=int32), parameter :: COLNUM_E_MIN   = 2
    integer(kind=int32), parameter :: COLNUM_E_MAX   = 3

!    character(len=*),    parameter :: extname = "EBOUNDS"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int64), parameter :: frow=1, felem=1

    integer(kind=int32)    :: status


    ierr = ERROR_LEVEL_OK
    status = 0
    call wrschatty (subroutinename, "Appending EBOUNDS extension")

    ! TODO check the status of the inputs
    if ( .not. ebounds%is_valid ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Invalid EBOUNDS data structure")
      return
    endif

    !-----------------------------------
    ! append a new binary table HDU
    call FTIBIN(unit,ebounds%num_rows,ncols,ttype,tform,tunit,trim(ebounds%hduinfo%key_extname%value),varidat,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------


    !-----------------------------------
    call srctool_ebounds_write_ebounds_keys(unit, ebounds, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing EBOUNDS fits header keys")
      return
    end if
    !-----------------------------------


    !-----------------------------------
    ! now write the column data to the file
    ! format: FTPCL[SLBIJKEDCM](unit,colnum,frow,felem,nelements,values, > status)
    ! TODO do this in chunks for better efficiency

    ! write CHANNEL column
    call  FTPCLJ(unit,COLNUM_CHANNEL,frow,felem,ebounds%num_rows,ebounds%pchannel,status)
    ! write EMIN column
    call  FTPCLE(unit,COLNUM_E_MIN,  frow,felem,ebounds%num_rows,ebounds%pemin,status)
    ! write EMAX column
    call  FTPCLE(unit,COLNUM_E_MAX,  frow,felem,ebounds%num_rows,ebounds%pemax,status)


    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    !-----------------------------------




  end subroutine srctool_ebounds_append_ebounds_to_file

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------END of EBOUNDS file handling------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

end module srctool_ebounds
