! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_evtlist

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         ! structs and datatypes
  use srctool_misc
  use srctool_gti
  use srctool_att
  use srctool_badpix
  use srctool_deadcor
  use srctool_cal

  implicit none

  character(len=*), private, parameter :: modulename='SRCTOOL_EVTLIST_MOD'

  private :: srctool_evtlist_empty_evtarrays
  private :: srctool_evtlist_empty
  public  :: srctool_evtlist_empty_evtlistlist
  private :: srctool_evtlist_allocate_arrays
  private :: srctool_evtlist_setup_merged_hduinfo
  public  :: srctool_evtlist_read_evtlists_from_file
  private :: srctool_evtlist_read_from_file
  private :: srctool_evtlist_read_from_chdu
  private :: srctool_evtlist_divide_by_inst

  public  :: srctool_evtlist_print_evt_headerline
  public  :: srctool_evtlist_print_evt

contains

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !------------------------START of EVTLIST file handling---------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_evtlist_empty_evtarrays(evtlist)
    type(evtlist_type), intent(inout)      :: evtlist
    if ( allocated(evtlist%ptime      )) deallocate ( evtlist%ptime      )
    if ( allocated(evtlist%pra        )) deallocate ( evtlist%pra        )
    if ( allocated(evtlist%pdec       )) deallocate ( evtlist%pdec       )
    if ( allocated(evtlist%px         )) deallocate ( evtlist%px         )
    if ( allocated(evtlist%py         )) deallocate ( evtlist%py         )
    if ( allocated(evtlist%penergy    )) deallocate ( evtlist%penergy    )
    if ( allocated(evtlist%pev_weight )) deallocate ( evtlist%pev_weight )
    if ( allocated(evtlist%prawx      )) deallocate ( evtlist%prawx      )
    if ( allocated(evtlist%prawy      )) deallocate ( evtlist%prawy      )
    if ( allocated(evtlist%psubx      )) deallocate ( evtlist%psubx      )
    if ( allocated(evtlist%psuby      )) deallocate ( evtlist%psuby      )
    if ( allocated(evtlist%ppha       )) deallocate ( evtlist%ppha       )
    if ( allocated(evtlist%ppat_typ   )) deallocate ( evtlist%ppat_typ   )
    if ( allocated(evtlist%ppat_inf   )) deallocate ( evtlist%ppat_inf   )
    if ( allocated(evtlist%ptm_nr     )) deallocate ( evtlist%ptm_nr     )
    if ( allocated(evtlist%pflag      )) deallocate ( evtlist%pflag      )
    if ( allocated(evtlist%pclass_mask)) deallocate ( evtlist%pclass_mask)
    evtlist%num_rows = 0
  end subroutine srctool_evtlist_empty_evtarrays
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_evtlist_empty(evtlist)
    type(evtlist_type), intent(inout)      :: evtlist
    call srctool_evtlist_empty_evtarrays(evtlist)
    call srctool_gti_empty_gtilist(evtlist%gtilist)
    call srctool_badpix_empty_badpix (evtlist%badpix)
    call srctool_att_empty_attlist(evtlist%attlist)
    call srctool_deadcor_empty_deadcor(evtlist%deadcor)
    evtlist%is_valid = .false.
!    evtlist%filename = "NONE"

    call fitswrap_init_hduinfo(evtlist%hduinfo)
    call fitswrap_init_key(evtlist%key_flagsel , "FLAGSEL" )
    call fitswrap_init_key(evtlist%key_pat_sel , "PAT_SEL" )
    evtlist%obs_mode = OBS_MODE_NULL

  end subroutine srctool_evtlist_empty
  !--------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_evtlist_empty_evtlistlist(evtlistlist)
    type(evtlistlist_type), intent(inout) :: evtlistlist
    integer(kind=int32) :: i
    do i=1,MAX_EVTFILES
      call srctool_evtlist_empty(evtlistlist%evtlist(i))
    end do
    evtlistlist%num_evtlists = 0
    call fitswrap_init_hduinfo(evtlistlist%hduinfo)

  end subroutine srctool_evtlist_empty_evtlistlist
  !--------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine srctool_evtlist_allocate_arrays(evtlist, num_events, ierr)
    character(len=*), parameter :: subroutinename='srctool_evtlist_allocate_arrays'

    type(evtlist_type),  intent(inout) :: evtlist
    integer(kind=int32), intent(in)    :: num_events
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status
    ierr = ERROR_LEVEL_OK
    status = 0
!    evtlist%array_length = num_events
    evtlist%num_rows     = 0

!    write (str_buffer, '(A,I0,A)'), "Allocating memory for ", num_events, " events"
!    call wrschatty (subroutinename, str_buffer)

!    write (str_buffer,'(A,I0)'), "Size of ptime: ", size(evtlist%ptime)
!    call wrschatty (subroutinename, str_buffer)

    allocate(evtlist%ptime(num_events), &
        evtlist%pra(num_events), &
        evtlist%pdec(num_events), &
        evtlist%px(num_events), &
        evtlist%py(num_events), &
        evtlist%penergy(num_events), &
        evtlist%pev_weight(num_events), &
        evtlist%prawx(num_events), &
        evtlist%prawy(num_events), &
        evtlist%psubx(num_events), &
        evtlist%psuby(num_events), &
        evtlist%ppha(num_events), &
        evtlist%ppat_typ(num_events), &
        evtlist%ppat_inf(num_events), &
        evtlist%ptm_nr(num_events), &
        evtlist%pflag(num_events), &
        evtlist%pclass_mask(num_events), &
        STAT = status)

    if ( check_allocate_status (subroutinename, status, ierr)) return

    ! initialise the values in the arrays to zero
    evtlist%ptime(:)      = 0.0d0
    evtlist%pra(:)        = 0.0d0
    evtlist%pdec(:)       = 0.0d0
    evtlist%px(:)         = 0
    evtlist%py(:)         = 0
    evtlist%penergy(:)    = 0.0e0
    evtlist%pev_weight(:) = 0.0e0
    evtlist%prawx(:)      = 0
    evtlist%prawy(:)      = 0
    evtlist%psubx(:)      = 0
    evtlist%psuby(:)      = 0
    evtlist%ppha(:)       = 0
    evtlist%ppat_typ(:)   = 0
    evtlist%ppat_inf(:)   = 0
    evtlist%ptm_nr(:)     = 0
    evtlist%pflag(:)      = 0
    evtlist%pclass_mask(:) = EVT_CLASS_FLAG_NULL

  end subroutine srctool_evtlist_allocate_arrays
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_evtlist_setup_merged_hduinfo(instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_evtlist_setup_merged_hduinfo'
    type(instlist_type), intent(inout),target   :: instlist
    integer(kind=int32), intent(out)            :: ierr
    integer(kind=int32)        :: i, e, ee
    type(evtlistlist_type), pointer  :: pevtlistlist

    integer(kind=int32)              :: tstart_min_index
    integer(kind=int32)              :: tstop_max_index
    integer(kind=int32)              :: date_obs_min_index
    integer(kind=int32)              :: date_end_max_index
    real(kind=real64)                :: tstart_min
    real(kind=real64)                :: tstop_max
    character(len=FITSIO_KEY_MAXLEN) :: date_obs_min
    character(len=FITSIO_KEY_MAXLEN) :: date_end_max
    character(len=STR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK

    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        pevtlistlist => instlist%inst(i)%evtlistlist
        if ( pevtlistlist%num_evtlists .gt. 0 ) then
          ! first copy all the hduinfo from the 1st evtlist
          pevtlistlist%hduinfo = pevtlistlist%evtlist(1)%hduinfo

          ! take a coded copy of the filter type
          instlist%inst(i)%filter_code = srctool_misc_filter_string_to_code(pevtlistlist%hduinfo%key_filter%value)

          ! verify that all evtlists for an instrument have the same FILTER keyword
          do e=1,pevtlistlist%num_evtlists
            if ( trim(pevtlistlist%evtlist(e)%hduinfo%key_filter%value) .ne. trim(pevtlistlist%hduinfo%key_filter%value) ) then
              ierr =  ERROR_LEVEL_SEVERE
              call wrserr (subroutinename, ierr, &
                  "Mismatch between FILTER keywords in eventlists for inst: " // instlist%inst(i)%name)
              do ee=1,pevtlistlist%num_evtlists
                call wrserr(subroutinename, ierr, "file=" // trim(pevtlistlist%evtlist(ee)%hduinfo%filename) &
                    // " => FILTER=" // trim(pevtlistlist%evtlist(ee)%hduinfo%key_filter%value))
              end do
              return
            endif
          end do


          ! now search the evtlists to find the appropriate values of various keywords after merging
          ! TSTART, TSTOP, DATE-OBS, DATE-END
          ! TODO TODO TODO TODO TODO

          ! do a simple sorting - numeric for TSTART, TSTOP, alphabetic for DATE-OBS, DATE-END
          ! check that the indices of the first and last eventfiles agree as determined using
          ! TSTART/TSTOP versus DATE-OBS/DATE-END

          tstart_min = huge(tstart_min) ! +1.0d300   ! set to extreme values
          tstop_max  = -huge(tstop_max) ! -1.0d300
          date_obs_min = "9999-12-31T23:59:59.9"
          date_end_max = "0000-01-01T00:00:00.0"

          ! search for the min and max values within the eventlists
          do e=1,pevtlistlist%num_evtlists
            if ( pevtlistlist%evtlist(e)%hduinfo%key_tstart%value .lt. tstart_min ) then
              tstart_min = pevtlistlist%evtlist(e)%hduinfo%key_tstart%value
              tstart_min_index =  e
            endif
            if ( pevtlistlist%evtlist(e)%hduinfo%key_tstop%value .gt. tstop_max ) then
              tstop_max = pevtlistlist%evtlist(e)%hduinfo%key_tstop%value
              tstop_max_index =  e
            endif

            ! use string comparision for DATE-* keys
            if ( LLT(trim(pevtlistlist%evtlist(e)%hduinfo%key_date_obs%value),trim(date_obs_min)) ) then
              date_obs_min = trim(pevtlistlist%evtlist(e)%hduinfo%key_date_obs%value)
              date_obs_min_index =  e
            endif

            if ( LGT(trim(pevtlistlist%evtlist(e)%hduinfo%key_date_end%value),trim(date_end_max)) ) then
              date_end_max = trim(pevtlistlist%evtlist(e)%hduinfo%key_date_end%value)
              date_end_max_index =  e
            endif

          end do

          if ((tstart_min_index .ne.  date_obs_min_index) .or. &
              (tstop_max_index  .ne.  date_end_max_index)) then
            ierr = ERROR_LEVEL_SEVERE
            call wrserr (subroutinename, ierr, "Mismatch between TSTART/TSTOP vs DATE-OBS/DATE-END in eventlists")
            do e=1,pevtlistlist%num_evtlists
              write (str_buffer, '(A,I0,A,A)') "eventlist(", e, ")=", trim(pevtlistlist%evtlist(e)%hduinfo%filename)
              call wrschatty(subroutinename, str_buffer)
            end do
            do e=1,pevtlistlist%num_evtlists
              write (str_buffer,'(A,I0,A,A,A,I0,A,F20.3,A,F20.3,A,A,A,A)') &
                  "Inst(", i,  ")=", trim(instlist%inst(i)%name), &
                  " eventlist(", e, &
                  ") tstart=", pevtlistlist%evtlist(e)%hduinfo%key_tstart%value, &
                  " tstop=",  pevtlistlist%evtlist(e)%hduinfo%key_tstop%value, &
                  " date-obs=", trim(pevtlistlist%evtlist(e)%hduinfo%key_date_obs%value), &
                  " date-end=", trim(pevtlistlist%evtlist(e)%hduinfo%key_date_end%value)
              call wrschatty(subroutinename, str_buffer)
            end do

            call wrschatty(subroutinename, "(tstart_min_index,date_obs_min_index)=" &
                // trim(i2s(tstart_min_index)) // " " // trim(i2s(date_obs_min_index)))
            call wrschatty(subroutinename, "(tstop_max_index,date_end_max_index)=" &
                // trim(i2s(tstop_max_index)) // " " // trim(i2s(date_end_max_index)))

            return
          endif

          ! write the extremal values to the hduinfo structure
          call fitswrap_set_key(pevtlistlist%hduinfo%key_tstart, tstart_min)
          call fitswrap_set_key(pevtlistlist%hduinfo%key_tstop,  tstop_max)
          call fitswrap_set_key(pevtlistlist%hduinfo%key_date_obs, date_obs_min)
          call fitswrap_set_key(pevtlistlist%hduinfo%key_date_end, date_end_max)

        end if
      end if
    end do

  end subroutine srctool_evtlist_setup_merged_hduinfo
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! read in a set of eventlists and their contained GTIn, BADPIXn, CORRATTn, DEADCORn extensions
  subroutine srctool_evtlist_read_evtlists_from_file(instlist, filenames, num_filenames, ierr)
    character(len=*), parameter :: subroutinename='srctool_evtlist_read_evtlists_from_file'

    type(instlist_type), target, intent(inout)  :: instlist
    character(len=*),    intent(in)             :: filenames(MAX_EVTFILES)
    integer(kind=int32), intent(in)             :: num_filenames
    integer(kind=int32), intent(out)            :: ierr

    integer(kind=int32)        :: i !, j
    character(len=STR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    do i=1,num_filenames
      call srctool_evtlist_read_from_file(instlist, filenames(i), ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem reading evtlist from file: " // trim(filenames(i)) )
        return
      end if
    end do

    ! combine the hduinfo from each input evtlist in a sensible way
    call srctool_evtlist_setup_merged_hduinfo(instlist, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem creating merged evtlist headers" )
      return
    end if


    ! now merge the attlists for each instrument
    do i=1,MAX_INST
      if ( instlist%inst(i)%is_valid ) then
        call srctool_att_merge_attlists_for_inst(instlist%inst(i), ierr)
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          write (str_buffer, *) "Couldn't merge CORRATT lists for inst: ", trim(instlist%inst(i)%name)
          call wrserr (subroutinename, ierr, str_buffer)
          return
        end if
      end if
    end do


    ! now merge the DEADCOR lists for each instrument
    do i=1,MAX_INST
      if ( instlist%inst(i)%is_valid ) then
        call srctool_deadcor_merge_deadcors_for_inst(instlist%inst(i), ierr)
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          write (str_buffer, *) "Couldn't merge DEADCOR lists for inst: ", trim(instlist%inst(i)%name)
          call wrserr (subroutinename, ierr, str_buffer)
          return
        end if
      end if
    end do

    ! now merge the gtilists for each instrument
    do i=1,MAX_INST
      if ( instlist%inst(i)%is_valid ) then
        call srctool_gti_merge_gtilists_for_inst(instlist%inst(i), ierr)
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          write (str_buffer, *) "Couldn't merge *GTI lists for inst: ", trim(instlist%inst(i)%name)
          call wrserr (subroutinename, ierr, str_buffer)
          return
        end if
        ! If there are no GTI entries, then do not consider this instrument
        if ( .not. instlist%inst(i)%gtilist%is_valid ) then
           write (str_buffer, *) "Warning: disabling generation of outputs for inst ", &
                trim(instlist%inst(i)%name), " due to no GTI"
           call wrslog (subroutinename, str_buffer)
           instlist%inst(i)%to_consider = .false.
        end if
      end if
    end do
    call srctool_gti_merge_gtilists_for_instlist(instlist, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      write (str_buffer, *) "Couldn't merge *GTI lists for inst: ", trim(instlist%inst(0)%name)
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end if



  end subroutine srctool_evtlist_read_evtlists_from_file
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! read in an eventlist and any contained GTIn, BADPIXn, CORRATTn, DEADCORn extensions
  subroutine srctool_evtlist_read_from_file(instlist, filename, ierr)
    !------------------------------------------------------------
    ! Read in an eventlist file, allocate space to contain events, and populate with
    ! data. Read in supplementary extensions, and relevant header keywords.
    !------------------------------------------------------------
    character(len=*), parameter :: subroutinename='srctool_evtlist_read_from_file'

    type(instlist_type), intent(inout)      :: instlist
    character(len=*), intent(in)            :: filename
    integer(kind=int32), intent(out)        :: ierr

    integer(kind=int32)        :: status
!    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)        :: unit
    character(len=LSTR_MAX) :: filename_copy

    type(evtlist_type) :: evtlist     ! local evtlist

    status = 0
    ierr = ERROR_LEVEL_OK

    ! initialise the local events list
    call srctool_evtlist_empty(evtlist)

    ! Open the input event files - read only
    ! get an i/o unit
    call FTGIOU(unit, status)
    ! open the evt file and move to whichever extension is specified
    call FTNOPN(unit, filename, FITSIO_READONLY, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    call wrschatty (subroutinename, "Successfully opened eventlist file: " //  trim(filename))

    ! move to the EVENTS extension - if it exists
    call FTMNHD(unit, FITSIO_BINARY_TBL, "EVENTS", FITSIO_EXTVER_IGNORE, status)
    if (status .ne. 0) then  ! perhaps the EVENTS extensions does not exist?
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    ! now read a copy of the evtlist table into the local evtlist data structure
    call srctool_evtlist_read_from_chdu(unit, evtlist, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) return

    if ( evtlist%hduinfo%key_tstart%flag .ne. FITSWRAP_KEY_FLAG_BLANK ) then
      instlist%min_tstart = MIN(instlist%min_tstart,evtlist%hduinfo%key_tstart%value)
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "No TSTART keyword present in evtlist: " // trim(filename))
    end if
    if ( evtlist%hduinfo%key_tstop%flag .ne. FITSWRAP_KEY_FLAG_BLANK ) then
      instlist%max_tstop = MAX(instlist%max_tstop,evtlist%hduinfo%key_tstop%value)
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "No TSTOP keyword present in evtlist: " //  trim(filename) )
    end if

    !tidy up
    !close the events file
    call FTCLOS(unit, status)
    !release the unit number
    call FTFIOU(unit, status)

    ! combine the command line flagsel mask with the values in the evtfile keyword
    ! Flag selection is combined with a logical OR
    evtlist%key_flagsel%value = IOR(evtlist%key_flagsel%value, instlist%cl_flagsel)

    ! combine the pattern selection with a logical AND
    ! we special case 0 here, as 0 should mean "unfiltered", so we use our parameter value
    if( evtlist%key_pat_sel%value /= 0 ) then
       evtlist%key_pat_sel%value = IAND(evtlist%key_pat_sel%value, instlist%cl_pat_sel)
    else
       evtlist%key_pat_sel%value = instlist%cl_pat_sel
    end if

    ! now count the number of eventlists for each OBS_MODE
    select case ( trim(evtlist%hduinfo%key_obs_mode%value) )
    case ( "SURVEY" )
      instlist%num_survey_mode   = instlist%num_survey_mode + 1
    case ( "POINTING" )
      instlist%num_pointing_mode = instlist%num_pointing_mode + 1
    case ( "SLEW" )
      instlist%num_slew_mode     = instlist%num_slew_mode + 1
    case ( "SCAN" )
      instlist%num_scan_mode     = instlist%num_scan_mode + 1
    case default
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,"Unknown OBS_MODE: '" // &
          trim(evtlist%hduinfo%key_obs_mode%value) // "' in evtlist: " // trim(filename) )
      return
    end select


    !-------------------------------------------------------------------------------
    ! assign the local events to the relevant instrument eventlists
    call srctool_evtlist_divide_by_inst(instlist, evtlist, ierr)
    !-------------------------------------------------------------------------------

    ! free the local events structure ASAP
    call srctool_evtlist_empty(evtlist)

    !now read the metadata files from the evtfile
    ! try to trim off any bits of the filename which are fits HDU/selection expressions
    call FTRTNM(trim(filename), filename_copy, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    call srctool_cal_read_cal_from_file(instlist, filename_copy, instlist%gtitype_to_read, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) return

    call srctool_cal_read_cal_from_file(instlist, filename_copy, SRCTOOL_CALTYPE_BADPIX, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) return

    call srctool_cal_read_cal_from_file(instlist, filename_copy, SRCTOOL_CALTYPE_CORRATT, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) return

    call srctool_cal_read_cal_from_file(instlist, filename_copy, SRCTOOL_CALTYPE_DEADCOR, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) return

  end subroutine srctool_evtlist_read_from_file
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_evtlist_read_from_chdu(unit, evtlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_evtlist_read_from_chdu'

    integer(kind=int32), intent(in)    :: unit
    type(evtlist_type),  intent(inout) :: evtlist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32) :: status
!    character(len=STR_MAX) :: str_buffer

    integer(kind=int32) :: colnum_time
    integer(kind=int32) :: colnum_ra
    integer(kind=int32) :: colnum_dec
    integer(kind=int32) :: colnum_x
    integer(kind=int32) :: colnum_y
    integer(kind=int32) :: colnum_pi         ! Used for PI or ENERGY
    integer(kind=int32) :: colnum_ev_weight
    integer(kind=int32) :: colnum_rawx
    integer(kind=int32) :: colnum_rawy
    integer(kind=int32) :: colnum_subx
    integer(kind=int32) :: colnum_suby
    integer(kind=int32) :: colnum_pha
    integer(kind=int32) :: colnum_pat_typ
    integer(kind=int32) :: colnum_pat_inf
    integer(kind=int32) :: colnum_tm_nr    ! used for TM_NR or CCDNR
    integer(kind=int32) :: colnum_flag

    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    character(len=STR_MAX) :: colname_pi
    character(len=STR_MAX) :: colname_tm_nr
    logical             :: col_exist_flag

    integer(kind=int8) , parameter :: nullval_int8  = 0
    integer(kind=int16), parameter :: nullval_int16  = 0
    integer(kind=int32), parameter :: nullval_int32  = 0
    integer(kind=int64), parameter :: nullval_int64  = 0
    real(kind=real32)  , parameter :: nullval_real32  = 0.0
    real(kind=real64)  , parameter :: nullval_real64  = 0.0d0

    character(len=16) :: key_name

    ierr = ERROR_LEVEL_OK
    status = 0

    ! get the hdu info/keys
!    call fitswrap_init_hduinfo(evtlist%hduinfo)
    call fitswrap_read_hduinfo(unit, evtlist%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    call wrschatty (subroutinename, "Reading evtlist from: " // trim(evtlist%hduinfo%filename))

!    ! determine the OBS_MODE
!    if ( evtlist%hduinfo%key_obs_mode%flag .ne. FITSWRAP_KEY_FLAG_SUPPLIED ) then
!      call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
!          "Warning, OBS_MODE keyword was not supplied in eventlist. Assuming default value (SURVEY)")
!      call fitswrap_set_key(evtlist%hduinfo%key_obs_mode, "SURVEY")
!    endif
!
!    str_buffer = trim(evtlist%hduinfo%key_obs_mode%value)
!    ! first capitalise the string
!    call to_upper(str_buffer)
!    ! now match against list of valid options
!    select case ( TRIM(str_buffer))
!      case ( "SURVEY" )
!        evtlist%obs_mode = OBS_MODE_SURVEY
!      case ( "POINTING" )
!        evtlist%obs_mode = OBS_MODE_POINTING
!      case ( "SLEW" )
!        evtlist%obs_mode = OBS_MODE_SLEW
!      case ( "SCAN" )
!        evtlist%obs_mode = OBS_MODE_SCAN
!      case default
!        ierr=ERROR_LEVEL_SEVERE
!        call wrserr(subroutinename,ierr,"Unknown OBS_MODE description: " // &
!            TRIM(evtlist%hduinfo%key_obs_mode%value))
!        return
!    end select
!    call wrschatty (subroutinename, "Eventlist has OBS_MODE= " // trim(OBS_MODE_STR(evtlist%obs_mode)))


!    call  fitswrap_set_key(evtlist%key_flagsel, 0)
    call  fitswrap_read_key(unit, evtlist%key_flagsel ,ierr)
    if ((ierr .ne. ERROR_LEVEL_OK) .or. &
        (evtlist%key_flagsel%flag .ne. FITSWRAP_KEY_FLAG_SUPPLIED) ) then
!      write (*,'(A)'), "Setting FLAGSEL to default"
      call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
          "Warning, FLAGSEL keyword was not supplied in eventlist. Assuming default value (0x00)")
      call  fitswrap_set_key(evtlist%key_flagsel, 0)
      ierr = ERROR_LEVEL_OK
    end if
    ! initialise the pattern mask to all 1s
    call  fitswrap_read_key(unit, evtlist%key_pat_sel, ierr)
    if ((ierr .ne. ERROR_LEVEL_OK) .or. &
        (evtlist%key_pat_sel%flag .ne. FITSWRAP_KEY_FLAG_SUPPLIED) ) then
      call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
          "Warning, PAT_SEL keyword was not supplied in eventlist. Assuming default value (0x0F)")
      call  fitswrap_set_key(evtlist%key_pat_sel ,INT(1+2+4+8,int16))
      ierr = ERROR_LEVEL_OK
    end if

    if ( evtlist%key_pat_sel%value .eq. 0 ) then
      call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
          "Warning, PAT_SEL keyword value is exactly zero, i.e. no patterns are valid for Eventlist: " // &
          trim(evtlist%hduinfo%filename))
    endif

    !get the number of rows in the table extension
    call FTGNRW(unit, evtlist%num_rows, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    ! check that have at least some rows, and if so read in the column data
    if ( evtlist%num_rows .gt. 0 ) then
      call wrschatty (subroutinename, "Eventlist contains: " // trim(i2s(evtlist%num_rows)) // " rows")



      ! work out if we should read PI or ENERGY column
      colname_pi = "PI"
      call fitswrap_test_for_col(unit, colname_pi, col_exist_flag)
!      call FTGCNO(unit, FITSIO_CASEINSENS, trim(colname_pi), col_pi_energy, status)
!      if ((status .ne. 0) .or. (col_pi_energy <= 0) ) then
      if ( .not. col_exist_flag ) then
!        status = 0
        colname_pi = "ENERGY"
        call fitswrap_test_for_col(unit, colname_pi, col_exist_flag)
!        call FTGCNO(unit, FITSIO_CASEINSENS, trim(colname_pi), col_pi_energy, status)

        if ( .not. col_exist_flag ) then
!        if ((status .ne. 0) .or. (col_pi_energy <= 0)) then
!          if (status .ne. 0) then
!            call fitswrap_write_fits_error(subroutinename, status)
!          end if
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, &
              "Problem finding a 'PI' (or 'ENERGY') column in eventlist:" // trim(evtlist%hduinfo%filename) )
          return
        else
          ierr = ERROR_LEVEL_WARNING
          call wrserr (subroutinename, ierr, &
              "Warning. Eventlist file contains deprecated 'ENERGY' column (expected 'PI')" )
!              // trim(evtlist%hduinfo%filename))
          ierr = ERROR_LEVEL_OK
        end if
      end if
      call wrschatty (subroutinename, "Reading event energy information from column: " // trim(colname_pi) )
      !


      ! https://wiki.mpe.mpg.de/eRosita/EroCat/eSASSdevel/FormChanges
      ! allow for older event files that call this column 'TM_NR'
      colname_tm_nr =  EVTLIST_INST_COL
      call fitswrap_test_for_col(unit, colname_tm_nr, col_exist_flag)
      if ( .not. col_exist_flag ) then
        colname_tm_nr = EVTLIST_INST_COL_OLD   ! i.e. "CCDNR"
        call fitswrap_test_for_col(unit, colname_tm_nr, col_exist_flag)
        if ( .not. col_exist_flag ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, &
              "Problem finding a '" // EVTLIST_INST_COL // "' (or '" // EVTLIST_INST_COL_OLD // &
              "') column in eventlist:" // trim(evtlist%hduinfo%filename) )
          return
        else
          call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
                      "Warning. Eventlist file contains the deprecated '" // EVTLIST_INST_COL_OLD // &
                      "' column (expected '" // EVTLIST_INST_COL // "').")
        endif
      endif
      call wrschatty (subroutinename, "Reading telescope module index from column: " // trim(colname_tm_nr) )

      call fitswrap_prep_col(unit, "TIME"             , evtlist%num_rows, colnum_time     , evtlist%ptime     , ierr)
      call fitswrap_prep_col(unit, "RA"               , evtlist%num_rows, colnum_ra       , evtlist%pra       , ierr)
      call fitswrap_prep_col(unit, "DEC"              , evtlist%num_rows, colnum_dec      , evtlist%pdec      , ierr)
      call fitswrap_prep_col(unit, "X"                , evtlist%num_rows, colnum_x        , evtlist%px        , ierr)
      call fitswrap_prep_col(unit, "Y"                , evtlist%num_rows, colnum_y        , evtlist%py        , ierr)
      call fitswrap_prep_col(unit, trim(colname_pi)   , evtlist%num_rows, colnum_pi       , evtlist%penergy   , ierr)
      call fitswrap_prep_col(unit, "EV_WEIGHT"        , evtlist%num_rows, colnum_ev_weight, evtlist%pev_weight, ierr)
      call fitswrap_prep_col(unit, "RAWX"             , evtlist%num_rows, colnum_rawx     , evtlist%prawx     , ierr)
      call fitswrap_prep_col(unit, "RAWY"             , evtlist%num_rows, colnum_rawy     , evtlist%prawy     , ierr)
      call fitswrap_prep_col(unit, "SUBX"             , evtlist%num_rows, colnum_subx     , evtlist%psubx     , ierr)
      call fitswrap_prep_col(unit, "SUBY"             , evtlist%num_rows, colnum_suby     , evtlist%psuby     , ierr)
      call fitswrap_prep_col(unit, "PHA"              , evtlist%num_rows, colnum_pha      , evtlist%ppha      , ierr)
      call fitswrap_prep_col(unit, "PAT_TYP"          , evtlist%num_rows, colnum_pat_typ  , evtlist%ppat_typ  , ierr)
      call fitswrap_prep_col(unit, "PAT_INF"          , evtlist%num_rows, colnum_pat_inf  , evtlist%ppat_inf  , ierr)
      call fitswrap_prep_col(unit, trim(colname_tm_nr), evtlist%num_rows, colnum_tm_nr    , evtlist%ptm_nr    , ierr)
      call fitswrap_prep_col(unit, "FLAG"             , evtlist%num_rows, colnum_flag     , evtlist%pflag     , ierr)
      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if

      ! get wcs (used when writing event files)
      write(key_name, '(A,I1)') 'TCRVL', colnum_x
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%rval_x, ierr)
      write(key_name, '(A,I1)') 'TCRVL', colnum_y
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%rval_y, ierr)
      write(key_name, '(A,I1)') 'TCRPX', colnum_x
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%rpix_x, ierr)
      write(key_name, '(A,I1)') 'TCRPX', colnum_y
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%rpix_y, ierr)
      write(key_name, '(A,I1)') 'TCDLT', colnum_x
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%delta_x, ierr)
      write(key_name, '(A,I1)') 'TCDLT', colnum_y
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%delta_y, ierr)
      write(key_name, '(A,I1)') 'TCTYP', colnum_x
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%type_x, ierr)
      write(key_name, '(A,I1)') 'TCTYP', colnum_y
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%type_y, ierr)
      write(key_name, '(A,I1)') 'TCUNI', colnum_x
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%unit_x, ierr)
      write(key_name, '(A,I1)') 'TCUNI', colnum_y
      call fitswrap_read_key_quick(unit, key_name, evtlist%wcs%unit_y, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) return

      ! get the chunk size
      call FTGRSZ(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      do row = 1, evtlist%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,evtlist%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVD(unit, colnum_time,     row,1,nrows_to_read,nullval_real64,evtlist%ptime(row:row_end),     anyf,status)
        call FTGCVD(unit, colnum_ra,       row,1,nrows_to_read,nullval_real64,evtlist%pra(row:row_end),       anyf,status)
        call FTGCVD(unit, colnum_dec,      row,1,nrows_to_read,nullval_real64,evtlist%pdec(row:row_end),      anyf,status)
        call FTGCVJ(unit, colnum_x,        row,1,nrows_to_read,nullval_int32, evtlist%px(row:row_end),        anyf,status)
        call FTGCVJ(unit, colnum_y,        row,1,nrows_to_read,nullval_int32, evtlist%py(row:row_end),        anyf,status)
        call FTGCVE(unit, colnum_pi       ,row,1,nrows_to_read,nullval_real32,evtlist%penergy(row:row_end),   anyf,status)
        call FTGCVE(unit, colnum_ev_weight,row,1,nrows_to_read,nullval_real32,evtlist%pev_weight(row:row_end),anyf,status)
        call FTGCVI(unit, colnum_rawx,     row,1,nrows_to_read,nullval_int16, evtlist%prawx(row:row_end),     anyf,status)
        call FTGCVI(unit, colnum_rawy,     row,1,nrows_to_read,nullval_int16, evtlist%prawy(row:row_end),     anyf,status)
        call FTGCVI(unit, colnum_subx,     row,1,nrows_to_read,nullval_int16, evtlist%psubx(row:row_end),     anyf,status)
        call FTGCVI(unit, colnum_suby,     row,1,nrows_to_read,nullval_int16, evtlist%psuby(row:row_end),     anyf,status)
        call FTGCVI(unit, colnum_pha,      row,1,nrows_to_read,nullval_int16, evtlist%ppha(row:row_end),      anyf,status)
        call FTGCVI(unit, colnum_pat_typ,  row,1,nrows_to_read,nullval_int16, evtlist%ppat_typ(row:row_end),  anyf,status)
        call FTGCVB(unit, colnum_pat_inf,  row,1,nrows_to_read,nullval_int8,  evtlist%ppat_inf(row:row_end),  anyf,status)
        call FTGCVB(unit, colnum_tm_nr,    row,1,nrows_to_read,nullval_int8,  evtlist%ptm_nr(row:row_end),    anyf,status)
        call FTGCVJ(unit, colnum_flag,     row,1,nrows_to_read,nullval_int32, evtlist%pflag(row:row_end),     anyf,status)

        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do

    else
      call wrslog (subroutinename, "Warning: no events found in extension")
    end if
    evtlist%is_valid = .true.
  end subroutine srctool_evtlist_read_from_chdu
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  ! divide events by instrument
  subroutine srctool_evtlist_divide_by_inst(instlist, evtlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_evtlist_divide_by_inst'

    type(instlist_type), intent(inout), target :: instlist
    type(evtlist_type),  intent(inout) :: evtlist
    integer(kind=int32), intent(out)   :: ierr

    type(evtlist_type),  pointer :: pevtlist

    integer(kind=int32) :: i, j
    integer(kind=int32) :: evtlist_index(0:MAX_INST)
    integer(kind=int32) :: evtlist_nevents(0:MAX_INST)
    integer(kind=int32) :: n, this_filter_code
    character(len=STR_MAX) :: str_buffer

    ierr  = ERROR_LEVEL_OK
    call wrschatty (subroutinename, "Dividing events by instrument")

    evtlist_nevents(0:MAX_INST) = 0
    ! work out the index of each destination evtlist
    do i=0, MAX_INST
       evtlist_index(i) = instlist%inst(i)%evtlistlist%num_evtlists + 1
    end do

    ! count the events for each instrument
    do j=1,evtlist%num_rows
      i = evtlist%ptm_nr(j)
      if ( (i .ge. 1) .and. (i .le. MAX_INST )) then
        evtlist_nevents(i) = evtlist_nevents(i) + 1
      else
        ! emit a warning
      end if
    end do

    !now allocate space for the event arrays
    do i=1,MAX_INST
      if ( (evtlist_nevents(i) .gt. 0) .and. instlist%inst(i)%to_consider ) then
        pevtlist => instlist%inst(i)%evtlistlist%evtlist(evtlist_index(i))
        call srctool_evtlist_allocate_arrays(pevtlist, &
            evtlist_nevents(i), ierr)
        nullify(pevtlist)
!        call srctool_evtlist_allocate_arrays(instlist%inst(i)%evtlistlist%evtlist(evtlist_index(i)), &
!            evtlist_nevents(i), ierr)
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          return
        end if

      end if
    end do

    ! copy some of the file metadata from the original list into the new lists
    ! note that we will over-write this info later for any instruments which did not receive events
    ! from this eventlist
    do i=1,MAX_INST
      pevtlist => instlist%inst(i)%evtlistlist%evtlist(evtlist_index(i))
      pevtlist%hduinfo = evtlist%hduinfo
      pevtlist%wcs = evtlist%wcs
      pevtlist%key_flagsel = evtlist%key_flagsel
      pevtlist%key_pat_sel = evtlist%key_pat_sel

      ! overwrite the instrume keyword to the appropriate value
      write (pevtlist%hduinfo%key_instrume%value, '(A,I0)') INST_STEM, i

      pevtlist%obs_mode = evtlist%obs_mode

      nullify(pevtlist)
    end do


    ! now distribute the events among the instruments
    do j=1,evtlist%num_rows
      i = evtlist%ptm_nr(j)
      if ( (i .ge. 1) .and. (i .le. MAX_INST )) then
        pevtlist => instlist%inst(i)%evtlistlist%evtlist(evtlist_index(i))
        if ( instlist%inst(i)%to_consider ) then
!          n = instlist%inst(i)%evtlistlist%evtlist(evtlist_index(i))%num_rows + 1
          n = pevtlist%num_rows + 1
          if ( n .gt. size(pevtlist%ptime) ) then
            ! emit error
            ierr = ERROR_LEVEL_SEVERE
            return
          end if
          pevtlist%ptime(n)      = evtlist%ptime(j)
          pevtlist%pra(n)        = evtlist%pra(j)
          pevtlist%pdec(n)       = evtlist%pdec(j)
          pevtlist%px(n)         = evtlist%px(j)
          pevtlist%py(n)         = evtlist%py(j)
          pevtlist%penergy(n)    = evtlist%penergy(j)
          pevtlist%pev_weight(n) = evtlist%pev_weight(j)
          pevtlist%prawx(n)      = evtlist%prawx(j)
          pevtlist%prawy(n)      = evtlist%prawy(j)
          pevtlist%psubx(n)      = evtlist%psubx(j)
          pevtlist%psuby(n)      = evtlist%psuby(j)
          pevtlist%ppha(n)       = evtlist%ppha(j)
          pevtlist%ppat_typ(n)   = evtlist%ppat_typ(j)
          pevtlist%ppat_inf(n)   = evtlist%ppat_inf(j)
          pevtlist%ptm_nr(n)     = evtlist%ptm_nr(j)
          pevtlist%pflag(n)      = evtlist%pflag(j)
!          pevtlist%pclass_mask(n) = evtlist%pclass_mask(j)

          pevtlist%num_rows = n
        end if
        nullify(pevtlist)
      end if
    end do

    ! increment the num_evtlists counter for each instrument which actually received events
    ! mark those as valid too

    ! update the filter_code for any instruments which received events and for the merged instrument
    this_filter_code = srctool_misc_filter_string_to_code(evtlist%hduinfo%key_filter%value)

    do i=1,MAX_INST
      pevtlist => instlist%inst(i)%evtlistlist%evtlist(evtlist_index(i))
      if ( (pevtlist%num_rows .gt. 0) .and. instlist%inst(i)%to_consider ) then
        instlist%inst(i)%is_valid = .true.
        pevtlist%is_valid = .true.
        instlist%inst(i)%evtlistlist%num_evtlists = evtlist_index(i)

        if ( instlist%inst(i)%filter_code .eq. ESASS_FILTER_CODE_NULL ) then
          instlist%inst(i)%filter_code = this_filter_code
        else if ( instlist%inst(i)%filter_code .ne. this_filter_code ) then
          instlist%inst(i)%filter_code = ESASS_FILTER_CODE_INDEF  ! mixed filters!
        endif

      end if
      nullify(pevtlist)
    end do


    if ( instlist%inst(0)%filter_code .eq. ESASS_FILTER_CODE_NULL ) then
      instlist%inst(0)%filter_code = this_filter_code
    else if ( instlist%inst(0)%filter_code .ne. this_filter_code ) then
      instlist%inst(0)%filter_code = ESASS_FILTER_CODE_INDEF  ! mixed filters!
    endif

    write (str_buffer, '(A6,1x,A8,1x,A10,1x,A10,1x,A10)') "Inst", "Consider", "Nevts", "NevtRead", "Nevtlists"
    call wrschatty (subroutinename, str_buffer)
    do i=1,MAX_INST
      if ( instlist%inst(i)%is_valid ) then
        write (str_buffer, '(A6,1x,A8,1x,I10,1x,I10,1x,I10)') &
            trim(instlist%inst(i)%name), &
            (logical2string(instlist%inst(i)%to_consider)), &
            evtlist_nevents(i), &
            instlist%inst(i)%evtlistlist%evtlist(evtlist_index(i))%num_rows, &
            instlist%inst(i)%evtlistlist%num_evtlists
        call wrschatty (subroutinename, str_buffer)
      end if
    end do


  end subroutine srctool_evtlist_divide_by_inst
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine srctool_evtlist_print_evt_headerline()
    character(len=*), parameter :: subroutinename='srctool_evtlist_print_evt'
    character(len=STR_MAX) :: str_buffer
    write (str_buffer, '(A6,1x,A5,1x,A30,1x,A8,1x,A6,1x,A3,1x,A12,1x,2(A10,1x) (A8,1x),A8,1x,A5)') &
        "#EVENT", "CLASS",&
        "FILENAME", "EXT",&
        "INDEX","CCD","TIME",&
        "RA","DEC",&
        "X","Y",&
        "RAWX","RAWY",&
        "ENERGY","PHA"
    call wrschatty (subroutinename, str_buffer)
  end subroutine srctool_evtlist_print_evt_headerline

  subroutine srctool_evtlist_print_evt(evtlist, index)
    character(len=*), parameter :: subroutinename='srctool_evtlist_print_evt'
    type(evtlist_type),  intent(in)    :: evtlist
    integer(kind=int32), intent(in)    :: index
    character(len=STR_MAX) :: str_buffer

    if ( (index .gt. evtlist%num_rows)  .or. (index .lt. 1) ) then
      call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
          "Error with supplied event index: " // trim(i2s(index)) // " (num_rows=" // trim(i2s(evtlist%num_rows)) // ")"  )
    else
      write (str_buffer, '(A6,1x,A5,1x,A30,1x,A8,1x,I6,1x,I3,1x,F12.4,1x,2(F10.6,1x) (I8,1x),F8.1,1x,I5)') &
          "EVENT", EVT_CLASS_FLAG_STR(evtlist%pclass_mask(index)), &
          trim(evtlist%hduinfo%filename),trim(evtlist%hduinfo%key_extname%value),&
          index,evtlist%ptm_nr(index),evtlist%ptime(index),&
          evtlist%pra(index),evtlist%pdec(index),&
          evtlist%px(index),evtlist%py(index),&
          evtlist%prawx(index),evtlist%prawy(index),&
          evtlist%penergy(index),evtlist%ppha(index)
      call wrschatty (subroutinename, str_buffer)
    end if
  end subroutine srctool_evtlist_print_evt
  !---------------------------------------------------------------------------------------------

end module srctool_evtlist
