! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_extmap

  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines handling extmaps or source masks
  !
  !-----------------------------------------------------------------------------

  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface

  use srctool_generic
  use srctool_generic_fitswrap

  implicit none

  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_EXTMAP_MOD'

  public

  type :: extmap_type
     type(fitskey_int32_type)  :: key_naxis
     type(fitskey_int32_type)  :: key_naxis1
     type(fitskey_int32_type)  :: key_naxis2
     type(fitskey_real64_type) :: key_crpix1
     type(fitskey_real64_type) :: key_crpix2
     type(fitskey_real64_type) :: key_crval1
     type(fitskey_real64_type) :: key_crval2
     type(fitskey_real64_type) :: key_pc1_1
     type(fitskey_real64_type) :: key_pc1_2
     type(fitskey_real64_type) :: key_pc2_1
     type(fitskey_real64_type) :: key_pc2_2
     type(fitskey_real64_type) :: key_cd1_1
     type(fitskey_real64_type) :: key_cd1_2
     type(fitskey_real64_type) :: key_cd2_1
     type(fitskey_real64_type) :: key_cd2_2
     type(fitskey_real64_type) :: key_cdelt1
     type(fitskey_real64_type) :: key_cdelt2
     type(fitskey_string_type) :: key_ctype1
     type(fitskey_string_type) :: key_ctype2
     type(fitskey_string_type) :: key_cunit1
     type(fitskey_string_type) :: key_cunit2

     type(wcs_type) :: wcs

     ! the naxis1,2 values are copied here from key_naxis1,2 keys for ease of access:
     integer(kind=int32) :: naxis1, naxis2

     ! 2d array to hold the map (note it's a little wasteful to use
     ! real64 for a mask, but we'd have to duplicate the code
     ! otherwise)
     real(kind=real64), dimension(:,:), allocatable :: ppix

     !    ! index of the pixels that bracket the non-zero pixels in the map
     integer(kind=int32) :: pix_x_min, pix_x_max
     integer(kind=int32) :: pix_y_min, pix_y_max

     real(kind=real64) :: original_image_sum

     real(kind=real64) :: src_x, src_y ! location of centre of source in image

     real(kind=real64) :: max_radius_deg  ! radius of non-zero pixels from src
  end type extmap_type

  public :: srctool_extmap_empty
  public :: srctool_extmap_read_from_file
  public :: srctool_extmap_update_source_coord

contains

  ! Empty extmap structure
  subroutine srctool_extmap_empty(extmap)
    type(extmap_type), intent(inout) :: extmap
    logical, parameter :: is_compulsory = .true.

    call fitswrap_init_key(extmap%key_naxis,       "NAXIS", is_compulsory)
    call fitswrap_init_key(extmap%key_naxis1,      "NAXIS1", is_compulsory)
    call fitswrap_init_key(extmap%key_naxis2,      "NAXIS2", is_compulsory)
    call fitswrap_init_key(extmap%key_crpix1,      "CRPIX1", is_compulsory)
    call fitswrap_init_key(extmap%key_crpix2,      "CRPIX2", is_compulsory)
    call fitswrap_init_key(extmap%key_crval1,      "CRVAL1", is_compulsory)
    call fitswrap_init_key(extmap%key_crval2,      "CRVAL2", is_compulsory)
    call fitswrap_init_key(extmap%key_pc1_1,       "PC1_1")
    call fitswrap_init_key(extmap%key_pc1_2,       "PC1_2")
    call fitswrap_init_key(extmap%key_pc2_1,       "PC2_1")
    call fitswrap_init_key(extmap%key_pc2_2,       "PC2_2")
    call fitswrap_init_key(extmap%key_cd1_1,       "CD1_1")
    call fitswrap_init_key(extmap%key_cd1_2,       "CD1_2")
    call fitswrap_init_key(extmap%key_cd2_1,       "CD2_1")
    call fitswrap_init_key(extmap%key_cd2_2,       "CD2_2")
    call fitswrap_init_key(extmap%key_ctype1,      "CTYPE1", is_compulsory)
    call fitswrap_init_key(extmap%key_ctype2,      "CTYPE2", is_compulsory)
    call fitswrap_init_key(extmap%key_cdelt1,      "CDELT1")
    call fitswrap_init_key(extmap%key_cdelt2,      "CDELT2")
    call fitswrap_init_key(extmap%key_cunit1,      "CUNIT1", is_compulsory)
    call fitswrap_init_key(extmap%key_cunit2,      "CUNIT2", is_compulsory)

    extmap%naxis1 = 0
    extmap%naxis2 = 0

    extmap%pix_x_min = -1
    extmap%pix_x_max = -1
    extmap%pix_y_min = -1
    extmap%pix_y_max = -1
    extmap%src_x = -huge(extmap%src_x)
    extmap%src_y = -huge(extmap%src_y)
    extmap%original_image_sum = 0.0d0
    extmap%max_radius_deg = -huge(extmap%max_radius_deg)

    call wcs_init(extmap%wcs)
    if (allocated(extmap%ppix))    deallocate(extmap%ppix)

  end subroutine srctool_extmap_empty

  ! Read extmap from file
  subroutine srctool_extmap_read_from_file(filename, extmap, normalise, ierr)

    character(len=*) , intent(in) :: filename
    type(extmap_type), intent(inout), target  :: extmap
    logical, intent(in) :: normalise
    integer(kind=int32) , intent(out)  :: ierr

    character(len=*), parameter :: subroutinename='srctool_extmap_read_from_file'

    integer(kind=int32)        :: unit
    integer(kind=int32)        :: status
    type(wcs_type), pointer    :: wcs
    integer(kind=int32)        :: num_pci_j, num_cdi_j, num_cdelti
    ! character(len=STR_MAX) :: str_buffer
    !    integer(kind=int32) :: bitpix
    integer(kind=int32) :: anyf
    integer(kind=int32), parameter :: group=0
    real(kind=real64), parameter   :: nullval = 0.0d0

    ! debug
    !    real(kind=real64) :: long, lat, src_x, src_y
    status = 0
    ierr = ERROR_LEVEL_OK
    wcs => extmap%wcs

    call srctool_extmap_empty(extmap)

    ! Open the files - read only
    ! get an i/o unit
    call FTGIOU(unit, status)
    ! open the file and move to whichever extension is specified or to first image extension
    call FTIOPN(unit, filename, FITSIO_READONLY, status)
    if (status /= 0) then
       call fitswrap_write_fits_error(subroutinename, status)
       ierr = ERROR_LEVEL_SEVERE
       return
    end if

    call wrschatty (subroutinename, "Reading extmap/mask from: " // trim(filename))

    ! We require all of the following to be present (these have already been flagged as 'compulsory')
    call fitswrap_read_key(unit, extmap%key_naxis ,ierr)
    call fitswrap_read_key(unit, extmap%key_naxis1 ,ierr)
    call fitswrap_read_key(unit, extmap%key_naxis2 ,ierr)
    call fitswrap_read_key(unit, extmap%key_crpix1 ,ierr)
    call fitswrap_read_key(unit, extmap%key_crpix2 ,ierr)
    call fitswrap_read_key(unit, extmap%key_crval1 ,ierr)
    call fitswrap_read_key(unit, extmap%key_crval2 ,ierr)
    call fitswrap_read_key(unit, extmap%key_ctype1 ,ierr)
    call fitswrap_read_key(unit, extmap%key_ctype2 ,ierr)
    call fitswrap_read_key(unit, extmap%key_cunit1 ,ierr)
    call fitswrap_read_key(unit, extmap%key_cunit2 ,ierr)

    ! we require one of the following sets of keywords to be present
    call fitswrap_read_key(unit, extmap%key_pc1_1  ,ierr)
    call fitswrap_read_key(unit, extmap%key_pc1_2  ,ierr)
    call fitswrap_read_key(unit, extmap%key_pc2_1  ,ierr)
    call fitswrap_read_key(unit, extmap%key_pc2_2  ,ierr)
    call fitswrap_read_key(unit, extmap%key_cdelt1 ,ierr)
    call fitswrap_read_key(unit, extmap%key_cdelt2 ,ierr)

    call fitswrap_read_key(unit, extmap%key_cd1_1  ,ierr)
    call fitswrap_read_key(unit, extmap%key_cd1_2  ,ierr)
    call fitswrap_read_key(unit, extmap%key_cd2_1  ,ierr)
    call fitswrap_read_key(unit, extmap%key_cd2_2  ,ierr)

    ! this will check that all compulsory keywords were supplied
    if ( ierr /= ERROR_LEVEL_OK ) then
       call wrserr(subroutinename,ierr,"Problem reading keywords from extmap/mask file: " // trim(filename))
       return
    endif

    ! check that the CTYPEi are ok
    if ((trim(extmap%key_ctype1%value) == "RA---TAN") .and. &
         (trim(extmap%key_ctype2%value) == "DEC--TAN") ) then
       wcs%projtype = PROJTYPE_CODE_TAN
       call wrschatty (subroutinename, "Working with projtype: TAN")
    else if ((trim(extmap%key_ctype1%value) == "RA---SIN") .and. &
         (trim(extmap%key_ctype2%value) == "DEC--SIN") ) then
       wcs%projtype = PROJTYPE_CODE_SIN
       call wrschatty (subroutinename, "Working with projtype: SIN")
    else
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename,ierr,&
            "Unsupported coordinate system CTYPE1,CTYPE2 = " &
            // trim(extmap%key_ctype1%value) // "," // trim(extmap%key_ctype2%value))
       return
    endif

    ! check that the CUNITi are ok
    if ((trim(extmap%key_cunit1%value) /= "deg") .or. &
         (trim(extmap%key_cunit2%value) /= "deg") ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename,ierr,&
            "Unsupported coordinate units CUNIT1,CUNIT2 = -->" &
            // trim(extmap%key_cunit1%value) // "," // trim(extmap%key_cunit2%value) // "<--")
       return
    endif

    ! make copies for convenience
    extmap%naxis1 = extmap%key_naxis1%value
    extmap%naxis2 = extmap%key_naxis2%value

    ! now check that have at least one PCi_j or CDi_j keyword
    num_pci_j = &
         b2i(extmap%key_pc1_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) + &
         b2i(extmap%key_pc1_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED) + &
         b2i(extmap%key_pc2_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) + &
         b2i(extmap%key_pc2_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED)

    num_cdi_j = &
         b2i(extmap%key_cd1_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) + &
         b2i(extmap%key_cd1_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED) + &
         b2i(extmap%key_cd2_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) + &
         b2i(extmap%key_cd2_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED)

    num_cdelti = &
         b2i(extmap%key_cdelt1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) + &
         b2i(extmap%key_cdelt2%flag == FITSWRAP_KEY_FLAG_SUPPLIED)

    call wrschatty (subroutinename, "Num PCi_j=" // trim(i2s(num_pci_j)) // &
         " Num CDi_j=" // trim(i2s(num_cdi_j)) )

    if ( (num_pci_j >= 1) .and. (num_cdi_j >= 1) ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename,ierr,&
            "You supplied both PCi_j and CDi_j keywords in header of " // trim(filename))
       return

    else if ( num_cdi_j >= 1 ) then
       ! any CDi_j keywords
       wcs%cdi_j(:,:) = 0.0d0
       ! note that any elements of CDi_j will default to 0 if not supplied
       if (extmap%key_cd1_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) wcs%cdi_j(1,1) = extmap%key_cd1_1%value
       if (extmap%key_cd1_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED) wcs%cdi_j(1,2) = extmap%key_cd1_2%value
       if (extmap%key_cd2_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) wcs%cdi_j(2,1) = extmap%key_cd2_1%value
       if (extmap%key_cd2_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED) wcs%cdi_j(2,2) = extmap%key_cd2_2%value

    else
       ! this is reached if no CDi_j keywords are set
       if ( num_cdelti /= 2 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr(subroutinename, ierr, &
               "You didn't supply CDELT1,CDELT2 in header of " // trim(filename))
          return
       end if

       wcs%cdi_j(:,:) = 0.0d0
       if (is_zero(extmap%key_cdelt1%value) .or. &
            is_zero(extmap%key_cdelt2%value) ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr(subroutinename, ierr, &
               "Invalid (zero) value for CDELT1,CDELT2 keyword(s) in header of " // trim(filename))
          return
       end if

       ! note that elements of PCi_j default to 1 for on-diagonal elements and 0 for off-diagonal elements
       if (extmap%key_pc1_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) then
          wcs%cdi_j(1,1) = extmap%key_pc1_1%value * extmap%key_cdelt1%value
       else
          wcs%cdi_j(1,1) = extmap%key_cdelt1%value
       end if

       if (extmap%key_pc1_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED) then
          wcs%cdi_j(1,2) = extmap%key_pc1_2%value * extmap%key_cdelt1%value
       end if

       if (extmap%key_pc2_1%flag == FITSWRAP_KEY_FLAG_SUPPLIED) then
          wcs%cdi_j(2,1) = extmap%key_pc2_1%value * extmap%key_cdelt2%value
       end if

       if (extmap%key_pc2_2%flag == FITSWRAP_KEY_FLAG_SUPPLIED) then
          wcs%cdi_j(2,2) = extmap%key_pc2_2%value * extmap%key_cdelt2%value
       else
          wcs%cdi_j(2,2) = extmap%key_cdelt2%value
       end if
    end if

    ! load the wcs_type structure
    wcs%cdi_j(:,:) = wcs%cdi_j(:,:)
    ! take inverse of matrix =>
    ! [ a  b ]^-1  =  1/(ad - bc) * [ d  -b ]
    ! [ c  d ]                      [-c   a ]
    wcs%inv_cdi_j(1,1) =        wcs%cdi_j(2,2)
    wcs%inv_cdi_j(1,2) = -1.0d0*wcs%cdi_j(1,2)
    wcs%inv_cdi_j(2,1) = -1.0d0*wcs%cdi_j(2,1)
    wcs%inv_cdi_j(2,2) =        wcs%cdi_j(1,1)
    wcs%inv_cdi_j(:,:) = wcs%inv_cdi_j(:,:)/ &
         (wcs%cdi_j(1,1)*wcs%cdi_j(2,2) - wcs%cdi_j(1,2)*wcs%cdi_j(2,1))

    wcs%crpix(1) = extmap%key_crpix1%value
    wcs%crpix(2) = extmap%key_crpix2%value
    wcs%crval(1) = extmap%key_crval1%value
    wcs%crval(2) = extmap%key_crval2%value

    wcs%cos_crval2 = cosd(wcs%crval(2))
    wcs%sin_crval2 = sind(wcs%crval(2))

    !    write (str_buffer, '(A,F15.7,x,F15.7,A)') &
    !        "WCS MATRIX CDi_j    = [", wcs%cdi_j(1,1), wcs%cdi_j(1,2), "]"
    !    call wrschatty (subroutinename, str_buffer)
    !    write (str_buffer, '(A,F15.7,x,F15.7,A)') &
    !        "WCS                   [", wcs%cdi_j(2,1), wcs%cdi_j(2,2), "]"
    !    call wrschatty (subroutinename, str_buffer)
    !
    !    write (str_buffer, '(A,F15.7,x,F15.7,A)') &
    !        "WCS MATRIX invCDi_j = [", wcs%inv_cdi_j(1,1), wcs%inv_cdi_j(1,2), "]"
    !    call wrschatty (subroutinename, str_buffer)
    !    write (str_buffer, '(A,F15.7,x,F15.7,A)') &
    !        "WCS                   [", wcs%inv_cdi_j(2,1), wcs%inv_cdi_j(2,2), "]"
    !    call wrschatty (subroutinename, str_buffer)
    !    write (*, '(A,A,x,ES15.7E2)') subroutinename, ": WCS CRPIX1 = ", wcs%crpix(1)
    !    write (*, '(A,A,x,ES15.7E2)') subroutinename, ": WCS CRPIX2 = ", wcs%crpix(2)
    !    write (*, '(A,A,x,ES15.7E2)') subroutinename, ": WCS CRVAL1 = ", wcs%crval(1)
    !    write (*, '(A,A,x,ES15.7E2)') subroutinename, ": WCS CRVAL2 = ", wcs%crval(2)

    !    ! read the datatype of the image
    !    call FTGIDT(unit, bitpix, status)
    !    if (status .ne. 0) then
    !      ierr = ERROR_LEVEL_SEVERE
    !      call wrserr (subroutinename, ierr, "Could not read datatype of " // trim(filename))
    !      return
    !    end if
    !    call wrschatty (subroutinename, "Bitpix of extmap is: " // i2s(bitpix))



    ! now allocate some memory for the image pixel data
    if ((extmap%naxis1 .gt. 1) .and. &
         (extmap%naxis2 .gt. 1) ) then
       allocate(extmap%ppix(extmap%naxis1,extmap%naxis2), STAT = status)

       if ( check_allocate_status(subroutinename, status, ierr) ) return
    else
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Strange image dimensions: " // i2s(extmap%naxis1) // "x" // i2s(extmap%naxis2) )
       return
    endif

    ! Read the pixel data as double precision reals
    call FTG2DD(unit,group,nullval,extmap%naxis1,extmap%naxis1,extmap%naxis2, extmap%ppix(:,:),anyf,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Problem reading pixel data from " // trim(filename))
       return
    end if

    !tidy up
    !close the input file
    call FTCLOS(unit, status)
    !release the unit number
    call FTFIOU(unit, status)

    ! make sure that there are no negative values in the image
    extmap%ppix(:,:) = MAX(0.0d0, extmap%ppix(:,:))

    extmap%original_image_sum = SUM(extmap%ppix(:,:))

    if ( normalise ) then
       if ( extmap%original_image_sum .gt. 0.0d0 ) then
          !      write (str_buffer, '(E12.6)') extmap%original_image_sum
          call wrschatty (subroutinename, &
               "Normalising extmap pixel values to unity => divide image by factor: " // &
               trim(r2s(extmap%original_image_sum)))
          extmap%ppix(:,:) = extmap%ppix(:,:) / extmap%original_image_sum
       else
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Sum of pixels is zero in extmap " // trim(filename))
          return
       end if
    end if

  end subroutine srctool_extmap_read_from_file

  ! Update the values in the extmap based on the source coordinates
  subroutine srctool_extmap_update_source_coord(extmap, src_coord)
    type(extmap_type), intent(inout), target  :: extmap
    type(coord_type) , intent(in)  :: src_coord

    character(len=*), parameter :: subroutinename='srctool_extmap_update_source'

    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i, j
    real(kind=real64) :: dist, pix_lat, pix_long

    call wcs_world2pix(extmap%wcs, src_coord%long, src_coord%lat, extmap%src_x, extmap%src_y)
    write (str_buffer, '(2(F15.9,x),A,2(F15.7,x))') &
         src_coord%long, src_coord%lat, " -> (X,Y)= ", extmap%src_x, extmap%src_y
    call wrschatty (subroutinename, &
         "Computed source pixel location from sky coords: (RA,DEC)= " // trim(str_buffer) )

    !DEBUG     ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG
    !DEBUG     ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG
    !DEBUG     ! now compute the coords of the source in the image
    !DEBUG     call RANDOM_SEED()
    !DEBUG     do i=1,30
    !DEBUG       call RANDOM_NUMBER(long)
    !DEBUG       call RANDOM_NUMBER(lat)
    !DEBUG       long = src_coord%long + extmap%naxis1*extmap%key_cdelt1%value*(long - 0.5d0)
    !DEBUG       lat  = src_coord%lat  + extmap%naxis2*extmap%key_cdelt2%value*(lat  - 0.5d0)
    !DEBUG       call wcs_world2pix(extmap%wcs, long, lat, src_x, src_y)
    !DEBUG       write (str_buffer, '(2(F15.9,x),A10,2(F15.7,x))') long, lat, " -> (X,Y)= ", src_x, src_y
    !DEBUG       call wrschatty (subroutinename, &
    !DEBUG           "Testing wcs_world2pix: (RA,DEC)= " // trim(str_buffer) )
    !DEBUG     end do
    !DEBUG
    !DEBUG     do i=1,5
    !DEBUG       do j=1,5
    !DEBUG         src_x = extmap%naxis1*0.2d0*(i-0.5d0)
    !DEBUG         src_y = extmap%naxis2*0.2d0*(j-0.5d0)
    !DEBUG         call wcs_pix2world(extmap%wcs, src_x, src_y, long, lat)
    !DEBUG         write (str_buffer, '(2(F15.7,x),A10,2(F15.9,x))') src_x, src_y, " -> (RA,DEC)= ", long, lat
    !DEBUG         call wrschatty (subroutinename, &
    !DEBUG             "Testing wcs_pix2world: (X,Y)=    " // trim(str_buffer) )
    !DEBUG        end do
    !DEBUG     end do
    !DEBUG     ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG
    !DEBUG     ! DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG DEBUG


    ! compute index of the pixels that bracket the non-zero pixels in the map
    call find_nonzero_pixel_bounds_real64 (extmap%ppix, 0.0d0, extmap%pix_x_min, &
         extmap%pix_x_max, extmap%pix_y_min, extmap%pix_y_max)
    call wrschatty (subroutinename, &
         "Non-zero pixel bounds: [x_min,x_max][y_min,y_max] = [" // &
         trim(i2s(extmap%pix_x_min)) // "," // &
         trim(i2s(extmap%pix_x_max)) // "][" // &
         trim(i2s(extmap%pix_y_min)) // "," // &
         trim(i2s(extmap%pix_y_max)) // "]" )


    ! compute radius of most distance non-zero pixel from source position
    ! scan along each of the bounding edges
    extmap%max_radius_deg = 0.0d0

    ! scan along the left and right hand bounds
    do i=extmap%pix_x_min,extmap%pix_x_max,(extmap%pix_x_max-extmap%pix_x_min)
       do j=extmap%pix_y_min,extmap%pix_y_max
          if ( extmap%ppix(i,j) .gt. 0.0d0 ) then
             call wcs_pix2world(extmap%wcs, i+0.5d0, j+0.5d0, pix_lat, pix_long)
             dist = calc_ang_sep(pix_lat, pix_long, src_coord%long, src_coord%lat)
             if ( dist .gt.  extmap%max_radius_deg ) then
                !            write (str_buffer, '(A,2(I8,x),2(F15.9,x),2(F15.10,x))') &
                !                "DEBUG max_radius: ", i, j, pix_lat, pix_long, extmap%max_radius_deg, dist
                !            call wrschatty (subroutinename, str_buffer)
                extmap%max_radius_deg = dist
             endif
          endif
       end do
    end do

    ! scan along the top and bottom bounds
    do j=extmap%pix_y_min,extmap%pix_y_max,(extmap%pix_y_max-extmap%pix_y_min)
       do i=extmap%pix_x_min,extmap%pix_x_max
          if ( extmap%ppix(i,j) .gt. 0.0d0 ) then
             call wcs_pix2world(extmap%wcs, i+0.5d0, j+0.5d0, pix_lat, pix_long)
             dist = calc_ang_sep(pix_lat, pix_long, src_coord%long, src_coord%lat)
             if ( dist .gt.  extmap%max_radius_deg ) then
                !            write (str_buffer, '(A,2(I8,x),2(F15.9,x),2(F15.10,x))') &
                !                "DEBUG max_radius: ", i, j, pix_lat, pix_long, extmap%max_radius_deg, dist
                !            call wrschatty (subroutinename, str_buffer)
                extmap%max_radius_deg = dist
             endif
          endif
       end do
    end do

    !    write (str_buffer, '(ES12.6E2)') extmap%max_radius_deg
    call wrschatty (subroutinename, &
         "Most distant non-zero pixel is: " // trim(r2s(extmap%max_radius_deg)) // &
         " degrees from source centre")

  end subroutine srctool_extmap_update_source_coord
  !---------------------------------------------------------------------------------------------

end module srctool_extmap
