! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_generic

!-----------------------------------------------------------------------------
! Author: T. Dwelly (dwelly@mpe.mpg.de)
!
! First version: 2013-02-19
! Current version: $Revision$
! CVS Date: $Date$
!-----------------------------------------------------------------------------
!
!-----------------------------------------------------------------------------
! Summary:
! A module containing very basic parameters and subroutines/functions
! - renamed from common_mod.f90 on 20.06.2014
!-----------------------------------------------------------------------------

  use iso_fortran_env ! replacement for numkind
  use param           ! standard eROSITA module handling parameter/logging/error interface

  implicit none     ! insist on explicit typing

!-------------------------------------------------------------------------------
  character(len=*), private, parameter :: str_modname='SRCTOOL_GENERIC_MOD'

!old  !!!!! This must be manually updated for each release  !!!!!!!!!!!!!!!!
!old  character(len=*), parameter :: CREATOR_STRING='eSASS-SRCTOOL v1.21'
!old  !!!!! This must be updated !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
  character(len=64) :: CREATOR_STRING   ! This variable is now set in set_creator_string() function,
                                        ! which is called at top of srctool_main.f90
                                        ! - piggybacks on the auto-generated version.for function

  public

  ! The following was used as the default kind of logicals. It was
  ! previously set to 1.  We now use the default logical kind, but
  ! change this to 4 for compatibility (4 is the default for
  ! gfortran).
  integer(kind=int32), parameter :: BOOL=4

  integer(kind=int32), parameter :: STR_MAX=256     ! normal string buffer length
  integer(kind=int32), parameter :: LSTR_MAX=1024   ! suitable for filenames etc
!  integer(kind=int32), parameter :: VLSTR_MAX=4096   ! suitable for long comments etc
  integer(kind=int32), parameter :: VLSTR_MAX=16384   ! suitable for long comments etc
!  real(kind=real32),   parameter :: NULLVAL_REAL32 = -9.999999e35
!  real(kind=real64),   parameter :: NULLVAL_REAL64 = -9.999999d300
  real(kind=real32),   parameter :: NULLVAL_REAL32 = -huge(1.0e0)
  real(kind=real64),   parameter :: NULLVAL_REAL64 = -huge(1.0d0)

  integer(kind=int32), parameter :: MAX_INST=7
  character(len=*)   , parameter :: INSTRUME_IS_MERGED  = "MERGED"

  real(kind=real64),   parameter :: PI             = abs(dacos(-1.0d0))
  real(kind=real64),   parameter :: DEG2RAD        = PI/180.0d0
  real(kind=real64),   parameter :: RAD2DEG        = 180.0d0/PI
  real(kind=real64),   parameter :: DEG2ARCMIN     = 60.0d0
  real(kind=real64),   parameter :: DEG2ARCSEC     = 3600.0d0
  real(kind=real64),   parameter :: ARCMIN2DEG     = 1.0d0/60.0d0
  real(kind=real64),   parameter :: ARCSEC2DEG     = 1.0d0/3600.0d0
  real(kind=real64),   parameter :: ARCMIN2ARCSEC  = 60.0d0
  real(kind=real64),   parameter :: ARCSEC2ARCMIN  = 1.0d0/60.0d0
  real(kind=real64),   parameter :: HOUR2DEG       = 15.0d0
  real(kind=real64),   parameter :: MINUTE2DEG     = 15.0d0/60.0d0
  real(kind=real64),   parameter :: SECOND2DEG     = 15.0d0/3600.0d0
  real(kind=real64),   parameter :: SQDEG2SQARCMIN = DEG2ARCMIN**2
  real(kind=real64),   parameter :: SQDEG2SQARCSEC = DEG2ARCSEC**2
  real(kind=real64),   parameter :: SQARCMIN2SQDEG = 1.0d0/(DEG2ARCMIN**2)
  real(kind=real64),   parameter :: SQARCSEC2SQDEG = 1.0d0/(DEG2ARCSEC**2)
  real(kind=real64),   parameter :: UM2METRES      = 1.0d-6
  real(kind=real64),   parameter :: METRES2UM      = 1.0d6
  real(kind=real64),   parameter :: EV2KEV         = 1.0d-3
  real(kind=real64),   parameter :: KEV2EV         = 1.0d3

  !standard wr error levels
  integer(kind=int32), parameter :: ERROR_LEVEL_OK      = 0
  integer(kind=int32), parameter :: ERROR_LEVEL_WARNING = 1
  integer(kind=int32), parameter :: ERROR_LEVEL_ERROR   = 2
  integer(kind=int32), parameter :: ERROR_LEVEL_SEVERE  = 3

  integer(kind=int32), parameter :: COORDSYS_CODE_ERROR   =-1  ! Error interpretting
  integer(kind=int32), parameter :: COORDSYS_CODE_NULL    = 0  ! Undefined
  integer(kind=int32), parameter :: COORDSYS_CODE_ICRS    = 1  ! anything given as "fk5" or "J2000" is assumed to actually be in ICRS
  integer(kind=int32), parameter :: COORDSYS_CODE_FK4     = 2  ! aka B1950
  integer(kind=int32), parameter :: COORDSYS_CODE_GAL     = 3  ! Galactic (epoch=2000)
  integer(kind=int32), parameter :: COORDSYS_CODE_ECL     = 4  ! Ecliptic (epoch=2000)
  integer(kind=int32), parameter :: COORDSYS_CODE_IMAGE   = 5  ! Pixels within a sky map
  integer(kind=int32), parameter :: COORDSYS_CODE_PHYSICAL= 6  ! Pixels within a CCD
  integer(kind=int32), parameter :: COORDSYS_CODE_UNITVEC = 7  ! unit vectors x,y,z


  character(len=*), parameter :: COORDSYS_STR(COORDSYS_CODE_ERROR:COORDSYS_CODE_UNITVEC) &
      = (/ "ERROR   ",  "NULL    ", "ICRS    ", "FK4     ", "GAL     ", &
           "ECL     ",  "IMAGE   ", "PHYSICAL", "UNITVEC " /)

  integer(kind=int32), parameter :: PROJTYPE_CODE_NULL    = 0  ! Undefined
  integer(kind=int32), parameter :: PROJTYPE_CODE_TAN     = 1  !
  integer(kind=int32), parameter :: PROJTYPE_CODE_SIN     = 2  !
!  integer(kind=int32), parameter :: PROJTYPE_CODE_*       = 3  !

  character(len=*), parameter :: PROJTYPE_STR(PROJTYPE_CODE_NULL:PROJTYPE_CODE_SIN) &
      = (/ "NULL", "TAN ", "SIN " /)



  character(len=*), parameter :: LIST_OF_UPPERCASE_LETTERS  = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
  character(len=*), parameter :: LIST_OF_LOWERCASE_LETTERS  = 'abcdefghijklmnopqrstuvwxyz'
  character(len=*), parameter :: LIST_OF_DECIMAL_DIGITS     = '0123456789'
  character(len=*), parameter :: LIST_OF_OCTAL_DIGITS       = '01234567'
  character(len=*), parameter :: LIST_OF_HEXADECIMAL_DIGITS = '0123456789ABCDEF'
  character(len=*), parameter :: LIST_OF_BINARY_DIGITS      = '01'

  character(len=*), parameter :: INST_STEM = 'TM'
  character(len=*), parameter :: INST_STEM_OLD = 'FM'

  character(len=*), parameter :: EVTLIST_INST_COL = 'TM_NR'
  character(len=*), parameter :: EVTLIST_INST_COL_OLD = 'CCDNR'



  integer(kind=int32),parameter :: BINARY_SEARCH_STATUS_ERROR = -1
  integer(kind=int32),parameter :: BINARY_SEARCH_STATUS_OK    =  0
  integer(kind=int32),parameter :: BINARY_SEARCH_STATUS_LTMIN =  1
  integer(kind=int32),parameter :: BINARY_SEARCH_STATUS_GTMAX =  2

  ! subroutine declarations and interfaces for module: srctool_generic
  public :: set_creator_string
  public :: convert_bitmask_string_to_int32
  public :: convert_bitmask_string_to_int16
  public :: convert_bitmask_string_to_int64
  public :: convert_string_array_to_string
  public :: convert_int32_array_to_string
  public :: convert_logical_array_to_string
  public :: convert_real32_array_to_string
  public :: convert_real64_array_to_string
  public :: to_upper, to_upper_fun
  public :: to_lower, to_lower_fun
  public :: calc_destination
  public :: calc_destination_fast
  public :: calc_destination_faster
  public :: interpolate_long_lat
  public :: interpolate_long_lat_fast
  private :: insertion_sort_real64
  private :: insertion_sort_real32
  private :: insertion_sort_int32
  private :: insertion_sort_int64
  private :: insertion_cosort_real64_real64
  private :: insertion_cosort_real64_real32
  private :: insertion_cosort_real64_real64_real64
  private :: insertion_cosort_real64_real64_real64_real64
  private :: insertion_cosort_3real64_2int32
  private :: insertion_cosort_4real64_2int32
  !public :: QSort
  public :: filter_unique_real64
  public :: filter_unique_real32
  public :: filter_unique_int32
  public :: convert_longitude_to_decimal
  public :: convert_latitude_to_decimal
  public :: convert_longlat_to_decimal
  public :: longlat2xy
  public :: longlat2xy_fast
  public :: longlat2xy_faster
  public :: longlats2xys
  public :: longlats2xys_fast
  public :: xy2longlat
  public :: xys2longlats
  public :: pathtools_extract_tail
  public :: make_necessary_dirs_for_new_file
  public :: bilinear_interp_real64
  public :: trilinear_interp_real64
  public :: get_interp_indices_real64
  public :: get_interp_indices_real32
  public :: get_interp_indices_weights_real64
  public :: empty_linear_interp_real64
  public :: setup_linear_interp_real64
  public :: evaluate_linear_interp_real64
  public :: search_and_linear_interp_real64_real64
  public :: search_and_linear_interp_real64_real64_real64
  public :: binary_search_real64
  public :: binary_search_real32
  public :: binary_search_with_hint_real64
  public :: geom_ra_dec_to_gal_l_b
  public :: geom_gal_l_b_to_ra_dec
  public :: geom_convert_coordsys_longlat_longlat
  public :: geom_convert_coordsys_coord_coord
  public :: geom_convert_coordsys_coord_longlat
  public :: geom_convert_coordsys_longlat_coord
  public :: srctool_generic_init_coord
  public :: check_allocate_status

  ! write-safe versions of wrlog, wrerr, wrchatty
  public :: wrslog
  public :: wrserr
  public :: wrschatty

  private :: i2s_int8
  private :: i2s_int16
  private :: i2s_int32
  private :: i2s_int64
  public  :: b2i
  private :: r2s_real64
  private :: r2s_real32
  public  :: logical2yesno

  public :: calc_ang_sep
  public :: calc_ang_sep_fast
  public :: calc_haversine_ang_sep

  private :: wcs_iworld2world_sin
  private :: wcs_iworld2world_tan
  private :: wcs_world2iworld_sin
  private :: wcs_world2iworld_tan
  private :: wcs_pix2iworld
  private :: wcs_iworld2pix
  public  :: wcs_pix2world
  public  :: wcs_world2pix
  public  :: wcs_calc_pixarea_atpixel
  public  :: wcs_init

  public  :: find_nonzero_pixel_bounds_real64

  public :: squash_string

  interface i2s
    module procedure i2s_int8
    module procedure i2s_int16
    module procedure i2s_int32
    module procedure i2s_int64
  end interface i2s

  interface r2s
    module procedure r2s_real64
    module procedure r2s_real32
  end interface r2s

  interface r2fs
    module procedure r2fs_real64
    module procedure r2fs_real32
  end interface r2fs

  interface convert_array_to_string
     module procedure convert_int32_array_to_string
     module procedure convert_string_array_to_string
     module procedure convert_real32_array_to_string
     module procedure convert_real64_array_to_string
  end interface

  interface convert_bitmask_string_to_int
     module procedure convert_bitmask_string_to_int16
     module procedure convert_bitmask_string_to_int32
     module procedure convert_bitmask_string_to_int64
  end interface

  interface is_zero
    module procedure is_zero_real32
    module procedure is_zero_real64
  end interface

  interface eq_real
    module procedure eq_real32
    module procedure eq_real64
  end interface

  interface near_real
    module procedure near_real32
    module procedure near_real64
  end interface

  interface insertion_sort
    module procedure insertion_sort_int32
    module procedure insertion_sort_int64
    module procedure insertion_sort_real32
    module procedure insertion_sort_real64
  end interface

  interface insertion_cosort
    module procedure insertion_cosort_real64_real64
    module procedure insertion_cosort_real64_real32
    module procedure insertion_cosort_real64_real64_real64
    module procedure insertion_cosort_real64_real64_real64_real64
    module procedure insertion_cosort_3real64_2int32
    module procedure insertion_cosort_4real64_2int32
  end interface

  interface filter_unique
    module procedure filter_unique_real64
    module procedure filter_unique_real32
    module procedure filter_unique_int32
  end interface


  interface trilinear_interp
    module procedure trilinear_interp_real64
  end interface

  interface get_interp_indices
    module procedure get_interp_indices_real64
    module procedure get_interp_indices_real32
  end interface

  interface get_interp_indices_weights
    module procedure get_interp_indices_weights_real64
  end interface

  interface empty_linear_interp
    module procedure empty_linear_interp_real64
  end interface
  interface setup_linear_interp
    module procedure setup_linear_interp_real64
  end interface
  interface evaluate_linear_interp
    module procedure evaluate_linear_interp_real64
  end interface

  interface geom_convert_coordsys
    module procedure  geom_convert_coordsys_longlat_longlat
    module procedure  geom_convert_coordsys_coord_coord
    module procedure  geom_convert_coordsys_coord_longlat
    module procedure  geom_convert_coordsys_longlat_coord
  end interface

  interface search_and_linear_interp
    module procedure search_and_linear_interp_real64_real64
    module procedure search_and_linear_interp_real64_real64_real64
  end interface search_and_linear_interp

  interface binary_search
    module procedure binary_search_real64
    module procedure binary_search_real32
  end interface

  interface n2s
    module procedure n2s_int32
    module procedure n2s_int64
  end interface

  interface expand_array
     module procedure expand_array_real64
  end interface expand_array

  type QSort_group
    integer(kind=int64) :: order    ! original order of unsorted data
    real(kind=real64) :: value       ! values to be sorted by
  end type QSort_group


  !-------------------------------------------------------------------------------
  type linear_interp_real64_type
    integer(kind=int32) :: n_ref        ! number of reference measurements
    integer(kind=int32) :: n_eval       ! number of evaluations
    real(kind=real64),  dimension(:), allocatable :: y_eval    ! interpolated outputs
    integer(kind=int32),dimension(:), allocatable :: index1, index2
    real(kind=real64),  dimension(:), allocatable :: weight1, weight2
  end type linear_interp_real64_type
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  type :: coord_type
    integer(kind=int32) :: coordsys_code    ! applies to long,lat only
    real(kind=real64)   :: long ! exact meaning is coordsys-specific
    real(kind=real64)   :: lat  !

    real(kind=real64)   :: x    ! sky map coords
    real(kind=real64)   :: y    !

    real(kind=real64)   :: rawx    ! physical pixel coords (fractional)
    real(kind=real64)   :: rawy    !
  end type coord_type
  !-------------------------------------------------------------------------------



  !- simple structure holding WCS parameters - minimum needed for TAN, SIN etc projections
  !- cdi_j is assumed to be scaled into units of degrees
  type :: wcs_type
    integer(kind=int32)  :: projtype  ! encodes the projection - see PROJTYPE_CODE_* options above
    real(kind=real64), dimension(1:2,1:2) :: cdi_j ! interpreted rotation+transformation matrix, includes cdelti factors

    real(kind=real64), dimension(1:2,1:2) :: inv_cdi_j ! inverse of cdi_j(:,:)
    real(kind=real64)  :: crpix(1:2)  ! reference pixel
    real(kind=real64)  :: crval(1:2)  ! reference wcs
    real(kind=real64)  :: sin_crval2  ! sine of reference wcs latitude
    real(kind=real64)  :: cos_crval2  ! cosine of reference wcs latitude
  end type wcs_type


contains

  !-----------------------------------------------
  subroutine set_creator_string()
    character(len=64) :: version
    CREATOR_STRING = trim(version())
  end subroutine set_creator_string
  !-----------------------------------------------


  !-----------------------------------------------
  ! call like this:
  !  if ( check_allocate_status(subroutinename, status, ierr) ) return

  recursive function check_allocate_status (subroutinename, status, ierr) result(flag)
    character(len=*), intent(in) ::  subroutinename
    integer(kind=int32), intent(in) :: status
    integer(kind=int32), intent(out)  :: ierr
    logical :: flag
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(trim(subroutinename), ierr, &
          "Could not allocate memory (error code=" //  trim(i2s(status)) // ")" )
      flag = .true.
    else
      flag = .false.
    end if
  end function check_allocate_status
  !-----------------------------------------------


  !-----------------------------------------------
  pure recursive function cosd (x)
    real(kind=real64), intent(in)   :: x
    real(kind=real64)               :: cosd
    cosd = cos(x*DEG2RAD)
  end function cosd

  pure recursive function sind (x)
    real(kind=real64), intent(in)   :: x
    real(kind=real64)               :: sind
    sind = sin(x*DEG2RAD)
  end function sind

  pure recursive function tand (x)
    real(kind=real64), intent(in)   :: x
    real(kind=real64)               :: tand
    tand = tan(x*DEG2RAD)
  end function tand

  pure recursive function acosd (x)
    real(kind=real64), intent(in)   :: x
    real(kind=real64)               :: acosd
    acosd = acos(x)*RAD2DEG
  end function acosd

  pure recursive function asind (x)
    real(kind=real64), intent(in)   :: x
    real(kind=real64)               :: asind
    asind = asin(x)*RAD2DEG
  end function asind

  pure recursive function atand (x)
    real(kind=real64), intent(in)   :: x
    real(kind=real64)               :: atand
    atand = atan(x)*RAD2DEG
  end function atand

  pure recursive function atan2d (y, x)
    real(kind=real64), intent(in)   :: x, y
    real(kind=real64)               :: atan2d
    atan2d = atan2(y, x)*RAD2DEG
  end function atan2d
  !-----------------------------------------------

  !-----------------------------------------------
  pure recursive function n2s_int32(n) result(s)
    integer(kind=int32), intent(in)   :: n
    character(len=1)     :: s
    if ( n .eq. 1 ) then
      s = " "
    else
      s = "s"
    endif
  end function n2s_int32
  !-----------------------------------------------
  pure recursive function n2s_int64(n) result(s)
    integer(kind=int64), intent(in)   :: n
    character(len=1)     :: s
    if ( n .eq. 1 ) then
      s = " "
    else
      s = "s"
    endif
  end function n2s_int64
  !-----------------------------------------------

  !-----------------------------------------------
  ! real variable comparison functions
  pure recursive function eq_real32 (x,y)
    real(kind=real32), intent(in)   :: x,y
    logical :: eq_real32
!    if (abs(x-y) .lt. precision(x)*max(abs(x),abs(y))) then
!    if (abs(x-y) .le. 1.0d-30*max(abs(x),abs(y))) then
    if (abs(x-y) .le. 1.0e-7*max(abs(x),abs(y))) then
      eq_real32 = .true.
    else
      eq_real32 = .false.
    endif
  end function eq_real32

  pure recursive function eq_real64 (x,y)
    real(kind=real64), intent(in)   :: x,y
    logical :: eq_real64
!    if (abs(x-y) .lt. precision(x)*max(abs(x),abs(y))) then
!    if (abs(x-y) .le. 1.0d-30*max(abs(x),abs(y))) then
    if (abs(x-y) .le. 1.0d-15*max(abs(x),abs(y))) then
      eq_real64 = .true.
    else
      eq_real64 = .false.
    endif
  end function eq_real64
  !-----------------------------------------------


  !-----------------------------------------------
  pure recursive function near_real32 (x,y,frac_tol)
    real(kind=real32), intent(in)   :: x,y,frac_tol
    logical :: near_real32
    if (abs(x-y) .le. frac_tol*max(abs(x),abs(y))) then
      near_real32 = .true.
    else
      near_real32 = .false.
    endif
  end function near_real32

  pure recursive function near_real64 (x,y,frac_tol)
    real(kind=real64), intent(in)   :: x,y,frac_tol
    logical :: near_real64
    if (abs(x-y) .le. frac_tol*max(abs(x),abs(y))) then
      near_real64 = .true.
    else
      near_real64 = .false.
    endif
  end function near_real64
  !-----------------------------------------------



  pure recursive function is_zero_real32 (x)
    real(kind=real32), intent(in)   :: x
    logical :: is_zero_real32
!    if ( abs(x) .le. tiny(x) ) then
    if ( abs(x) .le. epsilon(x) ) then
      is_zero_real32 = .true.
    else
      is_zero_real32 = .false.
    endif
  end function is_zero_real32

 pure recursive function is_zero_real64 (x)
    real(kind=real64), intent(in)   :: x
    logical :: is_zero_real64
    if ( abs(x) .le. epsilon(x) ) then
!      write (*,'(2(x,A,x,E15.7))'), "abs(x)=", abs(x), "epsilon=", epsilon(x)
      is_zero_real64 = .true.
    else
      is_zero_real64 = .false.
    endif
  end function is_zero_real64
  !-----------------------------------------------


  ! some basic simple functions

  pure recursive function logical2char (flag)
    logical, intent(in)  :: flag
    character :: logical2char
    if ( flag ) then
      logical2char = "T"
    else
      logical2char = "F"
    endif
  end function logical2char

  pure recursive function logical2string (flag)
    logical, intent(in)  :: flag
    character(len=5) :: logical2string
    if ( flag ) then
      logical2string = "TRUE"
    else
      logical2string = "FALSE"
    endif
  end function logical2string

  !----------------------------------------------

  pure recursive function logical2yesno(flag)
    logical, intent(in)  :: flag
    character(len=3) :: logical2yesno
    if ( flag ) then
      logical2yesno = "yes"
    else
      logical2yesno = "no"
    endif
  end function logical2yesno
  !----------------------------------------------



  ! convert a string like 6756567 or 0x17f3 or o172743 or b1010001111 to an integer
  ! assume a big endian format
  recursive subroutine convert_bitmask_string_to_int32(str_bitmask, bitmask, ierr)
    character(len=*), parameter :: subroutinename='convert_bitmask_string_to_int32'
    character(len=*),intent(in) ::  str_bitmask
    integer(kind=int32), intent(out) :: bitmask
    integer(kind=int32), intent(out) :: ierr
    character(len=STR_MAX) :: str_buffer
    character(len=STR_MAX) :: str_buffer2
    integer(kind=int32) :: i,j,base,ic
    integer(kind=int32), parameter :: max_digits_binary = 31
    integer(kind=int32), parameter :: max_digits_octal  = 11
    integer(kind=int32), parameter :: max_digits_decimal = 10
    integer(kind=int32), parameter :: max_digits_hexadecimal = 8
    integer(kind=int32) :: max_digits
    character(len=STR_MAX) :: valid_digits
    ierr = ERROR_LEVEL_OK
    bitmask = 0
    if ( LEN_TRIM(ADJUSTL(str_bitmask)) .gt. STR_MAX ) then
      ierr = ERROR_LEVEL_SEVERE
      return
    endif

    str_buffer = TRIM(ADJUSTL(str_bitmask))

    if ( LEN_TRIM(str_buffer) .lt. 1 ) then
      ierr = ERROR_LEVEL_SEVERE
      return
    endif

    ! first non-blank character(s) must be '0-9', 'b', '0x', 'x', or 'o'
    !---------------------------------------------------
    if ( str_buffer(1:1) .eq. 'b' ) then   ! treat as a binary number
      valid_digits = TRIM(LIST_OF_BINARY_DIGITS)
      base = 2
      max_digits = max_digits_binary
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( str_buffer(1:1) .eq. 'o' ) then   ! treat as an octal number
      valid_digits = TRIM(LIST_OF_OCTAL_DIGITS)
      base = 8
      max_digits = max_digits_octal
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( (str_buffer(1:1) .eq. 'x') ) then ! treat as a hexadecimal number
      valid_digits = TRIM(LIST_OF_HEXADECIMAL_DIGITS)
      base = 16
      max_digits = max_digits_hexadecimal
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( (str_buffer(1:2) .eq. '0x') ) then ! treat as a hexadecimal number
      valid_digits = TRIM(LIST_OF_HEXADECIMAL_DIGITS)
      base = 16
      max_digits = max_digits_hexadecimal
      str_buffer2 = TRIM(str_buffer(3:))
    else
      ic = INDEX(LIST_OF_DECIMAL_DIGITS, str_buffer(1:1))
      if ( ic .gt. 0 ) then ! treat as an integer
        valid_digits = TRIM(LIST_OF_DECIMAL_DIGITS)
        base = 10
        max_digits = max_digits_decimal
        str_buffer2 = TRIM(str_buffer)
      else
        ierr = ERROR_LEVEL_SEVERE
        return
      endif
   endif

   if ( LEN_TRIM(str_buffer2) .lt. 1 ) then
     ierr = ERROR_LEVEL_SEVERE
     return
   endif

   call to_upper(str_buffer2) ! deal with upper case characters only

   if ( LEN_TRIM(str_buffer2) .gt. max_digits) then ! too many bits!
     ierr = ERROR_LEVEL_SEVERE
     return
   endif
   j = 0
   do i=LEN_TRIM(str_buffer2),1,-1
     ic = INDEX(valid_digits,str_buffer2(i:i))
     if ( ic .gt. 0 ) then
       bitmask = bitmask + INT((ic-1)*(base**j),int32)
     else
       ierr = ERROR_LEVEL_SEVERE
       return
     endif
     j = j + 1
   end do

  end subroutine convert_bitmask_string_to_int32

  !same as above, but for int16
  recursive subroutine convert_bitmask_string_to_int16(str_bitmask, bitmask, ierr)
    character(len=*), parameter :: subroutinename='convert_bitmask_string_to_int16'
    character(len=*),intent(in) ::  str_bitmask
    integer(kind=int16), intent(out) :: bitmask
    integer(kind=int32), intent(out) :: ierr
    character(len=STR_MAX) :: str_buffer
    character(len=STR_MAX) :: str_buffer2
    integer(kind=int32) :: i,j,base,ic
    integer(kind=int32), parameter :: max_digits_binary = 15
    integer(kind=int32), parameter :: max_digits_octal  = 6
    integer(kind=int32), parameter :: max_digits_decimal = 5
    integer(kind=int32), parameter :: max_digits_hexadecimal = 4
    integer(kind=int32) :: max_digits
    character(len=STR_MAX) :: valid_digits
    ierr = ERROR_LEVEL_OK
    bitmask = 0
    if ( LEN_TRIM(ADJUSTL(str_bitmask)) .gt. STR_MAX ) then
      ierr = ERROR_LEVEL_SEVERE
      return
    endif

    str_buffer = TRIM(ADJUSTL(str_bitmask))

    if ( LEN_TRIM(str_buffer) .lt. 1 ) then
      ierr = ERROR_LEVEL_SEVERE
      return
    endif

    ! first non-blank character(s) must be '0-9', 'b', '0x', 'x', or 'o'
    !---------------------------------------------------
    if ( str_buffer(1:1) .eq. 'b' ) then   ! treat as a binary number
      valid_digits = TRIM(LIST_OF_BINARY_DIGITS)
      base = 2
      max_digits = max_digits_binary
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( str_buffer(1:1) .eq. 'o' ) then   ! treat as an octal number
      valid_digits = TRIM(LIST_OF_OCTAL_DIGITS)
      base = 8
      max_digits = max_digits_octal
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( (str_buffer(1:1) .eq. 'x') ) then ! treat as a hexadecimal number
      valid_digits = TRIM(LIST_OF_HEXADECIMAL_DIGITS)
      base = 16
      max_digits = max_digits_hexadecimal
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( (str_buffer(1:2) .eq. '0x') ) then ! treat as a hexadecimal number
      valid_digits = TRIM(LIST_OF_HEXADECIMAL_DIGITS)
      base = 16
      max_digits = max_digits_hexadecimal
      str_buffer2 = TRIM(str_buffer(3:))
    else
      ic = INDEX(LIST_OF_DECIMAL_DIGITS, str_buffer(1:1))
      if ( ic .gt. 0 ) then ! treat as an integer
        valid_digits = TRIM(LIST_OF_DECIMAL_DIGITS)
        base = 10
        max_digits = max_digits_decimal
        str_buffer2 = TRIM(str_buffer)
      else
        ierr = ERROR_LEVEL_SEVERE
        return
      endif
   endif

   if ( LEN_TRIM(str_buffer2) .lt. 1 ) then
     ierr = ERROR_LEVEL_SEVERE
     return
   endif

   call to_upper(str_buffer2) ! deal with upper case characters only

   if ( LEN_TRIM(str_buffer2) .gt. max_digits) then ! too many bits!
     ierr = ERROR_LEVEL_SEVERE
     return
   endif
   j = 0
   do i=LEN_TRIM(str_buffer2),1,-1
     ic = INDEX(valid_digits,str_buffer2(i:i))
     if ( ic .gt. 0 ) then
       bitmask = bitmask + INT((ic-1)*(base**j),int16)
     else
       ierr = ERROR_LEVEL_SEVERE
       return
     endif
     j = j + 1
   end do

 end subroutine convert_bitmask_string_to_int16


  !same as above, but for int64
  recursive subroutine convert_bitmask_string_to_int64(str_bitmask, bitmask, ierr)
    character(len=*), parameter :: subroutinename='convert_bitmask_string_to_int64'
    character(len=*),intent(in) ::  str_bitmask
    integer(kind=int64), intent(out) :: bitmask
    integer(kind=int32), intent(out) :: ierr
    character(len=STR_MAX) :: str_buffer
    character(len=STR_MAX) :: str_buffer2
    integer(kind=int32) :: i,j,base,ic
    integer(kind=int32), parameter :: max_digits_binary = 63
    integer(kind=int32), parameter :: max_digits_octal  = 22
    integer(kind=int32), parameter :: max_digits_decimal = 20
    integer(kind=int32), parameter :: max_digits_hexadecimal = 16
    integer(kind=int32) :: max_digits
    character(len=STR_MAX) :: valid_digits
    ierr = ERROR_LEVEL_OK
    bitmask = 0
    if ( LEN_TRIM(ADJUSTL(str_bitmask)) .gt. STR_MAX ) then
      ierr = ERROR_LEVEL_SEVERE
      return
    endif

    str_buffer = TRIM(ADJUSTL(str_bitmask))

    if ( LEN_TRIM(str_buffer) .lt. 1 ) then
      ierr = ERROR_LEVEL_SEVERE
      return
    endif

    ! first non-blank character(s) must be '0-9', 'b', '0x', 'x', or 'o'
    !---------------------------------------------------
    if ( str_buffer(1:1) .eq. 'b' ) then   ! treat as a binary number
      valid_digits = TRIM(LIST_OF_BINARY_DIGITS)
      base = 2
      max_digits = max_digits_binary
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( str_buffer(1:1) .eq. 'o' ) then   ! treat as an octal number
      valid_digits = TRIM(LIST_OF_OCTAL_DIGITS)
      base = 8
      max_digits = max_digits_octal
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( (str_buffer(1:1) .eq. 'x') ) then ! treat as a hexadecimal number
      valid_digits = TRIM(LIST_OF_HEXADECIMAL_DIGITS)
      base = 16
      max_digits = max_digits_hexadecimal
      str_buffer2 = TRIM(str_buffer(2:))
    else if ( (str_buffer(1:2) .eq. '0x') ) then ! treat as a hexadecimal number
      valid_digits = TRIM(LIST_OF_HEXADECIMAL_DIGITS)
      base = 16
      max_digits = max_digits_hexadecimal
      str_buffer2 = TRIM(str_buffer(3:))
    else
      ic = INDEX(LIST_OF_DECIMAL_DIGITS, str_buffer(1:1))
      if ( ic .gt. 0 ) then ! treat as an integer
        valid_digits = TRIM(LIST_OF_DECIMAL_DIGITS)
        base = 10
        max_digits = max_digits_decimal
        str_buffer2 = TRIM(str_buffer)
      else
        ierr = ERROR_LEVEL_SEVERE
        return
      endif
   endif

   if ( LEN_TRIM(str_buffer2) .lt. 1 ) then
     ierr = ERROR_LEVEL_SEVERE
     return
   endif

   call to_upper(str_buffer2) ! deal with upper case characters only

   if ( LEN_TRIM(str_buffer2) .gt. max_digits) then ! too many bits!
     ierr = ERROR_LEVEL_SEVERE
     return
   endif
   j = 0
   do i=LEN_TRIM(str_buffer2),1,-1
     ic = INDEX(valid_digits,str_buffer2(i:i))
     if ( ic .gt. 0 ) then
       bitmask = bitmask + INT((ic-1)*(base**j),int64)
     else
       ierr = ERROR_LEVEL_SEVERE
       return
     endif
     j = j + 1
   end do

 end subroutine convert_bitmask_string_to_int64



!  recursive subroutine strings_array_to_string  (array, length, str)
!    character(len=*),parameter :: separator = ","
!    character(len=*),dimension(:),intent(in) :: array
!    integer(kind=int32), intent(in) :: length
!    character(len=*),intent(inout) :: str
!    integer(kind=int32) :: i, j
!    str = ""
!    do i=1,length
!      j = LEN_TRIM(str)
!      if ( i .gt. 1 ) then
!        if ( j + len(separator) .le. len(str) ) then
!          str = str(1:j)//separator
!        else
!          return
!        endif
!      endif
!
!      j = LEN_TRIM(str)
!      if ( j + LEN_TRIM(array(i)) .le. len(str) ) then
!        str = str(1:j)//TRIM(array(i))
!      else
!        return
!      endif
!    end do
!  end subroutine strings_array_to_string

  recursive subroutine convert_string_array_to_string  (array, nelem, str, separator)
    character(len=*),dimension(:),intent(in) :: array
    integer(kind=int32), intent(in) :: nelem
    character(len=*),intent(in) :: separator
    character(len=*),intent(inout) :: str
    integer(kind=int32) :: i, j, jj
    str = ""
    do i=1,nelem
      j = LEN_TRIM(str)
      jj = j
      if ( i .gt. 1 ) then
        jj = j + LEN(separator)
        if ( jj .le. LEN(str) ) then
          str = str(1:j)//separator
        else
          return
        endif
      endif

      if ( jj + LEN_TRIM(array(i)) .le. LEN(str) ) then
        str = str(1:jj)//TRIM(array(i))
      else
        return
      endif
    end do
  end subroutine convert_string_array_to_string

  recursive subroutine convert_int32_array_to_string  (array, length, str, separator)
    integer(kind=int32),dimension(:),intent(in) :: array
    integer(kind=int32), intent(in) :: length
    character(len=*),intent(out) :: str
    character(len=*),intent(in) :: separator
    character(len=STR_MAX) :: str_temp
    integer(kind=int32) :: i, j, jj
    str = ""
    do i=1,length
      j = LEN_TRIM(str)
      jj = j
      if ( i .gt. 1 ) then
        jj = j + LEN(separator)
        if ( jj .le. LEN(str) ) then
          str = str(1:j)//separator
        else
          return
        endif
      endif

      write(str_temp, '(I0)') array(i)

      if ( jj + LEN_TRIM(str_temp) .le. len(str) ) then
        str = str(1:jj)//TRIM(str_temp)
      else
        return
      endif
    end do
  end subroutine convert_int32_array_to_string

  recursive subroutine convert_logical_array_to_string  (array, str, separator)
    logical, dimension(:), intent(in) :: array
    character(len=*),intent(out) :: str
    character(len=*),intent(in) :: separator
    character(len=STR_MAX) :: str_temp
    integer(kind=int32) :: i, j, jj
    str = ""
    do i=1,size(array)
      j = LEN_TRIM(str)
      jj = j
      if ( i .gt. 1 ) then
        jj = j + LEN(separator)
        if ( jj .le. len(str) ) then
          str = str(1:j)//separator
        else
          return
        endif
      endif

      str_temp = logical2char (array(i))
      if ( jj + LEN_TRIM(str_temp) .le. len(str) ) then
        str = str(1:jj)//TRIM(str_temp)
      else
        return
      endif
    end do
  end subroutine convert_logical_array_to_string

  !--------------------------------------------------------
  recursive subroutine convert_real32_array_to_string  (array, length, fmt, str, separator)
    real(kind=real32),dimension(:),intent(in) :: array
    integer(kind=int32), intent(in) :: length
    character(len=*),intent(in) :: fmt
    character(len=*),intent(out) :: str
    character(len=*),intent(in) :: separator
    character(len=STR_MAX) :: str_temp
    integer(kind=int32) :: i, j, jj
    str = ""
    do i=1,length
      j = LEN_TRIM(str)
      jj = j
      if ( i .gt. 1 ) then
        jj = j + LEN(separator)
        if ( jj .le. len(str) ) then
          str = str(1:jj)//separator
        else
          return
        endif
      endif

      write(str_temp, fmt) array(i)
      if ( jj + LEN_TRIM(str_temp) .le. len(str) ) then
        str = str(1:jj)//TRIM(str_temp)
      else
        return
      endif
    end do
  end subroutine convert_real32_array_to_string
  !--------------------------------------------------------


  !--------------------------------------------------------
  recursive subroutine convert_real64_array_to_string  (array, length, fmt, str, separator)
    real(kind=real64),dimension(:),intent(in) :: array
    integer(kind=int32), intent(in) :: length
    character(len=*),intent(in) :: fmt
    character(len=*),intent(out) :: str
    character(len=*),intent(in) :: separator
    character(len=STR_MAX) :: str_temp
    integer(kind=int32) :: i, j, jj
    str = ""
    do i=1,length
      j = LEN_TRIM(str)
      jj = j
      if ( i .gt. 1 ) then
        jj = j + LEN(separator)
        if ( jj .le. len(str) ) then
          str = str(1:j)//separator
        else
          return
        endif
      endif

      write(str_temp, fmt) array(i)
      if ( jj + LEN_TRIM(str_temp) .le. len(str) ) then
        str = str(1:jj)//TRIM(str_temp)
      else
        return
      endif
    end do
  end subroutine convert_real64_array_to_string
  !--------------------------------------------------------



  !--------------------------------------------------------
  ! Change any lower-case characters (A-Z) in a string to upper case
  ! adpted from here: http://stackoverflow.com/questions/10759375/how-can-i-write-a-to-upper-or-to-lower-function-in-f90
  recursive subroutine to_upper(string)
    character(*), intent(inout) :: string
    integer :: ic, i
    !   Capitalize each letter if it is lowercase
    do i = 1, LEN_TRIM(string)
        ic = INDEX(LIST_OF_LOWERCASE_LETTERS, string(i:i))
        if (ic > 0) string(i:i) = LIST_OF_UPPERCASE_LETTERS(ic:ic)
    end do
  end subroutine to_upper

  ! Change any upper-case characters (A-Z) in a string to lower case
  recursive subroutine to_lower(string)
    character(*), intent(inout) :: string
    integer :: ic, i
    do i = 1, LEN_TRIM(string)
        ic = INDEX(LIST_OF_UPPERCASE_LETTERS, string(i:i))
        if (ic > 0) string(i:i) = LIST_OF_LOWERCASE_LETTERS(ic:ic)
    end do
  end subroutine to_lower
  !--------------------------------------------------------

  !--------------------------------------------------------
  ! return a string with lowered characters
  recursive function to_lower_fun(string)
    character(:), allocatable :: to_lower_fun
    character(*), intent(in) :: string

    allocate(character(len(string)) :: to_lower_fun)
    to_lower_fun = string
    call to_lower(to_lower_fun)
  end function to_lower_fun
  !--------------------------------------------------------

  !--------------------------------------------------------
  ! return a string with uppered characters
  recursive function to_upper_fun(string)
    character(:), allocatable :: to_upper_fun
    character(*), intent(in) :: string

    allocate(character(len(string)) :: to_upper_fun)
    to_upper_fun = string
    call to_upper(to_upper_fun)
  end function to_upper_fun
  !--------------------------------------------------------


  !--------------------------------------------------------
  ! full angular separation function - law of haversines
  ! https://en.wikipedia.org/wiki/Haversine_formula
  ! input coords in degrees
  ! returned value is radial distance in degrees
  pure recursive function calc_ang_sep (long1,lat1,long2,lat2)
    real(kind=real64), intent(in) ::  long1,lat1,long2,lat2
    real(kind=real64) :: calc_ang_sep
    real(kind=real64) :: haversine2_dlat, haversine2_dlong
    real(kind=real64) :: haversine_ang_sep

    if ( ( eq_real64(long1,long2) ) .and. (eq_real64(lat1,lat2) )) then
      calc_ang_sep = 0.0d0
    else
      haversine2_dlat  = sind((lat2  - lat1 )*0.5d0)
      haversine2_dlong = sind((long2 - long1)*0.5d0)
      haversine2_dlat  = haversine2_dlat*haversine2_dlat
      haversine2_dlong = haversine2_dlong*haversine2_dlong
      haversine_ang_sep  = sqrt(haversine2_dlat + cosd(lat1)*cosd(lat2)*haversine2_dlong)
      calc_ang_sep = 2.0d0 * asind(haversine_ang_sep)
    endif

  end function calc_ang_sep
  !--------------------------------------------------------


  !--------------------------------------------------------
  ! Slightly faster version as it receives precomputed cosines of the latitutes
  pure recursive function calc_ang_sep_fast (long1,lat1,cos_lat1, long2, lat2, cos_lat2)
    real(kind=real64), intent(in) ::  long1,lat1,long2,lat2, cos_lat1, cos_lat2
    real(kind=real64) :: calc_ang_sep_fast
    real(kind=real64) :: haversine2_dlat, haversine2_dlong
    real(kind=real64) :: haversine_ang_sep

    if ( ( eq_real64(long1,long2) ) .and. (eq_real64(lat1,lat2) )) then
      calc_ang_sep_fast = 0.0d0
    else
      haversine2_dlat  = sind((lat2  - lat1 )*0.5d0)
      haversine2_dlong = sind((long2 - long1)*0.5d0)
      haversine2_dlat  = haversine2_dlat*haversine2_dlat
      haversine2_dlong = haversine2_dlong*haversine2_dlong
      haversine_ang_sep  = sqrt(haversine2_dlat + cos_lat1*cos_lat2*haversine2_dlong)
      calc_ang_sep_fast = 2.0d0 * asind(haversine_ang_sep)
    endif

  end function calc_ang_sep_fast
  !--------------------------------------------------------


  !--------------------------------------------------------
  ! faster version of above as it avoids the call to arcsin
  ! returned value is the haversine of the radial distance
  pure recursive function calc_haversine_ang_sep (long1,lat1,long2,lat2)
    real(kind=real64), intent(in) ::  long1,lat1,long2,lat2
    real(kind=real64) :: calc_haversine_ang_sep
    real(kind=real64) :: haversine2_dlat, haversine2_dlong

    if ( ( eq_real64(long1,long2) ) .and. (eq_real64(lat1,lat2) )) then
      calc_haversine_ang_sep = 0.0d0
    else
      haversine2_dlat  = sind((lat2  - lat1 )*0.5d0)
      haversine2_dlong = sind((long2 - long1)*0.5d0)
      haversine2_dlat  = haversine2_dlat*haversine2_dlat
      haversine2_dlong = haversine2_dlong*haversine2_dlong
      calc_haversine_ang_sep = sqrt(haversine2_dlat + cosd(lat1)*cosd(lat2)*haversine2_dlong)
    endif

  end function calc_haversine_ang_sep
  !--------------------------------------------------------

  !--------------------------------------------------------
  ! even faster version of above as it receives precomputed cosines of the latitutes
  ! returned value is the haversine of the radial distance
  pure recursive function calc_haversine_ang_sep_fast (long1,lat1,cos_lat1,long2,lat2,cos_lat2)
    real(kind=real64), intent(in) ::  long1,lat1,long2,lat2, cos_lat1, cos_lat2
    real(kind=real64) :: calc_haversine_ang_sep_fast
    real(kind=real64) :: haversine2_dlat, haversine2_dlong

    if ( ( eq_real64(long1,long2) ) .and. (eq_real64(lat1,lat2) )) then
      calc_haversine_ang_sep_fast = 0.0d0
    else
      haversine2_dlat  = sind((lat2  - lat1 )*0.5d0)
      haversine2_dlong = sind((long2 - long1)*0.5d0)
      haversine2_dlat  = haversine2_dlat*haversine2_dlat
      haversine2_dlong = haversine2_dlong*haversine2_dlong
      calc_haversine_ang_sep_fast = sqrt(haversine2_dlat + cos_lat1*cos_lat2*haversine2_dlong)
    endif

  end function calc_haversine_ang_sep_fast
  !--------------------------------------------------------





  !--------------------------------------------------------
  ! full angular separation function - law of cosines
  ! input coords in degrees
  ! returned value is radial distance in degrees
  ! not robust for small angles - so deprecated
  pure recursive function calc_ang_sep_cosines (long1,lat1,long2,lat2)
    real(kind=real64), intent(in) ::  long1,lat1,long2,lat2
    real(kind=real64) :: calc_ang_sep_cosines
    real(kind=real64) :: cos_ang_sep
    real(kind=real64) :: sin_lat1, cos_lat1
    real(kind=real64) :: sin_lat2, cos_lat2

    sin_lat1 = sind(lat1)
    cos_lat1 = cosd(lat1)
    sin_lat2 = sind(lat2)
    cos_lat2 = cosd(lat2)

    cos_ang_sep = sin_lat1*sin_lat2 + cos_lat1*cos_lat2*cosd(long2 - long1)
    if ( ( eq_real64(long1,long2) ) .and. (eq_real64(lat1,lat2) )) then
      calc_ang_sep_cosines = 0.0d0
    else
      calc_ang_sep_cosines = acosd(cos_ang_sep)
    endif

  end function calc_ang_sep_cosines
  !--------------------------------------------------------

  !--------------------------------------------------------
  ! returned value is bearing in degrees
  pure recursive function calc_bearing (long1, lat1, long2, lat2)
    real(kind=real64), intent(in) ::  long1,lat1,long2,lat2
    real(kind=real64) :: calc_bearing
    real(kind=real64) :: x, y
    real(kind=real64) :: sin_lat1, cos_lat1
    real(kind=real64) :: sin_lat2, cos_lat2
    sin_lat1 = sind(lat1)
    cos_lat1 = cosd(lat1)
    sin_lat2 = sind(lat2)
    cos_lat2 = cosd(lat2)

    y = sind(long2 - long1) * cos_lat2
    x = cos_lat1*sin_lat2 - sin_lat1*cos_lat2*cosd(long2 - long1)
    calc_bearing = atan2(y, x)*RAD2DEG
  end function calc_bearing

  !--------------------------------------------------------
  ! returned value is bearing in degrees
  pure recursive function calc_bearing_fast (long1, cos_lat1, sin_lat1, long2, cos_lat2, sin_lat2)
    real(kind=real64), intent(in) :: long1, sin_lat1, cos_lat1
    real(kind=real64), intent(in) :: long2, sin_lat2, cos_lat2
    real(kind=real64) :: calc_bearing_fast
    real(kind=real64) :: x, y
    real(kind=real64) :: sin_long2_long1
    real(kind=real64) :: cos_long2_long1

    sin_long2_long1 = sind(long2 - long1)
    cos_long2_long1 = cosd(long2 - long1)
    y = sin_long2_long1 * cos_lat2
    x = cos_lat1*sin_lat2 - sin_lat1*cos_lat2*cos_long2_long1
    calc_bearing_fast = atan2(y, x)*RAD2DEG
  end function calc_bearing_fast

  !calc new position from old position, bearing and distance
  !assume motion along a greatcircle
  ! from  http://www.movable-type.co.uk/scripts/latlong.html
  recursive subroutine calc_destination (long1, lat1, dist, bearing, long0, lat0)
    real(kind=real64), intent(in)  :: long1,lat1,dist,bearing
    real(kind=real64), intent(out) :: long0, lat0
    real(kind=real64)  :: sin_dist, cos_dist, sin_lat1, cos_lat1, sin_bearing, cos_bearing
    sin_dist = sind(dist)
    cos_dist = cosd(dist)
    sin_lat1 = sind(lat1)
    cos_lat1 = cosd(lat1)
    sin_bearing = sind(bearing)
    cos_bearing = cosd(bearing)
    lat0  = asind(sin_lat1*cos_dist + cos_lat1*sin_dist*cos_bearing)
    long0 = long1 + RAD2DEG*atan2(sin_bearing*sin_dist*cos_lat1, cos_dist - sin_lat1*sind(lat0))
    long0 = modulo(long0,360.0d0)
  end subroutine calc_destination


  recursive subroutine calc_destination_fast (long1, cos_lat1, sin_lat1, dist, bearing, long0, lat0)
    real(kind=real64), intent(in)  :: long1,dist,bearing, cos_lat1, sin_lat1
    real(kind=real64), intent(out) :: long0, lat0
    real(kind=real64)  :: sin_dist, cos_dist, sin_bearing, cos_bearing
    sin_dist = sind(dist)
    cos_dist = cosd(dist)
    sin_bearing = sind(bearing)
    cos_bearing = cosd(bearing)
    lat0  = asind(sin_lat1*cos_dist + cos_lat1*sin_dist*cos_bearing)
    long0 = long1 + RAD2DEG*atan2(sin_bearing*sin_dist*cos_lat1, cos_dist - sin_lat1*sind(lat0))
    long0 = modulo(long0,360.0d0)
  end subroutine calc_destination_fast

  recursive subroutine calc_destination_faster (long1, cos_lat1, sin_lat1, dist, &
      cos_bearing, sin_bearing, long0, lat0)
    real(kind=real64), intent(in)  :: long1, cos_lat1, sin_lat1
    real(kind=real64), intent(in)  :: dist, sin_bearing, cos_bearing
    real(kind=real64), intent(out) :: long0, lat0
    real(kind=real64)  :: sin_dist, cos_dist
    sin_dist = sind(dist)
    cos_dist = cosd(dist)
    lat0  = asind(sin_lat1*cos_dist + cos_lat1*sin_dist*cos_bearing)
    long0 = long1 + atan2d(sin_bearing*sin_dist*cos_lat1, cos_dist - sin_lat1*sind(lat0))
    long0 = modulo(long0,360.0d0)
  end subroutine calc_destination_faster
  !--------------------------------------------------------------------------


  ! find the point that is (t0-t1)/(t2-t1) of the way along the great circle that joins 1 and 2
  ! derived from c-code at 4most/OpSim/geometry_lib.c
  recursive subroutine interpolate_long_lat ( t1, long1, lat1, t2, long2, lat2, t0, long0, lat0, status )
    character(len=*), parameter :: subroutinename='interpolate_long_lat'
    real(kind=real64),  intent(in)  :: t1, long1, lat1, t2, long2, lat2, t0
    real(kind=real64),  intent(out) :: long0, lat0
    real(kind=real64) ::dist, bearing, dt21, dt01
    integer(kind=int32),  intent(out)   :: status
    real(kind=real64) :: sin_lat1, cos_lat1
    real(kind=real64) :: sin_lat2, cos_lat2
    status = 0
    sin_lat1 = sind(lat1)
    cos_lat1 = cosd(lat1)
    sin_lat2 = sind(lat2)
    cos_lat2 = cosd(lat2)

    bearing = calc_bearing_fast (long1, cos_lat1, sin_lat1, long2, cos_lat2, sin_lat2)
    dt21 = t2 - t1;
    dt01 = t0 - t1;
    if ( eq_real64(dt01,0.0d0) ) then ! we are at the starting point
      long0 = long1
      lat0  = lat1
    else if ( eq_real64(dt21,0.0d0) ) then ! the start and end points are the same
      status = 1
      long0 = huge(long0) !TODO
      lat0  = huge(lat0)
    else
      dist = calc_ang_sep_fast(long1, lat1, cos_lat1, long2, lat2, cos_lat2) * dt01/dt21
      call calc_destination_fast (long1, cos_lat1, sin_lat1, dist, bearing, long0, lat0)
    endif
  end subroutine interpolate_long_lat


  ! faster version of thw above which uses a pre-computed distance and bearing between the two end points
  recursive subroutine interpolate_long_lat_fast ( t1, long1, lat1, cos_lat1, sin_lat1, t2, long2, lat2, dist12, &
      cos_bearing, sin_bearing, t0, long0, lat0, status )
    character(len=*), parameter :: subroutinename='interpolate_long_lat'
    real(kind=real64),  intent(in)  :: t1, long1, lat1, t2, long2, lat2, dist12, cos_bearing, sin_bearing, t0
    real(kind=real64),  intent(in) :: sin_lat1, cos_lat1
    real(kind=real64),  intent(out) :: long0, lat0
    real(kind=real64) ::dt21, dt01
    integer(kind=int32),  intent(out)   :: status
    real(kind=real64) :: t_frac
    real(kind=real64), parameter :: max_dec_for_fast = 75.0d0 ! deg
    real(kind=real64), parameter :: max_dist_for_fast = 0.05d0 ! deg = 3arcmin
    real(kind=real64), parameter :: max_delta_ra_for_fast = 10.0d0 ! deg


    status = 0

    dt21 = t2 - t1;
    dt01 = t0 - t1;
    if ( is_zero(dt01) ) then ! we are at the starting point
      long0 = long1
      lat0  = lat1
    else if ( is_zero(dt21) ) then ! the start and end points are the same
      status = 1
      long0 = huge(long0)
      lat0  = huge(lat0)
    else
      ! do a simple linear interpolation under certain retricted circumstances
      t_frac = dt01/dt21
      if ((dist12           .lt. max_dist_for_fast) .and. & ! Only a small separation between consecutive attitude measurements
          (abs(lat1)        .lt. max_dec_for_fast)  .and. & ! Avoid the poles
          (abs(long2-long1) .lt. max_delta_ra_for_fast) ) then ! Avoid points which cross the RA=0,360deg boundary
        long0  = long1  + t_frac*(long2  - long1)
        lat0   = lat1   + t_frac*(lat2   - lat1)
      else
        ! fall back to the proper interpolation scheme
        call calc_destination_faster (long1, cos_lat1, sin_lat1, dist12*t_frac, cos_bearing, sin_bearing, long0, lat0)
      endif
    endif
  end subroutine interpolate_long_lat_fast


  ! linearly interpolate a periodic variable (note: here t is periodic, not v)
  ! assume that interpolate over the smallest distance
  recursive function interpolate_periodic_t ( t1, v1, t2, v2, t_period, t0)
    real(kind=real64),  intent(in)  :: t1,  v1, t2, v2, t_period, t0
    real(kind=real64) :: interpolate_periodic_t
    real(kind=real64) :: abs_dt12, tt1, tt2, tt0
    tt1 = modulo(t1, t_period)
    tt2 = modulo(t2, t_period)
    tt0 = modulo(t0, t_period)
    abs_dt12 = abs(tt2 - tt1)
    if ( abs_dt12 .eq. 0.0d0 ) then
      interpolate_periodic_t = 0.5d0*(v1+v2)
    else
      if ( abs_dt12 .gt. 0.5d0*t_period ) then
        if ( tt2 .gt. tt1 ) then
          tt1 = tt1 + t_period
        else
          tt2 = tt2 + t_period
        endif
      endif
      interpolate_periodic_t = modulo(v1 + (tt0-tt1)*(v2 - v1)/(tt2 - tt1) , t_period)
    endif
  end function interpolate_periodic_t

  ! linearly interpolate a periodic variable (note: here v is periodic, not t)
  ! assume that we will interpolate over the smallest distance
  recursive function interpolate_periodic_v ( t1, v1, t2, v2, v_period, t0)
    real(kind=real64),  intent(in)  :: t1,  v1, t2, v2, v_period, t0
    real(kind=real64) :: interpolate_periodic_v
    real(kind=real64) :: abs_dv12, vv1, vv2
    vv1 = modulo(v1, v_period)
    vv2 = modulo(v2, v_period)
    abs_dv12 = abs(vv2 - vv1)
    if ( abs_dv12 .gt. 0.5d0*v_period ) then
      if ( vv2 .gt. vv1 ) then
        vv1 = vv1 + v_period
      else
        vv2 = vv2 + v_period
      endif
    endif

    if ( abs(t2 - t1) .eq. 0.0d0 ) then
      interpolate_periodic_v = modulo(0.5d0*(vv1+vv2),v_period)
    else
      interpolate_periodic_v = modulo(vv1 + (t0-t1)*(vv2 - vv1)/(t2 - t1) , v_period)
    endif
  end function interpolate_periodic_v

  !--------------------------------------------------------
  !-----------------SORT functions-------------------------
  !--------------------------------------------------------
  !http://rosettacode.org/wiki/Sorting_algorithms/Insertion_sort#Fortran
  recursive subroutine insertion_sort_real64(a)
    real(kind=real64), intent(inout), dimension(:) :: a
    real(kind=real64) :: temp
    integer(kind=int64) :: i, j
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      temp = a(i)
!      write (*, '(A,3(1x,I8),2(1x,F12.7))'), "insertion_sort_real64()", size(a), i, j, a(i), a(j)
      do while ((j .ge. 1) .and. (a(j) .gt. temp))
        a(j+1) = a(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = temp
    end do
  end subroutine insertion_sort_real64
  !--------------------------------------------------------

  recursive subroutine insertion_sort_real32(a)
    real(kind=real32), intent(inout), dimension(:) :: a
    real(kind=real32) :: temp
    integer(kind=int64) :: i, j
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      temp = a(i)
      do while ((j .ge. 1) .and. (a(j) .gt. temp))
        a(j+1) = a(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = temp
    end do
  end subroutine insertion_sort_real32
  !--------------------------------------------------------

  recursive subroutine insertion_sort_int32(a)
    integer(kind=int32), intent(inout), dimension(:) :: a
    integer(kind=int32) :: temp
    integer(kind=int64) :: i, j
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      temp = a(i)
      do while ((j .ge. 1) .and. (a(j) .gt. temp))
        a(j+1) = a(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = temp
    end do
  end subroutine insertion_sort_int32
  !--------------------------------------------------------
  recursive subroutine insertion_sort_int64(a)
    integer(kind=int64), intent(inout), dimension(:) :: a
    integer(kind=int64) :: temp
    integer(kind=int64) :: i, j
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      temp = a(i)
      do while ((j .ge. 1) .and. (a(j) .gt. temp))
        a(j+1) = a(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = temp
    end do
  end subroutine insertion_sort_int64
  !--------------------------------------------------------


  !--------------------------------------------------------
  ! rearrange accompanying array b() at the same time as sorting array a()
  recursive subroutine insertion_cosort_real64_real64(a,b)
    real(kind=real64), intent(inout), dimension(:) :: a
    real(kind=real64), intent(inout), dimension(:) :: b
    real(kind=real64) :: tempa
    real(kind=real64) :: tempb
    integer(kind=int64) :: i, j
    if ( size(a) .ne. size(b) ) return
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      tempa = a(i)
      tempb = b(i)
      do while ((j .ge. 1) .and. (a(j) .gt. tempa))
        a(j+1) = a(j)
        b(j+1) = b(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = tempa
      b(j+1) = tempb
    end do
  end subroutine insertion_cosort_real64_real64

  recursive subroutine insertion_cosort_real64_real32(a,b)
    real(kind=real64), intent(inout), dimension(:) :: a
    real(kind=real32), intent(inout), dimension(:) :: b
    real(kind=real64) :: tempa
    real(kind=real32) :: tempb
    integer(kind=int64) :: i, j
    if ( size(a) .ne. size(b) ) return
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      tempa = a(i)
      tempb = b(i)
      do while ((j .ge. 1) .and. (a(j) .gt. tempa))
        a(j+1) = a(j)
        b(j+1) = b(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = tempa
      b(j+1) = tempb
    end do
  end subroutine insertion_cosort_real64_real32

  ! rearrange accompanying array b() and c() at the same time as sorting array a()
  recursive subroutine insertion_cosort_real64_real64_real64(a,b,c)
    real(kind=real64), intent(inout), dimension(:) :: a
    real(kind=real64), intent(inout), dimension(:) :: b
    real(kind=real64), intent(inout), dimension(:) :: c
    real(kind=real64) :: tempa
    real(kind=real64) :: tempb
    real(kind=real64) :: tempc
    integer(kind=int64) :: i, j
    if ( (size(a) .ne. size(b)) .or. (size(a) .ne. size(c))) return
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      tempa = a(i)
      tempb = b(i)
      tempc = c(i)
      do while ((j .ge. 1) .and. (a(j) .gt. tempa))
        a(j+1) = a(j)
        b(j+1) = b(j)
        c(j+1) = c(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = tempa
      b(j+1) = tempb
      c(j+1) = tempc
    end do
  end subroutine insertion_cosort_real64_real64_real64
  !--------------------------------------------------------
  ! rearrange accompanying array b() and c() and d() at the same time as sorting array a()
  recursive subroutine insertion_cosort_real64_real64_real64_real64(a,b,c,d)
    real(kind=real64), intent(inout), dimension(:) :: a
    real(kind=real64), intent(inout), dimension(:) :: b
    real(kind=real64), intent(inout), dimension(:) :: c
    real(kind=real64), intent(inout), dimension(:) :: d
    real(kind=real64) :: tempa
    real(kind=real64) :: tempb
    real(kind=real64) :: tempc
    real(kind=real64) :: tempd
    integer(kind=int64) :: i, j
    if ( (size(a) .ne. size(b)) .or. (size(a) .ne. size(c)) .or. (size(a) .ne. size(d))) return
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      tempa = a(i)
      tempb = b(i)
      tempc = c(i)
      tempd = d(i)
      do while ((j .ge. 1) .and. (a(j) .gt. tempa))
        a(j+1) = a(j)
        b(j+1) = b(j)
        c(j+1) = c(j)
        d(j+1) = d(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = tempa
      b(j+1) = tempb
      c(j+1) = tempc
      d(j+1) = tempd
    end do
  end subroutine insertion_cosort_real64_real64_real64_real64
  !--------------------------------------------------------

  !--------------------------------------------------------
  ! rearrange accompanying arrays b,c,d,e at same time as sorting array a()
  recursive subroutine insertion_cosort_3real64_2int32(a,b,c,d,e)
    real(kind=real64), intent(inout), dimension(:) :: a
    real(kind=real64), intent(inout), dimension(:) :: b
    real(kind=real64), intent(inout), dimension(:) :: c
    integer(kind=int32), intent(inout), dimension(:) :: d
    integer(kind=int32), intent(inout), dimension(:) :: e
    real(kind=real64) :: tempa
    real(kind=real64) :: tempb
    real(kind=real64) :: tempc
    integer(kind=int32) :: tempd
    integer(kind=int32) :: tempe
    integer(kind=int64) :: i, j
    if ((size(a) .ne. size(b)) .or. &
        (size(a) .ne. size(c)) .or. &
        (size(a) .ne. size(d)) .or. &
        (size(a) .ne. size(e)) ) return
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      tempa = a(i)
      tempb = b(i)
      tempc = c(i)
      tempd = d(i)
      tempe = e(i)
      do while ((j .ge. 1) .and. (a(j) .gt. tempa))
        a(j+1) = a(j)
        b(j+1) = b(j)
        c(j+1) = c(j)
        d(j+1) = d(j)
        e(j+1) = e(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = tempa
      b(j+1) = tempb
      c(j+1) = tempc
      d(j+1) = tempd
      e(j+1) = tempe
    end do
  end subroutine insertion_cosort_3real64_2int32
  !--------------------------------------------------------

  !--------------------------------------------------------
  ! rearrange accompanying array b() and c() and d() at the same time as sorting array a()
  recursive subroutine insertion_cosort_4real64_2int32(a,b,c,d,e,f)
    real(kind=real64), intent(inout), dimension(:) :: a
    real(kind=real64), intent(inout), dimension(:) :: b
    real(kind=real64), intent(inout), dimension(:) :: c
    real(kind=real64), intent(inout), dimension(:) :: d
    integer(kind=int32), intent(inout), dimension(:) :: e
    integer(kind=int32), intent(inout), dimension(:) :: f
    real(kind=real64) :: tempa
    real(kind=real64) :: tempb
    real(kind=real64) :: tempc
    real(kind=real64) :: tempd
    integer(kind=int32) :: tempe
    integer(kind=int32) :: tempf
    integer(kind=int64) :: i, j
    if ((size(a) .ne. size(b)) .or. &
        (size(a) .ne. size(c)) .or. &
        (size(a) .ne. size(d)) .or. &
        (size(a) .ne. size(e)) .or. &
        (size(a) .ne. size(f)) ) return
    if ( size(a) .lt. 2 ) return
    do i = 2, size(a)
      j = i - 1
      tempa = a(i)
      tempb = b(i)
      tempc = c(i)
      tempd = d(i)
      tempe = e(i)
      tempf = f(i)
      do while ((j .ge. 1) .and. (a(j) .gt. tempa))
        a(j+1) = a(j)
        b(j+1) = b(j)
        c(j+1) = c(j)
        d(j+1) = d(j)
        e(j+1) = e(j)
        f(j+1) = f(j)
        j = j - 1
        if ( j .lt. 1 ) exit
      end do
      a(j+1) = tempa
      b(j+1) = tempb
      c(j+1) = tempc
      d(j+1) = tempd
      e(j+1) = tempe
      f(j+1) = tempf
    end do
  end subroutine insertion_cosort_4real64_2int32
  !--------------------------------------------------------

  !--------------------------------------------------------

  !--------------------------------------------------------
  ! recursive subroutine QSort(a,na)

  !   ! DUMMY ARGUMENTS
  !   integer(kind=int64), intent(in) :: nA
  !   type (QSort_group), dimension(nA), intent(in out) :: A

  !   ! LOCAL VARIABLES
  !   integer(kind=int64) :: left, right
  !   real(kind=real64) :: random
  !   real(kind=real64) :: pivot
  !   type (QSort_group) :: temp
  !   integer(kind=int64) :: marker

  !   if (nA > 1) then

  !     call random_number(random)
  !     pivot = A(int(random*real(nA-1))+1)%value    ! random pivot (not best performance, but avoids worst-case)
  !     left = 0
  !     right = nA + 1

  !     do while (left < right)
  !       right = right - 1
  !       do while (A(right)%value > pivot)
  !         right = right - 1
  !       end do
  !       left = left + 1
  !       do while (A(left)%value < pivot)
  !         left = left + 1
  !       end do
  !       if (left < right) then
  !         temp = A(left)
  !         A(left) = A(right)
  !         A(right) = temp
  !       end if
  !     end do

  !     if (left == right) then
  !       marker = left + 1
  !     else
  !       marker = left
  !     end if

  !     call QSort(A(:marker-1),marker-1)
  !     call QSort(A(marker:),nA-marker+1)

  !   end if
  ! end subroutine QSort

  !-----------------------------------------------------
  ! shrink a previously sorted array so that it only contains unique values
  recursive subroutine filter_unique_real64(a,nunique)
    real(kind=real64), intent(inout), dimension(:) :: a
    integer(kind=int32), intent(out) :: nunique
    integer(kind=int32) :: j
    nunique = 1
    j = nunique + 1
    do while ( j .le. size(a))
!      do while ( a(nunique) == a(j) )
      do while ( eq_real64(a(nunique),a(j)) )
        j = j + 1
        if ( j .gt. size(a)) then
          return
        end if
      end do
      nunique = nunique + 1
      a(nunique) = a(j)
    end do
  end subroutine filter_unique_real64

  recursive subroutine filter_unique_real32(a,nunique)
    real(kind=real32), intent(inout), dimension(:) :: a
    integer(kind=int32), intent(out) :: nunique
    integer(kind=int32) :: j
    nunique = 1
    j = nunique + 1
    do while ( j .le. size(a))
      do while ( eq_real32(a(nunique),a(j)) )
        j = j + 1
        if ( j .gt. size(a)) then
          return
        end if
      end do
      nunique = nunique + 1
      a(nunique) = a(j)
    end do
  end subroutine filter_unique_real32

  recursive subroutine filter_unique_int32(a,nunique)
    integer(kind=int32), intent(inout), dimension(:) :: a
    integer(kind=int32), intent(out) :: nunique
    integer(kind=int32) :: j
    nunique = 1
    j = nunique + 1
    do while ( j .le. size(a))
      do while ( a(nunique) .eq. a(j) )
        j = j + 1
        if ( j .gt. size(a)) then
          return
        end if
      end do
      nunique = nunique + 1
      a(nunique) = a(j)
    end do
  end subroutine filter_unique_int32
  !--------------------------------------------------------

  !--------------------------------------------------------
  ! allow following formats
  ! DDD.DDDDD...
  ! HH:MM:SS.SSS...
  ! HHhMMmSS.SSS...s
  ! HH MM SS.SSS...
  recursive subroutine convert_longitude_to_decimal (str_long, long, success_flag)
    character(len=*), parameter :: subroutinename='convert_longitude_to_decimal'
    character(len=*),intent(in) :: str_long
    real(kind=real64),intent(out) :: long
    logical,intent(out) ::  success_flag
!    character(len=:), allocatable :: str_copy   ! removed this since it was not working under some compilers
    character(len=LSTR_MAX) :: str_copy

    character(len=STR_MAX) :: str_fmt
    integer(kind=int32)  ::  i, j, k, flag
    integer(kind=int32)  :: hh, mm
    real(kind=real64)  :: ss
    real(kind=real64)  :: val
    integer(kind=int32) :: status
    success_flag = .false.
    long = -99.9d0
    flag = 0
    status = 0
    i=-1
    j=-1
    k=-1

    if ( LEN_TRIM(ADJUSTL(str_long)) .gt. len(str_copy)) then
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: input string is too long")
      return
    endif

    if ( LEN_TRIM(ADJUSTL(str_long)) .lt. 1) then
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: input string is too short")
      return
    endif

    ! take a copy of the input string and convert to lowercase
    str_copy = TRIM(ADJUSTL(str_long))
    call to_lower(str_copy)


    !--------------------------------------------------------
    ! Now try to read as HH:MM:SS.SSS...
    ! First find locations of the ':' characters
    if ( flag .eq. 0 ) then
      i = index(str_copy,":")   ! 1st one
      if ( i .gt. 0 ) then
        j = index(str_copy(i+1:LEN_TRIM(str_copy)),":") ! 2nd one
        if ( j .gt. 0 ) then
!          k = len(str_copy(i+1+j+1:LEN_TRIM(str_copy)))
          k = len(str_copy(i+1+j:LEN_TRIM(str_copy)))
          write (str_fmt, '(A,I0,A,I0,A,I0,A)') &
              '(I', i-1, ',1x,I', j-1, ',1x,F', k, '.0)'
          read (str_copy, TRIM(str_fmt), iostat=status) hh,mm,ss

          if ( status .eq. 0 ) then
            flag = 1
          end if

        endif
      end if
    end if
    !--------------------------------------------------------

    !--------------------------------------------------------
    ! Now try to read as HH MM SS.SSS
    ! First find locations of the ' ' characters
    if ( flag .eq. 0 ) then
      i = index(str_copy," ")   ! 1st one
      if ( i .gt. 0 ) then
        j = index(str_copy(i+1:LEN_TRIM(str_copy))," ") ! 2nd one
        if ( j .gt. 0 ) then
!          k = len(str_copy(i+1+j+1:LEN_TRIM(str_copy)))
          k = len(str_copy(i+1+j:LEN_TRIM(str_copy)))
          write (str_fmt, '(A,I0,A,I0,A,I0,A)') &
              '(I', i-1, ',1x,I', j-1, ',1x,F', k, '.0)'
          read (str_copy, TRIM(str_fmt), iostat=status) hh,mm,ss

          if ( status .eq. 0 ) then
            flag = 1
          endif
        endif
      endif
    end if
    !--------------------------------------------------------

    !--------------------------------------------------------
    ! Now try to read as HHhMMmSS.SSS...s
    ! First find locations of the 'h,m,s' characters
    if ( flag .eq. 0 ) then
      i = index(str_copy,"h")   ! 1st one
      if ( i .gt. 0 ) then
        j = index(str_copy(i+1:LEN_TRIM(str_copy)),"m") ! 2nd one
        if ( j .gt. 0 ) then
          k = index(str_copy(i+1+j:LEN_TRIM(str_copy)),"s") ! 3rd one
          if ( k .gt. 0 ) then
            write (str_fmt, '(A,I0,A,I0,A,I0,A)') &
                '(I', i-1, ',1x,I', j-1, ',1x,F', k-1, '.0,1x)'
            read (str_copy, TRIM(str_fmt), iostat=status) hh,mm,ss

            if ( status .eq. 0 ) then
              flag = 1
            endif
          endif
        endif
      end if
    end if

    if ( flag .eq. 1 ) then
      if (hh .ge. 0     .and. hh .lt. 24 .and. &
          mm .ge. 0     .and. mm .lt. 60 .and. &
          ss .ge. 0.0d0 .and. ss .lt. 60.0d0 ) then
        long = hh * HOUR2DEG + &
            mm * MINUTE2DEG + &
            ss * SECOND2DEG
        success_flag = .true.
      endif
    endif

    !--------------------------------------------------------
    ! attempt to read the string as a real
    if ( flag .eq. 0 ) then
      write (str_fmt, '(A,I0,A)') '(F', len(str_copy), '.0)'
      read (str_copy, TRIM(str_fmt), iostat=status) val
      if ( status .eq. 0 ) then
        long = val
        success_flag = .true.
        flag = 1
      endif
    end if
    !--------------------------------------------------------

!    write (*,'(A,2(A,A,A),4(A,I0),A,A,A,I0,A,I0,A,F12.5,A,F15.9)'), &
!        subroutinename, &
!        " str_long='", TRIM(str_long)  , "'", &
!        " str_copy='", TRIM(str_copy) , "'", &
!        " i=", i, &
!        " j=", j, &
!        " k=", k, &
!        " len=", LEN_TRIM(str_copy), &
!        ' fmt=', TRIM(str_fmt), &
!        " => hh=", hh, &
!        " mm=", mm, &
!        " ss=", ss, &
!        "  ==> long=", long

    !--------------------------------------------------------
    ! free up the allocated space
!    if ( allocated(str_copy) ) DEALLOCATE(str_copy)


  end subroutine convert_longitude_to_decimal
  !--------------------------------------------------------


  !--------------------------------------------------------
  ! allow following formats
  ! +DD.DDDDD...
  ! +DD:AM:AS.SSS...
  ! +DD AM AS.SSS...
  ! +DDdAM'AS.SSS..."
  ! +DDdAMmAS.SSS...s
  recursive subroutine convert_latitude_to_decimal (str_lat, lat, success_flag)
    character(len=*), parameter :: subroutinename='convert_latitude_to_decimal '
    character(len=*),intent(in) :: str_lat
    real(kind=real64),intent(out) :: lat
    logical,intent(out) ::  success_flag
!    character(len=:), allocatable :: str_copy
    character(len=LSTR_MAX) :: str_copy
    character(len=STR_MAX) :: str_fmt
    integer(kind=int32)  :: status, i, j, k, flag
    integer(kind=int32)  :: sign, dd, mm
    real(kind=real64)  :: ss
    real(kind=real64)  :: val
    success_flag = .false.
    lat = -99.9d0
    flag = 0
    status = 0
    i=-1
    j=-1
    k=-1

    if ( LEN_TRIM(ADJUSTL(str_lat)) .gt. len(str_copy)) then
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: input string is too long")
      return
    endif

    if ( LEN_TRIM(ADJUSTL(str_lat)) .lt. 1) then
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: input string is too short")
      return
    endif

    ! take a copy of the input string and convert to lowercase
    str_copy = TRIM(ADJUSTL(str_lat))
    call to_lower(str_copy)


    !--------------------------------------------------------
    ! now check if there is a leading sign character
    ! get rid of the sign character if present
    if ( str_copy(1:1) .eq. '-' ) then
      sign = -1
      str_copy(1:) = str_copy(2:)
    else if ( str_copy(1:1) .eq. '+' ) then
      sign = +1
      str_copy(1:) = str_copy(2:)
    else
      sign = +1
    endif
    !--------------------------------------------------------

    !--------------------------------------------------------
    ! check the length again
    if ( LEN_TRIM(ADJUSTL(str_copy)) .lt. 1) then
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: input string is too short")
      return
    endif
    !--------------------------------------------------------


    !--------------------------------------------------------
    ! Now try to read as (+)DD:AM:AS.SSS...
    ! First find locations of the ':' characters
    if ( flag .eq. 0 ) then
      i = index(str_copy,":")   ! 1st one
      if ( i .gt. 0 ) then
        j = index(str_copy(i+1:LEN_TRIM(str_copy)),":") ! 2nd one
        if ( j .gt. 0 ) then
!          k = len(str_copy(i+1+j+1:LEN_TRIM(str_copy)))
          k = len(str_copy(i+1+j:LEN_TRIM(str_copy)))
          write (str_fmt, '(A,I0,A,I0,A,I0,A)') &
              '(I', i-1, ',1x,I', j-1, ',1x,F', k, '.0)'
          read (str_copy, TRIM(str_fmt), iostat=status) dd,mm,ss

          if ( status .eq. 0 ) then
            flag = 1
          endif
        endif
      endif
    end if
    !--------------------------------------------------------

    !--------------------------------------------------------
    ! Now try to read as (+)DD AM AS.SSS
    ! First find locations of the ' ' characters
    if ( flag .eq. 0 ) then
      i = index(str_copy," ")   ! 1st one
      if ( i .gt. 0 ) then
        j = index(str_copy(i+1:LEN_TRIM(str_copy))," ") ! 2nd one
        if ( j .gt. 0 ) then
!          k = len(str_copy(i+1+j+1:LEN_TRIM(str_copy)))
          k = len(str_copy(i+1+j:LEN_TRIM(str_copy)))
          write (str_fmt, '(A,I0,A,I0,A,I0,A)') &
              '(I', i-1, ',1x,I', j-1, ',1x,F', k, '.0)'
          read (str_copy, TRIM(str_fmt), iostat=status) dd,mm,ss

          if ( status .eq. 0 ) then
            flag = 1
          endif

        endif
      endif
    end if
    !--------------------------------------------------------

    !--------------------------------------------------------
    ! Now try to read as (+)DDdAM'AS.SSS..."
    ! First find locations of the 'd,',"' characters
    if ( flag .eq. 0 ) then
      i = index(str_copy,"d")   ! 1st one
      if ( i .gt. 0 ) then
        j = index(str_copy(i+1:LEN_TRIM(str_copy)),"'") ! 2nd one
        if ( j .gt. 0 ) then
          k = index(str_copy(i+1+j:LEN_TRIM(str_copy)),'"') ! 3rd one
          if ( k .gt. 0 ) then
            write (str_fmt, '(A,I0,A,I0,A,I0,A)') &
                '(I', i-1, ',1x,I', j-1, ',1x,F', k-1, '.0,1x)'
            read (str_copy, TRIM(str_fmt), iostat=status) dd,mm,ss
            if ( status .eq. 0 ) then
              flag = 1
            endif
          endif
        endif
      endif
    endif

    !--------------------------------------------------------
    ! Now try to read as (+)DDdAMmAS.SSS...s
    ! First find locations of the 'd,m,s' characters
    if ( flag .eq. 0 ) then
      i = index(str_copy,"d")   ! 1st one
      if ( i .gt. 0 ) then
        j = index(str_copy(i+1:LEN_TRIM(str_copy)),"m") ! 2nd one
        if ( j .gt. 0 ) then
          k = index(str_copy(i+1+j:LEN_TRIM(str_copy)),'s') ! 3rd one
          if ( k .gt. 0 ) then
            write (str_fmt, '(A,I0,A,I0,A,I0,A)') &
                '(I', i-1, ',1x,I', j-1, ',1x,F', k-1, '.0,1x)'
            read (str_copy, TRIM(str_fmt), iostat=status) dd,mm,ss
            if ( status .eq. 0 ) then
              flag = 1
            endif
          endif
        endif
      endif
    endif

    if ( flag .eq. 1 ) then
      if (dd .ge. 0     .and. dd .le. 90 .and. &
          mm .ge. 0     .and. mm .lt. 60 .and. &
          ss .ge. 0.0d0 .and. ss .lt. 60.0d0 ) then
        lat = REAL(sign,real64) * (dd + &
            mm * ARCMIN2DEG + &
            ss * ARCSEC2DEG )
        success_flag = .true.
      endif
    endif


    !--------------------------------------------------------
    ! attempt to read the string as a real
    if ( flag .eq. 0 ) then
      write (str_fmt, '(A,I0,A)') '(F', len(str_copy), '.0)'
      read (str_copy, TRIM(str_fmt), iostat=status) val
      if ( status .eq. 0 ) then
        lat =  REAL(sign,real64) * val
        success_flag = .true.
        flag = 1
      endif
    endif
    !--------------------------------------------------------

!    write (*,'(A,2(A,A,A),4(A,I0),A,A,3(A,I0),A,F12.5,A,F15.9)'), &
!        subroutinename, &
!        " str_lat='",  TRIM(str_lat)  , "'", &
!        " str_copy='", TRIM(str_copy) , "'", &
!        " i=", i, &
!        " j=", j, &
!        " k=", k, &
!        " len=", LEN_TRIM(str_copy), &
!        ' fmt=', TRIM(str_fmt), &
!        " => sign=",  sign, &
!        " dd=",  dd, &
!        " mm=",  mm, &
!        " ss=",  ss, &
!        "  ==> lat=", lat

    !--------------------------------------------------------
    ! free up the allocated space
!    if ( ALLOCATED(str_copy) ) DEALLOCATE(str_copy)

  end subroutine convert_latitude_to_decimal
  !--------------------------------------------------------


  !--------------------------------------------------------
  ! convert a string into a pair of decimal coordinates
  !
  ! Accept a one line description of the coordinates
  ! in one of the following formats:
  !   a) [coordsys;] LONGITUDE_DEG , LATITUDE_DEG
  !   b) [coordsys;] LONGITUDE_H:LONGITUDE_M:LONGITUDE_S , LATITUDE_D:LATITUDE_M:LATITUDE_S
  !   c) [coordsys;] LONGITUDE_H LONGITUDE_M LONGITUDE_S , LATITUDE_D LATITUDE_M LATITUDE_S
  !---------------------------------------------------------
  recursive subroutine convert_longlat_to_decimal (str_longlat, long, lat, coordsys_code, success_flag)
    character(len=*), parameter :: subroutinename='convert_longlat_to_decimal'
    character(len=*),intent(in) :: str_longlat
    real(kind=real64),intent(out) :: long
    real(kind=real64),intent(out) :: lat
    integer(kind=int32),intent(out) ::  coordsys_code
    logical,intent(out) ::  success_flag
!    character(len=:), allocatable :: str_copy
    character(len=LSTR_MAX) :: str_copy
    integer(kind=int32)  :: i
    integer(kind=int32)  :: status
    integer(kind=int32)  :: length_longlat


    status = 0
    success_flag = .false.
    length_longlat = LEN_TRIM(ADJUSTL(str_longlat))


    if ( length_longlat .gt. LEN(str_copy)) then
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: input string is too long" )
      return
    endif

    if ( length_longlat .lt. 3) then
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: input string is too short")
      return
    endif

    ! take a copy of the input string and convert it to lowercase
    str_copy = TRIM(ADJUSTL(str_longlat))
    call to_lower(str_copy)

    ! step 1 : strip off any comments after a "#" symbol
    i = INDEX(str_copy,"#")
    if ( i > 0 ) then
      str_copy(i:) = " "  ! convert to white space
    end if

    ! step 2 : look for coordinate system string
    ! look for first ';' character
    i = index(str_copy,";")
    if ( i > 1 ) then       ! convert the part before the ';'
      coordsys_code = interpret_coordsys(TRIM(str_copy(1:i-1)))
      if ( coordsys_code .eq. COORDSYS_CODE_ERROR ) then
        call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: interpreting coordinate system")
        return
      endif
    else
      ! leave the coordsys code as it is
!      coordsys_code = COORDSYS_CODE_ICRS ! default to icrs; format
      coordsys_code = COORDSYS_CODE_NULL ! no coordsys supplied...this is ok
    end if
    ! blank out any stuff before the ';' and then get rid of any leading space
    str_copy(1:i) = " "
    str_copy = TRIM(ADJUSTL(str_copy))


    ! step 3 : look for longitude coordinate string
    ! look for first ',' character
    i = INDEX(str_copy,",")
    if ( i > 1 ) then       ! try to convert the part before the ','
      call convert_longitude_to_decimal (TRIM(str_copy(1:i-1)), long, success_flag)
      if ( .not. success_flag ) then
        call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: interpreting longitude coordinate" )
        return
      end if
    else
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: Interpreting longitude coordinate")
      return
    end if
    ! blank out any stuff before the first ',' and then get rid of any leading space
    str_copy(1:i) = " "
    str_copy = TRIM(ADJUSTL(str_copy))

    ! step 4 : look for latitude coordinate string in remainder
    ! look for first ',' character
    i = LEN_TRIM(str_copy)
    if ( i > 1 ) then       ! try to convert the remainder
      call convert_latitude_to_decimal(TRIM(str_copy), lat, success_flag)
      if ( .not. success_flag ) then
        call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: interpreting latitude coordinate ")
        return
      end if
    else
      call wrserr(subroutinename, ERROR_LEVEL_SEVERE, "Error: Interpreting latitude coordinate")
      return
    end if

    ! exit successfully so can now free up the allocated space
!    if ( ALLOCATED(str_copy) ) DEALLOCATE(str_copy)

  end subroutine convert_longlat_to_decimal
  !--------------------------------------------------

  ! TODO It would be better to do these comparisons in a
  ! purely case-insensitive way
  pure recursive function interpret_coordsys ( string) result(code)
    character(len=*), parameter ::functionname='interpret_coordsys'
    character(len=*), intent(in)     :: string
    integer(kind=int32) :: code
    code = COORDSYS_CODE_NULL
    select case ( TRIM(string) )
    case ( "icrs","ICRS", "j2000","J2000", "fk5","FK5" )
      code = COORDSYS_CODE_ICRS
    case ( "fk4","b1950", "FK4","B1950" )
      code = COORDSYS_CODE_FK4
    case ( "gal","galactic","Gal","GAL","GALACTIC","Galactic" )
      code = COORDSYS_CODE_GAL
    case ( "ecl","ecliptic","ECL","Ecl", "ECLIPTIC","Ecliptic" )
      code = COORDSYS_CODE_ECL
    case ( "image", "IMAGE", "Image" )
      code = COORDSYS_CODE_IMAGE
    case ( "physical", "PHYSICAL", "Physical" )
      code = COORDSYS_CODE_PHYSICAL
    case ( "unitvec", "UNITVEC", "Unitvec" )
      code = COORDSYS_CODE_UNITVEC
    case  default
      code = COORDSYS_CODE_ERROR
!      write (*, '(A,A,A)'), functionname, "Unknown coordinate system: '",  TRIM(string), "'"
    end select

  end function interpret_coordsys




  !--------------------------------------------------------
  ! http://en.wikipedia.org/wiki/Orthographic_projection_%28cartography%29
  ! oa_x,oa_y give the pixel coords of the telescope optical axis
  ! inv_pixscale_* are in units of pix/deg
  ! long0, lat0, roll are sky coords of optical axis of telescope, in deg
  ! roll: in evatt, the relation between sky axes (with X pointing west and Y pointing north) and the RAWX/Y axes projected on the celestial sphere is such that a positive roll angle corresponds to a clockwise rotation of the projected RAWX/Y on the sky, the zero point corresponding to RAWX pointing south and RAWY pointing west.
  ! long,lat are the nominated sky position, in same system, in degrees
  ! x,y receive the on-chip pixel position
  recursive subroutine longlat2xy (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, long0, lat0, roll, long, lat, x, y )
    real(kind=real64),intent(in)  :: oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, long0, lat0, roll, long, lat
    real(kind=real64),intent(out) :: x, y

    real(kind=real64) :: cosa,sina, coslat,sinlat, coslat0,sinlat0, coslong_long0, sinlong_long0, xx,yy, cosc
    real(kind=real64) :: inv_pixscale_fromrad_x, inv_pixscale_fromrad_y, long_long0

    coslat0   = cosd(lat0)
    sinlat0   = sind(lat0)
    cosa      = cosd(roll)
    sina      = sind(roll)
    coslat    = cosd(lat)
    sinlat    = sind(lat)
    long_long0 = long - long0
    coslong_long0 = cosd(long_long0)
    sinlong_long0 = sind(long_long0)

    inv_pixscale_fromrad_x = RAD2DEG * inv_pixscale_x
    inv_pixscale_fromrad_y = RAD2DEG * inv_pixscale_y

    ! check that we are not on the opposite side of the sphere
    cosc = sinlat0*sinlat + coslat0*coslat*coslong_long0
    if ( cosc .lt. 0.0d0 ) then
      x = huge(x)
      y = huge(y)
    else
      ! use a tangent plane/orthographic projection
      xx = inv_pixscale_fromrad_x * coslat*sinlong_long0
      yy = inv_pixscale_fromrad_y * (coslat0*sinlat - sinlat0*coslat*coslong_long0)

      ! now apply the rotation and offsets
      x = oa_x + (sina*xx - cosa*yy)
      y = oa_y - (cosa*xx + sina*yy)
    end if

  end subroutine longlat2xy


  recursive subroutine longlat2xy_fast (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
      long0, lat0, roll, long, coslat, sinlat,x, y )
    real(kind=real64),intent(in)  :: oa_x, oa_y, inv_pixscale_x, inv_pixscale_y
    real(kind=real64),intent(in)  :: long0, lat0, roll, long, coslat, sinlat
    real(kind=real64),intent(out) :: x, y

    real(kind=real64) :: cosa,sina, coslat0,sinlat0, coslong_long0, sinlong_long0, xx,yy, cosc
    real(kind=real64) :: inv_pixscale_fromrad_x, inv_pixscale_fromrad_y, long_long0

    coslat0   = cosd(lat0)
    sinlat0   = sind(lat0)
    cosa      = cosd(roll)
    sina      = sind(roll)
    long_long0 = long - long0
    coslong_long0 = cosd(long_long0)
    sinlong_long0 = sind(long_long0)

    inv_pixscale_fromrad_x =  RAD2DEG * inv_pixscale_x
    inv_pixscale_fromrad_y =  RAD2DEG * inv_pixscale_y

    ! check that we are not on the opposite side of the sphere
    cosc = sinlat0*sinlat + coslat0*coslat*coslong_long0
    if ( cosc .lt. 0.0d0 ) then
      x = huge(x)
      y = huge(y)
    else
      ! use a tangent plane/orthographic projection
      xx = inv_pixscale_fromrad_x * coslat*sinlong_long0
      yy = inv_pixscale_fromrad_y * (coslat0*sinlat - sinlat0*coslat*coslong_long0)

      ! now apply the rotation and offsets
      x = oa_x + (sina*xx - cosa*yy)
      y = oa_y - (cosa*xx + sina*yy)
    end if

  end subroutine longlat2xy_fast



  recursive subroutine longlat2xy_faster (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
      long0, coslat0, sinlat0, cosroll, sinroll, long, coslat, sinlat,x, y )
    real(kind=real64),intent(in)  :: oa_x, oa_y, inv_pixscale_x, inv_pixscale_y
    real(kind=real64),intent(in)  :: long0, coslat0, sinlat0, cosroll, sinroll, long, coslat, sinlat
    real(kind=real64),intent(out) :: x, y

    real(kind=real64) :: coslong_long0, sinlong_long0, xx,yy, cosc
    real(kind=real64) :: inv_pixscale_fromrad_x, inv_pixscale_fromrad_y, long_long0

    long_long0 = long - long0
    coslong_long0 = cosd(long_long0)
    sinlong_long0 = sind(long_long0)

    inv_pixscale_fromrad_x =  RAD2DEG * inv_pixscale_x
    inv_pixscale_fromrad_y =  RAD2DEG * inv_pixscale_y

    ! check that we are not on the opposite side of the sphere
    cosc = sinlat0*sinlat + coslat0*coslat*coslong_long0
    if ( cosc .lt. 0.0d0 ) then
      x = huge(x)
      y = huge(y)
    else
      ! use a tangent plane/orthographic projection
      xx = inv_pixscale_fromrad_x * coslat*sinlong_long0
      yy = inv_pixscale_fromrad_y * (coslat0*sinlat - sinlat0*coslat*coslong_long0)

      ! now apply the rotation and offsets
      x = oa_x + (sinroll*xx - cosroll*yy)
      y = oa_y - (cosroll*xx + sinroll*yy)
    end if

  end subroutine longlat2xy_faster


  ! same as above but for an array of coords
  recursive subroutine longlats2xys (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, long0, lat0, roll, long, lat, x, y )
    character(len=*), parameter :: subroutinename='longlats2xys'
    real(kind=real64),intent(in)  :: oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, long0, lat0, roll
    real(kind=real64),intent(in),dimension(:)  :: long, lat
    real(kind=real64),intent(out),dimension(:) :: x, y

    real(kind=real64) :: cosa,sina, coslat,sinlat, coslat0,sinlat0, coslong_long0, sinlong_long0, xx,yy, cosc
    real(kind=real64) :: inv_pixscale_fromrad_x, inv_pixscale_fromrad_y, longi_long0
    integer(int32) :: i,dim

    if (size(long) /= size(lat) .or. &
        size(long) .gt. size(x) .or. &
        size(long) .gt. size(y) ) then
      write (*,'(A,1x,A)') subroutinename, "Warning: Array dimensions are mismatched!"
    endif
    dim = min(size(long),size(lat),size(x),size(y))

    coslat0   = cosd(lat0)
    sinlat0   = sind(lat0)
    cosa      = cosd(roll)
    sina      = sind(roll)
!    call sincos (lat0, sinlat0, coslat0)   ! unfortunately these do not exist for direct calls, but are
!    call sincos (roll, sina, cosa)         ! used when compiler optimisation switches are turned on
    inv_pixscale_fromrad_x =  RAD2DEG * inv_pixscale_x
    inv_pixscale_fromrad_y =  RAD2DEG * inv_pixscale_y

    do i=1,dim
      coslat    = cosd(lat(i))
      sinlat    = sind(lat(i))
      longi_long0 = long(i) - long0
      coslong_long0 = cosd(longi_long0)
      sinlong_long0 = sind(longi_long0)

      ! check that we are not on the opposite side of the sphere
      cosc = sinlat0*sinlat + coslat0*coslat*coslong_long0
      if ( cosc .lt. 0.0d0 ) then
        x(i) = huge(x(i))
        y(i) = huge(y(i))
      else

        ! use a tangent plane/orthographic projection
        xx = inv_pixscale_fromrad_x * (                         coslat*sinlong_long0)
        yy = inv_pixscale_fromrad_y * (coslat0*sinlat - sinlat0*coslat*coslong_long0)

        ! now apply the rotation and offsets
        x(i) = oa_x + (sina*xx - cosa*yy)
        y(i) = oa_y - (cosa*xx + sina*yy)
      end if
    end do
  end subroutine longlats2xys


  ! same as above but for an array of coords
  recursive subroutine longlats2xys_fast (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
      long0, lat0, roll, long, coslat, sinlat, x, y )
    character(len=*), parameter :: subroutinename='longlats2xys'
    real(kind=real64),intent(in)  :: oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, long0, lat0, roll
    real(kind=real64),intent(in),dimension(:)  :: long
    real(kind=real64),intent(in),dimension(:)  :: coslat, sinlat
    real(kind=real64),intent(out),dimension(:) :: x, y

    real(kind=real64) :: cosa,sina,  coslat0,sinlat0, coslong_long0, sinlong_long0, xx,yy, cosc
    real(kind=real64) :: inv_pixscale_fromrad_x, inv_pixscale_fromrad_y, longi_long0
    integer(int32) :: i,dim

    if (size(long) /= size(coslat) .or. &
        size(long) /= size(sinlat) .or. &
        size(long) .gt. size(x) .or. &
        size(long) .gt. size(y) ) then
      write (*,'(A,1x,A)') subroutinename, "Warning: Array dimensions are mismatched!"
    endif
    dim = min(size(long),size(coslat),size(sinlat),size(x),size(y))

    coslat0   = cosd(lat0)
    sinlat0   = sind(lat0)
    cosa      = cosd(roll)
    sina      = sind(roll)
!    call sincos (lat0, sinlat0, coslat0)   ! unfortunately these do not exist for direct calls, but are
!    call sincos (roll, sina, cosa)         ! used when compiler optimisation switches are turned on
    inv_pixscale_fromrad_x =  RAD2DEG * inv_pixscale_x
    inv_pixscale_fromrad_y =  RAD2DEG * inv_pixscale_y

    do i=1,dim
!      coslat    = cosd(lat(i))
!      sinlat    = sind(lat(i))
      longi_long0 = long(i) - long0
      coslong_long0 = cosd(longi_long0)
      sinlong_long0 = sind(longi_long0)

      ! check that we are not on the opposite side of the sphere
      cosc = sinlat0*sinlat(i) + coslat0*coslat(i)*coslong_long0
      if ( cosc .lt. 0.0d0 ) then
        x(i) = huge(x(i))
        y(i) = huge(y(i))
      else

        ! use a tangent plane/orthographic projection
        xx = inv_pixscale_fromrad_x * (                         coslat(i)*sinlong_long0)
        yy = inv_pixscale_fromrad_y * (coslat0*sinlat(i) - sinlat0*coslat(i)*coslong_long0)

        ! now apply the rotation and offsets
        x(i) = oa_x + (sina*xx - cosa*yy)
        y(i) = oa_y - (cosa*xx + sina*yy)
      end if
    end do
  end subroutine longlats2xys_fast




  !-------------------------------------------------------------------------------
  ! Assumes a projection with: RA---SIN,DEC--SIN
  ! http://en.wikipedia.org/wiki/Orthographic_projection_%28cartography%29
  ! oa_x,oa_y give the pixel coords of the telescope optical axis
  ! pixscale_* are in units of deg/pix
  ! long0, lat0, roll are sky coords of optical axis of telescope, in deg
  ! roll is angle between North and +ve X direction, measured counter clockwise
  ! x,y define the nominated on-chip pixel position
  ! long,lat are the returned sky position, in same system, in degrees

  recursive subroutine xy2longlat (oa_x, oa_y, pixscale_x, pixscale_y,long0, lat0, roll, x, y, long, lat )
    real(kind=real64),intent(in)  :: oa_x, oa_y, pixscale_x, pixscale_y, long0, lat0, roll
    real(kind=real64),intent(in)  :: x, y
    real(kind=real64),intent(out) :: long, lat
    real(kind=real64) :: cosa,sina, xp,yp,xx,yy, rho2, cosc, coslat0, sinlat0
    sina = sind(-roll)
    cosa = cosd(-roll)

    ! first subtract the offsets and multiply by pixel scale
    xp = (x - oa_x)*pixscale_x
    yp = (y - oa_y)*pixscale_y

    if ( eq_real64(xp,0.0d0) .and. eq_real64(yp,0.0d0) ) then
      long = long0
      lat = lat0
    else

      ! now apply the rotation and convert to radians
      xx =  DEG2RAD*(sina*xp - cosa*yp)
      yy = -DEG2RAD*(cosa*xp + sina*yp)

      rho2 = xx*xx + yy*yy
      if ( rho2 .gt. 0.0d0) then
        cosc = sqrt(1.0d0 - rho2)
        sinlat0 = sind(lat0)
        coslat0 = cosd(lat0)
        lat = asind(cosc*sinlat0 + yy*coslat0)
        long = long0 + atan2d(xx, coslat0*cosc - yy*sinlat0)
      else
        long = long0
        lat = lat0
      endif
    endif

  end subroutine xy2longlat
  !-------------------------------------------------------------------------------




  recursive subroutine xys2longlats (oa_x, oa_y, pixscale_x, pixscale_y,long0, lat0, roll, x, y, long, lat )
    character(len=*), parameter :: subroutinename='xys2longlats'
    real(kind=real64),intent(in)  :: oa_x, oa_y, pixscale_x, pixscale_y, long0, lat0, roll
    real(kind=real64),intent(in),dimension(:)  :: x, y
    real(kind=real64),intent(out),dimension(:) :: long, lat
    real(kind=real64) :: cosa,sina, xp,yp,xx,yy, rho2, cosc, coslat0, sinlat0
    integer(int32) :: i,dim

    if (size(x) /= size(y) .or. &
        size(x) .gt. size(long) .or. &
        size(x) .gt. size(lat) ) then
      write (*,'(A,1x,A)') subroutinename, "Warning: Array dimensions are mismatched!"
    endif
    dim = min(size(x),size(y),size(long),size(lat))

    sinlat0 = sind(lat0)
    coslat0 = cosd(lat0)
    sina = sind(-roll)
    cosa = cosd(-roll)

    do i=1,dim
      ! first subtract the offsets and multiply by pixel scale
      xp = (x(i) - oa_x)*pixscale_x
      yp = (y(i) - oa_y)*pixscale_y

      if ( eq_real64(xp,0.0d0) .and. eq_real64(yp,0.0d0) ) then
        long(i) = long0
        lat(i) = lat0
      else

        ! now apply the rotation and convert to radians
        xx =  DEG2RAD*(sina*xp - cosa*yp)
        yy = -DEG2RAD*(cosa*xp + sina*yp)

        rho2 = xx**2 + yy**2
        if ( rho2 .gt. 0.0d0) then
          cosc = sqrt(1.0d0 - rho2)
          lat(i) = asind(cosc*sinlat0 + yy*coslat0)
          long(i) = long0 + atan2d(xx, coslat0*cosc - yy*sinlat0)
        else
          long(i) = long0
          lat(i) = lat0
        endif
      endif
    end do

  end subroutine xys2longlats
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !------- General image WCS<->pix projection routines ---------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------


!  ! assumes an RA---SIN,DEC--SIN projection (Orthographic) , starting from intermediate world coordinates xx,yy in degrees
!  ! http://en.wikipedia.org/wiki/Orthographic_projection_%28cartography%29
!  recursive subroutine intwcs2longlat_sin(long0, lat0, xx, yy, long, lat )
!    real(kind=real64),intent(in)  :: long0, lat0, xx, yy
!    real(kind=real64),intent(out) :: long, lat
!    real(kind=real64) :: xr, yr, rho2, rho, c, cosc, sinc,  coslat0, sinlat0
!
!    xr = xx*DEG2RAD
!    yr = yy*DEG2RAD
!    rho2 = (xr*xr + yr*yr)
!    if ( rho2 .gt. 0.0d0) then
!      rho = sqrt(rho2)
!      c = asin(rho)
!      cosc = cos(c)
!      sinc = sin(c)
!      sinlat0 = sind(lat0)
!      coslat0 = cosd(lat0)
!
!      lat = asind(cosc*sinlat0 + yr*sinc*coslat0/rho)
!      long = long0 + atan2d(xr*sinc, rho*coslat0*cosc - yr*sinlat0*sinc)
!    else
!      long = long0
!      lat = lat0
!    endif
!
!  end subroutine intwcs2longlat_sin
!
!

  ! deprojects from an RA---SIN,DEC--SIN projection (Orthographic),
  ! starting from long,lat and ending in intermediate world coordinates xx,yy (all in degrees)
  recursive subroutine wcs_world2iworld_sin(wcs, long, lat, xx, yy)
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in) :: long, lat     ! in degrees
    real(kind=real64),intent(out)  :: xx, yy      ! in degrees
    real(kind=real64) ::cosc, coslat, sinlat, coslong_long0, sinlong_long0 ! xr, yr, rho2,cosc, c!, sinc, cosc

    coslat = cosd(lat)
    sinlat = sind(lat)
    coslong_long0 = cosd(long-wcs%crval(1))
    sinlong_long0 = sind(long-wcs%crval(1))
    ! compute cosc and check that we are not on the opposite side of the sphere
    cosc = wcs%sin_crval2*sinlat + wcs%cos_crval2*coslat*coslong_long0
    if ( cosc .lt. 0.0d0 ) then
      xx = huge(xx)
      yy = huge(yy)
    else
      xx = RAD2DEG* coslat*sinlong_long0
      yy = RAD2DEG* (wcs%cos_crval2*sinlat - wcs%sin_crval2*coslat*coslong_long0)
    endif
  end subroutine wcs_world2iworld_sin

  ! deprojects from an RA---TAN,DEC--TAN projection (Gnomonic),
  ! starting from long,lat and ending in intermediate world coordinates xx,yy (all in degrees)
  recursive subroutine wcs_world2iworld_tan(wcs, long, lat, xx, yy)
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in) :: long, lat     ! in degrees
    real(kind=real64),intent(out)  :: xx, yy      ! in degrees
    real(kind=real64) ::cosc, coslat, sinlat, coslong_long0, sinlong_long0 !

    coslat = cosd(lat)
    sinlat = sind(lat)
    coslong_long0 = cosd(long-wcs%crval(1))
    sinlong_long0 = sind(long-wcs%crval(1))
    ! compute cosc and check that we are not on the opposite side of the sphere
    cosc = wcs%sin_crval2*sinlat + wcs%cos_crval2*coslat*coslong_long0
    if ( cosc .le. 0.0d0 ) then
      xx = huge(xx)
      yy = huge(yy)
    else
      xx = (RAD2DEG/cosc)*coslat*sinlong_long0
      yy = (RAD2DEG/cosc)*(wcs%cos_crval2*sinlat - wcs%sin_crval2*coslat*coslong_long0)
    endif
  end subroutine wcs_world2iworld_tan


  recursive subroutine wcs_iworld2pix(wcs, xx, yy, x, y)
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in) :: xx, yy
    real(kind=real64),intent(out)  :: x, y
    real(kind=real64)  :: xp, yp

    ! go from intermediate world coordinates to x
    xp = (wcs%inv_cdi_j(1,1)*xx + wcs%inv_cdi_j(1,2)*yy)
    ! go from intermediate world coordinates to y
    yp = (wcs%inv_cdi_j(2,1)*xx + wcs%inv_cdi_j(2,2)*yy)

    x = xp + wcs%crpix(1)
    y = yp + wcs%crpix(2)
  end subroutine wcs_iworld2pix


  recursive subroutine wcs_world2pix (wcs, long, lat, x, y  )
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in) :: long, lat
    real(kind=real64),intent(out)  :: x, y
    real(kind=real64) :: xx,yy


    ! de project using given projection type
    select case (wcs%projtype)
    case(PROJTYPE_CODE_SIN)
      call wcs_world2iworld_sin(wcs, long, lat, xx, yy )
    case(PROJTYPE_CODE_TAN)
      call wcs_world2iworld_tan(wcs, long, lat, xx, yy )
    case default
      x = -huge(1.0d0)
      y  = -huge(1.0d0)
      return
    end select
!    write (*, '(A,2(F12.7,x),A,2(F12.6,x))') "wcs_world2pix: (long,lat)= ", long, lat, " -> (XX,YY)= ", xx, yy

    ! convert from intermediate world coordinates
    call wcs_iworld2pix(wcs, xx, yy, x, y)

  end subroutine wcs_world2pix

  !----------------------------------------------------------------------
  !----------------------------------------------------------------------
  !----------------------------------------------------------------------
  !----------------------------------------------------------------------
  !----------------------------------------------------------------------
  !----------------------------------------------------------------------
  !----------------------------------------------------------------------
  !----------------------------------------------------------------------
  !----------------------------------------------------------------------

  ! assumes an RA---SIN,DEC--SIN projection (Orthographic) , starting from intermediate world coordinates xx,yy in degrees
  ! http://en.wikipedia.org/wiki/Orthographic_projection_%28cartography%29
  recursive subroutine wcs_iworld2world_sin(wcs, xx, yy, long, lat )
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in)  :: xx, yy        ! in degrees
    real(kind=real64),intent(out) :: long, lat     ! in degrees
    real(kind=real64) :: xr, yr, rho2, cosc !, sinc, rho, c

    xr = xx*DEG2RAD
    yr = yy*DEG2RAD

    rho2 = (xr*xr + yr*yr)
    if ( rho2 .gt. 0.0d0) then
!      rho = sqrt(rho2)
!      c = asin(rho)
!      cosc = cos(c)
!!      sinc = sin(c)
!      sinc = rho
      ! c = asin(rho)
      ! sinc = rho
      ! sin2c = rho2
      ! 1 - cos2c = sin2c = rho2
      ! cosc = sqrt(1 - rho2)
      cosc = sqrt(1.0d0 - rho2)

!      lat = asind(cosc*wcs%sin_crval2 + yr*sinc*wcs%cos_crval2/rho)
!      long = wcs%crval(1) + atan2d(xr*sinc, rho*cosc*wcs%cos_crval2 - yr*sinc*wcs%sin_crval2)

      ! simplify to the following:
      lat = asind(cosc*wcs%sin_crval2 + yr*wcs%cos_crval2)
      long = wcs%crval(1) + atan2d(xr, cosc*wcs%cos_crval2 - yr*wcs%sin_crval2)
    else
      long = wcs%crval(1)
      lat = wcs%crval(2)
    endif
    !write (*, '(A,2(F12.3,x),A,2(F12.7,x))') "wcs_iworld2world_sin: (xx,yy)= ", xx, yy, " -> (long,lat)= ", long, lat
  end subroutine wcs_iworld2world_sin



  ! assumes an RA---TAN,DEC--TAN projection (Gnomonic), starting from intermediate world coordinates xx,yy in degrees
  ! http://mathworld.wolfram.com/GnomonicProjection.html - but the following is simpler:
  ! https://archive.lib.msu.edu/crcmath/math/math/g/g187.htm
  recursive subroutine wcs_iworld2world_tan(wcs, xx, yy, long, lat )
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in)  :: xx, yy        ! in degrees
    real(kind=real64),intent(out) :: long, lat     ! in degrees
    real(kind=real64) :: xr, yr, rho2,cosc !

    xr = xx*DEG2RAD
    yr = yy*DEG2RAD

    rho2 = (xr*xr + yr*yr)
    if ( rho2 .gt. 0.0d0) then
!      c = atan2(sqrt(rho2), 1.0d0)
!      cosc = cos(c)
!      sinc = sin(c)

      ! c = atan(rho)
      ! tanc = rho
      ! tan2c = rho2
      ! tan2c + 1 = sec2c
      ! cosc = 1/sqrt(tan2c + 1)
      cosc = 1.0d0/sqrt(rho2 + 1.0d0)
      ! cot2c + 1 = cosec2c
      ! sin2c = 1/(cot2c + 1)
      ! sin2c = tan2c/(1 + tan2c)
      ! sinc = sqrt(tan2c/(1 + tan2c))
      ! sinc = sqrt(rho2/(1 + rho2))
!      sinc = sqrt(rho2/(1.0d0 + rho2))

!      lat = asind(cosc*wcs%sin_crval2 + yr*sinc*wcs%cos_crval2/sqrt(rho2))

      ! simplify to the following:
      lat = asind(cosc*(wcs%sin_crval2 + yr*wcs%cos_crval2))
      long = wcs%crval(1) + atan2d(xr, wcs%cos_crval2 - yr*wcs%sin_crval2)

    else
      long = wcs%crval(1)
      lat = wcs%crval(2)
    endif
    !write (*, '(A,2(F12.3,x),A,2(F12.7,x))') "wcs_iworld2world_tan: (xx,yy)= ", xx, yy, " -> (long,lat)= ", long, lat


  end subroutine wcs_iworld2world_tan

  recursive subroutine wcs_pix2iworld(wcs, x, y, xx, yy)
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in)  :: x, y
    real(kind=real64),intent(out) :: xx, yy
    real(kind=real64)  :: xp, yp
    xp = x - wcs%crpix(1)
    yp = y - wcs%crpix(2)

    ! go from x,y to intermediate world coordinate axis 1
    xx = wcs%cdi_j(1,1)*xp + wcs%cdi_j(1,2)*yp
    ! go from x,y to intermediate world coordinate axis 2
    yy = wcs%cdi_j(2,1)*xp + wcs%cdi_j(2,2)*yp
    !write (*, '(A,2(F12.3,x),A,2(F12.7,x))') "wcs_pix2iworld: (x,y)=   ", x, y, " -> (XX,YY)=    ", xx, yy

  end subroutine wcs_pix2iworld


  recursive subroutine wcs_pix2world (wcs, x, y, long, lat )
    character(len=*), parameter :: subroutinename='wcs_pix2world'
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in)  :: x, y
    real(kind=real64),intent(out) :: long, lat
    real(kind=real64) :: xx,yy

    ! convert to intermediate world coordinates
    call wcs_pix2iworld(wcs, x, y, xx, yy)

    ! project using given projection type
    select case (wcs%projtype)
    case(PROJTYPE_CODE_SIN)
      call wcs_iworld2world_sin(wcs, xx, yy, long, lat )
    case(PROJTYPE_CODE_TAN)
      call wcs_iworld2world_tan(wcs, xx, yy, long, lat )
    case default
      long = -huge(1.0d0)
      lat  = -huge(1.0d0)
    end select

  end subroutine wcs_pix2world
  !----------------------------------------------------------------------


  recursive function wcs_calc_pixarea_atpixel (wcs, x, y )
    type(wcs_type), intent(in) :: wcs
    real(kind=real64),intent(in)  :: x, y
    real(kind=real64) :: xx1,xx2,yy1,yy2
    real(kind=real64) :: wcs_calc_pixarea_atpixel

    ! calc coords of opposite corners
    call wcs_pix2iworld(wcs, x+0.5d0, y+0.5d0, xx1, yy1)
    call wcs_pix2iworld(wcs, x-0.5d0, y-0.5d0, xx2, yy2)
    ! area is half of the square of the diagonal distance between these points
    wcs_calc_pixarea_atpixel =  0.5d0 * ((xx2-xx1)**2 + (yy2-yy1)**2)

  end function wcs_calc_pixarea_atpixel


  recursive subroutine wcs_init (wcs)
    type(wcs_type), intent(inout) :: wcs
    wcs%projtype = PROJTYPE_CODE_NULL
    wcs%cdi_j(1,1) = 1.0d0
    wcs%cdi_j(1,2) = 0.0d0
    wcs%cdi_j(2,1) = 0.0d0
    wcs%cdi_j(2,2) = 1.0d0
    wcs%inv_cdi_j(:,:) = wcs%cdi_j(:,:)
    wcs%crval(:) = huge(1.0d0)
    wcs%crpix(:) = huge(1.0d0)
    wcs%sin_crval2 = huge(1.0d0)
    wcs%cos_crval2 = huge(1.0d0)
  end subroutine wcs_init
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------















  !-------------------------------------------------------------------------------
  ! get the trailing filename part of a full filename+path
  ! remove any leading ! characters
  ! assume that everything after the last / or ! character is the filename
  ! assume that the filename string has enough space to hold the result
  recursive subroutine pathtools_extract_tail (path, tail)
    character(len=*), intent(in)   :: path
    character(len=*), intent(out)  :: tail
!    integer(kind=int32) :: j
!    integer(kind=int32) :: k
    integer(kind=int32) :: i

    tail = ""

    if ( len(path) .le. 0 ) then
      return      ! exit silently if path is blank
    endif

!    j = index(path,'/',.true.)
!    k = index(path,'!',.true.)
    i = MAX(index(path,'/',.true.),index(path,'!',.true.))

    if ( i .le. 0) then
      tail = path
    else if ( i .ge. len(path) ) then
      tail = ""
    else
      tail = path(i+1:)
    endif

  end subroutine pathtools_extract_tail



  !-------------------------------------------------------------------------------
  ! make all necessary directories so that we can create a new file
  ! return status > 0 if unsuccessful
  ! TODO - tilde expansion
  recursive subroutine make_necessary_dirs_for_new_file (filename, status)
    character(len=*), parameter :: subroutinename='make_necessary_dirs_for_new_file():'
    character(len=*),     intent(in)   :: filename
    integer(kind=int32),  intent(out)   :: status
    character(len=STR_MAX) :: str_home
    integer(kind=int32) :: i,j,k
    logical             :: exist_flag

    status = 0

    if ( LEN_TRIM(filename) .le. 0 ) then
      return     ! exit silently if filename is blank
    endif

    ! check for a '~/' sequence - expand to $HOME
    j = index(filename,'~/')
    if ( j .gt. 0 ) then
      ! get the home environment variable
      call GET_ENVIRONMENT_VARIABLE(NAME='HOME', VALUE=str_home,STATUS=status)
      if ( status /= 0 ) then
        write (*, '(A,1x,A,A,A,I0)') subroutinename, &
            "Error when expanding tilde in: '", filename ,"' status=", status
        return
      endif

      status = 1
      write (*, '(A,1x,A,A,A,I0)') subroutinename, &
         "Error, cannot handle tilde(~) expansion for: '", filename ,"' status=", status
     return
    else

    endif

!    ALLOCATE( character(len=len(filename)+LEN_TRIM(str_home)) :: str_copy(1), STATUS=status)
!    ALLOCATE( character(len=) :: str_copy(1), STAT=status)



    ! check for '/' characters
    i = 1
    do while ( i .le. len(filename) )
      j = index(filename(i:),'/')
      k = i + j - 1
!      write (*, '(A,A,A)'), "Directory name: '", filename(:k) ,"'"
      if ( j .eq. 0 ) then
        ! no more directories were found - so finish
        return
      else
        ! we found a '/' character, so attempt to make the directory if it doesn't already exist
        inquire (FILE=filename(:k), IOSTAT=status, exist=exist_flag)
        if ( status /= 0 ) then
          write (*, '(A,1x,A,A,A,I0)') subroutinename, &
              "Error when inquiring about: '", filename(:k) ,"' status=", status
          return
        endif
        if ( .not. exist_flag ) then
          ! try to create the directory
!          write (*, '(A,A,A)'), "Creating directory: '", filename(:k) ,"'"

          status = system( 'mkdir '//filename(:k) )
          if ( status /= 0 ) then
            write (*, '(A,1x,A,A,A,I0)') subroutinename, &
                "Error when creating directory: '", filename(:k) ,"' status=", status
            return
          endif
        else
!          write (*, '(A,A,A)'), "Existing directory: '", filename(:k) ,"'"
        endif
        i = k + 1
      endif
    end do

  end subroutine make_necessary_dirs_for_new_file


  !--------------------------------------------------------------------
  ! http://en.wikipedia.org/wiki/Bilinear_interpolation
  ! TODO test, and check if procedure for x1==x2 || y1==y2 is ok
  recursive subroutine bilinear_interp_real64 (v1_in, v2_in, f_in, v_eval, f_eval)
    real(kind=real64), dimension(2), intent(in) ::  v1_in, v2_in, v_eval
    real(kind=real64), dimension(2,2), intent(in) ::  f_in
    real(kind=real64), intent(out) ::  f_eval

    real(kind=real64) :: norm

    ! assume that v_out is bracketed by v1_in and v2_in
    norm = 1.0d0
    if ( .not. eq_real(v2_in(1),v1_in(1)) ) then
      norm = norm / ((v2_in(1) - v1_in(1)))
    endif

    if ( .not. eq_real(v2_in(2),v1_in(2)) ) then
      norm = norm / ((v2_in(2) - v1_in(2)))
    endif

    f_eval = norm * ( &
        f_in(1,1)*(v2_in(1) - v_eval(1))*(v2_in(2) - v_eval(2)) + &
        f_in(2,1)*(v_eval(1) - v1_in(1))*(v2_in(2) - v_eval(2)) + &
        f_in(1,2)*(v2_in(1) - v_eval(1))*(v_eval(2) - v1_in(2)) + &
        f_in(2,2)*(v_eval(1) - v1_in(1))*(v_eval(2) - v1_in(2)))

  end subroutine bilinear_interp_real64
  !--------------------------------------------------------------------


  !--------------------------------------------------------------------
  !http://en.wikipedia.org/wiki/Trilinear_interpolation
  recursive subroutine trilinear_interp_real64 (v1_in, v2_in, f_in, v_eval, f_eval)
    real(kind=real64), dimension(3), intent(in) ::  v1_in, v2_in, v_eval
    real(kind=real64), dimension(2,2,2), intent(in) ::  f_in
    real(kind=real64), intent(out) ::  f_eval

    real(kind=real64), dimension(3)   :: vd
    real(kind=real64), dimension(2,2) :: cc
    real(kind=real64), dimension(2)   :: c
    integer(kind=int32) :: i

    ! assume that v_out is bracketed by v1_in and v2_in
    ! and that  v2_in(:) != v1_in(:)
    do i=1,3
      if ( eq_real(v2_in(i),v1_in(i))) then
        vd(i) = 0.0d0
      else
        vd(i) = (v_eval(i) - v1_in(i)) / (v2_in(i) - v1_in(i))
      end if
    end do
    cc(1,1) = f_in(1,1,1)*(1.0d0 - vd(1)) + f_in(2,1,1)*vd(1)
    cc(1,2) = f_in(1,2,1)*(1.0d0 - vd(1)) + f_in(2,2,1)*vd(1)
    cc(2,1) = f_in(1,1,2)*(1.0d0 - vd(1)) + f_in(2,1,2)*vd(1)
    cc(2,2) = f_in(1,2,2)*(1.0d0 - vd(1)) + f_in(2,2,2)*vd(1)

    c(1)    = cc(1,1)*(1.0d0 - vd(2)) + cc(2,1)*vd(2)
    c(2)    = cc(1,2)*(1.0d0 - vd(2)) + cc(2,2)*vd(2)

    f_eval  = c(1)*(1.0d0 - vd(3))    + c(2)*vd(3)

  end subroutine trilinear_interp_real64
  !--------------------------------------------------------------------




  !-------------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------------
  ! find the indices of the two consecutive values in previously sorted array refx which bound each x(j) value
  recursive subroutine get_interp_indices_real64 ( refx, x, ilo, ihi)
    character(len=*), parameter :: subroutinename='get_interp_indices_real64'
    real(kind=real64), dimension(:), intent(in)      :: refx, x
    integer(kind=int32), dimension(:),  intent(out)   :: ilo, ihi
    integer(kind=int32) :: i,j, dimi, dimj
    dimj = min(size(x),size(ilo),size(ihi))
    dimi = size(refx)

    if (dimj .lt. size(x) .or. &
        dimj .lt. size(ilo) .or. &
        dimj .lt. size(ihi) ) then
      write (*,'(A,A,I0)') subroutinename, ": Warning, dimension mismatch: dimj=", dimj
    endif

    if ( dimi .eq. 1 ) then
      ilo(:) = 1
      ihi(:) = 1
    else if ( dimi .eq. 2 ) then
      ilo(:) = 1
      ihi(:) = 2
    else
      do j=1,dimj
        i = 1
        do while ( x(j) .gt. refx(i))
          i  = i + 1
          if ( i .gt. dimi) then
            exit
          endif
        end do

        if ( i .eq. 1 ) then
          ilo(j) = 1
        else if ( i .ge. dimi ) then
          ilo(j) = dimi - 1
        else
          ilo(j) = i - 1
        endif
        ihi(j) = ilo(j) + 1
      end do
    end if
  end subroutine get_interp_indices_real64

  !-------------------------------------------------------------------------------------------
  recursive subroutine get_interp_indices_real32 ( refx, x, ilo, ihi)
    character(len=*), parameter :: subroutinename='get_interp_indices_real32'
    real(kind=real32), dimension(:), intent(in)      :: refx, x
    integer(kind=int32), dimension(:),  intent(out)   :: ilo, ihi
    integer(kind=int32) :: i,j, dimi, dimj
    dimj = min(size(x),size(ilo),size(ihi))
    dimi = size(refx)

    if (dimj .lt. size(x) .or. &
        dimj .lt. size(ilo) .or. &
        dimj .lt. size(ihi) ) then
      write (*,'(A,A,I0)') subroutinename, ": Warning, dimension mismatch: dimj=", dimj
    endif

    if ( dimi .eq. 1 ) then
      ilo(:) = 1
      ihi(:) = 1
    else if ( dimi .eq. 2 ) then
      ilo(:) = 1
      ihi(:) = 2
    else
      do j=1,dimj
        i = 1
        do while ( x(j) .gt. refx(i))
          i  = i + 1
          if ( i .gt. dimi) then
            exit
          endif
        end do

        if ( i .eq. 1 ) then
          ilo(j) = 1
        else if ( i .ge. dimi ) then
          ilo(j) = dimi - 1
        else
          ilo(j) = i - 1
        endif
        ihi(j) = ilo(j) + 1
      end do
    end if
  end subroutine get_interp_indices_real32
  !-------------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------------

  ! prepare the weights for a linear interpolation
  ! the weights are such that interpolated value = wlo*val(index_lo) + whi*val(index_hi)
  recursive subroutine get_interp_indices_weights_real64 ( refx, x, ilo, ihi, wlo, whi)
    real(kind=real64), dimension(:), intent(in)      :: refx, x
    integer(kind=int32), dimension(:),  intent(out)   :: ilo, ihi
    real(kind=real64), dimension(:), intent(out)      :: wlo, whi
    integer(kind=int32) ::  dimj , j
    dimj = min(size(x),size(ilo),size(ihi),size(wlo),size(whi))

    if (dimj .lt. size(x) .or. &
        dimj .lt. size(ilo) .or. &
        dimj .lt. size(ihi) .or. &
        dimj .lt. size(wlo) .or. &
        dimj .lt. size(whi) ) then
      write (*,'(A,I0)') "get_interp_indices_weights_real64: Warning, dimension mismatch: dimj=", dimj
    endif

    if ( size(refx) .eq. 1 ) then
      wlo(:) = 1.0d0
      whi(:) = 0.0d0
      ilo = 1
      ihi = 1
    else
      call get_interp_indices_real64 (refx, x, ilo, ihi)
      do j=1,dimj
        wlo(j) = 1.0d0 - (x(j) - refx(ilo(j)))/(refx(ihi(j)) - refx(ilo(j)))
        whi(j) = (x(j) - refx(ilo(j)))/(refx(ihi(j)) - refx(ilo(j)))
      end do
!      wlo(:) = 1.0d0 - (x(:) - refx(ilo(:)))/(refx(ihi(:)) - refx(ilo(:)))
!      whi(:) = (x(:) - refx(ilo(:)))/(refx(ihi(:)) - refx(ilo(:)))
    endif

  end subroutine get_interp_indices_weights_real64

  recursive subroutine empty_linear_interp_real64(interp)
    type(linear_interp_real64_type), intent(inout) :: interp
!    if(allocated(interp%x_ref)) DEALLOCATE(interp%x_ref)
!    if(allocated(interp%x_eval)) DEALLOCATE(interp%x_eval)
    if(allocated(interp%y_eval)) DEALLOCATE(interp%y_eval)
    if(allocated(interp%index1)) DEALLOCATE(interp%index1)
    if(allocated(interp%index2)) DEALLOCATE(interp%index2)
    if(allocated(interp%weight1)) DEALLOCATE(interp%weight1)
    if(allocated(interp%weight2)) DEALLOCATE(interp%weight2)
    interp%n_ref  = 0
    interp%n_eval = 0
  end subroutine empty_linear_interp_real64

  recursive subroutine setup_linear_interp_real64(interp, x_ref, x_eval, status)
    character(len=*), parameter ::  subroutinename='setup_linear_interp_real64'
    type(linear_interp_real64_type), intent(inout) :: interp
    real(kind=real64), dimension(:), intent(in) :: x_ref
    real(kind=real64), dimension(:), intent(in) :: x_eval
    integer(kind=int32), intent(out) :: status
    integer(kind=int32)  :: ierr
    status = 0

    interp%n_ref = size(x_ref)
    interp%n_eval = size(x_eval)

    ALLOCATE(interp%index1(interp%n_eval), &
        interp%index2(interp%n_eval),  &
        interp%weight1(interp%n_eval), &
        interp%weight2(interp%n_eval), &
        interp%y_eval(interp%n_eval), &
        STAT = status)
    interp%y_eval(:) = 0.0d0
    interp%weight1(:) = 0.0d0
    interp%weight2(:) = 0.0d0
    interp%index1(:) = 0
    interp%index2(:) = 0

    if ( check_allocate_status(subroutinename, status, ierr) ) return

    call get_interp_indices_weights(x_ref(:), x_eval(:), &
        interp%index1(:),  interp%index2(:), &
        interp%weight1(:), interp%weight2(:))

!    write (*, '(A,I0,A,I0)'), "setup_linear_interp_real64: n_ref=", interp%n_ref, " n_eval=", interp%n_eval

  end subroutine setup_linear_interp_real64

  recursive subroutine evaluate_linear_interp_real64(interp, y_ref, status)
    character(len=*), parameter ::  subroutinename='evaluate_linear_interp_real64'
    type(linear_interp_real64_type), intent(inout) :: interp
    real(kind=real64), dimension(:), intent(in) :: y_ref
!    real(kind=real64), dimension(:), intent(out) :: y_eval
    integer(kind=int32), intent(out) :: status
!    integer(kind=int32) :: i
    status = 0
    if (size(y_ref)         .lt. interp%n_ref .or. &
        size(interp%y_eval) .lt. interp%n_eval ) then
       call wrserr(subroutinename, ERROR_LEVEL_SEVERE, &
           "evaluate_linear_interp_real64: Target arrays are too short" )
      status = 1
      return
    endif
!    write (*, '(A,I0,A,I0)'), "evaluate_linear_interp_real64: n_ref=", interp%n_ref, " size(y_ref)=", size(y_ref)

    interp%y_eval(:) = y_ref(interp%index1(:)) * interp%weight1(:) + &
        y_ref(interp%index2(:)) * interp%weight2(:)

  end subroutine evaluate_linear_interp_real64


  !-----------------------------------------------------------------------
  ! given row-matched arrays array_x(:) and array_y(:)
  ! find the value y corresponding to the position x
  ! assume that array_x is sorted
  ! use simple linear interpolation
  recursive subroutine search_and_linear_interp_real64_real64(array_x, array_y, x, y, status)
    real(kind=real64), dimension(:), intent(in) :: array_x
    real(kind=real64), dimension(:), intent(in) :: array_y
    real(kind=real64), intent(in)               :: x
    real(kind=real64), intent(out)              :: y
    integer(kind=int32), intent(out) :: status
    integer(kind=int32) :: i
    real(kind=real64) :: dx
    status = 0
    y = -1.0d-300

    if ( size(array_x) .ne.  size(array_y) ) then
      status = -1
      return
    endif

    ! find bounds of position x in array_x(:)
    call binary_search (array_x(:), x, i, status)

    ! check status
    if ( status .lt. 0 ) then
      return
    endif

    if ( i .lt. 1 ) then
      y = array_y(1)
    else if ( i .ge. size(array_x)) then
      y = array_y(size(array_x))
    else
      dx =  array_x(i+1) - array_x(i)
      if ( dx .le. 0.0 ) then
        y = array_y(i)
      else
        y = array_y(i) + &
            (array_y(i+1) - array_y(i)) * (x - array_x(i)) / dx
      endif
    endif

  end subroutine search_and_linear_interp_real64_real64

  !-----------------------------------------------------------------------
  ! Same as above but with a second y array
  ! find the values y1,y2 corresponding to the position x
  ! assume that array_x is sorted
  ! use simple linear interpolation
  recursive subroutine search_and_linear_interp_real64_real64_real64(array_x, array_y1,array_y2, x, y1, y2, status)
    real(kind=real64), dimension(:), intent(in) :: array_x
    real(kind=real64), dimension(:), intent(in) :: array_y1
    real(kind=real64), dimension(:), intent(in) :: array_y2
    real(kind=real64), intent(in)               :: x
    real(kind=real64), intent(out)              :: y1
    real(kind=real64), intent(out)              :: y2
    integer(kind=int32), intent(out) :: status
    integer(kind=int32) :: i
    real(kind=real64) :: dx
    status = 0
    y1 = -1.0d-300
    y2 = -1.0d-300

    if ((size(array_x) .ne. size(array_y1)) .or. &
        (size(array_x) .ne. size(array_y2)) ) then
      status = -1
      return
    endif

    ! find bounds of position x in array_x(:)
    call binary_search (array_x(:), x, i, status)

    ! check status
    if ( status .lt. 0 ) then
      return
    endif

    if ( i .lt. 1 ) then
      y1 = array_y1(1)
      y2 = array_y2(1)
    else if ( i .ge. size(array_x)) then
      y1 = array_y1(size(array_x))
      y2 = array_y2(size(array_x))
    else
      dx =  array_x(i+1) - array_x(i)
      if ( dx .le. 0.0 ) then
        y1 = array_y1(i)
        y2 = array_y2(i)
      else
        y1 = array_y1(i) + &
            (array_y1(i+1) - array_y1(i)) * (x - array_x(i)) / dx
        y2 = array_y2(i) + &
            (array_y2(i+1) - array_y2(i)) * (x - array_x(i)) / dx
      endif
    endif

  end subroutine search_and_linear_interp_real64_real64_real64



!-----------------------------------------------------------------------
! do a binary search for "val" on (pre-sorted) "array" (of type real64),
! find the indices such that array(index_lo) <= val < array(index_lo+1)
! Return values of status parameter:
!  status = BINARY_SEARCH_STATUS_OK    - everything went ok
!  status = BINARY_SEARCH_STATUS_ERROR - length of array is too short
!  status = BINARY_SEARCH_STATUS_LTMIN - val is smaller than minimum (first) value in array
!  status = BINARY_SEARCH_STATUS_GTMAX - val is larger  than maximum (last)  value in array
  recursive subroutine binary_search_real64(array, val, index_lo, status)
    real(kind=real64), dimension(:), intent(in) :: array
    real(kind=real64), intent(in)               :: val
    integer(kind=int32), intent(out)            :: index_lo
    integer(kind=int32), intent(out) :: status
    integer(kind=int32) :: imin, imax, imid
    imax = size(array)

    ! check for unbounded cases
    if (imax .le. 1 ) then  ! length of array is too short
      index_lo = imax
      status = BINARY_SEARCH_STATUS_ERROR
      return
    else if ( array(1) .gt. val ) then ! val is smaller than minimum (first) value in array
      index_lo = 1
      status = BINARY_SEARCH_STATUS_LTMIN
      return
    else if ( val .gt. array(imax)) then ! val is larger  than maximum (last)  value in array
      index_lo = imax
      status = BINARY_SEARCH_STATUS_GTMAX
      return
    endif

    ! nominal behaviour below:
    imin = 1
    ! continue searching while [imin,imax] is not empty
    do while ( imax .gt. imin+1 )
      imid = imin + ((imax - imin) / 2)       ! calculate the midpoint for roughly equal partition
      if (array(imid) .le. val ) then
        imin = imid     ! change min index to search upper subarray
      else
        imax = imid     ! change max index to search lower subarray
      endif
    end do
    ! success
    index_lo = imin
    status = BINARY_SEARCH_STATUS_OK
  end subroutine binary_search_real64

!-----------------------------------------------------------------------
! do a binary search for val on array, find the indices such that array(index_lo) <= val < array(index_lo+1)
  recursive subroutine binary_search_real32(array, val, index_lo, status)
    real(kind=real32), dimension(:), intent(in) :: array
    real(kind=real32), intent(in)               :: val
    integer(kind=int32), intent(out)            :: index_lo
    integer(kind=int32), intent(out) :: status
    integer(kind=int32) :: imin, imax, imid

    imax = size(array)

    ! check for unbounded cases
    if (imax .le. 1 ) then  ! length of array is too short
      index_lo = imax
      status = BINARY_SEARCH_STATUS_ERROR
      return
    else if ( array(1) .gt. val ) then ! val is smaller than minimum (first) value in array
      index_lo = 1
      status = BINARY_SEARCH_STATUS_LTMIN
      return
    else if ( val .gt. array(imax)) then ! val is larger  than maximum (last)  value in array
      index_lo = imax
      status = BINARY_SEARCH_STATUS_GTMAX
      return
    endif

    ! nominal behaviour below:
    imin = 1
    ! continue searching while [imin,imax] is not empty
    do while ( imax .gt. imin+1 )
      imid = imin + ((imax - imin) / 2)       ! calculate the midpoint for roughly equal partition
      if (array(imid) .le. val ) then
        imin = imid     ! change min index to search upper subarray
      else
        imax = imid     ! change max index to search lower subarray
      endif
    end do
    ! success
    index_lo = imin
    status = BINARY_SEARCH_STATUS_OK
  end subroutine binary_search_real32

  recursive subroutine binary_search_with_hint_real64(array, val, index_lo, index_hint, status)
    real(kind=real64), dimension(:), intent(in) :: array
    real(kind=real64), intent(in)               :: val
    integer(kind=int32), intent(out)            :: index_lo
    integer(kind=int32), intent(in)             :: index_hint
    integer(kind=int32), intent(out)            :: status
    status = BINARY_SEARCH_STATUS_OK
    if ((index_hint   .ge. 1) .and. &
        (index_hint+1 .le. size(array)) ) then
      if ((array(index_hint)   .le. val) .and. &
          (array(index_hint+1) .gt. val) ) then
        index_lo = index_hint
        return
      endif
    endif

    ! we need to find a new solution, so do a full binary search
    call binary_search_real64(array, val, index_lo, status)

  end subroutine binary_search_with_hint_real64
  !--------------------------------------------------------------------

  !--------------------------------------------------------------------
  ! taken from
  !  http://www.robertmartinayers.org/tools/coordinates.html
  !   From Fundamental Astronomy by Karttunen et al:
  !   The galactic coordinates can be obtained from the equatorial
  !   ones with the transformation equations
  !    sin (Ln - L) cos (B) = cos (delta) sin (alpha - AP)
  !    cos (Ln - L) cos (B) = - cos(delta) sin(deltaP) cos(alpha-AP) + sin(delta) cos (deltaP)
  !    sin(B) = cos(delta)* cos(deltaP)* cos(alpha-AP) + sin(delta)*sin(deltaP)
  !   where the direction of the Galactic north pole is
  !   alphaP = 12h 51.4m, deltaP = 27d 08', and the
  !   galactic longitude of the celestial pole Ln = 123.0 degrees.
  !
  !  now, converting this to the J2000 epoch:
  !  the galactic N pole in J2000 coords is:
  !  skycoor -n 20 -d -j 180 +90. Gal
  !  192.85948098333332723087  27.12825125722222097124 J2000
  !  galactic longitude of the J2000 equatorial pole is Ln ~ 123.0 degrees.
  !  skycoor -n 20 -d -g 0 +90. J2000
  !  122.93191813706228288083  27.12825125793921543504 galactic
  recursive subroutine geom_ra_dec_to_gal_l_b (j2000_ra, j2000_dec, gal_l, gal_b)
    real(kind=real64), parameter :: alphaP       = 192.85948098333332723087d0
    real(kind=real64), parameter :: l_of_eq_pole = 122.93191813706228288083d0
    real(kind=real64), parameter :: deltaP_rad   = 27.12825125722222097124d0 * DEG2RAD
    real(kind=real64), parameter :: sin_deltaP   = sin(deltaP_rad)
    real(kind=real64), parameter :: cos_deltaP   = cos(deltaP_rad)

    real(kind=real64), intent(in)  :: j2000_ra, j2000_dec
    real(kind=real64), intent(out) :: gal_l, gal_b
    real(kind=real64) :: cos_dec, sin_dec, cos_ra_alphaP, b, sin_l1_cosb, cos_l2_cosb, l

    cos_dec       = cosd(j2000_dec)
    sin_dec       = sind(j2000_dec)
    cos_ra_alphaP = cosd(j2000_ra - alphaP)
    sin_l1_cosb   = sind(j2000_ra - alphaP)*cos_dec
    cos_l2_cosb   = sin_dec*cos_deltaP - cos_dec*sin_deltaP*cos_ra_alphaP
    b             = asind(sin_dec*sin_deltaP + cos_dec*cos_deltaP*cos_ra_alphaP)

    ! if we are at the poles, we will return zero longitude
    if ( (b .ge. +90.0d0) .or.  &
         (b .le. -90.0d0) .or.  &
         is_zero(cos_l2_cosb) ) then
      l = 0.0d0
    else
      l = l_of_eq_pole - atan2d(sin_l1_cosb, cos_l2_cosb)
      do while ( l .lt. 0.0d0 )
        l = l + 360.0d0
      end do
    endif
    gal_b = b
    gal_l = l

  end subroutine geom_ra_dec_to_gal_l_b


  !  now, converting this to the J2000 epoch:
  !  the J2000 N pole in Gal coords is:
  !  skycoor -n 20 -d -g 180 +90. J2000
  !  122.93191813706228288083  27.12825125793922254047 galactic
  !  RA of the Galactic N pole is:
  !  skycoor -n 20 -d -j 0 +90. Gal
  !  192.85948098333332723087  27.12825125722222097124 J2000
  recursive subroutine geom_gal_l_b_to_ra_dec (gal_l, gal_b,  j2000_ra, j2000_dec)
    real(kind=real64), parameter :: alphaP         = 122.93191813706228288083d0
    real(kind=real64), parameter :: ra_of_gal_pole = 192.85948098333332723087d0
    real(kind=real64), parameter :: deltaP_rad     = 27.12825125722222097124d0 * DEG2RAD
    real(kind=real64), parameter :: sin_deltaP     = sin(deltaP_rad)
    real(kind=real64), parameter :: cos_deltaP     = cos(deltaP_rad)

    real(kind=real64), intent(in)  :: gal_l, gal_b
    real(kind=real64), intent(out) :: j2000_ra, j2000_dec
    real(kind=real64) :: cos_b, sin_b, cos_l_alphaP, dec, sin_l1_cosb, cos_l2_cosb, ra

    cos_b       = cosd(gal_b)
    sin_b       = sind(gal_b)
    cos_l_alphaP = cosd(gal_l - alphaP)
    sin_l1_cosb  = sind(gal_l - alphaP)*cos_b
    cos_l2_cosb  = sin_b*cos_deltaP - cos_b*sin_deltaP*cos_l_alphaP
    dec          = asind(sin_b*sin_deltaP + cos_b*cos_deltaP*cos_l_alphaP)

    !if we are at the poles, we will return zero longitude
    if ((dec .ge. +90.0d0) .or.  &
        (dec .le. -90.0d0) .or.  &
        is_zero(cos_l2_cosb) ) then
      ra = 0.0d0
    else
      ra = ra_of_gal_pole - atan2d(sin_l1_cosb, cos_l2_cosb)
      do while ( ra .lt.  0.0d0 )
        ra = ra + 360.0d0
      end do
      do while ( ra .gt. 360.0d0 )
        ra = ra - 360.0d0
      end do
    end if
    j2000_dec = dec
    j2000_ra =  ra

  end subroutine geom_gal_l_b_to_ra_dec
  !--------------------------------------------------------------------


  ! now do the same for generic conversions
  ! for each combination of coord systems we need
  ! alphaP1    = longitude (in deg, coordsys_in)  of the North pole of coordsys_out
  ! alphaP2    = longitude (in deg, coordsys_out) of the North pole of coordsys_in
  ! deltaP     = latitude  (in rad, coordsys_in)  of the North pole of coordsys_out
  ! sin_deltaP = sin(latitude (in deg, coordsys_in)  of the North pole of coordsys_out)
  ! cos_deltaP = cos(latitude (in deg, coordsys_in)  of the North pole of coordsys_out)
  ! gawk -v str_coord_list="fk5 fk4 Gal Ecl" -v str_c_list="j b g e" 'BEGIN {ncoord=split(str_coord_list,coord_list," ");split(str_c_list,c_list," "); for(i=1;i<=ncoord;i++){for(j=1;j<=ncoord;j++){str_sys=sprintf("skycoor -n 20 -d -%s 0.0d0 +90.0d0 %s", c_list[i], coord_list[j]); printf("  ! pole of %-4s in system %-4s = ", coord_list[j], coord_list[i]); system(str_sys);}}; exit }'
  ! pole of fk5  in system fk5  =   0.00000000000000000000  90.00000000000000000000 J2000
  ! pole of fk4  in system fk5  = 180.31660480000002166889  89.72174750944444099332 J2000
  ! pole of Gal  in system fk5  = 192.85948098333332723087  27.12825125722222097124 J2000
  ! pole of Ecl  in system fk5  = 270.00000000000000000000  66.56070888888888248403 J2000
  ! pole of fk5  in system fk4  = 359.67566772916666195670  89.72168712805556367584 B1950
  ! pole of fk4  in system fk4  =   0.00000000000000000000  90.00000000000000000000 B1950
  ! pole of Gal  in system fk4  = 192.25000000000000000000  27.39999999999999857891 B1950
  ! pole of Ecl  in system fk4  = 269.99962012916671483254  66.55418658416665778077 B1950
  ! pole of fk5  in system Gal  = 122.93191813706228288083  27.12825125793921543504 galactic
  ! pole of fk4  in system Gal  = 122.99999999999113242666  27.40000000001088054091 galactic
  ! pole of Gal  in system Gal  =   0.00000000000000000000  90.00000000000000000000 galactic
  ! pole of Ecl  in system Gal  =  96.38397063181592727688  29.81144003976618606089 galactic
  ! pole of fk5  in system Ecl  =  90.00000000000000000000  66.56070888888918091197 ecliptic
  ! pole of fk4  in system Ecl  =  89.99976885764567668957  66.55418823093982894079 ecliptic
  ! pole of Gal  in system Ecl  = 180.02321726958501812987  29.81144003928359254019 ecliptic
  ! pole of Ecl  in system Ecl  =   0.00000000000000000000  90.00000000000000000000 ecliptic

  ! gawk -v str_coord_list="fk5 fk4 Gal Ecl" -v str_c_list="j b g e" 'BEGIN {ncoord=split(str_coord_list,coord_list," ");split(str_c_list,c_list," "); for(j=1;j<=ncoord;j++){for(i=1;i<=ncoord;i++){str_sys=sprintf("skycoor -n 20 -d -%s 0.0d0 +90.0d0 %s", c_list[i], coord_list[j]); printf("  ! pole of %-4s in system %-4s = ", coord_list[j], coord_list[i]); system(str_sys);}}; exit }'

  ! pole of fk5  in system fk5  =   0.00000000000000000000  90.00000000000000000000 J2000
  ! pole of fk5  in system fk4  = 359.67566772916666195670  89.72168712805556367584 B1950
  ! pole of fk5  in system Gal  = 122.93191813706228288083  27.12825125793921543504 galactic
  ! pole of fk5  in system Ecl  =  90.00000000000000000000  66.56070888888918091197 ecliptic
  ! pole of fk4  in system fk5  = 180.31660480000002166889  89.72174750944444099332 J2000
  ! pole of fk4  in system fk4  =   0.00000000000000000000  90.00000000000000000000 B1950
  ! pole of fk4  in system Gal  = 122.99999999999113242666  27.40000000001088054091 galactic
  ! pole of fk4  in system Ecl  =  89.99976885764567668957  66.55418823093982894079 ecliptic
  ! pole of Gal  in system fk5  = 192.85948098333332723087  27.12825125722222097124 J2000
  ! pole of Gal  in system fk4  = 192.25000000000000000000  27.39999999999999857891 B1950
  ! pole of Gal  in system Gal  =   0.00000000000000000000  90.00000000000000000000 galactic
  ! pole of Gal  in system Ecl  = 180.02321726958501812987  29.81144003928359254019 ecliptic
  ! pole of Ecl  in system fk5  = 270.00000000000000000000  66.56070888888888248403 J2000
  ! pole of Ecl  in system fk4  = 269.99962012916671483254  66.55418658416665778077 B1950
  ! pole of Ecl  in system Gal  =  96.38397063181592727688  29.81144003976618606089 galactic
  ! pole of Ecl  in system Ecl  =   0.00000000000000000000  90.00000000000000000000 ecliptic


  !--------------------------------------------------------------------
  recursive subroutine geom_convert_coordsys_longlat_longlat (coordsys_in, long_in, lat_in,  coordsys_out, long_out, lat_out )
    real(kind=real64), parameter, dimension(1:COORDSYS_CODE_ECL*COORDSYS_CODE_ECL) :: alphaP_1D = (/ &
          0.00000000000000000000d0, 359.67566772916666195670d0, &
        122.93191813706228288083d0,  90.00000000000000000000d0, &
        180.31660480000002166889d0,   0.00000000000000000000d0, &
        122.99999999999113242666d0,  89.99976885764567668957d0, &
        192.85948098333332723087d0, 192.25000000000000000000d0, &
          0.00000000000000000000d0, 180.02321726958501812987d0, &
        270.00000000000000000000d0, 269.99962012916671483254d0, &
         96.38397063181592727688d0,   0.00000000000000000000d0 /)

    real(kind=real64), parameter, dimension(1:COORDSYS_CODE_ECL*COORDSYS_CODE_ECL) :: deltaP_1D = (/ &
        90.00000000000000000000d0*DEG2RAD, 89.72168712805556367584d0*DEG2RAD, &
        27.12825125793921543504d0*DEG2RAD, 66.56070888888918091197d0*DEG2RAD, &
        89.72174750944444099332d0*DEG2RAD, 90.00000000000000000000d0*DEG2RAD, &
        27.40000000001088054091d0*DEG2RAD, 66.55418823093982894079d0*DEG2RAD, &
        27.12825125722222097124d0*DEG2RAD, 27.39999999999999857891d0*DEG2RAD, &
        90.00000000000000000000d0*DEG2RAD, 29.81144003928359254019d0*DEG2RAD, &
        66.56070888888888248403d0*DEG2RAD, 66.55418658416665778077d0*DEG2RAD, &
        29.81144003976618606089d0*DEG2RAD, 90.00000000000000000000d0*DEG2RAD /)

    real(kind=real64), parameter, dimension(1:COORDSYS_CODE_ECL,1:COORDSYS_CODE_ECL) :: alphaP=reshape(alphaP_1D,shape(alphaP))
    real(kind=real64), parameter, dimension(1:COORDSYS_CODE_ECL,1:COORDSYS_CODE_ECL) :: deltaP=reshape(deltaP_1D,shape(deltaP))
    real(kind=real64), parameter, dimension(1:COORDSYS_CODE_ECL,1:COORDSYS_CODE_ECL) :: sin_deltaP=sin(deltaP)
    real(kind=real64), parameter, dimension(1:COORDSYS_CODE_ECL,1:COORDSYS_CODE_ECL) :: cos_deltaP=cos(deltaP)

    integer(kind=int32), intent(in) :: coordsys_in, coordsys_out
    real(kind=real64), intent(in)  :: long_in, lat_in
    real(kind=real64), intent(out) :: long_out, lat_out

    real(kind=real64) :: cos_lat_in, sin_lat_in, cos_long_in_alphaP1
    real(kind=real64) :: sin_l1_cosb, cos_l2_cosb

    if ( coordsys_in == coordsys_out ) then
      long_out = long_in
      lat_out  = lat_in
      return
    else if ( &
        (coordsys_in  .lt. COORDSYS_CODE_ICRS) .or. &
        (coordsys_in  .gt. COORDSYS_CODE_ECL) .or. &
        (coordsys_out .lt. COORDSYS_CODE_ICRS) .or. &
        (coordsys_out .gt. COORDSYS_CODE_ECL)) then
      long_out = NULLVAL_REAL64
      lat_out  = NULLVAL_REAL64
      return
    endif

    cos_lat_in          = cosd(lat_in)
    sin_lat_in          = sind(lat_in)
    cos_long_in_alphaP1 = cosd(long_in - alphaP(coordsys_in,coordsys_out))
    sin_l1_cosb         = sind(long_in - alphaP(coordsys_in,coordsys_out))*cos_lat_in
    cos_l2_cosb         = sin_lat_in*cos_deltaP(coordsys_in,coordsys_out) - &
        cos_lat_in*sin_deltaP(coordsys_in,coordsys_out)*cos_long_in_alphaP1

    lat_out             = asind(sin_lat_in*sin_deltaP(coordsys_in,coordsys_out) + &
        cos_lat_in*cos_deltaP(coordsys_in,coordsys_out)*cos_long_in_alphaP1)

    ! if we are at the poles, we will return zero longitude
    if ( (lat_out .ge. +90.0d0) .or.  &
         (lat_out .le. -90.0d0) .or.  &
         (abs(cos_l2_cosb) .lt. 1.0d-30) ) then
!         (is_zero(cos_l2_cosb) ) ) then
!      write (*,'(3(x,A,x,E15.7))'), "cos_l2_cosb=", cos_l2_cosb, "tiny=", tiny(cos_l2_cosb), "epsilon=", epsilon(cos_l2_cosb)
      long_out = 0.0d0
    else
      long_out = alphaP(coordsys_out,coordsys_in) - atan2d(sin_l1_cosb, cos_l2_cosb)
      do while ( long_out .lt. 0.0d0 )
        long_out = long_out + 360.0d0
      end do
      do while ( long_out .gt. 360.0d0 )
        long_out = long_out - 360.0d0
      end do
    endif

  end subroutine geom_convert_coordsys_longlat_longlat

  ! wrapper for the above with two coord_type structs
  recursive subroutine geom_convert_coordsys_coord_coord (coords_in, coords_out)
    type(coord_type), intent(in)  :: coords_in
    type(coord_type), intent(inout) :: coords_out
    call geom_convert_coordsys_longlat_longlat (coords_in%coordsys_code, coords_in%long, coords_in%lat, &
        coords_out%coordsys_code, coords_out%long, coords_out%lat )
  end subroutine geom_convert_coordsys_coord_coord

  ! wrapper for the above with one coord_type struct and one set of coord params
  recursive subroutine geom_convert_coordsys_coord_longlat (coords_in,  coordsys_out, long_out, lat_out)
    type(coord_type), intent(in)  :: coords_in
    integer(kind=int32), intent(in) :: coordsys_out
    real(kind=real64), intent(out) :: long_out, lat_out
    call geom_convert_coordsys_longlat_longlat (coords_in%coordsys_code, coords_in%long, coords_in%lat, &
        coordsys_out, long_out, lat_out )
  end subroutine geom_convert_coordsys_coord_longlat

  ! wrapper for the above with one coord_type struct and one set of coord params
  recursive subroutine geom_convert_coordsys_longlat_coord (coordsys_in, long_in, lat_in, coords_out )
    integer(kind=int32), intent(in) :: coordsys_in
    real(kind=real64), intent(in) :: long_in, lat_in
    type(coord_type), intent(inout)  :: coords_out
    call geom_convert_coordsys_longlat_longlat (coordsys_in, long_in, lat_in, &
        coords_out%coordsys_code, coords_out%long, coords_out%lat)
  end subroutine geom_convert_coordsys_longlat_coord


  !-------------------------------------------------------------------------------------
  recursive subroutine srctool_generic_init_coord(coord)
    type(coord_type),  intent(inout)    :: coord
    coord%coordsys_code = COORDSYS_CODE_NULL
    coord%long          = 0.0d0
    coord%lat           = 0.0d0
    coord%x             = 0.0d0
    coord%y             = 0.0d0
    coord%rawx          = 0.0d0
    coord%rawy          = 0.0d0
  end subroutine srctool_generic_init_coord
  !-------------------------------------------------------------------------------------

  ! Convert an integer to string inline
  ! from https://stackoverflow.com/questions/1262695/convert-integers-to-strings-to-create-output-filenames-at-run-time
  ! make sure to wrap the result in trim()
!  character(len=32) function i2s(k)
!    integer, intent(in) :: k
!    write (i2s, '(I0)') k
!    i2s = adjustl(i2s)
!  end function i2s

  character(len=32) pure recursive function i2s_int64(k)
    integer(kind=int64), intent(in) :: k
    write (i2s_int64, '(I0)') k
    i2s_int64 = adjustl(i2s_int64)
  end function i2s_int64

  character(len=32) pure recursive function i2s_int32(k)
    integer(kind=int32), intent(in) :: k
    write (i2s_int32, '(I0)') k
    i2s_int32 = adjustl(i2s_int32)
  end function i2s_int32

  character(len=32) pure recursive function i2s_int16(k)
    integer(kind=int16), intent(in) :: k
    write (i2s_int16, '(I0)') k
    i2s_int16 = adjustl(i2s_int16)
  end function i2s_int16

  character(len=32) pure recursive function i2s_int8(k)
    integer(kind=int8), intent(in) :: k
    write (i2s_int8, '(I0)') k
    i2s_int8 = adjustl(i2s_int8)
  end function i2s_int8


  ! Convert a logical to an integer inline
  integer pure recursive function b2i(b)
    logical, intent(in) :: b
    if ( b ) then
      b2i = 1
    else
      b2i = 0
    endif
  end function b2i

  ! Convert a real number to a string inline
  ! make sure to wrap the result in trim()
  character(len=12) pure recursive function r2s_real64(r)
    real(kind=real64), intent(in) :: r
    write (r2s_real64, '(ES12.6E2)') r
  end function r2s_real64

  character(len=12) pure recursive function r2s_real32(r)
    real(kind=real32), intent(in) :: r
    write (r2s_real32, '(ES12.6E2)') r
  end function r2s_real32

  character(len=12) pure recursive function r2fs_real64(r)
    real(kind=real64), intent(in) :: r
    write (r2fs_real64, '(F12.7)') r
  end function r2fs_real64

  character(len=12) pure recursive function r2fs_real32(r)
    real(kind=real32), intent(in) :: r
    write (r2fs_real32, '(F12.7)') r
  end function r2fs_real32





!
!  ! from here: https://stackoverflow.com/questions/22460220/fortran-character-returning-function-of-unknown-length
!  function int2str(i) result(res)
!    character(:),allocatable :: res
!    integer,intent(in) :: i
!    character(range(i)+2) :: tmp
!    write(tmp,'(i0)') i
!    res = trim(tmp)
!  end function
!



  !-----------------------------------------------
  !-----------------------------------------------
  recursive subroutine find_nonzero_pixel_bounds_real64 (ppix, thresh, x_min, x_max, y_min, y_max)
    real(kind=real64), intent(in) :: ppix(:,:)
    real(kind=real64), intent(in) :: thresh
    integer(kind=int32), intent(out) ::  x_min, x_max, y_min, y_max

    integer(kind=int32) :: i, j, naxis1, naxis2
    x_min = -1
    x_max = -1
    y_min = -1
    y_max = -1

    naxis1 = SIZE(ppix, 1)
    naxis2 = SIZE(ppix, 2)

    do i=1,naxis1
      do j=1,naxis2
        if ( ppix(i,j) .gt. thresh ) then
          x_min = i
          exit
        end if
      end do
      if ( x_min .gt. 0 ) then
        exit
      end if
    end do

    do j=1,naxis2
      do i=x_min,naxis1
        if ( ppix(i,j) .gt. thresh ) then
          y_min = j
          exit
        end if
      end do
      if ( y_min .gt. 0 ) then
        exit
      end if
    end do

    do i=naxis1,x_min,-1
      do j=y_min,naxis2
        if ( ppix(i,j) .gt. thresh ) then
          x_max = i
          exit
        end if
      end do
      if ( x_max .gt. 0 ) then
        exit
      end if
    end do

    do j=naxis2,y_min,-1
      do i=x_min,x_max
        if ( ppix(i,j) .gt. thresh ) then
          y_max = j
          exit
        end if
      end do
      if ( y_max .gt. 0 ) then
        exit
      end if
    end do

  end subroutine find_nonzero_pixel_bounds_real64
  !------------------------------------------------------------------

  ! remove white space at beginning of string, then reduce all consecutive whitespaces to single spaces
  ! this is about as inefficient as it could possibly be
  recursive subroutine squash_string (str)
    character(len=*), intent(inout) :: str
    integer(kind=int32) :: i,j,k,n
    i = 1
    j = 1
!    write(*,*) trim(str)
    str = adjustl(str)  ! remove leading spaces
!    write(*,*) trim(str)
    n = len_trim(str)
    do while  ( j .gt. 0 )
      j = INDEX(trim(str), "  ")
      if ( j .gt. 0 ) then
        k = 1
        do while ( j+k .lt. n )
          if ( str(j+k+1:j+k+1) .ne. " " ) exit
          k = k + 1
        end do
        str(j:) = str(j+k:)
        i = j
      endif
!      write(*,*) trim(str)
    enddo
!    write(*,*) trim(str)
  end subroutine squash_string

  ! expand array, saving contents
  ! this doubles the size by default and enforces a minimum size
  recursive subroutine expand_array_real64(arr)
    real(kind=real64), allocatable, intent(inout) :: arr(:)
    real(kind=real64), allocatable :: temp(:)
    integer(kind=int32) :: num

    num = size(arr)
    allocate(temp(num))
    temp = arr

    deallocate(arr)

    ! double array by default
    num = max(num, 32)
    num = num * 2
    allocate(arr(num))
    arr(:size(temp)) = temp

    deallocate(temp)

  end subroutine expand_array_real64

  !----------------------------------------------------------------------
  ! wrerr/wrlog/wrchatty cannot be called in parallel, as they are not
  ! recursive and write to the same files

  ! The below are "safe" wrapper functions which put the calls in a
  ! critical section, to ensure they are only called in one thread at
  ! a time

  ! Note the critical section is shared as wrerr/wrslog/wrserr share
  ! output files.

  recursive subroutine wrschatty(rtname, text)
    character(len=*), intent(in) :: rtname, text

    !$omp critical (param_logging)
    call wrchatty(rtname, text)
    !$omp end critical (param_logging)
  end subroutine wrschatty

  recursive subroutine wrslog(rtname, text)
    character(len=*), intent(in) :: rtname, text

    !$omp critical (param_logging)
    call wrlog(rtname, text)
    !$omp end critical (param_logging)
  end subroutine wrslog

  recursive subroutine wrserr(rtname, ierr, text)
    character(len=*), intent(in) :: rtname, text
    integer, intent(in) :: ierr

    !$omp critical (param_logging)
    call wrerr(rtname, ierr, text)
    !$omp end critical (param_logging)
  end subroutine wrserr
  !----------------------------------------------------------------------

end module srctool_generic
!--------------------------------------------------------------------
