! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_generic_regions

!-----------------------------------------------------------------------------
! Author: T. Dwelly (dwelly@mpe.mpg.de)
!
! First version: 2013-06-20
! Current version: $Revision$
! CVS Date: $Date$
!-----------------------------------------------------------------------------
!
!-----------------------------------------------------------------------------
! Summary:
! A module containing generic subroutines/functions handling region files/specifications/tests
!
!-----------------------------------------------------------------------------

  use iso_fortran_env ! replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface

  use srctool_generic   ! generic module with some universal constants
  use srctool_extmap    ! extmap module used for masks

  implicit none     ! insist on explicit typing

  !-------------------------------------------------------------------------------
  !some private constants
  character(len=*), private, parameter :: str_modname='SRCTOOL_GENERIC_REGIONS_MOD'

  public
  integer(kind=int32), parameter :: MAX_REGCOMP_PARS=16  ! max number of params used to describe a region component

  !-------------------------------------------------------------------------------
  !----- REGION DESCRIPTION TYPES ------------------------------------------------
  !-------------------------------------------------------------------------------
  type :: regcomp_type
    integer(kind=int32) :: shape_code       ! see list of defined REGCOMP_SHAPE_CODE_* options below
    integer(kind=int32) :: coordsys_code    ! see list of defined COORDSYS_CODE_* options in srctool_generic_mod.f90
    integer(kind=int32) :: num_pars
    integer(kind=int32) :: num_pars_read    ! number of paremeters that were read in
    real(kind=real64)   :: pars(1:MAX_REGCOMP_PARS) ! meaning is shape-specific
    logical           :: wildcards(1:MAX_REGCOMP_PARS) ! set to true if the param should be interpreted as wild card
    integer(kind=int32) :: weight      ! set to -1 if component is to be subtracted
    real(kind=real64)   :: max_radius  ! maximum distance of any point in regcomp from src coords - deg
                                       ! must be set to a negative number if not yet calculated

    character(len=STR_MAX)         :: mask_filename  ! filename in mask
    type(extmap_type), pointer     :: mask ! image used REGCOMP_SHAPE_CODE_MASK
  end type regcomp_type
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! a region consists of a number of regcomp_type components
  type :: region_type
    integer(kind=int32) :: num_regcomps
    type(regcomp_type), allocatable  :: pregcomp(:)
    real(kind=real64)   :: max_radius  ! maximum distance of any point in the region from src coords - deg
                                       ! must be set to a negative number if not yet calculated
  end type region_type
  !-------------------------------------------------------------------------------



  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_NULL       =0
  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_CIRCLE     =1
  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_ELLIPSE    =2
  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_BOX        =3
  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_ANNULUS    =4
  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_ELLANNULUS =5
  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_RECTANGLE  =6
  integer(kind=int32), parameter :: REGCOMP_SHAPE_CODE_MASK       =7
  character(len=*), parameter :: REGCOMP_SHAPE_STR(REGCOMP_SHAPE_CODE_NULL:REGCOMP_SHAPE_CODE_MASK) &
       = (/ "null      ", "circle    ", "ellipse   ", "box       ", "annulus   ", &
            "ellannulus", "rectangle ", "mask      " /)

  integer(kind=int32), parameter :: REGCOMP_PARS_GENERIC_LONG  =1
  integer(kind=int32), parameter :: REGCOMP_PARS_GENERIC_LAT   =2

  integer(kind=int32), parameter :: REGCOMP_PARS_CIRCLE_LONG  =REGCOMP_PARS_GENERIC_LONG
  integer(kind=int32), parameter :: REGCOMP_PARS_CIRCLE_LAT   =REGCOMP_PARS_GENERIC_LAT
  integer(kind=int32), parameter :: REGCOMP_PARS_CIRCLE_RADIUS=3
  integer(kind=int32), parameter :: REGCOMP_NUM_PARS_CIRCLE   =3

  integer(kind=int32), parameter :: REGCOMP_PARS_ANNULUS_LONG   =REGCOMP_PARS_GENERIC_LONG
  integer(kind=int32), parameter :: REGCOMP_PARS_ANNULUS_LAT    =REGCOMP_PARS_GENERIC_LAT
  integer(kind=int32), parameter :: REGCOMP_PARS_ANNULUS_RADIUS1=3
  integer(kind=int32), parameter :: REGCOMP_PARS_ANNULUS_RADIUS2=4
  integer(kind=int32), parameter :: REGCOMP_NUM_PARS_ANNULUS    =4

  integer(kind=int32), parameter :: REGCOMP_PARS_BOX_LONG   =REGCOMP_PARS_GENERIC_LONG
  integer(kind=int32), parameter :: REGCOMP_PARS_BOX_LAT    =REGCOMP_PARS_GENERIC_LAT
  integer(kind=int32), parameter :: REGCOMP_PARS_BOX_WIDTH  =3
  integer(kind=int32), parameter :: REGCOMP_PARS_BOX_HEIGHT =4
  integer(kind=int32), parameter :: REGCOMP_PARS_BOX_ANGLE  =5
  integer(kind=int32), parameter :: REGCOMP_NUM_PARS_BOX    =5

  integer(kind=int32), parameter :: REGCOMP_PARS_ELLIPSE_LONG     =REGCOMP_PARS_GENERIC_LONG
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLIPSE_LAT      =REGCOMP_PARS_GENERIC_LAT
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLIPSE_SEMIMAJOR=3
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLIPSE_SEMIMINOR=4
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLIPSE_ANGLE    =5
  integer(kind=int32), parameter :: REGCOMP_NUM_PARS_ELLIPSE      =5

  integer(kind=int32), parameter :: REGCOMP_PARS_RECTANGLE_LONG1     =1
  integer(kind=int32), parameter :: REGCOMP_PARS_RECTANGLE_LONG2     =2
  integer(kind=int32), parameter :: REGCOMP_PARS_RECTANGLE_LAT1      =3
  integer(kind=int32), parameter :: REGCOMP_PARS_RECTANGLE_LAT2      =4
  integer(kind=int32), parameter :: REGCOMP_NUM_PARS_RECTANGLE       =4

  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_LONG      =REGCOMP_PARS_GENERIC_LONG
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_LAT       =REGCOMP_PARS_GENERIC_LAT
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_SEMIMAJOR1=3
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_SEMIMINOR1=4
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_SEMIMAJOR2=5
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_SEMIMINOR2=6
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_ANGLE1    =7
  integer(kind=int32), parameter :: REGCOMP_PARS_ELLANNULUS_ANGLE2    =8
  integer(kind=int32), parameter :: REGCOMP_NUM_PARS_ELLANNULUS       =8

  integer(kind=int32), parameter :: REGCOMP_NUM_PARS_MASK             =0

  !-------------------------------------------------------------------------------
  ! subroutine declarations and interfaces for module: srctool_generic_regions
  public  :: srctool_regions_init_region
  public  :: srctool_regions_init_regcomp
  public  :: srctool_regions_add_regcomp
  private :: srctool_regions_sanity_check_regcomp
  private :: srctool_regions_interpret_regcomp_pars
  private :: srctool_regions_load_mask
  public  :: srctool_regions_interpret_regcomp_desc
  private :: srctool_regions_regcomp_calc_bounding_circle
  public  :: srctool_regions_region_calc_bounding_circle
  private :: srctool_regions_print_regcomp
  private :: srctool_regions_sprint_regcomp
  public  :: srctool_regions_print_region
  public  :: srctool_regions_sprint_region

  private :: srctool_test_if_coord_in_regcomp_circle
  private :: srctool_test_if_coord_in_regcomp_annulus
  private :: srctool_test_if_coord_in_regcomp_rectangle
  private :: srctool_test_if_coord_in_regcomp_ellipse
  private :: srctool_test_if_coord_in_regcomp_ellannulus
  private :: srctool_test_if_coord_in_regcomp_box
  private :: srctool_test_if_coord_in_regcomp_mask

  private :: srctool_regions_test_if_coord_in_regcomp
  public  :: srctool_regions_test_if_coord_in_region
  public  :: srctool_regions_resolve_region_wildcards
  private :: srctool_regions_resolve_regcomp_wildcards
  public  :: srctool_regions_extract_central_coords
  public  :: srctool_regions_test_harness


!-------------------------------------------------------------------------------
contains



  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------- START of REGION structure handling ---------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  recursive subroutine srctool_regions_init_regcomp(regcomp)
    character(len=*), parameter          :: subroutinename='srctool_regions_init_regcomp'
    type(regcomp_type),  intent(inout)    :: regcomp
    regcomp%shape_code    = REGCOMP_SHAPE_CODE_NULL
    regcomp%coordsys_code = COORDSYS_CODE_NULL
    regcomp%num_pars      = 0
    regcomp%num_pars_read = 0
    regcomp%pars(:)       = 0.0d0
    regcomp%wildcards(:)  = .false.
    regcomp%weight        = 0
    regcomp%max_radius    = -1.0d0
    nullify(regcomp%mask)
    regcomp%mask_filename = ''
  end subroutine srctool_regions_init_regcomp

  !-------------------------------------------------------------------------------------
  recursive subroutine srctool_regions_init_region(region)
    character(len=*), parameter          :: subroutinename='srctool_regions_init_region'
    type(region_type),  intent(inout)    :: region

    region%num_regcomps = 0
    region%max_radius = -1.0d0
  end subroutine srctool_regions_init_region

  !-------------------------------------------------------------------------------------
  recursive subroutine srctool_regions_add_regcomp(region, regcomp)
    type(region_type), intent(inout) :: region
    type(regcomp_type), intent(in) :: regcomp

    integer(kind=int32) :: i
    type(regcomp_type), allocatable :: temp(:)

    if( .not. allocated(region%pregcomp) ) then
       ! allocate some sensible initial size
       allocate(region%pregcomp(16))
       do i=1, size(region%pregcomp)
          call srctool_regions_init_regcomp(region%pregcomp(i))
       end do
    end if

    ! if number of components is too large, double allocated array
    ! this is clumsy, as I don't think there's a mechanism for swaping arrays
    if( region%num_regcomps == size(region%pregcomp) ) then
       ! move to temporary
       call move_alloc(region%pregcomp, temp)
       ! make larger
       allocate(region%pregcomp(region%num_regcomps*2))
       ! copy back
       region%pregcomp(:region%num_regcomps) = temp(:region%num_regcomps)
       ! delete
       deallocate(temp)

       ! initialise new members
       do i=region%num_regcomps+1, size(region%pregcomp)
          call srctool_regions_init_regcomp(region%pregcomp(i))
       end do
    end if

    ! add new value
    region%num_regcomps = region%num_regcomps + 1
    region%pregcomp(region%num_regcomps) = regcomp
  end subroutine srctool_regions_add_regcomp

  !-----------------------------------------------------------------------------------
  ! TODO - add lots more tests
  recursive subroutine srctool_regions_sanity_check_regcomp (str_regcomp, regcomp, ierr)
    character(len=*), parameter :: subroutinename='srctool_regions_sanity_check_regcomp'
    character(len=*), intent(in)   :: str_regcomp
    type(regcomp_type),  intent(in)  :: regcomp
    integer(kind=int32), intent(out)  :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK


    if ( regcomp%num_pars_read .ne. regcomp%num_pars ) then
      ierr = ERROR_LEVEL_SEVERE
      write (str_buffer, '(A,A,A)') &
          "Malformed region description (incorrect number of region params): '",  &
          trim(str_regcomp), "'"
      call wrserr (subroutinename, ierr, str_buffer)
    endif

    select case (regcomp%shape_code)

    case (REGCOMP_SHAPE_CODE_CIRCLE)
      if (regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS) .lt. 0.0d0 ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,A,F12.5,A)') "Illegal CIRLE parameter value(s):", &
            "radius=", regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS), 'deg'
        call wrserr (subroutinename, ierr, str_buffer)
        return
      endif

    case (REGCOMP_SHAPE_CODE_ANNULUS)
      if (regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS1) .gt. regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2)) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,2(A,F12.5,A))') "Illegal ANNULUS parameter value(s):", &
            " inner-radius=", regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS1), 'deg', &
            "> outer-radius=", regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2), 'deg'
        call wrserr (subroutinename, ierr, str_buffer)
        return
      endif

    case (REGCOMP_SHAPE_CODE_RECTANGLE)
      if (regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT1) .gt. regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT2)) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,2(A,F12.5,A))') "Illegal RECTANGLE parameter value(s):", &
            " lat1=", regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT1), 'deg', &
            "> lat2=", regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT2), 'deg'
        call wrserr (subroutinename, ierr, str_buffer)
        return
      endif
      if (regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1) .gt. regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2)) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,2(A,F12.5,A))') "Illegal RECTANGLE parameter value(s):", &
            " long1=", regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1), 'deg', &
            "> long2=", regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2), 'deg'
        call wrserr (subroutinename, ierr, str_buffer)
        return
      endif


    case (REGCOMP_SHAPE_CODE_ELLIPSE)
      if ((regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMAJOR) .lt. 0.0d0) .or. &
          (regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMINOR) .lt. 0.0d0) ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,2(A,F12.5,A))') "Illegal ELLIPSE parameter value(s):", &
            " semi-major=", regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMAJOR), 'deg', &
            " semi-minor=", regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMINOR), 'deg'
        call wrserr (subroutinename, ierr, str_buffer)
        return
      endif

    case (REGCOMP_SHAPE_CODE_ELLANNULUS)
      if ((regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR1) .lt. 0.0d0) .or. &
          (regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR1) .lt. 0.0d0) .or. &
          (regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR2) .lt. 0.0d0) .or. &
          (regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR2) .lt. 0.0d0) ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,4(A,F12.5,A))') "Illegal ELLANNULUS parameter value(s):", &
            " semi-major1=", regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR1), 'deg', &
            " semi-minor1=", regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR1), 'deg', &
            " semi-major2=", regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR2), 'deg', &
            " semi-minor2=", regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR2), 'deg'
        call wrserr (subroutinename, ierr, str_buffer)
        return
      endif

    case (REGCOMP_SHAPE_CODE_BOX)
      if ((regcomp%pars(REGCOMP_PARS_BOX_WIDTH)  .lt. 0.0d0) .or. &
          (regcomp%pars(REGCOMP_PARS_BOX_HEIGHT) .lt. 0.0d0) ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,2(A,F12.5,A))') "Illegal BOX parameter value(s):", &
            " width=", regcomp%pars(REGCOMP_PARS_BOX_WIDTH), 'deg', &
            " height=", regcomp%pars(REGCOMP_PARS_BOX_HEIGHT), 'deg'
        call wrserr (subroutinename, ierr, str_buffer)
        return
      endif

    case(REGCOMP_SHAPE_CODE_MASK)
       ! It's ok if it loaded

    case default
      write (str_buffer, '(A,I0)') "Unknown region component: ", regcomp%shape_code
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return

    end select

  end subroutine srctool_regions_sanity_check_regcomp

  !-----------------------------------------------------------------------------------

  recursive subroutine srctool_regions_interpret_regcomp_pars(str_regcomp, regcomp, src_coord, str, ierr)
    character(len=*), parameter :: subroutinename='srctool_regions_interpret_regcomp_pars'

    character(len=*), intent(in)        :: str_regcomp
    type(regcomp_type), intent(inout)   :: regcomp
    type(coord_type), intent(in)        :: src_coord
    character(len=LSTR_MAX), intent(in) :: str
    integer(kind=int32), intent(out)    :: ierr

    character(len=LSTR_MAX) :: str_buffer
    character(len=STR_MAX)  :: str_fmt
    integer(kind=int32)     :: i, j
    character(len=LSTR_MAX) :: str_copy
    integer(kind=int32)     :: status
    logical               :: success_flag
    logical               :: wildcard
    real(kind=real64)       :: par
    real(kind=real64)       :: units_in_deg

    ! make a copy (probably not needed, but means function does not
    ! modify this input)
    str_copy = str

    !----------------------------------------------------------------------
    ! step 6: now try to read the remaining values into the regcomp parameters
    ! expect to start with no leading space
    !    npars = 0
    regcomp%num_pars_read = 0
    do while ( len_trim(str_copy) >= 1 )
       ! look for first blank space
       i = index(str_copy," ")
       if ( i == 0 ) i = len_trim(str_copy)  ! ie all of str_copy is remaining argument

       ! check to see if this is a wildcard parameter
       wildcard = .false.
       if ( trim(str_copy(1:i-1)) == "*" ) then
          if ( regcomp%num_pars_read == 0 ) then
             par = src_coord%long
             wildcard = .true.
          else if ( regcomp%num_pars_read == 1 ) then
             par = src_coord%lat
             wildcard = .true.
          else
             ierr = ERROR_LEVEL_SEVERE
             write (str_buffer, '(A,A)') "Malformed region description (illegal wildcard): ",  trim(str_regcomp)
             call wrserr (subroutinename, ierr, str_buffer)
             return
          endif
          units_in_deg = 1.0d0

       else  ! interpret as a normal number
          ! check if the coordinate string matches any of the allowed input formats
          units_in_deg = 1.0d0
          success_flag = .false.
          if ( regcomp%num_pars_read == 0 ) then
             call convert_longitude_to_decimal (str_copy(1:i-1), par, success_flag)
          else if ( regcomp%num_pars_read == 1 ) then
             call convert_latitude_to_decimal (str_copy(1:i-1), par, success_flag)
          endif

          if ( .not. success_flag ) then
             status = 1
          else
             status = 0
          endif
          ! if conversion to decimal was unsuccessful,
          ! or if we are not looking for a coordinate value
          if ( status /= 0 .or. regcomp%num_pars_read > 1 ) then

             ! check for a single character trailing units indicator ("'drsm), interpret them as such
             j = i - 1
             select case (trim(str_copy(i-1:i-1)))
             case ( '0', '1':'9', '.' )
                units_in_deg = 1.0d0
                j = i   ! no need to decrement as no units are supplied
             case ( '"', 's', 'S' )
                units_in_deg = ARCSEC2DEG
             case ( "'", 'm', 'M' )
                units_in_deg = ARCMIN2DEG
             case ( 'd', 'D' )
                units_in_deg = 1.0d0
             case ( "r", 'R' )
                units_in_deg = RAD2DEG
             case default
                ierr = ERROR_LEVEL_SEVERE
                call wrserr (subroutinename, ierr, "Malformed size/angle units given in: '" &
                     // trim(str_regcomp) // "'")
                call wrserr (subroutinename, ierr, 'Valid size/angle units are: " or s - arcseconds')
                call wrserr (subroutinename, ierr, "                          : ' or m - arcminutes")
                call wrserr (subroutinename, ierr, "                          : d      - degrees")
                call wrserr (subroutinename, ierr, "                          : r      - radians")
                return
             end select

             ! just attempt to read the string as a real
             status = 0
             write (str_fmt, '(A,I0,A)') '(F', j-1, '.0)'
             read (str_copy(1:j-1), trim(str_fmt), iostat=status) par
          end if

          if ( status /= 0 ) then
             ierr = ERROR_LEVEL_SEVERE
             write (str_buffer, '(A,A,A,I0)') "Malformed region description: '",  &
                  trim(str_regcomp), "' iostat=", status
             call wrserr (subroutinename, ierr, str_buffer)
          endif

       end if

       ! now blank out any stuff before the first ' ' and then get rid of any leading space
       str_copy(1:i) = " "
       str_copy = trim(adjustl(str_copy))

!      write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_copy)='", trim(str_copy), "' len=",  &
!         len_trim(str_copy)
!      step = step + 1

       ! if success - increment the counter
       regcomp%num_pars_read = regcomp%num_pars_read + 1
       regcomp%pars(regcomp%num_pars_read) = par * units_in_deg
       regcomp%wildcards(regcomp%num_pars_read) = wildcard

!      write (*,'(A,I0,A,A,A,A,F12.7,A,F12.7,A,A,A,ES12.5,A,A)'), &
!          "Found parameter: index=", npars, &
!          " str='",  trim(str_copy(1:i-1)), "'", &
!          " val= ",  par, &
!          " val_deg= ",  regcomp%pars(npars), &
!          ' fmt=', trim(str_fmt), &
!          ' units_in_deg=', units_in_deg, &
!          ' wildcard=', logical2string(wildcard)

    end do

  end subroutine srctool_regions_interpret_regcomp_pars


  !-----------------------------------------------------------------------------------

  recursive subroutine srctool_regions_load_mask(regcomp, src_coord, filename, ierr)
    character(len=*), parameter :: subroutinename='srctool_regions_load_mask'

    type(regcomp_type), intent(inout) :: regcomp
    type(coord_type), intent(in)      :: src_coord
    character(len=*), intent(in)      :: filename
    integer(kind=int32), intent(out)  :: ierr

    regcomp%num_pars_read = 0
    regcomp%mask_filename = trim(filename)

    ! FIXME: we don't clean up this mask later, so we rely on it being
    ! freed on program exit. srctool makes copies of this, which
    ! causes a crash unless mask is a pointer.
    allocate(regcomp%mask)
    call srctool_extmap_empty(regcomp%mask)

    call srctool_extmap_read_from_file(filename, regcomp%mask, .false., ierr)
    if( ierr /= ERROR_LEVEL_OK ) then
       call wrserr (subroutinename, ierr, "Error: loading region mask")
       return
    end if

    call srctool_extmap_update_source_coord(regcomp%mask, src_coord)
  end subroutine srctool_regions_load_mask

  !-----------------------------------------------------------------------------------
  recursive subroutine srctool_regions_interpret_regcomp_desc (str_regcomp, regcomp, src_coord, ierr)

    character(len=*), parameter :: subroutinename='srctool_regions_interpret_regcomp_desc'
    character(len=*), intent(in)   :: str_regcomp
    type(regcomp_type),  intent(inout)  :: regcomp
    type(coord_type),     intent(in)  :: src_coord
    integer(kind=int32), intent(out)  :: ierr
    integer(kind=int32)  :: i, j !, npars
    integer(kind=int32)  :: status
    character(len=LSTR_MAX) :: str_copy
    character(len=LSTR_MAX) :: str_buffer
!    integer(kind=int32)  :: step

    ierr = ERROR_LEVEL_OK
    status = 0
!    step = 0

    call srctool_regions_init_regcomp(regcomp)

!    write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_regcomp)='", trim(str_regcomp), "' len=",  len_trim(str_regcomp)
!    step = step + 1
    ! make some space for a copy
!    allocate(character(len=len_trim(adjustl(str_regcomp))) :: str_copy, STAT = status)
    ! and make the copy
    str_copy = trim(adjustl(str_regcomp))

!    write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_copy)='", trim(str_copy), "' len=",  len_trim(str_copy)
!    step = step + 1

    ! expect to deal with a single region component suplied as a string

    !----------------------------------------------------------------------
    ! step 1 : strip off any comments after a "#" symbol
    i = index(str_copy,"#")
    if ( i > 0 ) then
      str_copy(i:) = " "  ! convert to white space
    end if
    !----------------------------------------------------------------------


!    write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_copy)='", trim(str_copy), "' len=",  len_trim(str_copy)
!    step = step + 1


    !----------------------------------------------------------------------
    ! step 2 : look for coordinate system string
    ! look for first ';' character
    i = index(str_copy,";")
    if ( i > 1 ) then
      ! select on the part before the ';'
      regcomp%coordsys_code = interpret_coordsys ( trim(str_copy(1:i-1)))
      if ( regcomp%coordsys_code .eq. COORDSYS_CODE_ERROR ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Error: interpreting coordinate system")
        return
      endif
    else
      ! default to icrs; format
      regcomp%coordsys_code = COORDSYS_CODE_ICRS
    end if
    ! blank out any stuff before the ';' and then get rid of any leading space
    str_copy(1:i) = " "
    str_copy = trim(adjustl(str_copy))
    !----------------------------------------------------------------------

!    write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_copy)='", trim(str_copy), "' len=",  len_trim(str_copy)
!    step = step + 1

    !----------------------------------------------------------------------
    ! step 3a : look for a minus/plus sign
    select case ( trim(str_copy(1:1)))
    case ( "-" )
      regcomp%weight=-1
      str_copy(1:1) = " "
    case ( "+" )
      regcomp%weight=1
      str_copy(1:1) = " "
    case default
      regcomp%weight=1
      j = 1
    end select
    str_copy = trim(adjustl(str_copy))
!   write (*,'(A,I0)'), "Found weight=", regcomp%weight
    !----------------------------------------------------------------------


!   write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_copy)='", trim(str_copy), "' len=",  len_trim(str_copy)
!    step = step + 1

    !----------------------------------------------------------------------
    ! step 4 : look for region type string
    !substitute a " " for any punctuation '(),' characters
    do i=1,len_trim(str_copy)
      if (str_copy(i:i) == '(' .or. &
          str_copy(i:i) == ')' .or. &
          str_copy(i:i) == ',') str_copy(i:i) = ' '
    end do
    !----------------------------------------------------------------------

!    write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_copy)='", trim(str_copy), "' len=",  len_trim(str_copy)
!    step = step + 1


    !----------------------------------------------------------------------
    ! step 5: look for first blank space
    i = index(str_copy," ")
    if ( i > 1 ) then
      ! select on the part before the ' '
      select case ( to_lower_fun(trim(str_copy(1:i-1))) )
        case ( "circle" )
          regcomp%shape_code = REGCOMP_SHAPE_CODE_CIRCLE
          regcomp%num_pars   = REGCOMP_NUM_PARS_CIRCLE
        case ( "ellipse" )
          regcomp%shape_code = REGCOMP_SHAPE_CODE_ELLIPSE
          regcomp%num_pars   = REGCOMP_NUM_PARS_ELLIPSE
        case ( "box" )
          regcomp%shape_code = REGCOMP_SHAPE_CODE_BOX
          regcomp%num_pars   = REGCOMP_NUM_PARS_BOX
        case ( "annulus" )
          regcomp%shape_code = REGCOMP_SHAPE_CODE_ANNULUS
          regcomp%num_pars   = REGCOMP_NUM_PARS_ANNULUS
        case ( "elliptannulus", "ellannulus" )
          regcomp%shape_code = REGCOMP_SHAPE_CODE_ELLANNULUS
          regcomp%num_pars   = REGCOMP_NUM_PARS_ELLANNULUS
        case ( "rectangle" )
          regcomp%shape_code = REGCOMP_SHAPE_CODE_RECTANGLE
          regcomp%num_pars   = REGCOMP_NUM_PARS_RECTANGLE
        case ( "mask" )
          regcomp%shape_code = REGCOMP_SHAPE_CODE_MASK
          regcomp%num_pars   = REGCOMP_NUM_PARS_MASK
        case default
          ierr = ERROR_LEVEL_SEVERE
          write (str_buffer, '(A,A)') "Unknown region type: ",  trim(str_copy(1:i-1))
          call wrserr (subroutinename, ierr, str_buffer)
          return
        end select
!        write (*,'(A,A)'), "Found shape_code=", trim(REGCOMP_SHAPE_STR(regcomp%shape_code))

      else
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,A)') "Malformed region description: ",  trim(str_regcomp)
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if
    ! blank out any stuff before the first ' ' and then get rid of any leading space
    str_copy(1:i) = " "
    str_copy = trim(adjustl(str_copy))
    !----------------------------------------------------------------------

!    write (*,'(A,I0,A,A,A,I0)'), "step ",step," : trim(str_copy)='", trim(str_copy), "' len=",  len_trim(str_copy)
!    step = step + 1


    !----------------------------------------------------------------------
    ! step 6: now try to read the remaining values into the regcomp parameters
    ! expect to start with no leading space
!    npars = 0

    if ( regcomp%shape_code /= REGCOMP_SHAPE_CODE_MASK ) then
       ! take numerical paramerters after region type
       call srctool_regions_interpret_regcomp_pars(str_regcomp, regcomp, src_coord, str_copy, ierr)
    else
       ! load mask image from fits file
       call srctool_regions_load_mask(regcomp, src_coord, str_copy, ierr)
    end if
    if( ierr /= ERROR_LEVEL_OK ) return

    !----------------------------------------------------------------------

!    write (*,'(A,I0,A,I0)'), "Found ", npars, " params, expected ", regcomp%num_pars


!    if ( npars .ne. regcomp%num_pars ) then
!      ierr = ERROR_LEVEL_SEVERE
!      write (str_buffer, '(A,A,A,I0)'), &
!          "Malformed region description (incorrect number of region params): '",  &
!          trim(str_regcomp), "' status=", status
!      call wrserr (subroutinename, ierr, str_buffer)
!    end if
! !   if (allocated(str_copy) ) deallocate(str_copy)

    ! now do some further sanity checks on the regcomp to make sure that the supplied values make sense.

    !-----------------------------------------------------------------------------
    ! special case - ellipse with exactly 7 params is shorthand for elliptannulus
    ! allows user to interpret/define using ds9
    if ((regcomp%shape_code .eq. REGCOMP_SHAPE_CODE_ELLIPSE) .and. &
        (regcomp%num_pars_read .eq. 7 )) then
      regcomp%shape_code = REGCOMP_SHAPE_CODE_ELLANNULUS
      regcomp%num_pars = REGCOMP_NUM_PARS_ELLANNULUS
      regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2) = regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1)
    endif
    !-----------------------------------------------------------------------------

    !-----------------------------------------------------------------------------
    ! convert any circle and annulus regions that are in fk4,Gal,Ecl coordsystems into the icrs system
    ! this will almost always speed up the processing later
    ! other regcomp types are non-trivial to convert because of rotation angles etc
    if ((regcomp%coordsys_code .eq. COORDSYS_CODE_FK4) .or. &
        (regcomp%coordsys_code .eq. COORDSYS_CODE_GAL) .or. &
        (regcomp%coordsys_code .eq. COORDSYS_CODE_ECL) ) then
      if ((regcomp%shape_code .eq. REGCOMP_SHAPE_CODE_CIRCLE) .or. &
          (regcomp%shape_code .eq. REGCOMP_SHAPE_CODE_ANNULUS)) then
        call geom_convert_coordsys (regcomp%coordsys_code, &
            regcomp%pars(REGCOMP_PARS_GENERIC_LONG), regcomp%pars(REGCOMP_PARS_GENERIC_LAT),  &
            COORDSYS_CODE_ICRS, &
            regcomp%pars(REGCOMP_PARS_GENERIC_LONG), regcomp%pars(REGCOMP_PARS_GENERIC_LAT))
        regcomp%coordsys_code = COORDSYS_CODE_ICRS

      end if
    end if
    !-----------------------------------------------------------------------------

    !-----------------------------------------------------------------------------
    call srctool_regions_sanity_check_regcomp (str_regcomp, regcomp, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Malformed region component: '" // trim(str_regcomp) // "'")
      return
    endif
    !-----------------------------------------------------------------------------

  end subroutine srctool_regions_interpret_regcomp_desc
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  ! Need to calculate the radius of a circle centred on the src position
  ! which completely contains the regcomp
  recursive subroutine srctool_regions_regcomp_calc_bounding_circle (regcomp, src_coord, ierr )
    character(len=*), parameter          :: subroutinename='srctool_regions_regcomp_calc_bounding_circle'
    type(regcomp_type),  intent(inout)   :: regcomp
    type(coord_type), intent(in)         :: src_coord
    integer(kind=int32), intent(out)     :: ierr
    character(len=STR_MAX) :: str_buffer
    real(kind=real64) :: radius(4)
    type(coord_type)  :: local_coord
    ierr = ERROR_LEVEL_OK

    ! check if we need to convert the coordinates to the region's native system
    if ( regcomp%coordsys_code .ne. src_coord%coordsys_code ) then
      local_coord%coordsys_code = regcomp%coordsys_code
      call geom_convert_coordsys (src_coord%coordsys_code, src_coord%long, src_coord%lat,  &
          local_coord%coordsys_code, local_coord%long, local_coord%lat)
    else
      local_coord = src_coord
    endif

    !result depends on region type

    select case (regcomp%shape_code)
    case (REGCOMP_SHAPE_CODE_CIRCLE)
      regcomp%max_radius = regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS) + &
          calc_ang_sep(regcomp%pars(REGCOMP_PARS_CIRCLE_LONG),regcomp%pars(REGCOMP_PARS_CIRCLE_LAT),&
          local_coord%long,local_coord%lat)
    case (REGCOMP_SHAPE_CODE_ANNULUS)
      regcomp%max_radius = regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2) + &
          calc_ang_sep(regcomp%pars(REGCOMP_PARS_ANNULUS_LONG),regcomp%pars(REGCOMP_PARS_ANNULUS_LAT),&
          local_coord%long,local_coord%lat)

    case (REGCOMP_SHAPE_CODE_RECTANGLE) ! find the most distant corner
      radius(1) = calc_ang_sep(regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1),regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT1),&
          local_coord%long,local_coord%lat)
      radius(2) = calc_ang_sep(regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1),regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT2),&
          local_coord%long,local_coord%lat)
      radius(3) = calc_ang_sep(regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2),regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT1),&
          local_coord%long,local_coord%lat)
      radius(4) = calc_ang_sep(regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2),regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT2),&
          local_coord%long,local_coord%lat)
!      regcomp%max_radius = MAX(radius(1),radius(2),radius(3),radius(4))
      regcomp%max_radius = MAXVAL(radius(:))

    case (REGCOMP_SHAPE_CODE_ELLIPSE)
      regcomp%max_radius = MAX(regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMAJOR),regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMINOR)) + &
          calc_ang_sep(regcomp%pars(REGCOMP_PARS_ELLIPSE_LONG),regcomp%pars(REGCOMP_PARS_ELLIPSE_LAT),&
          local_coord%long,local_coord%lat)

    case (REGCOMP_SHAPE_CODE_ELLANNULUS)
      regcomp%max_radius = MAX(regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR1),&
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR1), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR2), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR2)) + &
          calc_ang_sep(regcomp%pars(REGCOMP_PARS_ELLANNULUS_LONG),regcomp%pars(REGCOMP_PARS_ELLANNULUS_LAT),&
          local_coord%long,local_coord%lat)


    case (REGCOMP_SHAPE_CODE_BOX)
      regcomp%max_radius = &
          SQRT((0.5d0 * regcomp%pars(REGCOMP_PARS_BOX_WIDTH))**2 + &
          (0.5d0 * regcomp%pars(REGCOMP_PARS_BOX_HEIGHT))**2) + &
          calc_ang_sep(regcomp%pars(REGCOMP_PARS_BOX_LONG),regcomp%pars(REGCOMP_PARS_BOX_LAT),&
          local_coord%long,local_coord%lat)

    case (REGCOMP_SHAPE_CODE_MASK)
      regcomp%max_radius = regcomp%mask%max_radius_deg

    case default
      write (str_buffer, '(A,I0)') "Unknown region component: ", regcomp%shape_code
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return

    end select

  end subroutine srctool_regions_regcomp_calc_bounding_circle

  !-----------------------------------------------------------------------------------
  ! Need to calculate the radius of a circle centred on the src position
  ! which completely contains the source region
  recursive subroutine srctool_regions_region_calc_bounding_circle (region, src_coord, ierr )
    character(len=*), parameter          :: subroutinename='srctool_regions_region_calc_bounding_circle'
    type(region_type),  intent(inout)    :: region
    type(coord_type), intent(in)         :: src_coord
    integer(kind=int32), intent(out)     :: ierr
    integer(kind=int32)                  :: i
    ierr = ERROR_LEVEL_OK
    region%max_radius = 0.0d0
    ierr = ERROR_LEVEL_OK

    do i=1,region%num_regcomps
      call srctool_regions_regcomp_calc_bounding_circle (region%pregcomp(i), src_coord, ierr )
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr(subroutinename,ierr,"Problem calculating bounding circles")
        return
      endif
      if ( region%pregcomp(i)%weight .gt. 0 ) then ! only consider the additive regions here
        region%max_radius = max(region%max_radius,region%pregcomp(i)%max_radius)
      endif
    end do

  end subroutine srctool_regions_region_calc_bounding_circle
  !-----------------------------------------------------------------------------------





  !-----------------------------------------------------------------------------------
  recursive subroutine srctool_regions_print_regcomp (regcomp, ierr)
    character(len=*), parameter          :: subroutinename='srctool_regions_print_regcomp'
    type(regcomp_type),  intent(in)   :: regcomp
    integer(kind=int32), intent(out)     :: ierr
    character(len=LSTR_MAX) :: str_buffer
    str_buffer(:) = " "
    call srctool_regions_sprint_regcomp (regcomp, str_buffer, "DETAILS", ierr)
    write (*,'(A)') trim(str_buffer)
  end subroutine srctool_regions_print_regcomp

  !-----------------------------------------------------------------------------------
  recursive subroutine srctool_regions_sprint_regcomp (regcomp, str_buffer, mode, ierr)
    character(len=*), parameter       :: subroutinename='srctool_regions_sprint_regcomp'
    type(regcomp_type),  intent(in)   :: regcomp
    character(len=*), intent(out)     :: str_buffer
    character(len=*), intent(in)      :: mode
    integer(kind=int32), intent(out)  :: ierr
    character(len=STR_MAX) :: str_buffer1
    character(len=STR_MAX) :: str_buffer2
    character(len=1) :: weight_char
    integer(kind=int32) :: i,max_len

    str_buffer(:) = " "
    str_buffer1(:) = " "
    str_buffer2(:) = " "
    max_len = len(str_buffer)

    ierr = ERROR_LEVEL_OK
    if ( regcomp%weight .gt. 0 ) then
      weight_char = ' '
    else
      weight_char = '-'
    endif

!    write(str_buffer1, '(A,A,1x,A,A)'), &
!    write(str_buffer1, '(A,A,A,A)') &
!        trim(COORDSYS_STR(regcomp%coordsys_code)), ';', trim(weight_char), trim(REGCOMP_SHAPE_STR(regcomp%shape_code))
    if ( max_len .ge. (len_trim(COORDSYS_STR(regcomp%coordsys_code)) + &
        len_trim(';') + len_trim(weight_char) + len_trim(REGCOMP_SHAPE_STR(regcomp%shape_code))) ) then
      str_buffer = trim(COORDSYS_STR(regcomp%coordsys_code)) // ';' &
          // trim(weight_char) // trim(REGCOMP_SHAPE_STR(regcomp%shape_code))
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Supplied str_buffer was too short to hold region prefix string")
      call wrserr (subroutinename, ierr, "str_buffer='" // trim(str_buffer) // "' string_to_write='" // &
          trim(COORDSYS_STR(regcomp%coordsys_code)) // ';' &
          // trim(weight_char) // trim(REGCOMP_SHAPE_STR(regcomp%shape_code)) // "'")
      return
    endif


    select case (regcomp%shape_code)
    case (REGCOMP_SHAPE_CODE_CIRCLE)
      write(str_buffer1, '(2(x,F13.9),x,ES11.5E2)') &
          regcomp%pars(REGCOMP_PARS_CIRCLE_LONG),regcomp%pars(REGCOMP_PARS_CIRCLE_LAT), &
          regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS)

    case (REGCOMP_SHAPE_CODE_ANNULUS)
      write(str_buffer1, '(2(x,F13.9),2(x,ES11.5E2))') &
          regcomp%pars(REGCOMP_PARS_ANNULUS_LONG),regcomp%pars(REGCOMP_PARS_ANNULUS_LAT), &
          regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS1), &
          regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2)

    case (REGCOMP_SHAPE_CODE_ELLIPSE)
      write(str_buffer1, '(2(x,F13.9),3(x,ES11.5E2))') &
          regcomp%pars(REGCOMP_PARS_ELLIPSE_LONG),regcomp%pars(REGCOMP_PARS_ELLIPSE_LAT), &
          regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMAJOR), &
          regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMINOR), &
          regcomp%pars(REGCOMP_PARS_ELLIPSE_ANGLE)

    case (REGCOMP_SHAPE_CODE_BOX)
      write(str_buffer1, '(2(x,F13.9),3(x,ES11.5E2))') &
          regcomp%pars(REGCOMP_PARS_BOX_LONG),regcomp%pars(REGCOMP_PARS_BOX_LAT), &
          regcomp%pars(REGCOMP_PARS_BOX_WIDTH), &
          regcomp%pars(REGCOMP_PARS_BOX_HEIGHT),&
          regcomp%pars(REGCOMP_PARS_BOX_ANGLE)

    case (REGCOMP_SHAPE_CODE_ELLANNULUS)
      write(str_buffer1, '(2(x,F13.9),6(x,ES11.5E2))') &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_LONG),regcomp%pars(REGCOMP_PARS_ELLANNULUS_LAT), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR1), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR1), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR2), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR2), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1), &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2)

    case (REGCOMP_SHAPE_CODE_RECTANGLE)
      write(str_buffer1, '(4(x,F13.9),x)') &
          regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1), &
          regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2), &
          regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT1),  &
          regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT2)

    case (REGCOMP_SHAPE_CODE_MASK)
       write(str_buffer1, '(A,A,A)') '(', trim(regcomp%mask_filename), ')'

    case (REGCOMP_SHAPE_CODE_NULL)
      write(str_buffer1, '(x,A,x)') &
          "NULL component"

    case default
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Unknown region component code: " // trim(i2s(regcomp%shape_code)))
      return
    end select


    i = len_trim(str_buffer) + 1
    if ( max_len - i .ge. (len_trim(str_buffer1))) then
      str_buffer(i:) = trim(str_buffer1)
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Supplied str_buffer was too short to hold main region string")
      return
    endif

    i = len_trim(str_buffer) + 1
    if ( mode .eq. "DETAILS" ) then
      write(str_buffer2, '(A,I0,A,F12.8,A)') &
          " # text={weight=", regcomp%weight, " max_radius=", regcomp%max_radius, "deg}"

      if ( max_len - i .ge. (len_trim(str_buffer2))) then
        str_buffer(i:) = trim(str_buffer2)
      else
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Supplied str_buffer was too short to hold region details string")
        return
      endif
    endif


!    select case (regcomp%shape_code)
!    case (REGCOMP_SHAPE_CODE_CIRCLE)
!      write(str_buffer, '(A,2(x,F13.9),x,ES11.5E2,A)'), &
!          trim(str_buffer1), &
!          regcomp%pars(REGCOMP_PARS_CIRCLE_LONG),regcomp%pars(REGCOMP_PARS_CIRCLE_LAT), &
!          regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS), &
!          trim(str_buffer2)
!    case (REGCOMP_SHAPE_CODE_ANNULUS)
!      write(str_buffer, '(A,2(x,F13.9),2(x,ES11.5E2),A)'), &
!          trim(str_buffer1), &
!          regcomp%pars(REGCOMP_PARS_ANNULUS_LONG),regcomp%pars(REGCOMP_PARS_ANNULUS_LAT), &
!          regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS1), &
!          regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2), &
!          trim(str_buffer2)
!
!    case (REGCOMP_SHAPE_CODE_ELLIPSE)
!      write(str_buffer, '(A,2(x,F13.9),3(x,ES11.5E2),A)'), &
!          trim(str_buffer1), &
!          regcomp%pars(REGCOMP_PARS_ELLIPSE_LONG),regcomp%pars(REGCOMP_PARS_ELLIPSE_LAT), &
!          regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMAJOR), &
!          regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMINOR), &
!          regcomp%pars(REGCOMP_PARS_ELLIPSE_ANGLE), &
!          trim(str_buffer2)
!    case (REGCOMP_SHAPE_CODE_BOX)
!      write(str_buffer, '(A,2(x,F13.9),3(x,ES11.5E2),A)'), &
!          trim(str_buffer1), &
!          regcomp%pars(REGCOMP_PARS_BOX_LONG),regcomp%pars(REGCOMP_PARS_BOX_LAT), &
!          regcomp%pars(REGCOMP_PARS_BOX_WIDTH), &
!          regcomp%pars(REGCOMP_PARS_BOX_HEIGHT),&
!          regcomp%pars(REGCOMP_PARS_BOX_ANGLE), &
!          trim(str_buffer2)
!
!    case (REGCOMP_SHAPE_CODE_ELLANNULUS)
!      write(str_buffer, '(A,2(x,F13.9),6(x,ES11.5E2),A)'), &
!          trim(str_buffer1), &
!          regcomp%pars(REGCOMP_PARS_ELLANNULUS_LONG),regcomp%pars(REGCOMP_PARS_ELLANNULUS_LAT), &
!          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR1), &
!          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR1), &
!          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR2), &
!          regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR2), &
!          regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1), &
!          regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2), &
!          trim(str_buffer2)
!    case (REGCOMP_SHAPE_CODE_RECTANGLE)
!      write(str_buffer, '(A,4(x,F13.9),x,A)'), &
!          trim(str_buffer1), &
!          regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1), &
!          regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2), &
!          regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT1),  &
!          regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT2),  &
!          trim(str_buffer2)
!    case (REGCOMP_SHAPE_CODE_NULL)
!      write(str_buffer, '(A,x,A,x,A)') &
!          trim(str_buffer1), "NULL component", &
!          trim(str_buffer2)
!
!    case default
!      ierr = ERROR_LEVEL_SEVERE
!      call wrserr (subroutinename, ierr, "Unknown region component code: " // trim(i2s(regcomp%shape_code)))
!      return
!    end select



  end subroutine srctool_regions_sprint_regcomp


  !-----------------------------------------------------------------------------------
  recursive subroutine srctool_regions_print_region (region, ierr)
    character(len=*), parameter          :: subroutinename='srctool_regions_print_region'
    type(region_type),  intent(in)   :: region
    integer(kind=int32), intent(out)     :: ierr
!    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i
    ierr = ERROR_LEVEL_OK

    write (*, '(A,I0,A,F15.7,A)') &
        "#region made of ", region%num_regcomps, &
        " components, max_radius=", region%max_radius, " deg"

    do i=1,region%num_regcomps
      call srctool_regions_print_regcomp (region%pregcomp(i), ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr(subroutinename,ierr,"Problem printing region")
      end if
    end do

  end subroutine srctool_regions_print_region

  !-----------------------------------------------------------------------------------
  recursive subroutine srctool_regions_sprint_region (region, str_buffer, mode, ierr)
    character(len=*), parameter                :: subroutinename='srctool_regions_print_region'
    type(region_type),  intent(in)             :: region
    character(len=:), allocatable, intent(out) :: str_buffer
    character(len=*), intent(in)               :: mode
    integer(kind=int32), intent(out)           :: ierr

    integer(kind=int32) :: i, l, lin
    character(len=STR_MAX) :: header
    character(len=STR_MAX) :: items(region%num_regcomps)
    integer(kind=int32) :: sum_length

    ierr = ERROR_LEVEL_OK

    if( mode == "DETAILS") then
       write (header, '(A,I0,A,F15.7,A)') &
            "#region made of ", region%num_regcomps, &
            " components, max_radius=", region%max_radius, " deg"
    else
       header = ""
    end if
    sum_length = len_trim(header)

    ! collect all the component regions
    do i=1, region%num_regcomps
       call srctool_regions_sprint_regcomp(region%pregcomp(i), items(i), mode, ierr)
       if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr(subroutinename, ierr, "Region is too long to fit in buffer")
          return
       end if
       call squash_string(items(i))
       sum_length = sum_length + len_trim(items(i))
    end do

    ! add on spacers "; "
    sum_length = sum_length + region%num_regcomps*2

    ! setup output buffer
    allocate(character(sum_length) :: str_buffer)

    ! put header into buffer
    lin = len_trim(header)
    if(lin > 0) then
       str_buffer(:) = header(1:lin)
    end if
    l = 1 + lin

    ! then put region components in
    do i=1, region%num_regcomps
       ! add component
       lin = len_trim(items(i))
       if( lin > 0 ) then
          str_buffer(l:l+lin-1) = items(i)(1:lin)
       end if
       l = l + lin
       ! add spacer
       str_buffer(l:l+1) = "; "
       l = l + 2
    end do

  end subroutine srctool_regions_sprint_region
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  !some generic spatial tests
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  recursive function srctool_test_if_coord_in_regcomp_circle(coord, regcomp)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_regcomp_circle'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    logical :: srctool_test_if_coord_in_regcomp_circle
    real(kind=real64) :: temp1

    srctool_test_if_coord_in_regcomp_circle = .false.
    ! do cartesian comparisons using the square of the radius
    if ((regcomp%coordsys_code .eq. COORDSYS_CODE_IMAGE) .or.  &
        (regcomp%coordsys_code .eq. COORDSYS_CODE_PHYSICAL)) then
      if ( regcomp%coordsys_code .eq. COORDSYS_CODE_IMAGE ) then
        temp1 = (coord%x - regcomp%pars(REGCOMP_PARS_CIRCLE_LONG))**2 + &
            (coord%y - regcomp%pars(REGCOMP_PARS_CIRCLE_LAT))**2
      else  ! if ( regcomp%coordsys_code .eq. COORDSYS_CODE_PHYSICAL ) then
        temp1 = (coord%rawx - regcomp%pars(REGCOMP_PARS_CIRCLE_LONG))**2 + &
            (coord%rawy - regcomp%pars(REGCOMP_PARS_CIRCLE_LAT))**2
      end if
      if ( temp1 .le. regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS)**2) then
        srctool_test_if_coord_in_regcomp_circle = .true.
      end if
    else
      ! do a test that is valid over full sphere
      if ( calc_ang_sep(coord%long,coord%lat,&
          regcomp%pars(REGCOMP_PARS_CIRCLE_LONG),&
          regcomp%pars(REGCOMP_PARS_CIRCLE_LAT)) .le. &
          regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS)) then
        srctool_test_if_coord_in_regcomp_circle = .true.
      end if
    end if
  end function srctool_test_if_coord_in_regcomp_circle
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  recursive function srctool_test_if_coord_in_regcomp_annulus(coord, regcomp)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_regcomp_annulus'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    logical :: srctool_test_if_coord_in_regcomp_annulus
    real(kind=real64) :: temp1

    srctool_test_if_coord_in_regcomp_annulus = .false.
    ! do cartesian comparisons using the square of the radius
    if ((regcomp%coordsys_code .eq. COORDSYS_CODE_IMAGE) .or.  &
        (regcomp%coordsys_code .eq. COORDSYS_CODE_PHYSICAL)) then
      if ( regcomp%coordsys_code .eq. COORDSYS_CODE_IMAGE ) then
        temp1 = (coord%x - regcomp%pars(REGCOMP_PARS_ANNULUS_LONG))**2 + &
            (coord%y - regcomp%pars(REGCOMP_PARS_ANNULUS_LAT))**2
      else ! if ( regcomp%coordsys_code .eq. COORDSYS_CODE_PHYSICAL ) then
        temp1 = (coord%rawx - regcomp%pars(REGCOMP_PARS_ANNULUS_LONG))**2 + &
            (coord%rawy - regcomp%pars(REGCOMP_PARS_ANNULUS_LAT))**2
      end if
      if ((temp1 .ge. regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS1)**2) .and. &
          (temp1 .lt. regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2)**2)) then
        srctool_test_if_coord_in_regcomp_annulus = .true.
      end if
    else
      temp1 = calc_ang_sep(coord%long,coord%lat,&
          regcomp%pars(REGCOMP_PARS_ANNULUS_LONG),&
          regcomp%pars(REGCOMP_PARS_ANNULUS_LAT))
      if ((temp1 .ge. regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS1)) .and. &
          (temp1 .lt. regcomp%pars(REGCOMP_PARS_ANNULUS_RADIUS2))) then
        srctool_test_if_coord_in_regcomp_annulus = .true.
      end if
    end if
  end function srctool_test_if_coord_in_regcomp_annulus
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  recursive function srctool_test_if_coord_in_regcomp_rectangle(coord, regcomp)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_regcomp_rectangle'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    logical :: srctool_test_if_coord_in_regcomp_rectangle
    real(kind=real64) :: x, y
    srctool_test_if_coord_in_regcomp_rectangle = .false.

    if ( regcomp%coordsys_code .eq. COORDSYS_CODE_IMAGE ) then
      x = coord%x
      y = coord%y
    else if ( regcomp%coordsys_code .eq. COORDSYS_CODE_PHYSICAL ) then
      x = coord%rawx
      y = coord%rawy
    else
      x = coord%long
      y = coord%lat
    endif

    ! first test y limits, then x limits
    if ( y .ge. regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT1)) then
      if ( y .lt. regcomp%pars(REGCOMP_PARS_RECTANGLE_LAT2)) then
        ! normal case - i.e. not wrapped around RA=0.0deg
        if ( regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1) .le. regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2) ) then
          if ( x .ge. regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1)) then
            if ( x .lt. regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2)) then
              srctool_test_if_coord_in_regcomp_rectangle = .true.
            endif
          endif
        else   ! wrapped around case - i.e. wrapped around RA=0.0deg
          if ((x .ge. regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG1)) .or. &
              (x .lt. regcomp%pars(REGCOMP_PARS_RECTANGLE_LONG2))) then
            srctool_test_if_coord_in_regcomp_rectangle = .true.
          end if
        end if
      end if
    end if
  end function srctool_test_if_coord_in_regcomp_rectangle
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  ! first project the coord onto the tangent plane centred at ellipse centre
  recursive function srctool_test_if_coord_in_regcomp_ellipse(coord, regcomp)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_regcomp_ellipse'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    logical :: srctool_test_if_coord_in_regcomp_ellipse

    real(kind=real64), parameter  :: oa_x = 0.0d0
    real(kind=real64), parameter  :: oa_y = 0.0d0
    real(kind=real64), parameter  :: inv_pixscale_x =  1.0d0
    real(kind=real64), parameter  :: inv_pixscale_y = -1.0d0
    real(kind=real64) :: x, y
    real(kind=real64) :: angle

    srctool_test_if_coord_in_regcomp_ellipse = .false.

    if ( regcomp%coordsys_code == COORDSYS_CODE_IMAGE ) then
      ! project x,y along major and minimum axes of inner and outer ellipses
      ! TODO check that these rotations are in the correct sense
      x = coord%x * cosd(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
      y = coord%y * sind(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
    else if ( regcomp%coordsys_code == COORDSYS_CODE_PHYSICAL ) then
      x = coord%rawx * cosd(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
      y = coord%rawy * sind(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
    else
      ! change sense of angle to match what ds9 would do
      angle = 90 + regcomp%pars(REGCOMP_PARS_ELLIPSE_ANGLE)
      call longlat2xy (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
           regcomp%pars(REGCOMP_PARS_ELLIPSE_LONG), regcomp%pars(REGCOMP_PARS_ELLIPSE_LAT), &
           angle, coord%long, coord%lat, x, y)
    end if
    if ( (x/regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMAJOR))**2 + &
         (y/regcomp%pars(REGCOMP_PARS_ELLIPSE_SEMIMINOR))**2 <= 1.0d0 ) then
      srctool_test_if_coord_in_regcomp_ellipse = .true.
    end if
  end function srctool_test_if_coord_in_regcomp_ellipse
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  ! first project the coord onto the tangent plane centred at ellannulus centre
  recursive function srctool_test_if_coord_in_regcomp_ellannulus(coord, regcomp)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_regcomp_ellannulus'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    logical :: srctool_test_if_coord_in_regcomp_ellannulus

    real(kind=real64), parameter  :: oa_x = 0.0d0
    real(kind=real64), parameter  :: oa_y = 0.0d0
    real(kind=real64), parameter  :: inv_pixscale_x =  1.0d0
    real(kind=real64), parameter  :: inv_pixscale_y = -1.0d0
    real(kind=real64) :: x1, y1, x2, y2 !, temp1, temp2
    real(kind=real64) :: angle1, angle2

    srctool_test_if_coord_in_regcomp_ellannulus = .false.

    if ( regcomp%coordsys_code == COORDSYS_CODE_IMAGE ) then
      ! project x,y along major and minimum axes of inner and outer ellipses
      ! TODO check that these rotations are in the correct sense
      x1 = coord%x * cosd(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
      y1 = coord%y * sind(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
      x2 = coord%x * cosd(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2))
      y2 = coord%y * sind(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2))
    else if ( regcomp%coordsys_code == COORDSYS_CODE_PHYSICAL ) then
      x1 = coord%rawx * cosd(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
      y1 = coord%rawy * sind(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1))
      x2 = coord%rawx * cosd(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2))
      y2 = coord%rawy * sind(regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2))

    else
      ! change sense of angle to match what ds9 would do
      angle1 = 90 + regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE1)
      call longlat2xy (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
          regcomp%pars(REGCOMP_PARS_ELLANNULUS_LONG), regcomp%pars(REGCOMP_PARS_ELLANNULUS_LAT), &
          angle1, coord%long, coord%lat, x1, y1 )
      if ( eq_real64(angle1,angle2)) then
        x2 = x1
        y2 = y1
      else
        angle2 = 90 + regcomp%pars(REGCOMP_PARS_ELLANNULUS_ANGLE2)
        call longlat2xy (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
             regcomp%pars(REGCOMP_PARS_ELLANNULUS_LONG), regcomp%pars(REGCOMP_PARS_ELLANNULUS_LAT), &
             angle2, coord%long, coord%lat, x2, y2 )
      end if
    end if

    if ( ((x1/regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR1))**2 + &
        (y1/regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR1))**2) >= 1.0d0 ) then
      if ( ((x2/regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMAJOR2))**2 + &
          (y2/regcomp%pars(REGCOMP_PARS_ELLANNULUS_SEMIMINOR2))**2) <= 1.0d0 ) then
        srctool_test_if_coord_in_regcomp_ellannulus = .true.
      end if
    end if
  end function srctool_test_if_coord_in_regcomp_ellannulus
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  ! box is described by centre coords, width, and height
  ! first project the coord onto the tangent plane centred at box centre
  recursive function srctool_test_if_coord_in_regcomp_box(coord, regcomp)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_regcomp_box'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    logical :: srctool_test_if_coord_in_regcomp_box

    real(kind=real64), parameter  :: oa_x = 0.0d0
    real(kind=real64), parameter  :: oa_y = 0.0d0
    real(kind=real64), parameter  :: inv_pixscale_x =  1.0d0
    real(kind=real64), parameter  :: inv_pixscale_y = -1.0d0
    real(kind=real64) :: x, y
    real(kind=real64) :: angle

    srctool_test_if_coord_in_regcomp_box = .false.

    if ( regcomp%coordsys_code == COORDSYS_CODE_IMAGE ) then
      x = coord%x - regcomp%pars(REGCOMP_PARS_BOX_LONG)
      y = coord%y - regcomp%pars(REGCOMP_PARS_BOX_LAT)
    else if ( regcomp%coordsys_code == COORDSYS_CODE_PHYSICAL ) then
      x = coord%rawx - regcomp%pars(REGCOMP_PARS_BOX_LONG)
      y = coord%rawy - regcomp%pars(REGCOMP_PARS_BOX_LAT)
    else
      ! change sense of angle to match what ds9 would do
      angle = 90 + regcomp%pars(REGCOMP_PARS_BOX_ANGLE)
      call longlat2xy (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
           regcomp%pars(REGCOMP_PARS_BOX_LONG), regcomp%pars(REGCOMP_PARS_BOX_LAT), &
           angle, coord%long, coord%lat, x, y)
    end if

    if ((ABS(x) <= 0.5d0*regcomp%pars(REGCOMP_PARS_BOX_WIDTH)) .and. &
        (ABS(y) <= 0.5d0*regcomp%pars(REGCOMP_PARS_BOX_HEIGHT))) then
      srctool_test_if_coord_in_regcomp_box = .true.
    end if
  end function srctool_test_if_coord_in_regcomp_box
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  ! Is the coordinate inside the masked region?
  recursive function srctool_test_if_coord_in_regcomp_mask(coord, regcomp)
    character(len=*), parameter :: subroutinename='srctool_test_if_coord_in_regcomp_mask'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    logical :: srctool_test_if_coord_in_regcomp_mask

    integer(kind=int32) :: ix, iy
    real(kind=real64) :: x, y, a, b

    ! FIXME: unclear what to do here for IMAGE and PHYSICAL coordinates
    if ( regcomp%coordsys_code .eq. COORDSYS_CODE_IMAGE ) then
      a = coord%x
      b = coord%y
    else if ( regcomp%coordsys_code .eq. COORDSYS_CODE_PHYSICAL ) then
      a = coord%rawx
      b = coord%rawy
    else
      a = coord%long
      b = coord%lat
    endif

    call wcs_world2pix(regcomp%mask%wcs, a, b, x, y)
    ix = 1 + int(x, int32)
    iy = 1 + int(y, int32)
    if ( ix >= 1 .and. ix <= regcomp%mask%naxis1 .and. &
         iy >= 1 .and. iy <= regcomp%mask%naxis2 ) then

       srctool_test_if_coord_in_regcomp_mask = regcomp%mask%ppix(ix,iy) > 0
    else
       srctool_test_if_coord_in_regcomp_mask = .false.
    end if

  end function srctool_test_if_coord_in_regcomp_mask
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  ! test depends on regcomp type
  ! return 0 if outside the region
  recursive subroutine srctool_regions_test_if_coord_in_regcomp(coord, regcomp, flag, ierr)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_regcomp'
    type(coord_type),    intent(in)  :: coord
    type(regcomp_type),  intent(in)  :: regcomp
    integer(kind=int32), intent(out) :: flag
    integer(kind=int32), intent(out) :: ierr
    type(coord_type)  :: local_coord

    ! check if we need to convert the sky coordinates to the region's native system
    if ( (regcomp%coordsys_code .ne. coord%coordsys_code) .and. &
         (regcomp%coordsys_code .ne. COORDSYS_CODE_IMAGE) .and. &
         (regcomp%coordsys_code .ne. COORDSYS_CODE_PHYSICAL)) then
      local_coord%coordsys_code = regcomp%coordsys_code
      call geom_convert_coordsys (coord%coordsys_code, coord%long, coord%lat,  &
          local_coord%coordsys_code, local_coord%long, local_coord%lat)
    else
      local_coord = coord  ! just take a local copy
    endif
    flag = 0

    select case (regcomp%shape_code)

    case (REGCOMP_SHAPE_CODE_CIRCLE)
      if ( srctool_test_if_coord_in_regcomp_circle(local_coord, regcomp) ) then
        flag = regcomp%weight
      endif

    case (REGCOMP_SHAPE_CODE_ANNULUS)
      if ( srctool_test_if_coord_in_regcomp_annulus(local_coord, regcomp) ) then
        flag = regcomp%weight
      endif

    case (REGCOMP_SHAPE_CODE_RECTANGLE)
      if ( srctool_test_if_coord_in_regcomp_rectangle(local_coord, regcomp) ) then
        flag = regcomp%weight
      endif

    case (REGCOMP_SHAPE_CODE_ELLIPSE)
      if ( srctool_test_if_coord_in_regcomp_ellipse(local_coord, regcomp) ) then
        flag = regcomp%weight
      endif

    case (REGCOMP_SHAPE_CODE_ELLANNULUS)
      if ( srctool_test_if_coord_in_regcomp_ellannulus(local_coord, regcomp) ) then
        flag = regcomp%weight
      endif

    case (REGCOMP_SHAPE_CODE_BOX)
      if ( srctool_test_if_coord_in_regcomp_box(local_coord, regcomp) ) then
        flag = regcomp%weight
      endif

    case (REGCOMP_SHAPE_CODE_MASK)
      if ( srctool_test_if_coord_in_regcomp_mask(local_coord, regcomp) ) then
        flag = regcomp%weight
      endif

    case default
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr,  "Unknown region component: " // trim(i2s(regcomp%shape_code)) )
      return

    end select

  end subroutine srctool_regions_test_if_coord_in_regcomp
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  ! test depends on region type
  recursive subroutine srctool_regions_test_if_coord_in_region(coord, region, flag, ierr)
    character(len=*), parameter :: subroutinename='srctool_regions_test_if_coord_in_region'
    type(coord_type),   intent(in)  :: coord
    type(region_type),   intent(in)  :: region
    integer(kind=int32), intent(out) :: flag, ierr
    integer(kind=int32) :: i, temp_flag
    flag = 0
    ierr = ERROR_LEVEL_OK

    do i=1,region%num_regcomps
      call srctool_regions_test_if_coord_in_regcomp(coord, region%pregcomp(i), temp_flag, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem calling srctool_test_if_coord_in_regcomp()")
        return
      endif
      flag = flag + temp_flag
    end do

  end subroutine srctool_regions_test_if_coord_in_region
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  recursive subroutine srctool_regions_resolve_region_wildcards (src_coord, region )
    type(coord_type), intent(in)       :: src_coord
    type(region_type), intent(inout)   :: region
    integer(kind=int32)             :: i

    do i=1,region%num_regcomps
      call srctool_regions_resolve_regcomp_wildcards (src_coord, region%pregcomp(i) )
!      if ( region%pregcomp(i)%wildcards(REGCOMP_PARS_GENERIC_LONG) ) then
!        region%pregcomp(i)%pars(REGCOMP_PARS_GENERIC_LONG) = local_coord%long
!      endif
!      if ( region%pregcomp(i)%wildcards(REGCOMP_PARS_GENERIC_LAT) ) then
!        region%pregcomp(i)%pars(REGCOMP_PARS_GENERIC_LAT) = local_coord%lat
!      endif
    end do
  end subroutine srctool_regions_resolve_region_wildcards
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  recursive subroutine srctool_regions_resolve_regcomp_wildcards (src_coord, regcomp )
    type(coord_type), intent(in)       :: src_coord
    type(regcomp_type), intent(inout)   :: regcomp
    type(coord_type) :: local_coord

    ! check if we need to convert the coordinates to the region's native system
    if ( regcomp%coordsys_code .ne. src_coord%coordsys_code ) then
      local_coord%coordsys_code = regcomp%coordsys_code
      call geom_convert_coordsys (src_coord%coordsys_code, src_coord%long, src_coord%lat,  &
          local_coord%coordsys_code, local_coord%long, local_coord%lat)
    else
      local_coord = src_coord
    endif

    if ( regcomp%wildcards(REGCOMP_PARS_GENERIC_LONG) ) then
      regcomp%pars(REGCOMP_PARS_GENERIC_LONG) = local_coord%long
    endif
    if ( regcomp%wildcards(REGCOMP_PARS_GENERIC_LAT) ) then
      regcomp%pars(REGCOMP_PARS_GENERIC_LAT) = local_coord%lat
    endif
  end subroutine srctool_regions_resolve_regcomp_wildcards
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  ! goal is to remove all parts of a region description which are not the central coords
  ! require that str_out is at least as long as str_in
  recursive subroutine srctool_regions_extract_central_coords (str_in, str_out)
    character(len=*), parameter :: subroutinename='srctool_regions_extract_central_coords'
    character(len=*),intent(in)  :: str_in
    character(len=*),intent(out)  :: str_out
    integer(kind=int32)  :: i,j1,j2,j3,j4,j5
    integer(kind=int32)  :: ncommas
    j1=0
    j2=0
    j3=0
    j4=0
    j5=0

    if ((len_trim(str_in) .gt. len(str_out)) .or. &
        (len_trim(str_in) .lt. 1 )) then
      str_out(:) = "?"
      return
    endif

    str_out(:) = trim(str_in(:))
!    write (*,'(A)'), "DEBUG0"

!    write (*,'(A,A,A)'), "DEBUG1 ='", trim(str_out), "'"

    call to_lower(str_out)
!    write (*,'(A,A,A)'), "DEBUG2 ='", trim(str_out), "'"

    ! first remove any comments (anything after a '#' character)
    i = INDEX(str_out,"#")
    if ( i .gt. 0 ) then
      str_out(i:) = " "  ! convert to white space
    end if
!    write (*,'(A,A,A)'), "DEBUG3 ='", trim(str_out), "'"

    ! get rid of everything between last semicolon and first numeric character
    ! if no semicolons, then we can remove everything up to first numeric character
    j1 = INDEX(str_out,";",back=.true.)
!    if ( j1 .lt. 1 ) j1 = 0
    if ( (j1 .ge. 0) .and. (j1 .lt. len(str_out))) then
      j2 = SCAN(str_out(j1+1:),"1234567890.")
      if ( j2 .ge. 1 ) then
        str_out(j1+1:j1+j2-1) = " "
      endif
    endif
!    write (*,'(A,A,A)'), "DEBUG4 ='", trim(str_out), "'"

!    ! strip out brackets and commas
!    j1 = MAX(1,INDEX(str_out,";",back=.true.))
!    do i=j1,len_trim(str_out)
!      if ((str_out(i:i) .eq. ",") .or. &
!          (str_out(i:i) .eq. "(") .or. &
!          (str_out(i:i) .eq. ")") ) then
!        str_out(i:i) = " "
!      endif
!    end do
!    write (*,'(A,A,A)'), "DEBUG5='", trim(str_out), "'"

    ! strip out brackets and count commas
    ncommas = 0
    j1 = MAX(1,INDEX(str_out,";",back=.true.))
    do i=j1,len_trim(str_out)
      if ((str_out(i:i) .eq. "(") .or. &
          (str_out(i:i) .eq. ")") ) then
        str_out(i:i) = " "
      endif
      if (str_out(i:i) .eq. ",") then
        ncommas = ncommas + 1
      endif
    end do
!    write (*,'(A,A,A)'), "DEBUG5 ='", trim(str_out), "'"

    ! if we have commas then strip out everything after the second comma (if existing)
    if ( ncommas .ge. 1 ) then
      ncommas = 0
      j1 = MAX(1,INDEX(str_out,";",back=.true.))
      do i=j1,len_trim(str_out)
        if (str_out(i:i) .eq. ",") then
          ncommas = ncommas + 1
        endif
        if ( ncommas .ge. 2 ) then
          str_out(i:i) = " "
        endif
      end do
!      write (*,'(A,A,A)'), "DEBUG6 ='", trim(str_out), "'"
    else

      !insert a comma between longitude and latitude terms
      ! j1+j2+1 is index of first numeric value of first coordinate
      ! so scan from there onwards for the first white space
      if ( j1+j2+1 .ge. 1 ) then
        j3 = INDEX(str_out(j1+j2+1:)," ")
        if ( j3 .ge. 1 ) then
          str_out(j1+j2+j3:j1+j2+j3) = ","
        endif
      endif

!      write (*,'(A,A,A)'), "DEBUG7a='", trim(str_out), "'"

      !get rid of everything after the second coordinate
      ! j1+j2+j3+1 is index of first numeric value of second coordinate
      ! so scan from there onwards for the first white space
      j4 = SCAN(str_out(j1+j2+j3+1:),"1234567890.")
      if ( j4 .ge. 1 ) then
        j5 = INDEX(str_out(j1+j2+j3+j4+1:)," ")
        if ( j5 .ge. 1 ) then
          str_out(j1+j2+j3+j4+j5:) = " "
        endif
      endif
!      write (*,'(A,A,A)'), "DEBUG7b='", trim(str_out), "'"
    endif

!    write (*,'(A,A,A)'), "DEBUG8='", trim(str_out), "'"


  end subroutine srctool_regions_extract_central_coords
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  subroutine srctool_regions_test_harness(ierr)
    character(len=*), parameter :: subroutinename='srctool_regions_test_harness'
    integer(kind=int32), intent(inout) :: ierr

    type(region_type) :: region
    type(coord_type)  :: coord
    type(coord_type)  :: temp_coord
    integer(kind=int32) :: flag
    character(len=STR_MAX) :: str_in
    character(len=STR_MAX) :: str_out
    character(len=STR_MAX) :: normal_buffer
    character(len=96) :: short_buffer
    real(kind=real64) :: long, lat
    integer(kind=int32) :: coordsys_code
    logical           :: success_flag
    type(regcomp_type) :: regcomp
    integer(kind=int32) :: num_errs

    ! skip processing if have already found errors
    if (ierr .ne. ERROR_LEVEL_OK) return


    num_errs = 0

    call wrschatty(subroutinename, "Testing the generic regions module")

    call srctool_regions_init_region(region)
    call srctool_generic_init_coord(coord)
    call srctool_generic_init_coord(temp_coord)


    !------------------------------------------------------------
    ! Test 1: Check if point falls inside/outside simple region
    call srctool_regions_init_regcomp(regcomp)
    call srctool_regions_interpret_regcomp_desc ("icrs;circle(10.0,+10.0,1.0d)", &
        regcomp, temp_coord, ierr)
    call srctool_regions_add_regcomp(region, regcomp)

    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    ierr = ERROR_LEVEL_OK


    coord%coordsys_code=COORDSYS_CODE_ICRS
    coord%long=10.0d0
    coord%lat=+10.0d0
    flag = 0
    call srctool_regions_test_if_coord_in_region (coord, region, flag, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    if ( flag .ne. 1 ) then
      call srctool_regions_print_region(region, ierr)
      call wrserr(subroutinename, ierr, "Problem with srctool_regions_test_if_coord_in_region")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK

    coord%long=12.0d0
    coord%lat=+10.0d0
    flag = 0
    call srctool_regions_test_if_coord_in_region (coord, region, flag, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    if ( flag .ne. 0 ) then
      call srctool_regions_print_region(region, ierr)
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, "Problem with srctool_regions_test_if_coord_in_region")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK

    ! try some points and regions around poles
    call srctool_regions_interpret_regcomp_desc ("icrs;circle(12.00,+89.51,0.02d)", &
        region%pregcomp(1), temp_coord, ierr )
    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    ierr = ERROR_LEVEL_OK

    coord%long=12.00d0
    coord%lat=+89.50d0
    flag = 0
    call srctool_regions_test_if_coord_in_region (coord, region, flag, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    if ( flag .ne. 1 ) then
      call srctool_regions_print_region(region, ierr)
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, "Problem with srctool_regions_test_if_coord_in_region")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK

    ! try some points and regions around poles
    coord%long=00.00d0
    coord%lat=+89.50d0
    flag = 0
    call srctool_regions_test_if_coord_in_region (coord, region, flag, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    if ( flag .ne. 0 ) then
      call srctool_regions_print_region(region, ierr)
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, "Problem with srctool_regions_test_if_coord_in_region")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK



    ! Test a photon at origin and a region at opposite pole
    call srctool_regions_interpret_regcomp_desc ("icrs;circle(180.0,+0.0,1.0d)", &
        region%pregcomp(1), temp_coord, ierr )
    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    coord%long=0.0d0
    coord%lat=0.0d0
    flag = 0
    call srctool_regions_test_if_coord_in_region (coord, region, flag, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK )  num_errs = num_errs + 1
    if ( flag .ne. 0 ) then
      call srctool_regions_print_region(region, ierr)
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, "Problem with srctool_regions_test_if_coord_in_region")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK
    !------------------------------------------------------------

    if ( .false. ) then
      str_in = 'icrs;circle 01:23:45.67,+01:46:00.0,300.0"'
      call srctool_regions_extract_central_coords(str_in, str_out)
      call convert_longlat_to_decimal (TRIM(str_out), long, lat, coordsys_code, success_flag)
      write(*,'(5A,2(x,F12.6),x,A)') &
          'TEST1:', trim(str_in), ' => ',  trim(str_out),  ' => ', long, lat, trim(COORDSYS_STR(coordsys_code))

      str_in = 'icrs;circle 01 23 45.67,+01 46 00.0,300.0"'
      call srctool_regions_extract_central_coords(str_in, str_out)
      call convert_longlat_to_decimal (TRIM(str_out), long, lat, coordsys_code, success_flag)
      write(*,'(5A,2(x,F12.6),x,A)') &
          'TEST1:', trim(str_in), ' => ',  trim(str_out),  ' => ', long, lat, trim(COORDSYS_STR(coordsys_code))

      str_in = 'icrs;point(01 23 45.67,+01 46 00.0)'
      call srctool_regions_extract_central_coords(str_in, str_out)
      call convert_longlat_to_decimal (TRIM(str_out), long, lat, coordsys_code, success_flag)
      write(*,'(5A,2(x,F12.6),x,A)') &
          'TEST1:', trim(str_in), ' => ',  trim(str_out),  ' => ', long, lat, trim(COORDSYS_STR(coordsys_code))

      str_in = 'icrs;point(01h23m45.67s,+01d46m00.0s) # gubbings and blah,,,'
      call srctool_regions_extract_central_coords(str_in, str_out)
      call convert_longlat_to_decimal (TRIM(str_out), long, lat, coordsys_code, success_flag)
      write(*,'(5A,2(x,F12.6),x,A)') &
          'TEST1:', trim(str_in), ' => ',  trim(str_out),  ' => ', long, lat, trim(COORDSYS_STR(coordsys_code))

      str_in = 'point(01 23 45.67,+01 46 00.0)'
      call srctool_regions_extract_central_coords(str_in, str_out)
      call convert_longlat_to_decimal (TRIM(str_out), long, lat, coordsys_code, success_flag)
      write(*,'(5A,2(x,F12.6),x,A)') &
          'TEST1:', trim(str_in), ' => ',  trim(str_out),  ' => ', long, lat, trim(COORDSYS_STR(coordsys_code))

      str_in = '01 23 45.67,+01 46 00.0'
      call srctool_regions_extract_central_coords(str_in, str_out)
      call convert_longlat_to_decimal (TRIM(str_out), long, lat, coordsys_code, success_flag)
      write(*,'(5A,2(x,F12.6),x,A)') &
          'TEST1:', trim(str_in), ' => ',  trim(str_out),  ' => ', long, lat, trim(COORDSYS_STR(coordsys_code))

    endif



    call srctool_regions_init_regcomp(regcomp)
    regcomp%shape_code = REGCOMP_SHAPE_CODE_CIRCLE
    regcomp%coordsys_code = COORDSYS_CODE_ICRS
    regcomp%num_pars = REGCOMP_NUM_PARS_CIRCLE
    regcomp%num_pars_read = REGCOMP_NUM_PARS_CIRCLE
    regcomp%pars(REGCOMP_PARS_CIRCLE_LONG) = 123.456789d0
    regcomp%pars(REGCOMP_PARS_CIRCLE_LAT) = -12.3456789d0
    regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS) = 0.1234567d0
    regcomp%wildcards(:) = .false.
    regcomp%weight = 1
    regcomp%max_radius = regcomp%pars(REGCOMP_PARS_CIRCLE_RADIUS)
    call srctool_regions_print_regcomp(regcomp, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Errors found during test of srctool_regions_print_regcomp() ")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK


    ! we expect the string to fit here:
    call srctool_regions_sprint_regcomp(regcomp, normal_buffer, "DETAILS", ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Errors found during test1 of srctool_regions_sprint_regcomp()")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK

    ! but we expect the string to be too short here:
    call srctool_regions_sprint_regcomp(regcomp, short_buffer, "DETAILS", ierr)
    if (ierr .eq. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ERROR_LEVEL_SEVERE, "Errors found during test2 of srctool_regions_sprint_regcomp()")
      num_errs = num_errs + 1
    endif
    ierr = ERROR_LEVEL_OK


    if (num_errs .ne. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "At least one error found during region tests: num_errs=" // trim(i2s(num_errs)) )
    else
      ierr = ERROR_LEVEL_OK
    endif

  end subroutine srctool_regions_test_harness
  !-----------------------------------------------------------------------------------



end module srctool_generic_regions
