! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_lc

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-25
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling LCs
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         ! module with some srctool-specific constants and datatypes
  use srctool_misc              ! module with some srctool-specific constants and datatypes
  use srctool_gti               ! module with some srctool-specific constants and datatypes
  use srctool_owner             ! handle event ownership

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_LC_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_lc
  public  :: srctool_lc_empty_lc
  public  :: srctool_lc_empty_lclist
  private :: srctool_lc_write_lc_keys
  private :: srctool_lc_calc_lc_rates
  public  :: srctool_lc_calc_lclist_rates
  private :: srctool_lc_write_lc
  public  :: srctool_lc_write_lclist
  private :: srctool_lc_prepare_lc_template
  private :: srctool_lc_setup_lc
  public  :: srctool_lc_setup_lclist
  public  :: srctool_lc_accumulate_lc
  public  :: srctool_lc_accumulate_lclist
  public  :: srctool_lc_coadd_lclist
  public  :: srctool_lc_get_lc_index
  public  :: srctool_lc_get_next_lc_index


contains


  !-----------------------------------------------------------------------------------
  !-----------------------------------------------------------------------------------
  !--------------------------- START of Lightcurve handling --------------------------
  !-----------------------------------------------------------------------------------
  !-----------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_lc_empty_lc(lc)
    type(lc_type), intent(inout)   :: lc
    integer(kind=int32) :: e!, i
!    character(len=FITSIO_KEY_MAXLEN)             :: str_buffer1
!    character(len=FITSIO_KEY_MAXLEN)             :: str_buffer2
    call fitswrap_init_hduinfo(lc%hduinfo)

    lc%lctype      =  "NULL"
    lc%lctype_code =  LCTYPE_CODE_NULL
    lc%lcpars(:)   =  0.0d0
    lc%num_lcpars  =  0

    call fitswrap_init_key(lc%key_timedel , "TIMEDEL" , "" )
    call fitswrap_init_key(lc%key_timezero, "TIMEZERO", "[s]")

    call fitswrap_init_key(lc%key_telapse,  "TELAPSE" , "[s] Total observation duration"                 )
    call fitswrap_init_key(lc%key_ontime,   "ONTIME"  , "[s] On-target exposure time (pre deadtime corr)")
    call fitswrap_init_key(lc%key_livetime, "LIVETIME", "[s] On-target active-detector exposure time"    )
    call fitswrap_init_key(lc%key_exposure, "EXPOSURE", "[s] Fully corrected effective exposure time")

    call fitswrap_init_key(lc%key_tierrela, "TIERRELA", "")
    call fitswrap_init_key(lc%key_tierabso, "TIERABSO", "[s]")
    call fitswrap_init_key(lc%key_clockcor, "CLOCKCOR", "True if clock correction was applied"      )
    call fitswrap_init_key(lc%key_backapp , "BACKAPP" , "True if background was subtracted"         )
    call fitswrap_init_key(lc%key_deadapp , "DEADAPP" , "True if deadtime correction was applied"   )
    call fitswrap_init_key(lc%key_vignapp , "VIGNAPP" , "True if vignetting correction was applied" )

    call fitswrap_init_key(lc%key_flagsel , "FLAGSEL" , "Flag selection mask which has been applied" )
    call fitswrap_init_key(lc%key_pat_sel , "PAT_SEL" , "Pattern mask which has been applied")

    call fitswrap_init_key(lc%key_src_region,  "SRC_REG",  "Source extraction (ds9) region used")
    call fitswrap_init_key(lc%key_back_region, "BACK_REG", "Background extraction (ds9) region used")

    do e=1,MAX_LCBANDS
!      write(str_buffer1, '(A,I0)') "EMIN", e
!      write(str_buffer2, '(A,I0)') "[keV] Min. energy bound,band ", e
      call fitswrap_init_key(lc%key_emin(e), &
          "EMIN" // trim(i2s(e)), "[keV] Min. energy bound,band " // trim(i2s(e)))

!      write(str_buffer1, '(A,I0)') "EMAX", e
!      write(str_buffer2, '(A,I0)') "[keV] Max. energy bound,band ", e
!      call fitswrap_init_key(lc%key_emax(e), trim(str_buffer1), trim(str_buffer2))
      call fitswrap_init_key(lc%key_emax(e), &
          "EMAX" // trim(i2s(e)), "[keV] Max. energy bound,band " // trim(i2s(e)))

!      write(str_buffer1, '(A,I0)') "CTS_SC", e
!      write(str_buffer2, '(A,I0)') "Cts in source region,band ", e
!      call fitswrap_init_key(lc%key_ncounts_src(e), trim(str_buffer1), trim(str_buffer2))
      call fitswrap_init_key(lc%key_ncounts_src(e), &
          "CTS_SC" // trim(i2s(e)), "Cts in source region,band " // trim(i2s(e)))

!      write(str_buffer1, '(A,I0)') "CTS_BK", e
!      write(str_buffer2, '(A,I0)') "Cts in bkgnd region,band ", e
!      call fitswrap_init_key(lc%key_ncounts_back(e), trim(str_buffer1), trim(str_buffer2))
      call fitswrap_init_key(lc%key_ncounts_back(e), &
          "CTS_BK" // trim(i2s(e)), "Cts in bkgnd region,band " // trim(i2s(e)))
    end do

!    call fitswrap_init_key(lc%key_ninst, "NINST", "Number of instruments that were merged")
!    do i=1,MAX_INST
!!      write(str_buffer1, '(A,I0)') "INSTRUM", i
!     call fitswrap_init_key(lc%pkey_instrum(i), "INSTRUM" // trim(i2s(i)), "Contributing instrument name")
!    end do

    if ( allocated(lc%ptime         ))  deallocate (lc%ptime         )
    if ( allocated(lc%ptimedel      ))  deallocate (lc%ptimedel      )
    if ( allocated(lc%pcounts       ))  deallocate (lc%pcounts       )
    if ( allocated(lc%prate         ))  deallocate (lc%prate         )
    if ( allocated(lc%prate_err     ))  deallocate (lc%prate_err     )
    if ( allocated(lc%pback_counts  ))  deallocate (lc%pback_counts  )
    if ( allocated(lc%pfracarea     ))  deallocate (lc%pfracarea     )
    if ( allocated(lc%pfractime     ))  deallocate (lc%pfractime     )
    if ( allocated(lc%pfracexp      ))  deallocate (lc%pfracexp      )
    if ( allocated(lc%poffaxis      ))  deallocate (lc%poffaxis      )
    if ( allocated(lc%pbackratio    ))  deallocate (lc%pbackratio    )
    if ( allocated(lc%ptmin         ))  deallocate (lc%ptmin         )
    if ( allocated(lc%ptmax         ))  deallocate (lc%ptmax         )
    lc%num_rows   = 0
    lc%num_ebands = 0
    lc%to_consider = .false.

    lc%lcgamma = 2.0d0
    lc%int_nenergies = 1
    if ( allocated(lc%pint_e))       deallocate (lc%pint_e)
    if ( allocated(lc%pint_weights))  deallocate (lc%pint_weights)


  end subroutine srctool_lc_empty_lc
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_lc_empty_lclist(lclist)
    type(lclist_type), intent(inout)   :: lclist
    integer(kind=int32) :: i
    do i=0,MAX_INST
      call srctool_lc_empty_lc(lclist%lc(i))
    end do
    lclist%num_lcs = 0
  end subroutine srctool_lc_empty_lclist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! write lc fits keys info to the chdu
  ! TODO - rest of keywords - check existing ones too
  subroutine srctool_lc_write_lc_keys(unit, lc, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_write_lc_keys'
    integer(kind=int32), intent(in)  :: unit
    type(lc_type),       intent(in)  :: lc
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)    :: status
    integer(kind=int32)    :: e!, i
    ierr = ERROR_LEVEL_OK
    status = 0


    !-----------------------------------
    ! write the standard keywords to the binary table HDU:
    call fitswrap_write_hduinfo_quiet(unit, lc%hduinfo, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      return
    end if


    call fitswrap_write_key (unit, lc%key_timedel,  ierr)

    call fitswrap_write_key (unit, lc%key_telapse,  ierr)
    call fitswrap_write_key (unit, lc%key_ontime,   ierr)
    call fitswrap_write_key (unit, lc%key_livetime, ierr)
    call fitswrap_write_key (unit, lc%key_exposure, ierr)

    call fitswrap_write_key (unit, lc%key_tierrela, ierr)
    call fitswrap_write_key (unit, lc%key_tierabso, ierr)
    call fitswrap_write_key (unit, lc%key_clockcor, ierr)
    call fitswrap_write_key (unit, lc%key_backapp,  ierr)
    call fitswrap_write_key (unit, lc%key_deadapp,  ierr)
    call fitswrap_write_key (unit, lc%key_vignapp,  ierr)

    call fitswrap_write_key (unit, lc%key_flagsel, ierr)
    call fitswrap_write_key (unit, lc%key_pat_sel, ierr)

    ! write the energy band keywords
    do e=1,lc%num_ebands
      call fitswrap_write_key (unit, lc%key_emin(e), ierr)
      call fitswrap_write_key (unit, lc%key_emax(e), ierr)
      call fitswrap_write_key (unit, lc%key_ncounts_src(e), ierr)
      call fitswrap_write_key (unit, lc%key_ncounts_back(e), ierr)
    end do

!    if ( trim(lc%hduinfo%key_instrume%value) .eq. "MERGED" ) then
!      call fitswrap_write_key (unit, lc%key_ninst,    ierr)
!      do i=1,lc%key_ninst%value
!        call fitswrap_write_key (unit, lc%pkey_instrum(i), ierr)
!      end do
!   end if

    call fitswrap_write_key (unit, lc%key_src_region,    ierr)
    call fitswrap_write_key (unit, lc%key_back_region,   ierr)


    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    ! TODO - rest of keywords
    !-----------------------------------

  end subroutine srctool_lc_write_lc_keys


  !-------------------------------------------------------------------------------
  subroutine srctool_lc_calc_lc_rates(lc, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_calc_lc_rates'
    type(lc_type),  intent(inout) :: lc
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)    :: i, e , ei, j, ei1, ei2, ej1, ej2
    real(kind=real64)      :: temp_fracrsp    ! working variable
    real(kind=real64)      :: err_s, err_b    ! working variables
    logical             :: last_bin_empty
    logical             :: this_bin_empty
!    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK

    do i=1,lc%num_rows
      do e=1,lc%num_ebands
        ei = (i-1)*lc%num_ebands + e
        ! net rate = (cts_src - cts_back*backratio)/fracarea*fractime*timedel
        temp_fracrsp = REAL(lc%ptimedel(i) * lc%pfracexp(ei), real64)
        if ( temp_fracrsp > 0.0e0 ) then
          lc%prate(ei) = REAL((REAL(lc%pcounts(ei)) - &
              REAL(lc%pback_counts(ei))*lc%pbackratio(i))/ &
              temp_fracrsp, real32)

          ! Combine foreground and background errors providing we have enough counts
          ! (Gehrels doesn't work with error combination formula)
          if( lc%pcounts(ei) + lc%pback_counts(ei) >= 25 ) then
            err_s = sqrt(real(lc%pcounts(ei)))
            err_b = sqrt(real(lc%pback_counts(ei)))*lc%pbackratio(i)
            lc%prate_err(ei) = real(sqrt(err_s**2 + err_b**2) / &
                temp_fracrsp, real32)
          else
            lc%prate_err(ei) = NULLVAL_REAL32
          end if
        else
          lc%prate(ei)     = NULLVAL_REAL32 !
          lc%prate_err(ei) = NULLVAL_REAL32 !
          lc%pbackratio(i) = NULLVAL_REAL64 !
        end if
      end do
    end do

    ! If the lc typecode is LCTYPE_CODE_*MINUS then filter the lc to remove all the zero bins
    if ((lc%lctype_code .eq. LCTYPE_CODE_REGULAR_MINUS) .or. &
        (lc%lctype_code .eq. LCTYPE_CODE_SRCGTI_MINUS) ) then
      j = 0
      do i=1,lc%num_rows
        ei1 = (i-1)*lc%num_ebands + 1
        ei2 = (i-1)*lc%num_ebands + lc%num_ebands

        if ((MAXVAL(lc%pcounts(ei1:ei2))       .le. 0    ) .and. &
            (MAXVAL(lc%pback_counts(ei1:ei2))  .le. 0    ) .and. &
            (MAXVAL(lc%pfracexp(ei1:ei2))      .le. 0.0e0) ) then
          this_bin_empty = .true.
        else
          this_bin_empty = .false.
        end if

        if ( this_bin_empty ) then
        else
          j = j + 1
          ej1 = (j-1)*lc%num_ebands + 1
          ej2 = (j-1)*lc%num_ebands + lc%num_ebands
          lc%ptime(j)              = lc%ptime(i)
          lc%ptimedel(j)           = lc%ptimedel(i)
          lc%pcounts(ej1:ej2)      = lc%pcounts(ei1:ei2)
          lc%prate(ej1:ej2)        = lc%prate(ei1:ei2)
          lc%prate_err(ej1:ej2)    = lc%prate_err(ei1:ei2)
          lc%pback_counts(ej1:ej2) = lc%pback_counts(ei1:ei2)
          lc%pfracarea(ej1:ej2)    = lc%pfracarea(ei1:ei2)
          lc%pfractime(j)          = lc%pfractime(i)
          lc%pfracexp(ej1:ej2)     = lc%pfracexp(ei1:ei2)
          lc%poffaxis(j)           = lc%poffaxis(i)
          lc%pbackratio(j)         = lc%pbackratio(i)
        end if
      end do

      ! call wrschatty (subroutinename, &
      !     " DEBUG: Removed empty lc bins: new num_rows=" // &
      !     trim(i2s(j)) // " (was " // trim(i2s(lc%num_rows)) // ")" )
      lc%num_rows = j
    end if


    ! If the lc typecode is LCTYPE_CODE_REGULAR_PLUS then filter the lc to compact all the consecutive zero bins
    if ( lc%lctype_code .eq. LCTYPE_CODE_REGULAR_PLUS ) then
      last_bin_empty = .false.
      j = 0
      do i=1,lc%num_rows
        ei1 = (i-1)*lc%num_ebands + 1
        ei2 = (i-1)*lc%num_ebands + lc%num_ebands

        if ((MAXVAL(lc%pcounts(ei1:ei2))       .le. 0    ) .and. &
            (MAXVAL(lc%pback_counts(ei1:ei2))  .le. 0    ) .and. &
            (MAXVAL(lc%pfracexp(ei1:ei2))      .le. 0.0e0) ) then
          this_bin_empty = .true.
        else
          this_bin_empty = .false.
        end if

        if ( this_bin_empty .and.  last_bin_empty ) then
          lc%ptime(j)    = lc%ptime(j)    + 0.5d0*lc%ptimedel(i)
          lc%ptimedel(j) = lc%ptimedel(j) + lc%ptimedel(i)
          lc%ptmax(j)    = lc%ptmax(i)
        else
          j = j + 1
          ej1 = (j-1)*lc%num_ebands + 1
          ej2 = (j-1)*lc%num_ebands + lc%num_ebands
          lc%ptime(j)              = lc%ptime(i)
          lc%ptimedel(j)           = lc%ptimedel(i)
          lc%pcounts(ej1:ej2)      = lc%pcounts(ei1:ei2)
          lc%prate(ej1:ej2)        = lc%prate(ei1:ei2)
          lc%prate_err(ej1:ej2)    = lc%prate_err(ei1:ei2)
          lc%pback_counts(ej1:ej2) = lc%pback_counts(ei1:ei2)
          lc%pfracarea(ej1:ej2)    = lc%pfracarea(ei1:ei2)
          lc%pfractime(j)          = lc%pfractime(i)
          lc%pfracexp(ej1:ej2)     = lc%pfracexp(ei1:ei2)
          lc%poffaxis(j)           = lc%poffaxis(i)
          lc%pbackratio(j)         = lc%pbackratio(i)

        end if
        last_bin_empty = this_bin_empty
      end do

      ! call wrschatty (subroutinename, &
      !     " DEBUG: Compacted number of lc bins: new num_rows=" // &
      !     trim(i2s(j)) // " (was " // trim(i2s(lc%num_rows)) // ")" )

      lc%num_rows = j
    end if

  end subroutine srctool_lc_calc_lc_rates
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  subroutine srctool_lc_calc_lclist_rates(lclist, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_calc_lclist_rates'
    type(lclist_type),  intent(inout) :: lclist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    ierr = ERROR_LEVEL_OK

    do i=0,MAX_INST
      if ( (i .eq. 0) .or. lclist%lc(i)%to_consider ) then
        call srctool_lc_calc_lc_rates(lclist%lc(i), ierr)
      end if
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem calculating rate column for lightcurves")
        return
      end if
    end do

  end subroutine srctool_lc_calc_lclist_rates
  !-------------------------------------------------------------------------------






  !-------------------------------------------------------------------------------
  ! write a lc file to disk
  ! we expect that all relevant parameters are already set up
  subroutine srctool_lc_write_lc(lc, gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_write_lc'
    type(lc_type),      intent(inout) :: lc
    type(gtilist_type),  intent(in) :: gtilist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32), parameter :: blank_bitpix = 16
    integer(kind=int32), parameter :: blank_naxis = 0
    integer(kind=int32), parameter :: blank_naxes(1) = (/ 0 /)
!    logical(kind=log32), parameter :: simple = .true.
!    logical,             parameter :: simple = .true.

    integer(kind=int32), parameter :: COLNUM_TIME          =  1
    integer(kind=int32), parameter :: COLNUM_TIMEDEL       =  2
    integer(kind=int32), parameter :: COLNUM_COUNTS        =  3
    integer(kind=int32), parameter :: COLNUM_RATE          =  4
    integer(kind=int32), parameter :: COLNUM_RATE_ERR      =  5
    integer(kind=int32), parameter :: COLNUM_BACK_COUNTS   =  6
!    integer(kind=int32), parameter :: COLNUM_BACK_RATE     =  7
!    integer(kind=int32), parameter :: COLNUM_BACK_RATE_ERR =  8
    integer(kind=int32), parameter :: COLNUM_FRACAREA      =  7
    integer(kind=int32), parameter :: COLNUM_FRACTIME      =  8
    integer(kind=int32), parameter :: COLNUM_FRACEXP       =  9
    integer(kind=int32), parameter :: COLNUM_OFFAXIS       = 10
    integer(kind=int32), parameter :: COLNUM_BACKRATIO     = 11
    integer(kind=int32), parameter :: ncols=11



    character(len=*),    parameter :: ttype(ncols) = &
        (/"TIME         ","TIMEDEL      ","COUNTS       ","RATE         ",&
          "RATE_ERR     ","BACK_COUNTS  ",& !"BACK_RATE    ","BACK_RATE_ERR",&
          "FRACAREA     ","FRACTIME     ","FRACEXP      ","OFFAXIS      ",&
          "BACKRATIO    "/)
    character(len=*),    parameter :: tunit(ncols) = &
        (/"s    "        ,"s    "        ,"count"        ,"ct/s "        ,&
          "ct/s "        ,"count"        ,& !"ct/s "        ,"ct/s "        ,&
          "ratio"        ,"ratio"        ,"ratio"        ,"deg  "        ,&
          "ratio"        /)
    character(len=4)  :: tform(ncols) ! to be filled in depending on number of ebands

!    character(len=*),    parameter :: extname = "RATE"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int64), parameter :: frow=1, felem=1

!    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)    :: unit
    integer(kind=int32)    :: status


    ierr = ERROR_LEVEL_OK
    status = 0
    unit = 0
!    str_buffer = ""

    if ( len_trim(lc%hduinfo%filename) .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Illegal filename for lc file")
      return
    end if

!    write (str_buffer,'(A,A)')  "Writing lc file, filename= ", trim(lc%hduinfo%filename)
    call wrschatty (subroutinename,  "Writing lc file, filename= " // trim(lc%hduinfo%filename))

    ! set up the tform params
    write  (tform(COLNUM_TIME)         ,'(A)')    "D"
    write  (tform(COLNUM_TIMEDEL)      ,'(A)')    "D"
    write  (tform(COLNUM_COUNTS)       ,'(I0,A)') lc%num_ebands, "J"
    write  (tform(COLNUM_RATE)         ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_RATE_ERR)     ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_BACK_COUNTS)  ,'(I0,A)') lc%num_ebands, "E"
!    write  (tform(COLNUM_BACK_RATE)    ,'(I0,A)'), lc%num_ebands, "E"
!    write  (tform(COLNUM_BACK_RATE_ERR),'(I0,A)'), lc%num_ebands, "E"
    write  (tform(COLNUM_FRACAREA)     ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_FRACTIME)     ,'(A)')    "E"
    write  (tform(COLNUM_FRACEXP)      ,'(I0,A)') lc%num_ebands, "E"
    write  (tform(COLNUM_OFFAXIS)      ,'(A)')    "E"
    write  (tform(COLNUM_BACKRATIO)    ,'(A)')    "D"

    !-----------------------------------
    ! get a fortran unit
    call FTGIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT(unit,lc%hduinfo%filename,FITSIO_BLOCKSIZE, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

    ! append a new empty primary image hdu and write date keyword to the HDU
    call FTIIMG(unit,blank_bitpix,blank_naxis,blank_naxes, status)
    call FTPDAT(unit, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
!
!    !-----------------------------------
!    ! write the standard FITS header:
!    call FTPHPS(unit,bitpix,naxis,naxes,status)
!    if (status .ne. 0) then
!      ierr = ERROR_LEVEL_SEVERE
!      call fitswrap_write_fits_error(subroutinename, status)
!      return
!    end if
!    !-----------------------------------
!
!    !-----------------------------------
!    ! write the necessary keywords to the primary HDU
!    ! TODO
!    ! write date keyword
!    call  FTPDAT(unit, status)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------

    !-----------------------------------
    ! append a new binary table HDU
    !FTIBIN(unit,nrows,tfields,ttype,tform,tunit,extname,varidat > status)
    call FTIBIN(unit,lc%num_rows,ncols,ttype,tform,tunit,trim(lc%hduinfo%key_extname%value),varidat,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------
!
!    !-----------------------------------
!    ! append a new table HDU
!    call FTCRHD(unit, status)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------
!
!    !-----------------------------------
!    ! write header info for the binary table:
!    call FTPHBN(unit,lc%num_rows,ncols,ttype,tform,tunit,extname,varidat,status)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------

    ! update TSTART/TSTOP using GTI
    if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
       call fitswrap_set_key(lc%hduinfo%key_tstart, gtilist%pstart(1))
       call fitswrap_set_key(lc%hduinfo%key_tstop, gtilist%pstop(gtilist%num_rows))
    end if

    !-----------------------------------
    call srctool_lc_write_lc_keys(unit, lc, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing lc fits keys")
      return
    end if
    !-----------------------------------

    !-----------------------------------
    ! now write the column data to the file
    ! format: FTPCL[SLBIJKEDCM](unit,colnum,frow,felem,nelements,values, > status)

    ! TODO - write data in chunks

    call  FTPCLD(unit,COLNUM_TIME         ,frow,felem,lc%num_rows,              lc%ptime(:)       ,status)
    call  FTPCLD(unit,COLNUM_TIMEDEL      ,frow,felem,lc%num_rows,              lc%ptimedel(:)    ,status)
    call  FTPCLJ(unit,COLNUM_COUNTS       ,frow,felem,lc%num_rows*lc%num_ebands,lc%pcounts(:)     ,status)
!    call  FTPCLE(unit,COLNUM_RATE         ,frow,felem,lc%num_rows*lc%num_ebands,lc%prate(:)       ,status)
    call  FTPCNE(unit,COLNUM_RATE         ,frow,felem,lc%num_rows*lc%num_ebands,lc%prate(:),NULLVAL_REAL32,status)
    call  FTPCNE(unit,COLNUM_RATE_ERR     ,frow,felem,lc%num_rows*lc%num_ebands,lc%prate_err(:),NULLVAL_REAL32,status)
    call  FTPCLE(unit,COLNUM_BACK_COUNTS  ,frow,felem,lc%num_rows*lc%num_ebands,lc%pback_counts(:),status)
    call  FTPCLE(unit,COLNUM_FRACAREA     ,frow,felem,lc%num_rows*lc%num_ebands,lc%pfracarea(:)   ,status)
    call  FTPCLE(unit,COLNUM_FRACTIME     ,frow,felem,lc%num_rows,              lc%pfractime(:)   ,status)
    call  FTPCLE(unit,COLNUM_FRACEXP      ,frow,felem,lc%num_rows*lc%num_ebands,lc%pfracexp(:)    ,status)
    call  FTPCNE(unit,COLNUM_OFFAXIS      ,frow,felem,lc%num_rows,              lc%poffaxis(:),NULLVAL_REAL32,status)
    call  FTPCND(unit,COLNUM_BACKRATIO    ,frow,felem,lc%num_rows,              lc%pbackratio(:),NULLVAL_REAL64,status)



    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    !-----------------------------------


    !-----------------------------------
    ! append the gtilist as a new hdu
    if ( gtilist%is_valid ) then
      call srctool_gti_append_gtilist_hdu(unit, gtilist, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem appending GTI extension")
        return
      end if
    end if
    !-----------------------------------


    !-----------------------------------
    ! now close the fits file
    call FTCLOS(unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------


  end subroutine srctool_lc_write_lc
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! write a set of lc files to disk
  ! we expect that all relevant parameters are already set up
  subroutine srctool_lc_write_lclist(lclist, instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_lc_write_lclist'
    type(lclist_type),  intent(inout) :: lclist
    type(instlist_type),  intent(in) :: instlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    type(gtilist_type) :: dummy_gtilist
    ierr = ERROR_LEVEL_OK
    dummy_gtilist%is_valid = .false.

    do i=0,MAX_INST
      if ( i .eq. 0 ) then
        call srctool_lc_write_lc(lclist%lc(i), dummy_gtilist, ierr )
      else if ( instlist%inst(i)%to_consider ) then
        call srctool_lc_write_lc(lclist%lc(i), instlist%inst(i)%src_gtilist, ierr )
      end if
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem writing lightcurve file")
        return
      end if
    end do

  end subroutine srctool_lc_write_lclist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! set up binning for an lc
  ! assume that input params have already been copied into the lc structure
  subroutine srctool_lc_prepare_lc_template(lc, instlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_prepare_lc_template'
    type(lc_type),  intent(inout) :: lc
    type(instlist_type), intent(in),target :: instlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)                :: status
    real(kind=real64)        :: timedel, half_timedel
    real(kind=real64)        :: tstart, tstop
    real(kind=real64)        :: t1, t2
    integer(kind=int32)      :: nbins
    integer(kind=int32)      :: i,ii
    character(len=STR_MAX) :: str_buffer
    real(kind=real64), parameter :: nullval_real64  = -999.9d0
    type(gtilist_type), pointer :: pgtilist

    if ( allocated(lc%ptime))    deallocate(lc%ptime)
    if ( allocated(lc%ptimedel)) deallocate(lc%ptimedel)
    if ( allocated(lc%ptmin))    deallocate(lc%ptmin)
    if ( allocated(lc%ptmax))    deallocate(lc%ptmax)

    ! how to proceed depends on the lctype parameter
    select case ( lc%lctype_code )
    case ( LCTYPE_CODE_REGULAR, LCTYPE_CODE_REGULAR_PLUS, LCTYPE_CODE_REGULAR_MINUS ) ! evenly spaced bins
      ! interpretation depends on the number of parameters supplied
      select case ( lc%num_lcpars )
      case ( 1 ) ! param 1 is the bin time, tstart and tstop derived from eventlist TSTART and TSTOP
        timedel = lc%lcpars(1)
        tstart  = instlist%min_tstart
        tstop   = instlist%max_tstop
      case ( 2 ) !  param 1 is the bin time, param 2 is tstart, tstop derived from eventlist TSTOP
        timedel = lc%lcpars(1)
        tstart  = lc%lcpars(2)
        tstop   = instlist%max_tstop
      case ( 3 ) !  param 1 is the bin time, param 2 is tstart, param 3 is tstop
        timedel = lc%lcpars(1)
        tstart  = lc%lcpars(2)
        tstop   = lc%lcpars(3)

      case default
        write (str_buffer, '(A,I0,A)') &
            "Invalid number of lcpars (", lc%num_lcpars, ") were supplied: "
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end select


      ! make some sanity checks on the bin params
      if (  timedel .le. 0.0d0 ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,F18.4,A,I0,A)') &
            "Invalid value of LC binning param: timedel=", timedel, &
            " (npars_supplied=",  lc%num_lcpars, ")"
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if
      if (  tstart .ge. tstop ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(2(A,F18.4) ,I0,A)') &
            "Invalid value of LC binning params: tstart=", tstart, " tstop=", tstop, &
            " (npars_supplied=",  lc%num_lcpars, ")"
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if

      call fitswrap_set_key(lc%key_timedel, timedel)

      nbins = int((tstop - tstart ) / timedel, int32) + 1
!      nbins = int(((tstop - tstart ) / timedel)+0.5d0, int32)
      if ( tstart + timedel*REAL(nbins-1,kind=real64) .ge. tstop ) then
        nbins = nbins - 1
      endif
      half_timedel = 0.5d0 * timedel

      if (  nbins .lt. 0 ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,I0,A,I0,A)') &
            "Invalid value of LC binning params: nbins=", nbins, &
            " (npars_supplied=",  lc%num_lcpars, ")"
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if

      write (str_buffer, '(A,I0,A,I0,A)') &
          "Working with following ", lc%num_lcpars, " LC parameters (npars_supplied=",  lc%num_lcpars, "):"
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,F16.4)') "timedel= ", timedel
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,F16.4)') "tstart=  ", tstart
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,F16.4)') "tstop=   ", tstop
      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,I16)')   "nbins=   ", nbins
      call wrschatty (subroutinename, str_buffer)


      lc%num_rows = nbins
      allocate(lc%ptime(nbins), &
          lc%ptimedel(nbins), &
          lc%ptmin(nbins), &
          lc%ptmax(nbins), &
          STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
!      t1 = tstart - half_timedel
!      t2 = t1     + half_timedel
      t1 = tstart
      t2 = t1     + timedel
      do i=1,lc%num_rows
        t2 = MIN(t2,tstop)
        lc%ptimedel(i) = t2 - t1
        lc%ptime(i)    = 0.5d0*(t1 + t2)
        lc%ptmin(i)    = t1 ! lc%ptime(i) - 0.5d0*lc%ptimedel(i)
        lc%ptmax(i)    = t2 ! lc%ptime(i) + 0.5d0*lc%ptimedel(i)
        t1 = t2
        t2 = t2 + timedel
      end do
!      if ( lc%ptimedel(lc%num_rows) .le. 0.0d0 ) then
!        lc%num_rows = lc%num_rows -1
!      endif

      ! TODO - deal more efficiently with gaps between GTIs for REGULAR+ binning??

      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Error when interpreting lc descriptions")
        return
      end if


    case ( LCTYPE_CODE_SRCGTI, LCTYPE_CODE_GTI , LCTYPE_CODE_SRCGTI_MINUS )
      nullify(pgtilist)
      if ( (lc%lctype_code .eq. LCTYPE_CODE_SRCGTI) .or. &
           (lc%lctype_code .eq. LCTYPE_CODE_SRCGTI_MINUS ) ) then
        pgtilist => instlist%inst(0)%src_gtilist
      else if ( lc%lctype_code .eq. LCTYPE_CODE_GTI ) then
        pgtilist => instlist%inst(0)%gtilist
      endif
      ! Use bins which match the src gtis, with a gap between each pair of gtis
      lc%num_rows = (2 * pgtilist%num_rows) -1
      allocate(lc%ptime(lc%num_rows), &
          lc%ptimedel(lc%num_rows), &
          lc%ptmin(lc%num_rows), &
          lc%ptmax(lc%num_rows), &
          STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return

      ii = 1
      do i=1,lc%num_rows
        if ( MODULO(i,2) .eq. 1 ) then
          lc%ptmin(i) = pgtilist%pstart(ii)
          lc%ptmax(i) = pgtilist%pstop(ii)
          ii = ii + 1
        else
          lc%ptmin(i) = lc%ptmax(i-1)       ! end of the last bin
          lc%ptmax(i) = pgtilist%pstart(ii) ! start of the next bin
        endif
        lc%ptime(i) = 0.5d0 * (lc%ptmin(i) + lc%ptmax(i))
        lc%ptimedel(i) = (lc%ptmax(i) - lc%ptmin(i))
      end do
      call fitswrap_set_key_null(lc%key_timedel)

    case ( LCTYPE_CODE_FOLDED )
      write (str_buffer, '(A,A,A,I0,A)') "Unimplemented LC type: ", trim(lc%lctype), " (status=", status, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return

    case default
      write (str_buffer, '(A,A,A,I0,A)') "Unknown LC type: ", trim(lc%lctype), " (status=", status, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end select      ! end select on lctype_code
    !--------------------------------

  end subroutine srctool_lc_prepare_lc_template

  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for an lc file
  subroutine srctool_lc_setup_lc(lc, lc_template, inst, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_setup_lc'
    type(lc_type),      intent(inout) :: lc
    type(lc_type),      intent(in)    :: lc_template
    type(inst_type),     intent(in),target :: inst
    type(src_type),      intent(in)    :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)                :: status
!    character(len=STR_MAX)             :: str_buffer
!    character(len=STR_MAX)             :: str_buffer2
    integer(kind=int32)      :: i,j,e,flag
    real(kind=real64)        :: emid, weight_sum
    type(arf_type),    pointer :: prefarf
    ierr = ERROR_LEVEL_OK
    status = 0
    lc%to_consider = .true.

    ! base the lc structure on the lc template
    lc%lctype_code = lc_template%lctype_code
    lc%num_rows   = lc_template%num_rows
    lc%num_ebands = lc_template%num_ebands
!    lc%int_nenergies = lc_template%int_nenergies

    if ( .not. allocated(lc%ptime) ) then
      allocate(lc%ptime(lc%num_rows),&
          lc%ptimedel      (lc%num_rows),&
          lc%ptmin         (lc%num_rows),&
          lc%ptmax         (lc%num_rows),&
          lc%pcounts       (lc%num_rows*lc%num_ebands),&
          lc%prate         (lc%num_rows*lc%num_ebands),&
          lc%prate_err     (lc%num_rows*lc%num_ebands),&
          lc%pback_counts  (lc%num_rows*lc%num_ebands),&
          lc%pfracarea     (lc%num_rows*lc%num_ebands),&
          lc%pfractime     (lc%num_rows), &
          lc%pfracexp      (lc%num_rows*lc%num_ebands),&
          lc%poffaxis      (lc%num_rows), &
          lc%pbackratio    (lc%num_rows), &
          STAT = status)

      if ( check_allocate_status(subroutinename, status, ierr) ) return

    end if

    ! initialise the values in the arrays
    lc%ptime(:)          = 0.0d0
    lc%ptimedel(:)       = 0.0d0
    lc%ptmin(:)          = 0.0d0
    lc%ptmax(:)          = 0.0d0
    lc%pcounts(:)        = 0
    lc%prate(:)          = 0.0e0
    lc%prate_err(:)      = 0.0e0
    lc%pback_counts(:)   = 0
    lc%pfracarea(:)      = 0.0e0
    lc%pfractime(:)      = 0.0e0
    lc%pfracexp(:)       = 0.0e0
    lc%poffaxis(:)       = 0.0e0
    lc%pbackratio(:)     = 0.0d0
    ! copy the time bins over from the lc template
    lc%ptime(:)       = lc_template%ptime(:)
    lc%ptimedel(:)    = lc_template%ptimedel(:)
    lc%lcgamma        = lc_template%lcgamma
    lc%ptmin(:)       = lc_template%ptmin(:)
    lc%ptmax(:)       = lc_template%ptmax(:)
!    ! and the integration binning stuff
!    lc%pint_e      = lc_template%pint_e
!    lc%pint_weights = lc_template%pint_weights
!    lc%pint_index1 = lc_template%pint_index1
!    lc%pint_index2 = lc_template%pint_index2

    ! and copy over the relevant keywords from the lc template
    lc%key_timedel    = lc_template%key_timedel
    lc%key_timezero   = lc_template%key_timezero

    lc%key_telapse    = lc_template%key_telapse
    lc%key_ontime     = lc_template%key_ontime
    lc%key_livetime   = lc_template%key_livetime
    lc%key_exposure   = lc_template%key_exposure

    lc%key_tierrela   = lc_template%key_tierrela
    lc%key_tierabso   = lc_template%key_tierabso
    lc%key_clockcor   = lc_template%key_clockcor
    lc%key_backapp    = lc_template%key_backapp
    lc%key_deadapp    = lc_template%key_deadapp
    lc%key_vignapp    = lc_template%key_vignapp
    lc%key_emin       = lc_template%key_emin
    lc%key_emax       = lc_template%key_emax

    if ( inst%index .eq. 0 ) then
!
!      !TODO
!      ! we put some placeholder values here for now, but should eventually derive these from TM1-7 values
!      call fitswrap_set_key(lc%hduinfo%key_telescop, "TELESCOP" ,  "eROSITA", "Telescope name")
!      call fitswrap_set_key(lc%hduinfo%key_instrume, "INSTRUME" ,      "TM0", "Instrument name")
!      call fitswrap_set_key(lc%hduinfo%key_filter,   "FILTER"   ,  "UNKNOWN", "CCD filter used")
!      call fitswrap_set_key(lc%hduinfo%key_tstart,   "TSTART"   ,      0.0d0, "[s]")
!      call fitswrap_set_key(lc%hduinfo%key_tstop,    "TSTOP"    ,      1.0d0, "[s]")
!      call fitswrap_set_key(lc%hduinfo%key_date_obs, "DATE-OBS" , "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")
!      call fitswrap_set_key(lc%hduinfo%key_date_end, "DATE-END" , "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")
    else
      ! copy the hduinfo from the first evtlist in the inst structure
      ! but first check that there is at least one valid evtlist
      if ( inst%evtlistlist%num_evtlists .le. 0  ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Instrument structure contains no evtlists")
        return
      end if
      lc%hduinfo = inst%evtlistlist%hduinfo
    end if

    ! set the output filename and DSTYPE
    call srctool_misc_set_filename(lc%hduinfo%filename, inst, src, SRCTOOL_FILETYPE_CODE_LC )
    call fitswrap_set_key(lc%hduinfo%key_dstype, ESASS_DSTYPE_LC )

    call fitswrap_set_key(lc%hduinfo%key_dscode, &
        srctool_misc_dscode(inst%index,ESASS_FILTER_STR(inst%filter_code),0) )



    ! now set some more appropriate values for an OGIP compliant lc file
    call fitswrap_set_key(lc%hduinfo%key_extname,   "RATE"      )
    call fitswrap_set_key(lc%hduinfo%key_hduclass,  "OGIP"      )
    call fitswrap_set_key(lc%hduinfo%key_hduclas1,  "LIGHTCURVE")
    call fitswrap_set_key(lc%hduinfo%key_hduclas2,  "TOTAL"     )
    call fitswrap_set_key(lc%hduinfo%key_hduclas3,  "COUNT"     )

    call fitswrap_set_key(lc%hduinfo%key_hduvers,  "1994Nov30" )
!    call fitswrap_set_key(lc%hduinfo%key_creator,  "")

!   TODO: Check if these keywords should just be populated from the input evtlist
    call fitswrap_set_key(lc%hduinfo%key_object  , trim(src%name))
    call fitswrap_set_key(lc%hduinfo%key_ra_obj  , src%coord%long)
    call fitswrap_set_key(lc%hduinfo%key_dec_obj , src%coord%lat)
    call fitswrap_set_key(lc%hduinfo%key_radecsys, COORDSYS_STR(src%coord%coordsys_code))
    call fitswrap_set_key(lc%hduinfo%key_equinox , 2000.0d0)

    ! ownership
    call fitswrap_set_key(lc%hduinfo%key_owner, srctool_owner_getkeyword_owner(src%owner))

!    ! TODO - put the correct values in the following
!    call fitswrap_set_key(lc%hduinfo%key_date_end, "DATE-END" , "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")

    call fitswrap_set_key(lc%key_clockcor, .false.)
    call fitswrap_set_key(lc%key_backapp,  .false.)
    call fitswrap_set_key(lc%key_deadapp,  .false.)
    call fitswrap_set_key(lc%key_vignapp,  .false.)
    call fitswrap_set_key(lc%key_timezero,  0.0d0)

    call fitswrap_set_key(lc%key_telapse,  0.0d0) ! these get set later
    call fitswrap_set_key(lc%key_ontime,   0.0d0)
    call fitswrap_set_key(lc%key_livetime, 0.0d0)
    call fitswrap_set_key(lc%key_exposure, 0.0d0)

    call fitswrap_set_key(lc%key_tierrela, 0.0d0)
    call fitswrap_set_key(lc%key_tierabso, 0.0d0)

    call fitswrap_set_key(lc%key_src_region,  trim(adjustl(src%str_src_region)))
    call fitswrap_set_key(lc%key_back_region, trim(adjustl(src%str_back_region)))

    ! set the FILTER keyword
    call fitswrap_set_key(lc%hduinfo%key_filter, ESASS_FILTER_STR(inst%filter_code) )

    !--------------------------------
    ! now set up the lc weighting and energies arrays
    ! Base this on the refarf and the lcgamma param
    ! weight(E) = E^-gamma * ARF(E) * dE / (Emax - Emin)
    ! where dE is the bin width
    call srctool_misc_select_refarf_pointer(inst,prefarf,ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Could not find suitable reference ARF")
      return
    end if

    ! TODO TODO
    ! count the number of arf energy bins that fall within at least one of the lc ebands
    lc%int_nenergies = 0
    do i=1,prefarf%num_rows
      flag = 0
      do e=1,lc%num_ebands
        emid = 0.5d0*(prefarf%pelo(i) + prefarf%pehi(i))
        if (emid .ge. lc%key_emin(e)%value .and. &
            emid .lt. lc%key_emax(e)%value   ) then
          flag = 1
          exit
        end if
      end do
      if ( flag .eq. 1 ) lc%int_nenergies = lc%int_nenergies + 1
    end do

    if ( .not. allocated(lc%pint_e) ) then
      allocate(lc%pint_e(lc%int_nenergies), &
          lc%pint_weights(lc%num_ebands,lc%int_nenergies), &
          STAT = status)

      if ( check_allocate_status(subroutinename, status, ierr) ) return
    end if

    ! now set the energies and weights
    j = 1
    lc%pint_weights(:,:) = 0.0d0
    do i=1,prefarf%num_rows
      flag = 0
      do e=1,lc%num_ebands
        emid = 0.5d0*(prefarf%pelo(i) + prefarf%pehi(i))
        if (emid .ge. lc%key_emin(e)%value .and. &
            emid .lt. lc%key_emax(e)%value   ) then
          flag = 1
          lc%pint_e(j) = emid
          lc%pint_weights(e,j) = prefarf%presp(i)* &
              (emid**(-1.0d0 * lc%lcgamma)) * &
              (prefarf%pehi(i) - prefarf%pelo(i))
        end if
      end do
      if ( flag .eq. 1 ) j = j + 1
    end do

    ! now normalise the weights to unity in each lc band
    do e=1,lc%num_ebands
      weight_sum = sum(lc%pint_weights(e,:))
      if ( weight_sum .gt. 0.0d0 ) then
        lc%pint_weights(e,:) = lc%pint_weights(e,:) / weight_sum
      end if
    end do

    do e=1,lc%num_ebands
!      write (str_buffer,'(A,I0)'), "CTS_SC", e
!      write (str_buffer2,'(A,I0)'), "Cts in source reg,band ", e
!      call fitswrap_set_key(lc%key_ncounts_src(e), trim(str_buffer),   0, trim(str_buffer2))
!      write (str_buffer,'(A,I0)'), "CTS_BK", e
!      write (str_buffer2,'(A,I0)'), "Cts in bkgnd reg,band ", e
!      call fitswrap_set_key(lc%key_ncounts_back(e), trim(str_buffer),  0, trim(str_buffer2))
      call fitswrap_set_key(lc%key_ncounts_src(e),  0)
      call fitswrap_set_key(lc%key_ncounts_back(e), 0)
    end do


!    if ( inst%index .eq. 1 ) then
!      write (*, '(A,1x,A6,2(1x,A6,1x,A8,1x,A12))'), "#DEBUG_LCINT", &
!          "INST", "lc_e", "lc_emin", "lc_emax", &
!          "j", "e(j)", "weight(e,j)"
!      do e=1,lc%num_ebands
!        do j=1,lc%int_nenergies
!          write (*, '(A,1x,A6,2(1x,I6,1x,F8.3,1x,F12.8))'), "DEBUG_LCINT ", &
!              trim(inst%name), e, lc%key_emin(e)%value, lc%key_emax(e)%value, &
!              j, lc%pint_e(j), lc%pint_weights(e,j)
!        end do
!      end do
!    end if

  end subroutine srctool_lc_setup_lc
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for a lclist
  subroutine srctool_lc_setup_lclist(lclist, lc_template, instlist, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_setup_lclist'
    type(lclist_type),    intent(inout) :: lclist
    type(lc_type),        intent(inout) :: lc_template
    type(instlist_type),  intent(in)    :: instlist
    type(src_type),       intent(in)    :: src
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32)                 :: i
    integer(kind=int32)                 :: nok
!    character(len=STR_MAX)             :: str_buffer


    ierr = ERROR_LEVEL_OK
    call srctool_lc_prepare_lc_template(lc_template, instlist, ierr )
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error preparing lc template structure")
      return
    end if

    do i=0,MAX_INST
      if ( (i .eq. 0) .or. instlist%inst(i)%to_consider ) then
        call srctool_lc_setup_lc(lclist%lc(i), lc_template, instlist%inst(i), src, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error setting up src lc data structure")
          return
        end if
      end if
    end do

!    ! now set up the hduinfo values in the merged instrument lc
!    ! derive them from the first valid lc - with INSTRUME and filename adapted accordingly
!    do i=1,MAX_INST
!      if ( (instlist%inst(i)%to_consider ) ) then
!        lclist%lc(0)%hduinfo = lclist%lc(i)%hduinfo
!        call fitswrap_set_key(lclist%lc(0)%hduinfo%key_instrume, "INSTRUME",trim(instlist%inst(0)%name),"Instrument name")
!        ! sort out the output filename
!        call srctool_misc_set_filename(lclist%lc(0)%hduinfo%filename, &
!            instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_LC )
!        exit
!      end if
!    end do

    ! copy the basic keywords from the first valid instrument for the merged instrument files
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        lclist%lc(0)%hduinfo  = lclist%lc(i)%hduinfo
        exit
      end if
    end do

    ! now set up the hduinfo values in the "MERGED" lc files
    ! derive them from the first valid lc file - with INSTRUME and filename adapted accordingly
    call fitswrap_set_key(lclist%lc(0)%hduinfo%key_instrume, INSTRUME_IS_MERGED)
    call srctool_misc_set_filename(lclist%lc(0)%hduinfo%filename, &
          instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_LC )
    ! now deal with the NINST and INSTRUMn keywords
    nok = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        nok = nok + 1
        !write (str_buffer, '(A,I0)') "INSTRUM", nok
        call fitswrap_set_key(lclist%lc(0)%hduinfo%pkey_instrum(nok), trim(instlist%inst(i)%name))
      end if
    end do
    call fitswrap_set_key(lclist%lc(0)%hduinfo%key_ninst, nok)

  end subroutine srctool_lc_setup_lclist
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! gather lc info from an evtlist and add to a lc file
  ! assume that the lc file has been setup
  subroutine srctool_lc_accumulate_lc(lc, evtlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_accumulate_lc'
    type(lc_type),      intent(inout) :: lc
    type(evtlist_type),  intent(in)    :: evtlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    integer(kind=int32) :: tbin, ebin, ei
    integer(kind=int32) :: status
    real(kind=real64)   :: t
    real(kind=real32)   :: e_keV
    real(kind=real64)   :: t1,t2
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32), dimension(1:MAX_LCBANDS)  :: ncounts_src, ncounts_back  !counters

    ierr = ERROR_LEVEL_OK
    status = 0

    ncounts_src(:) = 0
    ncounts_back(:) = 0

    if ( evtlist%num_rows .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Error accumulating lc (no events)")
      return
    endif

    if ( lc%num_rows .gt. 0 ) then

  !    t1 = lc%ptime(1) - 0.5d0*lc%ptimedel(1)
  !    t2 = lc%ptime(lc%num_rows) + 0.5d0*lc%ptimedel(lc%num_rows)
      t1 = lc%ptmin(1)
      t2 = lc%ptmax(lc%num_rows)
      do i=1,evtlist%num_rows
        if ( evtlist%pclass_mask(i) .ne. EVT_CLASS_FLAG_NULL ) then
          t = evtlist%ptime(i)
          if ( (t .ge. t1) .and.  (t .lt. t2) ) then
          ! determine the lc bin (tbin) in which this event falls
            if ( lc%num_rows .gt. 1 ) then
              call binary_search(lc%ptmin(:), t, tbin, status)
              if ((status .ne. 0) .and. (status .ne. 2) ) then
                tbin = 0  ! make sure that this event is not recorded
  !              ierr = ERROR_LEVEL_SEVERE
  !              call wrserr (subroutinename, ierr, "Could not find lc bin")
  !              return
              end if
            else
              tbin = 1
            endif
            if ( tbin .gt. 0 ) then
              ! now refine the bin:
              tbin=MAX(1,tbin-1)
              do tbin=tbin,lc%num_rows
                !            if ((t .ge. (lc%ptime(tbin) - 0.5*lc%ptimedel(tbin))) .and. &
                !                (t .lt. (lc%ptime(tbin) + 0.5*lc%ptimedel(tbin))) ) then
                if ((t .ge. lc%ptmin(tbin)) .and. &
                    (t .lt. lc%ptmax(tbin))) then
                  exit ! get out of do loop
                end if
              end do

              ! work out which energy bands the event falls within
              e_keV = REAL(EV2KEV * evtlist%penergy(i), real32)
              do ebin=1,lc%num_ebands
                if ((e_keV .ge. lc%key_emin(ebin)%value) .and. &
                    (e_keV .lt. lc%key_emax(ebin)%value)) then
                  ei = (tbin-1)*lc%num_ebands + ebin
                  if ( IAND(evtlist%pclass_mask(i),EVT_CLASS_FLAG_SRC) .ne. 0 ) then
                    lc%pcounts(ei) = lc%pcounts(ei) + 1
                    ncounts_src(ebin) = ncounts_src(ebin) + 1
                  end if
                  if ( IAND(evtlist%pclass_mask(i),EVT_CLASS_FLAG_BACK) .ne. 0 ) then
                    lc%pback_counts(ei) = lc%pback_counts(ei) + 1
                    ncounts_back(ebin) = ncounts_back(ebin) + 1
                  end if
                end if
              end do
           endif
          else
            ! silently ignore any out-of-bounds events
          end if

       end if
      end do
    endif

!     ! debug
! !    write (str_buffer, '(A,A6,3(1x,I8))'), &
! !       "DEBUG LC_ACCUM: ",  trim(evtlist%hduinfo%key_instrume%value), &
! !        evtlist%num_rows, ncounts_src, ncounts_back
!     write (str_buffer, '(A,2(1x,A8),2(1x,I8))'), &
!         "DEBUG LC_ACCUM: ",  "SRC", trim(evtlist%hduinfo%key_instrume%value), &
!         evtlist%num_rows, ncounts_src
!     call wrschatty (subroutinename, str_buffer)
!     write (str_buffer, '(A,2(1x,A8),2(1x,I8))'), &
!         "DEBUG LC_ACCUM: ",  "BACK", trim(evtlist%hduinfo%key_instrume%value), &
!         evtlist%num_rows, ncounts_back
!     call wrschatty (subroutinename, str_buffer)

    ! copy the flagsel and pat_sel keywords from the evtlist
    lc%key_flagsel = evtlist%key_flagsel
    lc%key_pat_sel = evtlist%key_pat_sel

    do ebin=1,lc%num_ebands
!      write (str_buffer, '(A,2(1x,A10),2(1x,F10.3),1(1x,I10))'), &
!          "DEBUG LC_ACCUM: ",  "SRC", trim(evtlist%hduinfo%key_instrume%value), &
!          lc%key_emin(ebin)%value, lc%key_emax(ebin)%value, &
!          ncounts_src(ebin)
!      call wrschatty (subroutinename, str_buffer)
!      write (str_buffer, '(A,2(1x,A10),2(1x,F10.3),1(1x,I10))'), &
!          "DEBUG LC_ACCUM: ",  "BACK", trim(evtlist%hduinfo%key_instrume%value), &
!          lc%key_emin(ebin)%value, lc%key_emax(ebin)%value, &
!          ncounts_back(ebin)
!      call wrschatty (subroutinename, str_buffer)
      write (str_buffer, '(A,1(1x,A10),2(1x,F10.3),2(1x,I10))') &
          "DEBUG LC_ACCUM: ", trim(evtlist%hduinfo%key_instrume%value), &
          lc%key_emin(ebin)%value, lc%key_emax(ebin)%value, &
          ncounts_src(ebin), ncounts_back(ebin)
      call wrschatty (subroutinename, str_buffer)

      call fitswrap_set_key(lc%key_ncounts_src(ebin),  ncounts_src(ebin))
      call fitswrap_set_key(lc%key_ncounts_back(ebin), ncounts_back(ebin))

    end do

  end subroutine srctool_lc_accumulate_lc
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! gather lc info from an evtlist and add to a lc file
  ! assume that the lc file has been setup
  subroutine srctool_lc_accumulate_lclist(lclist, instlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_lc_accumulate_lc'
    type(lclist_type),    intent(inout) :: lclist
    type(instlist_type),  intent(in)    :: instlist
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32) :: i,j
    character(len=STR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    write (str_buffer, '(A,5(1x,A10))') &
        "DEBUG LC_ACCUM: ", &
        "INSTR", "Emin(keV)", "Emax(keV)", "Src_cnts", "Back_cnts"
    call wrschatty (subroutinename, str_buffer)

    !-----------------------------------------------
    do i=1,MAX_INST
      do j=1,instlist%inst(i)%evtlistlist%num_evtlists
        call srctool_lc_accumulate_lc(lclist%lc(i), instlist%inst(i)%evtlistlist%evtlist(j), ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error accumulating lc")
          return
        end if
      end do
    end do
    !-----------------------------------------------

!    ! now add up the counts to make the the merged instrument lc
!    ! sum the lcs over the seven telescopes
!    call srctool_coadd_lclist(lclist, instlist )

  end subroutine srctool_lc_accumulate_lclist
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! coadd the lc data across all 7 telescope modules
  subroutine srctool_lc_coadd_lclist(lclist, instlist )
    character(len=*), parameter :: subroutinename='srctool_lc_coadd_lclist'
    type(lclist_type),    intent(inout) :: lclist
    type(instlist_type),  intent(in) :: instlist
    integer(kind=int32) :: i, e
    real(kind=real32) :: nok
    real(kind=real64) :: tstart, tstop
    integer(kind=int32) :: first_valid_inst
    first_valid_inst = -1


    nok = 0.0e0
    lclist%lc(0)%pcounts(:)       = 0
    lclist%lc(0)%pback_counts(:)  = 0
    lclist%lc(0)%pfracarea(:)     = 0.0e0
    lclist%lc(0)%pfractime(:)     = 0.0e0
    lclist%lc(0)%pfracexp(:)      = 0.0e0
    lclist%lc(0)%poffaxis(:)      = 0.0d0
    lclist%lc(0)%pbackratio(:)    = 0.0d0
!    lclist%lc(0)%prate(:)         = 0.0d0

    tstart = huge(0.0)
    tstop = -huge(0.0)

    lclist%lc(0)%key_ncounts_src(:)%value  = 0
    lclist%lc(0)%key_ncounts_back(:)%value  = 0

    ! TODO - proper weighting of the fractional exposure columns etc
    !-----------------------------------------------
    do i=1,MAX_INST
      if (instlist%inst(i)%to_consider ) then
        if ( lclist%lc(i)%to_consider ) then
          nok = nok + 1.0e0
          if ( first_valid_inst .lt. 0 ) first_valid_inst = i
          lclist%lc(0)%pcounts(:)       = lclist%lc(0)%pcounts(:)      + lclist%lc(i)%pcounts(:)
          lclist%lc(0)%pback_counts(:)  = lclist%lc(0)%pback_counts(:) + lclist%lc(i)%pback_counts(:)
          lclist%lc(0)%pfracarea(:)     = lclist%lc(0)%pfracarea(:)    + lclist%lc(i)%pfracarea(:)
          lclist%lc(0)%pfractime(:)     = lclist%lc(0)%pfractime(:)    + lclist%lc(i)%pfractime(:)
          lclist%lc(0)%pfracexp(:)      = lclist%lc(0)%pfracexp(:)     + lclist%lc(i)%pfracexp(:)
          lclist%lc(0)%poffaxis(:)      = lclist%lc(0)%poffaxis(:)     + lclist%lc(i)%poffaxis(:)
          lclist%lc(0)%pbackratio(:)    = lclist%lc(0)%pbackratio(:)   + lclist%lc(i)%pbackratio(:)
!          lclist%lc(0)%prate(:)         = lclist%lc(0)%prate(:)        + lclist%lc(i)%prate(:)
          do e=1,lclist%lc(0)%num_ebands
            lclist%lc(0)%key_ncounts_src(e)%value  = &
                lclist%lc(0)%key_ncounts_src(e)%value + lclist%lc(i)%key_ncounts_src(e)%value
            lclist%lc(0)%key_ncounts_back(e)%value  = &
                lclist%lc(0)%key_ncounts_back(e)%value + lclist%lc(i)%key_ncounts_back(e)%value
          end do

          ! combine tstart/tstop with individual gti
          associate(gtilist => instlist%inst(i)%src_gtilist)
            if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
               tstart = min(gtilist%pstart(1), tstart)
               tstop = max(gtilist%pstop(gtilist%num_rows), tstop)
            end if
          end associate

        end if
      end if
    end do
    !-----------------------------------------------

    ! now take an average of the relevant columns
    if ( nok .gt. 0.0e0 ) then
      lclist%lc(0)%pfracarea(:)   = lclist%lc(0)%pfracarea(:)    / nok
      lclist%lc(0)%pfractime(:)   = lclist%lc(0)%pfractime(:)    / nok
      lclist%lc(0)%pfracexp(:)    = lclist%lc(0)%pfracexp(:)     / nok
      lclist%lc(0)%poffaxis(:)    = lclist%lc(0)%poffaxis(:)     / nok
      lclist%lc(0)%pbackratio(:)  = lclist%lc(0)%pbackratio(:)   / nok
!      lclist%lc(0)%prate(:)       = lclist%lc(0)%prate(:)        / nok   ! TODO - weighted averages?

!      lclist%lc(0)%poffaxis(:) = MIN(lclist%lc(0)%poffaxis(:), NULLVAL_REAL32)    ! Do this better
    end if

    do i=1,size(lclist%lc(0)%poffaxis)
      if ( lclist%lc(0)%poffaxis(i) .le. NULLVAL_REAL32 ) then
        lclist%lc(0)%poffaxis(i) = NULLVAL_REAL32
      endif
    end do

    if ( first_valid_inst .gt. 0 ) then
      ! copy the flagsel and pat_sel keywords from the first valid evtlist
      lclist%lc(0)%key_flagsel  = lclist%lc(first_valid_inst)%key_flagsel
      lclist%lc(0)%key_pat_sel  = lclist%lc(first_valid_inst)%key_pat_sel
    end if

    ! set the filter keyword
    call fitswrap_set_key(lclist%lc(0)%hduinfo%key_filter, ESASS_FILTER_STR(instlist%inst(0)%filter_code))

    ! update tstart/tstop
    if(tstart /= huge(0.0)) call fitswrap_set_key(lclist%lc(0)%hduinfo%key_tstart, tstart)
    if(tstop /= -huge(0.0)) call fitswrap_set_key(lclist%lc(0)%hduinfo%key_tstop, tstop)

    !---------------------------------------------------------------



  end subroutine srctool_lc_coadd_lclist
  !-------------------------------------------------------------------------------


  recursive subroutine srctool_lc_get_lc_index (lc, t, i)
    character(len=*), parameter :: subroutinename='srctool_lc_get_lc_index'
    type(lc_type),    intent(in)      :: lc
    real(kind=real64), intent(in)  :: t
    integer(kind=int32), intent(out) :: i
    integer(kind=int32) :: status
    status = 0
    i = 0

    if ( lc%to_consider ) then

      ! check if t is outside bounds of lc
      if ((t .lt. lc%ptmin(1)) .or. &
          (t .ge. lc%ptmax(lc%num_rows))) then
        i = 0
      else if ( lc%num_rows .eq. 1 ) then
        i = 1
      else
        ! do binary search to find index of lc bin spanning time=t
        call binary_search(lc%ptmin(:),t,i,status)
        if ((status .ne. BINARY_SEARCH_STATUS_OK) .and. &
            (status .ne. BINARY_SEARCH_STATUS_GTMAX) ) then
          ! this should never happen
          i = -1
          call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
              "Problem moving to next lcbin (code:" // trim(i2s(status)) // ")" )
          return
        endif
      endif
    endif

  end subroutine srctool_lc_get_lc_index


  recursive subroutine srctool_lc_get_next_lc_index (lc, t, i)
    character(len=*), parameter :: subroutinename='srctool_lc_get_next_lc_index'
    type(lc_type),    intent(in)      :: lc
    real(kind=real64), intent(in)  :: t
    integer(kind=int32), intent(out) :: i
    integer(kind=int32) :: status
    status = 0
    i = -1
    if ( lc%to_consider ) then

      if ( lc%num_rows .gt. 1 ) then
        call binary_search(lc%ptmax(1:lc%num_rows), t, i, status)
        if ((status .eq. BINARY_SEARCH_STATUS_OK ) .or. &
            (status .eq. BINARY_SEARCH_STATUS_LTMIN )) then
          if (t .gt. lc%ptmin(i+1))  then
            i = -1
          endif
        else
          i = -1
        endif
      else if ( t .lt. lc%ptmax(1) ) then  ! t is before fist and only time bin
        i = 1
      end if
    endif

  end subroutine srctool_lc_get_next_lc_index

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- END of Lightcurve handling ------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------



end module srctool_lc
