! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_misc

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-24
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of misc routines
  !
  !-----------------------------------------------------------------------------

  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
!  use srctool_generic_fitswrap  ! generic fits file handling routines
!  use srctool_generic_regions   ! generic region handling routines
  use srctool_datatypes         ! module with some srctool-specific constants and datatypes
  use srctool_generic_avign     ! generic analytic vignetting file handling routines

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_MISC_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_misc
  public ::  srctool_misc_set_filename
  public ::  srctool_misc_select_ebounds_pointer
  public ::  srctool_misc_select_psf_pointer
  public ::  srctool_misc_select_detmap_pointer
  public ::  srctool_misc_select_badpix_pointer
  public ::  srctool_misc_select_instpar_pointer
  public ::  srctool_misc_select_refrmfset_pointer
  public ::  srctool_misc_select_gtilist_pointer
!  public ::  srctool_misc_select_vign_pointer
  public ::  srctool_misc_select_avign_pointer
  public ::  srctool_misc_select_refarf_pointer
  public ::  srctool_misc_select_deadcor_pointer

  public :: srctool_misc_filter_string_to_code
  public :: srctool_misc_dscode

  interface srctool_misc_select_cal_pointer
    module procedure srctool_misc_select_ebounds_pointer
    module procedure srctool_misc_select_psf_pointer
    module procedure srctool_misc_select_detmap_pointer
    module procedure srctool_misc_select_badpix_pointer
    module procedure srctool_misc_select_instpar_pointer
    module procedure srctool_misc_select_refrmfset_pointer
    module procedure srctool_misc_select_gtilist_pointer
!    module procedure srctool_misc_select_vign_pointer
    module procedure srctool_misc_select_avign_pointer
    module procedure srctool_misc_select_refarf_pointer
    module procedure srctool_misc_select_deadcor_pointer
  end interface srctool_misc_select_cal_pointer

contains

  pure recursive function srctool_misc_filter_string_to_code (str_filter) result(code)
    character(len=*),    intent(in)   :: str_filter
    integer(kind=int32) :: code

    ! http://erosita.mpe.mpg.de/eROdoc/pipeline/SASS-products_file_naming_04.html
    if ( trim(str_filter) .eq. "CLOSED" ) then
      code = ESASS_FILTER_CODE_CLOSED
    else if ( trim(str_filter) .eq. "OPEN" ) then
      code = ESASS_FILTER_CODE_OPEN
    else if ( trim(str_filter) .eq. "FILTER" ) then
      code = ESASS_FILTER_CODE_FILTER
    else if ( trim(str_filter) .eq. "Fe55" ) then
      code = ESASS_FILTER_CODE_FE55
    else
      code = ESASS_FILTER_CODE_INDEF
    endif
  end function srctool_misc_filter_string_to_code


  recursive function srctool_misc_dscode (inst_index, str_filter, eband_index ) result(dscode)
    integer(kind=int32), intent(in)   :: inst_index
    character(len=*),    intent(in)   :: str_filter
    integer(kind=int32), intent(in)   :: eband_index
    character(len=3) :: dscode
    character(len=1) :: inst_code
    character(len=1) :: eband_code

    if ( inst_index .ge. 0 .and. inst_index .le. MAX_INST ) then
      write (inst_code, '(I1)') inst_index
    else
      inst_code = '-'
    endif

    if ( eband_index .ge. 0 .and. eband_index .le. 9 ) then
      write (eband_code, '(I1)') eband_index
    else
      eband_code = '-'
    endif

    dscode = inst_code // ESASS_FILTER_CHAR(srctool_misc_filter_string_to_code(str_filter)) // eband_code
  end function srctool_misc_dscode



  !-------------------------------------------------------------------------------
  ! see https://wiki.mpe.mpg.de/eROSITA/EroCat/GenDataAnal?action=AttachFile&do=view&target=SASS-products_file_naming_03.html
  recursive subroutine srctool_misc_set_filename(filename, inst, src, filetype_code )
    character(len=*), parameter :: subroutinename='srctool_misc_set_filename'
    character(len=*),    intent(out)   :: filename
    type(inst_type),     intent(in)    :: inst
    type(src_type),      intent(in)    :: src
    integer(kind=int32), intent(in)    :: filetype_code
!    character(len=*), parameter :: str_filter_code = '0'   ! TODO - adjust
!    character(len=*), parameter :: str_band_code   = '0'
    character(len=STR_MAX) :: str_filetype
    character(len=3) :: dscode

    ! sort out the output filename
    !   - compound of prefix, instrument and source name
!    select case ( filetype_code )
!    case ( SRCTOOL_FILETYPE_CODE_SRC_SPEC )
!!      write (filename, '(A,A,A,A,A)'), &
!!          trim(prefix),trim(src%name),"_src_",trim(inst%name),".pha"
!!    write (filename, '(A,A,A,A,A)'), &
!!          trim(prefix),trim(src%name),"_",trim(inst%name),"_spec_src.fits"
!    case ( SRCTOOL_FILETYPE_CODE_BACK_SPEC )
!!      write (filename, '(A,A,A,A,A)'), &
!!          trim(prefix),trim(src%name),"_back_",trim(inst%name),".pha"
!      write (filename, '(A,A,A,A,A)'), &
!          trim(prefix),trim(src%name),"_",trim(inst%name),"_spec_back.fits"
!    case ( SRCTOOL_FILETYPE_CODE_ARF )
!      write (filename, '(A,A,A,A,A)'), &
!          trim(prefix),trim(src%name),"_",trim(inst%name),"_arf.fits"
!    case ( SRCTOOL_FILETYPE_CODE_LC )
!      write (filename, '(A,A,A,A,A)'), &
!          trim(prefix),trim(src%name),"_",trim(inst%name),"_lc.fits"
!    case ( SRCTOOL_FILETYPE_CODE_EVT )
!      write (filename, '(A,A,A,A,A)'), &
!          trim(prefix),trim(src%name),"_",trim(inst%name),"_evt.fits"
!
!    case default
!      write (filename, '(A,I0)'), "Unknown_filetype_code=", filetype_code
!    end select



!    write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_SourceSpec_', trim(src%name),trim(src%str_suffix),'.fits'
!    case ( SRCTOOL_FILETYPE_CODE_BACK_SPEC )
!      write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_BackgrSpec_', trim(src%name),trim(src%str_suffix),'.fits'
!    case ( SRCTOOL_FILETYPE_CODE_ARF )
!      write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_ARF_',        trim(src%name),trim(src%str_suffix),'.fits'
!    case ( SRCTOOL_FILETYPE_CODE_RMF )
!      write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_RMF_',        trim(src%name),trim(src%str_suffix),'.fits'
!    case ( SRCTOOL_FILETYPE_CODE_LC )
!      write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_LightCurve_', trim(src%name),trim(src%str_suffix),'.fits'
!    case ( SRCTOOL_FILETYPE_CODE_EVT )
!      write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_EventList_',  trim(src%name),trim(src%str_suffix),'.fits'
!    case ( SRCTOOL_FILETYPE_CODE_SRC_SAMPLELIST )
!      write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_SourceSampleList_', trim(src%name),trim(src%str_suffix),'.fits'
!    case ( SRCTOOL_FILETYPE_CODE_BACK_SAMPLELIST )
!      write (filename, '(A,I0,A,A,A,A,A)'), &
!          trim(src%str_prefix),inst%index, filter_code,'0_BackgrSampleList_', trim(src%name),trim(src%str_suffix),'.fits'
!
!    case default
!      write (filename, '(A,I0)'), "Unknown_filetype_code=", filetype_code
!    end select

    select case ( filetype_code )
    case ( SRCTOOL_FILETYPE_CODE_SRC_SPEC )
      str_filetype = ESASS_DSTYPE_SRC_SPEC
    case ( SRCTOOL_FILETYPE_CODE_BACK_SPEC )
      str_filetype = ESASS_DSTYPE_BACK_SPEC
    case ( SRCTOOL_FILETYPE_CODE_ARF )
      str_filetype = ESASS_DSTYPE_ARF
    case ( SRCTOOL_FILETYPE_CODE_RMF )
      str_filetype = ESASS_DSTYPE_RMF
    case ( SRCTOOL_FILETYPE_CODE_LC )
      str_filetype = ESASS_DSTYPE_LC
    case ( SRCTOOL_FILETYPE_CODE_SRC_EVT )
      str_filetype = ESASS_DSTYPE_SRC_EVT
    case ( SRCTOOL_FILETYPE_CODE_BACK_EVT )
      str_filetype = ESASS_DSTYPE_BACK_EVT
    case ( SRCTOOL_FILETYPE_CODE_SRC_SAMPLELIST )
      str_filetype = 'SourceSampleList'        ! debug filetype
    case ( SRCTOOL_FILETYPE_CODE_BACK_SAMPLELIST )
      str_filetype = 'BackgrSampleList'        ! debug filetype

    case default
      write (filename, '(A,I0)') "Unknown_filetype_code=", filetype_code
      return
    end select

!    write (filename, '(A,I0,A,A,A,A,A,A)') &
!        trim(src%str_prefix),&
!        inst%index, &
!        trim(str_filter_code), &
!        trim(str_band_code), &
!        trim(str_filetype), &
!        trim(src%name),&
!        trim(src%str_suffix),&
!        '.fits'
    dscode = srctool_misc_dscode(inst%index, ESASS_FILTER_STR(inst%filter_code), 0)

    filename = &
        trim(src%str_prefix) // &
        dscode // &
        '_' // trim(str_filetype) // '_' // &
        trim(src%name) // &
        trim(src%str_suffix)

    ! add .fits if there's no extension in suffix
    if( index(src%str_suffix, ".") == 0 ) then
       filename = trim(filename) // ".fits"
    end if

  end subroutine srctool_misc_set_filename
  !-------------------------------------------------------------------------------



  recursive subroutine srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_instpar_pointer'
    type(inst_type), intent(in), target :: inst
    type(instpar_type), intent(out), pointer :: pinstpar
    integer(kind=int32), intent(out)   :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK

    if ( inst%instpar%is_valid ) then
      pinstpar => inst%instpar
    else if ( inst%pinst0%instpar%is_valid ) then
      pinstpar => inst%pinst0%instpar
    else
      ierr=ERROR_LEVEL_SEVERE
      write (str_buffer, '(A,A,A)') &
          "No valid instpar description was found for inst='", trim(inst%name),"'"
      call wrserr(subroutinename,ierr,str_buffer)
      return
    end if
  end subroutine srctool_misc_select_instpar_pointer



 !-------------------------------------------------------------------------------
  recursive subroutine srctool_misc_select_refarf_pointer(inst,prefarf,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_refarf_pointer'
    type(inst_type), intent(in), target :: inst
    type(arf_type), intent(out), pointer :: prefarf
    integer(kind=int32), intent(out)   :: ierr
    ierr=ERROR_LEVEL_OK
    nullify(prefarf)

!    write (*, '(A,I0)'), "inst%index = ", inst%index
!    write (*, '(A,A)'), "inst%name = ", trim(inst%name)
!    write (*, '(A,A)'), "inst%refarf%is_valid = ", logical2char(inst%refarf%is_valid)
!    write (*, '(A,A)'), "inst%pinst0%refarf%is_valid = ", logical2char(inst%pinst0%refarf%is_valid)

    if ( inst%refarf%is_valid ) then
      prefarf => inst%refarf
    else if ( inst%pinst0%refarf%is_valid ) then
      prefarf => inst%pinst0%refarf
    else
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,"No valid refarf description was found for inst=" // trim(inst%name))
      return
    end if
  end subroutine srctool_misc_select_refarf_pointer

 !-------------------------------------------------------------------------------
  recursive subroutine srctool_misc_select_refrmfset_pointer(inst,prefrmfset,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_refrmfset_pointer'
    type(inst_type), intent(in), target :: inst
    type(rmfset_type), intent(out), pointer :: prefrmfset
    integer(kind=int32), intent(out)   :: ierr
    ierr=ERROR_LEVEL_OK
    nullify(prefrmfset)

    if ( inst%refrmfset%is_valid ) then
      prefrmfset => inst%refrmfset
    else if ( inst%pinst0%refrmfset%is_valid ) then
      prefrmfset => inst%pinst0%refrmfset
    else
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,"No valid refrmfset description was found for inst=" // trim(inst%name))
      return
    end if
  end subroutine srctool_misc_select_refrmfset_pointer



  recursive subroutine srctool_misc_select_detmap_pointer(inst,pdetmap,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_detmap_pointer'
    type(inst_type), intent(in), target :: inst
    type(detmap_type), intent(out), pointer :: pdetmap
    integer(kind=int32), intent(out)   :: ierr
    ierr=ERROR_LEVEL_OK

    if ( inst%detmap%is_valid ) then
      pdetmap => inst%detmap
    else if ( inst%pinst0%detmap%is_valid ) then
      pdetmap => inst%pinst0%detmap
    else
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,"No valid detmap description was found for inst=" // trim(inst%name))
      return
    end if
  end subroutine srctool_misc_select_detmap_pointer




  recursive subroutine srctool_misc_select_psf_pointer(inst,ppsf,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_psf_pointer'
    type(inst_type), intent(in), target :: inst
    type(psf_type), intent(out), pointer :: ppsf
    integer(kind=int32), intent(out)   :: ierr
    ierr=ERROR_LEVEL_OK
    nullify(ppsf)

    if ( inst%psf%is_valid ) then
      ppsf => inst%psf
    else if ( inst%pinst0%psf%is_valid ) then
      ppsf => inst%pinst0%psf
    else
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,"No valid psf description was found for inst=" // trim(inst%name))
      return
    end if
  end subroutine srctool_misc_select_psf_pointer

  ! TODO - this currently only selects the cal badpix structure and ignores the
  ! badpix structures attached to individual evtlists
  recursive subroutine srctool_misc_select_badpix_pointer(inst,pbadpix,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_badpix_pointer'
    type(inst_type), intent(in), target :: inst
    type(badpix_type), intent(out), pointer :: pbadpix
    integer(kind=int32), intent(out)   :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK

    nullify(pbadpix)
    if ( inst%badpix%is_valid ) then
      pbadpix => inst%badpix
    else if ( inst%pinst0%badpix%is_valid ) then
      pbadpix => inst%pinst0%badpix
    else
      ierr=ERROR_LEVEL_SEVERE
      write(str_buffer, '(A,I0,A,A,A)') "No valid badpix description was found for inst index=", &
          inst%index, " (name=", trim(inst%name), ")"
      call wrserr(subroutinename,ierr,str_buffer)
      return
    end if
  end subroutine srctool_misc_select_badpix_pointer


  !-------------------------------------------------------------------------------------
  recursive subroutine srctool_misc_select_deadcor_pointer(inst,pdeadcor,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_deadcor_pointer'
    type(inst_type), intent(in), target :: inst
    type(deadcor_type), intent(out), pointer :: pdeadcor
    integer(kind=int32), intent(out)   :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK
    nullify(pdeadcor)
    if ( inst%deadcor%is_valid ) then
      pdeadcor => inst%deadcor
    else if ( inst%pinst0%deadcor%is_valid ) then
      pdeadcor => inst%pinst0%deadcor
    else
      ierr=ERROR_LEVEL_WARNING
      write (str_buffer,'(A,A)') "No valid deadcor description was found for inst: ", trim(inst%name)
      call wrserr(subroutinename,ierr,str_buffer)
      return
    end if
  end subroutine srctool_misc_select_deadcor_pointer
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  recursive subroutine srctool_misc_select_ebounds_pointer(inst,pebounds,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_ebounds_pointer'
    type(inst_type), intent(in), target :: inst
    type(ebounds_type), intent(out), pointer :: pebounds
    integer(kind=int32), intent(out)   :: ierr
    ierr=ERROR_LEVEL_OK
    nullify(pebounds)

    if ( inst%ebounds%is_valid ) then
      pebounds => inst%ebounds
    else if ( inst%pinst0%ebounds%is_valid ) then
      pebounds => inst%pinst0%ebounds
    else
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,"No valid ebounds description was found for inst:" // trim(inst%name))
      return
    end if
  end subroutine srctool_misc_select_ebounds_pointer
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !use the first valid gti extension of the first eventlist as a template
  recursive subroutine srctool_misc_select_gtilist_pointer(inst,pgtilist,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_gtilist_pointer'
    type(inst_type), intent(in), target :: inst
    type(gtilist_type), intent(out), pointer :: pgtilist
    integer(kind=int32), intent(out)   :: ierr
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i
    ierr=ERROR_LEVEL_OK
    nullify(pgtilist)

    do i=1,inst%evtlistlist%num_evtlists
      if ( inst%evtlistlist%evtlist(i)%is_valid ) then
        if ( inst%evtlistlist%evtlist(i)%gtilist%is_valid ) then
          pgtilist => inst%evtlistlist%evtlist(i)%gtilist
          exit
        end if
      end if
    end do

    if ( .not. associated(pgtilist) ) then
      ierr=ERROR_LEVEL_SEVERE
      write (str_buffer,'(A,A)') "No valid input gtilist description was found for inst: ", trim(inst%name)
      call wrserr(subroutinename,ierr,str_buffer)
      return
    end if


  end subroutine srctool_misc_select_gtilist_pointer
  !-------------------------------------------------------------------------------------

!  !-------------------------------------------------------------------------------
!  recursive subroutine srctool_misc_select_vign_pointer(inst,pvign,ierr)
!    character(len=*), parameter :: subroutinename='srctool_misc_select_vign_pointer'
!    type(inst_type), intent(in), target :: inst
!    type(vign_type), intent(out), pointer :: pvign ! ok, leave this here for now
!    integer(kind=int32), intent(out)   :: ierr
!    ierr=ERROR_LEVEL_OK
!    nullify(pvign)
!
!    if ( inst%vign%is_valid ) then
!      pvign => inst%vign
!    else if ( inst%pinst0%vign%is_valid ) then
!      pvign => inst%pinst0%vign
!    else
!      ierr=ERROR_LEVEL_SEVERE
!      call wrserr(subroutinename,ierr,"No valid tabulated vignetting description was found")
!      return
!    end if
!  end subroutine srctool_misc_select_vign_pointer
!  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_misc_select_avign_pointer(inst,pavign,ierr)
    character(len=*), parameter :: subroutinename='srctool_misc_select_avign_pointer'
    type(inst_type), intent(in), target :: inst
    type(avign_type), intent(out), pointer :: pavign
    integer(kind=int32), intent(out)   :: ierr
    ierr=ERROR_LEVEL_OK
    nullify(pavign)

    if ( inst%avign%is_valid ) then
      pavign => inst%avign
    else if ( inst%pinst0%avign%is_valid ) then
      pavign => inst%pinst0%avign
    else
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,"No valid analytic vignetting description was found")
      return
    end if
  end subroutine srctool_misc_select_avign_pointer
  !-------------------------------------------------------------------------------





end module srctool_misc
