! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_engine

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2013-02-14
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of work routines for srctool
  !
  !-----------------------------------------------------------------------------
  !

  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc
  use iso_c_binding, only: C_NULL_PTR

#ifdef _OPENMP
  use omp_lib
#endif


  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_generic_regions   ! generic region handling routines
  use srctool_generic_avign     ! generic analytic vignetting file handling routines
  use srctool_datatypes         ! structs and datatypes
  use srctool_misc              !
  use srctool_instpar           !
  use srctool_gti               !
  use srctool_ebounds           !
  use srctool_spec              !
  use srctool_att               !
!  use srctool_vign              !
  use srctool_param             !
  use srctool_lc                !
  use srctool_badpix            !
  use srctool_pattcal           !
  use srctool_psf               !
  use srctool_rmf               !
  use srctool_arf               !
  use srctool_detmap            !
  use srctool_deadcor           !
  use srctool_cal               !
  use srctool_samplelist        !
  use srctool_evtlist

  implicit none     ! insist on explicit typing

  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_MOD'

  ! ------------------------------------------------------------------------------
  ! ------------------------------------------------------------------------------
  ! Notes on input data file and data structure hierarchy
  !
  ! ---------
  ! EVTFILES
  ! ---------
  ! We expect to read in up to MAX_EVENTFILES eventfiles We must
  ! expect that in general each events file can contain events from
  ! just one, a subset of, or all instruments
  !
  ! Each eventfile contains :
  !   - Exactly one EVENTS extension
  !   - Up to MAX_INST gti (GTIn|QUALGTIn|FLAREGTIn) extensions
  !   - Up to MAX_INST badpix (BADPIXn) extensions
  !   - Up to MAX_INST attitude (CORRATTn) extensions
  !   - Up to MAX_INST deadtime correction (DEADCORn) extensions
  !
  ! ----------------
  ! ADITTIONAL FILES
  ! ----------------
  ! We also expect to read in files containing:
  !  - One reference ARF extension per instrument considered (SPECRESP)
  !  - One vignetting description extension per instrument considered (VIGNET)
  !  - One PSF description per instrument considered (PSFDESC)
  !  - One additional "calibration" bad pixel extension per instrument considered (BADPIXn)
  !  - One INSTPAR description per instrument giving basic detector array geometry info
  !  - One DETMAP  description per instrument giving on/off sensitivity map of each focal plane
  !  - One energy bounds extension per instrument considered (EBOUNDS)
  ! A calibration file can be supplied for all instruments by giving the dummy instrument name "TM0" (previously "FM0")
  !
  ! ----------------
  ! RULES
  ! ----------------
  !
  ! Notes:
  !  - GTIn|QUALGTIn|FLAREGTIn, BADPIXn, CORRATTn, DEADCORn extensions are allocated to a
  !    specific eROSITA instrument via their "n" digit (1...MAX_INST)
  !  - A maxium of one GTIn|QUALGTIn|FLAREGTIn, BADPIXn, CORRATTn,
  !    or DEADCORn extension is allowed per eROSITA instrument
  !  - Each eventfile is read into a temporary buffer, and then divided up between instruments
  !  - We read all SPECRESP, EBOUNDS, VIGNET, and PSFDESC extensions found in the
  !    supplied filename lists, or from the CALDB
  !  - If a SPECRESP, EBOUNDS, VIGNET, or PSFDESC extension has keyword INSTRUME = "TM0"  (previously "FM0"),
  !    then we assume that it applies to all instruments.
  !  - The INSTPAR and DETMAP data files are always read from the CALDB
  !  - The SPECRESP, EBOUNDS, VIGNET, and PSFDESC extensions are associated
  !    with instruments via the INSTRUME keyword
  !
  !  - If zero DEADCORn extensions are supplied then no deadcor correction is applied
  !  - If zero BADPIXn extensions are supplied then no badpixel correction is applied

  ! A fatal error occurs when reading the evtlists:
  !  - If the value of the INSTRUME keyword doesn't agree with the "n" in the EXTNAME
  !  - If zero CORRATTn extensions are supplied/found
  !  - If zero GTIn|QUALGTIn|FLAREGTIn extensions are supplied/found
  !
  ! A fatal error occurs when reading calibration data :
  !  - If zero SPECRESP descriptions are found
  !  - If zero EBOUNDS descriptions are found
  !  - If zero PSFDESC descriptions are found
  !  - If zero INSTPAR or DETMAP descriptions are found
  !  - If more than one SPECRESP, EBOUNDS, VIGNET, or PSFDESC HDU is supplied for an instrument
  ! Warnings:
  !  - The user is warned if there are any missing BADPIXn or DEADCORn extensions
  !  - If zero VIGNET extensions are supplied then the program emits a warning and
  !    assumes that there is zero vignetting
  !
  ! ------------------------------------------------------------------------------
  ! ------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  ! PSF handling, interpolation and effective area calculations
  !
  ! For each time step we will have to calculate the local PSF shape
  ! at the position of the source, at a number of energies, and modulo
  ! vignetting etc. We then sum up the values of the PSF over the
  ! (non-bad) pixels which fall within the extraction region.  For the
  ! ARF calculation, the fracresp values are integrated over time
  ! giving a vector in energy channels. For the LCCORR calculation, the
  ! summation is carried out over a finite energy band, giving a vector in time bins.
  !
  !  ARF_i(e)      = sum_t {fracrsp_i(e,t)}
  !  LCCORR_i(E,t) = sum_e {fracrsp_i(e,t)}
  !  where i=inst, e=bnarrow energy channel, E=broad energy channel, t=time step
  !
  !  -----------------------------------------------------------------------------
  !
  !  Interpolating "small map" style PSFs, and gridded vignetting model
  !
  !  PSF maps are supplied at a small number of energy ranges
  !  (Elo-Ehi), off-axis angles (theta), and position angles (phi)
  !  Finding the value of these maps at a given location in
  !  (E,theta,phi) requires a cubic interpolation This interpolation
  !  should be done the minimum number of times possible for efficiency.
  !  The interpolated PSF must then be rotated, translated and resampled
  !  into the physical pixel grid within the extraction region
  !
  !  The PSF will typically be specified at only a small number of
  !  energy levels.  Therefore our approach is to calculate the
  !  fraction of the instantaneous PSF falling in the extraction
  !  ragion at its native energy levels (interpolating in theta,phi to
  !  do this), and then to interpolate these fractional responses in
  !  energy later:
  !    fracrsp_i(e,t) = fracpsf_i(e,t) * vigneting_i(e,t)
  !  where fracpsf_i(e,t) and vigneting_i(e,t) are each interpolated
  !  from a number of fixed points in energy:
  !  fracpsf_i(e,t) =
  !     interp(e){fracpsf_i(e1,t),fracpsf_i(e2,t),...,fracpsf_i(eN,t)}
  !
  !  Note that the PSF and vignetting systems are not necessarily
  !  defined at the same grid of energies, therefore we first
  !  determine the grid of unique energies at which either the
  !  vignetting or the PSF model is defined, and then interpolate the
  !  fracpsf and vignetting at each of these unique energies finally
  !  we interpolate to the output energy pixelisation
  !  -------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !  Estimate of calculation steps for ARF:
  !
  !  SubFact= pixel resampling factor = num sub-pixels per physical pixel ~ 4-16
  !  Npix   = typical number of physical pixels in src extract region ~100 (<<384^2)
  !  Nsubpix= typical number of sub-pixels in src extract region = Npix*SubFact ~ 400->1600
  !  Tstep  = tstep   default ~ 50ms (~frametime)
  !  Texp   = typical exposure time ~ 2-10ks in all sky survey mode (eRASS8)
  !  Ntstep = total number of time steps = Texp/Tstep = 40000 -> 200000
  !  Nebins = typical number of energy bins to consider = ~10keV/(~10eV) ~ 1000
  !  Ninst  = 7 instruments
  !
  !  Nsamp  = Ninst * Nebins *  Ntstep * Nsubpix
  !         ~ 7     * 1e3    *  1e5    * 1e3
  !         ~ 1e12 -> which is quite a big number!
  !
  ! However, considering that for the small maps PSF models (and also
  ! vignetting descriptions) are only supplied at ~3 energies we can
  ! reduce the number of times that need to evaluate the PSF by a
  ! factor ~300 (Nebins_old/Nebins_new = 1000/3) This brings the Nsamp
  ! down to a much more manageable ~1e9-1e10
  !
  ! We will have to revisit this for the other types of PSF description (e.g. shapelets)
  !-------------------------------------------------------------------------------

  !--------------------------------------------------------------------------
  ! Notes on standard exposure time keywords:
  ! http://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/ofwg_recomm/r11.html
  !--------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  ! subroutine declarations and interfaces for module: srctool_engine
  private :: srctool_allocate_inst
  private :: srctool_deallocate_inst
  private :: srctool_empty_inst
  public  :: srctool_empty_instlist
  public :: srctool_allocate_instlist
  public :: srctool_deallocate_instlist
  private :: srctool_select_events_parallel
  private :: srctool_write_source_specific_evtlist
  private :: srctool_write_source_specific_evtlists
  private :: srctool_empty_intervallist
  public  :: srctool_setup_intervallist
  private :: srctool_set_spec_keys_from_arf
  private :: srctool_set_speclist_keys_from_arflist
  private :: srctool_increment_tstep
  private :: srctool_increment_calc_srcgti_tstep

!  private :: debug_write_detailed_tstep_info
  private :: srctool_calc_fracrsp_tstep
  private :: srctool_init_lcbin
  private :: srctool_finish_lcbin
  private :: srctool_calc_fracrsp
  public  :: srctool_calc_fracrsps_parallel
  public  :: srctool_calc_fracrsps_serial
  private :: srctool_calc_srcgti
  private :: srctool_merge_srcgtis
  private :: srctool_calc_srcgtis_parallel
  private :: srctool_make_dummy_srcgtis
  public  :: srctool_write_regions
  private :: srctool_prepare_extent_model_params
  private :: srctool_prepare_extent_model
  public  :: srctool_setup_extended_source_handling
  public  :: srctool_do_all_processing_for_source
!  public  :: srctool_test_harness
  private :: srctool_build_survey_psfs_for_inst
  public  :: srctool_build_survey_psfs_for_instlist

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------


contains

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !------------------------ SUBROUTINES ------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------







  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !------------------ START of inst/instlist structure handling ------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  ! allocate inst structure - we do this rather than use normal variables to avoid a
  ! gfortran bug in gfortran 7/8/9 where enormous amounts of code are generated to
  ! initialise the data structure (scales like N^2, where N is the number of allocatable
  ! arrays).
  ! See https://gcc.gnu.org/bugzilla/show_bug.cgi?id=87352
  ! using pointers breaks the links between the data types and prevents this scaling.
  subroutine srctool_allocate_inst(inst, ierr)
    character(len=*), parameter :: subroutinename='srctool_allocate_inst'
    type(inst_type), intent(inout)      :: inst
    integer(kind=int32), intent(out)    :: ierr
    integer(kind=int32) :: status
    allocate(inst%evtlistlist, &
         inst%refarf, &
         inst%refrmfset, &
         inst%ebounds, &
         inst%avign, &
         inst%psf, &
         inst%survey_psf, &
         inst%badpix, &
         inst%detmap, &
         inst%instpar, &
         inst%attlist, &
         inst%deadcor, &
         inst%gtilist, &
         inst%pattcal, &
         inst%effmap, &
         inst%src_gtilist, &
         inst%samplelist_src, &
         inst%samplelist_back, &
         STAT=status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return
  end subroutine srctool_allocate_inst

  !-------------------------------------------------------------------------------------
  subroutine srctool_deallocate_inst(inst)
    type(inst_type), intent(inout)      :: inst
    deallocate(inst%evtlistlist, &
         inst%refarf, &
         inst%refrmfset, &
         inst%ebounds, &
         inst%avign, &
         inst%psf, &
         inst%survey_psf, &
         inst%badpix, &
         inst%detmap, &
         inst%instpar, &
         inst%attlist, &
         inst%deadcor, &
         inst%gtilist, &
         inst%pattcal, &
         inst%effmap, &
         inst%src_gtilist, &
         inst%samplelist_src, &
         inst%samplelist_back &
         )
  end subroutine srctool_deallocate_inst

  !------------------------------------------------------------------------------------
  subroutine srctool_empty_inst(inst)
    type(inst_type), intent(inout)      :: inst
    inst%is_valid    = .false.
    inst%to_consider = .false.
    nullify(inst%pinst0)

    call srctool_evtlist_empty_evtlistlist(inst%evtlistlist)
    call srctool_arf_empty_arf(inst%refarf)
    call srctool_rmf_empty_rmfset(inst%refrmfset)
    call srctool_ebounds_empty_ebounds(inst%ebounds)
    call avign_empty_avign(inst%avign)
    call srctool_psf_empty_psf(inst%psf)
    call srctool_psf_empty_psf(inst%survey_psf)
    call srctool_badpix_empty_badpix(inst%badpix)
    call srctool_detmap_empty_detmap(inst%detmap)
    call srctool_instpar_empty_instpar(inst%instpar)
    call srctool_att_empty_attlist(inst%attlist)
    call srctool_gti_empty_gtilist(inst%gtilist)
    call srctool_deadcor_empty_deadcor(inst%deadcor)
    call srctool_pattcal_empty_pattcal(inst%pattcal)
    call srctool_badpix_empty_effmap(inst%effmap)
    call srctool_gti_empty_gtilist(inst%src_gtilist)
    call srctool_samplelist_empty(inst%samplelist_src)
    call srctool_samplelist_empty(inst%samplelist_back)
    inst%tstep      = 1.0d0
    inst%xgrid(:)   = 1.0d0
    inst%filter_code = ESASS_FILTER_CODE_NULL
    inst%use_fov_mask = .false.

  end subroutine srctool_empty_inst
  !------------------------------------------------------------------------------------

  !------------------------------------------------------------------------------------
  subroutine srctool_allocate_instlist(instlist, ierr)
    type(instlist_type), intent(inout)  :: instlist
    integer(kind=int32), intent(out)    :: ierr
    integer(kind=int32) :: i
    do i=0, MAX_INST
       call srctool_allocate_inst(instlist%inst(i), ierr)
    end do
  end subroutine srctool_allocate_instlist

  !------------------------------------------------------------------------------------
  subroutine srctool_deallocate_instlist(instlist)
    type(instlist_type), intent(inout)  :: instlist
    integer(kind=int32) :: i
    do i=0, MAX_INST
       call srctool_deallocate_inst(instlist%inst(i))
    end do
  end subroutine srctool_deallocate_instlist

  !------------------------------------------------------------------------------------
  subroutine srctool_empty_instlist(instlist)
    type(instlist_type), intent(inout),target  :: instlist
    integer(kind=int32) :: i
    do i=0,MAX_INST
      call srctool_empty_inst(instlist%inst(i))
      instlist%inst(i)%index = i
      write (instlist%inst(i)%name, '(A,I0)') INST_STEM, i
      ! set up the pointer to the generic instrument
!      if ( i == 0 ) then
!        instlist%inst(i)%pinst0 => instlist%inst(1)      ! TODO make this point to first valid instrument!!
!      else
        instlist%inst(i)%pinst0 => instlist%inst(0)
!      endif
    end do
    instlist%num_inst_to_consider = 0
    instlist%min_tstart =  HUGE(instlist%min_tstart) ! very big +ve number
    instlist%max_tstop  = -HUGE(instlist%max_tstop)  ! very big -ve number
    call srctool_empty_intervallist(instlist%intervallist)
    instlist%gtitype_to_read = SRCTOOL_CALTYPE_NULL

    instlist%num_survey_mode = 0
    instlist%num_pointing_mode = 0
    instlist%num_scan_mode = 0
    instlist%num_slew_mode = 0

    instlist%fft_npix = 0
    instlist%fft_pixscale = 0
    instlist%fftw_plan_r2c = C_NULL_PTR
    instlist%fftw_plan_c2r = C_NULL_PTR

  end subroutine srctool_empty_instlist
  !------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of of inst/instlist structure handling --------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------










  !--------------------------------------------------------------------------------------
  ! same as above but run each instrument in parallel
  recursive subroutine srctool_select_events_parallel(instlist, src, ierr)
    character(len=*), parameter :: subroutinename='srctool_select_events_parallel'
    type(instlist_type),intent(inout),target    :: instlist
    type(src_type),intent(inout)      :: src
    integer(kind=int32), intent(out)  :: ierr

    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i,e,j, num_valid, flag
!    integer(kind=int32) :: nevents_src_i(MAX_INST), nevents_back_i(MAX_INST)
!    integer(kind=int32) :: nevents_src, nevents_back

    type(coord_type) :: evt_coord
    type(inst_type),pointer :: pinst
    type(evtlist_type),pointer :: pevtlist
    integer(kind=int32)  :: ierr_i(MAX_INST)
    integer(kind=int16)  :: pat_typ_int16
#ifdef _OPENMP
    integer(kind=int32)  :: flag_int32
!    integer(kind=int32)  :: max_threads
#endif
    integer(kind=int32)  :: ii, iidx
    integer(kind=int32)  :: i_ok(1:MAX_INST)
    integer(kind=int32)  :: max_i_ok

    integer(kind=int32)  :: fov_flag

!    integer(kind=int32), parameter :: flag_mask_int32 = 2147479552 ! =(2**31 - 2**12) - only the "rejection" bits (bits 12-32)
    num_valid = 0
    src%nevents_src_i(:) = 0
    src%nevents_back_i(:) = 0
    ierr = ERROR_LEVEL_OK
    ierr_i(:)=ERROR_LEVEL_OK

    ! initialise ownership details for source
    call srctool_owner_setup_owner(src%owner)

!#ifdef _OPENMP
!    max_threads = omp_get_max_threads()
!    write (str_buffer, '(A,I0)') &
!        "Running with max threads: ",max_threads
!    call wrschatty(subroutinename, str_buffer)
!#endif

    write (str_buffer, '(A,I0,A,A,1x,F12.7,1x,F12.7,A)'  ) &
        "Selecting events for src: ", src%index, " (",trim(src%name), &
    src%coord%long, src%coord%lat,")"
    call wrschatty(subroutinename, str_buffer)

    ! for each instrument
    ! Start of FOMP handling

    ! make a list of valid instruments
    max_i_ok = 0
    do i=1,MAX_INST
       if ( instlist%inst(i)%to_consider .and. instlist%inst(i)%is_valid ) then
          max_i_ok = max_i_ok + 1
          i_ok(max_i_ok) = i
       end if
    end do

    ! set FoV flag for each instrument
    do ii=1,max_i_ok
       pinst => instlist%inst(i_ok(ii))
       if(pinst%to_consider .and. pinst%is_valid) then

          ! set flag for filtering events inside FoV
          ! -1=unset, 0=out_of_fov filtered, 1=out_of_fov unfiltered, -2=mismatch
          fov_flag = -1
          do e = 1, pinst%evtlistlist%num_evtlists
             pevtlist => pinst%evtlistlist%evtlist(e)
             if( iand(pevtlist%key_flagsel%value, FLAG_OUT_OF_FOV) == 0 ) then
                ! flag is not set, so we don't filter
                if(fov_flag == 0 .or. fov_flag == -1) then
                   fov_flag = 0
                else
                   fov_flag = -2
                end if
             else
                ! flag is set, so we do filter
                if(fov_flag == 1 .or. fov_flag == -1) then
                   fov_flag = 1
                else
                   fov_flag = -2
                end if
             end if
          end do
          if(fov_flag < 0) then
             call wrserr(subroutinename,ierr_i(i), "Mismatch in OUT_OF_FOV flag for event lists")
             fov_flag = 1
          end if
          pinst%use_fov_mask = fov_flag /= 0
       end if
    end do

    !$OMP PARALLEL DO shared(src, instlist) private (pinst, pevtlist, pat_typ_int16, flag_int32, evt_coord, iidx, j, flag, e)
    do ii=1,max_i_ok
      iidx = i_ok(ii)
      pinst => instlist%inst(iidx)
      ! if instrument should be considered
      if ( pinst%to_consider .and. pinst%is_valid ) then

        ! for each eventlist for this inst
        do e=1,pinst%evtlistlist%num_evtlists
          pevtlist => pinst%evtlistlist%evtlist(e)
          ! reset all the evt code flags to null
          pevtlist%pclass_mask(:) = EVT_CLASS_FLAG_NULL
          ! for each input event
          do j=1,pevtlist%num_rows
            evt_coord%coordsys_code = COORDSYS_CODE_ICRS
            evt_coord%long = pevtlist%pra(j)
            evt_coord%lat  = pevtlist%pdec(j)
            evt_coord%x    = pevtlist%px(j)
            evt_coord%y    = pevtlist%py(j)
            evt_coord%rawx = pevtlist%prawx(j) ! TODO maybe also need to take account of subx,suby
            evt_coord%rawx = pevtlist%prawy(j)


            ! test if event meets the flag selection criteria
            ! - only include events with any flag bits set which do not match the mask
!            ! - take notice of only the "rejection" bits (bits 12-32)
!            flag_int32 = IAND(pevtlist%pflag(j), flag_mask_int32)
!            if ( IAND(flag_int32, pevtlist%key_flagsel%value) == 0 ) then
            if ( IAND(pevtlist%pflag(j), pevtlist%key_flagsel%value) == 0 ) then

              ! test if event meets the pattern selection criteria
              ! - only include events with pattern bits set which match at least one mask bit
              ! convert the pattern type into a bitmask
              pat_typ_int16 = INT(2**(pevtlist%ppat_typ(j)-1), int16)
              if ( pevtlist%ppat_typ(j) >= 1 .and. pevtlist%ppat_typ(j) <= 4 .and. &
                   IAND(pat_typ_int16, pevtlist%key_pat_sel%value) /= 0 ) then

                ! NEW:  test if event is within a valid GTI interval
                if ( srctool_gti_test_if_in_gtilist(pinst%src_gtilist, pevtlist%ptime(j)) ) then

                  ! test if event is in src region
                  call srctool_regions_test_if_coord_in_region(evt_coord, src%src_region, flag, ierr_i(iidx))
                  if ( ierr_i(iidx) /= ERROR_LEVEL_OK ) then
                    exit
                  end if
                  if ( flag > 0 ) then
                    pevtlist%pclass_mask(j) = IOR(pevtlist%pclass_mask(j),EVT_CLASS_FLAG_SRC)
                    src%nevents_src_i(iidx) = src%nevents_src_i(iidx) + 1
                    call srctool_owner_applyflag_owner(src%owner, pevtlist%pflag(j))
                  end if

                  ! test if event is in background region
                  if ( src%do_back ) then
                    call srctool_regions_test_if_coord_in_region(evt_coord, src%back_region, flag, ierr_i(iidx))
                    if ( ierr_i(iidx) /= ERROR_LEVEL_OK ) then
                      exit
                    end if
                    if ( flag > 0 ) then
                      pevtlist%pclass_mask(j) = IOR(pevtlist%pclass_mask(j),EVT_CLASS_FLAG_BACK)
                      src%nevents_back_i(iidx) = src%nevents_back_i(iidx) + 1
                      call srctool_owner_applyflag_owner(src%owner, pevtlist%pflag(j))
                    end if
                  end if
                end if
              end if
            end if
          end do

          if ( ierr_i(iidx) /= ERROR_LEVEL_OK ) then
            exit
          end if
        end do
      end if
    end do
    !$OMP END PARALLEL DO

    do i=1,MAX_INST
      if ( ierr_i(i) /= ERROR_LEVEL_OK ) then
        call wrserr(subroutinename,ierr_i(i),"Problem selecting events")
        ierr = ierr_i(i)
        return
      end if
    end do

    call wrschatty(subroutinename, "Source=" // trim(src%name) // &
        ": Selected " // trim(i2s(sum(src%nevents_src_i(:)))) //  " source events" )

    ! give a warning if the ownership details are invalid
    call srctool_owner_warnon_owner(src%owner)

    ! update prefix based on ownership if in pipeline mode
    call srctool_owner_modify_src_prefix(src)

    if ( src%do_back ) then
      call wrschatty(subroutinename, &
          "Source=" // trim(src%name) // ": Selected " // trim(i2s(sum(src%nevents_back_i(:)))) // " background events" )
    endif
  end subroutine srctool_select_events_parallel
  !---------------------------------------------------------------------------------------------





!  !-------------------------------------------------------------------------------
!  ! write a source-specific event list to file
!  ! use a virtual memory file to speed things up and to allow standard row selection routines
!  ! this will allow copying of
!  subroutine srctool_write_source_specific_evtlist_new(inst, src, gtilist, ierr)
!    character(len=*), parameter :: subroutinename='srctool_write_source_specific_evtlist_new'
!
!    type(inst_type),  intent(in), target :: inst
!    type(src_type), intent(in) :: src
!    type(gtilist_type),  intent(in) :: gtilist
!    integer(kind=int32), intent(out)  :: ierr
!
!
!  end subroutine srctool_write_source_specific_evtlist_new
!  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! write a source-specific event list to file
  ! TODO - write standard headers
  ! TODO - replace this with a much improved routine or a call to evtool?
  subroutine srctool_write_source_specific_evtlist(inst, src, gtilist, write_source, ierr)
    character(len=*), parameter :: subroutinename='srctool_write_source_specific_evtlist'

    type(inst_type),  intent(in), target :: inst
    type(src_type), intent(in) :: src
    type(gtilist_type),  intent(in) :: gtilist
    logical, intent(in) :: write_source   ! true=source, false=background
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32), parameter :: blank_bitpix = 8
    integer(kind=int32), parameter :: blank_naxis = 0
    integer(kind=int32), parameter :: blank_naxes(1) = (/ 0 /)

    integer(kind=int32), parameter :: COLNUM_TIME          =  1
    integer(kind=int32), parameter :: COLNUM_RA            =  2
    integer(kind=int32), parameter :: COLNUM_DEC           =  3
    integer(kind=int32), parameter :: COLNUM_X             =  4
    integer(kind=int32), parameter :: COLNUM_Y             =  5
    integer(kind=int32), parameter :: COLNUM_ENERGY        =  6
    integer(kind=int32), parameter :: COLNUM_EV_WEIGHT     =  7
    integer(kind=int32), parameter :: COLNUM_RAWX          =  8
    integer(kind=int32), parameter :: COLNUM_RAWY          =  9
    integer(kind=int32), parameter :: COLNUM_SUBX          = 10
    integer(kind=int32), parameter :: COLNUM_SUBY          = 11
    integer(kind=int32), parameter :: COLNUM_PHA           = 12
    integer(kind=int32), parameter :: COLNUM_PAT_TYP       = 13
    integer(kind=int32), parameter :: COLNUM_PAT_INF       = 14
    integer(kind=int32), parameter :: COLNUM_TM_NR         = 15
    integer(kind=int32), parameter :: COLNUM_FLAG          = 16
    integer(kind=int32), parameter :: ncols=16

    character(len=*),    parameter :: ttype(ncols) = &
        (/"TIME     ",&
          "RA       ","DEC      ",&
          "X        ","Y        ",&
          "ENERGY   ","EV_WEIGHT",&
          "RAWX     ","RAWY     ",&
          "SUBX     ","SUBY     ",&
          "PHA      ","PAT_TYP  ",&
          "PAT_INF  ","TM_NR    ",&
          "FLAG     "/)
    character(len=*),    parameter :: tunit(ncols) = &
        (/"s     ",&
          "deg   ","deg   ",&
          "skypix","skypix",&
          "eV    ","ratio ",&
          "pix   ","pix   ",&
          "pix   ","pix   ",&
          "adu   ","code  ",&
          "info  ","index ",&
          "flag  "/)
    character(len=*),    parameter :: tform(ncols) = &
        (/"D",&
          "D","D",&
          "J","J",&
          "E","E",&
          "I","I",&
          "I","I",&
          "I","I",&
          "B","B",&
          "J"/)

    character(len=*),    parameter :: extname = "EVENTS"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int64) :: frow
    integer(kind=int64), parameter ::  felem = 1

    character(len=LSTR_MAX)   :: str_filename

    integer(kind=int32) :: i, j
    integer(kind=int32) :: unit
    integer(kind=int32)    :: status
    integer(kind=int32) :: nevents
    type(hduinfo_type) :: hduinfo

    character(len=8)   :: key_name

    ierr = ERROR_LEVEL_OK
    status = 0
    unit = 0

    if( write_source ) then
       call srctool_misc_set_filename(str_filename, inst, src, SRCTOOL_FILETYPE_CODE_SRC_EVT)
    else
       call srctool_misc_set_filename(str_filename, inst, src, SRCTOOL_FILETYPE_CODE_BACK_EVT)
    end if

!    call fitswrap_set_key(evt%hduinfo%key_dstype, ESASS_DSTYPE_EVT )

    if ( len_trim(str_filename) <= 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Illegal filename for source specific evtlist file")
      return
    end if

    call wrschatty (subroutinename, "Writing source specific evtlist file, filename= " // trim(str_filename))

    !-----------------------------------
    ! get a fortran unit
    call FTGIOU( unit, status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT(unit,str_filename,FITSIO_BLOCKSIZE, status)
    if (status /= 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

    ! append a new empty primary image hdu and write date keyword to the HDU
    call FTIIMG(unit,blank_bitpix,blank_naxis,blank_naxes, status)
    call FTPDAT(unit, status)
    if (status /= 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
!
!    !-----------------------------------
!    ! write the standard FITS header:
!    call FTPHPS(unit,bitpix,naxis,naxes,status)
!    if (status /= 0) then
!      ierr = ERROR_LEVEL_SEVERE
!      call fitswrap_write_fits_error(subroutinename, status)
!      return
!    end if
!    !-----------------------------------
!
!    !-----------------------------------
!    ! write the compulsory keywords to the primary HDU
!    ! write date keyword
!    call  FTPDAT(unit, status)
!    if (status /= 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------

    ! now set up the format for the table extension
    if(write_source) then
       nevents = src%nevents_src_i(inst%index)
    else
       nevents = src%nevents_back_i(inst%index)
    end if

    !-----------------------------------
    ! append a new binary table HDU
    call FTIBIN(unit,nevents,ncols,ttype,tform,tunit,extname,varidat,status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

!    !-----------------------------------
!    ! append a new table HDU
!    call FTCRHD(unit, status)
!    if (status /= 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------

!    !-----------------------------------
!    ! write header info for the binary table:
!    call FTPHBN(unit,nevents,ncols,ttype,tform,tunit,extname,varidat,status)
!    if (status /= 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!    end if
!    !-----------------------------------
!
    ! TODO copy over appropriate header keywords
!    !-----------------------------------
!    call srctool_write_evtlist_keys(unit,inst%evtlistlist, ierr)
!    if ( ierr /= ERROR_LEVEL_OK ) then
!      call wrserr (subroutinename, ierr, "Problem writing evtlist fits keys")
!      return
!    end if
!    !-----------------------------------

    ! set header for events
    ! copy input event list
    if ( inst%evtlistlist%num_evtlists <= 0  ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Instrument structure contains no evtlists")
       return
    end if
    hduinfo = inst%evtlistlist%hduinfo

    call fitswrap_set_key(hduinfo%key_owner, &
         trim(srctool_owner_getkeyword_owner(src%owner)))
    call fitswrap_set_key(hduinfo%key_object  , trim(src%name) )
    call fitswrap_set_key(hduinfo%key_ra_obj  , src%coord%long)
    call fitswrap_set_key(hduinfo%key_dec_obj , src%coord%lat)

    ! update TSTART/TSTOP using GTI
    if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
       call fitswrap_set_key(hduinfo%key_tstart, gtilist%pstart(1))
       call fitswrap_set_key(hduinfo%key_tstop, gtilist%pstop(gtilist%num_rows))
    end if

    call fitswrap_write_hduinfo_quiet(unit, hduinfo, ierr)
    if (ierr /= ERROR_LEVEL_OK) then
       call wrserr (subroutinename, ierr, "Problem writing HDU header")
       return
    end if

    ! write wcs information about X and Y columns
    if (size(inst%evtlistlist%evtlist) >= 1) then
       associate ( wcs => inst%evtlistlist%evtlist(1)%wcs )
         write(key_name, '(A,I1)') 'TCRVL', COLNUM_X
         call FTPKYD(unit, key_name, wcs%rval_x, FITSIO_DECIMALS, "", status)
         write(key_name, '(A,I1)') 'TCRVL', COLNUM_Y
         call FTPKYD(unit, key_name, wcs%rval_y, FITSIO_DECIMALS, "", status)
         write(key_name, '(A,I1)') 'TCRPX', COLNUM_X
         call FTPKYD(unit, key_name, wcs%rpix_x, FITSIO_DECIMALS, "", status)
         write(key_name, '(A,I1)') 'TCRPX', COLNUM_Y
         call FTPKYD(unit, key_name, wcs%rpix_y, FITSIO_DECIMALS, "", status)
         write(key_name, '(A,I1)') 'TCDLT', COLNUM_X
         call FTPKYD(unit, key_name, wcs%delta_x, FITSIO_DECIMALS, "", status)
         write(key_name, '(A,I1)') 'TCDLT', COLNUM_Y
         call FTPKYD(unit, key_name, wcs%delta_y, FITSIO_DECIMALS, "", status)
         write(key_name, '(A,I1)') 'TCTYP', COLNUM_X
         call FTPKYS(unit, key_name, trim(wcs%type_x), "", status)
         write(key_name, '(A,I1)') 'TCTYP', COLNUM_Y
         call FTPKYS(unit, key_name, trim(wcs%type_y), "", status)
         write(key_name, '(A,I1)') 'TCUNI', COLNUM_X
         call FTPKYS(unit, key_name, trim(wcs%unit_x), "", status)
         write(key_name, '(A,I1)') 'TCUNI', COLNUM_Y
         call FTPKYS(unit, key_name, trim(wcs%unit_y), "", status)
       end associate
    end if

    ! copy the rows from the input evtfiles to the output file
    frow = 1
    do i=1,inst%evtlistlist%num_evtlists
       associate ( evtlist => inst%evtlistlist%evtlist(i) )
         if ( evtlist%is_valid ) then
            do j=1,evtlist%num_rows
               if( ( write_source .and. &
                     iand(evtlist%pclass_mask(j),EVT_CLASS_FLAG_SRC) /= 0 ) .or. &
                    (.not. write_source .and. &
                     iand(evtlist%pclass_mask(j),EVT_CLASS_FLAG_BACK) /= 0 ) ) then

                  call FTPCLD(unit, COLNUM_TIME     ,frow,felem,1, evtlist%ptime(j), status)
                  call FTPCLD(unit, COLNUM_RA       ,frow,felem,1, evtlist%pra(j), status)
                  call FTPCLD(unit, COLNUM_DEC      ,frow,felem,1, evtlist%pdec(j), status)
                  call FTPCLJ(unit, COLNUM_X        ,frow,felem,1, evtlist%px(j), status)
                  call FTPCLJ(unit, COLNUM_Y        ,frow,felem,1, evtlist%py(j), status)
                  call FTPCLE(unit, COLNUM_ENERGY   ,frow,felem,1, evtlist%penergy(j), status)
                  call FTPCLE(unit, COLNUM_EV_WEIGHT,frow,felem,1, evtlist%pev_weight(j), status)
                  call FTPCLI(unit, COLNUM_RAWX     ,frow,felem,1, evtlist%prawx(j), status)
                  call FTPCLI(unit, COLNUM_RAWY     ,frow,felem,1, evtlist%prawy(j), status)
                  call FTPCLI(unit, COLNUM_SUBX     ,frow,felem,1, evtlist%psubx(j), status)
                  call FTPCLI(unit, COLNUM_SUBY     ,frow,felem,1, evtlist%psuby(j), status)
                  call FTPCLI(unit, COLNUM_PHA      ,frow,felem,1, evtlist%ppha(j), status)
                  call FTPCLI(unit, COLNUM_PAT_TYP  ,frow,felem,1, evtlist%ppat_typ(j), status)
                  call FTPCLB(unit, COLNUM_PAT_INF  ,frow,felem,1, evtlist%ppat_inf(j), status)
                  call FTPCLB(unit, COLNUM_TM_NR    ,frow,felem,1, evtlist%ptm_nr(j), status)
                  call FTPCLI(unit, COLNUM_FLAG     ,frow,felem,1, evtlist%pflag(j), status)
                  frow = frow + 1
               end if
            end do
         end if
       end associate
    end do

    !-----------------------------------
    ! append the gtilist as a new hdu
    if ( gtilist%is_valid ) then
      call srctool_gti_append_gtilist_hdu(unit, gtilist, ierr)
      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem appending GTI extension")
        return
      end if
    end if
    !-----------------------------------



    !-----------------------------------
    ! now close the fits file
    call FTCLOS(unit, status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU( unit, status)
    if (status /= 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

  end subroutine srctool_write_source_specific_evtlist
 !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  !write an set of source-specific event lists to file
  subroutine srctool_write_source_specific_evtlists(instlist, src, write_source, ierr)
    character(len=*), parameter :: subroutinename='srctool_write_source_specific_evtlist'

    type(instlist_type),  intent(in) :: instlist
    type(src_type), intent(in) :: src
    logical, intent(in) :: write_source   ! true=source, false=background
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: i

    ierr = ERROR_LEVEL_OK

    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        call srctool_write_source_specific_evtlist(instlist%inst(i), src, &
             instlist%inst(i)%src_gtilist, write_source, ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem writing source specific evtlist file")
          return
        end if
      end if
    end do

  end subroutine srctool_write_source_specific_evtlists
 !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !------------------------- END of EVENTLIST file handling-----------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------










  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- START of INTERVALLIST handling --------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_empty_intervallist(intervallist)
    type(intervallist_type), intent(inout)   :: intervallist
    if ( allocated(intervallist%pinterval)) deallocate(intervallist%pinterval)
    intervallist%num_rows = 0
    intervallist%num_tsteps = 0
    intervallist%total_time = 0.0d0
  end subroutine srctool_empty_intervallist
  !-------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------
  ! setup the intervallist based on the gtis defined within the instlist
  subroutine srctool_setup_intervallist(intervallist, instlist, tstep, ierr)
    character(len=*), parameter :: subroutinename='srctool_setup_intervallist'
    type(intervallist_type), intent(inout)   :: intervallist
    type(instlist_type), intent(in),target   :: instlist
    real(kind=real64), intent(in)        :: tstep
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32) :: i,j,k,m,n,status
    character(len=STR_MAX) :: str_buffer2
    real(kind=real64), allocatable, dimension(:) :: pt
    type(inst_type), pointer :: pinst

    ierr = ERROR_LEVEL_OK
    status = 0
    call wrschatty(subroutinename, "Setting up the integration intervals")

    ! make a list of all START and STOP times, then sort it, throwing away any
    ! matches. This list then defines the instances when any change of state occurs.
    n = 0
    ! first count the number of intervals
    do i=1,MAX_INST
      pinst => instlist%inst(i)
      if ( pinst%to_consider ) then
        do j=1,pinst%evtlistlist%num_evtlists
          n = n + 2*pinst%evtlistlist%evtlist(j)%gtilist%num_rows
        end do
      end if
    end do

    call wrschatty(subroutinename, "There are a total of " // trim(i2s(n)) // " interval bounds")
    ! now sort and filter for matches
    allocate(pt(n), STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return


    m = 1
    do i=1,MAX_INST
      pinst => instlist%inst(i)
      if ( pinst%to_consider ) then
        do j=1,pinst%evtlistlist%num_evtlists
          do k=1,pinst%evtlistlist%evtlist(j)%gtilist%num_rows
            pt(m) = pinst%evtlistlist%evtlist(j)%gtilist%pstart(k)
            m = m + 1
            pt(m) = pinst%evtlistlist%evtlist(j)%gtilist%pstop(k)
            m = m + 1
          end do
        end do
      end if
    end do

!    do i=1,n
!     write (str_buffer, '(A,I8,1x,F12.4)'), "DEBUG intervals unsorted: ", i, pt(i)
!     call wrschatty(subroutinename, str_buffer)
!    end do
    call insertion_sort(pt(:))

!    do i=1,n
!     write (str_buffer, '(A,I8,1x,F12.4)'), "DEBUG intervals   sorted: ", i, pt(i)
!     call wrschatty(subroutinename, str_buffer)
!    end do

    call  filter_unique_real64(pt,n)
!    do i=1,n
!     write (str_buffer, '(A,I8,1x,F12.3)'), "DEBUG intervals filtered:", i, pt(i)
!     call wrschatty(subroutinename, str_buffer)
!    end do

!    intervallist%num_rows = n - 1
    intervallist%num_rows = n - 1
    allocate(intervallist%pinterval(intervallist%num_rows), STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    do m=1,intervallist%num_rows
      intervallist%pinterval(m)%inst_flag = .false.
      ! set the start and stop times of all potential intervals
      intervallist%pinterval(m)%t1 = pt(m)
      intervallist%pinterval(m)%t2 = pt(m+1)
      intervallist%pinterval(m)%tmid = 0.5*(pt(m) + pt(m+1))
!      intervallist%pinterval(m)%tstep = &
!          instlist%inst(1)%evtlistlist%hduinfo%key_frametim%value * tstep * 1e-3 ! TODO check this is valid
      intervallist%pinterval(m)%tstep = tstep

      ! now see which instruments are valid in which interval
      do i=1,MAX_INST
        pinst => instlist%inst(i)
        if ( pinst%to_consider ) then
          do j=1,pinst%evtlistlist%num_evtlists
            intervallist%pinterval(m)%inst_flag(i) = &
                srctool_gti_test_if_in_gtilist(pinst%evtlistlist%evtlist(j)%gtilist, intervallist%pinterval(m)%tmid)
            if ( intervallist%pinterval(m)%inst_flag(i) ) then
              exit ! don't need to keep checking
            end if
          end do
        end if
      end do
    end do

    ! now count the number of tsteps and the total duration within valid intervals
    do m=1,intervallist%num_rows
      intervallist%pinterval(m)%ninst = 0
      do i=1,MAX_INST
        if ( intervallist%pinterval(m)%inst_flag(i) ) then
          intervallist%pinterval(m)%ninst =  intervallist%pinterval(m)%ninst + 1
        end if
      end do
      if ( intervallist%pinterval(m)%ninst >= 1 ) then
        if ( intervallist%pinterval(m)%tstep > 0.0 ) then
          intervallist%num_tsteps = intervallist%num_tsteps + &
              nint((intervallist%pinterval(m)%t2 - intervallist%pinterval(m)%t1)/intervallist%pinterval(m)%tstep)
        end if
        intervallist%total_time = intervallist%total_time + (intervallist%pinterval(m)%t2 - intervallist%pinterval(m)%t1)
      end if
    end do


    ! write (str_buffer, '(A,1x,A5,1x,A15,1x,A15,1x,A15,1x,A15,1x,A)') &
    !     "DEBUG intervallist :", "INDEX", &
    !     "START(s)", "STOP(s)", "DURATION(s)", "TSTEP(s)", "INSTS"
    ! call wrschatty(subroutinename, str_buffer)
    do m=1,intervallist%num_rows
      call convert_logical_array_to_string(intervallist%pinterval(m)%inst_flag(1:MAX_INST), str_buffer2, "")
      do i=1,MAX_INST
        pinst => instlist%inst(i)
        if ( .not. pinst%to_consider ) then
          str_buffer2(i:i) = '_'
        end if
      end do

      ! if ( intervallist%pinterval(m)%ninst >= 1 ) then
      !   write (str_buffer, '(A,1x,I5,1x,F15.3,1x,F15.3,1x,F15.3,1x,F15.5,1x,A)') &
      !       "DEBUG intervallist :", m, &
      !       intervallist%pinterval(m)%t1, &
      !       intervallist%pinterval(m)%t2, &
      !       intervallist%pinterval(m)%t2 -intervallist%pinterval(m)%t1, &
      !       intervallist%pinterval(m)%tstep, &
      !       trim(str_buffer2)
      !   call wrschatty(subroutinename, str_buffer)
      ! endif

      ! if ( m == 20 ) then
      !   write (str_buffer, '(A,1x,A)') "DEBUG intervallist :", "...........truncated..........."
      !   call wrschatty(subroutinename, str_buffer)
      !   exit
      ! endif

    end do

    ! write (str_buffer, '(A,1x,A5,1x,F15.3,1x,F15.3,1x,F15.3,1x,A15,1x,A)') &
    !     "DEBUG intervallist :", "TOTAL", &
    !     intervallist%pinterval(1)%t1, &
    !     intervallist%pinterval(intervallist%num_rows)%t2, &
    !     intervallist%total_time, &
    !     "-", &
    !     "-"
    !     call wrschatty(subroutinename, str_buffer)



    ! write (str_buffer, '(A,A,F15.3,A,I0)') &
    !     "DEBUG intervallist : TOTALS: ", &
    !     " total time=", intervallist%total_time, "s, total number of integration steps=", intervallist%num_tsteps
    ! call wrschatty(subroutinename, str_buffer)

    if(allocated(pt)) deallocate(pt)


  end subroutine srctool_setup_intervallist
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !----------------------------- END of INTERVALLIST handling --------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_set_spec_keys_from_arf(arf, spec, ierr)
    character(len=*), parameter :: subroutinename='srctool_set_spec_keys_from_arf'
    type(arf_type), intent(in)        :: arf
    type(spec_type), intent(inout)     :: spec
    integer(kind=int32), intent(out)  :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK
    call fitswrap_set_key(spec%key_telapse,  arf%key_telapse%value )
    call fitswrap_set_key(spec%key_ontime,   arf%key_ontime%value  )
    call fitswrap_set_key(spec%key_livetime, arf%key_livetime%value)
    call fitswrap_set_key(spec%key_exposure, arf%key_livetime%value)

    if ( spec%evt_class == EVT_CLASS_FLAG_SRC ) then
!      call fitswrap_set_key(spec%key_backscal, FITSWRAP_SET_KEY_BLANK, arf%backscale_src,  FITSWRAP_SET_KEY_BLANK)
!      call fitswrap_set_key(spec%key_backscal, arf%backratio)
      call fitswrap_set_key(spec%key_backscal, arf%backscale_src)
      call fitswrap_set_key(spec%key_geomregarea, arf%regarea_src)
      call fitswrap_set_key(spec%key_geomregmodarea, arf%regmodarea_src)

      call pathtools_extract_tail(trim(arf%hduinfo%filename), str_buffer)
      call fitswrap_set_key(spec%key_ancrfile, trim(str_buffer))
    else  if ( spec%evt_class == EVT_CLASS_FLAG_BACK ) then
!      call fitswrap_set_key(spec%key_backscal, FITSWRAP_SET_KEY_BLANK, arf%backscale_back, FITSWRAP_SET_KEY_BLANK)
!      call fitswrap_set_key(spec%key_backscal, 1.0d0)
      call fitswrap_set_key(spec%key_backscal, arf%backscale_back)
      call fitswrap_set_key(spec%key_geomregarea, arf%regarea_back)
      call fitswrap_set_key(spec%key_geomregmodarea, arf%regmodarea_back)
    end if

  end subroutine srctool_set_spec_keys_from_arf
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  subroutine srctool_set_speclist_keys_from_arflist(instlist, arflist, speclist, ierr)
    character(len=*), parameter :: subroutinename='srctool_set_speclist_keys_from_arflist'
    type(instlist_type), intent(in)   :: instlist
    type(arflist_type), intent(in)   :: arflist
    type(speclist_type), intent(inout)   :: speclist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)        :: i,j
! TODO reinstate this when the valid response filename can be read from the CALDB
!    type(ebounds_type), pointer :: pebounds
    ierr=ERROR_LEVEL_OK
    j = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then

        if ( j == 0 ) then
          j = i
        end if

        call srctool_set_spec_keys_from_arf(arflist%arf(i),speclist%src_spec(i),ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr(subroutinename,ierr,"Problem setting spec keywords")
          return
        end if
        call srctool_set_spec_keys_from_arf(arflist%arf(i),speclist%back_spec(i),ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr(subroutinename,ierr,"Problem setting spec keywords")
          return
        end if
! TODO reinstate this when the valid response filename can be read from the CALDB
!        ! if the ebounds file has been supplied then assume that it
!        ! is also a valid rmf file
!        call srctool_select_ebounds_pointer(instlist%inst(i),pebounds,ierr)
!        if ( ierr == ERROR_LEVEL_OK ) then
!          if ( pebounds%is_valid ) then
!            call fitswrap_set_key(speclist%src_spec(i)%key_respfile, FITSWRAP_SET_KEY_BLANK, &
!                trim(pebounds%hduinfo%filename),FITSWRAP_SET_KEY_BLANK)
!          end if
!        else
!          ierr =  ERROR_LEVEL_OK
!        end if

      end if
    end do

    ! now set up the summed spectrum keywords (using first valid instrument's values)
    i = 0
    call srctool_set_spec_keys_from_arf(arflist%arf(i),speclist%src_spec(i),ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr(subroutinename,ierr,"Problem setting spec keywords")
      return
    end if
    call srctool_set_spec_keys_from_arf(arflist%arf(i),speclist%back_spec(i),ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr(subroutinename,ierr,"Problem setting spec keywords")
      return
    end if
! TODO reinstate this when the valid response filename can be read from the CALDB
!    ! if the ebounds file has been supplied then assume that it
!    ! is also a valid rmf file
!    if ( j > 0 ) then
!      call srctool_select_ebounds_pointer(instlist%inst(j),pebounds,ierr)
!      if ( ierr == ERROR_LEVEL_OK ) then
!        if ( pebounds%is_valid ) then
!          call fitswrap_set_key(speclist%src_spec(i)%key_respfile, FITSWRAP_SET_KEY_BLANK, &
!              trim(pebounds%hduinfo%filename),FITSWRAP_SET_KEY_BLANK)
!        end if
!      else
!        ierr =  ERROR_LEVEL_OK
!      end if
!    end if


  end subroutine srctool_set_speclist_keys_from_arflist
  !-------------------------------------------------------------------------------







  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- START of FRACRSP handling -------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_increment_calc_srcgti_tstep(tstep, interval, time_increment, ierr)
    character(len=*), parameter :: subroutinename='srctool_increment_calc_srcgti_tstep'
    type(tstep_type),  intent(inout)  :: tstep
    type(interval_type), intent(in) :: interval
    real(kind=real64),  intent(in)  :: time_increment
    integer(kind=int32), intent(out)   :: ierr

    ierr=ERROR_LEVEL_OK

    tstep%tmin = tstep%tmax
    tstep%tmax = min(tstep%tmax + time_increment, interval%t2) !

    tstep%tmid = 0.5d0*(tstep%tmin + tstep%tmax)
    tstep%dt   = tstep%tmax - tstep%tmin

    tstep%dt_live = tstep%dt
    !-------------------------------------------------------
  end subroutine srctool_increment_calc_srcgti_tstep
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_increment_tstep(tstep, pdeadcor, time_increment, ierr)
    character(len=*), parameter :: subroutinename='srctool_increment_tstep'
    type(tstep_type),  intent(inout)  :: tstep
    type(deadcor_type),  intent(in), pointer :: pdeadcor
    real(kind=real64),  intent(in)  :: time_increment
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status, j

    ierr=ERROR_LEVEL_OK
    status = 0

    tstep%tmin = tstep%tmax
    tstep%tmax = tstep%tmax + time_increment !

    tstep%tmid = 0.5d0*(tstep%tmin + tstep%tmax)
    tstep%dt   = tstep%tmax - tstep%tmin

    !-------------------------------------------------------
    tstep%dt_live = tstep%dt
    if ( associated(pdeadcor) ) then
      if ( pdeadcor%is_valid ) then
        ! work out the deadcor in this time step - use the middle of the time bin
        j = 0
        call binary_search_real64(pdeadcor%pstart(:), tstep%tmid, j, status)
        if (status == BINARY_SEARCH_STATUS_ERROR ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Could not find valid DEADCOR value")
          return
        end if
        tstep%dt_live = tstep%dt * real(pdeadcor%pdeadc(j),real64)
      end if
    end if
    !-------------------------------------------------------
  end subroutine srctool_increment_tstep
  !-------------------------------------------------------------------------------

!  recursive subroutine debug_write_detailed_tstep_info(prefix, samplelist, tstep, lcbin, &
!      this_tstep, this_gti, write_flag)
!    character(len=*), parameter :: subroutinename='srctool_calc_fracrsp_debug'
!    character(len=*), intent(in) :: prefix
!    type(samplelist_type), intent(in) :: samplelist
!    type(tstep_type), intent(in) :: tstep
!    type(lcbin_type), intent(in) :: lcbin
!    real(kind=real64), intent(in) :: this_tstep
!    type(current_gti_type), intent(in) :: this_gti
!    integer(kind=int32), intent(in) :: write_flag
!
!    character(len=STR_MAX) :: str_buffer
!
!    !-------------------------------------------------------------------------!
!    !---------------------Detailed debug--------------------------------------!
!    if (IAND(write_flag,8) /= 0 ) then
!      call wrslog (subroutinename, "")
!    endif
!
!
!    if (IAND(write_flag,2) /= 0 ) then
!      write (str_buffer,&
!          '(A,A10,5(x,A12),x,A4,4(x,A12),x,A12,2(x,A15),x,A5,(x,A15),x,A15)' ) &
!          "#", trim(prefix), "ts%tmin", "ts%tmax", "ts%tmid", "ts%dt", "ts%dtlive", &
!          "lb%i", "lb%tstart", "lb%tstop", "lb%ttot", "lb%ttotlive", "this_tstep", &
!          "gti%tstart", "gti%tstop",  "inGTI", "t_tot_live", "val"
!      call wrslog (subroutinename, str_buffer)
!    endif
!
!
!    if (IAND(write_flag,1) /= 0 ) then
!      write (str_buffer,&
!          '(x,A10,3(x,F12.3),2(x,F12.6),x,I4,4(x,F12.3),x,F12.3,2(x,F15.3),x,A5,(x,F15.3),x,ES15.5E2)' ) &
!          trim(prefix), tstep%tmin, tstep%tmax, tstep%tmid, tstep%dt, tstep%dt_live, &
!          lcbin%i, lcbin%tstart, lcbin%tstop, lcbin%t_tot, lcbin%t_tot_live, &
!          this_tstep, this_gti%tstart, this_gti%tstop, logical2string(this_gti%is_valid), &
!          samplelist%t_tot_live, tstep%backscale_src
!      call wrslog (subroutinename, str_buffer)
!    endif
!
!    if (IAND(write_flag,4) /= 0 ) then
!      write (str_buffer,&
!          '(A,A10,5(x,A12),x,A4,4(x,A12),x,A12,2(x,A15),x,A5,(x,A15),x,A15)' ) &
!          "#", trim(prefix), "ts%tmin", "ts%tmax", "ts%tmid", "ts%dt", "ts%dtlive", &
!          "lb%i", "lb%tstart", "lb%tstop", "lb%ttot", "lb%ttotlive", "this_tstep", &
!          "gti%tstart", "gti%tstop",  "inGTI", "t_tot_live", "val"
!      call wrslog (subroutinename, str_buffer)
!    endif
!
!    if (IAND(write_flag,16) /= 0 ) then
!      call wrslog (subroutinename, "")
!    endif
!
!    !---------------------Detailed debug--------------------------------------!
!    !-------------------------------------------------------------------------!
!  end subroutine debug_write_detailed_tstep_info


!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!
!-!   !-------------------------------------------------------------------------!

  ! all steps done within a tstep
  ! assume that the time step: tsetp%tmin -> tstep%tmax is completely  within a single GTI and a single lcbin
  ! assume that the effmap has been updated already
  recursive subroutine srctool_calc_fracrsp_tstep(inst, instpar, psf, avign, &
      src, sl_src, sl_back, tstep, interp_eval_psf, ierr, lcbin )
    character(len=*), parameter :: subroutinename='srctool_calc_fracrsp_tstep'
    type(inst_type), intent(in), target   :: inst
    type(src_type), intent(in)      :: src
    type(tstep_type), intent(inout) :: tstep
    type(samplelist_type), intent(inout) :: sl_src
    type(samplelist_type), intent(inout) :: sl_back
    type(instpar_type), intent(in) :: instpar
    type(psf_type),   intent(in)  :: psf
    type(avign_type), intent(in)  :: avign
    type(linear_interp_real64_type), intent(inout) :: interp_eval_psf
    integer(kind=int32), intent(out)   :: ierr
    type(lcbin_type), intent(inout), optional :: lcbin

    integer(kind=int32)    :: e ! energy index
    integer(kind=int32)    :: status

    logical                :: srcreg_in_fov_flag

    ierr=ERROR_LEVEL_OK
    status = 0
    !-------------------------------------------------------
    ! calc the current ra0,dec0,roll at middle of tstep
    call srctool_att_get_current_att(inst%attlist, tstep%tmid, &
        sl_src%ra0, sl_src%dec0, sl_src%roll, ierr)
    if ( src%do_back ) then ! copy the above into the background samplelist
      sl_back%ra0  = sl_src%ra0
      sl_back%dec0 = sl_src%dec0
      sl_back%roll = sl_src%roll
    endif

    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error getting current attitude")
      return
    end if

    ! calc the (fractional) pixel coords of the src
    call srctool_att_radec2rawxy_a (instpar, &
        sl_src%ra0, sl_src%dec0, sl_src%roll, &
        sl_src%src_ra, sl_src%src_dec, &
        sl_src%src_rawx, sl_src%src_rawy, &
        sl_src%src_on_chip_flag )
    ! calc offaxis of the src in the focal plane
    sl_src%src_theta_arcmin = DEG2ARCMIN * &
        calc_ang_sep(sl_src%ra0, sl_src%dec0, &
        sl_src%src_ra, sl_src%src_dec)
    !-------------------------------------------------------

    ! we want to know if any src sample point could possibly lie within the FOV
    ! so must be very conservative.
    ! FIXME FIXME! JSS inserted factor of 2x here as could be larger than fov, if FoV flag is switched off
    if ( sl_src%src_on_chip_flag .or. &
        (sl_src%src_theta_arcmin < &
        sl_src%largest_sample_radius_deg*DEG2ARCMIN + 2*instpar%fov_radius_arcmin )) then

      ! calc azimuthal angle of the src in the focal plane
      sl_src%src_phi = &
          atan2d((sl_src%src_rawy - instpar%oa_rawy)*instpar%pixscale_y, &
          (sl_src%src_rawx - instpar%oa_rawx)*instpar%pixscale_x)

      !----------------------------------------------------------------------------------------
      ! Get the detector coordinates of the sample points - both src and back
      call srctool_att_radec2rawxy_as2_fast (instpar, &
          sl_src%ra0, &
          sl_src%dec0, &
          sl_src%roll, &
          sl_src%pra(:),   &
          sl_src%pcosdec(:), &
          sl_src%psindec(:), &
          sl_src%prawx(:), &
          sl_src%prawy(:), &
          sl_src%prawx_int(:), &
          sl_src%prawy_int(:), &
          sl_src%pon_chip_flag(:) )
      if ( src%do_back ) then
        call srctool_att_radec2rawxy_as2_fast (instpar, &
            sl_back%ra0, &
            sl_back%dec0, &
            sl_back%roll, &
            sl_back%pra(:),   &
            sl_back%pcosdec(:), &
            sl_back%psindec(:), &
            sl_back%prawx(:), &
            sl_back%prawy(:), &
            sl_back%prawx_int(:), &
            sl_back%prawy_int(:), &
            sl_back%pon_chip_flag(:) )
      endif
      !----------------------------------------------------------------------------------------

      !----------------------------------------------------------------------------------------
      ! Get the detector effmap values at the locations of the sample points at the PSF energies
      ! i.e. return 1.0 for full sensitivity good pixels, and 0.0 for completely bad pixels
      ! this changes the inst%samplelist_xxx%peffmap(:) arrays
      call srctool_badpix_get_effmap(inst%effmap, sl_src)
      if ( src%do_back ) then
        call srctool_badpix_get_effmap(inst%effmap, sl_back)
      end if
      !----------------------------------------------------------------------------------------


      !-------------------------------------------
      ! count the number of samples that landed in good pixels, and add up their areas
      ! and multiply by the live time interval
      ! This implementation is energy band independent
      tstep%backscale_src  = sum(sl_src%peffmap(:))

      if ( tstep%backscale_src > 0.0d0 ) then
        srcreg_in_fov_flag = .true.
      else
        srcreg_in_fov_flag = .false.
      end if

      tstep%backscale_src  = tstep%backscale_src * sl_src%area_per_point * tstep%dt_live
      tstep%backscale_back = 0.0d0
      if ( src%do_back ) then
        tstep%backscale_back = sl_back%area_per_point * sum(sl_back%peffmap(:))* tstep%dt_live
      endif

      if ( tstep%backscale_back > 0.0d0 ) then
        tstep%backratio =  ( tstep%backscale_src / tstep%backscale_back) * tstep%dt_live !* frac_tstep_in_gti
      else
        tstep%backratio = 0.0d0
      end if
      !-------------------------------------------


      !----------------------------------------------------------------------------------------
      ! only proceed into further calculations if we have at least one src sample lying on the CCD
      if ( srcreg_in_fov_flag ) then

        !-------------------------------------------
        ! work out the vignetting at each sample point and energy
        ! - depending on psf method used
        if ( psf%class == SRCTOOL_PSF_CLASS_NONE ) then
          ! Compute energy dependent vignetting independently at each sample point
          call avign_calc_vign_at_xye(avign, &
              sl_src%prawx(:), &
              sl_src%prawy(:), &
              sl_src%peval_e(:), &
              sl_src%pvign(:,:), &
              ierr)
        else
          ! Consider only a single vignetting value (per energy):
          ! First get the vignetting values at the source location
          ! using the first sample point as the destination array,
          ! then duplicate the at-source vignetting for remaining sample points
          call avign_calc_vign_at_xye(avign, &
              sl_src%src_rawx, &
              sl_src%src_rawy, &
              sl_src%peval_e(:), &
              sl_src%pvign(1,:), ierr)
          ! we can skip this part if we are careful later
          !                  if ( sl_src%num_points >= 2 ) then
          !                    sl_src%pvign(2:sl_src%num_points,:) = sl_src%pvign(1,:)
          !                  endif
        endif

        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr(subroutinename,ierr,"Error calculating vignetting")
          return
        end if
        !-------------------------------------------


        if ( psf%class == SRCTOOL_PSF_CLASS_NONE ) then
          ! what we want is the sum over points of the product of effmap * model * vignetting
          ! divided by number of sample points
          do e=1,sl_src%num_eval_e
            sl_src%pfracarea(e) = sum(sl_src%peffmap(:)*sl_src%pvign(:,e)*sl_src%pmodel(:)) &
                / REAL(MAX(1,sl_src%num_points),real64)

            sl_src%eval_sum_psf(e) = sl_src%eval_sum_psf(e) + &
                 sum(sl_src%peffmap(:)*sl_src%pmodel(:)) / &
                 REAL(MAX(1,sl_src%num_points),real64) * tstep%dt_live
            sl_src%eval_sum_vign(e) = sl_src%eval_sum_vign(e) + &
                 sum(sl_src%pvign(:,e)*sl_src%pmodel(:)) / &
                 REAL(MAX(1,sl_src%num_points),real64) * tstep%dt_live
          end do
        else
          ! compute the sum of the PSF (convolved with the source extent model) that
          ! falls within the source extraction region

          ! Get the summed PSFxExtent values at the locations of the src
          ! sample points at the native energies of the calibration psf maps
          call srctool_psf_get_psf_values(inst, psf, sl_src)

          ! calc the summed psf measurements at the list of unique calibration evaluation energies
          ! using the pre-prepared interpolation struct
          call evaluate_linear_interp(interp_eval_psf, sl_src%ppsfsum(:), status)
          if ( status /= 0 ) then
            ierr = ERROR_LEVEL_SEVERE
            call wrserr (subroutinename, ierr, "Error evaluating interpolator - psf")
            return
          end if

          ! multiply by the source vignetting values and the time step here -
          sl_src%pfracarea(:) = sl_src%pvign(1,:) * interp_eval_psf%y_eval(:)
          ! keep track of PSF and vignetting contributions
          sl_src%eval_sum_psf = sl_src%eval_sum_psf + interp_eval_psf%y_eval(:)*tstep%dt_live
          sl_src%eval_sum_vign = sl_src%eval_sum_vign + sl_src%pvign(1,:)*tstep%dt_live
        endif



        !----------------------------------------------------------------------------------------
        if ( IAND(src%todo_code,TODO_CODE_SAMPLE) /= TODO_CODE_NULL) then
          call srctool_samplelist_accumulate_expmap (sl_src, tstep )
          if ( src%do_back ) then
            call srctool_samplelist_accumulate_expmap (sl_back, tstep )
          endif
        endif
        !----------------------------------------------------------------------------------------


        ! now calc the contribution to the frac resp - time weighted
        sl_src%eval_values(:) = sl_src%pfracarea(:) * tstep%dt_live


        ! add contribution to running totals of arf and lc
        sl_src%eval_values_arf(:) = sl_src%eval_values_arf(:) + sl_src%eval_values(:)

        if ( present(lcbin) ) then
          sl_src%eval_values_lc(:) = &
              sl_src%eval_values_lc(:) + sl_src%eval_values(:)

          ! add the contribution to the mean offaxis during this lc bin
          lcbin%offaxis =  &
              lcbin%offaxis + sl_src%src_theta_arcmin * tstep%dt_live
        end if


        ! deal with the backscale/background ratio stuff
        sl_src%backscale =  sl_src%backscale + tstep%backscale_src
        if ( src%do_back ) then
          sl_back%backscale =  sl_back%backscale + tstep%backscale_back
        endif

        if (  present(lcbin) ) then
          lcbin%backratio   =  lcbin%backratio  + tstep%backratio
        end if

        ! keep track of the total time
        sl_src%t_tot       = sl_src%t_tot      + tstep%dt      !
        sl_src%t_tot_live  = sl_src%t_tot_live + tstep%dt_live !
        if ( src%do_back ) then
          sl_back%t_tot       = sl_back%t_tot      + tstep%dt
          sl_back%t_tot_live  = sl_back%t_tot_live + tstep%dt_live
        endif
        if ( present(lcbin) ) then
          lcbin%t_tot      = lcbin%t_tot      + tstep%dt
          lcbin%t_tot_live = lcbin%t_tot_live + tstep%dt_live
        end if
      end if  ! if at least one sample point was inside a live FOV
      !----------------------------------------------------------------------------------------
    end if  ! if a conservative check on source being inside FOV was met


  end subroutine srctool_calc_fracrsp_tstep
  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! Notes on Time steps and intervals, gtis and lc bins
  ! the integration is done over small time steps (tstep)
  ! the  time steps have a default length given on the command line
  ! In order to increase accuracy, we dynamically adjust (shorten) the
  ! length of time steps if their default size would make them overlap with
  ! the boundary of an interval, a GTI, or a lightcurve bin.
  ! In order to increase efficiency, we dynamically adjust (lengthen) the
  ! time steps to take us to the beginning of a valid GTI

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_init_lcbin(lcbin)
    type(lcbin_type), intent(out)  :: lcbin
    lcbin%i = -1
    lcbin%tstart      = -huge(lcbin%tstart)
    lcbin%tstop       = -huge(lcbin%tstop)
    lcbin%t_tot       = 0.0d0
    lcbin%t_tot_live  = 0.0d0
    lcbin%backratio   = 0.0d0
    lcbin%offaxis     = 0.0d0

  end subroutine srctool_init_lcbin
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  recursive subroutine srctool_finish_lcbin(lc, lcbin, samplelist, interp_vignpsf_lc, ierr)
    character(len=*), parameter :: subroutinename='srctool_finish_lcbin'
    type(lc_type),  intent(inout) :: lc
    type(lcbin_type), intent(inout)  :: lcbin
    type(samplelist_type), intent(inout)   :: samplelist
    type(linear_interp_real64_type), intent(inout) :: interp_vignpsf_lc
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32)    :: status
    integer(kind=int32)    :: e , ei

    ierr=ERROR_LEVEL_OK
    status = 0

    if ( lcbin%i < 1 ) then  ! this lcbin is not valid, so get out quickly

      samplelist%eval_values_lc(:) = 0.0d0  ! TODO move this into the lcbin type?
      call srctool_init_lcbin (lcbin)
      return
    endif

    ! now do any actions to finish off this bin
    ! calc the fracrsp for this tbin
    ! interpolate from the evaluation points
    call evaluate_linear_interp(interp_vignpsf_lc, samplelist%eval_values_lc(:), status)
    if ( status /= 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Error evaluating interpolator - lc")
      return
    end if

    if ( lc%ptimedel(lcbin%i) > 0.0d0 ) then
      lc%pfractime(lcbin%i) = REAL(lcbin%t_tot_live / lc%ptimedel(lcbin%i), real32)
    else
      lc%pfractime(lcbin%i) = -1.0e0
    end if

    if ( lcbin%t_tot_live > 0.0d0 ) then
      lc%pbackratio(lcbin%i) = lcbin%backratio / lcbin%t_tot_live
      lc%poffaxis(lcbin%i)   = REAL(ARCMIN2DEG *  lcbin%offaxis / lcbin%t_tot_live, real32)
    else
      lc%pbackratio(lcbin%i) = 0.0d0
      lc%poffaxis(lcbin%i)   = NULLVAL_REAL32
    end if

    ! note that we eventually want to just write the fracexp
    !... i.e. the product of fracarea and fractime
    if ( lcbin%t_tot_live > 0.0d0 ) then
      do e=1,lc%num_ebands
        ei = ((lcbin%i-1)*lc%num_ebands) + e
        lc%pfracarea(ei) = &
            REAL(SUM(interp_vignpsf_lc%y_eval(:)*lc%pint_weights(e,:))/lcbin%t_tot_live)
      end do
    end if

    ! the following calculates the fracrsp = fracarea*fractime
    if ( lc%ptimedel(lcbin%i) > 0.0d0 ) then
      do e=1,lc%num_ebands
        ei = ((lcbin%i-1)*lc%num_ebands) + e
        lc%pfracexp(ei) = &
            REAL(SUM(interp_vignpsf_lc%y_eval(:)*lc%pint_weights(e,:))/lc%ptimedel(lcbin%i))
      end do
    end if


    ! reset the values for the next bin
    samplelist%eval_values_lc(:) = 0.0d0  ! TODO move this into the lcbin type?
    call srctool_init_lcbin (lcbin)

  end subroutine srctool_finish_lcbin
  !-------------------------------------------------------------------------------

  ! Compute a suitable tstep to use.
  ! interval [tmin:tstop] is spanned by an inger number of tsteps
  ! tstep has maximum length = nominal_tstep
  ! the tstep_tolerance parameter allows for numerical innacuracies
  recursive subroutine srctool_calc_integration_steps(tmin, tmax, nominal_tstep, tstep, nsteps, tstep_tolerance)
    character(len=*), parameter :: subroutinename='srctool_calc_integration_steps'
    real(kind=real64), intent(in) :: tmin, tmax, nominal_tstep
    real(kind=real64), intent(out) :: tstep
    real(kind=real64), optional, intent(out) :: tstep_tolerance
    integer(kind=int32), intent(out) :: nsteps
    nsteps = MAX(1,CEILING((tmax - tmin)/nominal_tstep, int32))
    tstep = (tmax - tmin)/REAL(nsteps,real64)
    if ( present(tstep_tolerance) ) then
      tstep_tolerance = MAX(1.0d-6*tstep,10.0d0*ABS((tmin+(nsteps*tstep) - tmax)))
    endif
  end subroutine srctool_calc_integration_steps

  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------
  recursive subroutine srctool_calc_fracrsp(instlist, inst, arf, lc, src, ierr)
    character(len=*), parameter :: subroutinename='srctool_calc_fracrsp'

    type(instlist_type), intent(in) :: instlist
    type(inst_type), intent(inout), target   :: inst
    !type(intervallist_type), intent(in), target :: intervallist
    type(arf_type), intent(inout)      :: arf
    type(lc_type),  intent(inout)      :: lc
    type(src_type), intent(inout)      :: src
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32)    :: status
    character(len=STR_MAX) :: str_buffer

    type(interval_type),    pointer :: pinterval   ! pointers to data sub-structures within "inst","intervallist"
    type(avign_type),       pointer :: pavign
    type(badpix_type),      pointer :: pbadpix
    type(psf_type),         pointer :: ppsf
    type(arf_type),         pointer :: prefarf
    type(detmap_type),      pointer :: pdetmap
    type(instpar_type),     pointer :: pinstpar
    type(deadcor_type),     pointer :: pdeadcor
    type(gtilist_type),     pointer :: pgtilist

    type(samplelist_type), pointer :: sl_src
    type(samplelist_type), pointer :: sl_back

    type(tstep_type) :: tstep   ! working data structures
    type(lcbin_type) :: lcbin
    type(psf_type), target   :: psf_convolved

    ! this is used when interpolating PSF values onto the eval energy grid
    type(linear_interp_real64_type) :: interp_eval_psf
    type(linear_interp_real64_type) :: interp_vignpsf_lc

    real(kind=real64)  :: this_tstep
    real(kind=real64)  :: tend
    integer(kind=int32)  :: nsteps
    integer(kind=int32)  :: i,j
    integer(kind=int32)  :: lc_i, next_lc_i

!    type(current_gti_type) :: debug_this_gti
    logical            :: finish_gti_after_loop
    logical            :: finish_lcbin_after_loop


    logical           , parameter :: emit_detailed_debug = .false.
    real(kind=real64), parameter  :: time_tolerance = 1.0d-6  !=1us, used to avoid floating point rounding problems

    ierr=ERROR_LEVEL_OK
    status = 0
!    this_tstep = 0.0d0
    nullify(sl_src)
    nullify(sl_back)


    call  srctool_psf_empty_psf(psf_convolved)

    call srctool_misc_select_cal_pointer(inst,pavign,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_cal_pointer(inst,pbadpix,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_cal_pointer(inst,ppsf,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_cal_pointer(inst,prefarf,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_cal_pointer(inst,pdetmap,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_cal_pointer(inst,pinstpar,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_cal_pointer(inst,pdeadcor,ierr)
    if ( ierr >= ERROR_LEVEL_SEVERE ) return
    ierr = ERROR_LEVEL_OK

    !----------------------------------------------------------------------------------------
    !----------------------------------------------------------------------------------------
    !----------------------------------------------------------------------------------------
    ! setup everything that needs to be called once per call

    !----------------------------------------------------------------------------------------
    ! Now prepare the folded PSFs for extended sources
    ! Only do this for extended sources
    ! call wrschatty (subroutinename, "Convolving PSF model with source extent model: " // EXTTYPE_CODE_STR(src%extent%code))
    if      (src%extent%code == EXTTYPE_CODE_POINT ) then
      ! srctool will just use the original psf model
    else if (ppsf%class == SRCTOOL_PSF_CLASS_NONE ) then
      ! srctool will just use the original psf model
    else

      !------------------------------------------------------------------------------!
      !------------------------------Debug output------------------------------------!
      if ( IAND(src%todo_code,TODO_CODE_TEST) /= TODO_CODE_NULL) then
        do j=1,ppsf%psfmaplist%num_psfmaps
          write (str_buffer, '(A,I0.3,A,I0.3,A)') "!test_psfmap_orig_", src%index, '_', j, ".fits"
          call srctool_psf_write_psfmap(str_buffer, ppsf%psfmaplist%psfmap(j), ierr)
          if ( ierr /= ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Problem writing PSF model to file")
            return
          end if
        end do
      endif
      !------------------------------Debug output------------------------------------!
      !------------------------------------------------------------------------------!

      ! convolve the instrumental PSF model with the (circular) beta model
      !call wrschatty (subroutinename, "Convolving PSF model with source extent model: BETA")
      call srctool_psf_convolve_psf_with_extent(instlist, ppsf, psf_convolved, src%extent, ierr)
      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem convolving PSF model with source extent model")
        return
      end if
      ppsf => psf_convolved ! srctool will use the convolved psf model

      !------------------------------------------------------------------------------!
      !------------------------------Debug output------------------------------------!
      if ( IAND(src%todo_code,TODO_CODE_TEST) /= TODO_CODE_NULL) then
        do j=1,psf_convolved%psfmaplist%num_psfmaps
          write (str_buffer, '(A,I0.3,A,I0.3,A)') "!test_psfmap_conv_", src%index, '_', j, ".fits"
          call srctool_psf_write_psfmap(str_buffer, psf_convolved%psfmaplist%psfmap(j), ierr)
           if ( ierr /= ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Problem writing PSF model to file")
            return
          end if
        end do
      endif
      !------------------------------Debug output------------------------------------!
      !------------------------------------------------------------------------------!
    end if
    !----------------------------------------------------------------------------------------

    ! do j=1,ppsf%psfmaplist%num_psfmaps
    !   write (str_buffer, '(A,I0,A,I2.2,A,F0.1,A,F0.1,A)') "!psfmap_orig_tm", inst%index, '_', j, '_', ppsf%psfmaplist%psfmap(j)%energy_eff, '_', ppsf%psfmaplist%psfmap(j)%theta_eff, ".fits"
    !   call srctool_psf_write_psfmap(str_buffer, ppsf%psfmaplist%psfmap(j), ierr)
    !   if ( ierr /= ERROR_LEVEL_OK ) then
    !     call wrserr (subroutinename, ierr, "Problem writing PSF model to file")
    !     return
    !   end if
    ! end do

    !----------------------------------------------------------------------------------------
    ! Setup the list of spatial sample points, one for src, one for back (if needed)
    sl_src => inst%samplelist_src
    call srctool_setup_samplelist(inst, src, sl_src, ppsf, EVT_CLASS_FLAG_SRC, ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem setting up src samplelist")
      return
    end if
    if ( src%do_back ) then
      sl_back => inst%samplelist_back
      call srctool_setup_samplelist(inst, src, sl_back, ppsf, EVT_CLASS_FLAG_BACK, ierr)
      if ( ierr /=  ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem setting up back samplelist")
        return
      end if
      ! write(str_buffer, '(A,A,A,I0, A,I0,A,I0,A,I0,A, A,F10.3,A,F10.3,A, A,F8.6,A,F8.6,A,I0)') &
      !     "DEBUG: Sample pattern inst=", trim(inst%name), " src=", src%index, &
      !     " Np(src,back|num_e)=(", &
      !     sl_src%num_points, ",", sl_back%num_points,"|", sl_src%num_eval_e, ")", &
      !     " Area/sample(src,back)=(", &
      !     sl_src%area_per_point_sqarcsec,",",sl_back%area_per_point_sqarcsec,")arcsec^2", &
      !     " Weight/sample(src,back)=(", &
      !     sl_src%weight_per_point, ",", sl_back%weight_per_point, ") Ntsteps=", intervallist%num_tsteps
      ! call wrschatty(subroutinename, str_buffer)
    else
      ! write(str_buffer, '(A,A,A,I0, A,I0,A,I0,A, A,F10.3,A, A,F8.6,A,I0)') &
      !     "DEBUG: Sample pattern inst=", trim(inst%name), " src=", src%index, &
      !     " Np(src|num_e)=(", &
      !     sl_src%num_points, "|", sl_src%num_eval_e, ")", &
      !     " Area/sample(src)=(", &
      !     sl_src%area_per_point*SQDEG2SQARCSEC,")arcsec^2", &
      !     " Weight/sample(src)=(", &
      !     sl_src%weight_per_point, ") Ntsteps=", intervallist%num_tsteps
      ! call wrschatty(subroutinename, str_buffer)

    endif
    !----------------------------------------------------------------------------------------

    !----------------------------------------------------------------------------------------
    ! accumulate images from the source and back events on the same pixel scale as the samplelist
    ! Do this only when actually requested
    if ( IAND(src%todo_code,TODO_CODE_SAMPLE) /= TODO_CODE_NULL) then
      call srctool_samplelist_accumulate_image(inst, sl_src, ierr)
      if ( ierr /=  ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem accumulating src image")
        return
      end if
      if ( src%do_back ) then
        call srctool_samplelist_accumulate_image(inst, sl_back, ierr)
        if ( ierr /=  ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem accumulating back image")
          return
        end if
      endif
    endif
    !----------------------------------------------------------------------------------------



    !----------------------------------------------------------------------------------------
    ! initialise values which will be summed over the time steps
    ! and get the eval linear interpolator ready
    !----------------------------------------------------------------------------------------

    !----------------------------------------------------------------------------------------
    ! this will be used to interpolate from the native psf energy grid to
    ! the intermediate evaluation energy grid
    call empty_linear_interp (interp_eval_psf)
    if ( ppsf%class /= SRCTOOL_PSF_CLASS_NONE ) then
      call setup_linear_interp (interp_eval_psf, &
          sl_src%ppsf_e(:), &   ! input energy grid
          sl_src%peval_e(:), &  ! target energy grid
          status)
      if ( status /= 0 ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Error setting up interpolators - psf")
        return
      end if
    endif
    !-------
    !----------------------------------------------------------------------------------------


    !----------------------------------------------------------------------------------------
    ! get the lc linear interpolator ready
    ! this will be used to interpolate from the intermediate evaluation grid to the lc bins
    call empty_linear_interp (interp_vignpsf_lc)
    if ( lc%to_consider ) then
      call setup_linear_interp (interp_vignpsf_lc, &
                                sl_src%peval_e(:), &             ! input energy grid
                                lc%pint_e(1:lc%int_nenergies), & ! target energy grid
                                status )
      if ( status /= 0 ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Error setting up interpolators - lc")
        return
      end if
    end if
    !----------------------------------------------------------------------------------------


    pgtilist => inst%src_gtilist

    ! reset the effmap for the start of the first GTI
    call srctool_badpix_calculate_effmap(inst%effmap, pdetmap, pgtilist%pstart(1), &
         inst%use_fov_mask, ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return

    !----------------------------------------------------------------------------------------
    !----------------------------------------------------------------------------------------
    !--------------------------Start of Main loop--------------------------------------------
    !----------------------------------------------------------------------------------------
    !----------------------------------------------------------------------------------------

    if ( pgtilist%num_rows == 1 ) then
      if (pgtilist%pstop(1)-pgtilist%pstart(1) < time_tolerance) then
        call wrserr(subroutinename, ERROR_LEVEL_WARNING, "zero or infinitesimal GTI")
      endif
    endif


    ! A) For each src_gti i
    !  1) find index of lcbin that contains tstep%tmin = pgtilist%pstart(i)
    !  2) if this is is different from old lcbin then finish off old one and initialise new one
    !  3) for each loop of an infinite while loop
    !    i) determine what comes first (and set flag), end of gti or end of lcbin:  tend=MIN(this_gti%tstop, lcbin%tstop)
    !    ii) count number (nsteps) and length (this_tstep) of integration steps until tend
    !    iii) do i=1,nsteps
    !         a) integrate over step: srctool_calc_fracrsp_tstep()
    !    iv) if lcbin has finished, then finish off old one and initialise new lcbin
    !     v) if gti has finished, then move to next one
    ! B) finish lcbin if it is not already finished
    call srctool_init_lcbin (lcbin)
    finish_lcbin_after_loop = .false.
    finish_gti_after_loop   = .false.

    do i=1,pgtilist%num_rows ! for each (source) GTI
!      debug_this_gti%is_valid = .true.
!      debug_this_gti%tstart = pgtilist%pstart(i)
!      debug_this_gti%tstop = pgtilist%pstop(i)

      tstep%tmax = pgtilist%pstart(i)
!      tstep%tmin = 0.0d0 ! DEBUG
!      tstep%tmid = 0.0d0

      if ( lc%to_consider ) then
        call srctool_lc_get_lc_index(lc, tstep%tmax+time_tolerance, lc_i)

        ! if old bin was valid and new one is different (or invalid) then finish off lcbin
        if ( lcbin%i >= 1 .and. lc_i /= lcbin%i ) then
          call srctool_finish_lcbin (lc, lcbin, sl_src, interp_vignpsf_lc, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Error finishing lcbin")
            return
          endif
        endif

        if ( lc_i >= 1 .and. lc_i /= lcbin%i ) then ! if now in valid lcbin
          ! copy new lc limits into lcbin struct
          lcbin%i      = lc_i
          lcbin%tstart = lc%ptmin(lc_i)
          lcbin%tstop  = lc%ptmax(lc_i)
        endif

      endif


      !----------------------------------------------------------------------------------------
      !----------------------------------------------------------------------------------------
      !--------------------------Start of Middle loop------------------------------------------
      !----------------------------------------------------------------------------------------
      do while ( .true. )
        tend = pgtilist%pstop(i)
        finish_gti_after_loop   = .true.
        if ( lc%to_consider ) then
          if ( lcbin%i >= 1 ) then
            if ( lcbin%tstop < tend ) then
              ! lightcurve bin ends before end of GTI (Case A)
              tend = lcbin%tstop
              finish_gti_after_loop   = .false.
              finish_lcbin_after_loop = .true.
            else if ( eq_real(lcbin%tstop,tend)) then
              ! both lightcurve and GTI end at same time (Case B)
              finish_gti_after_loop   = .true.
              finish_lcbin_after_loop = .true.
            else
              ! just part of the lc bin in the GTI (Case C)
              finish_lcbin_after_loop = .false.
            end if
          else
            ! if we are not currently in an lcbin then see if the start
            ! of the next one is before the end of this gti
            call srctool_lc_get_next_lc_index(lc, tstep%tmax+time_tolerance, next_lc_i)
            if ( next_lc_i >= 1 ) then
              if ( lc%ptmin(next_lc_i) < tend ) then
                ! There's a lcbin within this GTI, but not immediately (Case D)
                tend = lc%ptmin(next_lc_i)
                finish_gti_after_loop   = .false.
                finish_lcbin_after_loop = .true.
              else
                ! lcbin is not before end of current GTI (Case E)
              end if
            else
              ! out of range of lcbins (case F)
            end if ! next_lc_i >= 1
          end if ! lcbin%i >= 1
        end if ! lc%to_consider

        call srctool_calc_integration_steps(tstep%tmax, tend, inst%tstep, this_tstep, nsteps)
        if ( this_tstep <= 0.0d0 ) then
          ierr = ERROR_LEVEL_WARNING
          call wrserr (subroutinename, ierr, "Error calculating tstep")
          return
        end if

        !----------------------------------------------------------------------------------------
        !--------------------------Start of Inner loop-------------------------------------------

        do j=1,nsteps
          ! move the integration step forward
          call srctool_increment_tstep(tstep, pdeadcor, this_tstep, ierr )

          ! check if any instrument params have changed
          ! 1) Check to see if we need to update the effmap (badpixels)
          if ( tstep%tmid >= inst%effmap%time_of_next_update ) then
            call srctool_badpix_calculate_effmap(inst%effmap, pdetmap, tstep%tmid, &
                 inst%use_fov_mask, ierr)
            if ( ierr /=  ERROR_LEVEL_OK ) return
          end if
          ! 2) in future can also check for time-variable ARF, RMF etc

          ! do the integration for this tstep - only send an lcbin if necessary
          if ( lc%to_consider .and. (lcbin%i >= 1) ) then
            call srctool_calc_fracrsp_tstep(inst, pinstpar, ppsf, pavign, &
                src, sl_src, sl_back, tstep,  &
                interp_eval_psf, ierr, lcbin )
          else
            call srctool_calc_fracrsp_tstep(inst, pinstpar, ppsf, pavign, &
                src, sl_src, sl_back, tstep,  &
                interp_eval_psf, ierr)
          end if
        end do
        !--------------------------End of Inner loop------------------------------------------!
        !-------------------------------------------------------------------------------------!


        ! take actions if we finished a lc bin
        if ( lc%to_consider ) then
          if ( finish_lcbin_after_loop ) then
            if ( lcbin%i >= 1 ) then

              ! check that it is safe+correct to move in to next LC bin
              next_lc_i = -1
              if ( lcbin%i < lc%num_rows ) then
                next_lc_i = lcbin%i + 1
                if ( lc%ptmin(next_lc_i) - lcbin%tstop > time_tolerance ) then
                  next_lc_i = -1
                endif
              endif

              ! finish the old bin
              call srctool_finish_lcbin (lc, lcbin, sl_src, interp_vignpsf_lc, ierr)
              if ( ierr /= ERROR_LEVEL_OK ) then
                call wrserr (subroutinename, ierr, "Error finishing lcbin")
                return
              endif

              ! copy new lc limits into lcbin struct
              if (next_lc_i >= 1) then
                lcbin%i      = next_lc_i
                lcbin%tstart = lc%ptmin(next_lc_i)
                lcbin%tstop  = lc%ptmax(next_lc_i)
              endif

            else
              ! asked to finish an lc bin (see Case D above), but we're not in an lc bin
              ! therefore we need to start a new one
              call srctool_lc_get_lc_index(lc, tstep%tmax+time_tolerance, next_lc_i)

              if ( next_lc_i >= 1 ) then ! if now in valid lcbin (should always be true)
                ! copy new lc limits into lcbin struct
                call srctool_init_lcbin(lcbin)
                lcbin%i      = next_lc_i
                lcbin%tstart = lc%ptmin(next_lc_i)
                lcbin%tstop  = lc%ptmax(next_lc_i)
              end if

            end if ! lcbin%i >= 1
          end if ! finish_lcbin_after loop
        end if ! to_consider

        if ( finish_gti_after_loop ) then
          exit  ! from the infinite  do while (.true.) loop
        endif

      end do
      !--------------------------------------------------------------------------------------!
      !--------------------------End of Middle loop------------------------------------------!
      !--------------------------------------------------------------------------------------!
      !--------------------------------------------------------------------------------------!

    end do
    !----------------------------------------------------------------------------------------!
    !----------------------------------------------------------------------------------------!
    !--------------------------End of Main loop----------------------------------------------!
    !----------------------------------------------------------------------------------------!
    !----------------------------------------------------------------------------------------!

    ! finish off everything that needs to be done once per call

    if ( lc%to_consider ) then
      ! if old bin was valid and unfinished then finish off lcbin
      if ( lcbin%i >= 1 ) then
        call srctool_finish_lcbin (lc, lcbin, sl_src, interp_vignpsf_lc, ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error finishing lcbin")
          return
        endif
      endif
    endif

    arf%backscale_src = sl_src%backscale
    arf%regarea_src = sl_src%num_points_region * sl_src%area_per_point
    arf%regmodarea_src = sl_src%num_points_region_model * sl_src%area_per_point
    if ( src%do_back ) then
      arf%backscale_back = sl_back%backscale
      arf%regarea_back = sl_back%num_points_region * sl_back%area_per_point
      arf%regmodarea_back = sl_back%num_points_region_model * sl_back%area_per_point
    endif


    call srctool_samplelist_compute_arf (sl_src, prefarf, arf, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error computing final ARF")
      return
    end if


    !----------------------------------------------------------------------------------------
    ! do any further processing of light curve results
    if ( lc%to_consider ) then
      ! set the exposure time values in the lc struct
!      call fitswrap_set_key(lc%key_telapse,  intervallist%total_time     )  ! TODO - implement this
      call fitswrap_set_key(lc%key_telapse,  sl_src%t_tot     )
      call fitswrap_set_key(lc%key_ontime,   sl_src%t_tot     )
      call fitswrap_set_key(lc%key_livetime, sl_src%t_tot_live)
      call fitswrap_set_key(lc%key_exposure, sl_src%t_tot_live)
    end if
    !----------------------------------------------------------------------------------------


    ! tidy up
    call empty_linear_interp (interp_eval_psf)
    call empty_linear_interp (interp_vignpsf_lc)
    call srctool_psf_empty_psf(psf_convolved)
    nullify(pinterval)
    nullify(pavign)
    nullify(pbadpix)
    nullify(ppsf)
    nullify(prefarf)
    nullify(pinstpar)
    nullify(pdetmap)
    nullify(pdeadcor)
    nullify(pgtilist)
    nullify(sl_src)
    nullify(sl_back)

  end subroutine srctool_calc_fracrsp

  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------
  !----------------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------
  subroutine srctool_calc_fracrsps_parallel(instlist, arflist, lclist, src, ierr)
    character(len=*), parameter :: subroutinename='srctool_calc_fracrsps_parallel'
    type(instlist_type), intent(inout)   :: instlist
    type(arflist_type), intent(inout)   :: arflist
    type(lclist_type), intent(inout)   :: lclist
    type(src_type), intent(inout)   :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)  :: i, iidx
    integer(kind=int32)  :: ierr_i(1:MAX_INST)
    integer(kind=int32)  :: num_arfs_to_consider
    integer(kind=int32)  :: ii
    integer(kind=int32)  :: i_ok(1:MAX_INST)
    integer(kind=int32)  :: max_i_ok

!#ifdef _OPENMP
!    character(len=STR_MAX) :: str_buffer
!    integer(kind=int32)  :: max_threads
!#endif
!    type(fracrsp_workarea_type) :: workarea(MAX_INST)
    ierr=ERROR_LEVEL_OK
    ierr_i(:)=ERROR_LEVEL_OK

!#ifdef _OPENMP
!    max_threads = omp_get_max_threads()
!    write (str_buffer, '(A,I0,A)') &
!        "Running with: ",max_threads, " threads"
!    call wrschatty(subroutinename, str_buffer)
!#endif

    ! make a list of valid instruments
    max_i_ok = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        max_i_ok = max_i_ok + 1
        i_ok(max_i_ok) = i
      end if
    end do



    !$OMP PARALLEL DO shared(src, instlist) private(iidx)
    do ii=1,max_i_ok
      iidx = i_ok(ii)
      if ( instlist%inst(iidx)%to_consider ) then
        call srctool_calc_fracrsp(instlist, &
            instlist%inst(iidx),arflist%arf(iidx),lclist%lc(iidx), &
            src, ierr_i(iidx))
      end if
    end do
    !$OMP END PARALLEL DO


    do i=1,MAX_INST
      if ( ierr_i(i) /= ERROR_LEVEL_OK ) then !.and. ierr_i(i) /= ERROR_LEVEL_WARNING) then
        call wrserr(subroutinename,ierr_i(i),"Problem calculating effective areas")
        ierr = ierr_i(i)
        return
      end if
    end do

    num_arfs_to_consider= 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider .and. arflist%arf(i)%to_consider ) then
         num_arfs_to_consider = num_arfs_to_consider + 1
      endif
    end do

    if ( num_arfs_to_consider > 0 ) then
      call srctool_arf_coadd_arflist(arflist, instlist )
    endif
  end subroutine srctool_calc_fracrsps_parallel

  !-------------------------------------------------------------------------------
  subroutine srctool_calc_fracrsps_serial(instlist, arflist, lclist, src, ierr)
    character(len=*), parameter :: subroutinename='srctool_calc_fracrsps_serial'
    type(instlist_type), intent(inout)   :: instlist
    type(arflist_type), intent(inout)   :: arflist
    type(lclist_type), intent(inout)   :: lclist
    type(src_type), intent(inout)   :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)        :: i
    integer(kind=int32) :: num_arfs_to_consider

    ierr=ERROR_LEVEL_OK

    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        call srctool_calc_fracrsp(instlist, instlist%inst(i), arflist%arf(i), &
             lclist%lc(i), src, ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr(subroutinename,ierr,"Problem calculating effective areas")
          return
        end if
      end if
    end do

    num_arfs_to_consider= 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider .and. arflist%arf(i)%to_consider ) then
        num_arfs_to_consider = num_arfs_to_consider + 1
      endif
    end do

    if ( num_arfs_to_consider > 0 ) then
      call srctool_arf_coadd_arflist(arflist, instlist )
    endif
  end subroutine srctool_calc_fracrsps_serial


  !-------------------------------------------------------------------------------
  subroutine srctool_merge_srcgtis(instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_merge_srcgtis'
    type(instlist_type), intent(inout),target   :: instlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)        :: i,j,k
    real(kind=real64), allocatable, dimension(:) :: tstart,tstop
    integer(kind=int32)  :: num_rows
    integer(kind=int32)  :: status
    integer(kind=int32)  :: nok
    type(gtilist_type), pointer :: merged_src_gtilist

    status = 0
    ierr=ERROR_LEVEL_OK

    ! merge the src GTIs from the separate instruments
    ! to give a merged GTI which covers intervals when the source region is covered by >=1 eROSITA instrument
    num_rows = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        num_rows = num_rows +  instlist%inst(i)%src_gtilist%num_rows
      end if
    end do
    merged_src_gtilist => instlist%inst(0)%src_gtilist

    if ( num_rows <= 0 ) then
!      ierr = ERROR_LEVEL_SEVERE
!      call wrserr (subroutinename, ierr, "Could not find any valid GTIs")
!      return
      call wrserr (subroutinename, ERROR_LEVEL_WARNING, "Could not find any valid GTIs, writing zero rows")

      merged_src_gtilist%num_rows = 0
      merged_src_gtilist%is_valid = .true.
      return
    end if

    if ( allocated(tstart)) deallocate(tstart)
    if ( allocated(tstop))  deallocate(tstop)

    allocate(tstart(num_rows),tstop(num_rows), &
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! copy the data to the new array
    k = 1
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        do j=1,instlist%inst(i)%src_gtilist%num_rows
          tstart(k) = instlist%inst(i)%src_gtilist%pstart(j)
          tstop(k)  = instlist%inst(i)%src_gtilist%pstop(j)
          k = k + 1
        end do
      end if
    end do

    call insertion_cosort(tstart(:),tstop(:))

    ! first go through and make sure that any intervals with matching tstart are filtered out
    ! also check for overlapping/consecutive gtis and merge them
    k = 1
    do i=1,num_rows
      if ( tstart(i) <= tstart(k)) then
        tstop(k) = MAX(tstop(i),tstop(k))
      else
        if (tstart(i) <= tstop(k)) then
          tstop(k) = MAX(tstop(i),tstop(k))
        else
          k = k + 1
          tstart(k) = tstart(i)
          tstop(k) = tstop(i)
        end if
      end if
    end do
    num_rows = k

    if ( num_rows <= 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not find any valid GTIs")
      return
    end if

    ! now copy over the merged gtis to the instlist%inst(0)%src_gtilist
    allocate(merged_src_gtilist%pstart(num_rows),&
        merged_src_gtilist%pstop(num_rows),&
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    merged_src_gtilist%num_rows = num_rows
    merged_src_gtilist%pstart(:) = tstart(1:num_rows)
    merged_src_gtilist%pstop(:)  = tstop(1:num_rows)

    call fitswrap_set_key(merged_src_gtilist%hduinfo%key_instrume, "MERGED")
    call fitswrap_set_key(merged_src_gtilist%hduinfo%key_filter,   &
        trim(ESASS_FILTER_STR(instlist%inst(0)%filter_code)))
    call fitswrap_set_key(merged_src_gtilist%hduinfo%key_extname, trim(SRCTOOL_CALTYPE_SRCGTI%extroot) // "0" )

    ! now deal with the NINST and INSTRUMn keywords
    nok = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        nok = nok + 1
        call fitswrap_set_key(merged_src_gtilist%hduinfo%pkey_instrum(nok),  trim(instlist%inst(i)%name))
      end if
    end do
    call fitswrap_set_key(merged_src_gtilist%hduinfo%key_ninst, nok)

    merged_src_gtilist%is_valid = .true.

    if ( allocated(tstart)) deallocate(tstart)
    if ( allocated(tstop))  deallocate(tstop)
    nullify(merged_src_gtilist)

  end subroutine srctool_merge_srcgtis



  !-------------------------------------------------------------------------------
  ! -- calculate the source GTIs and the total exposure time
  recursive subroutine srctool_calc_srcgti(inst, intervallist, src, src_gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_calc_srcgti'
    type(inst_type), intent(inout), target   :: inst
!    type(inst_type), intent(in), target   :: inst
    type(intervallist_type), intent(in), target :: intervallist
    type(src_type), intent(in)         :: src
    type(gtilist_type), intent(inout)  :: src_gtilist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32)    :: i,j ! indices
    integer(kind=int32)    :: status
    character(len=LSTR_MAX) :: str_buffer
    logical                :: srcreg_in_fov_flag

    type(interval_type),    pointer :: pinterval   ! pointers to data sub-structures within "inst","intervallist"
    type(badpix_type),      pointer :: pbadpix
    type(detmap_type),      pointer :: pdetmap
    type(instpar_type),     pointer :: pinstpar
    type(gtilist_type),     pointer :: pgtilist_in
!    type(deadcor_type), pointer :: null_pdeadcor
!    type(psf_type)   :: null_psf
    type(samplelist_type) :: samplelist_src  ! we can safely discard this at the end of the subroutine

    type(tstep_type) :: tstep   ! working data structures

    ! the following are used to keep a record of the summed ontime as we step through the timeline
    real(kind=real64)  :: t_tot          ! sum over the 'tstep%dt's
    real(kind=real64)  :: this_tstep
    integer(kind=int32)  :: nsteps
    integer(kind=int32)  :: k
    real(kind=real64)  :: effmap_val

!    real(kind=real64)  :: debug_meanpoint
!    integer(kind=int32)  ::  debug_npoints, debug_maxpoint
!    debug_meanpoint = 0.0d0
!    debug_npoints = 0
!    debug_maxpoint = 0

    ierr=ERROR_LEVEL_OK
    status = 0
    ! setup the dummy psf and pdeadcor types
!    nullify(null_pdeadcor)

!    call srctool_psf_empty_psf(null_psf)
!    null_psf%is_valid = .false.
!    null_psf%class = SRCTOOL_PSF_CLASS_NONE

    write (str_buffer, '(A,A,A,I0,A,A,A)') &
        "Calculating source GTIs for src=", trim(src%name), &
        " inst index=",inst%index, " (name=", trim(inst%name), ")"
    call wrschatty (subroutinename, str_buffer)


    call srctool_misc_select_badpix_pointer(inst,pbadpix,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_detmap_pointer(inst,pdetmap,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    call srctool_misc_select_gtilist_pointer(inst,pgtilist_in,ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return

    !----- setup some space for the src gtis based on the intervallist
    call srctool_gti_setup_src_gtilist(src_gtilist, pgtilist_in, intervallist, ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return
    !----------------------------------------------------------------------------------------
    ! Setup the list of spatial sample points - only do at one energy
!    call srctool_samplelist_empty(samplelist_src)
!    samplelist_src%evt_class  = EVT_CLASS_FLAG_SRC
!    call srctool_setup_samplelist(inst, src, samplelist_src, null_psf, EVT_CLASS_FLAG_SRC, ierr)
    call srctool_setup_samplelist_for_gti(inst, src, samplelist_src, ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem setting up samplelist for GTI computation")
      return
    end if

    ! write(str_buffer, '(A,A,A,I0, A,I0,A,I0,A, A,F10.3,A, A,F8.6,A)') &
    !     "DEBUG: Sample pattern inst=", trim(inst%name), &
    !     " src=", src%index, &
    !     " Np(src|num_e)=(", &
    !     samplelist_src%num_points, "|", samplelist_src%num_eval_e, ")", &
    !     " Area/sample(src)=(", &
    !     samplelist_src%area_per_point_sqarcsec,")arcsec^2", &
    !     " Weight/sample(src)=(", &
    !     samplelist_src%weight_per_point, ")"
    ! call wrschatty(subroutinename, str_buffer)
    !----------------------------------------------------------------------------------------

    !----------------------------------------------------------------------------------------
    ! initialise values which will be summed over the time steps
    !----------------------------------------------------------------------------------------
    t_tot           = 0.0d0
    !----------------------------------------------------------------------------------------

    ! reset the effmap to the start of the first interval
    call srctool_badpix_calculate_effmap(inst%effmap, pdetmap, &
         intervallist%pinterval(1)%t1, inst%use_fov_mask, ierr)
    if ( ierr /=  ERROR_LEVEL_OK ) return

    !----------------------------------------------------------------------------------------
    !  START OF THE MAIN LOOP OVER THE TIME INTERVALS
    !----------------------------------------------------------------------------------------
    ! now work through the intervallist
    do i=1,intervallist%num_rows ! for each interval defined in the intervallist
      pinterval => intervallist%pinterval(i)
      srcreg_in_fov_flag = .false.

      if ( pinterval%inst_flag(inst%index) ) then  ! if this instrument is valid in this interval

        tstep%tmax = pinterval%t1 ! the start of the interval
!        call srctool_increment_tstep_old(tstep, pinterval, null_pdeadcor, 0.0d0, ierr)

        ! step over the interval using an integer number of steps
        call srctool_calc_integration_steps(pinterval%t1, pinterval%t2, inst%tstep, this_tstep, nsteps)
        if ( this_tstep <= 0.0d0 ) then
          ierr = ERROR_LEVEL_WARNING
          call wrserr (subroutinename, ierr, "Error calculating tstep")
        endif


        do k=1,nsteps
!        do while ( tstep%tmin < pinterval%t2 )  ! for each time step within the interval
          call srctool_increment_calc_srcgti_tstep(tstep, pinterval, this_tstep, ierr)
          if ( ierr /=  ERROR_LEVEL_OK ) return

          ! check to see if we need to update the effmap
          if ( tstep%tmid >= inst%effmap%time_of_next_update ) then
            call srctool_badpix_calculate_effmap(inst%effmap, pdetmap, tstep%tmid, &
                 inst%use_fov_mask, ierr)
            if ( ierr /=  ERROR_LEVEL_OK ) return
          endif

          !-------------------------------------------------------
          ! calc the current ra0,dec0,roll at middle of tstep
          call srctool_att_get_current_att(inst%attlist, tstep%tmid, &
              samplelist_src%ra0, samplelist_src%dec0, samplelist_src%roll, ierr)
          if ( ierr /= ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Error getting current attitude")
            return
          end if

          ! calc the (fractional) pixel coords of the src
          call srctool_att_radec2rawxy_a (pinstpar, &
              samplelist_src%ra0, samplelist_src%dec0, samplelist_src%roll, &
              samplelist_src%src_ra, samplelist_src%src_dec, &
              samplelist_src%src_rawx, samplelist_src%src_rawy, &
              samplelist_src%src_on_chip_flag )
          !-------------------------------------------------------

          ! calc offaxis of the src in the focal plane
          samplelist_src%src_theta_arcmin = DEG2ARCMIN * &
              calc_ang_sep(samplelist_src%ra0, samplelist_src%dec0, &
                           samplelist_src%src_ra, samplelist_src%src_dec)

          ! Before we do a more detailed calculation, we want to know if any src sample
          ! point could possibly lie within the FOV
          ! so must allow a conservatively search radius .
          ! FIXME! FIXME! JSS inserted factor 2x as radius could be larger than FoV if FoV flag switched off
          if ( samplelist_src%src_on_chip_flag .or. &
              (samplelist_src%src_theta_arcmin < &
              samplelist_src%largest_sample_radius_deg*DEG2ARCMIN + 2*pinstpar%fov_radius_arcmin )) then

            ! calc azimuthal angle of the src in the focal plane
            samplelist_src%src_phi = &
                atan2d((samplelist_src%src_rawy - pinstpar%oa_rawy)*pinstpar%pixscale_y, &
                (samplelist_src%src_rawx - pinstpar%oa_rawx)*pinstpar%pixscale_x)

            ! compute these once to avoid repeats in inner loop
            samplelist_src%sindec0 = sind(samplelist_src%dec0)
            samplelist_src%cosdec0 = cosd(samplelist_src%dec0)
            samplelist_src%sinroll = sind(samplelist_src%roll)
            samplelist_src%cosroll = cosd(samplelist_src%roll)

            ! Treat each sample point separately since we want to break the loop as soon as possible
            srcreg_in_fov_flag = .false.
            do j=1,samplelist_src%num_points
              !----------------------------------------------------------------------------------------
              ! Get the detector coordinates of the src sample points
              call srctool_att_radec2rawxy_a2_faster (pinstpar, &
                  samplelist_src%ra0, &
                  samplelist_src%cosdec0, &
                  samplelist_src%sindec0, &
                  samplelist_src%cosroll, &
                  samplelist_src%sinroll, &
                  samplelist_src%pra(j),  &
                  samplelist_src%pcosdec(j), &
                  samplelist_src%psindec(j), &
                  samplelist_src%prawx(j), &
                  samplelist_src%prawy(j), &
                  samplelist_src%prawx_int(j), &
                  samplelist_src%prawy_int(j), &
                  samplelist_src%pon_chip_flag(j) )
              !----------------------------------------------------------------------------------------

              call srctool_badpix_get_effmap(inst%effmap, samplelist_src%pon_chip_flag(j), &
                  samplelist_src%prawx_int(j), samplelist_src%prawy_int(j), effmap_val)
              if ( effmap_val > 0.0d0 ) then
                srcreg_in_fov_flag = .true.

!                ! keep a count of how far we have to go into this loop on average
!                debug_meanpoint = debug_meanpoint + REAL(j, real64)
!                debug_npoints = debug_npoints + 1
!                if ( j > debug_maxpoint ) debug_maxpoint = j
                EXIT  ! break out of loop ASAP
              endif

!              if ( samplelist_src%pon_chip_flag(j) ) then
!                if ( inst%effmap%pmap(samplelist_src%prawx_int(j),&
!                                      samplelist_src%prawy_int(j)) > 0.0e0 ) then
!                  srcreg_in_fov_flag = .true.
!                  EXIT  ! break out of loop ASAP
!                endif
!              endif

            end do

            ! check if we are starting or ending a src gti here
            call srctool_gti_update_src_gtilist (src_gtilist, srcreg_in_fov_flag, tstep%tmin, tstep%tmax, ierr)
            if ( ierr /= ERROR_LEVEL_OK ) then
              call wrserr (subroutinename, ierr, "Error updating src gtilist")
              return
            end if

            if ( srcreg_in_fov_flag ) then

              ! add the contribution totals
              t_tot = t_tot + tstep%dt
            end if
          else
            ! check if we are ending a src gti here
            call srctool_gti_update_src_gtilist (src_gtilist, .false., tstep%tmin, tstep%tmax, ierr)
            if ( ierr /= ERROR_LEVEL_OK ) then
              call wrserr (subroutinename, ierr, "Error updating src gtilist")
              return
            end if
          end if

!          k = k + 1
          !increment the tstep by one interval
!          call srctool_increment_tstep(tstep, pinterval, null_pdeadcor, pinterval%tstep, ierr)
!          call srctool_increment_calc_srcgti_tstep(tstep, pinterval, pinterval%tstep, ierr)
!          call srctool_increment_calc_srcgti_tstep(tstep, pinterval, this_tstep, ierr)
!          if ( ierr /=  ERROR_LEVEL_OK ) return
        end do
        ! check if we are ending a src gti here
        call srctool_gti_update_src_gtilist (src_gtilist, .false., tstep%tmin, tstep%tmax, ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error updating src gtilist")
          return
        end if
      end if

    end do
    !----------------------------------------------------------------------------------------
    !  END OF THE MAIN LOOP OVER THE TIME INTERVALS
    !----------------------------------------------------------------------------------------


    !----------------------------------------------------------------------------------------
    !decrement the number of rows in the src_gtilist by one
    src_gtilist%num_rows = src_gtilist%num_rows - 1
    !----------------------------------------------------------------------------------------

    ! add a fix so that we don't get errors later when we have zero valid src gtis
    if ( (src_gtilist%num_rows < 1) .or. &
         (t_tot <= 0.0d0) ) then
      src_gtilist%num_rows = 0
!      src_gtilist%num_rows = 1
!      src_gtilist%pstart(1) = intervallist%pinterval(1)%t1
!      src_gtilist%pstop(1)  = intervallist%pinterval(1)%t1
      t_tot = 0.0d0
      !- emit a warning to the user

      write (str_buffer, '(A,A,A,I0,A,A,A)') &
          "Warning: Zero length source GTIs for src=", trim(src%name), &
          " inst index=",inst%index, " (name=", trim(inst%name), ")"
      call wrserr (subroutinename, ERROR_LEVEL_WARNING, str_buffer)

    endif


!    debug_meanpoint = debug_meanpoint / REAL(MAX(1,debug_npoints),real64)
!    write (str_buffer, '(A,x,F10.5,3(x,I10))') &
!        'DEBUG CALC_SRC_GTI_LOOP_DEPTH (mean,nloops,max,total) = ', &
!        debug_meanpoint, debug_npoints, debug_maxpoint, samplelist_src%num_points
!    call wrschatty (subroutinename, str_buffer)

    !----------------------------------------------------------------------------------------
    ! set the exposure time values in the lc struct
    call fitswrap_set_key(src_gtilist%key_telapse, t_tot)
    call fitswrap_set_key(src_gtilist%key_ontime,  t_tot)
    !----------------------------------------------------------------------------------------

    call fitswrap_set_key(src_gtilist%hduinfo%key_instrume, trim(inst%name))
    call fitswrap_set_key(src_gtilist%hduinfo%key_filter,   trim(ESASS_FILTER_STR(inst%filter_code)))

    call fitswrap_set_key(src_gtilist%hduinfo%key_extname, trim(SRCTOOL_CALTYPE_SRCGTI%extroot) // trim(i2s(inst%index)))


!    ! debug reporting of source GTIs
!    do i=1,src_gtilist%num_rows
!      write (str_buffer, '(A,x,I8,3(x,F12.2))'), "SOURCE_GTI", &
!          i, src_gtilist%pstart(i),  src_gtilist%pstop(i),src_gtilist%pstop(i)-src_gtilist%pstart(i)
!      call wrschatty (subroutinename, str_buffer)
!    enddo


    ! tidy up
    call srctool_samplelist_empty(samplelist_src)
    nullify(pinterval)
    nullify(pbadpix)
    nullify(pinstpar)
    nullify(pdetmap)
    nullify(pgtilist_in)

  end subroutine srctool_calc_srcgti
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_calc_srcgtis_parallel(instlist, src, ierr)
    character(len=*), parameter :: subroutinename='srctool_calc_srcgtis_parallel'
    type(instlist_type), intent(inout)   :: instlist
    type(src_type), intent(inout)   :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)        :: i, iidx
    integer(kind=int32)  :: ierr_i(MAX_INST)
    character(len=LSTR_MAX) :: str_buffer
    integer(kind=int32)  :: status
    integer(kind=int32)  :: ii
    integer(kind=int32)  :: i_ok(1:MAX_INST)
    integer(kind=int32)  :: max_i_ok

!#ifdef _OPENMP
!    integer(kind=int32)  :: max_threads
!#endif

    status = 0
    ierr=ERROR_LEVEL_OK
    ierr_i(:)=ERROR_LEVEL_OK

!#ifdef _OPENMP
!    max_threads = omp_get_max_threads()
!    write (str_buffer, '(A,I0,A)') &
!        "Running with: ",max_threads, " threads"
!    call wrschatty(subroutinename, str_buffer)
!#endif


    ! make a list of valid instruments
    max_i_ok = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        max_i_ok = max_i_ok + 1
        i_ok(max_i_ok) = i
      end if
    end do

    ! calculate the src GTIs for each instrument - spawn multiple threads if allowed
    !$OMP PARALLEL DO shared(src, instlist) private(iidx)
    do ii=1,max_i_ok
      iidx = i_ok(ii)
      if ( instlist%inst(iidx)%to_consider ) then
        call srctool_calc_srcgti( &
            instlist%inst(iidx), instlist%intervallist, &
            src, instlist%inst(iidx)%src_gtilist, ierr_i(iidx))
      end if
    end do
    !$OMP END PARALLEL DO

    ! check for errors in any of the threads
    do i=1,MAX_INST
      if ( ierr_i(i) /= ERROR_LEVEL_OK ) then
        ierr = ierr_i(i)
        call wrserr(subroutinename,ierr,&
            "Problem calculating source GTIs for:" // trim(instlist%inst(i)%name) )
        return
      end if
    end do


    ! merge the individual src gtis into a src gti for merged instrument
    call srctool_merge_srcgtis(instlist, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem merging source GTIs")
      return
    endif

    ! emit verbose details
    do i=0,MAX_INST
      if ( instlist%inst(i)%to_consider .or. (i == 0) ) then
        write (str_buffer, '(A,I8,A,A,A,I0,A,A,A)') &
            "There were ", instlist%inst(i)%src_gtilist%num_rows, &
            " source GTIs for src=", trim(src%name), &
            " inst index=",instlist%inst(i)%index, " (name=", trim(instlist%inst(i)%name), ")"
        call wrschatty (subroutinename, str_buffer)
      end if
    end do

  end subroutine srctool_calc_srcgtis_parallel


  !-------------------------------------------------------------------------------
  subroutine srctool_make_dummy_srcgtis(instlist, src, ierr)
    character(len=*), parameter :: subroutinename='srctool_make_dummy_srcgtis'
    type(instlist_type), intent(inout)   :: instlist
    type(src_type), intent(inout)   :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)        :: i
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)  :: status

    status = 0
    ierr=ERROR_LEVEL_OK

    ! copy GTIs for each instrument
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        call srctool_gti_duplicate_gtilist(instlist%inst(i)%gtilist, &
            instlist%inst(i)%src_gtilist, ierr)
      end if
    end do

    ! merge the individual src gtis into a src gti for the merged instrument
    call srctool_merge_srcgtis(instlist, ierr)
    if ( ierr /= ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem merging source GTIs")
      return
    endif

    ! emit verbose description
    do i=0,MAX_INST
      if ( instlist%inst(i)%to_consider .or. (i == 0) ) then
        write (str_buffer, '(A,I8,A,A,A,I0,A,A,A)') &
            "There were ", instlist%inst(i)%src_gtilist%num_rows, &
            " source GTIs for src=", trim(src%name), &
            " inst index=",instlist%inst(i)%index, " (name=", trim(instlist%inst(i)%name), ")"
        call wrschatty (subroutinename, str_buffer)
      end if
    end do

!    do i=1,instlist%inst(0)%src_gtilist%num_rows
!      write (str_buffer, '(A,A,A,I6,x,F12.6,x,F12.6)'), &
!          "DEBUG DUMMY SRCGTI ", INST_STEM, "0: ", &
!          i, &
!          instlist%inst(0)%src_gtilist%pstart(i), &
!          instlist%inst(0)%src_gtilist%pstop(i)
!      call wrschatty (subroutinename, str_buffer)
!    end do

  end subroutine srctool_make_dummy_srcgtis

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- END of FRACRSP handling ---------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !- Write regions to a file
  !-------------------------------------------------------------------------------

  subroutine srctool_write_regions(filename, srclist, clobber, sourcereg, ierr)
    character(len=*), parameter  :: subroutinename='srctool_write_regions_src'

    character(len=*), intent(in) :: filename    ! output filename
    type(srclist_type), intent(in) :: srclist   ! input sources
    logical, intent(in) :: clobber   ! overwrite output file
    logical, intent(in) :: sourcereg ! true=source, false=background
    integer, intent(out) :: ierr                ! if any error

    integer unit, iostat
    integer src_i
    character(len=7) :: status

    if(clobber) then
       ! always overwrite output file
       status = "replace"
    else
       ! do no overwrite output file
       status = "new"
    end if

    ! open output file
    ierr = ERROR_LEVEL_OK
    iostat = 0
    open(newunit=unit, file=trim(filename), status=trim(status), &
         action='write', iostat=iostat)
    if( iostat /= 0 ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Could not open file")
       return
    end if

    ! write regions for each source
    do src_i=1,srclist%num_srcs
       if(sourcereg) then
          ! is source
          write(unit, '(a)', iostat=iostat) trim(srclist%psrc(src_i)%str_src_region)
       else
          ! is background
          write(unit, '(a)', iostat=iostat) trim(srclist%psrc(src_i)%str_back_region)
       end if
       if( iostat /= 0 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Could not write to file")
          return
       end if
    end do

    ! finish up
    close(unit, iostat=iostat)
    if( iostat /= 0 ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Could not close file")
       return
    end if

  end subroutine srctool_write_regions

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- Extended source routines --------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  !------------------------------------------------
  ! calculate model parameters and number of pixels
  subroutine srctool_prepare_extent_model_params(src, fft_pixscale, ierr)
    character(len=*), parameter  :: subroutinename='srctool_prepare_extent_model_params'
    type(src_type), intent(inout), target :: src
    real(kind=real64), intent(in)   :: fft_pixscale ! arcsec/pix
    integer(kind=int32), intent(inout)   :: ierr

    type(extent_type), pointer :: extent ! simple shortcut

    real(kind=real64) :: r1,r2,rm,dr, f1,f2,fm
    integer(kind=int32) :: nstep, k

    ierr = ERROR_LEVEL_OK

    extent => src%extent

    ! add missing parameters for beta and gaussian
    if ((extent%code == EXTTYPE_CODE_BETA) .or. &
         (extent%code == EXTTYPE_CODE_GAUSSIAN) ) then
       if ( extent%num_pars == 1 ) then
          extent%num_pars = 2
          extent%pars(2) = EXTENDED_SOURCE_DEFAULT_NSIGMAS
       end if
    end if

    ! calculate integration normalisation and maximum dimensions
    ! now calculate the integration normalisation
    select case ( extent%code )

    case ( EXTTYPE_CODE_POINT )
       extent%norm_const = 1.0d0
       extent%num_pix = 0

    case ( EXTTYPE_CODE_GAUSSIAN)
       ! https://www.wolframalpha.com/input/?i=integrate+2*pi*x*
       !    exp(-0.5*(x%2Fsigma)**2)+from+0+to+c
       ! definite integral = const = 1 / [2 pi sigma^2 *
       !                             (1 - exp(-0.5*(c/sigma)^2)) ]
       extent%norm_const = &
            1.0d0/(2.0d0*PI*extent%pars(1)**2 * &
            (1.0d0 - exp(-0.5d0*(extent%pars(2))**2)))
       ! use diameter given by product of r_c/sigma and Nr_c/Nsigma
       extent%num_pix = nint(2.0 * extent%pars(1) * extent%pars(2) / &
            fft_pixscale)

    case ( EXTTYPE_CODE_BETA )
       ! first need to do integral of beta function over r=[0,r_c]
       ! use simpson's rule
       ! https://en.wikipedia.org/wiki/Simpson%27s_rule
       extent%norm_const = 0.0d0  !
       nstep = 1000
       dr = extent%pars(1) * extent%pars(2) / nstep
       do k=1,nstep
          r1 = (k-1)*dr
          r2 = (k)*dr
          rm = (r1+r2)*0.5
          f1 =  r1 * (1.0d0 + (r1/extent%pars(1))**2)**(-1.5)
          f2 =  r2 * (1.0d0 + (r2/extent%pars(1))**2)**(-1.5)
          fm =  rm * (1.0d0 + (rm/extent%pars(1))**2)**(-1.5)
          extent%norm_const = extent%norm_const + &
               (dr/6.0d0)*2.0d0 * PI * (f1 + 4.0d0*fm + f2)
       end do
       extent%norm_const = 1.0d0 / extent%norm_const
       !  TODO
       ! use diameter given by product of r_c/sigma and Nr_c/Nsigma
       extent%num_pix = nint(2.0 * extent%pars(1) * &
            extent%pars(2) /  fft_pixscale)

    case ( EXTTYPE_CODE_TOPHAT )
       ! integral = pi r**2
       extent%norm_const = 1.0d0/(PI*extent%pars(1)**2)   !
       extent%num_pix = nint(2.0 * extent%pars(1) / fft_pixscale)

    case ( EXTTYPE_CODE_MAP )
       ! we will assume that the map has been pre-normalised
       extent%norm_const = 1.0d0  !
       extent%num_pix = nint(2.0 * src%extmap%max_radius_deg * &
            DEG2ARCSEC / fft_pixscale)

    case default
       ierr = ERROR_LEVEL_SEVERE
       call wrserr(subroutinename, ierr, "Unknown extent code")
       return
    end select

    nullify(extent)

  end subroutine srctool_prepare_extent_model_params

  !--------------------------------------------------------
  ! set up the extent model, ready for convolution with PSF
  subroutine srctool_prepare_extent_model(src, fft_pixscale, ierr)
    character(len=*), parameter  :: subroutinename='srctool_prepare_extent_model'
    type(src_type), intent(inout), target :: src
    real(kind=real64), intent(in)   :: fft_pixscale ! arcsec/pix
    integer(kind=int32), intent(inout)   :: ierr

    integer(kind=int32) :: i, j
    integer(kind=int32) :: status
    real(kind=real64)   :: i0, j0
    real(kind=real64)   :: inv_rc_pix2, inv_sigma_pix2, nrc2, nsigma2, cut_radius2

    type(extent_type), pointer :: extent ! simple shortcut

    ierr = ERROR_LEVEL_OK
    status = 0

    extent => src%extent

    ! Exit if a point source
    if( extent%code == EXTTYPE_CODE_POINT ) then
       return
    end if

    !------------------------------------------------------
    if ( allocated(extent%xy) ) then
      call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
          "Memory previously allocated for extent model in xy plane")
      return
    end if

    ! prepare the extent model
    call wrschatty(subroutinename, "Setting up extent model for source: " &
        // trim(src%name) // " exttype=" // EXTTYPE_CODE_STR(extent%code) )

    ! allocate and initialise
    allocate(extent%xy(extent%num_pix,extent%num_pix), stat=status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return
    extent%xy = 0.0

    ! Calculate the central pixel coords of the analytic function
    i0 = real(extent%num_pix / 2.0, kind=real64)
    j0 = real(extent%num_pix / 2.0, kind=real64)

    ! Build the real-space map of the source
    ! write the model into the lower left corner of the allocated array
    select case ( extent%code )

    case  ( EXTTYPE_CODE_BETA ) ! just use value at middle of pixels
       !---------------------------------------------------------------
       ! - beta model, par(1) = rc (arcsec)
       !               par(2) = Nrc (unitless) - cutoff radius
       inv_rc_pix2 = (extent%pars(1)/fft_pixscale)**(-2)
       cut_radius2 = extent%pars(2)*extent%pars(2)
       do i=1,extent%num_pix
          do j=1,extent%num_pix
             nrc2 = ((i-i0)*(i-i0) + (j-j0)*(j-j0))*inv_rc_pix2
             ! only fill the model out to the requested maximum number of critical radii
             if ( nrc2 <= cut_radius2 ) then
                extent%xy(i,j) = real((1.0 + nrc2)**(-1.5), kind=real32)
             end if
          end do
       end do

    case  ( EXTTYPE_CODE_GAUSSIAN )
       !---------------------------------------------------------------
       ! - Gaussian model, par(1) = sigma (arcsec)
       !                   par(2) = Nsigma (unitless) - cutoff radius
       inv_sigma_pix2 = (extent%pars(1)/fft_pixscale)**(-2)
       cut_radius2    = extent%pars(2)*extent%pars(2)
       do i=1,extent%num_pix
          do j=1,extent%num_pix
             nsigma2 = ((i-i0)*(i-i0) + (j-j0)*(j-j0))*inv_sigma_pix2
             if ( nsigma2 <= cut_radius2 ) then
                extent%xy(i,j) = real(exp( -0.5d0 * nsigma2 ), kind=real32)
             end if
          end do
       end do

    case  ( EXTTYPE_CODE_TOPHAT )  ! just use value at middle of pixels
       !---------------------------------------------------------------
       ! - Tophat model, par(1) = radius (arcsec)
       !
       cut_radius2 = (extent%pars(1)/fft_pixscale)**2
       do i=1,extent%num_pix
          do j=1,extent%num_pix
             if ( (i-i0)*(i-i0) + (j-j0)*(j-j0) <= cut_radius2 ) then
                extent%xy(i,j) = 1.0
             endif
          end do
       end do

    case default
       ierr = ERROR_LEVEL_ERROR
       call wrserr (subroutinename, ierr, &
            "I cannot handle this class of extent model: " // EXTTYPE_CODE_STR(extent%code))
       return
    end select

    ! always normalise extent model to unity (safely)
    extent%xy = extent%xy / max(1.0,sum(extent%xy))

    nullify(extent)

  end subroutine srctool_prepare_extent_model


  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !----------------- Routine which prepares for extended sources -----------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  subroutine srctool_setup_extended_source_handling (instlist, srclist, psftype_code, ierr )
    character(len=*), parameter :: subroutinename='srctool_setup_extended_source_handling'
    type(instlist_type), intent(inout), target :: instlist
    type(srclist_type),  intent(inout), target :: srclist
    integer(kind=int32) , intent(in)  :: psftype_code
    type(psf_type),  pointer :: ppsf
    type(src_type),  pointer :: psrc
    integer(kind=int32), intent(out)   :: ierr
!    integer(kind=int32)                :: n_extended
    integer(kind=int32)                :: largest_psf_map
!    integer(kind=int32)                :: extent_width
    integer(kind=int32)                :: max_dim
    integer(kind=int32) :: i, j
    character(len=STR_MAX) :: str_buffer
    real(kind=real64) :: pix2arcmin

    ierr = ERROR_LEVEL_OK

    if ( (srclist%n_extended_to_extract == 0) .and. &
         (srclist%n_extended_to_exclude == 0)) then
      ! we can skip this section since we have no extended sources
      call wrschatty (subroutinename, "There are no extended sources to consider")
      return
    else
      call wrschatty (subroutinename, "Setting up extended source handling")
    endif

    !---------------------------------------------------------------------
    ! scan the PSFs and work out the pixel scale and the
    ! size of the largest map (in pixels)
    largest_psf_map = 0
    do i=0,MAX_INST
       if ( instlist%inst(i)%to_consider ) then
          call srctool_misc_select_psf_pointer(instlist%inst(i),ppsf,ierr)
          if ( ierr /= ERROR_LEVEL_OK ) then
             call wrserr (subroutinename, ierr, "Problem finding PSF model for instrument")
             return
          end if

          select case ( ppsf%class)

          case ( SRCTOOL_PSF_CLASS_NONE )
             ! do nothing
             instlist%fft_pixscale = 1.0d0
          case ( SRCTOOL_PSF_CLASS_MAP )
             ! TODO make this more robust
             instlist%fft_pixscale = ppsf%psfmaplist%psfmap(1)%pixscale_x * DEG2ARCSEC
             ! TODO check that all maps are on the same scale
             do j=1,ppsf%psfmaplist%num_psfmaps
                max_dim = max(ppsf%psfmaplist%psfmap(j)%naxis1,&
                     ppsf%psfmaplist%psfmap(j)%naxis1)
                if ( max_dim > largest_psf_map ) then
                   largest_psf_map = max_dim
                end if
             end do
          case default
             ierr = ERROR_LEVEL_ERROR
             call wrserr (subroutinename, ierr, &
                  "I cannot handle this class of PSF description: " // &
                  SRCTOOL_PSF_CLASS_STR(ppsf%class))
             return
          end select

          nullify(ppsf)
       end if
    end do

    !---------------------------------------------------------------------
    ! prepare extents (calculate size of extent and images)
    do i=1,srclist%num_srcs
       psrc => srclist%psrc(i)
       if ( psrc%cat_extract .or. psrc%cat_exclude ) then
          call srctool_prepare_extent_model_params(psrc, instlist%fft_pixscale, ierr)
          if (ierr /= ERROR_LEVEL_OK) then
             call wrserr(subroutinename, ierr, "Problem preparing extent model parameters")
             return
          end if
       end if
    end do

    !---------------------------------------------------------------------
    if ( psftype_code ==  SRCTOOL_PSF_CLASS_NONE ) then
      ! we do not need to do any FFT work in this case
      call wrschatty (subroutinename, "No FFT preparation is needed for psftype=" // &
          trim(SRCTOOL_PSF_CLASS_STR(psftype_code)))
      return
    end if

    !---------------------------------------------------------------------
    ! setup plans for FFT
    call srctool_psf_setup_fftw_plans(instlist, largest_psf_map)

    ! information about the PSFs
    write (str_buffer, '(A,F10.7,A)') &
         "Setting up extended sources: fft_pixscale=", instlist%fft_pixscale, " (arcsec/pix)"
    call wrschatty (subroutinename, str_buffer)
    pix2arcmin = instlist%fft_pixscale / 60.0d0
    write (str_buffer, '(A,3(A,I0,A,F6.2,A))') &
         "Setting up extended sources:", &
         " largest_psf_map=",largest_psf_map,  " pix (=",largest_psf_map  *pix2arcmin," arcmin)", &
         " => fft_npix=",    instlist%fft_npix," pix (=",instlist%fft_npix*pix2arcmin," arcmin)"
    call wrschatty (subroutinename, str_buffer)

    !---------------------------------------------------------------------------
    ! now calculate extents
    do i=1,srclist%num_srcs
       psrc => srclist%psrc(i)
       if ( psrc%cat_extract .or. psrc%cat_exclude ) then
          call srctool_prepare_extent_model(psrc, instlist%fft_pixscale, ierr)
          if (ierr /= ERROR_LEVEL_OK) then
             call wrserr(subroutinename, ierr, "Problem preparing extent model parameters")
             return
          end if
       end if
    end do

  end subroutine srctool_setup_extended_source_handling



  !---------------------------------------------------------------------------
  ! Construct a set of survey PSFs for an instrument using the combination
  ! of instantaneous PSFs, vignetting function and a Detmap
  subroutine srctool_build_survey_psfs_for_inst(inst, ierr )
    character(len=*), parameter :: subroutinename='srctool_build_survey_psfs_for_inst'
    type(inst_type), target, intent(inout) :: inst
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: e
    integer(kind=int32) :: i,j,k, itheta,itheta1,itheta2,iphi, ii, jj, ri
    integer(kind=int32) :: status
    character(len=STR_MAX) :: str_buffer
    type(psfmaplist_type), pointer   :: psfmaplist_in
    type(psfmaplist_type), pointer   :: psfmaplist_out
    type(psfmap_type), pointer   :: psfmap_in
    type(psfmap_type), pointer   :: psfmap_out

    type(psf_type), pointer   :: ppsf
!    type(vign_type), pointer   :: pvign
    type(avign_type), pointer   :: pavign
    type(detmap_type), pointer :: pdetmap
    type(instpar_type), pointer :: pinstpar

    real(kind=real64), dimension(:,:,:), allocatable :: psf_weights
    real(kind=real64), dimension(:), allocatable :: vign_values

    real(kind=real64), dimension(:), allocatable :: annulus_npix
    real(kind=real64), dimension(:), allocatable :: annulus_sumpsf
    real(kind=real64), dimension(:), allocatable :: annulus_meanpsf
    integer(kind=int32) :: nannuli
    real(kind=real64) :: annulus_dr
    real(kind=real64) :: w1,w2,theta_arcmin,phi_deg, npix


    ierr = ERROR_LEVEL_OK
    status = 0

    call srctool_psf_empty_psf(inst%survey_psf)

    call srctool_misc_select_psf_pointer(inst,ppsf,ierr)
    if (ierr /= ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable PSF description")
      return
    end if

!    call srctool_misc_select_vign_pointer(inst,pvign,ierr)
!    if (ierr /= ERROR_LEVEL_OK) then
!      call wrserr (subroutinename, ierr, "Could not find suitable vignetting description")
!      return
!    end if
    call srctool_misc_select_avign_pointer(inst,pavign,ierr)
    if (ierr /= ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable vignetting description")
      return
    end if

    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr /= ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable INSTPAR description")
      return
    end if


    call srctool_misc_select_detmap_pointer(inst,pdetmap,ierr)
    if (ierr /= ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable DETMAP description")
      return
    end if



    ! check that we have all the necessary info
    if ( .not. ppsf%is_valid     .or. &
!         (.not. pvign%is_valid ) .or. &
         .not. pavign%is_valid   .or. &
         .not. pinstpar%is_valid .or. &
         .not. pdetmap%is_valid ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, &
          "Not enough information to build survey PSFs for inst: " // trim(inst%name))
      return
    endif

    ! take copies of some params
    inst%survey_psf%class   = ppsf%class
    inst%survey_psf%hduinfo = ppsf%hduinfo  ! check which keywords should be updated

    select case ( ppsf%class )
    case ( SRCTOOL_PSF_CLASS_MAP )
      psfmaplist_in => ppsf%psfmaplist
      psfmaplist_out => inst%survey_psf%psfmaplist

      ! make space for the weights of each instantaneous PSFmap
      allocate(psf_weights(psfmaplist_in%num_e,psfmaplist_in%num_theta,psfmaplist_in%num_phi), &
               vign_values(psfmaplist_in%num_e), &
               STAT=status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
      psf_weights(:,:,:) = 0.0d0
      vign_values(:) = 0.0d0



!      write (str_buffer, '(A,A,x,I0,2(x,F10.6))'), "Instpar x-axis details (npix,oa,pixscale) for: ", &
!          trim(inst%name), pdetmap%naxis1, pinstpar%oa_rawx, pinstpar%pixscale_x
!      call wrschatty (subroutinename, str_buffer)
!      write (str_buffer, '(A,A,x,I0,2(x,F10.6))'), "Instpar y-axis details (npix,oa,pixscale) for: ", &
!          trim(inst%name), pdetmap%naxis2, pinstpar%oa_rawy, pinstpar%pixscale_y
!      call wrschatty (subroutinename, str_buffer)

      ! now step over the detmap to assess what amount of each PSF we should add to the
      ! final survey PSF map
      ! TODO make more efficient by assuming circular symmetry
      phi_deg=0.0d0  ! TODO use phi
      npix = 0.0d0
      do i=1,pdetmap%naxis1
        do j=1,pdetmap%naxis2
          if ( pdetmap%pmap(i,j) > 0.0e0) then
            npix = npix + 1.0d0
            theta_arcmin = DEG2ARCMIN * sqrt( &
                ((REAL(i,kind=real64) - pinstpar%oa_rawx)*pinstpar%pixscale_x)**2 + &
                ((REAL(j,kind=real64) - pinstpar%oa_rawy)*pinstpar%pixscale_y)**2 )

            !----------------------------------------------------------------------------------------
            ! Get the vignetting values at the pixel location for each energy
!            call srctool_vign_calc_vign_values_old(inst, REAL(i,kind=real64), REAL(j,kind=real64), &
!                psfmaplist_in%pe(:), vign_values(:), ierr)
            call avign_calc_vign_at_xye(pavign, REAL(i,kind=real64), REAL(j,kind=real64), &
                psfmaplist_in%pe(:), vign_values(:), ierr)


            if ( ierr /= ERROR_LEVEL_OK ) then
              call wrserr(subroutinename,ierr,"Error calculating vignetting")
              return
            end if



            !----------------------------------------------------------------------------------------

            ! choose the PSF maps which bound this point in theta
            iphi = 1 ! ! TODO use phi properly
            call binary_search(psfmaplist_in%ptheta_arcmin,  theta_arcmin, k, status)
            if ( status == 0 ) then
              itheta1 = k
              itheta2 = k+1
            else if ( status == 1 .or. status == 2 ) then
              itheta1 = k
              itheta2 = k
            else
              ierr = ERROR_LEVEL_SEVERE
              call wrserr(subroutinename,ierr,"Error interpolating theta")
              return
            endif

!            do k=1,psfmaplist_in%num_theta
!              if ( theta_arcmin < psfmaplist_in%ptheta_arcmin(k) ) then
!                exit
!              end if
!            end do
!            k = min(k,psfmaplist_in%num_theta)
!            itheta1 = max(1,k-1)
!            itheta2 = max(1,k)
            if ( itheta1 == itheta2 ) then
              w1 = 0.5d0
              w2 = 0.5d0
            else
              ! calculate the relative weights of each PSF map - based on proximity in radial direction
              w1 = abs((theta_arcmin - psfmaplist_in%ptheta_arcmin(itheta1))/ &
                  (psfmaplist_in%ptheta_arcmin(itheta1) - psfmaplist_in%ptheta_arcmin(itheta2)))
              w2 = 1.0d0 - w1
            endif

            if ((w1 < 0.0d0) .or. (w1 > 1.0d0) .or. &
                (w2 < 0.0d0) .or. (w2 > 1.0d0) ) then
              ierr = ERROR_LEVEL_SEVERE
              write (str_buffer, '(A,F12.7,A,F12.7,A)') " (w1=", w1, " w2=", w2, ")"
              call wrserr(subroutinename,ierr,"Error calculating weights: " // trim(str_buffer))
              return
            endif

            ! add the contributions to the sum
            do e=1,psfmaplist_in%num_e
              psf_weights(e,itheta1,iphi) = psf_weights(e,itheta1,iphi) + vign_values(e)*w1
              psf_weights(e,itheta2,iphi) = psf_weights(e,itheta2,iphi) + vign_values(e)*w2
            end do

          endif
        end do
      end do

      psf_weights(:,:,:) = psf_weights(:,:,:) / (npix)


!      ! DEBUG DEBUG
!      ! report the weights
!      do e=1,psfmaplist_in%num_e
!        do itheta=1,psfmaplist_in%num_theta
!          write (str_buffer, '(A,x,I0,x,I0,x,I0,x,F12.7)'), &
!              "DEBUG SURVEY PSF (npix,e,itheta,psf_weight(e,i))=", &
!              INT(npix,kind=int32),e, itheta, psf_weights(e,itheta,1)
!          call wrschatty (subroutinename, str_buffer)
!        end do
!      end do
!      ! DEBUG DEBUG


      ! output maps will have same energy resolution as the input PSF model
      ! but no dependence on position in FoV

      psfmaplist_out%num_e = psfmaplist_in%num_e
      psfmaplist_out%num_theta = 1
      psfmaplist_out%num_phi   = 1
      psfmaplist_out%num_psfmaps = psfmaplist_out%num_e * psfmaplist_out%num_theta * psfmaplist_out%num_phi

      allocate(psfmaplist_out%pe(psfmaplist_out%num_e), &
               psfmaplist_out%ptheta_arcmin(psfmaplist_out%num_theta), &
               psfmaplist_out%pphi(psfmaplist_out%num_phi), &
               psfmaplist_out%pindex(psfmaplist_out%num_e,psfmaplist_out%num_theta,psfmaplist_out%num_phi), &
               psfmaplist_out%psfmap(psfmaplist_out%num_psfmaps), &
          STAT=status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return

      psfmaplist_out%pe(:) = psfmaplist_in%pe(:)
      psfmaplist_out%ptheta_arcmin(:) = 0.0d0
      psfmaplist_out%pphi(:) = 0.0d0
      do e=1,psfmaplist_out%num_e
        psfmaplist_out%pindex(e,1,1) = e
      end do

      do e=1,psfmaplist_out%num_e
        psfmap_in  => psfmaplist_in%psfmap(psfmaplist_in%pindex(e,1,1))
        psfmap_out => psfmaplist_out%psfmap(e)
        ! write (str_buffer, '(A,A,A,I0,A,F8.3,A)') &
        !     "DEBUG Calc Survey PSF for: ", trim(inst%name), &
        !     " eband= ", e, " energy= " , psfmaplist_out%pe(e), " keV"
        ! call wrschatty (subroutinename, str_buffer)

        ! take a copy of the first psfmap
        call srctool_psf_copy_psfmap(psfmap_in, psfmap_out, ierr)
        if (ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem duplicating PSFmap")
          return
        endif

        ! set pixels to zero, and reset limits
        psfmap_out%ppix(:,:) = 0.0e0 ! 0.0d0
        psfmap_out%pix_x_min = 1
        psfmap_out%pix_x_max =  psfmap_out%naxis1
        psfmap_out%pix_y_min = 1
        psfmap_out%pix_y_max =  psfmap_out%naxis2

        ! update some keywords
        call fitswrap_set_key(psfmap_out%hduinfo%key_extname, trim(psfmap_out%hduinfo%key_extname%value) // "_survey")
        call fitswrap_set_key(psfmap_out%key_caldb_theta,"0.0e0")
        call fitswrap_set_key(psfmap_out%key_caldb_obsmode,"SURVEY")

        ! now, for each input PSF map
        ! 1) make PSF circularly symmetric (average over annuli)
        !  1a) define set of annuli
        !  1b) measure average PSF in each annulus
        ! 2) add to total map, weighted by psf_weights(e,itheta,iphi)

        do itheta=1,psfmaplist_in%num_theta
          do iphi=1,psfmaplist_in%num_phi
            ! weight = psf_weights(e,itheta,iphi)
            psfmap_in => psfmaplist_in%psfmap(psfmaplist_in%pindex(e,itheta,iphi))

            nannuli = MAX(psfmap_in%naxis1,psfmap_in%naxis2) ! double sampling
            annulus_dr = 0.5d0 !*MAX(psfmap_in%pixscale_x,psfmap_in%pixscale_y)

!            write (str_buffer, '(A,x,I0,x,I0,x,I0,x,A,x,I0,2(x,F12.7))'), &
!                "DEBUG Add contrib from PSF (e,itheta,iphi)=", &
!                e, itheta, iphi, &
!                "(nannuli,annulus_dr,psf_weights(e,itheta,iphi))=", &
!                nannuli, annulus_dr, psf_weights(e,itheta,iphi)
!            call wrschatty (subroutinename, str_buffer)

            allocate(annulus_npix(nannuli), &
                annulus_sumpsf(nannuli), &
                annulus_meanpsf(nannuli), &
                STAT=status)
            if ( check_allocate_status(subroutinename, status, ierr) ) return
            annulus_npix(:)   = 0.0d0
            annulus_sumpsf(:) = 0.0d0

            do ii=1,psfmap_in%naxis1
              do jj=1,psfmap_in%naxis2
                ri = CEILING(sqrt( &
                    ((REAL(ii,kind=real64) - psfmap_in%key_crpix1%value))**2 + &
                    ((REAL(jj,kind=real64) - psfmap_in%key_crpix2%value))**2 ) / annulus_dr, kind=int32)
                 if (ri >= 1 .and. ri <= nannuli ) then
                   annulus_npix(ri) =  annulus_npix(ri) + 1.0d0
                   annulus_sumpsf(ri) = annulus_sumpsf(ri) + REAL(psfmap_in%ppix(ii,jj),kind=real64)
                 endif
              end do
            end do

            annulus_meanpsf(:) = annulus_sumpsf(:) / MAX(1.0d0, annulus_npix(:))

            ! now add back to the survey PSF
            do ii=1,psfmap_out%naxis1
              do jj=1,psfmap_out%naxis2
                 ri = CEILING(sqrt( &
                    ((REAL(ii,kind=real64) - psfmap_out%key_crpix1%value))**2 + &
                    ((REAL(jj,kind=real64) - psfmap_out%key_crpix2%value))**2 ) / annulus_dr, kind=int32)
                  if (ri >= 1 .and. ri <= nannuli ) then
                    psfmap_out%ppix(ii,jj) = REAL(psfmap_out%ppix(ii,jj) + &
                        annulus_meanpsf(ri) * psf_weights(e,itheta,iphi), kind=real32)
                 endif
              end do
            end do

            if (allocated(annulus_npix)) deallocate(annulus_npix)
            if (allocated(annulus_sumpsf)) deallocate(annulus_sumpsf)
            if (allocated(annulus_meanpsf)) deallocate(annulus_meanpsf)

          end do
        end do


        ! now renormalise to unity
        psfmap_out%ppix(:,:) = REAL(psfmap_out%ppix(:,:) / sum(psfmap_out%ppix(:,:)), kind=real32)

        ! now compute the RPSF/REEF
        call srctool_psf_calc_rpsf(psfmap_out, psfmap_out%rpsf,ierr)
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem computing  RPSF/REEF from PSFMAP")
          return
        end if


      end do
    case default
      ierr = ERROR_LEVEL_ERROR
      call wrserr (subroutinename, ierr, &
          "I cannot handle this class of PSF desc: " // SRCTOOL_PSF_CLASS_STR(inst%psf%class))
      return
    end select

    if (allocated(psf_weights)) deallocate(psf_weights)
    if (allocated(vign_values)) deallocate(vign_values)

    inst%survey_psf%is_valid = .true.
    nullify(psfmaplist_in)
    nullify(psfmaplist_out)
    nullify(psfmap_in)
    nullify(psfmap_out)
    nullify(ppsf)
    nullify(pavign)
    nullify(pdetmap)
    nullify(pinstpar)

  end subroutine srctool_build_survey_psfs_for_inst
  !---------------------------------------------------------------------------



  !---------------------------------------------------------------------------
  ! Construct a set of survey PSFs for each telescope
  ! Also compute a eROSITA average over all telescopes weighting by ARFs
  ! this is only used for computation of automatic regions in survey mode data
  subroutine srctool_build_survey_psfs_for_instlist(instlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_build_survey_psfs_for_instlist'
    type(instlist_type), intent(inout),target :: instlist
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i,j
    type(inst_type), pointer :: pinst0
    type(inst_type), pointer :: pinsti
    real(kind=real32)  :: weight, sum_weights

    ierr = ERROR_LEVEL_OK

    do i=1,MAX_INST
      pinsti =>  instlist%inst(i)
      if ( pinsti%to_consider ) then
        call srctool_build_survey_psfs_for_inst(pinsti, ierr )
        if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, &
              "Problem building survey PSF for inst: " // trim(pinsti%name))
          return
        end if
      end if
      nullify(pinsti)
    end do

    ! now create a set of `eROSITA total' PSFs, summed over all consdered telescopes
    ! assume equal weighting
    ! TODO - take account of relative ARFs, exposure times
    pinst0 =>  instlist%inst(0)
    sum_weights = 0.0e0
    call srctool_psf_empty_psf(pinst0%survey_psf)
    do i=1,MAX_INST
      pinsti =>  instlist%inst(i)
      if ( pinsti%to_consider ) then
        weight = 1.0e0 ! TODO use proper weighting
        if ( pinsti%survey_psf%is_valid ) then
          if ( sum_weights <= 0.0e0 ) then
            call wrschatty(subroutinename, "Building merged PSF over instruments1a")
            call srctool_psf_copy_psf(pinsti%survey_psf, pinst0%survey_psf, ierr)
            do j=1,pinst0%survey_psf%psfmaplist%num_psfmaps
              pinst0%survey_psf%psfmaplist%psfmap(j)%ppix(:,:) = 0.0e0
            end do
          endif


          ! add in the new PSF maps, with weighting
          do j=1,pinst0%survey_psf%psfmaplist%num_psfmaps
            pinst0%survey_psf%psfmaplist%psfmap(j)%ppix(:,:) =  &
                pinst0%survey_psf%psfmaplist%psfmap(j)%ppix(:,:) + &
                weight * pinsti%survey_psf%psfmaplist%psfmap(j)%ppix(:,:)
          end do

          sum_weights = sum_weights + weight
        end if
      end if
      nullify(pinsti)
    end do

    ! divide psfmaps by the summed weight
    if ( sum_weights > 0.0e0 ) then
      do j=1,pinst0%survey_psf%psfmaplist%num_psfmaps
        pinst0%survey_psf%psfmaplist%psfmap(j)%ppix(:,:) = &
            pinst0%survey_psf%psfmaplist%psfmap(j)%ppix(:,:) / sum_weights
      end do
    end if

    ! generate a new rpsf/reef
    do j=1,pinst0%survey_psf%psfmaplist%num_psfmaps
      call srctool_psf_calc_rpsf(pinst0%survey_psf%psfmaplist%psfmap(j), &
          pinst0%survey_psf%psfmaplist%psfmap(j)%rpsf,ierr)
      if ( ierr /= ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem computing  RPSF/REEF from PSFMAP")
        return
      end if
    end do

    call fitswrap_set_key(pinst0%survey_psf%hduinfo%key_instrume, INST_STEM // "0")
    nullify(pinst0)


    ! to save memory, we can now free up the space used by the
    ! survey PSFs for the individual telescopes as these are no longer used
    do i=1,MAX_INST
      pinsti => instlist%inst(i)
      if ( pinsti%to_consider ) then
        if ( pinsti%survey_psf%is_valid ) then
          call srctool_psf_empty_psf(pinsti%survey_psf)
        end if
      end if
      nullify(pinsti)
    end do


  end subroutine srctool_build_survey_psfs_for_instlist
  !---------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- END of extended source routines -------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------





  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- Main work routine ---------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  ! step through the various requested tasks for each input source
  subroutine srctool_do_all_processing_for_source(instlist, src,  lc_template, ierr )
    character(len=*), parameter :: subroutinename='srctool_do_all_processing_for_source'
    type(instlist_type), intent(inout) :: instlist
    type(src_type),      intent(inout) :: src
    type(lc_type),       intent(inout) :: lc_template
    integer(kind=int32), intent(out)   :: ierr

    !  this will hold the output spectra (if required)
    type(speclist_type)         :: speclist
    !  this will hold the output arf files (if required)
    type(arflist_type)          :: arflist
    !  this will hold the output rmf files (if required)
!    type(rmfsetlist_type)       :: rmfsetlist
    type(rmflist_type)       :: rmflist
    !  this will hold the output lcs (if required)
    type(lclist_type)           :: lclist

    character(len=LSTR_MAX) :: str_buffer
    integer(kind=int32) :: clock_count, clock_count_src, clock_rate ! clock_count0
    integer(kind=int32) :: i
    ierr = ERROR_LEVEL_OK
    ! init the clock
    call system_clock(clock_count_src, clock_rate)

    call srctool_spec_empty_speclist(speclist)
    call srctool_lc_empty_lclist(lclist)
    call srctool_arf_empty_arflist(arflist)
!    call srctool_rmf_empty_rmfsetlist(rmfsetlist)
    call srctool_rmf_empty_rmflist(rmflist)

    do i=0,MAX_INST
      call srctool_gti_empty_gtilist(instlist%inst(i)%src_gtilist)
      call srctool_samplelist_empty(instlist%inst(i)%samplelist_src)
      call srctool_samplelist_empty(instlist%inst(i)%samplelist_back)
    end do

    call wrschatty(subroutinename,"--")
    write (str_buffer, '(A,A,A)') &
        "---------------------------------------- Working on Source: '", &
        trim(src%name), "' ---------------------------------------------------"
    call wrschatty(subroutinename,str_buffer )

    ! TODO - only emit this when working through a catalogue of sources
    write(str_buffer, '(A,A,A,F10.2,A,A,F10.5,A,A,A,A,F6.1,A)') &
        "  Catalogue details for src: ", trim(src%name), &
        " SCTS=", src%cat_scts, " cts", &
        " BG_MAP=", src%cat_bg_map/SQDEG2SQARCMIN, " cts/arcmin2", &
        " EXTTYPE=", trim(EXTTYPE_CODE_STR(src%extent%code)), &
        " EXT=", src%cat_ext*DEG2ARCSEC, " arcsec"
    call wrschatty(subroutinename, str_buffer)


    write(str_buffer, '(A,A,A,A,A,F10.1,A,A,F10.3)') &
        "  Individual details for src: ", trim(src%name), &
        " EXTTYPE=", trim(EXTTYPE_CODE_STR(src%extent%code)), &
        " EXTPARS(1)=", src%extent%pars(1), " arcsec", &
        " EXTPARS(2)=", src%extent%pars(2)
    call wrschatty(subroutinename, str_buffer)


    ! write the regions into strings which can be subsequently written to header keywords
    call srctool_regions_sprint_region(src%src_region,  src%str_src_region, "MINIMAL", ierr)
    call srctool_regions_sprint_region(src%back_region, src%str_back_region, "MINIMAL", ierr)

    call wrschatty(subroutinename, "SRC_REGION:  " // src%str_src_region)
    call wrschatty(subroutinename, "BACK_REGION: " // src%str_back_region)




!    call srctool_regions_sprint_region(src%src_region, str_buffer, ierr)
!    call wrschatty(subroutinename,"Source  region: " // trim(str_buffer) )
!    call srctool_regions_sprint_region(src%back_region, str_buffer, ierr)
!    call wrschatty(subroutinename,"Backgnd region: " // trim(str_buffer) )

    !---------------------------------------------------------------------
    ! First calculate the time intervals when at least one of the source sample points
    ! falls on an active part of the detector
    ! this information is used later when selecting events and in calculating frac response
    if ( IAND(src%todo_code,TODO_CODE_SRCGTI) /= TODO_CODE_NULL ) then
      call srctool_calc_srcgtis_parallel(instlist, src, ierr)
    else
      ! just copy the input GTIs to the output src GTIs
      call srctool_make_dummy_srcgtis(instlist, src, ierr)
    endif
    if (ierr /= ERROR_LEVEL_OK) then
      call wrserr(subroutinename, ierr, "Problem calculating src GTIs")
      return
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to extract a spectrum, a lightcurve or an eventlist
    if (IAND(src%todo_code,TODO_CODE_SPEC)   /= TODO_CODE_NULL .or. &
        IAND(src%todo_code,TODO_CODE_LC)     /= TODO_CODE_NULL .or. &
        IAND(src%todo_code,TODO_CODE_EVENTS) /= TODO_CODE_NULL  ) then

      call srctool_select_events_parallel(instlist, src, ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem selecting events")
        return
      endif
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to make spectra
    if (IAND(src%todo_code,TODO_CODE_SPEC)   /= TODO_CODE_NULL ) then
      call srctool_spec_setup_speclist(speclist, instlist, src, ierr )
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem setting up spec data structures")
        return
      endif

      call srctool_spec_accumulate_speclist(speclist, instlist, ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem accumulating spectra")
        return
      endif
    endif
    !---------------------------------------------------------------------


    !---------------------------------------------------------------------
    ! see if we need to make light curves
    if (IAND(src%todo_code,TODO_CODE_LC)   /= TODO_CODE_NULL ) then
      call srctool_lc_setup_lclist(lclist, lc_template, instlist, src, ierr )
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem setting up lc data structures")
        return
      endif

      call srctool_lc_accumulate_lclist(lclist, instlist, ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem accumulating light curves")
        return
      endif
    endif
    !---------------------------------------------------------------------



!test123!    !---------------------------------------------------------------------
!test123!    ! see if we need to set up the extended source handling
!test123!    if ((src%extent%code /= EXTTYPE_CODE_POINT) .and. &
!test123!        (IAND(src%todo_code,TODO_CODE_ARF)    /= TODO_CODE_NULL .or. &
!test123!         IAND(src%todo_code,TODO_CODE_LCCORR) /= TODO_CODE_NULL)  ) then
!test123!      call srctool_prepare_extent_model(src, instlist%fft_npix, &
!test123!          instlist%fft_pixscale, ierr)
!test123!        if (ierr /= ERROR_LEVEL_OK) then
!test123!        call wrserr(subroutinename, ierr, "Problem preparing extent model")
!test123!        return
!test123!      endif
!test123!    endif
!test123!
!test123!    !---------------------------------------------------------------------



    !---------------------------------------------------------------------
    ! see if we need to set up an arf
    if (IAND(src%todo_code,TODO_CODE_ARF   ) /= TODO_CODE_NULL ) then
      call srctool_arf_setup_arflist(arflist, instlist, src, ierr )
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem setting up arf data structures")
        return
      endif
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to set up an rmflist
    if (IAND(src%todo_code,TODO_CODE_RMF   ) /= TODO_CODE_NULL ) then
!      call srctool_rmf_setup_rmfsetlist(rmfsetlist, instlist, src, ierr )
      call srctool_rmf_setup_rmflist(rmflist, instlist, src, ierr )
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem setting up output RMF data structures")
        return
      endif
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to make an ARF or calculate a corrected LC
    if (IAND(src%todo_code,TODO_CODE_ARF   ) /= TODO_CODE_NULL .or. &
        IAND(src%todo_code,TODO_CODE_LCCORR) /= TODO_CODE_NULL ) then

      call srctool_calc_fracrsps_parallel(instlist, arflist, lclist, src, ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem calculating effective areas")
        return
      endif
    endif
    !---------------------------------------------------------------------



    !---------------------------------------------------------------------
    ! see if we need to do any coadding
    if (IAND(src%todo_code,TODO_CODE_ARF   ) /= TODO_CODE_NULL ) then
      call srctool_arf_coadd_arflist(arflist, instlist )
    endif
    if (IAND(src%todo_code,TODO_CODE_RMF   ) /= TODO_CODE_NULL ) then
!      call srctool_rmf_coadd_rmfsetlist(rmfsetlist, arflist, instlist )
      call srctool_rmf_coadd_rmflist(rmflist, arflist, instlist )
    endif
    if (IAND(src%todo_code,TODO_CODE_LC    ) /= TODO_CODE_NULL .or. &
        IAND(src%todo_code,TODO_CODE_LCCORR) /= TODO_CODE_NULL ) then
      call srctool_lc_coadd_lclist(lclist, instlist )
    endif

    call system_clock(clock_count)
    write (str_buffer, '(A,A,A,A,F12.5,A)') PROCTIME_PREFIX, &
        "------------------ Finished Source: '", trim(src%name), "' calculation time: ", &
        real((clock_count - clock_count_src), real32)/real(clock_rate, real32), &
        "s -----------------------"
    call wrschatty(subroutinename,str_buffer )
    !---------------------------------------------------------------------



    !---------------------------------------------------------------------
    ! see if we need to write out any spectral files
    if (IAND(src%todo_code,TODO_CODE_SPEC)   /= TODO_CODE_NULL ) then
      ! see if need to apply the arf keywords to the spectrum files
      if (IAND(src%todo_code,TODO_CODE_ARF) /= TODO_CODE_NULL ) then
        call srctool_set_speclist_keys_from_arflist(instlist, arflist, speclist, ierr)
        if (ierr /= ERROR_LEVEL_OK) then
          call wrserr(subroutinename, ierr, "Problem setting spec keywords")
          return
        endif
      endif

      call srctool_spec_coadd_speclist(speclist, instlist )
      call srctool_spec_write_speclist(speclist, instlist, ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem writing spec files")
        return
      endif
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to finish off and write lightcurve files
    if (IAND(src%todo_code,TODO_CODE_LC    ) /= TODO_CODE_NULL .or. &
        IAND(src%todo_code,TODO_CODE_LCCORR) /= TODO_CODE_NULL ) then

      call srctool_lc_calc_lclist_rates(lclist,ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem finishing off lightcurves")
        return
      endif

      call srctool_lc_write_lclist(lclist, instlist, ierr)
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem writing lc files")
        return
      endif
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to write arf files
    if (IAND(src%todo_code,TODO_CODE_ARF   ) /= TODO_CODE_NULL ) then
      call srctool_arf_write_arflist(arflist, instlist, ierr )
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem writing output ARF files")
        return
      endif
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to write rmf files
    if (IAND(src%todo_code,TODO_CODE_RMF   ) /= TODO_CODE_NULL ) then
!      call srctool_rmf_write_rmfsetlist(rmfsetlist, instlist, ierr )
      call srctool_rmf_write_rmflist(rmflist, instlist, ierr )
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem writing output RMF files")
        return
      endif
    endif
    !---------------------------------------------------------------------

    !---------------------------------------------------------------------
    ! see if we need to write source specific evtlist files
    if (IAND(src%todo_code,TODO_CODE_EVENTS ) /= TODO_CODE_NULL ) then
      ! write source
       call srctool_write_source_specific_evtlists(instlist, src, .true., ierr)
       if (ierr /= ERROR_LEVEL_OK) then
          call wrserr(subroutinename, ierr, "Problem writing source eventlist files")
          return
       endif

       ! write background
       call srctool_write_source_specific_evtlists(instlist, src, .false., ierr)
       if (ierr /= ERROR_LEVEL_OK) then
          call wrserr(subroutinename, ierr, "Problem writing background eventlist files")
          return
       endif
    endif

    !---------------------------------------------------------------------


    !---------------------------------------------------------------------
    ! This is currently for debug use only
    ! see if we need to write samplelist files
    if (IAND(src%todo_code,TODO_CODE_SAMPLE ) /= TODO_CODE_NULL ) then
      call srctool_samplelist_write_samplelists(instlist, src, ierr )
      if (ierr /= ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem writing samplelist files")
        return
      endif
    endif
    !---------------------------------------------------------------------



    ! tidy up
    call srctool_spec_empty_speclist(speclist)
    call srctool_lc_empty_lclist(lclist)
    call srctool_arf_empty_arflist(arflist)
!    call srctool_rmf_empty_rmfsetlist(rmfsetlist)
    call srctool_rmf_empty_rmflist(rmflist)

    call system_clock(clock_count)
    write (str_buffer, '(A,A,A,A,F12.5,A)') PROCTIME_PREFIX, &
        "------------------ Finished Source: '", trim(src%name), "' total proc time: ", &
        real((clock_count - clock_count_src), real32)/real(clock_rate, real32), &
        "s -----------------------"
    call wrschatty(subroutinename,str_buffer )

  end subroutine srctool_do_all_processing_for_source

  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  !--------------------------- Main work routine ---------------------------------
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  ! note, subroutine srctool_test_harness moved to srctool_test_mod.f90

!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
end module  srctool_engine
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
!-------------------------------------------------------------------------------
