! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_param

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-24
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling input params
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_regions   ! generic region handling routines
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !
  use srctool_lc                !
  use srctool_extmap

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_PARAM_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_param
  private :: srctool_param_interpret_insts_par
  private :: srctool_param_interpret_todo_par
  private :: srctool_param_interpret_exttype_par
  private :: srctool_param_interpret_extent_pars
  private :: srctool_param_interpret_lctype_par
  private :: srctool_param_interpret_psftype_par
  private :: srctool_param_init_src
  private :: srctool_param_read_srclist_from_fits
  private :: srctool_param_read_srclist_from_string
  private :: srctool_param_read_srclist_from_ascii
  private :: srctool_param_interpret_region_desc
  private :: srctool_param_copy_regions_across_sources
  private :: srctool_param_interpret_src_pars

  public ::  srctool_param_empty_srclist
  public ::  srctool_param_empty_extent
  public ::  srctool_param_print_pars
  public ::  srctool_param_interpret_lc_pars
  public ::  srctool_param_interpret_pars
  public ::  srctool_param_read_pars
  public ::  srctool_param_verify_pars
  public ::  srctool_param_init_pars

  public ::  srctool_param_env_real64

contains

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of PARAMETER handling -----------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_param_init_pars(paramlist,ierr)

    !---------------------------------------------------------
    ! Calls subroutine OPAST to open parameter and log files
    !---------------------------------------------------------
    character(len=*), parameter :: subroutinename='srctool_param_init_pars'

    type(srctool_paramlist_type), intent(inout) :: paramlist
    integer(kind=int32)          , intent(out)   :: ierr

    integer(kind=int32)                        :: status

    ierr=ERROR_LEVEL_OK
    status=0

    call opast("srctool",paramlist%unit,status)

    paramlist%num_evtfiles = 0
    paramlist%num_todo     = 0
    paramlist%num_insts    = 0
!    paramlist%num_refarfs  = 0
!!    paramlist%num_ebounds  = 0
!    paramlist%num_vigndesc = 0
!    paramlist%num_psfdesc  = 0
!    paramlist%num_badpix   = 0
    paramlist%num_extpars  = 0
    paramlist%num_lcpars   = 0
    paramlist%num_lcemin   = 0
    paramlist%num_lcemax   = 0
    paramlist%num_xgrid    = 0

    if (status .ne. 0) then
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,'Error opening parameter file')
    end if

  end subroutine srctool_param_init_pars
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_param_read_pars(paramlist,ierr)

    !------------------------------------------------------------
    ! Reads each of the input parameters in turn
    !------------------------------------------------------------
    character(len=*), parameter :: subroutinename='srctool_param_read_pars'

    type(srctool_paramlist_type) , intent(inout)  :: paramlist
    integer(kind=int32)       , intent(out)       :: ierr
    integer(kind=int32)                           :: status

    ierr = ERROR_LEVEL_OK
    status = 0

    call rgetca(paramlist%unit, "eventfiles" , MAX_EVTFILES, paramlist%evtfiles , paramlist%num_evtfiles ,status)
    if (status .ne. 0) ierr = max(ierr,status)
    if ( (paramlist%num_evtfiles .eq. 1) .and. (TRIM(paramlist%evtfiles(1)) .eq. "") ) then
      paramlist%num_evtfiles = 0
    end if

    call rgetc(paramlist%unit, "srccoord" , paramlist%srccoord , status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "prefix" , paramlist%prefix , status)
    ierr = max(ierr,status)
    call rgetc(paramlist%unit, "suffix" , paramlist%suffix , status)
    ierr = max(ierr,status)


    call rgetca(paramlist%unit, "todo", MAX_TODO_TASKS, paramlist%todo, paramlist%num_todo , status)
    ierr = max(ierr,status)
    if ( (paramlist%num_todo .eq. 1) .and. (TRIM(paramlist%todo(1)) .eq. "") ) then
      paramlist%num_todo = 0
    end if

    call rgetia(paramlist%unit, "insts", MAX_INST, paramlist%insts, paramlist%num_insts, status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "srcreg"  , paramlist%srcreg  , status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "backreg" , paramlist%backreg , status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "exttype"  , paramlist%exttype , status)
    ierr = max(ierr,status)

    call rgetda(paramlist%unit,"extpars" ,MAX_EXTPARS, paramlist%extpars, paramlist%num_extpars, status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "extmap"  , paramlist%extmap , status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "lctype"  , paramlist%lctype  , status)
    ierr = max(ierr,status)

    call rgetda(paramlist%unit,"lcpars" ,MAX_LCPARS, paramlist%lcpars, paramlist%num_lcpars, status)
    ierr = max(ierr,status)

    call rgetda(paramlist%unit,"lcemin" ,MAX_LCBANDS, paramlist%lcemin, paramlist%num_lcemin, status)
    ierr = max(ierr,status)

    call rgetda(paramlist%unit,"lcemax" ,MAX_LCBANDS, paramlist%lcemax, paramlist%num_lcemax, status)
    ierr = max(ierr,status)

    call rgetd(paramlist%unit, "lcgamma", paramlist%lcgamma, status)
    ierr = max(ierr,status)

    call rgetd(paramlist%unit, "tstep", paramlist%tstep, status)
    ierr = max(ierr,status)

!    call rgetd(paramlist%unit, "xgrid", paramlist%xgrid, status)
!    if (status .ne. 0) ierr = max(ierr,status)
    call rgetda(paramlist%unit, "xgrid", MAX_XGRIDPARS, paramlist%xgrid,  paramlist%num_xgrid,status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "gtitype" , paramlist%gtitype , status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "psftype" , paramlist%psftype , status)
    ierr = max(ierr,status)

    ! read the flagsel parameters as strings and the interpret later
    call rgetc(paramlist%unit, "pat_sel" , paramlist%str_pat_sel, status)
    ierr = max(ierr,status)
    call rgetc(paramlist%unit, "flagsel" , paramlist%str_flagsel, status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "pat_sel" , paramlist%str_pat_sel, status)
    ierr = max(ierr,status)

    call rgetc(paramlist%unit, "outsrcreg", paramlist%out_srcreg, status)
    ierr = max(ierr,status)
    call rgetc(paramlist%unit, "outbackreg", paramlist%out_backreg, status)
    ierr = max(ierr,status)

    call rgetl(paramlist%unit, "continue_on_error", &
         paramlist%continue_on_error, status)
    ierr = max(ierr,status)

    call rgetl(paramlist%unit, "clobber", paramlist%clobber, status)
    ierr = max(ierr,status)

    if (ierr .ne. ERROR_LEVEL_OK ) call wrserr(subroutinename, ierr, "Error reading parameters")

  end subroutine srctool_param_read_pars
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_param_print_pars(paramlist)

    !------------------------------------------------------------
    ! Do some debug reporting of the input parameters
    !------------------------------------------------------------
    character(len=*), parameter :: subroutinename='srctool_param_print_pars'
    character(len=*), parameter :: str_stem='Interpreted input parameters:'

    type(srctool_paramlist_type) , intent(inout)  :: paramlist
    character(len=LSTR_MAX) :: str_buffer
    character(len=LSTR_MAX) :: str_buffer2

!    CALL GET_COMMAND(str_buffer)
!    call wrschatty(subroutinename, "Full command line: " // str_buffer)


    write (str_buffer, '(A,1X,A12,A)'  ) str_stem, "PARNAME",    "=VALUE"
    call wrschatty(subroutinename, str_buffer)
    call convert_array_to_string  (paramlist%evtfiles, paramlist%num_evtfiles, str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "eventfiles","='", TRIM(str_buffer2) ,"'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "prefix",    "='", TRIM(paramlist%prefix) ,"'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "suffix",    "='", TRIM(paramlist%suffix) ,"'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "srccoord",  "='", TRIM(paramlist%srccoord) ,"'"
    call wrschatty(subroutinename, str_buffer)

    call convert_array_to_string  (paramlist%todo, paramlist%num_todo, str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "todo",      "='", TRIM(ADJUSTL(str_buffer2)) ,"'"
    call wrschatty(subroutinename, str_buffer)
    call convert_array_to_string  (paramlist%insts, paramlist%num_insts, str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "insts",     "='", TRIM(ADJUSTL(str_buffer2)) ,"'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "srcreg",    "='", TRIM(paramlist%srcreg) ,"'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "backreg",   "='", TRIM(paramlist%backreg) ,"'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "exttype",   "='", TRIM(paramlist%exttype) ,"'"
    call wrschatty(subroutinename, str_buffer)
    call convert_array_to_string  (paramlist%extpars, paramlist%num_extpars, '(F12.7)', str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)' ) str_stem,  "extpars",   "='", TRIM(ADJUSTL(str_buffer2)), "'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "extmap",    "='", TRIM(paramlist%extmap) ,"'"
    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "lctype",    "='", TRIM(paramlist%lctype) ,"'"
    call wrschatty(subroutinename, str_buffer)

    call convert_array_to_string  (paramlist%lcpars, paramlist%num_lcpars, '(F12.3)', str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "lcpars",    "='", TRIM(ADJUSTL(str_buffer2)) ,"'"
    call wrschatty(subroutinename, str_buffer)

    call convert_array_to_string  (paramlist%lcemin, paramlist%num_lcemin, '(F9.4)', str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "lcemin",    "='", TRIM(ADJUSTL(str_buffer2)) ,"'"
    call wrschatty(subroutinename, str_buffer)

    call convert_array_to_string  (paramlist%lcemax, paramlist%num_lcemax, '(F9.4)', str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "lcemax",    "='", TRIM(ADJUSTL(str_buffer2)) ,"'"
    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A,1X,A12,A,F10.5)') str_stem, "lcgamma",   "=",  paramlist%lcgamma
    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A,1X,A12,A,F10.5,A)') str_stem, "tstep",   "=",  paramlist%tstep
    call wrschatty(subroutinename, str_buffer)

    call convert_array_to_string  (paramlist%xgrid, paramlist%num_xgrid, '(F9.5)', str_buffer2, " ")
    write (str_buffer, '(A,1X,A12,A,A,A)' ) str_stem, "xgrid",      "='", TRIM(ADJUSTL(str_buffer2)) ,"'"
    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "gtitype",   "='", TRIM(paramlist%gtitype) ,"'"
    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "psftype",   "='", TRIM(paramlist%psftype) ,"'"
    call wrschatty(subroutinename, str_buffer)

!    write (str_buffer, '(A,1X,A12,A,A,A)'  ), str_stem, "pat_sel",   "='", TRIM(paramlist%str_pat_sel) ,"'"
!    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "flagsel",   "='", TRIM(paramlist%str_flagsel) ,"'"
    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "outsrcreg", "='", TRIM(paramlist%out_srcreg) ,"'"
    call wrschatty(subroutinename, str_buffer)
    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "outbackreg","='", TRIM(paramlist%out_backreg) ,"'"
    call wrschatty(subroutinename, str_buffer)

    write (str_buffer, '(A,1X,A12,A,A,A)'  ) str_stem, "clobber",   "='", logical2yesno(paramlist%clobber) ,"'"
    call wrschatty(subroutinename, str_buffer)


    ! write the same info into a single string for insertion into headers of all output products

  end subroutine srctool_param_print_pars
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_param_verify_pars(paramlist,ierr)

    !------------------------------------------------------------
    ! Verify that the supplied params are sensible and complete
    ! set ierr = ERROR_LEVEL_OK if all is well
    !------------------------------------------------------------
    character(len=*), parameter :: subroutinename='srctool_param_verify_pars'

    type(srctool_paramlist_type) , intent(inout)  :: paramlist
    integer(kind=int32)           , intent(out)  :: ierr

    ierr = ERROR_LEVEL_OK
    if ( paramlist%num_evtfiles .gt. 1 ) then
    end if
    ! TODO actually run some checks!
    ! TODO
    ! TODO
    ! TODO
    ! TODO
    ! TODO
    ! TODO
    ! TODO
    ! TODO
    ! TODO
    ! TODO
    ! TODO

  end subroutine srctool_param_verify_pars
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! interpret the value of the list of instrument names/ids
  subroutine srctool_param_interpret_insts_par(instlist, insts, num_insts, ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_insts_par'
    type(instlist_type), intent(inout)  :: instlist
    integer(kind=int32), intent(in)     :: insts(1:MAX_INST)
    integer(kind=int32), intent(in)     :: num_insts
    integer(kind=int32), intent(out)    :: ierr
    integer(kind=int32)        :: i
    character(len=STR_MAX) :: str_buffer
    character(len=STR_MAX) :: str_inst
    ierr=ERROR_LEVEL_OK
    ! init to false
    instlist%inst(:)%to_consider = .false.

    if ((num_insts .gt. MAX_INST) .or. &
        (num_insts .lt. 1) ) then
      ierr = ERROR_LEVEL_SEVERE
      write (str_buffer, '(A,I0,A,I0,A)') &
          "Error interpreting 'insts' param, the number of supplied instruments, ", num_insts, ", is out of the valid range [1-", MAX_INST, "]"
      call wrserr(subroutinename,ierr,str_buffer)
    end if

    !check special case
    if ((num_insts .eq. 1 ) .and. &
        (insts(1)  .eq. 0 ) ) then
      instlist%inst(:)%to_consider = .true. ! shortcut to allow all to be set to true
    else
      do i=1,num_insts
        if ( (insts(i) .ge. 1       ) .and. &
             (insts(i) .le. MAX_INST) ) then
          instlist%inst(insts(i))%to_consider = .true.
        else
          write (str_buffer, '(A,I0,A,I0,A)') &
              "Error interpreting this instrument:", insts(i), ", It is out of the valid range [1-", MAX_INST, "]"
          ierr = ERROR_LEVEL_SEVERE
          call wrserr(subroutinename,ierr,str_buffer)
          return
        end if
      end do
    end if

    write(str_buffer,*) "Using the following instruments (TM*): "
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        write(str_inst,'(I0)') i
      else
        str_inst="-"
      end if
      write(str_buffer,'(A,A)') TRIM(str_buffer), TRIM(str_inst)
    end do
    call wrschatty (subroutinename, str_buffer)

    ! record the number of valid instruments
    instlist%num_inst_to_consider = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        instlist%num_inst_to_consider = instlist%num_inst_to_consider + 1
      end if
    end do

  end subroutine srctool_param_interpret_insts_par
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! interpret the value of the todo list
  subroutine srctool_param_interpret_todo_par(todo, num_todo, todo_code, ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_todo_par'
    character(len=STR_MAX), intent(in)  :: todo(MAX_TODO_TASKS)
    integer(kind=int32), intent(in)     :: num_todo
    integer(kind=int32), intent(out)    :: todo_code
    integer(kind=int32), intent(out)    :: ierr
    integer(kind=int32)        :: i
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK
    todo_code = TODO_CODE_NULL
    todo_code = IOR(todo_code,TODO_CODE_SRCGTI)  ! do SRCGTI by default

    if ( num_todo > MAX_TODO_TASKS ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, "Error interpreting todo parameter")
      return
    end if

    do i=1,num_todo
      ! match against list of valid options
      select case ( to_upper_fun(trim(todo(i))) )
      case ( "ALL" )
        todo_code = IOR(todo_code,TODO_CODE_ALL)
      case ( "SPEC" )
        todo_code = IOR(todo_code,TODO_CODE_SPEC)
      case ( "ARF" )
        todo_code = IOR(todo_code,TODO_CODE_ARF)
      case ( "RMF" )
        todo_code = IOR(todo_code,TODO_CODE_RMF)
      case ( "LC" )
        todo_code = IOR(todo_code,TODO_CODE_LC)
      case ( "LCCORR" )
        todo_code = IOR(todo_code,TODO_CODE_LCCORR)
      case ( "EVENTS" )
        todo_code = IOR(todo_code,TODO_CODE_EVENTS)
      case ( "SAMPLE" )
        todo_code = IOR(todo_code,TODO_CODE_SAMPLE)
      case ( "STACK" )
        todo_code = IOR(todo_code,TODO_CODE_STACK)
      case ( "SRCGTI" )
        todo_code = IOR(todo_code,TODO_CODE_SRCGTI)
      case ( "NOSRCGTI" )
        ! note the opposite sense of this one
        todo_code = IAND(todo_code,NOT(TODO_CODE_SRCGTI))
      case ( "TEST" )
        todo_code = IOR(todo_code,TODO_CODE_TEST)
      case default
        ierr = ERROR_LEVEL_SEVERE
        write(str_buffer,*) "Unknown task in todolist: ", trim(todo(i))
        call wrserr(subroutinename,ierr,str_buffer)
      end select
    end do

  end subroutine srctool_param_interpret_todo_par
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! interpret the value of the exttype parameter
  subroutine srctool_param_interpret_exttype_par(exttype, exttype_code, ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_exttype_par'
    character(len=*), intent(in)        :: exttype
    integer(kind=int32), intent(out)    :: exttype_code
    integer(kind=int32), intent(out)    :: ierr

    ierr = ERROR_LEVEL_OK

    ! match against list of valid options
    select case ( to_upper_fun(trim(exttype)) )
    case ( "POINT", "NONE", "DELTA", "" )
      exttype_code = EXTTYPE_CODE_POINT
    case ( "GAUSSIAN", "GAUSS" )
      exttype_code = EXTTYPE_CODE_GAUSSIAN
    case ( "BETA" )
      exttype_code = EXTTYPE_CODE_BETA
    case ( "TOPHAT" )
      exttype_code = EXTTYPE_CODE_TOPHAT
    case ( "MAP", "IMAGE" )
      exttype_code = EXTTYPE_CODE_MAP
!      flag = .true.
!    case ( "PHOTON", "PHOT" )
!      exttype_code = EXTTYPE_CODE_PHOTON
!      flag = .true.
!    case ( "CATALOGUE", "CAT", "CATALOG", "ERMLDET" )
!      exttype_code = EXTTYPE_CODE_CATALOGUE
!      flag = .true.
    case default
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Unknown exttype description: " // TRIM(exttype))
      return
    end select

  end subroutine srctool_param_interpret_exttype_par
  !---------------------------------------------------------------------------------------------






  !---------------------------------------------------------------------------------------------
  subroutine srctool_param_interpret_extent_pars(paramlist,srclist,ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_extent_pars'
    type(srctool_paramlist_type), intent(inout)  :: paramlist
    type(srclist_type),   intent(inout), target  :: srclist
    integer(kind=int32) , intent(out)  :: ierr
    logical  :: numpar_flag
    logical  :: parval_flag
    character(len=STR_MAX) :: str_buffer

    ierr=ERROR_LEVEL_OK
    numpar_flag = .false.
    parval_flag = .false.

    ! assumes that have already interpretted the exttype param into a code

    ! check that other extent params are consistent with choice of exttype
    if ( paramlist%exttype_code .ne. EXTTYPE_CODE_MAP ) then
      if ( trim(paramlist%extmap) .ne. "" ) then
        call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
            "Warning. You supplied an extraneous extent map parameter for exttype=" // &
             EXTTYPE_CODE_STR(paramlist%exttype_code))
      endif
    endif

    str_buffer = ""
    select case ( paramlist%exttype_code )
    case ( EXTTYPE_CODE_POINT )
      if ( paramlist%num_extpars .gt. 0 ) then
        numpar_flag = .true.
      endif
    case ( EXTTYPE_CODE_GAUSSIAN, EXTTYPE_CODE_BETA )
      if ( (paramlist%num_extpars .ne. 1) .and. (paramlist%num_extpars .ne. 2) ) then
        numpar_flag = .true.
      else
        if ( paramlist%extpars(1) .le. 0.0d0 ) then
          parval_flag = .true.
          write (str_buffer, '(A,E10.4)') &
              "Illegal extent parameter value: extpar[1] = ", paramlist%extpars(1)
          call wrserr(subroutinename, ierr, str_buffer )
        endif
        if ( paramlist%num_extpars .eq. 2) then
          if ( paramlist%extpars(2) .le. 0.0d0 ) then
            parval_flag = .true.
            write (str_buffer, '(A,E10.4)') &
                "Illegal extent parameter value: extpar[2] = ", paramlist%extpars(2)
            call wrserr(subroutinename, ierr, str_buffer )
          endif
        endif
      endif
    case ( EXTTYPE_CODE_TOPHAT )
      if ( paramlist%num_extpars .ne. 1 ) then
        numpar_flag = .true.
      else
        if ( paramlist%extpars(1) .le. 0.0d0 ) then
          parval_flag = .true.
          write (str_buffer, '(A,E10.4)') &
              "Illegal extent parameter value: extpar[1] = ", paramlist%extpars(1)
          call wrserr(subroutinename, ierr, str_buffer )
        endif
      endif
    case ( EXTTYPE_CODE_MAP )
      if ( paramlist%num_extpars .ne. 0 ) then
        numpar_flag = .true.
      endif
    case default
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr, "Unknown exttype description: " // TRIM(paramlist%exttype))
      return
    end select

    if ( numpar_flag ) then
      call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
            "Warning.You supplied an illegal number of extent parameter(s) for exttype=" &
             // EXTTYPE_CODE_STR(paramlist%exttype_code) )
    endif

    if ( parval_flag ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, &
            "You supplied at least one illegal extent parameter for exttype=" &
             // EXTTYPE_CODE_STR(paramlist%exttype_code))
     return
    endif






    !----------------------------------------------------------------------------
    ! If exttype = MAP then attempt to read the map describing the source extent
    if ( paramlist%exttype_code .eq. EXTTYPE_CODE_MAP ) then
       call srctool_extmap_read_from_file(paramlist%extmap, srclist%psrc(1)%extmap, &
            .true., ierr)
       if ( ierr /= ERROR_LEVEL_OK ) then
          call wrserr(subroutinename, ierr, "Error interpreting extmap parameter")
          return
       end if
       call srctool_extmap_update_source_coord(srclist%psrc(1)%extmap, &
            srclist%psrc(1)%coord)
    end if
    !----------------------------------------------------------------------------



  end subroutine srctool_param_interpret_extent_pars
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  subroutine srctool_param_interpret_src_pars(paramlist,srclist,ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_src_pars'
    type(srctool_paramlist_type), intent(in)  :: paramlist
    type(srclist_type),   intent(inout), target  :: srclist

    integer(kind=int32) , intent(out)  :: ierr
    integer                 :: status
    logical                 :: exist_flag
    character(len=STR_MAX)  :: read_state
!    character(len=LSTR_MAX) :: str_buffer
    integer(kind=int32)     :: unit
    logical                 :: is_fits
!    type(coord_type)        :: local_coord
    integer(kind=int32)     :: i
    character(len=LSTR_MAX) :: srccoord_copy

    ierr=ERROR_LEVEL_OK
    status = 0

    call wrschatty (subroutinename, &
        "Interpreting source description: '" // TRIM(paramlist%srccoord) // "'")


    if (LEN_TRIM(paramlist%srccoord) .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Source description parameter is blank!")
      return
    end if

    !The srccoord param can contain one of the following:
    ! 1) The filename of a file containing a list of source
    !    coordinates in one of the following formats:
    !   1a) ascii - each line interpreted as 2) or 3) below
    !   1b) fits
    ! 2) A one line description of the source coordinates
    !    coordinates in one of the following formats:
    !   2a) [coordsys;] LONGITUDE_DEG , LATITUDE_DEG
    !   2b) [coordsys;] LONGITUDE_H:M:S , LATITUDE_D:M:S
    !   2c) [coordsys;] LONGITUDE_H LONGITUDE_M LONGITUDE_S , LATITUDE_D LATITUDE_M LATITUDE_S
    ! 3) Description of the source coordinates in DS9 region file format
    !   3a) [coordsys;] region_type  LONGITUDE  LATITUDE [additional region specifications]
    !   3b) [coordsys;] region_type  LONGITUDE, LATITUDE, [additional region specifications]
    !   3c) [coordsys;] region_type (LONGITUDE, LATITUDE, [additional region specifications]


    ! first check to see if the original srccoord param contains a valid filename
    inquire(FILE=TRIM(paramlist%srccoord), IOSTAT=status, EXIST=exist_flag, READ=read_state)

!    write(*,'(A,A,A,I0,A,A)'), "STEP1: filename='", &
!        TRIM(paramlist%srccoord), "' status= ", status, " exist_flag=", logical2yesno(exist_flag)

    ! If a file with this exact name doesn't exist then perhaps it's a fits
    ! file with an extension specification and/or a selection expression
    if ((status .eq. 0) .and. .not. exist_flag ) then
      ! try to trim off any bits of the filename which are fits HDU/selection expressions
      call FTRTNM(TRIM(paramlist%srccoord), srccoord_copy, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if
      ! now see if the trimmed version of the filename exists
      inquire(FILE=TRIM(srccoord_copy), IOSTAT=status, EXIST=exist_flag, READ=read_state)
!      write(*,'(A,A,A,I0,A,A)'), "STEP2: filename='", &
!          TRIM(srccoord_copy), "' status= ", status, " exist_flag=",  logical2yesno(exist_flag)
    endif

    if ((status .eq. 0) .and. exist_flag ) then    ! ok great, I found the name of a file which exists
      ! check that can actually read this file:
      if ( TRIM(read_state) .ne. "YES" ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, &
            "SRCTOOL does not have read access to the file named in 'srccoord' param: '" // TRIM(paramlist%srccoord) // "'" )
        return
      end if

      call wrschatty (subroutinename, "Interpreted 'srccoord' parameter as a filename")

      ! get an i/o unit - use the fitsio routine for this
      call FTGIOU(unit, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      ! now try to open as a fits file
      ! open the evt file and move to whichever extension is specified
      call FTTOPN(unit, TRIM(paramlist%srccoord), FITSIO_READONLY, status)
      if (status .eq. 0) then
        is_fits = .true.
      else
        is_fits = .false.
      end if
      status = 0

      !close the file
      call FTCLOS(unit, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      !and release the unit number
      call FTFIOU(unit, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if


      if ( is_fits ) then
        srclist%type_code = SRCLIST_TYPE_CODE_FITSCAT
        call wrschatty (subroutinename, "Interpreted srccoord as a FITS file")

        call srctool_param_read_srclist_from_fits(TRIM(paramlist%srccoord), srclist, ierr)

        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem reading from srccoord fits file: " // TRIM(paramlist%srccoord) )
          return
        end if

      else
        status = 0
        srclist%type_code = SRCLIST_TYPE_CODE_ASCIICAT
      ! if not fits then try as an ascii file
      ! first need to determine the number of sources so that can allocate the correct amount of memory.

        call wrschatty (subroutinename, "Interpreted srccoord as an ascii catalogue")
!        write ( *, '(A,A,A,I0)'), &
!            "Couldn't open file as fits: '", TRIM(paramlist%srccoord), "' status=", status

        call srctool_param_read_srclist_from_ascii(TRIM(paramlist%srccoord), srclist, ierr)

        ! set all of the sources to have the same extent type
        srclist%psrc(:)%extent%code   = paramlist%exttype_code

        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem reading from srccoord ascii file: " // TRIM(paramlist%srccoord) )
          return
        end if

      end if


    else  ! now try as a coordinate string describing a single source
      srclist%type_code = SRCLIST_TYPE_CODE_INLINE
      call wrschatty (subroutinename, "Interpreted srccoord as a single coordinate pair")
      call srctool_param_read_srclist_from_string(TRIM(paramlist%srccoord), srclist, ierr)

      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem reading from srccoord string: " // TRIM(paramlist%srccoord))
        return
      end if

      ! copy the generic extent parameters into the source structure
      srclist%psrc(1)%extent%code       = paramlist%exttype_code
      srclist%psrc(1)%extent%num_pars   = paramlist%num_extpars  ! TODO - move this to srctool_param_interpret_extent_pars()
      srclist%psrc(1)%extent%pars(:)    = paramlist%extpars(:)   ! TODO - move this to srctool_param_interpret_extent_pars()


    end if

    ! this makes life easier later (so that don't have to pass the suffix and prefix strings from function to function)
    if ( paramlist%clobber ) then
      srclist%psrc(:)%str_prefix = "!" // TRIM(paramlist%prefix)
!      write (*,*), "  I will clobber the output"
!      call wrschatty (subroutinename,  "Any existing output files will be clobbered!")
      call wrserr (subroutinename,  ERROR_LEVEL_WARNING, "Any existing output files will be clobbered!")
    else
      srclist%psrc(:)%str_prefix = TRIM(paramlist%prefix)
!      write (*,*), "  I will NOT clobber the output"
    endif
    srclist%psrc(:)%str_suffix = TRIM(paramlist%suffix)

!    ! now convert the src coords into ICRS if necessary
!    do i=1, srclist%num_srcs
!      local_coord = srclist%psrc(i)%coord
!      srclist%psrc(i)%coord%coordsys_code = COORDSYS_CODE_ICRS
!      call geom_convert_coordsys (local_coord, srclist%psrc(i)%coord )
!   end do


    ! now try to interpret the regions described in the srcreg,backreg strings
    call srctool_param_interpret_region_desc (paramlist%srcreg,  srclist, REG_CLASS_FLAG_SRC, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error when interpreting SRC region descriptions")
      return
    end if
    call srctool_param_interpret_region_desc (paramlist%backreg, srclist, REG_CLASS_FLAG_BACK, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Error when interpreting BACK region descriptions")
      return
    end if

    ! count the number of extended sources, num to extract, num to exclude
    srclist%n_to_extract = 0
    srclist%n_to_exclude = 0
    srclist%n_extended   = 0
    srclist%n_extended_to_extract = 0
    srclist%n_extended_to_exclude = 0
    do i=1,srclist%num_srcs
       associate( src => srclist%psrc(i) )
         if ( src%cat_extract ) srclist%n_to_extract = srclist%n_to_extract + 1
         if ( src%cat_exclude ) srclist%n_to_exclude = srclist%n_to_exclude + 1
         if ( src%extent%code /= EXTTYPE_CODE_POINT ) then
            srclist%n_extended = srclist%n_extended + 1
            if ( src%cat_extract ) then
               srclist%n_extended_to_extract = srclist%n_extended_to_extract + 1
            end if
            if ( src%cat_exclude ) then
               srclist%n_extended_to_exclude = srclist%n_extended_to_exclude + 1
            end if
         end if
       end associate
    end do

    ! take a copy of the todo_code parameter
    srclist%psrc(:)%todo_code = paramlist%todo_code

!    call wrschatty(subroutinename,"DEBUG Number of sources in srclist:    " // trim(i2s(srclist%num_srcs)))
!    call wrschatty(subroutinename,"DEBUG Number of sources to extract:    " // trim(i2s(srclist%n_to_extract )))
!    call wrschatty(subroutinename,"DEBUG Number of sources to exclude:    " // trim(i2s(srclist%n_to_exclude )))
!    call wrschatty(subroutinename,"DEBUG Number of extended sources:      " // trim(i2s(srclist%n_extended )))
!    call wrschatty(subroutinename,"DEBUG Num extended sources to extract: " // trim(i2s(srclist%n_extended_to_extract )))
!    call wrschatty(subroutinename,"DEBUG Num extended sources to exclude: " // trim(i2s(srclist%n_extended_to_exclude )))

!    ! DEBUG printout section:
!    do i=1,srclist%num_srcs
!      write(*,'(A)'),    "-----------------------------------------------"
!      write(*,'(A,I0)'), "-----------Here is the src region description for src= ", i
!      call srctool_print_region (srclist%psrc(i)%src_region, ierr)
!      write(*,'(A,I0)'), "-----------Here is the back region description for src= ", i
!      call srctool_print_region (srclist%psrc(i)%back_region, ierr)
!      write(*,'(A)'),    "-----------------------------------------------"
!    end do

  end subroutine srctool_param_interpret_src_pars
  !---------------------------------------------------------------------------------------------





  !---------------------------------------------------------------------------------------------
  ! interpret the value of the lctype parameter
  subroutine srctool_param_interpret_lctype_par(lctype, lctype_code, ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_lctype_par'
    character(len=*), intent(in)        :: lctype
    integer(kind=int32), intent(out)    :: lctype_code
    integer(kind=int32), intent(out)    :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK

    ! match against list of valid options
    select case ( to_upper_fun(trim(lctype)) )
    case ( "REGULAR" )
      lctype_code = LCTYPE_CODE_REGULAR
    case ( "REGULAR+" )
      lctype_code = LCTYPE_CODE_REGULAR_PLUS
    case ( "REGULAR-")
      lctype_code = LCTYPE_CODE_REGULAR_MINUS
    case ( "SRCGTI" )
      lctype_code = LCTYPE_CODE_SRCGTI
    case ( "SRCGTI-" )
      lctype_code = LCTYPE_CODE_SRCGTI_MINUS
    case ( "GTI" )
      lctype_code = LCTYPE_CODE_GTI
    case ( "FOLDED" )
      lctype_code = LCTYPE_CODE_FOLDED
    case default
      ierr = ERROR_LEVEL_SEVERE
      write(str_buffer,*) "Unknown lctype type: ", trim(lctype)
      call wrserr(subroutinename,ierr,str_buffer)
    end select

  end subroutine srctool_param_interpret_lctype_par
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! interpret the value of the psftype parameter
  subroutine srctool_param_interpret_psftype_par(psftype, psftype_code, ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_psftype_par'
    character(len=*), intent(in)        :: psftype
    integer(kind=int32), intent(out)    :: psftype_code
    integer(kind=int32), intent(out)    :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK

    select case ( to_upper_fun(trim(psftype)) )
    case ( "NONE", "NO", "FALSE", "-", "" )
      psftype_code = SRCTOOL_PSF_CLASS_NONE
    case ( "2D", "2D_PSF", "PSF_2D", "2D_MAP", "PSFMAP", "MAP" )
      psftype_code = SRCTOOL_PSF_CLASS_MAP
    case ( "SHAPELET", "SHAPELETS" )
      psftype_code = SRCTOOL_PSF_CLASS_SHAPELETS
    case ( "ANALYTIC" )
      psftype_code = SRCTOOL_PSF_CLASS_ANALYTIC
    case default
      ierr = ERROR_LEVEL_SEVERE
      write(str_buffer,*) "Unknown psftype type: ", trim(psftype)
      call wrserr(subroutinename,ierr,str_buffer)
    end select

  end subroutine srctool_param_interpret_psftype_par
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine srctool_param_interpret_lc_pars(paramlist,lc_template,ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_lc_pars'
    type(srctool_paramlist_type), intent(inout)  :: paramlist
    type(lc_type),   intent(inout)  :: lc_template
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32)      :: status
    integer(kind=int32)      :: e
    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK
    status = 0

    write (str_buffer, '(A,A,A)') "Setting up lc params (lctype= ",TRIM(paramlist%lctype),")"
    call wrschatty (subroutinename, str_buffer)

    call srctool_param_interpret_lctype_par(paramlist%lctype, paramlist%lctype_code, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Problem interpreting 'lctype' param")
      return
    end if

    ! check that the paramlist%lctype_code is valid
    select case ( paramlist%lctype_code )
    case ( LCTYPE_CODE_REGULAR, LCTYPE_CODE_REGULAR_PLUS, LCTYPE_CODE_REGULAR_MINUS, &
        LCTYPE_CODE_SRCGTI, LCTYPE_CODE_SRCGTI_MINUS, LCTYPE_CODE_GTI )
      ! Ok, so do nothing
    case ( LCTYPE_CODE_FOLDED )
      write (str_buffer, '(A,A,A,I0,A)') "Unimplemented LC type: ", TRIM(paramlist%lctype), " (status=", status, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    case default
      write (str_buffer, '(A,A,A,I0,A)') "Unknown LC type: ", TRIM(paramlist%lctype), " (status=", status, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end select      ! end select on lctype_code
    !--------------------------------

    lc_template%lctype      =  TRIM(paramlist%lctype)
    lc_template%lctype_code =  paramlist%lctype_code
    lc_template%num_lcpars  =  paramlist%num_lcpars
    lc_template%lcpars(:)   =  paramlist%lcpars(:)

    !--------------------------------
    ! now interpret the lcebands
    lc_template%lcgamma = paramlist%lcgamma

    if ( paramlist%num_lcemin .ne. paramlist%num_lcemax ) then
      ierr = ERROR_LEVEL_SEVERE
      write (str_buffer, '(A,I0,A,I0)') &
          "Mismatched number of lc energy bands: num_lcmin=", paramlist%num_lcemin , &
          "num_lcmax=", paramlist%num_lcemax
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end if
    if ( paramlist%num_lcemin .le. 0 ) then
      ! set to the default (full) energy band
      lc_template%num_ebands = 1
      call fitswrap_set_key(lc_template%key_emin(1), "E_MIN1", DEFAULT_E_MIN, &
          "[keV] Minimum energy bound")
      call fitswrap_set_key(lc_template%key_emax(1), "E_MAX1", DEFAULT_E_MAX, &
          "[keV] Maximum energy bound")
    else
      lc_template%num_ebands = paramlist%num_lcemin
      do e=1,lc_template%num_ebands
        write (str_buffer, '(A,I0)') "E_MIN", e
        call fitswrap_set_key(lc_template%key_emin(e), str_buffer, paramlist%lcemin(e), &
            "[keV] Minimum energy bound")
        write (str_buffer, '(A,I0)') "E_MAX", e
        call fitswrap_set_key(lc_template%key_emax(e), str_buffer, paramlist%lcemax(e), &
            "[keV] Maximum energy bound")
      end do
    end if
    !--------------------------------


  end subroutine srctool_param_interpret_lc_pars
  !---------------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  subroutine srctool_param_interpret_pars(paramlist,instlist,srclist,ierr)
    character(len=*), parameter :: subroutinename='srctool_param_interpret_pars'
    type(srctool_paramlist_type), intent(inout)  :: paramlist
    type(instlist_type),  intent(inout)  :: instlist
    type(srclist_type),   intent(inout)  :: srclist
    integer(kind=int32)           , intent(out)  :: ierr
    integer(kind=int32) :: status
    ierr=ERROR_LEVEL_OK

    !----------------------------------------------------------------------------
    !now interpret the todo strings
    call srctool_param_interpret_todo_par(paramlist%todo, paramlist%num_todo, paramlist%todo_code, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr, "Error interpreting 'todo' parameter")
      return
    end if
    ! check to see if we are in test mode,
    ! if so then return before processing other parameters
    if ( IAND(paramlist%todo_code,TODO_CODE_TEST)  .ne. 0 ) then
      call wrschatty(subroutinename, "We will run the internal test harness")
      return
    endif
    !----------------------------------------------------------------------------

    !----------------------------------------------------------------------------
    ! make directories using the prefix
    call make_necessary_dirs_for_new_file (TRIM(paramlist%prefix), status)
    if ( status .ne. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, "Error preparing output directory structure")
      return
    end if
    !----------------------------------------------------------------------------

    !----------------------------------------------------------------------------
    ! interpret the instlist param
    call srctool_param_interpret_insts_par(instlist, paramlist%insts, paramlist%num_insts, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr(subroutinename, ierr, "Error interpreting 'insts' parameter")
      return
    end if
    !----------------------------------------------------------------------------


    !----------------------------------------------------------------------------
    call srctool_param_interpret_exttype_par(paramlist%exttype, paramlist%exttype_code, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr, "Error interpreting 'exttype' parameter")
      return
    end if
    !----------------------------------------------------------------------------


    !----------------------------------------------------------------------------
    ! interpret the params describing the source(s) and extraction regions
    call srctool_param_interpret_src_pars(paramlist, srclist, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrstop(paramlist%unit, subroutinename, "Error interpreting source and/or region parameters")
    endif
    !----------------------------------------------------------------------------


    !----------------------------------------------------------------------------
    ! interpret the params describing the source extent
    call srctool_param_interpret_extent_pars(paramlist, srclist, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr, "Error interpreting extent parameters")
      return
    end if
    !----------------------------------------------------------------------------


    !----------------------------------------------------------------------------
    call srctool_param_interpret_lctype_par(paramlist%lctype, paramlist%lctype_code, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr, "Error interpreting 'lctype' parameter")
      return
    end if
    !----------------------------------------------------------------------------

    !----------------------------------------------------------------------------
    call to_upper(paramlist%gtitype)
    if      ( paramlist%gtitype .eq. "GTI" ) then
      instlist%gtitype_to_read = SRCTOOL_CALTYPE_GTI
    else if ( (paramlist%gtitype .eq. "QUALGTI")  .or. (paramlist%gtitype .eq. "QUAL") ) then
      instlist%gtitype_to_read = SRCTOOL_CALTYPE_QUALGTI
    else if ( (paramlist%gtitype .eq. "FLAREGTI") .or. (paramlist%gtitype .eq. "FLARE") ) then
      instlist%gtitype_to_read = SRCTOOL_CALTYPE_FLAREGTI
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr, "Error interpreting 'gtitype' parameter")
      return
    end if
    !----------------------------------------------------------------------------

    !----------------------------------------------------------------------------
    call srctool_param_interpret_psftype_par(paramlist%psftype, paramlist%psftype_code, ierr)
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr, "Error interpreting 'psftype' parameter")
      return
    end if
    !----------------------------------------------------------------------------

    !----------------------------------------------------------------------------
    call convert_bitmask_string_to_int(paramlist%str_flagsel, instlist%cl_flagsel, ierr)
!    write (*,'(A,A,A,I0)'), "FLAGSEL: input='", trim(paramlist%str_flagsel), "' out=", instlist%cl_flagsel
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr, "Error interpreting 'flagsel' parameter")
      return
    end if
    call convert_bitmask_string_to_int(paramlist%str_pat_sel, instlist%cl_pat_sel, ierr)
!    write (*,'(A,A,A,I0)'), "PAT_SEL: input='", trim(paramlist%str_pat_sel), "' out=", instlist%cl_pat_sel
    if (ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr(subroutinename, ierr, "Error interpreting pat_sel parameter")
      return
    end if
    !----------------------------------------------------------------------------

    !----------------------------------------------------------------------------
    ! take copies of the tstep and xgrid params for ease of use later on
    instlist%inst(:)%tstep = paramlist%tstep
    if (  paramlist%num_xgrid .eq. 1 ) then
      paramlist%xgrid(XGRIDPAR_BACK) = paramlist%xgrid(XGRIDPAR_SRC)
    end if
    instlist%inst(:)%xgrid(XGRIDPAR_SRC)  = paramlist%xgrid(XGRIDPAR_SRC)
    instlist%inst(:)%xgrid(XGRIDPAR_BACK) = paramlist%xgrid(XGRIDPAR_BACK)
    !----------------------------------------------------------------------------


  end subroutine srctool_param_interpret_pars
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of PARAMETER handling -----------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- START of SRCLIST handling ----------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !--------------------------------------------------
  subroutine srctool_param_empty_srclist(srclist)
    type(srclist_type), intent(inout)      :: srclist
    integer(kind=int32) :: i
    if ( allocated(srclist%psrc) ) then
      do i=1,size(srclist%psrc)
        !        if ( allocated(srclist%psrc(i)%uv)) deallocate(srclist%psrc(i)%uv)
        call srctool_param_init_src(srclist%psrc(i))
      end do
      deallocate(srclist%psrc)
    endif
    srclist%num_srcs = 0
    srclist%type_code         = SRCLIST_TYPE_CODE_NULL
    srclist%autoreg_code_src  = AUTOREG_CODE_NULL
    srclist%autoreg_code_back = AUTOREG_CODE_NULL
    srclist%auto_columns_supplied = .false.
    srclist%n_extended   = 0
    srclist%n_to_extract = 0
    srclist%n_to_exclude = 0
  end subroutine srctool_param_empty_srclist
  !------------------------------------------------------------------------------------


  subroutine srctool_param_empty_extent(extent)
    character(len=*), parameter :: subroutinename='srctool_param_empty_extent'
    type(extent_type),   intent(inout)  :: extent

    extent%code  = EXTTYPE_CODE_POINT
    extent%pars(:)   = 0.0d0
    extent%num_pars  = 0
    if (allocated(extent%xy)) deallocate(extent%xy)
    extent%num_pix = 0
    extent%norm_const = 0.0d0

  end subroutine srctool_param_empty_extent


  !-----------------------------------------------------------------------------------
  subroutine srctool_param_init_src(src)
    character(len=*), parameter :: subroutinename='srctool_param_init_src'
    type(src_type),   intent(inout)  :: src
    src%name = ""
    src%index = 0
    call srctool_generic_init_coord(src%coord)
!    src%coord%coordsys_code = COORDSYS_CODE_NULL
    src%coord%long = -huge(src%coord%long) ! would prefer these to be nulls
    src%coord%lat  = -huge(src%coord%lat)  ! would prefer these to be nulls

    call srctool_param_empty_extent(src%extent)
!    src%extent%code  = EXTTYPE_CODE_POINT
!    src%extent%pars(:)   = 0.0d0
!    src%extent%num_pars  = 0
    call srctool_extmap_empty(src%extmap)

    src%do_back = .true.

    call srctool_regions_init_region(src%src_region)
    call srctool_regions_init_region(src%back_region)

    src%str_src_region = ""
    src%str_back_region = ""
    src%nevents_src_i(:) = 0
    src%nevents_back_i(:) = 0
    src%str_prefix = ""
    src%str_suffix = ""
    src%todo_code = TODO_CODE_NULL

    src%cat_scts   = 0.0d0
    src%cat_bg_map = 0.0d0
    src%cat_ext    = 0.0d0
    src%cat_extract = .true.
    src%cat_exclude = .true.

!    if ( allocated(src%prpsf)) deallocate(src%prpsf)
!    if (associated(src%prpsf)) then
      nullify(src%prpsf)
!    endif
  end subroutine srctool_param_init_src
  !-----------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  ! read a sourcelist that is in the standard eSASS CATPREP format
  ! http://erosita.mpe.mpg.de/eROdoc/products/SourceCatalogs_doc.html
  subroutine srctool_param_read_srclist_from_fits(str_filename,srclist,ierr)
    character(len=*), parameter :: subroutinename='srctool_param_read_srclist_from_fits'
    character(len=*), intent(in) ::  str_filename
    type(srclist_type),   intent(inout)  :: srclist
    integer(kind=int32) , intent(out)  :: ierr
    integer                :: status
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)     :: unit

    integer(kind=int32) :: colnum_id
    integer(kind=int32) :: colnum_ra
    integer(kind=int32) :: colnum_dec
    integer(kind=int32) :: colnum_scts
    integer(kind=int32) :: colnum_bg_map
    integer(kind=int32) :: colnum_ext
    integer(kind=int32) :: colnum_extract
    integer(kind=int32) :: colnum_exclude
!    integer(kind=int32) :: colnum_x_ima
!    integer(kind=int32) :: colnum_y_ima

    integer(kind=int32), dimension(:), allocatable :: pid
    real(kind=real64),   dimension(:), allocatable :: pra
    real(kind=real64),   dimension(:), allocatable :: pdec
    real(kind=real64),   dimension(:), allocatable :: pscts
    real(kind=real64),   dimension(:), allocatable :: pbg_map
    real(kind=real64),   dimension(:), allocatable :: pext
    integer(kind=int32), dimension(:), allocatable :: pextract
    integer(kind=int32), dimension(:), allocatable :: pexclude

    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    integer(kind=int32)    :: i
    real(kind=real64)  , parameter :: nullval_real64  = 0.0d0
    integer(kind=int64), parameter :: nullval_int64   = 0
    integer(kind=int32), parameter :: nullval_int32   = 0
    logical           :: exists
    character(len=STR_MAX) :: try_colname

!    type(fitskey_real64_type)  :: key_tcdltx
!    type(fitskey_real64_type)  :: key_tcdlty
!!    real(kind=real64) :: pixel_area_arcsec2
!    real(kind=real64) :: pixel_area_deg2

    status = 0
    ierr=ERROR_LEVEL_OK
!    call fitswrap_init_key(key_tcdltx, "-")
!    call fitswrap_init_key(key_tcdlty, "-")

    ! get an i/o unit - use the fitsio routine for this
    call FTGIOU(unit, status)
    if (status /= 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    ! now try to open the file
    ! open the file and move to whichever extension is specified
    call FTTOPN(unit, TRIM(str_filename), FITSIO_READONLY, status)
    if (status /= 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    !get the number of rows in the extension
    call FTGNRW(unit, srclist%num_srcs, status)
    if (status /= 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    write (str_buffer, '(A,I0,A,A)') &
         "Reading in: ", srclist%num_srcs, " sources from file: ", TRIM(str_filename)
    call wrschatty (subroutinename, str_buffer)

    ! New column name to denote ID of source:
    call fitswrap_prep_col(unit, "ID_SRC", srclist%num_srcs, colnum_id,   pid,  ierr)
    call fitswrap_prep_col(unit, "RA"  ,   srclist%num_srcs, colnum_ra,   pra,  ierr)
    call fitswrap_prep_col(unit, "DEC" ,   srclist%num_srcs, colnum_dec,  pdec, ierr)

    if (ierr /= ERROR_LEVEL_OK ) then
       call wrserr (subroutinename, ierr, "Problem preparing ID_SRC,RA,DEC column(s) for reading from HDU")
       return
    end if

    ! now get the optional columns

    ! The total source counts
    try_colname = "ML_CTS_0"
    call fitswrap_prep_col(unit, try_colname , srclist%num_srcs, colnum_scts,  pscts, ierr)
    if (ierr /= ERROR_LEVEL_OK ) then
       call wrschatty (subroutinename, "There is no ML_CTS_0 column in catalogue file")
       ierr = ERROR_LEVEL_OK
       colnum_scts = 0
    else
       call wrschatty (subroutinename, &
            "Successfully found optional column ML_CTS_0 in catalogue file")
    end if

    ! The local background map count density
    try_colname = "ML_BKG_0"
    call fitswrap_prep_col(unit, try_colname , srclist%num_srcs, colnum_bg_map,  pbg_map, ierr)
    if (ierr /= ERROR_LEVEL_OK ) then
       call wrschatty (subroutinename, "There is no ML_BKG_0 column in catalogue file")
       ierr = ERROR_LEVEL_OK
       colnum_bg_map = 0
    else
       call wrschatty (subroutinename, &
            "Successfully found optional column ML_BKG_0 in catalogue file")
    end if

    ! The source extent column
    call fitswrap_prep_col(unit, "EXT" , srclist%num_srcs, colnum_ext,  pext, ierr)
    if (ierr /= ERROR_LEVEL_OK ) then
       call wrschatty (subroutinename, "There is no EXT column in catalogue file")
       ierr = ERROR_LEVEL_OK
       colnum_ext = 0
    else
       call wrschatty (subroutinename, &
            "Successfully found optional column EXT in catalogue file")
    end if

    ! The extract and exclude columns
    try_colname = "AUTO_EXTRACT"
    colnum_extract = 0
    call fitswrap_test_for_col(unit, try_colname, exists)
    if ( exists ) then
       call fitswrap_prep_col(unit, try_colname , srclist%num_srcs, colnum_extract,  pextract, ierr)
       if (ierr /= ERROR_LEVEL_OK ) then
          call wrschatty (subroutinename, &
               "Problem reading " // trim(try_colname) // " column from catalogue file")
          ierr = ERROR_LEVEL_OK
       else
          call wrschatty (subroutinename, &
               "Successfully found optional column " // trim(try_colname) // " in catalogue file")
       end if
    else
       call wrschatty (subroutinename, "There is no " // trim(try_colname) // " column in catalogue file")
    end if

    try_colname = "AUTO_EXCLUDE"
    colnum_exclude = 0
    call fitswrap_test_for_col(unit, try_colname, exists)
    if ( exists ) then
       call fitswrap_prep_col(unit, try_colname , srclist%num_srcs, colnum_exclude,  pexclude, ierr)
       if (ierr /= ERROR_LEVEL_OK ) then
          call wrschatty (subroutinename, &
               "Problem reading " // trim(try_colname) // " column from catalogue file")
          ierr = ERROR_LEVEL_OK
       else
          call wrschatty (subroutinename, &
               "Successfully found optional column " // trim(try_colname) // " in catalogue file")
       end if
    else
       call wrschatty (subroutinename, "There is no " // trim(try_colname) // " column in catalogue file")
    end if

    ! get the chunksize
    call FTGRSZ(unit, nrows_per_chunk, status)
    if (status /= 0) then
       call fitswrap_write_fits_error(subroutinename, status)
       ierr = ERROR_LEVEL_SEVERE
       return
    end if

    ! flag that we have both columns from the source catalogue
    if ( colnum_scts   > 0 .and. &
         colnum_bg_map > 0 .and. &
         colnum_ext    > 0 ) then
       srclist%auto_columns_supplied = .true.
    end if

    do row = 1, srclist%num_srcs, nrows_per_chunk
       nrows_to_read = min(nrows_per_chunk,srclist%num_srcs-row+1)
       row_end = row+nrows_to_read-1
       call FTGCVJ(unit, colnum_id, row,1,nrows_to_read,nullval_int32, pid(row:row_end), anyf,status)
       call FTGCVD(unit, colnum_ra, row,1,nrows_to_read,nullval_real64,pra(row:row_end), anyf,status)
       call FTGCVD(unit, colnum_dec,row,1,nrows_to_read,nullval_real64,pdec(row:row_end),anyf,status)
       if ( colnum_scts > 0 ) then
          call FTGCVD(unit, colnum_scts, row,1,nrows_to_read,nullval_real64,pscts(row:row_end), anyf,status)
       end if
       if ( colnum_bg_map > 0 ) then
          call FTGCVD(unit, colnum_bg_map, row,1,nrows_to_read,nullval_real64,pbg_map(row:row_end), anyf,status)
       end if
       if ( colnum_ext > 0 ) then
          call FTGCVD(unit, colnum_ext, row,1,nrows_to_read,nullval_real64,pext(row:row_end), anyf,status)
       end if

       if ( colnum_extract > 0 ) then
          call FTGCVJ(unit, colnum_extract, row,1,nrows_to_read,nullval_int32,pextract(row:row_end), anyf,status)
       end if
       if ( colnum_exclude > 0 ) then
          call FTGCVJ(unit, colnum_exclude, row,1,nrows_to_read,nullval_int32,pexclude(row:row_end), anyf,status)
       end if

       if (status /= 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
       end if
    end do

    ! make some space for the src structures
    allocate(srclist%psrc(srclist%num_srcs), STAT=status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! initialise variables for each source and copy source parameters into
    ! src structure
    do i=1,srclist%num_srcs
       associate( src => srclist%psrc(i) )
         call srctool_param_init_src(src)
         src%index = i

         ! new ID_SRC column
         ! https://wiki.mpe.mpg.de/eRosita/EroCat/GenDataAnal?action=AttachFile&do=view&target=SourceLists.html
         ! http://erosita.mpe.mpg.de/eROdoc/pipeline/SASS-products_file_naming_04.html
         write (src%name, '(I5.5)') pid(i)
         src%coord%long           = pra(i)
         src%coord%lat            = pdec(i)

         ! we can assume that all eROSITA source files will use ICRS coords
         src%coord%coordsys_code = COORDSYS_CODE_ICRS

         if ( colnum_scts > 0 ) then
            src%cat_scts = pscts(i)
         else
            src%cat_scts = -1.0d30 ! null value
         end if

         ! we now assume that thebgmap is supplied in counts/arcmin2 so can convert directly to counts/deg2
         if ( colnum_bg_map > 0 ) then
            src%cat_bg_map = pbg_map(i) * SQDEG2SQARCMIN
         else
            src%cat_bg_map = -1.0d30 ! null value
         end if

         ! default is to assume a point source
         src%cat_ext = 0.0d0 ! default value
         src%extent%code= EXTTYPE_CODE_POINT
         if ( colnum_ext > 0 ) then
            src%cat_ext = pext(i) * ARCSEC2DEG
            if ( src%cat_ext > 0.0d0 ) then
               src%extent%code = EXTTYPE_CODE_BETA
               src%extent%num_pars = 1
               src%extent%pars(1) = pext(i) ! units in extent%pars(1) should be arcsec
            end if
         end if

         src%cat_extract = .true. ! default value
         if ( colnum_extract > 0 ) then
            if (pextract(i) == 0 ) then
               src%cat_extract = .false.
            end if
         end if

         src%cat_exclude = .true. ! default value
         if ( colnum_exclude > 0 ) then
            if (pexclude(i) == 0 ) then
               src%cat_exclude = .false.
            end if
         end if

       end associate
    end do

    ! deallocate the temp arrays
    if (allocated(pid))      deallocate(pid)
    if (allocated(pra))      deallocate(pra)
    if (allocated(pdec))     deallocate(pdec)
    if (allocated(pscts))    deallocate(pscts)
    if (allocated(pbg_map))  deallocate(pbg_map)
    if (allocated(pext))     deallocate(pext)
    if (allocated(pextract)) deallocate(pextract)
    if (allocated(pexclude)) deallocate(pexclude)

    !close the catalogue file
    call FTCLOS(unit, status)
    if (status /= 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    !release the unit number
    call FTFIOU(unit, status)
    if (status /= 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

  end subroutine srctool_param_read_srclist_from_fits
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine srctool_param_read_srclist_from_string(str_srccoord, srclist, ierr)
    character(len=*), parameter :: subroutinename='srctool_param_read_srclist_from_string'
    type(srclist_type),   intent(inout), target  :: srclist
    character(len=*), intent(in) ::  str_srccoord
    integer(kind=int32) , intent(out)  :: ierr
!    character(len=LSTR_MAX) :: str_buffer
    integer                :: status
    type(coord_type)        :: local_coord
    type(src_type),   pointer  :: psrc
!    character(len=:), allocatable :: str_copy
    character(len=LSTR_MAX) :: str_copy

    logical             :: success_flag
    ierr=ERROR_LEVEL_OK

    ! we will only have one source, so can set up some space now
    srclist%num_srcs = 1
    status = 0
    allocate(srclist%psrc(srclist%num_srcs), STAT=status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! ititialise variables for the src
    call srctool_param_init_src(srclist%psrc(1))

    psrc => srclist%psrc(1)


!    ! ititialise variables for the src
!    call srctool_param_init_src(psrc)
    psrc%index = 1

!    ALLOCATE(character(len=LEN_TRIM(ADJUSTL(str_srccoord))) :: str_copy, &
!        STAT = status)
!    if ( check_allocate_status(subroutinename, status, ierr) ) return
    if ( LEN_TRIM(ADJUSTL(str_srccoord)) .gt. len(str_copy) ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "srccoord string is too long to process")
      return
    endif

    write (psrc%name, '(I5.5)') psrc%index

    ! strip out parts of the string which are region info
    call srctool_regions_extract_central_coords (str_srccoord, str_copy)
!    write (*,'(A,A,A)'), "DEBUG='", trim(str_copy), "'"

    ! try to interpret the resultant string as a decimal coordinate pair
    call convert_longlat_to_decimal (TRIM(str_copy), &
        psrc%coord%long, psrc%coord%lat, &
        psrc%coord%coordsys_code, success_flag)
    ! quietly assume that coords are equatorial if not explicitly stated
    if ( psrc%coord%coordsys_code .eq. COORDSYS_CODE_NULL ) then
      psrc%coord%coordsys_code = COORDSYS_CODE_ICRS
    endif

    if (.not. success_flag) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not interpret 'srccoord' parameter as a coordinate string")
      return
    end if



    ! now convert the src coords into ICRS if necessary
    if ( psrc%coord%coordsys_code .ne. COORDSYS_CODE_ICRS ) then
      if ( (psrc%coord%coordsys_code .eq. COORDSYS_CODE_FK4 ) .or. &
          (psrc%coord%coordsys_code .eq. COORDSYS_CODE_GAL ) .or. &
          (psrc%coord%coordsys_code .eq. COORDSYS_CODE_ECL ) ) then
        local_coord = psrc%coord
        psrc%coord%coordsys_code = COORDSYS_CODE_ICRS
        call geom_convert_coordsys (local_coord, psrc%coord )
        call wrschatty (subroutinename, "Converting input source coords into ICRS system")
      else
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Cannot handle this coordinate system type: '" // &
            COORDSYS_STR(psrc%coord%coordsys_code) // "'")
        return
      endif
    endif

!    if ( psrc%coord%coordsys_code .eq. COORDSYS_CODE_ICRS ) then
!      psrc%coord%equinox = 2000.0d0  ! set the equinox to 2000.0
!    else
!      ierr = ERROR_LEVEL_SEVERE
!      call wrserr (subroutinename, ierr, "Cannot handle source coordinates that are not in the icrs system")
!      return
!    end if



!!    write(str_buffer, '(A,F12.7,A,F12.7,A,A,A,F6.1,A)'), &
!    write(str_buffer, '(A,F12.7,A,F12.7,A,A,A)') &
!        "Interpreted srccoord as: (RA,Dec)= ", &
!        psrc%coord%long, ", ", psrc%coord%lat, " (deg,", &
!        TRIM(COORDSYS_STR(psrc%coord%coordsys_code)), ")"
!!        " equinox=", psrc%coord%equinox, ")"
    call wrschatty (subroutinename, "Interpreted srccoord as: (RA,Dec)= " // &
        r2fs(psrc%coord%long) // ", " // r2fs(psrc%coord%lat) // " (deg," // &
        TRIM(COORDSYS_STR(psrc%coord%coordsys_code)) // ")")

    nullify(psrc)
!    if ( allocated(str_copy) ) deallocate(str_copy)


  end subroutine srctool_param_read_srclist_from_string

  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine srctool_param_read_srclist_from_ascii(str_filename, srclist,ierr)
    character(len=*), parameter :: subroutinename='srctool_param_read_srclist_from_ascii'
    character(len=*), intent(in) ::  str_filename
    type(srclist_type),   intent(inout)  :: srclist
    integer(kind=int32) , intent(out)  :: ierr

    integer(kind=int32)     :: unit
    character(len=LSTR_MAX) :: read_buffer
    character(len=LSTR_MAX) :: read_buffer_copy
    character(len=LSTR_MAX) :: str_buffer
    integer                 :: status
    integer                 :: iostatus
    integer(kind=int32)     :: i, j
    logical                 :: success_flag
    type(coord_type)        :: local_coord
    integer(kind=int32)     :: default_coordsys
    ierr=ERROR_LEVEL_OK
    status = 0
    iostatus = 0
    default_coordsys = COORDSYS_CODE_ICRS
    ! get an i/o unit - use the fitsio routine for this
    call FTGIOU(unit, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

    open (unit, FILE=TRIM(str_filename), IOSTAT=iostatus, ACTION="read")
    if ( iostatus .ne. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "I couldn't open srccoord ascii file: " // TRIM(str_filename) )
      return
    end if

    srclist%num_srcs = 0
    iostatus = 0
    do while ( iostatus .eq. 0 )
      read_buffer = ""
      read(unit, '(A)', IOSTAT=iostatus) read_buffer
      if ( iostatus .lt. 0 ) then  ! end of file condition - ok
        exit  ! leave the do loop
      else if ( iostatus .gt. 0 ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Error reading input from srccoord ascii file: " // TRIM(str_filename) )
        return
      else ! iostat .eq. 0
        ! add checking for comment lines and blank lines


        ! first remove any comments (anything after a '#' character)
        j = INDEX(read_buffer,"#")
        if ( j .gt. 0 ) then
          read_buffer(j:) = " "  ! convert to white space
        end if

        ! check for global statements in ds9 region files and ignore them
        if ( INDEX(ADJUSTL(read_buffer),"global") .eq. 1 ) then
          read_buffer(:) = " "  ! convert to white space
        endif

        ! check for valid coordinate system statements in ds9 region files and ignore them for now
        str_buffer = ADJUSTL(read_buffer(1:MIN(LEN_TRIM(read_buffer),LSTR_MAX)))
        call to_upper(str_buffer)
        if ( (TRIM(str_buffer) .eq. "FK5") .or. &
             (TRIM(str_buffer) .eq. "J2000") .or. &
             (TRIM(str_buffer) .eq. "GALACTIC") .or. &
             (TRIM(str_buffer) .eq. "PHYSICAL ") .or. &
             (TRIM(str_buffer) .eq. "IMAGE ") .or. &
             (TRIM(str_buffer) .eq. "FK4") .or. &
             (TRIM(str_buffer) .eq. "B1950") .or. &
             (TRIM(str_buffer) .eq. "ECLIPTIC") .or. &
             (TRIM(str_buffer) .eq. "ICRS") .or. &
             (TRIM(str_buffer) .eq. "LINEAR") .or. &
             (TRIM(str_buffer) .eq. "AMPLIFIER") .or. &
             (TRIM(str_buffer) .eq. "DETECTOR") ) then
          read_buffer(:) = " "  ! convert to white space
        endif

!        call wrschatty(subroutinename, "DEBUG:-->" // TRIM(ADJUSTL(read_buffer)) // "<--")

        ! check for blank lines
        if ( LEN_TRIM(ADJUSTL(read_buffer)) .ge. 1 ) then
          srclist%num_srcs = srclist%num_srcs + 1
        endif
      end if
    end do

    call wrschatty (subroutinename, "I found " // trim(i2s(srclist%num_srcs)) // &
        " potential sources in srccoord ascii file: " // TRIM(str_filename) )

    ! make some space
    allocate(srclist%psrc(srclist%num_srcs), STAT=status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! ititialise variables for each of the sources
    do i=1,srclist%num_srcs
      call srctool_param_init_src(srclist%psrc(i))
      srclist%psrc(i)%index = i
      write (srclist%psrc(i)%name, '(I5.5)') srclist%psrc(i)%index
    end do

    ! rewind the file, assume it works
    rewind (unit)
    ! now read through the file again
    iostatus = 0
    i = 0
    do while ( iostatus .eq. 0 )
      read_buffer = ""
      read(unit, '(A)', IOSTAT=iostatus) read_buffer
      if ( iostatus .lt. 0 ) then  ! end of file condition - ok
        exit  ! leave the do loop
      else if ( iostatus .gt. 0 ) then
        ierr = ERROR_LEVEL_SEVERE
        write (str_buffer, '(A,A)') "Error reading input from srccoord ascii file: ",TRIM(str_filename)
        call wrserr (subroutinename, ierr, str_buffer)
        return
      else ! iostat .eq. 0

        ! first remove any comments (anything after a '#' character)
        j = INDEX(read_buffer,"#")
        if ( j > 0 ) then
          read_buffer(j:) = " "  ! convert to white space
        end if

        ! check for global statements in ds9 region files and ignore them
        if ( INDEX(ADJUSTL(read_buffer),"global") .eq. 1 ) then
          read_buffer(:) = " "  ! convert to white space
        endif

       ! check for coordinate system statements in ds9 region files and set the default accordingly
        str_buffer = ADJUSTL(read_buffer(1:MIN(LEN_TRIM(read_buffer),LSTR_MAX)))
        call to_upper(str_buffer)
        select case (TRIM(str_buffer) )
          case ( "FK5", "J2000" , "ICRS" )
            default_coordsys = COORDSYS_CODE_ICRS  ! close enough !
            read_buffer(:) = " "  ! convert to white space
          case ( "FK4", "B1950" )
            default_coordsys = COORDSYS_CODE_FK4
            read_buffer(:) = " "  ! convert to white space
          case ( "GALACTIC" )
            default_coordsys = COORDSYS_CODE_GAL
            read_buffer(:) = " "  ! convert to white space
          case ( "ECLIPTIC" )
            default_coordsys = COORDSYS_CODE_ECL
            read_buffer(:) = " "  ! convert to white space
          case ( "PHYSICAL", "IMAGE", "LINEAR", "AMPLIFIER", "DETECTOR" )
            ierr=ERROR_LEVEL_SEVERE
            call wrserr(subroutinename,ierr, "Unsupported coordinate system: " // TRIM(str_buffer) )
            return
          case default
          end select

        ! check for blank lines
        if ( len_trim(adjustl(read_buffer)) .gt. 0 ) then

          ! increment the counter and check we have space left
          i = i + 1
          if ( i .gt. srclist%num_srcs ) then
            ierr = ERROR_LEVEL_SEVERE
            call wrserr(subroutinename,ierr, &
                "Error, I found more sources than expected in the srccoord ascii file: " // TRIM(str_filename))
            return
          end if


          ! strip out any parts of the string which are region info
          call srctool_regions_extract_central_coords (read_buffer, read_buffer_copy)

          ! get the coordinates from the string
          call convert_longlat_to_decimal (TRIM(read_buffer_copy), &
              srclist%psrc(i)%coord%long, srclist%psrc(i)%coord%lat, &
              srclist%psrc(i)%coord%coordsys_code, success_flag)

          if (.not. success_flag) then
            ierr = ERROR_LEVEL_SEVERE
            call wrserr (subroutinename, ierr, "Could not interpret srccoord as a coordinate string: '" // TRIM(read_buffer) // "'")
            return
          end if

          ! quietly use the default coordsys if none supplied
          if ( srclist%psrc(i)%coord%coordsys_code .eq. COORDSYS_CODE_NULL ) then
            srclist%psrc(i)%coord%coordsys_code = default_coordsys
          endif

          ! now convert the src coords into ICRS if necessary
          if ( srclist%psrc(i)%coord%coordsys_code .ne. COORDSYS_CODE_ICRS ) then
            if ( (srclist%psrc(i)%coord%coordsys_code .eq. COORDSYS_CODE_FK4 ) .or. &
                (srclist%psrc(i)%coord%coordsys_code .eq. COORDSYS_CODE_GAL ) .or. &
                (srclist%psrc(i)%coord%coordsys_code .eq. COORDSYS_CODE_ECL ) ) then
              local_coord = srclist%psrc(i)%coord
              srclist%psrc(i)%coord%coordsys_code = COORDSYS_CODE_ICRS
              call geom_convert_coordsys (local_coord, srclist%psrc(i)%coord )
              call wrschatty (subroutinename, "Converting input src coords into ICRS system")
            else
              ierr = ERROR_LEVEL_SEVERE
              call wrserr (subroutinename, ierr,  "Cannot handle this coordinate system type: '" &
                  // TRIM(COORDSYS_STR(srclist%psrc(i)%coord%coordsys_code)) // "'")
              return
            endif
          endif

!        if ( srclist%psrc(i)%coord%coordsys_code .eq. COORDSYS_CODE_ICRS ) then
!          srclist%psrc(i)%coord%equinox = 2000.0d0  ! set the equinox to 2000.0
!        else
!          ierr = ERROR_LEVEL_SEVERE
!          call wrserr (subroutinename, ierr, "I cannot handle source coordinates that are not in the icrs system.")
!          return
!        end if

          !        write(str_buffer, '(A,I0,A,F12.7,A,F12.7,A,A)'), &
          write(str_buffer, '(A,I0,A,F12.7,A,F12.7,A,A,A,F6.1,A)') &
              "Interpretd srccoord ", i, " as: (RA,Dec)= ", &
              srclist%psrc(i)%coord%long, ", ", srclist%psrc(i)%coord%lat, " (deg,", &
              TRIM(COORDSYS_STR(srclist%psrc(i)%coord%coordsys_code))
          !            " equinox=", srclist%psrc(i)%coord%equinox, ")"
          call wrschatty (subroutinename, str_buffer)
        end if
      end if
    end do


    ! tidy up
    close(unit, IOSTAT=iostatus)
    if ( iostatus .ne. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr,  "I couldn't close the 'srccoord' ASCII file: " // TRIM(str_filename) )
      return
    end if


    !release the unit number
    call FTFIOU(unit, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if

  end subroutine srctool_param_read_srclist_from_ascii
  !---------------------------------------------------------------------------------------------







  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !--------------------------- END of SRCLIST handling ------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------- START of REGION structure handling ---------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  ! want to follow ds9 naming conventions as closely as possible.
  ! allow wild cards in position coordinates which will be replaced
  ! by the appropriate coordinates of the source .
  ! Composite regions are supplied with one region component on each line
  ! region components to be subtracted are prefixxed with a "-" character.
  ! Coordinate system defaults to icrs
  subroutine srctool_param_interpret_region_desc (str_reg, srclist, class, ierr )
    character(len=*), parameter :: subroutinename='srctool_param_interpret_region_desc'
    character(len=*), intent(in)   :: str_reg
    type(srclist_type), intent(inout), target  :: srclist
    integer(kind=int32), intent(in)  :: class
    integer(kind=int32), intent(out)  :: ierr
    type(src_type),     pointer  :: psrc1
    type(src_type),     pointer  :: psrci
    type(region_type),  pointer :: pregion1
    type(region_type),  pointer :: pregioni
    type(regcomp_type)  :: regcomp
    character(len=STR_MAX) :: read_state
!    character(len=STR_MAX) :: str_buffer
    character(len=STR_MAX) :: str_class
    character(len=LSTR_MAX) :: read_buffer
    integer                :: iostatus
    logical                :: exist_flag
    integer(kind=int32)    :: unit
    integer(kind=int32)    :: i,j
    integer(kind=int32)    :: status
    integer(kind=int32)    :: autoreg_code
    character(len=LSTR_MAX) :: reg_copy
    ierr = ERROR_LEVEL_OK
    iostatus = 0
    status = 0

    if ( class .eq. REG_CLASS_FLAG_SRC ) then
      str_class = "SRC"
    else if ( class .eq. REG_CLASS_FLAG_BACK ) then
      str_class = "BACK"
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Unknown region class")
      return
    end if


!    write (str_buffer, '(A,A,A,A,A)'), "Interpreting ", TRIM(str_class), " region description: '",TRIM(str_reg),"'"
!    call wrschatty (subroutinename, str_buffer)
    call wrschatty (subroutinename, "Interpreting " // TRIM(str_class) // &
        " region description: '" // TRIM(str_reg) // "'")

    if ( srclist%num_srcs <= 0 ) then
       ! not a hard error - just ignore
       return
    end if

    reg_copy = adjustl(trim(str_reg(1:MIN(LSTR_MAX,len(str_reg)))))
    call to_upper(reg_copy)

    ! check if we are using a null background region
    ! note that null regions are not allowed for sources
    if ( class .eq. REG_CLASS_FLAG_BACK ) then
      if ((trim(reg_copy) .eq. "") .or. &
          (trim(reg_copy) .eq. "-") .or.  &
          (trim(reg_copy) .eq. "NONE") .or.  &
          (trim(reg_copy) .eq. "NULL") ) then
        ! no background region is needed - so set the flags and then return, easy!
        srclist%psrc(:)%do_back = .false.
        srclist%autoreg_code_back = AUTOREG_CODE_NULL
        call wrschatty (subroutinename, "bkgnd_reg region is null, so we won't extract any background")
        return
      end if
    endif


    if ( LEN_TRIM(str_reg) .lt. 1 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Region string is too short")
      return
    end if

    ! ---------------------------------------------------------------------------
    ! first see if the region matches one of the special cases: AUTO, AUTO1
    autoreg_code = AUTOREG_CODE_NULL
    ! now match against list of valid options
    if      ( (TRIM(reg_copy) .eq. "AUTO1") .or. (TRIM(reg_copy) .eq. "AUTO") ) then
      autoreg_code = AUTOREG_CODE_AUTO1
!    else if ( (TRIM(reg_copy) .eq. "AUTO2") ) then
!      autoreg_code = AUTOREG_CODE_AUTO2
    end if

    if ( autoreg_code .ne. AUTOREG_CODE_NULL ) then
      if ( class .eq. REG_CLASS_FLAG_SRC ) then
        srclist%autoreg_code_src = autoreg_code
      else if ( class .eq. REG_CLASS_FLAG_BACK ) then
        srclist%autoreg_code_back = autoreg_code
      else
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Unknown region class")
        return
      end if

      ! check that we also have a suitable input catalogue to work from
      if ( srclist%type_code .ne. SRCLIST_TYPE_CODE_FITSCAT ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, &
            "Computation of automatic regions requires a FITS format catalogue (you supplied a " // &
            trim(SRCLIST_TYPE_CODE_STR(srclist%type_code)) // " format source description)" )
        return
      endif

      if ( .not. srclist%auto_columns_supplied ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, &
            "Computation of automatic regions requires catalogue columns SCTS_0 and BG_MAP_0")
        return
      endif


      ! Signal that we will compute the auto regions - do this within main module
      call wrschatty(subroutinename, "We will generate " &
          // trim(REG_CLASS_FLAG_STR(class)) // " regions automatically")

      return
    endif
    ! ---------------------------------------------------------------------------

    ! If we are not doing auto regions, then next see if the region is a readable filename
    inquire(FILE=TRIM(str_reg), IOSTAT=iostatus, EXIST=exist_flag, READ=read_state)

    if ((status .eq. 0) .and. .not. exist_flag ) then
      ! try to trim off any bits of the filename which are fits HDU/selection expressions
      call FTRTNM(TRIM(str_reg), reg_copy, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if
      ! now see if this version exists as well
      inquire(FILE=TRIM(reg_copy), IOSTAT=status, EXIST=exist_flag, READ=read_state)
!      write(*,'(A,A,A,I0,A,A)'), "STEP2: filename='", &
!          TRIM(reg_copy), "' status= ", status, " exist_flag=",  logical2yesno(exist_flag)
    endif



    !--------------------------------------------------------------------------
    !--------------------------------------------------------------------------
    if ( (iostatus .eq. 0 ) .and. exist_flag ) then    ! ok great, I found the name of a file
      ! check that can actually read this file:
      if ( TRIM(read_state) .ne. "YES" ) then
!        write (str_buffer, '(A,A,A)') "I do not have read access to region file: '",TRIM(str_reg),"'"
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "I do not have read access to region file: '" // TRIM(str_reg) // "'")
        return
      end if

!      write (str_buffer, '(A,A,A)') "Interpreting ", TRIM(str_class), " region description as a filename"
      call wrschatty (subroutinename, "Interpreted " // TRIM(str_class) // " region description as a filename")

      ! try to open file

      !-----------------------------------
      ! use fitsio to get a fortran unit
      call FTGIOU( unit, status)
      if (status .ne. 0) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if
      !-----------------------------------


      open (unit, FILE=TRIM(str_reg), IOSTAT=iostatus, ACTION="read")
      if ( iostatus .ne. 0 ) then
!        write (str_buffer, '(A,A)') "I couldn't open region file: ",TRIM(str_reg)
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr,  "I couldn't open region file: " // TRIM(str_reg) )
        return
      end if

      ! now read and interpret the lines of the region file
      ! read the first "stanza" normally for the 1st source
      ! if there are no more lines in the file then we should interpret
      ! this region as a template for all sources in the srclist
      iostatus = 0
      i = 1
      psrci => srclist%psrc(i)
      if ( class .eq. REG_CLASS_FLAG_SRC ) then
        pregioni => psrci%src_region
      else if ( class .eq. REG_CLASS_FLAG_BACK ) then
        pregioni => psrci%back_region
      else
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Unknown region class")
        return
      end if
      pregioni%num_regcomps = 0

      do while ( iostatus .eq. 0 )
        read_buffer = ""
        read(unit, '(A)', IOSTAT=iostatus) read_buffer
        if ( iostatus .lt. 0 ) then  ! end of file condition - ok
          exit
        else if ( iostatus .gt. 0 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Error reading input from region file")
          return
        else ! iostat .eq. 0
!          write (*,'(A,A,A,I0)'), "I just read this line: '", TRIM(read_buffer), "' iostat=", iostatus
          if ( LEN_TRIM(ADJUSTL(read_buffer)) .eq. 0 ) then  ! reached a blank line - i.e. end of stanza
!            write (*, '(A)'), "Blank line - end of stanza"

            ! increment the source counter
            i = i + 1
            if ( i .gt. srclist%num_srcs ) then
              ierr = ERROR_LEVEL_SEVERE
              call wrserr (subroutinename, ierr, "You supplied more region descriptions than sources!")
              return
            end if
            psrci => srclist%psrc(i)
            if ( class .eq. REG_CLASS_FLAG_SRC ) then
              pregioni => psrci%src_region
            else if ( class .eq. REG_CLASS_FLAG_BACK ) then
              pregioni => psrci%back_region
            else
              ierr = ERROR_LEVEL_SEVERE
              call wrserr (subroutinename, ierr, "Unknown region class")
              return
            end if
            pregioni%num_regcomps = 0
          else

            ! check for "global" statements in ds9 region files and treat them as comment lines
            if ( INDEX(ADJUSTL(read_buffer),"global") .eq. 1 ) then
              read_buffer(1:1) = "#"  ! convert line to a comment
            endif

            ! check if this is a comment line - i.e. if there is nothing but white space before the first hash
            j = INDEX(ADJUSTL(read_buffer), '#')
            if ( j .ne. 1 ) then
              call srctool_regions_init_regcomp(regcomp)
              call srctool_regions_interpret_regcomp_desc(trim(read_buffer), regcomp, psrci%coord, ierr)
              if ( ierr .ne. ERROR_LEVEL_OK ) then
                call wrserr (subroutinename, ierr, "Problem interpretting region description")
                return
              end if
              call srctool_regions_add_regcomp(pregioni, regcomp)

              call srctool_regions_resolve_region_wildcards (psrci%coord,psrci%src_region )
              call srctool_regions_resolve_region_wildcards (psrci%coord,psrci%back_region )

              call srctool_regions_region_calc_bounding_circle (pregioni, psrci%coord, ierr )
              if ( ierr .ne. ERROR_LEVEL_OK ) then
                call wrserr(subroutinename,ierr,"Problem calculating bounding circles")
                return
              end if
            else
!              write (*, '(A)'), "Comment line - skipping"
            end if

          end if
        end if
      end do

      ! if region descriptions were read in for only one source, then copy the regions for the remainder
      if ( i .eq. 1 .and. srclist%num_srcs .gt. 1 ) then

        call srctool_param_copy_regions_across_sources (srclist, class, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr(subroutinename,ierr,"Problem copying region descriptions")
          return
        end if

      else if ( i .ne. srclist%num_srcs ) then
        ierr = ERROR_LEVEL_SEVERE
!        write (str_buffer, '(A,I0,A,I0,A)') &
!            "You supplied ", i, " regions, but there are ", srclist%num_srcs, " sources!"
        call wrserr (subroutinename, ierr, "You supplied "// trim(i2s(i)) // &
             " regions, but there are " // trim(i2s(srclist%num_srcs)) // " sources!")
        return
      end if


      ! tidy up
      close(unit, IOSTAT=iostatus)
      if ( iostatus .ne. 0 ) then
!        write (str_buffer, '(A,A)') "I couldn't close region file: ",TRIM(str_reg)
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "I couldn't close region file: " // TRIM(str_reg))
        return
      end if

      !release the unit number
      call FTFIOU(unit, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      !-----------------------------------

!      ierr = ERROR_LEVEL_SEVERE
!      call wrserr (subroutinename, ierr, "I have not yet finished implementing this section of the code")
!      return

    !--------------------------------------------------------------------------
    !--------------------------------------------------------------------------
    else  ! we can assume that the region is just supplied on the command line as a single component string

!      write (str_buffer, '(A,A,A)') "Interpreting ", TRIM(str_class), " region description as an inline region"
      call wrschatty (subroutinename, "Interpreted " // TRIM(str_class) // " region description as an inline region")
      ! read the region as if it were only for the first src
      nullify(psrc1)
      nullify(pregion1)
      psrc1 => srclist%psrc(1)
      if ( class .eq. REG_CLASS_FLAG_SRC ) then
        pregion1 => psrc1%src_region
      else if ( class .eq. REG_CLASS_FLAG_BACK ) then
        pregion1 => psrc1%back_region
      else
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Unknown region class")
        return
      end if

      call srctool_regions_init_regcomp(regcomp)
      call srctool_regions_interpret_regcomp_desc(str_reg, regcomp, psrc1%coord, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem interpretting region description")
        return
      end if
      pregion1%num_regcomps = 0
      call srctool_regions_add_regcomp(pregion1, regcomp)

      call srctool_regions_resolve_region_wildcards (psrc1%coord,psrc1%src_region )
      call srctool_regions_resolve_region_wildcards (psrc1%coord,psrc1%back_region )
      call srctool_regions_region_calc_bounding_circle (pregion1, psrc1%coord, ierr )
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr(subroutinename,ierr,"Problem calculating bounding circles")
        return
      end if

      call srctool_param_copy_regions_across_sources (srclist, class, ierr )
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr(subroutinename,ierr,"Problem copying region descriptions")
        return
      end if

    end if

  end subroutine srctool_param_interpret_region_desc
  !-----------------------------------------------------------------------------------


  !-----------------------------------------------------------------------------------
  subroutine srctool_param_copy_regions_across_sources (srclist, class, ierr )
    character(len=*), parameter :: subroutinename='srctool_param_copy_regions_across_sources'
    type(srclist_type), intent(inout), target  :: srclist
    integer(kind=int32), intent(in)  :: class
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32)  :: i
    character(len=STR_MAX) :: str_buffer

    type(src_type),     pointer  :: psrc1
    type(src_type),     pointer  :: psrci
    type(region_type),  pointer :: pregion1
    type(region_type),  pointer :: pregioni

    ierr = ERROR_LEVEL_OK

     ! now copy the region description to the other sources
    if ( srclist%num_srcs .ge. 2 ) then

      write (str_buffer, '(A,A,A,I0,A)')  &
          "Deriving ", REG_CLASS_FLAG_STR(class), " regions for sources #2-#", &
          srclist%num_srcs, " from region description supplied for source #1"
      call wrschatty (subroutinename,str_buffer)

      psrc1 => srclist%psrc(1)
      if ( class .eq. REG_CLASS_FLAG_SRC ) then
        pregion1 => psrc1%src_region
      else if ( class .eq. REG_CLASS_FLAG_BACK ) then
        pregion1 => psrc1%back_region
      else
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Unknown region class")
        return
      end if

      do i=2,srclist%num_srcs
        nullify(psrci)
        nullify(pregioni)
        psrci => srclist%psrc(i)
        if ( class .eq. REG_CLASS_FLAG_SRC ) then
          pregioni => psrci%src_region
        else if ( class .eq. REG_CLASS_FLAG_BACK ) then
          pregioni => psrci%back_region
        else
          ierr = ERROR_LEVEL_SEVERE
          call wrserr(subroutinename,ierr,"Unknown region class")
          return
        end if

        pregioni = pregion1 ! do the copy
        call srctool_regions_resolve_region_wildcards (psrci%coord,psrci%src_region )
        call srctool_regions_resolve_region_wildcards (psrci%coord,psrci%back_region )

        call srctool_regions_region_calc_bounding_circle (pregioni, psrci%coord, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr(subroutinename,ierr,"Problem calculating bounding circles")
          return
        end if

      end do
    end if

  end subroutine srctool_param_copy_regions_across_sources
  !-----------------------------------------------------------------------------------

  !-----------------------------------------------------------------------------------
  !-----------------------------------------------------------------------------------
  !---------------------------- END of REGION structure handling ---------------------
  !-----------------------------------------------------------------------------------
  !-----------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! Get environment variable with default value (real64)
  ! Returns default value if cannot be found or text cannot be interpreted
  real(kind=real64) function srctool_param_env_real64(name, defval)

    character(len=*), parameter :: subroutinename='srctool_param_env_real64'
    character(len=*), intent(in) :: name
    real(kind=real64), intent(in) :: defval

    character(len=:),allocatable :: string
    integer :: length, status
    real(kind=real64) :: val

    srctool_param_env_real64 = defval

    ! get string length
    call get_environment_variable(name, length=length, status=status)
    if(length==0 .or. status/=0) then
       return
    end if

    ! allocate and get value
    allocate(character(len=length) :: string)
    call get_environment_variable(name, string)

    read(string,*,err=999) val
    srctool_param_env_real64 = val
    deallocate(string)
    return

999 call wrserr(subroutinename, ERROR_LEVEL_WARNING, &
         "Warning: could not interpret value of environment variable " // &
         trim(name))
    deallocate(string)

  end function srctool_param_env_real64

end module srctool_param
