! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_psf

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-25
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling instpar cal data
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use iso_c_binding

  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         !
  use srctool_misc              !
  use srctool_fftw              ! Wrap FFTW3 functions

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_PSF_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_psf
  private :: srctool_psf_empty_psfmaplist
  private :: srctool_psf_read_psfmap_from_chdu
  private :: srctool_psf_index_psfmaplist_caldb
  private :: srctool_psf_index_psf
  private :: srctool_psf_get_psfmap_pointers
  private :: srctool_psf_get_psfmap_values
  private :: srctool_psf_convolve_psfmaplist_with_extent
  private :: srctool_psf_copy_psfmaplist

  public ::  srctool_psf_copy_psfmap
  public ::  srctool_psf_index_psfs
  public ::  srctool_psf_convolve_psfmap_with_extent
  public ::  srctool_psf_copy_psf
  public ::  srctool_psf_convolve_psf_with_extent
  public ::  srctool_psf_empty_psf
  public ::  srctool_psf_empty_rpsf
  public ::  srctool_psf_empty_psfmap
  public ::  srctool_psf_calc_rpsf
  public ::  srctool_psf_write_psfmap
  public ::  srctool_psf_get_psf_values
  public ::  srctool_psf_read_psf_from_chdu
  public ::  srctool_psf_report_psf_info
  public ::  srctool_psf_get_psfmap_pointer

  public ::  srctool_psf_setup_fftw_plans

  private :: srctool_get_efficient_fft_size
  private :: fftw_alloc_real_array
  private :: fftw_free_real_array
  private :: fftw_alloc_complex_array
  private :: fftw_free_complex_array

contains



  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of PSF file handling------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  recursive subroutine srctool_psf_empty_rpsf(rpsf)
    type(rpsf_type), intent(inout)   :: rpsf
    rpsf%is_valid = .false.
    rpsf%dradius_deg = 0.0d0
    rpsf%radius_50pc_deg = -1.0d0
    rpsf%radius_90pc_deg = -1.0d0
    rpsf%radius_99pc_deg = -1.0d0
    if ( allocated(rpsf%pradius))  deallocate(rpsf%pradius)
    if ( allocated(rpsf%parea))    deallocate(rpsf%parea)
    if ( allocated(rpsf%prpsf))    deallocate(rpsf%prpsf)
    if ( allocated(rpsf%preef))    deallocate(rpsf%preef)

  end subroutine srctool_psf_empty_rpsf


  recursive subroutine srctool_psf_empty_psfmap(psfmap)
    type(psfmap_type), intent(inout)   :: psfmap

    call fitswrap_init_hduinfo(psfmap%hduinfo)

    ! init extra information keys
!    call fitswrap_init_key(psfmap%key_energy,   "ENERGY")
!    call fitswrap_init_key(psfmap%key_energ_lo, "ENERG_LO")
!    call fitswrap_init_key(psfmap%key_energ_hi, "ENERG_HI")
!    call fitswrap_init_key(psfmap%key_theta,    "THETA")
!    call fitswrap_init_key(psfmap%key_phi,      "PHI")
    call fitswrap_init_key(psfmap%key_energy,      "ENERGY")
    call fitswrap_init_key(psfmap%key_backgrnd,    "BACKGRND")
    call fitswrap_init_key(psfmap%key_crpix1,      "CRPIX1")
    call fitswrap_init_key(psfmap%key_crpix2,      "CRPIX2")
    call fitswrap_init_key(psfmap%key_crval1,      "CRVAL1")
    call fitswrap_init_key(psfmap%key_crval2,      "CRVAL2")
    call fitswrap_init_key(psfmap%key_cdelt1,      "CDELT1")
    call fitswrap_init_key(psfmap%key_cdelt2,      "CDELT2")
    call fitswrap_init_key(psfmap%key_cunit1,      "CUNIT1")
    call fitswrap_init_key(psfmap%key_cunit2,      "CUNIT2")
    call fitswrap_init_key(psfmap%key_naxis1,      "NAXIS1")
    call fitswrap_init_key(psfmap%key_naxis2,      "NAXIS2")
    call fitswrap_init_key(psfmap%key_caldb_energ, "CBD10001")
    call fitswrap_init_key(psfmap%key_caldb_theta, "CBD20001")
    call fitswrap_init_key(psfmap%key_caldb_obsmode,"CBD30001")



    if ( allocated(psfmap%ppix))    deallocate(psfmap%ppix)
    psfmap%naxis1 = 0
    psfmap%naxis2 = 0
    psfmap%area_per_pix = 0.0d0

    psfmap%pix_x_min = -1
    psfmap%pix_x_max = -1
    psfmap%pix_y_min = -1
    psfmap%pix_y_max = -1


    psfmap%energy_eff = -huge(1d0)
    psfmap%energy_lo =  huge(1d0)
    psfmap%energy_hi = -huge(1d0)
    psfmap%theta_eff = -huge(1d0)
    psfmap%theta_lo =  huge(1d0)
    psfmap%theta_hi = -huge(1d0)
    psfmap%phi_eff = -huge(1d0)
    psfmap%phi_lo =  huge(1d0)
    psfmap%phi_hi = -huge(1d0)

    call srctool_psf_empty_rpsf(psfmap%rpsf)

    psfmap%is_valid = .false.
  end subroutine srctool_psf_empty_psfmap


  recursive subroutine srctool_psf_empty_psfmaplist(psfmaplist)
    type(psfmaplist_type), intent(inout)   :: psfmaplist
    integer(kind=int32) :: i
    if ( allocated(psfmaplist%psfmap)) then
       do i=1,size(psfmaplist%psfmap)
          call srctool_psf_empty_psfmap(psfmaplist%psfmap(i))
       end do
       deallocate (psfmaplist%psfmap)
    end if
    psfmaplist%num_psfmaps = 0
    psfmaplist%is_valid = .false.

    if (allocated(psfmaplist%pe))            deallocate(psfmaplist%pe)
    if (allocated(psfmaplist%ptheta_arcmin)) deallocate(psfmaplist%ptheta_arcmin)
    if (allocated(psfmaplist%pphi))          deallocate(psfmaplist%pphi)
    if (allocated(psfmaplist%pindex))        deallocate(psfmaplist%pindex)
    psfmaplist%num_e     = 0
    psfmaplist%num_theta = 0
    psfmaplist%num_phi   = 0
!    psfmaplist%focal_length = -1.0d0

  end subroutine srctool_psf_empty_psfmaplist

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_psf_empty_psf(psf)
    type(psf_type), intent(inout)   :: psf

    call srctool_psf_empty_psfmaplist(psf%psfmaplist)
    ! TODO other structures

    call fitswrap_init_hduinfo(psf%hduinfo)
    psf%class = SRCTOOL_PSF_CLASS_NONE
    psf%is_valid = .false.
  end subroutine srctool_psf_empty_psf
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the psf map from the current hdu of the fits file identified by unit
  ! assume that callling routine has done some file checking
  subroutine srctool_psf_read_psfmap_from_chdu(unit, psfmap, ierr)
    character(len=*), parameter :: subroutinename='srctool_psf_read_psfmap_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(psfmap_type), intent(inout)  :: psfmap
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
    integer(kind=int32) :: anyf
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: bitpix
    integer(kind=int32) :: p1,p2,p3
    integer(kind=int32) :: i,j

    integer(kind=int32), parameter :: group=0
    real(kind=real64), parameter   :: nullval = 0.0d0
    real(kind=real64), parameter   :: psf_thresh = 1.0d-9 ! only used for statistics at the moment
!    real(kind=real32), parameter   :: psf_thresh = 1.0e-6


    ierr = ERROR_LEVEL_OK
    status = 0

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, psfmap%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

   ! read the other keywords


!    call  fitswrap_read_key(unit, psfmap%key_energy     ierr)
!    if ( ierr .ne. ERROR_LEVEL_OK ) return
!    ! convert to eV !!!
!    call fitswrap_set_key(psfmap%key_energy, FITSWRAP_SET_KEY_BLANK, psfmap%key_energy%value*1d-3, "[keV]")

!    call  fitswrap_read_key(unit, psfmap%key_energ_lo   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, psfmap%key_energ_hi   , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return


!    call  fitswrap_read_key(unit, psfmap%key_theta      , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
!    call  fitswrap_read_key(unit, psfmap%key_phi        , ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_energy     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_backgrnd   ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_crpix1     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_crpix2     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_crval1     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_crval2     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_cdelt1     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_cdelt2     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_cunit1     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_cunit2     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_naxis1     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_naxis2     ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_caldb_energ,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_caldb_theta,  ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return
    call  fitswrap_read_key(unit, psfmap%key_caldb_obsmode,ierr); if ( ierr .ne. ERROR_LEVEL_OK ) return

    ! copy the naxis keyword values to easier format
!    psfmap%naxis1 = psfmap%hduinfo%key_naxis1%value
!    psfmap%naxis2 = psfmap%hduinfo%key_naxis2%value
    psfmap%naxis1 = psfmap%key_naxis1%value
    psfmap%naxis2 = psfmap%key_naxis2%value


    ! calculate the effective enrgy of the PSF map.
    ! TODO need a better approxmiation of the effective central energy of the PSF
!    psfmap%energy_eff = 0.5d0*(psfmap%key_energ_lo%value + psfmap%key_energ_hi%value)

    ! extract the Energy range from the CALDB keyword
    p1 = scan(psfmap%key_caldb_energ%value,"(")
    p2 = scan(psfmap%key_caldb_energ%value,"-")
    p3 = scan(psfmap%key_caldb_energ%value,")")
    read(psfmap%key_caldb_energ%value(p1+1:p2-1), '(F10.0)') psfmap%energy_lo
    read(psfmap%key_caldb_energ%value(p2+1:p3-1), '(F10.0)') psfmap%energy_hi

    ! convert from eV to keV
    psfmap%energy_lo = psfmap%energy_lo * EV2KEV
    psfmap%energy_hi = psfmap%energy_hi * EV2KEV

!    write(str_buffer,'(A,A,A10,A,F12.3)'), &
!        trim(psfmap%key_caldb_energ%value), " => '", psfmap%key_caldb_energ%value(p1+1:p2-1), "' =>", psfmap%energy_lo
!    call wrschatty(subroutinename, str_buffer)
!    write(str_buffer,'(A,A,A10,A,F12.3)'), &
!        trim(psfmap%key_caldb_energ%value), " => '", psfmap%key_caldb_energ%value(p2+1:p3-1), "' =>", psfmap%energy_hi
!    call wrschatty(subroutinename, str_buffer)
    if (psfmap%energy_lo >=  0.0d0   .and. &
        psfmap%energy_lo  < 20.0d0   .and. &
        psfmap%energy_hi >   0.0d0   .and. &
        psfmap%energy_hi  < 20.0d4   .and. &
        psfmap%energy_lo  <= psfmap%energy_hi ) then
      ! TODO calculate the proper response-weighted average over the energy range
      ! psfmap%energy_eff = 0.5*(psfmap%energy_lo + psfmap%energy_hi)
       psfmap%energy_eff = psfmap%key_energy%value / 1.d3
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "PSF map energy parameters failed sanity check")
      return
    endif

    ! calculate the effective theta (off axis angle) of the PSF map from the CALDB keyword
    p1 = scan(psfmap%key_caldb_theta%value,"(")
    p2 = scan(psfmap%key_caldb_theta%value,"-")
    p3 = scan(psfmap%key_caldb_theta%value,")")
    read(psfmap%key_caldb_theta%value(p1+1:p2-1), '(F10.0)') psfmap%theta_lo
    read(psfmap%key_caldb_theta%value(p2+1:p3-1), '(F10.0)') psfmap%theta_hi
!    write(str_buffer,'(A,A,A10,A,F12.3)'), &
!        trim(psfmap%key_caldb_theta%value), " => '", psfmap%key_caldb_theta%value(p1+1:p2-1), "' =>", psfmap%theta_lo
!    call wrschatty(subroutinename, str_buffer)
!    write(str_buffer,'(A,A,A10,A,F12.3)'), &
!        trim(psfmap%key_caldb_theta%value), " => '", psfmap%key_caldb_theta%value(p2+1:p3-1), "' =>", psfmap%theta_hi
!    call wrschatty(subroutinename, str_buffer)
    if (psfmap%theta_lo >= 0.0d0             .and. &
        psfmap%theta_lo <= 60.0d0           .and. &
        psfmap%theta_hi >= 0.0d0             .and. &
        psfmap%theta_hi <= 60.0d0           .and. &
        psfmap%theta_lo <= psfmap%theta_hi ) then
      ! TODO calculate the proper response-weighted average over the energy range
      psfmap%theta_eff = 0.5d0*(psfmap%theta_lo + psfmap%theta_hi)
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "PSF map off-axis angle parameters failed sanity check")
      return
    endif


    ! calculate the effective phi (azimuthal angle) of the PSF map
    ! TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO
    psfmap%phi_eff = 0.0d0  ! TODO look for keywords if available
    ! TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO TODO

    ! Keywords for CALDB 2D_PSF maps should have units arcsec/pix -> convert to deg/pix
    if ( psfmap%key_cunit1%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      if ( trim(psfmap%key_cunit1%value) .eq. "arcsec" ) then
        psfmap%pixscale_x = psfmap%key_cdelt1%value * ARCSEC2DEG
      endif
    endif
    if ( psfmap%key_cunit2%flag .eq. FITSWRAP_KEY_FLAG_SUPPLIED ) then
      if ( trim(psfmap%key_cunit2%value) .eq. "arcsec" ) then
        psfmap%pixscale_y = psfmap%key_cdelt2%value * ARCSEC2DEG
      endif
    endif
    psfmap%inv_pixscale_x = 1.0d0 / psfmap%pixscale_x
    psfmap%inv_pixscale_y = 1.0d0 / psfmap%pixscale_y

    ! calc the area per pixel - in deg^2
    psfmap%area_per_pix = abs(psfmap%pixscale_x * psfmap%pixscale_y)


!    ! convert the cdelt1,2 into pixelscales using the known focal length
!    psfmap%pixscale_x = rad2deg* psfmap%key_cdelt1%value / EROSITA_FOCAL_LENGTH
!    psfmap%pixscale_y = rad2deg* psfmap%key_cdelt2%value / EROSITA_FOCAL_LENGTH
!    psfmap%inv_pixscale_x = 1.0d0 / psfmap%pixscale_x
!    psfmap%inv_pixscale_y = 1.0d0 / psfmap%pixscale_y

!    ! calc the area per pixel - in deg^2
!    psfmap%area_per_pix = abs(psfmap%pixscale_x * psfmap%pixscale_y)

    ! read the datatype of the image
    call FTGIDT(unit, bitpix, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not read datatype")
      return
    end if

    ! now allocate some space
    if ((psfmap%naxis1 .ge. 1) .and. &
        (psfmap%naxis2 .ge. 1) ) then
      allocate(psfmap%ppix(psfmap%naxis1,psfmap%naxis2), STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return

      ! now read the pixel data as single precision reals
      call FTG2DE(unit,group,nullval,psfmap%naxis1,psfmap%naxis1,psfmap%naxis2, psfmap%ppix(:,:),anyf,status)
      if (status .ne. 0) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Could not read pixel data")
        return
      end if

      !------------------------------------------------------------------
      ! Now calculate the minimum and maximum pixels of the above-threshold part of the image
      ! this is currently a really clumsy algorithm
      do i=1,psfmap%naxis1
        do j=1,psfmap%naxis2
          if ( psfmap%ppix(i,j) .gt. psf_thresh ) then
            psfmap%pix_x_min = i
            exit
          end if
        end do
        if ( psfmap%pix_x_min .gt. 0 ) then
          exit
        end if
      end do

      do j=1,psfmap%naxis2
        do i=1,psfmap%naxis1
          if ( psfmap%ppix(i,j) .gt. psf_thresh ) then
            psfmap%pix_y_min = j
            exit
          end if
        end do
        if ( psfmap%pix_y_min .gt. 0 ) then
          exit
        end if
      end do

      do i=psfmap%naxis1,1,-1
        do j=1,psfmap%naxis2
          if ( psfmap%ppix(i,j) .gt. psf_thresh ) then
            psfmap%pix_x_max = i
            exit
          end if
        end do
        if ( psfmap%pix_x_max .gt. 0 ) then
          exit
        end if
      end do

      do j=psfmap%naxis2,1,-1
        do i=1,psfmap%naxis1
          if ( psfmap%ppix(i,j) .gt. psf_thresh ) then
            psfmap%pix_y_max = j
            exit
          end if
        end do
        if ( psfmap%pix_y_max .gt. 0 ) then
          exit
        end if
      end do
      !------------------------------------------------------------------

      write (str_buffer, '(A,I0,A,I0,A,I0,A,A,A,A,A,A,F6.3,A,F6.3,A,A,6(I0,A))') &
          "Reading in: ", psfmap%naxis1, "x", psfmap%naxis2, &
          " [bitpix=", bitpix, "] pixels of PSFMAP info from extname='", &
          trim(psfmap%hduinfo%key_extname%value), "' instrument='", &
          trim(psfmap%hduinfo%key_instrume%value), "' ", &
!          "pixscale=[", psfmap%pixscale_x*deg2arcsec, ",", psfmap%pixscale_x*deg2arcsec, "] arcsec/pix"
!          "pixscale=[", psfmap%key_cdelt1%value*METRES2UM, ",", psfmap%key_cdelt2%value*METRES2UM, "] um/pix"
          "pixscale=[", psfmap%key_cdelt1%value, ",", psfmap%key_cdelt2%value, "] arcsec/pix", &
          " above-thresh bounds=[", psfmap%pix_x_min, ",", psfmap%pix_x_max, &
          "][", psfmap%pix_y_min, ",", psfmap%pix_y_max, "] = ", &
          (psfmap%pix_x_max-psfmap%pix_x_min+1), "x", &
          (psfmap%pix_y_max-psfmap%pix_y_min+1), " pix"
      call wrschatty (subroutinename, str_buffer)


    else
      write (str_buffer, '(A,I0,A,I0,A)') &
          "strange image dimensions: (", psfmap%naxis1, "x", psfmap%naxis2, ")"
!      call wrschatty (subroutinename, str_buffer)
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end if


    psfmap%is_valid = .true.

  end subroutine srctool_psf_read_psfmap_from_chdu
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the psf description from the current hdu of the fits file identified by unit
  subroutine srctool_psf_read_psf_from_chdu(unit, psf, ierr)
    character(len=*), parameter :: subroutinename='srctool_psf_read_psf_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(psf_type), intent(inout)     :: psf
    integer(kind=int32), intent(out)  :: ierr

    type(psfmap_type), allocatable, dimension(:) :: temp_psfmaps

    integer(kind=int32) :: i
    integer(kind=int32) :: status
    integer(kind=int32) :: p
    character(len=STR_MAX) :: str_buffer
    type(hduinfo_type) :: hduinfo

    ierr = ERROR_LEVEL_OK
    status = 0
    call fitswrap_init_hduinfo(hduinfo)

    ! get the hdu info/keys
    call fitswrap_read_hduinfo_quick(unit, hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    !- check the following keader keywords are present - TODO
    ! for psf%type .eq. SRCTOOL_PSF_TYPE_MAP
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1       'Image'
    !- HDUCLAS2       'PSF'
    !- HDUCLAS3       'PREDICTED'
    !- HDUCLAS4       'NET'

    ! Determine the PSF description type from the keywords
    ! The PSF type must be the same in all PSF hdus for each eROSITA instrument

    ! For now we only allow PSF maps type
    psf%class = SRCTOOL_PSF_CLASS_MAP

    select case ( psf%class )
    case ( SRCTOOL_PSF_CLASS_MAP )
      if ( hduinfo%hdu_type .ne. FITSIO_IMAGE_HDU ) then
        ierr = ERROR_LEVEL_WARNING
        write (str_buffer, '(A,I0)') "Expected image HDU but hdu_type= ", hduinfo%hdu_type
        call wrserr (subroutinename, ierr, str_buffer)
        return
      end if

      if( .not. allocated(psf%psfmaplist%psfmap) ) then
         allocate(psf%psfmaplist%psfmap(16)) ! reasonable initial size
         do i=1, size(psf%psfmaplist%psfmap)
            call srctool_psf_empty_psfmap(psf%psfmaplist%psfmap(i))
         end do
      end if

      p = psf%psfmaplist%num_psfmaps + 1
      if( p > size(psf%psfmaplist%psfmap) ) then
         ! deal with expanding size of psf list if necessary
         ! copy to temporary
         call move_alloc(psf%psfmaplist%psfmap, temp_psfmaps)
         allocate(psf%psfmaplist%psfmap(psf%psfmaplist%num_psfmaps*2))
         ! replace and initialise
         psf%psfmaplist%psfmap(1:psf%psfmaplist%num_psfmaps) = temp_psfmaps
         do i=psf%psfmaplist%num_psfmaps+1, size(psf%psfmaplist%psfmap)
            call srctool_psf_empty_psfmap(psf%psfmaplist%psfmap(i))
         end do
         ! clean temporary
         deallocate(temp_psfmaps)
      end if

      call srctool_psf_read_psfmap_from_chdu(unit, psf%psfmaplist%psfmap(p), ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem reading PSFMAP")
        return
      end if
      psf%psfmaplist%num_psfmaps = p
      psf%psfmaplist%is_valid = .true.
      psf%is_valid = .true.

      ! now compute the RPSF/REEF
      call srctool_psf_calc_rpsf(psf%psfmaplist%psfmap(p), psf%psfmaplist%psfmap(p)%rpsf,ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem computing  RPSF/REEF from PSFMAP")
        return
      end if



    case default
      ierr = ERROR_LEVEL_ERROR
      call wrserr (subroutinename, ierr, &
          "I cannot handle this class of PSF desc: " // SRCTOOL_PSF_CLASS_STR(psf%class))
      return
    end select

  end subroutine srctool_psf_read_psf_from_chdu
  !---------------------------------------------------------------------------------------------




  !---------------------------------------------------------------------------------------------
  ! organise the list of unique energies, theta and phis represented in
  ! a list of PSFMAPs
  ! version that handles the way that ENERGY and THETA are expressed in CALDB 2D_PSF files
  subroutine srctool_psf_index_psfmaplist_caldb(psfmaplist, ierr)
    character(len=*), parameter :: subroutinename='srctool_psf_index_psfmaplist_caldb'
    type(psfmaplist_type), intent(inout), target    :: psfmaplist
!    type(instpar_type), intent(in), pointer   :: pinstpar
    integer(kind=int32), intent(out) :: ierr
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i, j, ne, ntheta, nphi, ie, itheta, iphi, npsfs
    type(psfmap_type), pointer    :: ppsfmapi,ppsfmapj
    integer(kind=int32) :: status
    ierr = ERROR_LEVEL_OK
    status = 0

    psfmaplist%num_e     = 0
    psfmaplist%num_theta = 0
    psfmaplist%num_phi   = 0
    do i=1,psfmaplist%num_psfmaps
      ppsfmapi => psfmaplist%psfmap(i)
      ne     = 0
      ntheta = 0
      nphi   = 0
      do j=1,i-1
        ppsfmapj => psfmaplist%psfmap(j)
        ! check to see if the new map is unique in e,theta or phi
!        if ( eq_real(ppsfmapi%key_energy%value,ppsfmapj%key_energy%value) ) then
        if ( eq_real(ppsfmapi%energy_eff,ppsfmapj%energy_eff) ) then
          ne = ne + 1
        end if
        if ( eq_real(ppsfmapi%theta_eff,ppsfmapj%theta_eff) ) then
!        if ( eq_real(ppsfmapi%key_theta%value,ppsfmapj%key_theta%value) ) then
          ntheta = ntheta + 1
        end if
        if ( eq_real(ppsfmapi%phi_eff,ppsfmapj%phi_eff) ) then
!        if ( eq_real(ppsfmapi%key_phi%value,ppsfmapj%key_phi%value) ) then
          nphi = nphi + 1
        end if
      end do
      if ( ne .eq. 0 ) then
        psfmaplist%num_e = psfmaplist%num_e + 1
      end if
      if ( ntheta .eq. 0 ) then
        psfmaplist%num_theta = psfmaplist%num_theta + 1
      end if
      if ( nphi .eq. 0 ) then
        psfmaplist%num_phi = psfmaplist%num_phi + 1
      end if
    end do

    if (psfmaplist%num_e     .le. 0 .or. &
        psfmaplist%num_theta .le. 0 .or. &
        psfmaplist%num_phi   .le. 0 ) then
      write (str_buffer, '(A,I0,A,I0,A,I0,A)') &
          "Strange dimensions in PSFmap space (ne,ntheta,nphi)=(", &
          psfmaplist%num_e, ",", &
          psfmaplist%num_theta, ",", &
          psfmaplist%num_phi, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end if
    npsfs = psfmaplist%num_e * psfmaplist%num_theta * psfmaplist%num_phi
    ! make sure that have the correct number of psfmaps to fill the grid
    if ( psfmaplist%num_psfmaps .lt. npsfs  ) then
      write (str_buffer, '(A,I0,A,I0,A,I0,A,I0,A,I0,A)') &
          "There were too few PSFmaps (", psfmaplist%num_psfmaps, "<", npsfs, &
          ") to fill the grid in Energy-Theta-Phi space (ne,ntheta,nphi)=(", &
          psfmaplist%num_e, ",", &
          psfmaplist%num_theta, ",", &
          psfmaplist%num_phi, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    else if ( psfmaplist%num_psfmaps .gt. npsfs ) then
      write (str_buffer, '(A,I0,A,I0,A,I0,A,I0,A,I0,A)') &
          "There were too many PSFmaps (", psfmaplist%num_psfmaps, ">", npsfs, &
          ") to fill the grid in Energy-Theta-Phi space (ne,ntheta,nphi)=(", &
          psfmaplist%num_e, ",", &
          psfmaplist%num_theta, ",", &
          psfmaplist%num_phi, ")"
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end if


    ! make some space
    allocate ( psfmaplist%pe(psfmaplist%num_e), &
        psfmaplist%ptheta_arcmin(psfmaplist%num_theta), &
        psfmaplist%pphi(psfmaplist%num_phi), &
        psfmaplist%pindex(psfmaplist%num_e,psfmaplist%num_theta,psfmaplist%num_phi), &
        STAT = status )

    if ( check_allocate_status(subroutinename, status, ierr) ) return

    psfmaplist%pe     = -huge(psfmaplist%pe)  ! debug check
    psfmaplist%ptheta_arcmin = -huge(psfmaplist%ptheta_arcmin)
    psfmaplist%pphi   = -huge(psfmaplist%pphi)
    psfmaplist%pindex = 0

    ! now do a second pass and write the values to the arrays
    ie = 0
    itheta = 0
    iphi = 0
    do i=1,psfmaplist%num_psfmaps
      ppsfmapi => psfmaplist%psfmap(i)
      ne     = 0
      ntheta = 0
      nphi   = 0
      do j=1,i-1
        ppsfmapj => psfmaplist%psfmap(j)
!        if ( eq_real(ppsfmapi%key_energy%value,ppsfmapj%key_energy%value) ) then
        if ( eq_real(ppsfmapi%energy_eff,ppsfmapj%energy_eff) ) then
          ne = ne + 1
        end if
!        if ( eq_real(ppsfmapi%key_theta%value,ppsfmapj%key_theta%value) ) then
        if ( eq_real(ppsfmapi%theta_eff,ppsfmapj%theta_eff) ) then
          ntheta = ntheta + 1
        end if
!        if ( eq_real(ppsfmapi%key_phi%value,ppsfmapj%key_phi%value) ) then
        if ( eq_real(ppsfmapi%phi_eff,ppsfmapj%phi_eff) ) then
          nphi = nphi + 1
        end if
      end do
      if ( ne .eq. 0 ) then
        ie = ie + 1
!        psfmaplist%pe(ie) = ppsfmapi%key_energy%value
        psfmaplist%pe(ie) = ppsfmapi%energy_eff
      end if
      if ( ntheta .eq. 0 ) then
        itheta = itheta + 1
!        psfmaplist%ptheta_arcmin(itheta) = ppsfmapi%key_theta%value   ! TODO make sure this is in the right units
        psfmaplist%ptheta_arcmin(itheta) = ppsfmapi%theta_eff
      end if
      if ( nphi .eq. 0 ) then
        iphi = iphi + 1
!        psfmaplist%pphi(iphi) = ppsfmapi%key_phi%value
        psfmaplist%pphi(iphi) = ppsfmapi%phi_eff
      end if
    end do

    ! now make sure they are sorted in ascending order
    call insertion_sort(psfmaplist%pe)
    call insertion_sort(psfmaplist%ptheta_arcmin)
    call insertion_sort(psfmaplist%pphi)

    ! now do a third pass to set up the indices
    psfmaplist%pindex = 0
    do i=1,psfmaplist%num_psfmaps
      ppsfmapi => psfmaplist%psfmap(i)
      do ie=1,psfmaplist%num_e
!        if ( eq_real(psfmaplist%pe(ie),ppsfmapi%key_energy%value) ) then
!        if ( near_real(psfmaplist%pe(ie),ppsfmapi%key_energy%value, 1.0d-6) ) then
        if ( near_real(psfmaplist%pe(ie),ppsfmapi%energy_eff, 1.0d-6) ) then
          exit
        end if
      end do
      do itheta=1,psfmaplist%num_theta
!        if ( eq_real(psfmaplist%ptheta_arcmin(itheta),ppsfmapi%key_theta%value) ) then
!        if ( near_real(psfmaplist%ptheta_arcmin(itheta),ppsfmapi%key_theta%value, 1.0d-6) ) then
        if ( near_real(psfmaplist%ptheta_arcmin(itheta),ppsfmapi%theta_eff, 1.0d-6) ) then
          exit
        end if
      end do
      do iphi=1,psfmaplist%num_phi
!        if ( eq_real(psfmaplist%pphi(iphi),ppsfmapi%key_phi%value) ) then
!        if ( near_real(psfmaplist%pphi(iphi),ppsfmapi%key_phi%value, 1.0d-6) ) then
        if ( near_real(psfmaplist%pphi(iphi),ppsfmapi%phi_eff, 1.0d-6) ) then
          exit
        end if
      end do
      ! check that have not already read in a psfmap at this grid location
      if ( psfmaplist%pindex(ie,itheta,iphi) .ne. 0 ) then
        write (str_buffer, '(A,I0,A,I0,A,I0,A)') &
            "There was a repeated PSFmap at (ne,ntheta,nphi)=(", &
            ie, ",", &
            itheta, ",", &
            iphi, ")"
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, str_buffer)
      return

      end if
      psfmaplist%pindex(ie,itheta,iphi) = i
    end do


  end subroutine srctool_psf_index_psfmaplist_caldb
  !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  ! organise the list of unique energies, theta and phis represented in
  ! a list of PSFs
  subroutine srctool_psf_index_psf(psf, ierr)
    character(len=*), parameter :: subroutinename='srctool_psf_index_psf'
    type(psf_type), intent(inout)    :: psf
!    type(instpar_type), intent(in), pointer   :: pinstpar
    integer(kind=int32), intent(out) :: ierr
    character(len=STR_MAX) :: str_buffer
    ierr = ERROR_LEVEL_OK

    select case ( psf%class )
    case ( SRCTOOL_PSF_CLASS_NONE )
      ! do nothing
    case ( SRCTOOL_PSF_CLASS_MAP )

!      call srctool_psf_index_psfmaplist(psf%psfmaplist,pinstpar,ierr)
      call srctool_psf_index_psfmaplist_caldb(psf%psfmaplist,ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem indexing PSFMAPLIST")
        return
      end if

    case default
      ierr = ERROR_LEVEL_ERROR
      write (str_buffer, '(A,A)') &
          "I cannot handle this class of PSF description: ", SRCTOOL_PSF_CLASS_STR(psf%class)
      call wrserr (subroutinename, ierr, str_buffer)
      return
    end select

  end subroutine srctool_psf_index_psf
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  ! organise the list of unique energies, theta and phis represented in
  ! a list of PSFs
  subroutine srctool_psf_index_psfs(instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_psf_index_psfs'
    type(instlist_type), intent(inout) :: instlist
    integer(kind=int32), intent(out) :: ierr
    integer(kind=int32)        :: i
    integer(kind=int32)        :: i_first
!    type(instpar_type), pointer :: pinstpar
    i_first = 0
    ierr = ERROR_LEVEL_OK
    ! first do the TM1...7 insts
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider  ) then
        if ( i_first .eq. 0 ) i_first = i

        if ( instlist%inst(i)%psf%is_valid  ) then
          !get the appropriate instpar pointer
!          call srctool_misc_select_instpar_pointer(instlist%inst(i),pinstpar,ierr)
!          if (ierr .ne. ERROR_LEVEL_OK) then
!            call wrserr (subroutinename, ierr, "Could not find suitable instpar")
!            return
!          end if

          call srctool_psf_index_psf(instlist%inst(i)%psf, ierr)
          if (ierr .ne. ERROR_LEVEL_OK) then
            call wrserr (subroutinename, ierr, "Error indexing psfs")
            return
          end if
        end if
      end if
    end do

    ! now do the merged inst
    if ( i_first .gt. 0 ) then
!      call srctool_misc_select_instpar_pointer(instlist%inst(i_first),pinstpar,ierr)
!      if (ierr .ne. ERROR_LEVEL_OK) then
!        call wrserr (subroutinename, ierr, "Could not find suitable instpar")
!        return
!      end if
      call srctool_psf_index_psf(instlist%inst(0)%psf, ierr)
      if (ierr .ne. ERROR_LEVEL_OK) then
        call wrserr (subroutinename, ierr, "Error indexing psfs")
        return
      end if

    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Could not find suitable inst")
      return
    end if


  end subroutine srctool_psf_index_psfs
  !---------------------------------------------------------------------------------------------



  !---------------------------------------------------------------------------------------------
  ! report the psf info for all instruments
  subroutine srctool_psf_report_psf_info(instlist)
    character(len=*), parameter :: subroutinename='srctool_psf_report_psf_info'
    type(instlist_type), intent(in), target  :: instlist
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i, p, j
    type(psf_type), pointer  :: ppsf
    type(psfmap_type), pointer  :: ppsfmap
    type(psfmaplist_type), pointer  :: ppsfmaplist
!    integer(kind=int32)    :: ierr
!    ierr=ERROR_LEVEL_OK

    j = 0
    do i=0,MAX_INST
      if ( instlist%inst(i)%psf%is_valid ) then
        ppsf => instlist%inst(i)%psf
        j = j + 1
        select case ( ppsf%class )
        case ( SRCTOOL_PSF_CLASS_NONE )
          ! do nothing

        case ( SRCTOOL_PSF_CLASS_MAP )
          if ( j .eq. 1 ) then
            write(str_buffer, '(A12,1x,A6,1x,A5,3(1x,A10))') "#PSFmapInfo", &
                "Inst", "PSF", "effEnergy", "Theta", "Phi"
            call wrschatty (subroutinename, str_buffer)
            write(str_buffer, '(A12,1x,A6,1x,A5,3(1x,A10))') "#PSFmapInfo", &
                "-", "Index", "(keV)", "(arcmin)", "(deg)"
            call wrschatty (subroutinename, str_buffer)
          end if

          ppsfmaplist => ppsf%psfmaplist
          do p=1,ppsf%psfmaplist%num_psfmaps
            ppsfmap => ppsf%psfmaplist%psfmap(p)
            write(str_buffer, '(A12,1x,A6,1x,I5,3(1x,F10.3))') &
                "PSFmapInfo", &
                trim(instlist%inst(i)%name), &
                p, &
!                ppsfmap%key_energ_lo%value, &
!                ppsfmap%key_energ_hi%value, &
                ppsfmap%energy_eff, &
!                ppsfmap%key_energy%value, &
!                ppsfmap%key_theta%value, &
!                ppsfmap%key_phi%value
                ppsfmap%theta_eff, &
                ppsfmap%phi_eff
            call wrschatty (subroutinename, str_buffer)
          end do

          write (str_buffer, '(A12,1x,A6,1x,A,I0)') &
              "PSFmapInfo", trim(instlist%inst(i)%name), "Num unique energy values = ",  ppsfmaplist%num_e
          call wrschatty (subroutinename, str_buffer)
          write (str_buffer, '(A12,1x,A6,1x,A,I0)') &
              "PSFmapInfo", trim(instlist%inst(i)%name), "Num unique theta  values = ",  ppsfmaplist%num_theta
          call wrschatty (subroutinename, str_buffer)
          write (str_buffer, '(A12,1x,A6,1x,A,I0)') &
              "PSFmapInfo", trim(instlist%inst(i)%name), "Num unique phi    values = ",  ppsfmaplist%num_phi
          call wrschatty (subroutinename, str_buffer)

!    do i=1,psfmaplist%num_e
!      write (*, '(A,I0,A,F10.4)'), "DEBUG_PSF:Energy value (", i,") = ", psfmaplist%pe(i)
!    end do
!    do i=1,psfmaplist%num_theta
!      write (*, '(A,I0,A,F10.4)'), "DEBUG_PSF:Theta  value (", i,") = ", psfmaplist%ptheta_arcmin(i)
!    end do
!    do i=1,psfmaplist%num_phi
!      write (*, '(A,I0,A,F10.4)'), "DEBUG_PSF:Phi    value (", i,") = ", psfmaplist%pphi(i)
!    end do


        case default
          return
        end select
      end if
    end do

  end subroutine srctool_psf_report_psf_info
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! find the most appropriate set of psfmaps from a list
  ! assume that the maps have previously been indexed
  ! we want to return pairs of maps which bound the given theta value
  ! we will return a pair of maps for each available energy
  recursive subroutine srctool_psf_get_psfmap_pointers(psfmaplist, theta_arcmin, phi_deg, ppsfmaps )
    character(len=*), parameter  :: subroutinename='srctool_psf_get_psfmap_pointers'
    type(psfmaplist_type), intent(in), target :: psfmaplist
    real(kind=real64), intent(in) :: theta_arcmin, phi_deg
    type(psfmap_pointer), intent(out), dimension(:,:)  :: ppsfmaps
    integer(kind=int32) :: i, itheta1, itheta2, iphi,e

    if ( phi_deg .ge. 0.0d0 ) then
      iphi = 1  ! TODO - actually use the phi values
    else
      iphi = 1  ! TODO - actually use the phi values
    end if

!    ! work out the correct offaxis angles to use
!    ! - find the indices of the maps which bound the given theta value
!    i = 1
!    do while ( theta_arcmin .le. psfmaplist%ptheta_arcmin(i) )
!      if (i .eq. psfmaplist%num_theta ) then
!        exit
!      end if
!      i = i + 1
!    end do


    do i=1,psfmaplist%num_theta
      if ( theta_arcmin .lt. psfmaplist%ptheta_arcmin(i) ) then
        exit
     end if
    end do

    i = min(i,psfmaplist%num_theta)
    itheta1 = max(1,i-1)
    itheta2 = max(1,i)

!    write (*,'(A,1x,F10.4,2(1x,I6),2(1x,F10.4))'), subroutinename, &
!       theta_arcmin, itheta1, itheta2, &
!        psfmaplist%ptheta_arcmin(itheta1), psfmaplist%ptheta_arcmin(itheta2)

    do e=1,psfmaplist%num_e
      ppsfmaps(e,1)%p => psfmaplist%psfmap(psfmaplist%pindex(e,itheta1,iphi))
      ppsfmaps(e,2)%p => psfmaplist%psfmap(psfmaplist%pindex(e,itheta2,iphi))
    end do

  end subroutine srctool_psf_get_psfmap_pointers



  !-------------------------------------------------------------------------------
  ! find the most appropriate single psfmap from a list
  ! assume that the maps have previously been indexed
  ! we want to return map which is nearest the given theta,phi,energy values
  recursive subroutine srctool_psf_get_psfmap_pointer(psfmaplist, theta_arcmin, phi_deg, energy_keV, ppsfmap )
    character(len=*), parameter  :: subroutinename='srctool_psf_get_psfmap_pointer'
    type(psfmaplist_type), intent(in), target :: psfmaplist
    real(kind=real64), intent(in) :: theta_arcmin, phi_deg, energy_keV
    type(psfmap_type), pointer, intent(out)  :: ppsfmap
    integer(kind=int32) :: status, itheta1, itheta2, iphi, ie1, ie2, ie, itheta

    nullify(ppsfmap)

    if ( phi_deg .ge. 0.0d0 ) then
      iphi = 1  ! TODO - actually use the phi values
    else
      iphi = 1  ! TODO - actually use the phi values
    end if

    call binary_search (psfmaplist%ptheta_arcmin(:), theta_arcmin, itheta1, status)
    if ( status .eq. 0 ) then
      itheta2 = min(psfmaplist%num_theta,itheta1+1)
    else if (status .eq. 1 ) then
      itheta1 = 1
      itheta2 = min(psfmaplist%num_theta,2)
    else if (status .eq. 2 ) then
      itheta1 = max(psfmaplist%num_theta-1,1)
      itheta2 = psfmaplist%num_theta
    else
      itheta1 = 1
      itheta2 = 1
      ! this should never happen
    endif

    if (abs(theta_arcmin-psfmaplist%ptheta_arcmin(itheta1)) .lt. &
        abs(theta_arcmin-psfmaplist%ptheta_arcmin(itheta2)) ) then
      itheta = itheta1
    else
      itheta = itheta2
    endif


    call binary_search (psfmaplist%pe(:), energy_keV, ie1, status)
    if ( status .eq. 0 ) then
      ie2 = min(psfmaplist%num_e,ie1+1)
    else if (status .eq. 1 ) then
      ie1 = 1
      ie2 = min(psfmaplist%num_e,2)
    else if (status .eq. 2 ) then
      ie1 = max(psfmaplist%num_e-1,1)
      ie2 = psfmaplist%num_e
    else
      ie1 = 1 ! this should never happen
      ie2 = 1
    endif

    if (abs(energy_keV-psfmaplist%pe(ie1)) .lt. &
        abs(energy_keV-psfmaplist%pe(ie2)) ) then
      ie = ie1
    else
      ie = ie2
    endif

    ppsfmap => psfmaplist%psfmap(psfmaplist%pindex(ie,itheta,iphi))

!    write (*,'(A,1x,F10.4,2(1x,I6),2(1x,F10.4))'), subroutinename, &
!        theta_arcmin, itheta1, itheta2, &
!        psfmaplist%ptheta_arcmin(itheta1), psfmaplist%ptheta_arcmin(itheta2)

  end subroutine srctool_psf_get_psfmap_pointer


  !-------------------------------------------------------------------------------
  ! resample the psf maps at each point in the samplelist
  ! currently we are doing nearest neighbour resampling
  recursive subroutine srctool_psf_get_psfmap_values(inst, psfmaplist, samplelist)
    character(len=*), parameter          :: subroutinename='srctool_psf_get_psfmap_values'
    type(inst_type),       intent(in) :: inst
    type(psfmaplist_type), intent(in), target :: psfmaplist
    type(samplelist_type), intent(inout) :: samplelist
    integer(kind=int32) :: i,e,t
    real(kind=real32)   :: psf_val(2)
    real(kind=real64)   :: pix_x(2), pix_y(2)
    integer(kind=int32) :: pix_x_int(2), pix_y_int(2)
    real(kind=real64)   :: dx, dy  ! coords of sample point relative to src
    real(kind=real64)   :: cosa, sina
    real(kind=real64)   :: rotx,roty
!    type(psfmap_pointer), dimension(:,:), allocatable :: ppsfmaps
!    real(kind=real64), dimension(:,:), allocatable :: sum_psf_val
    type(psfmap_pointer), dimension(size(psfmaplist%psfmap),2) :: ppsfmaps
    real(kind=real32), dimension(size(psfmaplist%psfmap),2) :: sum_psf_val
    type(instpar_type), pointer :: pinstpar


!    integer(kind=int32)        :: counter = 0
!    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: status
    integer(kind=int32) :: ierr  ! TOD should be input param

    ierr = ERROR_LEVEL_OK
!    cosa = cosd(samplelist%src_phi - 90.0d0)
!    sina = sind(samplelist%src_phi - 90.0d0)

    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable instpar")
      return
    end if

    cosa = cosd(samplelist%src_phi)
    sina = sind(samplelist%src_phi)
    sina = -1.0d0*sina
    ! debug
!    if (  mod(counter,10) .eq. 1 ) then
!      write (*, '(A,1x,A4,1x,4(1x,A9),1x,A6,4(1x,A9),1x,A6,1x,A8,1x,A16,2(1x,A9),2(1x,A9),4(1x,A12))'), &
!          "#PSFMAP_SAMP",  "INST",  &
!          "src_rawx", "src_rawy", "theta", "phi", &
!          "samp_i", "rawx(i)", "rawy(i)", "dx_deg", "dy_deg", &
!          "samp_e", "energy", &
!          "PSFEXT", &
!          "pix_x", "pix_y", &
!          "theta1", "theta2", &
!          "psf_val1", "psf_val2", "psf_interp", &
!          "pval(i,e)"
!    end if

    status = 0
    do e=1,psfmaplist%num_e
      do t=1,2
        nullify(ppsfmaps(e,t)%p)
      end do
    end do
!    nullify(ppsfmaps(:,:)%p) ! this throws a compiler error

    ! first choose the set of psfmaps for this theta,phi
    call srctool_psf_get_psfmap_pointers(psfmaplist, &
        samplelist%src_theta_arcmin, samplelist%src_phi, &
        ppsfmaps(:,:) )

    sum_psf_val(:,:) = 0.0e0  ! initialise the sums

    ! now step through the sample points
    do i=1,samplelist%num_points
      if ( samplelist%pon_chip_flag(i) ) then

        ! no point getting psf map value if already in a bad pixel (energy independent):
        if ( samplelist%peffmap(i) .gt. 0.0d0 ) then

          dx = (samplelist%prawx(i) - samplelist%src_rawx)*pinstpar%pixscale_x
          dy = (samplelist%prawy(i) - samplelist%src_rawy)*pinstpar%pixscale_y

          rotx = dx*cosa - dy*sina
          roty = dx*sina + dy*cosa

          ! go through the list of PSF energies
          do e=1,psfmaplist%num_e ! = samplelist%num_psf_e

            ! no point getting psf map value if already in a bad pixel:
            ! This is energy sensitive - assumes that size(samplelist%peffmap(i,:)) = psfmaplist%num_psf_e
!            if ( samplelist%peffmap(i,e) .gt. 0.0d0 ) then

            ! read the values from the two psfmaps which bound the point in theta
            do t=1,2

              ! now find the coords of the sample point in the psf image
              ! TODO account for non linear terms properly - if necessary
              ! use the rawx,rawy coords of sample point and of the src
              ! rotate, scale and offset dx,dy into psf map
!              pix_x(t) = ppsfmaps(e,t)%p%key_crpix1%value + (dx*cosa - dy*sina)*ppsfmaps(e,t)%p%inv_pixscale_x
!              pix_y(t) = ppsfmaps(e,t)%p%key_crpix2%value + (dx*sina + dy*cosa)*ppsfmaps(e,t)%p%inv_pixscale_y
              pix_x(t) = ppsfmaps(e,t)%p%key_crpix1%value + rotx * ppsfmaps(e,t)%p%inv_pixscale_x
              pix_y(t) = ppsfmaps(e,t)%p%key_crpix2%value + roty * ppsfmaps(e,t)%p%inv_pixscale_y

!              pix_x_int(t) = nint(pix_x(t), int32)
!              pix_y_int(t) = nint(pix_y(t), int32)
              pix_x_int(t) = int(pix_x(t)+0.5d0)  ! This is a faster real->integer procedure, should be fine for
              pix_y_int(t) = int(pix_y(t)+0.5d0)  ! real->integer conversions in the range of interest


              ! check if the coord lands inside the psf map and then get the value
              ! should be marginally faster if we nest the if statements
              psf_val(t) = 0.0e0
              if ( pix_x_int(t) .ge. 1 ) then
                if ( pix_y_int(t) .ge. 1 ) then
                  if ( pix_x_int(t) .le. ppsfmaps(e,t)%p%naxis1 ) then
                    if ( pix_y_int(t) .le. ppsfmaps(e,t)%p%naxis2 ) then
                      psf_val(t) = ppsfmaps(e,t)%p%ppix(pix_x_int(t),pix_y_int(t))
                    end if
                  end if
                end if
              end if

              ! better and faster to just interpolate the summed psf values at the end
              ! this unfortunately means that it's not possible to write out a mean psf.
              sum_psf_val(e,t) = sum_psf_val(e,t)  + psf_val(t)
            end do  !t loop

          end do
        end if
      end if
    end do



!    !----------------------------------------------------------------------------------------
!    ! Sum the PSF values over all the sample points
!    samplelist%psum(:)  = sum(samplelist%pval(1:samplelist%num_points,:))
!    !----------------------------------------------------------------------------------------

    ! For each energy value interpolate the summed PSF values between the two theta values
    do e=1,psfmaplist%num_e
!      if ( eq_real(ppsfmaps(e,1)%p%key_theta%value,ppsfmaps(e,2)%p%key_theta%value)) then
      if ( eq_real(ppsfmaps(e,1)%p%theta_eff,ppsfmaps(e,2)%p%theta_eff)) then
        samplelist%ppsfsum(e) = 0.5d0*(sum_psf_val(e,1) + sum_psf_val(e,2))
      else
        samplelist%ppsfsum(e) = sum_psf_val(e,1) + &
            (sum_psf_val(e,2) - sum_psf_val(e,1)) * (samplelist%src_theta_arcmin - ppsfmaps(e,1)%p%theta_eff)/ &
            (ppsfmaps(e,2)%p%theta_eff - ppsfmaps(e,1)%p%theta_eff)
      end if
      samplelist%ppsfsum(e) = samplelist%ppsfsum(e) * &
          samplelist%area_per_point / (0.5*(ppsfmaps(e,1)%p%area_per_pix + ppsfmaps(e,2)%p%area_per_pix ))
      samplelist%ppsfsum(e) = MAX(0.0d0, samplelist%ppsfsum(e))
    end do

!    counter = counter + 1
!    if ( allocated(ppsfmaps) ) deallocate(ppsfmaps)
!    if ( allocated(sum_psf_val) ) deallocate(sum_psf_val)
  end subroutine srctool_psf_get_psfmap_values

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_psf_get_psf_values(inst, psf, samplelist)
    character(len=*), parameter          :: subroutinename='srctool_psf_get_psf_values'
    type(inst_type), intent(in)          :: inst
    type(psf_type), intent(in)           :: psf
    type(samplelist_type), intent(inout) :: samplelist

    select case ( psf%class )
    case ( SRCTOOL_PSF_CLASS_NONE )
      samplelist%ppsfsum(:) = 1.0d0
      ! do nothing
    case ( SRCTOOL_PSF_CLASS_MAP )
      call srctool_psf_get_psfmap_values(inst, psf%psfmaplist, samplelist)
    case ( SRCTOOL_PSF_CLASS_ANALYTIC )
      ! not yet implemented so should never get here
      samplelist%ppsfsum(:) = -1.0d30
    case ( SRCTOOL_PSF_CLASS_SHAPELETS )
      ! not yet implemented so should never get here
      samplelist%ppsfsum(:) = -2.0d30
    case default
      ! should never get here
      samplelist%ppsfsum(:) = -99.0d30
    end select

  end subroutine srctool_psf_get_psf_values

  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! Convolve the PSF map with the source extent model
  recursive subroutine srctool_psf_convolve_psfmap_with_extent(instlist, psfmap_in, psfmap_out, extent, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_convolve_psfmap_with_extent'
    type(instlist_type), intent(in) :: instlist
    type(psfmap_type), intent(inout) :: psfmap_in
    type(psfmap_type), intent(inout) :: psfmap_out
    type(extent_type), intent(in) :: extent
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32) :: status
    integer(kind=int32) :: fft_npix, out_pix, num_conv
    real(kind=real32) :: fft_scale
    real(kind=real64) :: buffer_pix

    integer(kind=int32) :: fft_npix_d2
    integer(kind=int32) :: cx, cy, dx, dy, ix, iy

    real(kind=real32), pointer :: real_arr(:,:)
    complex(kind=4), pointer :: complex_arr(:,:)
    complex(kind=4), pointer :: psf_fft(:,:)

    ierr = ERROR_LEVEL_OK
    status = 0

    fft_npix = instlist%fft_npix
    fft_scale=1.0/real((fft_npix*fft_npix), real32)
    fft_npix_d2 = fft_npix/2

    ! output size is size of extent + psf size (half of fft size)
    out_pix = fft_npix_d2 + extent%num_pix
    psfmap_out%naxis1 = out_pix
    psfmap_out%naxis2 = out_pix

    ! allocate and initialise output
    allocate(psfmap_out%ppix(out_pix, out_pix), stat=status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return
    psfmap_out%ppix = 0

    ! temporary allocations for doing convoltions
    call fftw_alloc_real_array(fft_npix, fft_npix, real_arr)
    call fftw_alloc_complex_array(fft_npix/2+1, fft_npix, complex_arr)
    call fftw_alloc_complex_array(fft_npix/2+1, fft_npix, psf_fft)

    ! get the transform of the PSF
    real_arr = 0
    real_arr(:size(psfmap_in%ppix,1), :size(psfmap_in%ppix,2)) = psfmap_in%ppix
    call fftwf_execute_dft_r2c(instlist%fftw_plan_r2c, real_arr, complex_arr)
    psf_fft = complex_arr

    ! use the overlap-add method for doing the convolution so we
    ! don't have to do one huge convolution

    ! we split the extent into multiple parts of size of psf
    ! each output convolution is added to the total

    ! get number of convolutions (note we move across the image by half psf)
    num_conv = extent%num_pix / fft_npix_d2
    if( mod(extent%num_pix, fft_npix_d2) /= 0 ) num_conv = num_conv+1

    do cx=0, num_conv-1
       do cy=0, num_conv-1

          ! take extent model in (fft_npix_d2, fft_npix_d2) and zero pad
          real_arr = 0
          do dx=1, fft_npix_d2
             do dy=1, fft_npix_d2
                ix = cx*fft_npix_d2 + dx
                iy = cy*fft_npix_d2 + dy

                if( ix<=extent%num_pix .and. iy<=extent%num_pix ) then
                   real_arr(dx,dy) = extent%xy(ix,iy)
                end if
             end do
          end do

          ! now take fft and put into psfmap_in%temp_uv
          call fftwf_execute_dft_r2c(instlist%fftw_plan_r2c, real_arr, complex_arr)

          ! multiply by psf in fft space (taking into account the scale factor)
          complex_arr = complex_arr * psf_fft * fft_scale

          ! put back into real space
          call fftwf_execute_dft_c2r(instlist%fftw_plan_c2r, complex_arr, real_arr)

          ! now add everything back onto output
          do dx=1, fft_npix
             do dy=1, fft_npix
                ix = cx*fft_npix_d2 + dx
                iy = cy*fft_npix_d2 + dy

                if( ix<=out_pix .and. iy<=out_pix ) then
                   psfmap_out%ppix(ix, iy) = psfmap_out%ppix(ix, iy) + real_arr(dx, dy)
                end if
             end do
          end do

       end do
    end do

    call fftw_free_complex_array(psf_fft)
    call fftw_free_complex_array(complex_arr)
    call fftw_free_real_array(real_arr)

    ! the new PSF will be offset by half of the size of the extent in pixels
    buffer_pix = extent%num_pix/2.0d0

    ! now copy the meta data
    psfmap_out%hduinfo  = psfmap_in%hduinfo

!    psfmap_out%hduinfo%key_naxis1       = psfmap_in%hduinfo%key_naxis1
!    psfmap_out%hduinfo%key_naxis2       = psfmap_in%hduinfo%key_naxis2
!    psfmap_out%hduinfo%key_naxis1%value = psfmap_out%naxis1
!    psfmap_out%hduinfo%key_naxis2%value = psfmap_out%naxis2

    psfmap_out%key_backgrnd     = psfmap_in%key_backgrnd
    psfmap_out%key_crpix1       = psfmap_in%key_crpix1
    psfmap_out%key_crpix2       = psfmap_in%key_crpix2
    psfmap_out%key_crpix1%value = psfmap_in%key_crpix1%value  + buffer_pix
    psfmap_out%key_crpix2%value = psfmap_in%key_crpix2%value  + buffer_pix
    psfmap_out%key_crval1       = psfmap_in%key_crval1
    psfmap_out%key_crval2       = psfmap_in%key_crval2
    psfmap_out%key_cdelt1       = psfmap_in%key_cdelt1
    psfmap_out%key_cdelt2       = psfmap_in%key_cdelt2
    psfmap_out%key_cunit1       = psfmap_in%key_cunit1
    psfmap_out%key_cunit2       = psfmap_in%key_cunit2
    psfmap_out%key_naxis1       = psfmap_in%key_naxis1
    psfmap_out%key_naxis2       = psfmap_in%key_naxis2
    psfmap_out%key_naxis1%value = psfmap_out%naxis1      ! make sure to use the new values here
    psfmap_out%key_naxis2%value = psfmap_out%naxis2      !
    psfmap_out%key_caldb_energ  = psfmap_in%key_caldb_energ
    psfmap_out%key_caldb_theta  = psfmap_in%key_caldb_theta
    psfmap_out%key_caldb_obsmode= psfmap_in%key_caldb_obsmode

    psfmap_out%focal_length   = psfmap_in%focal_length
    psfmap_out%pixscale_x     = psfmap_in%pixscale_x
    psfmap_out%pixscale_y     = psfmap_in%pixscale_y
    psfmap_out%inv_pixscale_x = psfmap_in%inv_pixscale_x
    psfmap_out%inv_pixscale_y = psfmap_in%inv_pixscale_y
    psfmap_out%area_per_pix   = psfmap_in%area_per_pix
    psfmap_out%energy_lo      = psfmap_in%energy_lo
    psfmap_out%energy_hi      = psfmap_in%energy_hi
    psfmap_out%energy_eff     = psfmap_in%energy_eff
    psfmap_out%theta_lo       = psfmap_in%theta_lo
    psfmap_out%theta_hi       = psfmap_in%theta_hi
    psfmap_out%theta_eff      = psfmap_in%theta_eff
    psfmap_out%phi_lo         = psfmap_in%phi_lo
    psfmap_out%phi_hi         = psfmap_in%phi_hi
    psfmap_out%phi_eff        = psfmap_in%phi_eff

    psfmap_out%is_valid  = .true.


  end subroutine srctool_psf_convolve_psfmap_with_extent
  !-------------------------------------------------------------------------------------





  !-------------------------------------------------------------------------------
  ! Convolve the PSF maplist with the source extent model
  recursive subroutine srctool_psf_convolve_psfmaplist_with_extent(instlist, psfmaplist_in, psfmaplist_out, extent, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_convolve_psfmaplist_with_extent'
    type(instlist_type), intent(in) :: instlist
    type(psfmaplist_type), intent(inout) :: psfmaplist_in
    type(psfmaplist_type), intent(inout) :: psfmaplist_out
    type(extent_type), intent(in) :: extent
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    integer(kind=int32) :: status
    ierr=ERROR_LEVEL_OK
    status = 0

    psfmaplist_out%num_psfmaps = psfmaplist_in%num_psfmaps
    psfmaplist_out%num_e       = psfmaplist_in%num_e
    psfmaplist_out%num_theta   = psfmaplist_in%num_theta
    psfmaplist_out%num_phi     = psfmaplist_in%num_phi

    allocate(psfmaplist_out%pe(psfmaplist_out%num_e), &
             psfmaplist_out%ptheta_arcmin(psfmaplist_out%num_theta), &
             psfmaplist_out%pphi(psfmaplist_out%num_phi), &
             psfmaplist_out%pindex(psfmaplist_out%num_e, &
             psfmaplist_out%num_theta, &
             psfmaplist_out%num_phi), &
             psfmaplist_out%psfmap(psfmaplist_in%num_psfmaps), &
             STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    psfmaplist_out%pe(:)             = psfmaplist_in%pe(:)
    psfmaplist_out%ptheta_arcmin(:)  = psfmaplist_in%ptheta_arcmin(:)
    psfmaplist_out%pphi(:)           = psfmaplist_in%pphi(:)
    psfmaplist_out%pindex(:,:,:)     = psfmaplist_in%pindex(:,:,:)

    do i=1,psfmaplist_in%num_psfmaps
      call srctool_psf_empty_psfmap(psfmaplist_out%psfmap(i))

      call srctool_psf_convolve_psfmap_with_extent(instlist, psfmaplist_in%psfmap(i), &
           psfmaplist_out%psfmap(i), extent, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem convolving PSF map with source extent model")
        return
      end if
    end do

    psfmaplist_out%is_valid   = .true.

  end subroutine srctool_psf_convolve_psfmaplist_with_extent
  !-------------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! Convolve the PSF model with the source extent model
  recursive subroutine srctool_psf_convolve_psf_with_extent(instlist, psf_in, psf_out, extent, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_convolve_psf_with_extent'
    type(instlist_type), intent(in) :: instlist
    type(psf_type), intent(inout) :: psf_in
    type(psf_type), intent(inout) :: psf_out
    type(extent_type), intent(in) :: extent
    integer(kind=int32), intent(out)   :: ierr
    ierr=ERROR_LEVEL_OK

    ! initialise the structure
    call srctool_psf_empty_psf(psf_out)

    ! do convolution differently depending on how the PSF is described
    select case ( psf_in%class )
    case ( SRCTOOL_PSF_CLASS_NONE )
      ! do nothing
    case ( SRCTOOL_PSF_CLASS_MAP )

      call srctool_psf_convolve_psfmaplist_with_extent(instlist, psf_in%psfmaplist, &
           psf_out%psfmaplist, extent, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem convolving PSF maplist")
        return
      end if

    case default
      ierr = ERROR_LEVEL_ERROR
      call wrserr (subroutinename, ierr, &
          "I cannot handle this class of PSF description: " // &
          SRCTOOL_PSF_CLASS_STR(psf_in%class))
      return
    end select
    psf_out%class = psf_in%class
    psf_out%hduinfo = psf_in%hduinfo
    psf_out%is_valid = .true.

  end subroutine srctool_psf_convolve_psf_with_extent
  !-------------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------------
  subroutine srctool_psf_write_psfmap(filename, psfmap, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_write_psfmap'
    character(len=*), intent(in) :: filename
    type(psfmap_type), intent(in) :: psfmap
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)    :: unit
    integer(kind=int32) :: status

    integer(kind=int32) :: bitpix
    integer(kind=int32), parameter :: naxes = 2
    integer(kind=int32) :: naxis(2)

    ierr=ERROR_LEVEL_OK
    status = 0

    call wrschatty(subroutinename, "Writing out psfmap to => " // trim(filename))
    !-----------------------------------
    ! get a fortran unit
    call FTGIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT(unit,filename,FITSIO_BLOCKSIZE, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

    !-----------------------------------
    ! append a new primary image hdu and write date keyword to the HDU
    naxis(1) = psfmap%naxis1
    naxis(2) = psfmap%naxis2
    bitpix = -64
    call FTIIMG(unit,bitpix,naxes,naxis(:), status)
    call FTPDAT(unit, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
!    call fitswrap_write_hduinfo_quiet(unit, psfmap%hduinfo, ierr)
!    if ( ierr .ne. ERROR_LEVEL_OK ) then
!      return
!    endif
    ! write the PSF xy data itself
!    call FTP2DD(unit, 0, naxis(1),naxis(1),naxis(2),psfmap%ppix(:,:), status)
    call FTP2DE(unit, 0, naxis(1),naxis(1),naxis(2),psfmap%ppix(:,:), status)

    call fitswrap_write_key(unit, psfmap%key_crpix1, ierr)
    call fitswrap_write_key(unit, psfmap%key_crpix2, ierr)
    call fitswrap_write_key(unit, psfmap%key_crval1, ierr)
    call fitswrap_write_key(unit, psfmap%key_crval2, ierr)
    call fitswrap_write_key(unit, psfmap%key_cdelt1, ierr)
    call fitswrap_write_key(unit, psfmap%key_cdelt2, ierr)
    call fitswrap_write_key(unit, psfmap%key_cunit1, ierr)
    call fitswrap_write_key(unit, psfmap%key_cunit2, ierr)

    !-----------------------------------


    !-----------------------------------
    ! now close the fits file
    call FTCLOS(unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------



  end subroutine srctool_psf_write_psfmap
  !-------------------------------------------------------------------------------------


  subroutine srctool_psf_calc_rpsf(psfmap, rpsf, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_calc_rpsf'
    type(psfmap_type), intent(inout), target :: psfmap
    type(rpsf_type), intent(inout), target :: rpsf
    integer(kind=int32), intent(out)   :: ierr
    real(kind=real64)      :: max_radius_deg, x0, y0, r, max_radius_deg_x, max_radius_deg_y
    integer(kind=int32) :: nsteps, i, j, rbin
!    type(rpsf_type), pointer :: prpsf
!    real(kind=real64), parameter :: dradius_deg = 1.0d0/3600.0d0 ! one arcsec
    integer(kind=int32)                :: status

!    real(kind=real64),parameter :: oversample_factor = 0.5d0
    real(kind=real64),parameter :: oversample_factor = sqrt(2.0d0)/2.0d0

    character(len=STR_MAX) :: str_buffer
    ierr=ERROR_LEVEL_OK
    status = 0

    x0 = psfmap%key_crpix1%value  ! assume that these keys give the pixel coords of the PSF centroid
    y0 = psfmap%key_crpix2%value

!    prpsf => psfmap%rpsf
!    prpsf%dradius_deg = dradius_deg
    ! oversample the pixels, being cautious not to leave empty annuli
    rpsf%dradius_deg = oversample_factor*MIN(psfmap%pixscale_x,psfmap%pixscale_y)
    ! first compute the radial steps to use
    ! use the maximum radius that will fit completely within the PSF map, taking account of the
    ! possible off-centre location of the centroid
    max_radius_deg_x = psfmap%pixscale_x * max(abs(psfmap%key_naxis1%value-x0), &
         abs(x0-psfmap%key_naxis1%value))
    max_radius_deg_y = psfmap%pixscale_y * max(abs(psfmap%key_naxis2%value-y0), &
         abs(y0-psfmap%key_naxis2%value))
!    max_radius_deg = rpsf%dradius_deg * MIN(psfmap%key_naxis1%value, psfmap%key_naxis2%value)
    max_radius_deg = max(max_radius_deg_x,max_radius_deg_y)
    nsteps = ceiling(max_radius_deg / rpsf%dradius_deg, kind=int32)


!    write (str_buffer, '(A,A,A,I0)'),    "RPSF for PSFmap:", trim(psfmap%hduinfo%key_extname%value), " nsteps=", nsteps
!    call wrschatty (subroutinename, str_buffer )
!    write (str_buffer, '(A,A,A,F10.7)'), "RPSF for PSFmap:", trim(psfmap%hduinfo%key_extname%value), " max_radius_deg=", max_radius_deg
!    call wrschatty (subroutinename, str_buffer )
!    write (str_buffer, '(A,A,A,F10.7)'), "RPSF for PSFmap:", trim(psfmap%hduinfo%key_extname%value), " dradius_deg=", rpsf%dradius_deg
!    call wrschatty (subroutinename, str_buffer )

!    call fitswrap_sprint_key(psfmap%key_naxis1, str_buffer)
!    call wrschatty (subroutinename, str_buffer )
!    call fitswrap_sprint_key(psfmap%key_naxis2, str_buffer)
!    call wrschatty (subroutinename, str_buffer )
!    call fitswrap_sprint_key(psfmap%key_crpix1, str_buffer)
!    call wrschatty (subroutinename, str_buffer )
!    call fitswrap_sprint_key(psfmap%key_crpix2, str_buffer)
!    call wrschatty (subroutinename, str_buffer )

    allocate (&
        rpsf%pradius(nsteps), &
        rpsf%parea(nsteps), &
        rpsf%prpsf(nsteps), &
        rpsf%preef(nsteps), &
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    do i=1,nsteps
      rpsf%pradius(i) = real(i,kind=real64) * rpsf%dradius_deg
      rpsf%parea(i)   = PI * (rpsf%pradius(i)**2 - (rpsf%pradius(i) - rpsf%dradius_deg)**2)
    end do
    rpsf%prpsf(:) = 0.0d0
    rpsf%preef(:) = 0.0d0

    do i=1,psfmap%key_naxis1%value
      do j=1,psfmap%key_naxis2%value
        r = sqrt( &
             ((real(i,kind=real64) - x0)*psfmap%pixscale_x)**2 + &
             ((real(j,kind=real64) - y0)*psfmap%pixscale_y)**2 )
        rbin = max(1,floor(r / rpsf%dradius_deg, kind=int32))
        if ( (rbin .ge. 1) .and. (rbin .le. nsteps) ) then
          rpsf%prpsf(rbin) = rpsf%prpsf(rbin) + real(psfmap%ppix(i,j),kind=real32)
        endif
      end do
    end do

    rpsf%preef(1) = rpsf%prpsf(1)
    do i=2,nsteps
      rpsf%preef(i) = rpsf%preef(i-1) + rpsf%prpsf(i)
    end do

!    ! DEBUG Write the RPSF to stdout
!    call wrschatty (subroutinename, "DEBUG RPSF for PSFmap: " // trim(psfmap%hduinfo%key_extname%value))
!    write (str_buffer, '(A,x,A,x,A8,x,A10,x,A10,x,A10,x,A10,A10,x,A10)'), &
!        "#RPSF", trim(psfmap%hduinfo%key_extname%value), "i", "r_in", "r_out", "RPSF", "REEF", "Energy", "Theta"
!      call wrschatty (subroutinename, str_buffer)
!    do i=1,nsteps
!      write (str_buffer, '(A,x,A,x,I8,x,F10.7,x,F10.7,x,F10.7,x,F10.7,F10.3,x,F10.3)'), &
!          "RPSF ", trim(psfmap%hduinfo%key_extname%value), &
!          i, rpsf%pradius(i)-rpsf%dradius_deg, rpsf%pradius(i), rpsf%prpsf(i), rpsf%preef(i), &
!          psfmap%energy_eff, psfmap%theta_eff
!      call wrschatty (subroutinename, str_buffer)
!    end do


    ! TODO handle error conditions, and interpolate
    call binary_search(rpsf%preef(:), 0.5d0, i, status)
    rpsf%radius_50pc_deg = rpsf%pradius(i)
    call binary_search(rpsf%preef(:), 0.9d0, i, status)
    rpsf%radius_90pc_deg = rpsf%pradius(i)
    call binary_search(rpsf%preef(:), 0.99d0, i, status)
    rpsf%radius_99pc_deg = rpsf%pradius(i)
    write (str_buffer, '(A,A24,A,F7.2,A,F7.2,A,F7.2,A,F7.2,A)') &
        "PSFmap:", trim(psfmap%hduinfo%key_extname%value), &
        " HEW= ",           rpsf%radius_50pc_deg*2.0d0*DEG2ARCSEC, &
        " arcsec; 90%EW= ", rpsf%radius_90pc_deg*2.0d0*DEG2ARCSEC, &
        " arcsec; 99%EW= ", rpsf%radius_99pc_deg*2.0d0*DEG2ARCSEC, &
        " arcsec; max_radius= ", max_radius_deg*DEG2ARCSEC, " arcsec"
    call wrschatty (subroutinename, str_buffer )


    ! set flag to show that the rpsf is now ready for use
    rpsf%is_valid = .true.

  end subroutine srctool_psf_calc_rpsf


  ! duplicate a psfmap, allocating space as necessary
  subroutine srctool_psf_copy_psfmap(psfmap_in, psfmap_out, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_copy_psfmap'
    type(psfmap_type), intent(in) :: psfmap_in
    type(psfmap_type), intent(inout) :: psfmap_out
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status
    ierr = ERROR_LEVEL_OK
    status = 0

    psfmap_out%is_valid          = psfmap_in%is_valid
    psfmap_out%hduinfo           = psfmap_in%hduinfo
    psfmap_out%key_backgrnd      = psfmap_in%key_backgrnd
    psfmap_out%key_crpix1        = psfmap_in%key_crpix1
    psfmap_out%key_crpix2        = psfmap_in%key_crpix2
    psfmap_out%key_crval1        = psfmap_in%key_crval1
    psfmap_out%key_crval2        = psfmap_in%key_crval2
    psfmap_out%key_cdelt1        = psfmap_in%key_cdelt1
    psfmap_out%key_cdelt2        = psfmap_in%key_cdelt2
    psfmap_out%key_cunit1        = psfmap_in%key_cunit1
    psfmap_out%key_cunit2        = psfmap_in%key_cunit2
    psfmap_out%key_naxis1        = psfmap_in%key_naxis1
    psfmap_out%key_naxis2        = psfmap_in%key_naxis2
    psfmap_out%key_caldb_energ   = psfmap_in%key_caldb_energ
    psfmap_out%key_caldb_theta   = psfmap_in%key_caldb_theta
    psfmap_out%key_caldb_obsmode = psfmap_in%key_caldb_obsmode
    psfmap_out%naxis1            = psfmap_in%naxis1
    psfmap_out%naxis2            = psfmap_in%naxis2
    psfmap_out%pix_x_min         = psfmap_in%pix_x_min
    psfmap_out%pix_x_max         = psfmap_in%pix_x_max
    psfmap_out%pix_y_min         = psfmap_in%pix_y_min
    psfmap_out%pix_y_max         = psfmap_in%pix_y_max
    psfmap_out%focal_length      = psfmap_in%focal_length
    psfmap_out%pixscale_x        = psfmap_in%pixscale_x
    psfmap_out%pixscale_y        = psfmap_in%pixscale_y
    psfmap_out%inv_pixscale_x    = psfmap_in%inv_pixscale_x
    psfmap_out%inv_pixscale_y    = psfmap_in%inv_pixscale_y
    psfmap_out%area_per_pix      = psfmap_in%area_per_pix
    psfmap_out%energy_lo         = psfmap_in%energy_lo
    psfmap_out%energy_hi         = psfmap_in%energy_hi
    psfmap_out%energy_eff        = psfmap_in%energy_eff
    psfmap_out%theta_lo          = psfmap_in%theta_lo
    psfmap_out%theta_hi          = psfmap_in%theta_hi
    psfmap_out%theta_eff         = psfmap_in%theta_eff
    psfmap_out%phi_lo            = psfmap_in%phi_lo
    psfmap_out%phi_hi            = psfmap_in%phi_hi
    psfmap_out%phi_eff           = psfmap_in%phi_eff

    ! make space for map and copy values
    if ( allocated(psfmap_in%ppix) .and. status .eq. 0 ) then
      allocate(psfmap_out%ppix(psfmap_out%naxis1,psfmap_out%naxis2), STAT = status)
      psfmap_out%ppix(:,:) = psfmap_in%ppix(:,:)
    end if

    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! don't allocate/copy the uv, temp_uv,temp_xy arrays
    ! don't copy the rpsf structure


  end subroutine srctool_psf_copy_psfmap




  ! duplicate a psfmaplist, allocating space as necessary
  subroutine srctool_psf_copy_psfmaplist(psfmaplist_in, psfmaplist_out, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_copy_psfmaplist'
    type(psfmaplist_type), intent(in) :: psfmaplist_in
    type(psfmaplist_type), intent(inout) :: psfmaplist_out
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status, i
    ierr = ERROR_LEVEL_OK
    status = 0

    psfmaplist_out%num_psfmaps    = psfmaplist_in%num_psfmaps
    psfmaplist_out%num_e          = psfmaplist_in%num_e
    psfmaplist_out%num_theta      = psfmaplist_in%num_theta
    psfmaplist_out%num_phi        = psfmaplist_in%num_phi

    allocate(psfmaplist_out%pe(psfmaplist_out%num_e), &
             psfmaplist_out%ptheta_arcmin(psfmaplist_out%num_theta),&
             psfmaplist_out%pphi(psfmaplist_out%num_phi), &
             psfmaplist_out%pindex(psfmaplist_out%num_e, &
             psfmaplist_out%num_theta,psfmaplist_out%num_phi), &
             psfmaplist_out%psfmap(psfmaplist_out%num_psfmaps), &
             STAT=status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    do i=1,psfmaplist_in%num_psfmaps
      call srctool_psf_copy_psfmap(psfmaplist_in%psfmap(i), &
           psfmaplist_out%psfmap(i), ierr)
      if (ierr .ne. ERROR_LEVEL_OK) then
        call wrserr(subroutinename, ierr, "Problem copying psfmaplist")
      endif
    end do

    psfmaplist_out%pe(:)     = psfmaplist_in%pe(:)
    psfmaplist_out%ptheta_arcmin(:) = psfmaplist_in%ptheta_arcmin(:)
    psfmaplist_out%pphi(:)   = psfmaplist_in%pphi(:)
    psfmaplist_out%pindex(:,:,:) = psfmaplist_in%pindex(:,:,:)

    psfmaplist_out%is_valid   = psfmaplist_in%is_valid

  end subroutine srctool_psf_copy_psfmaplist


  ! duplicate a psf_type, allocating space as necessary
  subroutine srctool_psf_copy_psf(psf_in, psf_out, ierr)
    character(len=*), parameter  :: subroutinename='srctool_psf_copy_psf'
    type(psf_type), intent(in) :: psf_in
    type(psf_type), intent(inout) :: psf_out
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: status
    ierr = ERROR_LEVEL_OK
    status = 0

    psf_out%hduinfo = psf_in%hduinfo
    psf_out%class   = psf_in%class

    call srctool_psf_copy_psfmaplist(psf_in%psfmaplist, psf_out%psfmaplist, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr(subroutinename, ierr, "Problem copying psfmaplist")
    endif

    psf_out%is_valid   = psf_in%is_valid

  end subroutine srctool_psf_copy_psf

  ! make a real array using an alligned allocation
  recursive subroutine fftw_alloc_real_array(i1, i2, arr)
    integer, intent(in) :: i1, i2  ! dimensions
    real(kind=real32), pointer, intent(out) :: arr(:,:) ! output
    type(C_PTR) :: c_p

    c_p = fftwf_alloc_real(int(i1*i2, C_SIZE_T))
    call c_f_pointer(c_p, arr, [i1, i2])
  end subroutine fftw_alloc_real_array

  ! free above real array
  recursive subroutine fftw_free_real_array(arr)
    real(kind=real32), pointer, intent(inout) :: arr(:,:)

    call fftwf_free(c_loc(arr))
    nullify(arr)
  end subroutine fftw_free_real_array

  ! make a real array using an aligned allocation
  recursive subroutine fftw_alloc_complex_array(i1, i2, arr)
    integer, intent(in) :: i1, i2  ! dimensions
    complex(kind=4), pointer, intent(out) :: arr(:,:) ! output
    type(C_PTR) :: c_p

    c_p = fftwf_alloc_complex(int(i1*i2, C_SIZE_T))
    call c_f_pointer(c_p, arr, [i1, i2])
  end subroutine fftw_alloc_complex_array

  ! free above complex array
  recursive subroutine fftw_free_complex_array(arr)
    complex(kind=4), pointer, intent(inout) :: arr(:,:)

    call fftwf_free(c_loc(arr))
    nullify(arr)
  end subroutine fftw_free_complex_array



  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  ! get an efficient FFTW size which divides by 2**a 3**b 5**c 7**d
  ! this also scales up the input to be twice as big as the output in each dim
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  recursive function srctool_get_efficient_fft_size(insize)
    integer(kind=int32), intent(in) :: insize
    integer(kind=int32) :: srctool_get_efficient_fft_size

    integer(kind=int32) :: isize, div

    ! isize = 2
    ! do while (isize < insize)
    !    isize = isize * 2
    ! end do
    ! srctool_get_efficient_fft_size = isize
    ! return

    isize = insize
    do
       div = isize
       do while( mod(div,2)==0 ); div=div/2; end do
       do while( mod(div,3)==0 ); div=div/3; end do
       do while( mod(div,5)==0 ); div=div/5; end do
       do while( mod(div,7)==0 ); div=div/7; end do

       ! was a multiple of primes
       if( div == 1 ) then
          exit
       end if
       isize = isize + 1
    end do

    ! double so we have twice the PSF dimension
    srctool_get_efficient_fft_size = isize*2

  end function srctool_get_efficient_fft_size


  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------
  ! calculate efficient plans for the FFT of the PSF
  !-------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------

  subroutine srctool_psf_setup_fftw_plans(instlist, largest_psf_map)
    type(instlist_type), intent(inout) :: instlist
    integer(kind=int32), intent(in) :: largest_psf_map

    integer(kind=int32) :: fft_npix
    real(kind=real32), pointer :: real_arr(:,:)
    complex(kind=4), pointer :: complex_arr(:,:)

    fft_npix = srctool_get_efficient_fft_size(largest_psf_map)
    instlist%fft_npix = fft_npix

    call fftw_alloc_real_array(fft_npix, fft_npix, real_arr)
    call fftw_alloc_complex_array(fft_npix/2+1, fft_npix, complex_arr)

    real_arr = 0
    complex_arr = 0

    instlist%fftw_plan_r2c = fftwf_plan_dft_r2c_2d(fft_npix, fft_npix, &
         real_arr, complex_arr, FFTW_MEASURE)
    instlist%fftw_plan_c2r = fftwf_plan_dft_c2r_2d(fft_npix, fft_npix, &
         complex_arr, real_arr, FFTW_MEASURE)

    call fftw_free_complex_array(complex_arr)
    call fftw_free_real_array(real_arr)
  end subroutine srctool_psf_setup_fftw_plans


end module srctool_psf

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------END of PSF file handling------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
