! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_rmf


  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2015-19-02
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling RMFs
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_ebounds           !
  use srctool_datatypes         !
  use srctool_misc              !

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_RMF_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_rmf
  private :: srctool_rmf_empty_rmf
  private :: srctool_rmf_setup_rmf_for_src
  private :: srctool_rmf_setup_rmf
  private :: srctool_rmf_write_rmf_keys
  private :: srctool_rmf_write_rmf

  public ::  srctool_rmf_empty_rmfset
  public ::  srctool_rmf_write_rmflist
  public ::  srctool_rmf_setup_rmflist
  public ::  srctool_rmf_empty_rmflist
  public ::  srctool_rmf_coadd_rmflist
  public ::  srctool_rmf_read_rmf_from_chdu
  public ::  srctool_rmf_report_rmf_info
  public ::  srctool_rmf_merge_rmfs


contains

  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of RMF file handling------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_rmf_empty_rmf(rmf)
    type(rmf_type), intent(inout)      :: rmf
!    integer(kind=int32) :: i
!    character(len=STR_MAX) :: str_buffer

    rmf%is_valid = .false.
    call fitswrap_init_hduinfo(rmf%hduinfo)
    rmf%patt_class = SRCTOOL_PATT_CLASS_NULL
    rmf%patt_mask  = SRCTOOL_PATT_MASK_NULL

    rmf%num_rows = 0
    if ( allocated(rmf%pelo))  deallocate (rmf%pelo)
    if ( allocated(rmf%pehi))  deallocate (rmf%pehi)
    if ( allocated(rmf%pn_grp)) deallocate (rmf%pn_grp)
!    if ( allocated(rmf%pf_chan)) deallocate (rmf%pf_chan)
!    if ( allocated(rmf%pn_chan)) deallocate (rmf%pn_chan)
    if ( allocated(rmf%pmatrix)) deallocate (rmf%pmatrix)
    if ( allocated(rmf%prowsum)) deallocate (rmf%prowsum)

    call fitswrap_init_key(rmf%key_phafile,  "PHAFILE", "Name of corresponding spectral file")
    call fitswrap_init_key(rmf%key_chantype, "CHANTYPE", "Raw or corrected channels?")
    call fitswrap_init_key(rmf%key_detchans, "DETCHANS", "Total number of raw detector PHA channels")
!    call fitswrap_init_key(rmf%key_tlmin,    "TLMIN")
!    call fitswrap_init_key(rmf%key_lo_thres, "L_THRES") ! TODO switch to "LO_THRES"
    call fitswrap_init_key(rmf%key_lo_thres, "LO_THRES", "Min. prob. threshold in matrix")
    call fitswrap_init_key(rmf%key_numgrp,   "NUMGRP",  "Total number of channel subsets")
    call fitswrap_init_key(rmf%key_numelt,   "NUMELT", "Total number of response elements")

!    call fitswrap_init_key(rmf%key_ninst, "NINST", "Number of instruments that were merged")
!    do i=1,MAX_INST
!     write(str_buffer, '(A,I0)') "INSTRUM", i
!      call fitswrap_init_key(rmf%pkey_instrum(i), trim(str_buffer), "Contributing instrument name")
!    end do

  end subroutine srctool_rmf_empty_rmf
  !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_rmf_empty_rmfset(rmfset)
    type(rmfset_type), intent(inout) :: rmfset
    integer(kind=int32) :: p
    rmfset%is_valid = .false.
    do p=SRCTOOL_PATT_CLASS_M,SRCTOOL_PATT_NCLASS
      call srctool_rmf_empty_rmf(rmfset%rmf(p))
    end do
  end subroutine srctool_rmf_empty_rmfset

  !-------------------------------------------------------------------------------
  subroutine srctool_rmf_empty_rmflist(rmflist)
    type(rmflist_type), intent(inout) :: rmflist
    integer(kind=int32) :: i
    do i=0,MAX_INST
      call srctool_rmf_empty_rmf(rmflist%rmf(i))
    end do
    rmflist%num_rmfs = 0
  end subroutine srctool_rmf_empty_rmflist

  !-------------------------------------------------------------------------------
  !read the rmf from the current hdu of the fits file identified by unit
  subroutine srctool_rmf_read_rmf_from_chdu(unit, rmfset, ierr)
    character(len=*), parameter :: subroutinename='srctool_rmf_read_rmf_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(rmfset_type), intent(inout), target :: rmfset
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
    character(len=STR_MAX) :: str_buffer
    type(fitskey_string_type) :: temp_fitskey


    integer(kind=int32) :: colnum_elo
    integer(kind=int32) :: colnum_ehi
    integer(kind=int32) :: colnum_n_grp
    integer(kind=int32) :: colnum_f_chan
    integer(kind=int32) :: colnum_n_chan
    integer(kind=int32) :: colnum_matrix

    integer(kind=int32) :: nrows_per_chunk
    integer(kind=int32) :: nrows_to_read
    integer(kind=int32) :: row
    integer(kind=int32) :: row_end
    integer(kind=int32) :: anyf
    integer(kind=int32) :: grp, file_f_elem, f_chan, n_chan
    integer(kind=int16), parameter :: nullval_int16  = 0
    integer(kind=int32), parameter :: nullval_int32  = 0
    real(kind=real32)  , parameter :: nullval_real32  = 0.0
    type(rmf_type), pointer :: prmf
    integer(kind=int32) :: patt_class
    integer(kind=int16) :: patt_mask


    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in RMF extension from CHDU")

    !read the EXTNAME into the temp key
    call fitswrap_init_key(temp_fitskey, "EXTNAME" )
    call  fitswrap_read_key(unit, temp_fitskey,   ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading EXTNAME from HDU")
      return
    end if


    !determine the pattern type based on the EXTNAME
    if (      trim(temp_fitskey%value) .eq. 'SPECRESP_1PX' ) then
      patt_class = SRCTOOL_PATT_CLASS_S
      patt_mask  = SRCTOOL_PATT_MASK_S
    else if ( trim(temp_fitskey%value) .eq. 'SPECRESP_2PX' ) then
      patt_class = SRCTOOL_PATT_CLASS_D
      patt_mask  = SRCTOOL_PATT_MASK_D
    else if ( trim(temp_fitskey%value) .eq. 'SPECRESP_3PX' ) then
      patt_class = SRCTOOL_PATT_CLASS_T
      patt_mask  = SRCTOOL_PATT_MASK_T
    else if ( trim(temp_fitskey%value) .eq. 'SPECRESP_4PX' ) then
      patt_class = SRCTOOL_PATT_CLASS_Q
      patt_mask  = SRCTOOL_PATT_MASK_Q
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Unknown pattern type derived from EXTNAME of RMF")
      return
    endif

    prmf => rmfset%rmf(patt_class)
    prmf%patt_class = patt_class
    prmf%patt_mask  = patt_mask

    !get the number of rows in the extension
    call FTGNRW(unit, prmf%num_rows, status)
    if ( fitswrap_check_status (subroutinename, status, ierr)) return

    ! get the remaining hdu info/keys
    call fitswrap_read_hduinfo(unit, prmf%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if

    call fitswrap_read_key(unit, prmf%key_phafile,  ierr)
    call fitswrap_read_key(unit, prmf%key_chantype, ierr)
    call fitswrap_read_key(unit, prmf%key_detchans, ierr)
!    call fitswrap_read_key(unit, prmf%key_tlmin, ierr)
    call fitswrap_read_key(unit, prmf%key_lo_thres, ierr)
    call fitswrap_read_key(unit, prmf%key_numgrp,   ierr)
    call fitswrap_read_key(unit, prmf%key_numelt,   ierr)

    if (prmf%num_rows           .le. 0 .or. &
        prmf%key_detchans%value .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "RMF matrix contains illegal number of elements")
      return
    end if

    !- check the required header keywords are present
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1       'RESPONSE'
    !- HDUCLAS2       'RSP_MATRIX'
    !- HDUCLAS3       'REDIST'
    !- HDUVERS        '1.3.0'

    ! make some space for the matrix (uncompressed version)
    ! format: prmf%pmatrix(E_LO:E_HI,CHAN_LO:CHAN_HI)
    allocate (prmf%pmatrix(1:prmf%num_rows,1:prmf%key_detchans%value), &
        prmf%prowsum(1:prmf%num_rows), &
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return
    prmf%pmatrix(:,:) = 0.0e0


    !check that have at least some entries and then read in the column data
    if ( prmf%num_rows .gt. 0 ) then
      write (str_buffer, '(A,I0,A,A,A,A,A,A,A)') &
          "Reading in: ", prmf%num_rows, " rows of RMF info from extname='", &
          trim(prmf%hduinfo%key_extname%value), "' instrument='", &
          trim(prmf%hduinfo%key_instrume%value), "' pattern='", &
          trim(SRCTOOL_PATT_CLASS_STR(prmf%patt_class)), "'"
      call wrschatty (subroutinename, str_buffer)

      ! prepare and read the fixed length columns
      call fitswrap_prep_col(unit, "ENERG_LO", prmf%num_rows, colnum_elo, prmf%pelo, ierr)
      call fitswrap_prep_col(unit, "ENERG_HI", prmf%num_rows, colnum_ehi, prmf%pehi, ierr)
      call fitswrap_prep_col(unit, "N_GRP",    prmf%num_rows, colnum_n_grp,prmf%pn_grp,ierr)

      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem preparing column(s) for reading from HDU")
        return
      end if

      ! get the chunksize
      call FTGRSZ(unit, nrows_per_chunk, status)
      if (status .ne. 0) then
        call fitswrap_write_fits_error(subroutinename, status)
        ierr = ERROR_LEVEL_SEVERE
        return
      end if

      ! read the ENERG_LO,ENERG_HI,N_GRP columns
      do row = 1, prmf%num_rows, nrows_per_chunk
        nrows_to_read = min(nrows_per_chunk,prmf%num_rows-row+1)
        row_end = row+nrows_to_read-1
        call FTGCVE(unit, colnum_elo,   row,1,nrows_to_read,nullval_real32,prmf%pelo(row:row_end),  anyf,status)
        call FTGCVE(unit, colnum_ehi,   row,1,nrows_to_read,nullval_real32,prmf%pehi(row:row_end),  anyf,status)
        call FTGCVI(unit, colnum_n_grp, row,1,nrows_to_read,nullval_int16, prmf%pn_grp(row:row_end),anyf,status)

        if (status .ne. 0) then
          call fitswrap_write_fits_error(subroutinename, status)
          ierr = ERROR_LEVEL_SEVERE
          return
        end if
      end do

      ! now need to read and process the variable length columns
      ! do this row by row (rather than in chunks)

      ! get the column numbers
      call fitswrap_get_colnum(unit, "F_CHAN", colnum_f_chan, ierr)
      call fitswrap_get_colnum(unit, "N_CHAN", colnum_n_chan, ierr)
      call fitswrap_get_colnum(unit, "MATRIX", colnum_matrix, ierr)
      if ((colnum_f_chan  <= 0) .or. &
          (colnum_n_chan  <= 0) .or. &
          (colnum_matrix <= 0) .or. &
          ierr .ne.  ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem reading RMF columns")
        return
      end if

      if (status /= 0) then
        ierr = ERROR_LEVEL_SEVERE
        call fitswrap_write_fits_error(subroutinename, status)
        return
      end if

!      write (str_buffer, '(A,1x,A8,1x,A4,1x,A8,1x,A8,5(1x,A4))') &
!          "#RMF read", "PATTERN", &
!          "row", "elo(row)", "ehi(row)", "ngrp", &
!          "grp", "fchn", "nchn", "f_elem"
!          call wrschatty (subroutinename, str_buffer)
      do row = 1, prmf%num_rows
        file_f_elem = 1
        do grp = 1, prmf%pn_grp(row)
          ! read the f_chan and n_chan values
          call FTGCVI(unit, colnum_f_chan, row,grp,1,nullval_int16, f_chan, anyf,status)
          call FTGCVI(unit, colnum_n_chan, row,grp,1,nullval_int16, n_chan, anyf,status)

!          write (str_buffer, '(A,1x,A8,1x,I4,1x,F8.3,1x,F8.3,5(1x,I4))') &
!          "RMF read:", trim(SRCTOOL_PATT_CLASS_STR(prmf%patt_class)), &
!          row, prmf%pelo(row), prmf%pehi(row), prmf%pn_grp(row), &
!          grp, f_chan, n_chan, file_f_elem
!          call wrschatty (subroutinename, str_buffer)

          ! now read a stripe of data from the matrix
          call FTGCVE(unit, colnum_matrix,row,file_f_elem,n_chan,nullval_real32,&
              prmf%pmatrix(row,f_chan:(f_chan+n_chan-1)),  anyf,status)
          file_f_elem = file_f_elem + colnum_n_chan
        end do
        prmf%prowsum(row) = sum(prmf%pmatrix(row,:))  ! take the sum across the row
      end do

    end if
    prmf%is_valid = .true.

  end subroutine srctool_rmf_read_rmf_from_chdu
  !-------------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! merge the RMFs across the pattern types according to the bitwise PAT_SEL flag
  subroutine srctool_rmf_merge_rmfs(rmfset, inst, pat_sel, ierr)
    character(len=*), parameter :: subroutinename='srctool_rmf_merge_rmfs'
    type(rmfset_type), intent(inout)       :: rmfset
    type(inst_type),     intent(in),target :: inst
    integer(kind=int16), intent(in)        :: pat_sel
    integer(kind=int32), intent(out)       :: ierr

    integer(kind=int32) :: p, nvalid
!    character(len=STR_MAX) :: str_buffer

    ierr = ERROR_LEVEL_OK
    if ( pat_sel .eq. 0 ) then
      call wrserr (subroutinename, ERROR_LEVEL_WARNING, &
          "Warning, PAT_SEL=0x00 (i.e. no event patterns are considered to be valid)")
    endif
    call wrschatty (subroutinename, &
        "Merging RMFs across patterns (mask=" // &
        trim(i2s(INT(pat_sel,kind=int32))) // ") for inst: " // trim(inst%name))


    ! Empty the merged pattern RMF
    call srctool_rmf_empty_rmf(rmfset%rmf(SRCTOOL_PATT_CLASS_M))
    rmfset%rmf(SRCTOOL_PATT_CLASS_M)%patt_class = SRCTOOL_PATT_CLASS_M

    nvalid = 0
    do p=SRCTOOL_PATT_CLASS_S,SRCTOOL_PATT_CLASS_Q
      if ( rmfset%rmf(p)%is_valid ) then
        if ( IAND(rmfset%rmf(p)%patt_mask, pat_sel) .ne. 0 ) then
          if ( nvalid .eq. 0 ) then
            call srctool_rmf_setup_rmf(rmfset%rmf(SRCTOOL_PATT_CLASS_M), rmfset%rmf(p), inst, ierr )
            if ( ierr .ne.  ERROR_LEVEL_OK ) then
              call wrserr (subroutinename, ierr, "Problem merging RMFs across patterns")
              return
            end if
          endif

          !add the matrix values to the merged rmf matrix
          rmfset%rmf(SRCTOOL_PATT_CLASS_M)%pmatrix(:,:) = rmfset%rmf(SRCTOOL_PATT_CLASS_M)%pmatrix(:,:) + &
              rmfset%rmf(p)%pmatrix(:,:)
          rmfset%rmf(SRCTOOL_PATT_CLASS_M)%prowsum(:)   = rmfset%rmf(SRCTOOL_PATT_CLASS_M)%prowsum(:) + &
              rmfset%rmf(p)%prowsum(:)

          ! update the pattern mask too
          rmfset%rmf(SRCTOOL_PATT_CLASS_M)%patt_mask    = &
              IOR(rmfset%rmf(SRCTOOL_PATT_CLASS_M)%patt_mask, rmfset%rmf(p)%patt_mask)
          nvalid = nvalid + 1
        endif
      endif
    end do
    if ( nvalid .le. 0 ) then
      ierr =  ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "No valid input RMFs were found when merging across patterns")
      if ( pat_sel .eq. 0 ) then
        call wrserr(subroutinename, ierr, &
            "This is likely because PAT_SEL=0x00 (i.e. no event patterns are considered to be valid)")
      endif
      return
    endif

    ! The input CALDB split-by-pattern RMFs have HDUCLAS1="PAT_RESP", HDUCLAS2="MATRIX"
    ! Modify the HDUCLAS1/HDUCLAS2/HDUCLAS3 keywords so that they can be correctly interpreted by XSPEC
    ! https://heasarc.gsfc.nasa.gov/docs/heasarc/ofwg/docs/ofwg_recomm/hduclas.html
    call fitswrap_set_key(rmfset%rmf(SRCTOOL_PATT_CLASS_M)%hduinfo%key_hduclas1, &
        FITSWRAP_SET_KEY_BLANK, "RESPONSE",   "Conforms to 'The OGIP RMF & ARF File Formats'")
    call fitswrap_set_key(rmfset%rmf(SRCTOOL_PATT_CLASS_M)%hduinfo%key_hduclas2, &
        FITSWRAP_SET_KEY_BLANK, "RSP_MATRIX", "detector photon redist. matrix")
    call fitswrap_set_key(rmfset%rmf(SRCTOOL_PATT_CLASS_M)%hduinfo%key_hduclas3, &
        FITSWRAP_SET_KEY_BLANK, "REDIST",     "detector ph. redist. matrix w/o other corrections")


    rmfset%rmf(SRCTOOL_PATT_CLASS_M)%is_valid = .true.
    rmfset%is_valid = .true.
!    call wrschatty (subroutinename, "done")

  end subroutine srctool_rmf_merge_rmfs



  !---------------------------------------------------------------------------------------------
  ! report the rmf info for all instruments
  subroutine srctool_rmf_report_rmf_info(instlist)
    character(len=*), parameter :: subroutinename='srctool_rmf_report_rmf_info'
    type(instlist_type), intent(in), target  :: instlist
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i,p
    type(rmf_type), pointer :: prmf
    write(str_buffer, '(A12,1x,A6,1x,A7,4(1x,A10))') &
        "#RMFInfo", &
        "Inst", "Pattern", "NumChans", "NumEbins", "Emin", "Emax"
    call wrschatty (subroutinename, str_buffer)
    write(str_buffer, '(A12,1x,A6,1x,A7,4(1x,A10))') &
        "#RMFInfo", &
        "-", "-", "-", "-", "(keV)", "(keV)"
    call wrschatty (subroutinename, str_buffer)
    do i=0,MAX_INST
      do p=SRCTOOL_PATT_CLASS_S,SRCTOOL_PATT_CLASS_Q
        prmf => instlist%inst(i)%refrmfset%rmf(p)
        if ( prmf%is_valid ) then
          if ( prmf%num_rows .gt. 0 ) then
            write(str_buffer, '(A12,1x,A6,1x,A7,2(1x,I10),2(1x,F10.3))') &
                "RMFInfo", &
                trim(instlist%inst(i)%name), &
                trim(SRCTOOL_PATT_CLASS_STR(prmf%patt_class)), &
                prmf%key_detchans%value, &
                prmf%num_rows, &
                prmf%pelo(1), &
                prmf%pehi(prmf%num_rows)
            call wrschatty (subroutinename, str_buffer)
          end if
        end if
      end do
    end do
  end subroutine srctool_rmf_report_rmf_info
  !---------------------------------------------------------------------------------------------


  subroutine srctool_rmf_setup_rmf_for_src(rmf, inst, src, ierr)
    character(len=*), parameter :: subroutinename='srctool_rmf_setup_rmf_for_src'
    type(rmf_type),      intent(inout) :: rmf
    type(inst_type),     intent(in),target :: inst
    type(src_type),      intent(in)    :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)                :: status
    character(len=STR_MAX)             :: str_buffer
    character(len=STR_MAX)             :: str_buffer2

    ierr = ERROR_LEVEL_OK
    status = 0

    ! set the output filename + DSTYPE
    call srctool_misc_set_filename(rmf%hduinfo%filename, inst, src, SRCTOOL_FILETYPE_CODE_RMF )
    call fitswrap_set_key(rmf%hduinfo%key_dstype, ESASS_DSTYPE_RMF )

    call fitswrap_set_key(rmf%hduinfo%key_dscode, &
        srctool_misc_dscode(inst%index,ESASS_FILTER_STR(inst%filter_code),0) )

    call fitswrap_set_key(rmf%hduinfo%key_filter, ESASS_FILTER_STR(inst%filter_code))

!    call fitswrap_set_key(rmf%hduinfo%key_object  , "OBJECT",   src%name ,      "[deg] Target name")
!    call fitswrap_set_key(rmf%hduinfo%key_ra_obj  , "RA-OBJ",   src%coord%long, "[deg] RA of target")
!    call fitswrap_set_key(rmf%hduinfo%key_dec_obj , "DEC-OBJ",  src%coord%lat,  "[deg] DEC of target")
!    call fitswrap_set_key(rmf%hduinfo%key_radecsys, "RADECSYS", COORDSYS_STR(src%coord%coordsys_code), FITSWRAP_SET_KEY_BLANK)
!!    call fitswrap_set_key(arf%hduinfo%key_equinox , "EQUINOX",  src%coord%equinox, FITSWRAP_SET_KEY_BLANK)
!    call fitswrap_set_key(rmf%hduinfo%key_equinox , "EQUINOX",  2000.0d0, FITSWRAP_SET_KEY_BLANK)

    call fitswrap_set_key(rmf%hduinfo%key_object  , src%name)
    call fitswrap_set_key(rmf%hduinfo%key_ra_obj  , src%coord%long) ! TODO - do properly
    call fitswrap_set_key(rmf%hduinfo%key_dec_obj , src%coord%lat)  ! TODO
    call fitswrap_set_key(rmf%hduinfo%key_radecsys, COORDSYS_STR(src%coord%coordsys_code)) ! TODO
    call fitswrap_set_key(rmf%hduinfo%key_equinox , 2000.0d0)

    call srctool_misc_set_filename(str_buffer, inst, src, SRCTOOL_FILETYPE_CODE_SRC_SPEC )
    call pathtools_extract_tail(trim(str_buffer), str_buffer2)
!    call fitswrap_set_key(rmf%key_phafile , "PHAFILE",  trim(str_buffer2), FITSWRAP_SET_KEY_BLANK)
    call fitswrap_set_key(rmf%key_phafile , trim(str_buffer2))


  end subroutine srctool_rmf_setup_rmf_for_src
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for an output rmf file
  subroutine srctool_rmf_setup_rmf(rmf, rmf_template, inst, ierr)
    character(len=*), parameter :: subroutinename='srctool_rmf_setup_rmf'
    type(rmf_type),      intent(inout) :: rmf
    type(rmf_type),      intent(in)    :: rmf_template
    type(inst_type),     intent(in),target :: inst
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)                :: status
!    character(len=STR_MAX)             :: str_buffer

    ierr = ERROR_LEVEL_OK
    status = 0

!    rmf%to_consider = .true.

    ! base the rmf structure on the rmf template
    rmf%num_rows     = rmf_template%num_rows
    rmf%key_detchans = rmf_template%key_detchans
    allocate(rmf%pelo(rmf%num_rows),&
        rmf%pehi(rmf%num_rows),&
        rmf%pn_grp(rmf%num_rows),&
        rmf%pmatrix(1:rmf%num_rows,1:rmf%key_detchans%value),&
        rmf%prowsum(rmf%num_rows),&
        STAT = status)

    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! copy some rmf specific keywords over (some will be updated later)
    rmf%key_chantype = rmf_template%key_chantype
    rmf%key_lo_thres = rmf_template%key_lo_thres
    rmf%key_phafile  = rmf_template%key_phafile
    rmf%key_numgrp   = rmf_template%key_numgrp
    rmf%key_numelt   = rmf_template%key_numelt

    ! initialise the values in the arrays
    rmf%pn_grp(:)          = 0
    rmf%pmatrix(:,:)       = 0.0e0
    rmf%prowsum(:)         = 0.0e0

    ! copy the energy bins etc over from the rmf template
    rmf%pelo(:) = rmf_template%pelo(:)
    rmf%pehi(:) = rmf_template%pehi(:)

    ! now copy the hduinfo
    rmf%hduinfo = rmf_template%hduinfo

    call fitswrap_set_key(rmf%hduinfo%key_extname,  'MATRIX')

!    ! now set some appropriate values for an OGIP compliant rmf
!
!    call fitswrap_set_key(rmf%hduinfo%key_creator,  "CREATOR", CREATOR_STRING, FITSWRAP_SET_KEY_BLANK)
    call fitswrap_set_key(rmf%hduinfo%key_date_obs, trim(inst%evtlistlist%hduinfo%key_date_obs%value))
    call fitswrap_set_key(rmf%hduinfo%key_date_end, trim(inst%evtlistlist%hduinfo%key_date_end%value))
    call fitswrap_set_key(rmf%hduinfo%key_tstart, inst%evtlistlist%hduinfo%key_tstart%value)
    call fitswrap_set_key(rmf%hduinfo%key_tstop,  inst%evtlistlist%hduinfo%key_tstop%value)

!    ! prevent the extname key being written twice
!    call fitswrap_init_key(rmf%hduinfo%key_extname, "EXTNAME")

  end subroutine srctool_rmf_setup_rmf
  !-------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------
  ! write rmf fits keys info to the chdu
  subroutine srctool_rmf_write_rmf_keys(unit, rmf, ierr)
    character(len=*), parameter :: subroutinename='srctool_rmf_write_rmf_keys'
    integer(kind=int32), intent(in)  :: unit
    type(rmf_type),      intent(in)  :: rmf
    integer(kind=int32), intent(out)   :: ierr
!    integer(kind=int32)    :: status
!    integer(kind=int32)    :: i
    ierr = ERROR_LEVEL_OK
!    status = 0

    !-----------------------------------
    ! write the standard keywords to the binary table HDU:
    call fitswrap_write_hduinfo_quiet(unit, rmf%hduinfo, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      return
    end if

!    !-----------------------------------
!    ! write the compulsory keywords to the binary table HDU:
!    call fitswrap_write_key (unit, rmf%hduinfo%key_telescop, ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_instrume, ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_filter,   ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_hduclass, ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_hduclas1, ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_hduclas2, ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_hduvers,  ierr)
!    if (status .ne. 0) then
!       ierr = ERROR_LEVEL_SEVERE
!       call fitswrap_write_fits_error(subroutinename, status)
!       return
!     end if
!    ! now the optional keys
!    call fitswrap_write_key (unit, rmf%hduinfo%key_creator,  ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_object,   ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_ra_obj,   ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_dec_obj,  ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_equinox,  ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_radecsys, ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_tstart,   ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_tstop,    ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_date_obs, ierr)
!    call fitswrap_write_key (unit, rmf%hduinfo%key_date_end, ierr)
!
!    if ( trim(rmf%hduinfo%key_instrume%value) .eq. "MERGED" ) then
!      call fitswrap_write_key (unit, rmf%key_ninst,    ierr)
!      do i=1,rmf%key_ninst%value
!        call fitswrap_write_key (unit, rmf%pkey_instrum(i), ierr)
!      end do
!    end if

    ! now the RMF-specific keys
    call fitswrap_write_key (unit, rmf%key_phafile, ierr)
    call fitswrap_write_key (unit, rmf%key_chantype, ierr)
    call fitswrap_write_key (unit, rmf%key_detchans, ierr)
!    call fitswrap_write_key (unit, rmf%key_tlmin, ierr)
    call fitswrap_write_key (unit, rmf%key_lo_thres, ierr)
    call fitswrap_write_key (unit, rmf%key_numgrp, ierr)
    call fitswrap_write_key (unit, rmf%key_numelt, ierr)

    if (ierr .ne. ERROR_LEVEL_OK) then
       return
    end if
    !-----------------------------------
    ! TODO - add any more keywords?
    !-----------------------------------


  end subroutine srctool_rmf_write_rmf_keys
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  !write an rmf to file
  subroutine srctool_rmf_write_rmf(rmf, ebounds, gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_rmf_write_rmf'

    type(rmf_type), intent(inout) :: rmf
    type(ebounds_type), intent(in) :: ebounds
    type(gtilist_type), intent(in) :: gtilist
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32), parameter :: blank_bitpix = 16
    integer(kind=int32), parameter :: blank_naxis = 0
    integer(kind=int32), parameter :: blank_naxes(1) = (/ 0 /)
    integer(kind=int32), parameter :: ncols=6
    character(len=*),    parameter :: ttype(ncols) = &
        (/ "ENERG_LO", "ENERG_HI", "N_GRP   ", "F_CHAN  ", "N_CHAN  ", "MATRIX  " /)
    character(len=*),    parameter :: tunit(ncols) = &
        (/ "keV"    , "keV"     ,"   "    , "   "    , "   "    , "   "     /)
!    character(len=*),    parameter :: tform(ncols) = &
!        (/ "E  "     , "E  "      ,"I  "     , "1PI"     , "1PI"     , "1PE"      /)
    character(len=*),    parameter :: tform(ncols) = &
        (/ "E "     , "E "      ,"I "     , "PI"     , "PI"     , "PE"      /)
    integer(kind=int32), parameter :: COLNUM_ENERG_LO  =  1
    integer(kind=int32), parameter :: COLNUM_ENERG_HI  =  2
    integer(kind=int32), parameter :: COLNUM_N_GRP     =  3
    integer(kind=int32), parameter :: COLNUM_F_CHAN    =  4
    integer(kind=int32), parameter :: COLNUM_N_CHAN    =  5
    integer(kind=int32), parameter :: COLNUM_MATRIX    =  6

!    character(len=*),    parameter :: extname = "SPECRESP"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int64), parameter :: frow=1, felem=1

    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)    :: unit
    integer(kind=int32)    :: status
    integer(kind=int64)    :: row
    integer(kind=int64) :: n_grp
    integer(kind=int16) :: pf_chan(1)
    integer(kind=int16) :: pn_chan(1)
    integer(kind=int32) :: f_chan
    integer(kind=int32) :: l_chan
    integer(kind=int64) :: n_chan

    type(fitskey_int32_type)  :: key_tlmin    !special key - index is col number for the F_CHAN column
    type(fitskey_int32_type)  :: key_tlmax    !special key - index is col number for the F_CHAN column

    ierr = ERROR_LEVEL_OK
    status = 0
    unit = 0
    str_buffer = ""

    if ( .not. rmf%is_valid .or. .not. ebounds%is_valid ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Invalid RMF or EBOUNDS data structures")
      return
    endif


    if ( len_trim(rmf%hduinfo%filename) .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
!      write (str_buffer, '(A,A)'), "Illegal filename for rmf file: ", trim(rmf%hduinfo%filename)
      call wrserr (subroutinename, ierr, "Illegal filename for rmf file: " // trim(rmf%hduinfo%filename))
      return
    end if

!    write (str_buffer,'(A,A)'),  "Writing RMF file, filename= ", trim(rmf%hduinfo%filename)
!    call wrschatty (subroutinename, str_buffer)
    call wrschatty (subroutinename, "Writing RMF file, filename= " // trim(rmf%hduinfo%filename))


    !-----------------------------------
    ! get a fortran unit
    call FTGIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------


    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT(unit,rmf%hduinfo%filename,FITSIO_BLOCKSIZE, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

    ! append a new empty primary image hdu and write date keyword to the HDU
    call FTIIMG(unit,blank_bitpix,blank_naxis,blank_naxes, status)
    call FTPDAT(unit, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if

    !-----------------------------------
    ! append a new binary table HDU
    !FTIBIN(unit,nrows,tfields,ttype,tform,tunit,extname,varidat > status)
    call FTIBIN(unit,rmf%num_rows,ncols,ttype,tform,tunit,trim(rmf%hduinfo%key_extname%value),varidat,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------


    ! update TSTART/TSTOP using GTI
    if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
       call fitswrap_set_key(rmf%hduinfo%key_tstart, gtilist%pstart(1))
       call fitswrap_set_key(rmf%hduinfo%key_tstop, gtilist%pstop(gtilist%num_rows))
    end if

    call srctool_rmf_write_rmf_keys(unit, rmf, ierr)
    !-----------------------------------
    write (str_buffer,'(A,I0)') "TLMIN", COLNUM_F_CHAN
    call fitswrap_init_key (key_tlmin,  trim(str_buffer), "The first channel in the response")
    call fitswrap_set_key(key_tlmin,  INT(1,int32))
    call fitswrap_write_key (unit, key_tlmin,  ierr)
    !-----------------------------------
    write (str_buffer,'(A,I0)') "TLMAX", COLNUM_F_CHAN
    call fitswrap_init_key (key_tlmax,  trim(str_buffer), "The last channel in the response")
    call fitswrap_set_key(key_tlmax,  INT(rmf%num_rows,int32))
    call fitswrap_write_key (unit, key_tlmax,  ierr)
    !-----------------------------------
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing rmf fits keys")
      return
    end if
    !-----------------------------------

    !-----------------------------------

    !-----------------------------------
    ! now write the column data to the file
    ! format: FTPCL[SLBIJKEDCM](unit,colnum,frow,felem,nelements,values, > status)

    ! write ENERG_LO column
    call  FTPCLE(unit,COLNUM_ENERG_LO,frow,felem,rmf%num_rows,rmf%pelo(:),status)
    ! write ENERG_HI column
    call  FTPCLE(unit,COLNUM_ENERG_HI,frow,felem,rmf%num_rows,rmf%pehi(:),status)
    ! write N_GRP column
    call  FTPCLI(unit,COLNUM_N_GRP,   frow,felem,rmf%num_rows,rmf%pn_grp(:),status)

    ! TODO write remaining columns
    ! prepare the data structures:
    ! simplify internal handling by always writing a single contiguous range of channels for each energy row
!    allocate (pmatrix(1:rmf%key_detchans%value), STAT = status)
!    if ( check_allocate_status(subroutinename, status, ierr) ) return
!    pmatrix(:) = 0.0e0
!
    ! write the variable length columns one row at a time
    n_grp = INT(1,int64)   ! always write one group per row
    do row=1,rmf%num_rows

      ! find the first valid entry
      do f_chan=1,rmf%key_detchans%value
        if ( rmf%pmatrix(row,f_chan) .ge. rmf%key_lo_thres%value ) then
          exit
        endif
      end do
      ! find the last valid entry
      do l_chan=rmf%key_detchans%value,1,-1
        if ( rmf%pmatrix(row,l_chan) .ge. rmf%key_lo_thres%value ) then
          exit
        endif
      end do
      n_chan = INT(l_chan-f_chan+1,int64)
      if ( n_chan .lt. 0 ) then
        n_chan = 0
        f_chan = 1
        l_chan = 1
      endif

      pf_chan(1) = INT(f_chan,int16)
      pn_chan(1) = INT(n_chan,int16)
      call  FTPCLI(unit,COLNUM_F_CHAN, row,felem,n_grp,pf_chan(:),status)
      call  FTPCLI(unit,COLNUM_N_CHAN, row,felem,n_grp,pn_chan(:),status)

      call  FTPCLE(unit,COLNUM_MATRIX, row,felem,n_chan,rmf%pmatrix(row,f_chan:l_chan),status)

      if (status .ne. 0) then
         ierr = ERROR_LEVEL_SEVERE
         call fitswrap_write_fits_error(subroutinename, status)
         return
      end if
    end do

    !-----------------------------------
    ! TODO add EBOUNDS extension
    call srctool_ebounds_append_ebounds_to_file(unit, ebounds, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem appending EBOUNDS extension")
      return
    end if
    !-----------------------------------


    !-----------------------------------
    ! now close the fits file
    call FTCLOS(unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

  end subroutine srctool_rmf_write_rmf
  !---------------------------------------------------------------------------------------------




 !-------------------------------------------------------------------------------
  !write a list of rmfs to file(s)
  subroutine srctool_rmf_write_rmflist(rmflist, instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_rmf_write_rmflist'

    type(rmflist_type), intent(inout) :: rmflist
    type(instlist_type),  intent(in), target :: instlist
    integer(kind=int32), intent(out)  :: ierr
    integer(kind=int32) :: i
    type(ebounds_type),   pointer      :: pebounds

    ierr = ERROR_LEVEL_OK

    do i=0,MAX_INST
      if ( (i .eq. 0 ) .or. instlist%inst(i)%to_consider ) then
        if ( rmflist%rmf(i)%is_valid ) then
          call srctool_misc_select_ebounds_pointer(instlist%inst(i),pebounds,ierr)
          if ( ierr .ne. ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Could not find suitable reference EBOUNDS for inst:" &
                // trim(instlist%inst(i)%name) )
            return
          end if

          call srctool_rmf_write_rmf(rmflist%rmf(i), pebounds, instlist%inst(i)%src_gtilist, ierr )
          if ( ierr .ne. ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Problem writing output RMF file")
            return
          end if
          nullify(pebounds)
        end if
      end if
    end do

  end subroutine srctool_rmf_write_rmflist
 !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for an rmflist
  subroutine srctool_rmf_setup_rmflist(rmflist, instlist, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_rmf_setup_rmflist'
    type(rmflist_type),    intent(inout) :: rmflist
    type(instlist_type),  intent(in), target :: instlist
    type(src_type),       intent(in)    :: src
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32)                 :: i
    type(rmfset_type),    pointer :: prefrmfset
!    character(len=STR_MAX)              :: str_buffer
    integer(kind=int32)                 :: nok
    integer(kind=int32)                 :: first_valid_i

    first_valid_i =  0
    ierr = ERROR_LEVEL_OK
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        if ( first_valid_i .eq. 0 ) first_valid_i = i

        call srctool_misc_select_refrmfset_pointer(instlist%inst(i),prefrmfset,ierr)
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Could not find suitable reference RMFs")
          return
        end if

        ! TODO - check this works correctly in all situations
        call srctool_rmf_setup_rmf(rmflist%rmf(i), &
            prefrmfset%rmf(SRCTOOL_PATT_CLASS_M), instlist%inst(i), ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, &
              "Error setting up RMF data structure for inst: " // trim(instlist%inst(i)%name))
          return
        end if
        call srctool_rmf_setup_rmf_for_src(rmflist%rmf(i), &
            instlist%inst(i), src, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, &
              "Error setting up RMF data structure for inst: " // trim(instlist%inst(i)%name))
          return
        end if

!        ! now set up the filename in the "MERGED" across patterns rmf file
!        call srctool_misc_set_filename(rmfsetlist%rmfset(i)%rmf(SRCTOOL_PATT_CLASS_M)%hduinfo%filename, &
!            instlist%inst(i), src, SRCTOOL_FILETYPE_CODE_RMF )

      end if
    end do

    ! now set up the inst(0) version using the first valid inst as a template
    if ( first_valid_i .gt. 0 ) then
      prefrmfset => instlist%inst(first_valid_i)%refrmfset

      ! TODO - check this works correctly in all situations
      call srctool_rmf_setup_rmf(rmflist%rmf(0), &
          prefrmfset%rmf(SRCTOOL_PATT_CLASS_M), instlist%inst(0), ierr )
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, &
            "Error setting up RMF data structure for inst: " // trim(instlist%inst(0)%name))
        return
      end if
      call srctool_rmf_setup_rmf_for_src(rmflist%rmf(0), &
          instlist%inst(0), src, ierr )
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, &
            "Error setting up RMF data structure for inst: " // trim(instlist%inst(0)%name))
        return
      end if
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, &
          "Error setting up RMF data structure for inst: " // trim(instlist%inst(0)%name))
      return
    endif


    ! copy the basic keywords from the first valid instrument for the merged instrument file
    if ( first_valid_i .gt. 0 ) then
        rmflist%rmf(0)%hduinfo  = rmflist%rmf(first_valid_i)%hduinfo
    endif

    ! now set up the hduinfo values in the "MERGED" across instruments rmf file
    ! derive them from the first valid rmf file - with INSTRUME and filename adapted accordingly
    call fitswrap_set_key(rmflist%rmf(0)%hduinfo%key_instrume, INSTRUME_IS_MERGED)
    call fitswrap_set_key(rmflist%rmf(0)%hduinfo%key_dscode, &
        srctool_misc_dscode(instlist%inst(0)%index,ESASS_FILTER_STR(instlist%inst(0)%filter_code),0) )

!    call fitswrap_set_key(rmflist%rmf(0)%hduinfo%key_instrume, &
!          "INSTRUME","MERGED","Instrument name")
    call srctool_misc_set_filename(rmflist%rmf(0)%hduinfo%filename, &
          instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_RMF )
    ! now deal with the NINST and INSTRUMn keywords
    nok = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        nok = nok + 1
        call fitswrap_set_key(rmflist%rmf(0)%hduinfo%pkey_instrum(nok), trim(instlist%inst(i)%name))
      end if
    end do
    call fitswrap_set_key(rmflist%rmf(0)%hduinfo%key_ninst, nok )

  end subroutine srctool_rmf_setup_rmflist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! coadd the rmf data across all valid telescopes
  ! first work out the RMF for each instrument
  ! this is just a simple copy operation currently but in the future could account for the
  ! time varying RMF.
  ! Then form the RMF for the merged-across-instruments spectrum by taking the
  ! weighted sum of the rmfs for the other telescopes
  subroutine srctool_rmf_coadd_rmflist(rmflist, arflist, instlist)
    character(len=*), parameter :: subroutinename='srctool_rmf_coadd_rmflist'
    type(rmflist_type),   intent(inout) :: rmflist
    type(arflist_type),   intent(in) :: arflist
    type(instlist_type),  intent(in) :: instlist
    integer(kind=int32) :: i, c, e
!    integer(kind=int32) :: nok
    real(kind=real64) :: weight, summed_weight, summed_value

    call wrschatty(subroutinename, "Coadding RMF for merged instrument")

    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider .and. &
           instlist%inst(i)%refrmfset%rmf(SRCTOOL_PATT_CLASS_M)%is_valid ) then
        rmflist%rmf(i)%pn_grp(:) = 1
        rmflist%rmf(i)%pmatrix(:,:) = &
          instlist%inst(i)%refrmfset%rmf(SRCTOOL_PATT_CLASS_M)%pmatrix(:,:)
        rmflist%rmf(i)%is_valid = .true.
      endif
    end do

    !-----------------------------------------------
    ! weight the summed-over-instruments rmf by the product of the
    ! effective exposure livetime times the arf value at each energy
    ! do this element by element
    rmflist%rmf(0)%pn_grp(:) = 1
    rmflist%rmf(0)%is_valid = .true.

    rmflist%rmf(0)%pmatrix(:,:) = 0.0e0

    ! TODO proper weighting: rmf_mean=weighted_mean(rmf,weights=arf,texp)
    do e=1,rmflist%rmf(0)%num_rows
      do c=1,rmflist%rmf(0)%key_detchans%value
        summed_value  = 0.0d0
        summed_weight = 0.0d0
        do i=1,MAX_INST
          if ( instlist%inst(i)%to_consider ) then
            weight = arflist%arf(i)%presp(e) * arflist%arf(i)%key_livetime%value
            summed_value  = summed_value +  (weight * instlist%inst(i)%refrmfset%rmf(SRCTOOL_PATT_CLASS_M)%pmatrix(e,c))
            summed_weight = summed_weight + weight
          end if
        end do
        if (summed_weight .gt. 0.0 ) then
          rmflist%rmf(0)%pmatrix(e,c) = REAL(summed_value / summed_weight, real32)
        endif
      end do
    end do






!     do j=0,2,1
!       write (*,'(A,1x,A4,1x,A8,1x,A3,2(1x,A8),1x,A6,1(1x,A8))'), &
!           "#DEBUG_RMF", "INST", "PATTERN", "ROW", "ELO", "EHI", "N_GRP", "RMF_VALS"
!       do i=1,5
!         write (*,'(A,1x,A4,1x,A8,1x,I3,2(1x,F8.3),1x,I6,10(1x,F8.5))'), " DEBUG_RMF", &
!             trim(instlist%inst(j)%name), SRCTOOL_PATT_CLASS_STR(SRCTOOL_PATT_CLASS_M), i, &
!             rmflist%rmf(j)%pelo(i), &
!             rmflist%rmf(j)%pehi(i), &
!             rmflist%rmf(j)%pn_grp(i), &
!             rmflist%rmf(j)%pmatrix(i,1), &
!             rmflist%rmf(j)%pmatrix(i,2), &
!             rmflist%rmf(j)%pmatrix(i,3), &
!             rmflist%rmf(j)%pmatrix(i,4), &
!             rmflist%rmf(j)%pmatrix(i,6), &
!             rmflist%rmf(j)%pmatrix(i,7), &
!             rmflist%rmf(j)%pmatrix(i,8), &
!             rmflist%rmf(j)%pmatrix(i,9), &
!             rmflist%rmf(j)%pmatrix(i,10)
!       end do
!     end do

  end subroutine srctool_rmf_coadd_rmflist
  !-------------------------------------------------------------------------------


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !---------------------------END of RMF file handling------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

end module srctool_rmf
