! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_samplelist

  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_datatypes         ! structs and datatypes
  use srctool_misc
  use srctool_att
  use srctool_generic
  use srctool_generic_regions

  implicit none     ! insist on explicit typing

  public :: srctool_samplelist_empty
  private :: srctool_samplelist_setup_spatial
  private :: srctool_samplelist_setup_energies
  private :: srctool_samplelist_setup_model
  private :: srctool_samplelist_write
  public :: srctool_samplelist_write_samplelists
  public :: srctool_samplelist_accumulate_image
  public :: srctool_samplelist_accumulate_expmap
  public :: srctool_samplelist_compute_arf

contains

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_samplelist_empty(samplelist)
    character(len=*), parameter          :: subroutinename='srctool_samplelist_empty'
    type(samplelist_type), intent(inout) :: samplelist
    if (allocated(samplelist%peval_e))       deallocate(samplelist%peval_e)
    if (allocated(samplelist%ppsf_e))        deallocate(samplelist%ppsf_e)
    if (allocated(samplelist%ppsfsum))       deallocate(samplelist%ppsfsum)
    if (allocated(samplelist%pfracarea))     deallocate(samplelist%pfracarea)

    if (allocated(samplelist%pra))           deallocate(samplelist%pra)
    if (allocated(samplelist%pdec))          deallocate(samplelist%pdec)
    if (allocated(samplelist%pcosdec))       deallocate(samplelist%pcosdec)
    if (allocated(samplelist%psindec))       deallocate(samplelist%psindec)
    if (allocated(samplelist%pdist))         deallocate(samplelist%pdist)
    if (allocated(samplelist%prawx))         deallocate(samplelist%prawx)
    if (allocated(samplelist%prawy))         deallocate(samplelist%prawy)
    if (allocated(samplelist%prawx_int))     deallocate(samplelist%prawx_int)
    if (allocated(samplelist%prawy_int))     deallocate(samplelist%prawy_int)
    if (allocated(samplelist%pon_chip_flag)) deallocate(samplelist%pon_chip_flag)
!    if (allocated(samplelist%pin_region_flag)) deallocate(samplelist%pin_region_flag)
    if (allocated(samplelist%pi))            deallocate(samplelist%pi)
    if (allocated(samplelist%pj))            deallocate(samplelist%pj)
    if (allocated(samplelist%pimage))        deallocate(samplelist%pimage)
    if (allocated(samplelist%pmodel))        deallocate(samplelist%pmodel)
!    if (allocated(samplelist%parea))         deallocate(samplelist%parea)

    if (allocated(samplelist%peffmap))       deallocate(samplelist%peffmap)
    if (allocated(samplelist%pexpmap))       deallocate(samplelist%pexpmap)
    if (allocated(samplelist%pvign))         deallocate(samplelist%pvign)


    if (allocated(samplelist%eval_values))      deallocate(samplelist%eval_values)
    if (allocated(samplelist%eval_values_arf))  deallocate(samplelist%eval_values_arf)
    if (allocated(samplelist%eval_sum_psf))     deallocate(samplelist%eval_sum_psf)
    if (allocated(samplelist%eval_sum_vign))    deallocate(samplelist%eval_sum_vign)
    if (allocated(samplelist%eval_values_lc))   deallocate(samplelist%eval_values_lc)

    samplelist%num_eval_e = 0
    samplelist%num_psf_e = 0

    samplelist%area_per_point = 0.0d0
    samplelist%area_per_point_sqarcsec = 0.0d0
    samplelist%weight_per_point = 1.0d0
    samplelist%num_points = 0
    samplelist%num_points_region = 0
    samplelist%num_points_region_model = 0
    samplelist%nx = 0
    samplelist%ny = 0
    samplelist%dx = 0.0d0
    samplelist%dy = 0.0d0
    samplelist%smallest_sample_radius_deg =  HUGE(samplelist%smallest_sample_radius_deg) ! 1.0d30
    samplelist%largest_sample_radius_deg  = -HUGE(samplelist%largest_sample_radius_deg)  !-1.0d30

    samplelist%t_tot = 0.0d0
    samplelist%t_tot_live = 0.0d0
    samplelist%backscale = 0.0d0

    samplelist%ra0  = 0.0d0
    samplelist%dec0 = 0.0d0
    samplelist%roll = 0.0d0
    samplelist%cosdec0 = 0.0d0
    samplelist%sindec0 = 0.0d0
    samplelist%cosroll = 0.0d0
    samplelist%sinroll = 0.0d0


  end subroutine srctool_samplelist_empty

  !-------------------------------------------------------------------------------
  !
  recursive subroutine srctool_samplelist_accumulate_image(inst, samplelist, ierr)
    character(len=*), parameter          :: subroutinename='srctool_samplelist_accumulate_image'
    type(inst_type), intent(in), target  :: inst
    type(samplelist_type), intent(inout) :: samplelist
    integer(kind=int32), intent(out)     :: ierr
    type(evtlist_type),  pointer         :: pevtlist
    real(kind=real64)  :: oa_x
    real(kind=real64)  :: oa_y
    real(kind=real64)  :: inv_pixscale_x
    real(kind=real64)  :: inv_pixscale_y
    real(kind=real64), parameter  :: rotation_angle = 0.0d0
    real(kind=real64) :: x, y
    integer(kind=int32) :: i, j, ix, iy
    ierr=ERROR_LEVEL_OK
    samplelist%pimage(:) = 0
    inv_pixscale_x = 1.0d0 / samplelist%dx
    inv_pixscale_y = 1.0d0 / samplelist%dy
    oa_x = 0.5d0 + REAL(samplelist%nx,real64)*0.5d0
    oa_y = 0.5d0 + REAL(samplelist%ny,real64)*0.5d0

    do j=1,inst%evtlistlist%num_evtlists
       pevtlist => inst%evtlistlist%evtlist(j)
       do i=1,pevtlist%num_rows
          if ( IAND(pevtlist%pclass_mask(i),samplelist%evt_class) .ne. 0 ) then
             ! project the event onto the tangent plane centred at the src position
             ! x,y are in units of pixels in the samplelist
             call longlat2xy (oa_x, oa_y, inv_pixscale_x, inv_pixscale_y, &
                  samplelist%src_ra, samplelist%src_dec, rotation_angle, &
                  pevtlist%pra(i), pevtlist%pdec(i), x, y )
             ix = INT(x + 0.5d0)
             iy = INT(y + 0.5d0)
             if ((ix .ge. 1) .and. (ix .le. samplelist%nx) .and. &
                  (iy .ge. 1) .and. (iy .le. samplelist%ny)) then
                samplelist%pimage(ix + samplelist%nx * (iy-1)) = &
                     samplelist%pimage(ix + samplelist%nx * (iy-1)) + 1
             end if
          end if
       end do
    end do
  end subroutine srctool_samplelist_accumulate_image
  !-------------------------------------------------------------------------------


  recursive subroutine srctool_samplelist_setup_energies(samplelist, avign, psf, ierr)
    character(len=*), parameter          :: subroutinename='srctool_samplelist_setup_energies'
    type(samplelist_type), intent(inout) :: samplelist
    type(avign_type), intent(in)         :: avign
    type(psf_type), intent(in)           :: psf
    integer(kind=int32), intent(out)     :: ierr

    integer(kind=int32) :: temp_num_eval_e, status
    real(kind=real64), dimension(:), allocatable :: temp_energies   ! temp energy values - keV
    integer(kind=int32) :: e,last_e ! indices
    status = 0
    ierr=ERROR_LEVEL_OK

    !--------------------------------------------------------------------------
    !--------------------------------------------------------------------------
    !--------------------------------------------------------------------------
    !--------------------------------------------------------------------------
    ! set up the list of eval and psf energies
    ! this sets the energy grid over which we will do our fractional response calculations
    ! the aim is to use at least as much energy resolution as is available in the the caldb
    ! i.e. use the union of the PSF energies and the vignetting energies
    ! for analytic vignetting consider the finite defined grid of energies

    ! for background we only need to consider a single energy band
    if (  samplelist%evt_class .eq. EVT_CLASS_FLAG_BACK ) then
       samplelist%num_psf_e = 0
       samplelist%num_eval_e = 1
       allocate(samplelist%peval_e(samplelist%num_eval_e), &
            STAT = status)
       if ( check_allocate_status(subroutinename, status, ierr) ) return
       samplelist%peval_e(1) = 1.0d0 ! nominal value, not used anywere
    else

       samplelist%num_psf_e = 0
       select case ( psf%class )
       case ( SRCTOOL_PSF_CLASS_NONE )
          temp_num_eval_e = size(avign%e_sample(:))
       case ( SRCTOOL_PSF_CLASS_MAP )
          temp_num_eval_e = size(avign%e_sample(:)) + MAX(1,psf%psfmaplist%num_e)
          samplelist%num_psf_e = psf%psfmaplist%num_e

          allocate(samplelist%ppsf_e(samplelist%num_psf_e), &
               STAT = status)
          if ( check_allocate_status(subroutinename, status, ierr) ) return

          samplelist%ppsf_e(:) = psf%psfmaplist%pe(:)
       case default
          ierr = ERROR_LEVEL_ERROR
          call wrserr (subroutinename, ierr, "I cannot handle this class of PSF desc: " // trim(i2s(psf%class)))
          return
       end select

       allocate(temp_energies(temp_num_eval_e), STAT = status)
       if ( check_allocate_status(subroutinename, status, ierr) ) return

       temp_energies(1:size(avign%e_sample(:))) = avign%e_sample(:)
       if ( psf%class .eq. SRCTOOL_PSF_CLASS_MAP .and. samplelist%num_psf_e .gt. 0 ) then
          temp_energies(size(avign%e_sample(:))+1:) = psf%psfmaplist%pe(1:)
       end if

       call insertion_sort(temp_energies(:))
       last_e = 1
       if ( temp_num_eval_e .gt. 1 ) then
          do e=2,temp_num_eval_e
             if ( temp_energies(e) .gt. temp_energies(last_e) + EVAL_ENERGY_MIN_DELTAE ) then
                last_e = last_e + 1
                temp_energies(last_e) = temp_energies(e)
             end if
          end do
       end if

       ! now allocate the proper eval energy array
       samplelist%num_eval_e = last_e
       allocate(samplelist%peval_e(samplelist%num_eval_e), STAT = status)
       if ( check_allocate_status(subroutinename, status, ierr) ) return
       samplelist%peval_e(1:samplelist%num_eval_e) = temp_energies(1:samplelist%num_eval_e)

       ! write (*,'(A,9999F6.3)') "Energies", samplelist%peval_e(1:samplelist%num_eval_e)

       ! we can get rid of the temp array
       if (allocated(temp_energies)) deallocate(temp_energies)
    end if
    !--------------------------------------------------------------------------
    !--------------------------------------------------------------------------
    !--------------------------------------------------------------------------

  end subroutine srctool_samplelist_setup_energies
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_samplelist_setup_spatial(samplelist, src, instpar, xgrid, ierr)
    character(len=*), parameter          :: subroutinename='srctool_samplelist_setup_spatial'
    type(samplelist_type), intent(inout) :: samplelist
    type(src_type), intent(in), target   :: src
    type(instpar_type), intent(in)       :: instpar
    real(kind=real64), intent(in)        :: xgrid
    integer(kind=int32), intent(out)   :: ierr

    type(region_type), pointer :: pregion
    real(kind=real64), parameter  :: roll_eq_zero = 0.0d0
    type(coord_type) :: sample_coord
    real(kind=real64) :: dist_arcsec, rawx, rawy
    integer(kind=int32) :: flag, i, j, k, status

    ierr=ERROR_LEVEL_OK
    status = 0

    if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
       pregion => src%src_region
    else if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_BACK ) then
       pregion => src%back_region
    else
       ierr=ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Unrecognised evt_class: " // trim(i2s(samplelist%evt_class)))
       return
    end if

    samplelist%src_ra    = src%coord%long
    samplelist%src_dec   = src%coord%lat

    samplelist%dx = instpar%pixscale_x * xgrid
    samplelist%dy = instpar%pixscale_y * xgrid

    if ( (samplelist%dx .le. 0.0d0) .or. (samplelist%dy .le. 0.0d0) ) then
       ierr=ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Illegal spatial step size")
       return
    end if

    samplelist%area_per_point = samplelist%dx * samplelist%dy
    samplelist%area_per_point_sqarcsec = samplelist%area_per_point * SQDEG2SQARCSEC
    samplelist%nx = 2 + ceiling(2.0 * pregion%max_radius / samplelist%dx)
    samplelist%ny = 2 + ceiling(2.0 * pregion%max_radius / samplelist%dy)


    ! number of sample points
    samplelist%num_points = samplelist%nx * samplelist%ny

    allocate( &
         samplelist%pra            (samplelist%num_points), &
         samplelist%pdec           (samplelist%num_points), &
         samplelist%pdist          (samplelist%num_points), &
         samplelist%pi             (samplelist%num_points), &
         samplelist%pj             (samplelist%num_points), &
         STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    samplelist%num_points_region = 0
!    samplelist%pin_region_flag(k) = .false.
    sample_coord%coordsys_code = COORDSYS_CODE_ICRS

    samplelist%smallest_sample_radius_deg =  HUGE(samplelist%smallest_sample_radius_deg) ! 1.0d30
    samplelist%largest_sample_radius_deg  = -HUGE(samplelist%largest_sample_radius_deg)  !-1.0d30

    k = 0
    do i=1,samplelist%nx
       do j=1,samplelist%ny
          k = k + 1
          samplelist%pi(k) = i
          samplelist%pj(k) = j
          rawx = instpar%oa_rawx + (REAL(i,real64) - 0.5d0*(REAL(samplelist%nx,real64)+1.0d0))*xgrid
          rawy = instpar%oa_rawy + (REAL(j,real64) - 0.5d0*(REAL(samplelist%ny,real64)+1.0d0))*xgrid
          ! convert back to ra,dec coords - assuming central position is on-axis (zero roll)
          call srctool_att_rawxy2radec_a (instpar, &
               src%coord%long, src%coord%lat, roll_eq_zero, &
               rawx, rawy, &
               sample_coord%long, sample_coord%lat)


          ! and check if inside the region - use a temp coord_type structure
          sample_coord%coordsys_code = COORDSYS_CODE_ICRS

          sample_coord%x    = -huge(sample_coord%x)   ! reminders that x,y,rawx,rawy are not yet defined here
          sample_coord%y    = -huge(sample_coord%y)   ! MAYBE - implement x,y and rawx,rawy filtering for samplelist
          sample_coord%rawx = -huge(sample_coord%rawx)
          sample_coord%rawy = -huge(sample_coord%rawy)

          ! test if sample is in region
          call srctool_regions_test_if_coord_in_region(sample_coord, pregion, flag, ierr)
          if ( ierr .ne. ERROR_LEVEL_OK ) then
             call wrserr (subroutinename, ierr, "Problem selecting sample points")
             return
          end if


          if ( flag .ge. 1 ) then     ! if sample is inside region
!          samplelist%pin_region_flag(k) = .true.
             samplelist%num_points_region = samplelist%num_points_region + 1

             samplelist%pra(k) = sample_coord%long
             samplelist%pdec(k) = sample_coord%lat

             ! measure min and max dist from sample point to src coords
             dist_arcsec = DEG2ARCSEC * calc_ang_sep(samplelist%pra(k), samplelist%pdec(k), &
                  samplelist%src_ra, samplelist%src_dec)

             samplelist%pdist(k) = dist_arcsec * ARCSEC2ARCMIN
             if ( samplelist%pdist(k) * ARCMIN2DEG .gt. samplelist%largest_sample_radius_deg ) then
                samplelist%largest_sample_radius_deg = samplelist%pdist(k)
             end if
             if ( samplelist%pdist(k) * ARCMIN2DEG .lt. samplelist%smallest_sample_radius_deg ) then
                samplelist%smallest_sample_radius_deg = samplelist%pdist(k)
             end if


!          ! now do a debug report of the sample points
!          write (*, '(A,2(1x,I6),(1x,A6),4(1x,F13.7),3(1x,F12.4),(1x,I4),1x)'), &
!              "SAMPLELIST ", src%index, k, EVT_CLASS_FLAG_STR(samplelist%evt_class), &
!              src%coord%long, src%coord%lat, &
!              samplelist%pra(k), samplelist%pdec(k), &
!              rawx, rawy, dist_arcsec, flag
!
          else
             k = k - 1 ! decrement counter if sample point lies outside valid region
          end if

       end do
    end do
    samplelist%num_points = k

    ! now sort the sample points in descending order of their distance from the source centre
    ! only need to retain info about pdist, pra, pdec, pmodel and pi,pj
    ! negate the points to invert natural sorting order

    samplelist%pdist(1:samplelist%num_points) = -1.0d0 * samplelist%pdist(1:samplelist%num_points)
    call insertion_cosort(samplelist%pdist(1:samplelist%num_points), &
         samplelist%pra(1:samplelist%num_points), &
         samplelist%pdec(1:samplelist%num_points), &
         samplelist%pi(1:samplelist%num_points), &
         samplelist%pj(1:samplelist%num_points))
    ! revert the dist entries to their original values
    samplelist%pdist(1:samplelist%num_points) = -1.0d0 * samplelist%pdist(1:samplelist%num_points)

  end subroutine srctool_samplelist_setup_spatial
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  recursive subroutine srctool_samplelist_setup_model(samplelist, src, ierr)
    character(len=*), parameter          :: subroutinename='srctool_samplelist_setup_model'
    type(samplelist_type), intent(inout) :: samplelist
    type(src_type), intent(in), target   :: src
    integer(kind=int32), intent(out)     :: ierr
    integer(kind=int32) :: pt
    real(kind=real64) :: dist_arcsec, nsigmas, x, y,  extmap_pixarea_sqarcsec
    integer(kind=int32) :: mindistpt
    integer(kind=int32) :: ix, iy
    integer(kind=int32) :: status
    logical :: make_model
    real(kind=real64) :: modval

    ierr = ERROR_LEVEL_OK

    ! This routine is only for when processing with a source
    if ( samplelist%evt_class /= EVT_CLASS_FLAG_SRC ) then
      return
    end if

    ! check that relevant arrays have already been allocated
    if ( .not. allocated(samplelist%pra)  .or. &
         .not. allocated(samplelist%pdec) .or. &
         .not. allocated(samplelist%pi)   .or. &
         .not. allocated(samplelist%pj)   .or. &
         .not. allocated(samplelist%pdist) ) then
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Vital columns were not allocated")
       return
    end if

    ! We only want to output a model if PSF convolution is switched
    ! off. In that case we use a separate model for the source list.

    ! Whether we're making a model or just counting samples
    make_model = samplelist%psf_class == SRCTOOL_PSF_CLASS_NONE

    if( make_model ) then
      status = 0
      allocate(samplelist%pmodel(samplelist%num_points), &
          STAT=status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return
      samplelist%pmodel(:) = 0.0d0
    end if

    if( src%extent%code == EXTTYPE_CODE_POINT ) then
       ! special case a point, where there's only a single sample -
       ! make the most central point 1
       mindistpt = minloc(samplelist%pdist(1:samplelist%num_points), dim=1, &
            mask=samplelist%pdist(1:samplelist%num_points)>=0)

       if(mindistpt >= 1) then
         if( make_model ) samplelist%pmodel(mindistpt) = 1.0d0
         samplelist%num_points_region_model = 1
       else
         ierr = ERROR_LEVEL_SEVERE
         call wrserr(subroutinename, ierr, "Could not identify centre of source")
       end if

       return
    end if

    samplelist%num_points_region_model = 0
    do pt = 1, samplelist%num_points
       dist_arcsec = samplelist%pdist(pt) * ARCMIN2ARCSEC
       modval = 0.0d0

       select case (  src%extent%code )

       case ( EXTTYPE_CODE_BETA )
          nsigmas = dist_arcsec / src%extent%pars(1)
          ! only sample points inside the cutoff (expressed in number of sigmas) are considered
          if (nsigmas <= src%extent%pars(2) ) then
             modval = src%extent%norm_const * &
                  (1.0d0 + nsigmas*nsigmas)**(-1.5d0) * &
                  samplelist%area_per_point_sqarcsec
          end if

       case ( EXTTYPE_CODE_GAUSSIAN )
          nsigmas = dist_arcsec / src%extent%pars(1)
          ! only sample points inside the cutoff (expressed in number of sigmas) are considered
          if (nsigmas <= src%extent%pars(2) ) then
             modval = src%extent%norm_const * exp(-0.5d0 * nsigmas*nsigmas) * &
                  samplelist%area_per_point_sqarcsec
          end if

       case ( EXTTYPE_CODE_TOPHAT )
          ! only sample points within the tophat are considered
          if (dist_arcsec <= src%extent%pars(1) ) then
             modval = src%extent%norm_const * samplelist%area_per_point_sqarcsec
          end if

       case ( EXTTYPE_CODE_MAP )
          ! use the MAP values at each sample point
          ! assume that map is already normalised to unity, so need to convert to [arcsec**-2]

          call wcs_world2pix(src%extmap%wcs, samplelist%pra(pt), samplelist%pdec(pt), x, y)
          ix = 1 + int(x, int32)
          iy = 1 + int(y, int32)
          if ( ix >= 1 .and. ix <= src%extmap%naxis1 .and. &
               iy >= 1 .and. iy <= src%extmap%naxis2 ) then
             modval = src%extmap%ppix(ix,iy)
          end if

          ! TODO TODO
          ! TODO TODO
          ! Caveat
          ! We should really take an average of the model over the area belonging to the sample point
          ! this will mitigate effects of steep gradients in sources
          ! we can do a good job on this by using the 2D version of Simpson's rule:
          ! e.g. http://mathfaculty.fullerton.edu/mathews/n2003/SimpsonsRule2DMod.html
          ! TODO TODO
          ! TODO TODO

          ! compute pixel area at sample point - in deg**2
          extmap_pixarea_sqarcsec = wcs_calc_pixarea_atpixel(src%extmap%wcs, x, y) * SQDEG2SQARCSEC
          modval = (modval / extmap_pixarea_sqarcsec) * samplelist%area_per_point_sqarcsec

       case default
          ierr = ERROR_LEVEL_SEVERE
          call wrserr(subroutinename,ierr, "Unknown exttype description: " // &
               TRIM(EXTTYPE_CODE_STR(src%extent%code)))
          return
       end select

       if( make_model ) then
         ! record model if making a model
         samplelist%pmodel(pt) = modval
       end if
       if( modval > 0.0d0 ) then
         ! count samples with non-zero models values
         samplelist%num_points_region_model = samplelist%num_points_region_model + 1
       end if

    end do

  end subroutine srctool_samplelist_setup_model

  !-------------------------------------------------------------------------------
  ! setup a grid of sample points that lie within the src region, only for gti calculations
  ! this only allocates a minimal set of the samplelist columns
  recursive subroutine srctool_setup_samplelist_for_gti(inst, src, samplelist, ierr)
    character(len=*), parameter   :: subroutinename='srctool_setup_samplelist_for_gti'
    type(inst_type), intent(in)          :: inst
    type(samplelist_type), intent(out)   :: samplelist
    type(src_type), intent(in), target   :: src
    integer(kind=int32), intent(out)     :: ierr
    type(instpar_type), pointer :: pinstpar
    integer(kind=int32) :: status
!    integer(kind=int32) :: num_sample_points_in_region

    nullify(pinstpar)
    ierr=ERROR_LEVEL_OK
    status = 0

    call srctool_samplelist_empty(samplelist)
    samplelist%evt_class = EVT_CLASS_FLAG_SRC
    samplelist%num_eval_e = 1
    samplelist%num_psf_e = 0

    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable instpar")
      return
    end if

    call srctool_samplelist_setup_spatial(samplelist, src, pinstpar, inst%xgrid(XGRIDPAR_SRC), ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not set up the samplelist spatial grid")
      return
    end if

!    call wrschatty(subroutinename, trim(inst%name) // &
!        ": Number of sample points in region:        " // trim(i2s(samplelist%num_points)))
    if ( samplelist%num_points .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr(subroutinename, ierr,  trim(inst%name) // &
          ": Number of sample points used to determine source GTIs is zero!")
      return
    end if


    ! now init the real sample list, allocate memory and and copy values into it
    allocate(&
!        samplelist%pmodel(samplelist%num_points), &
        samplelist%pcosdec        (samplelist%num_points), &
        samplelist%psindec        (samplelist%num_points), &
        samplelist%prawx          (samplelist%num_points), &
        samplelist%prawy          (samplelist%num_points), &
        samplelist%prawx_int      (samplelist%num_points), &
        samplelist%prawy_int      (samplelist%num_points), &
        samplelist%pon_chip_flag  (samplelist%num_points), &
        samplelist%peffmap        (samplelist%num_points), &
        STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    ! calc the cos and sin of the Dec
    samplelist%pcosdec(1:samplelist%num_points) = cos(DEG2RAD*samplelist%pdec(1:samplelist%num_points))
    samplelist%psindec(1:samplelist%num_points) = sin(DEG2RAD*samplelist%pdec(1:samplelist%num_points))


  end subroutine srctool_setup_samplelist_for_gti
  !-------------------------------------------------------------------------------





  !-------------------------------------------------------------------------------
  ! setup a grid of sample points that lie within the src or back regions
  !
  recursive subroutine srctool_setup_samplelist(inst, src, samplelist, psf, evt_class, ierr)
    character(len=*), parameter          :: subroutinename='srctool_setup_samplelist'
    type(inst_type), intent(in)          :: inst
    type(samplelist_type), intent(out)   :: samplelist
    type(src_type), intent(in), target   :: src
    type(psf_type), intent(in)           :: psf
    integer(kind=int8), intent(in)       :: evt_class
    integer(kind=int32), intent(out)     :: ierr
    character(len=STR_MAX) :: str_prefix
    integer(kind=int32) :: status
    type(instpar_type), pointer :: pinstpar
    real(kind=real64) :: xgrid
    integer(kind=int32) :: num_sample_points_in_region
    real(kind=real64) :: sum_of_model_in_region
    type(samplelist_type) :: sl

    status = 0
    ierr=ERROR_LEVEL_OK
    nullify(pinstpar)

    ! We will attempt to place points in a box centred on the src position
    ! which bounds the source region - half width=src%src_region%max_radius
    ! Then need to check if each test point actually lies inside the region

    ! we will start with a temporary samplelist and then after doing calculations will
    ! copy it into the output samplelist
    call srctool_samplelist_empty(sl)

    ! need to pay attention to evt class (and psf class later)
    sl%evt_class = evt_class
    sl%psf_class = psf%class
    write(str_prefix, '(A,x,4A,x,A)') trim(inst%name), EVT_CLASS_FLAG_STR(sl%evt_class), ":"

    if ( sl%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
       xgrid = inst%xgrid(XGRIDPAR_SRC)
    else if ( sl%evt_class .eq. EVT_CLASS_FLAG_BACK ) then
       xgrid = inst%xgrid(XGRIDPAR_BACK)
    else
       ierr=ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Unrecognised evt_class: " // trim(i2s(sl%evt_class)))
       return
    end if

    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
       call wrserr (subroutinename, ierr, "Could not find suitable instpar")
       return
    end if

    call srctool_samplelist_setup_energies(sl, inst%avign, psf, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
       call wrserr (subroutinename, ierr, "Could not set up the samplelist energies")
       return
    end if

    call srctool_samplelist_setup_spatial(sl, src, pinstpar, xgrid, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
       call wrserr (subroutinename, ierr, "Could not set up the samplelist spatial grid")
       return
    end if
    num_sample_points_in_region = sl%num_points

    call srctool_samplelist_setup_model(sl, src, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
       call wrserr (subroutinename, ierr, "Could not set up the samplelist model grid")
       return
    end if

    sl%weight_per_point = 1.0d0 / REAL(MAX(1,sl%num_points,real64))

!    call wrschatty(subroutinename, trim(str_prefix) // &
!        " Number of sample points in region:        " // trim(i2s(num_sample_points_in_region)))
!    call wrschatty(subroutinename, trim(str_prefix) // &
!        " Number of valid sample points:            " // trim(i2s(sl%num_points)))

    if ( sl%num_points .le. 0 ) then
       call wrserr(subroutinename, ERROR_LEVEL_WARNING, trim(str_prefix) // &
            " Number of sample points in extraction region is zero!")
    end if

    if ( sl%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
       if ( sl%psf_class .eq. SRCTOOL_PSF_CLASS_NONE ) then
!        call wrschatty(subroutinename,  trim(str_prefix) // &
!            " Sum of model for sample points in region: " // &
!            trim(r2s(sum(sl%pmodel(1:sl%num_points)))))

        ! now renormalise such that the pmodel(:) values have
        ! a mean equal to the fraction of the source model that
        ! lies within the extraction region
          if ( allocated(sl%pmodel)) then
             sum_of_model_in_region = sum(sl%pmodel(1:sl%num_points))
             if ( sum_of_model_in_region .gt. 0.0d0 ) then
                sl%pmodel(1:sl%num_points) = &
                     sl%pmodel(1:sl%num_points) * REAL(sl%num_points,real64) / sum_of_model_in_region

                if ( sum_of_model_in_region .gt. 1.05d0 ) then
                   call wrserr(subroutinename, ERROR_LEVEL_WARNING, trim(str_prefix) // &
                        "Sum of extent model at sample points in source region is significantly greater than unity." )
                   call wrserr(subroutinename, ERROR_LEVEL_WARNING, trim(str_prefix) // &
                        "This is an indication that your spatial sampling is too coarse. " )
                   call wrserr(subroutinename, ERROR_LEVEL_WARNING, trim(str_prefix) // &
                        "Consider using a finer spatial sampling scale (via input parameter xgrid)")
                end if
             else
                call wrserr(subroutinename, ERROR_LEVEL_WARNING,  trim(str_prefix) // &
                     ": Sum of model for sample points in region is less than or equal zero!")
             end if
          end if
       end if
    end if

    ! JSS: Is this copying really necessary?
    samplelist%evt_class  = sl%evt_class
    samplelist%psf_class  = sl%psf_class
    samplelist%num_points  = sl%num_points
    samplelist%num_points_region = sl%num_points_region
    samplelist%num_points_region_model = sl%num_points_region_model
    samplelist%nx  = sl%nx
    samplelist%ny  = sl%ny
    samplelist%area_per_point  = sl%area_per_point
    samplelist%area_per_point_sqarcsec  = sl%area_per_point_sqarcsec
    samplelist%dx  = sl%dx
    samplelist%dy  = sl%dy
    samplelist%weight_per_point  = sl%weight_per_point
    samplelist%num_eval_e  = sl%num_eval_e
    samplelist%num_psf_e  = sl%num_psf_e
    samplelist%src_ra  = sl%src_ra
    samplelist%src_dec  = sl%src_dec
    samplelist%smallest_sample_radius_deg  = sl%smallest_sample_radius_deg
    samplelist%largest_sample_radius_deg  = sl%largest_sample_radius_deg

    ! now init the real sample list, allocate memory and and copy values into it
    allocate(&
         samplelist%peval_e        (sl%num_eval_e), &
         samplelist%pra            (sl%num_points), &
         samplelist%pdec           (sl%num_points), &
         samplelist%pcosdec        (sl%num_points), &
         samplelist%psindec        (sl%num_points), &
         samplelist%pdist          (sl%num_points), &
         samplelist%prawx          (sl%num_points), &
         samplelist%prawy          (sl%num_points), &
         samplelist%prawx_int      (sl%num_points), &
         samplelist%prawy_int      (sl%num_points), &
         samplelist%pon_chip_flag  (sl%num_points), &
         samplelist%pi             (sl%num_points), &
         samplelist%pj             (sl%num_points), &
         samplelist%peffmap        (sl%num_points), &
         samplelist%pexpmap        (sl%num_points,sl%num_eval_e), &
         samplelist%pimage         (sl%nx*sl%ny),   &
         STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    samplelist%peval_e(:) = sl%peval_e(:)
    samplelist%pra(:)     = sl%pra(1:sl%num_points)
    samplelist%pdec(:)    = sl%pdec(1:sl%num_points)
    samplelist%pdec(:)    = sl%pdec(1:sl%num_points)
    samplelist%pi(:)      = sl%pi(1:sl%num_points)
    samplelist%pj(:)      = sl%pj(1:sl%num_points)
    samplelist%pdist(:)   = sl%pdist(1:sl%num_points)

    ! initialise some stuff to zero
    samplelist%prawx_int  = 0
    samplelist%prawy_int  = 0
    samplelist%pon_chip_flag  = .false.
    samplelist%pexpmap(:,:) = 0.0d0
    samplelist%peffmap(:)   = 0.0d0
    samplelist%pimage(:)    = 0

    ! calc the cos and sin of the Declination
    samplelist%pcosdec(:) = cos(DEG2RAD*samplelist%pdec(:))
    samplelist%psindec(:) = sin(DEG2RAD*samplelist%pdec(:))

    ! allocate and copy additional columns if this is a source samplelist
    if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_SRC ) then

      allocate( &
           samplelist%pfracarea      (sl%num_eval_e), &
           samplelist%pmodel         (sl%num_points), &
           samplelist%pvign          (sl%num_points,sl%num_eval_e), &
           samplelist%eval_values    (sl%num_eval_e), &
           samplelist%eval_values_arf(sl%num_eval_e), &
           samplelist%eval_sum_psf   (sl%num_eval_e), &
           samplelist%eval_sum_vign  (sl%num_eval_e), &
           samplelist%eval_values_lc (sl%num_eval_e), &
           STAT = status)
      if ( check_allocate_status(subroutinename, status, ierr) ) return

      samplelist%pfracarea(:) = 0.0d0
      samplelist%pvign(:,:) = 0.0d0

      if ( allocated(sl%pmodel)) then
         samplelist%pmodel(:)  = sl%pmodel(1:sl%num_points)
      else
         samplelist%pmodel(:)  = 0.0d0
      end if
      samplelist%eval_values(:)     = 0.0d0
      samplelist%eval_values_arf(:) = 0.0d0
      samplelist%eval_sum_psf(:)    = 0.0d0
      samplelist%eval_sum_vign(:)   = 0.0d0
      samplelist%eval_values_lc(:)  = 0.0d0

      if ( sl%num_psf_e .gt. 0 ) then
         allocate( &
              samplelist%ppsf_e         (sl%num_psf_e), &
              samplelist%ppsfsum        (sl%num_psf_e), &
              STAT = status)
         if ( check_allocate_status(subroutinename, status, ierr) ) return
         samplelist%ppsf_e(:)  = sl%ppsf_e(:)
         samplelist%ppsfsum(:)  = 0.0d0
      end if

    end if

    call srctool_samplelist_empty(sl)

  end subroutine srctool_setup_samplelist
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  subroutine srctool_samplelist_write(inst, src, samplelist, ierr)
    character(len=*), parameter          :: subroutinename='srctool_samplelist_write'
    type(inst_type), intent(in), target  :: inst
    type(samplelist_type), intent(in)    :: samplelist
    type(src_type), intent(in), target   :: src
    integer(kind=int32), intent(out)     :: ierr

    type(instpar_type), pointer :: pinstpar
    integer(kind=int32)    :: unit
    integer(kind=int32)    :: status
    character(len=LSTR_MAX) :: str_filename
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32)    :: e,i,j,k

    integer(kind=int32), parameter :: real64_bitpix = -64
    integer(kind=int32), parameter :: int32_bitpix  =  32
    integer(kind=int32), parameter :: naxis = 2
    integer(kind=int32), parameter :: fpixel = 1
    integer(kind=int32), parameter :: group = 0
    integer(kind=int32) :: naxes(naxis)

    integer(kind=int32), parameter :: blank_bitpix = 8
    integer(kind=int32), parameter :: blank_naxis = 1
    integer(kind=int32), parameter :: blank_naxes(blank_naxis) = (/ 0 /)

    real(kind=real64), dimension(:), allocatable :: pdata

    integer(kind=int32), parameter :: COLNUM_RA            =  1
    integer(kind=int32), parameter :: COLNUM_DEC           =  2
    integer(kind=int32), parameter :: COLNUM_I             =  3
    integer(kind=int32), parameter :: COLNUM_J             =  4
    integer(kind=int32), parameter :: COLNUM_EXPMAP        =  5
    integer(kind=int32), parameter :: COLNUM_MODEL         =  6
    integer(kind=int32), parameter :: COLNUM_DIST          =  7
    integer(kind=int32), parameter :: ncols=7

    character(len=*),    parameter :: ttype(ncols) = &
         (/"RA      ","DEC     ",&
           "I       ","J       ",&
           "EXPMAP  ","MODEL   ",&
           "DIST    "/)


    character(len=*),    parameter :: tunit(ncols) = &
         (/"deg     ","deg     ",&
           "pix     ","pix     ",&
           "s       ","fraction",&
           "arcmin  "/)

    character(len=*),    parameter :: extname = "SAMPLELIST"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int64), parameter ::  felem = 1
    character(len=6) :: tform(ncols)

    ierr = ERROR_LEVEL_OK
    status = 0
    unit = 0

    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
       call wrserr (subroutinename, ierr, "Could not find suitable instpar")
       return
    end if

    ! get the output filename
    if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
       call srctool_misc_set_filename(str_filename, inst, src, SRCTOOL_FILETYPE_CODE_SRC_SAMPLELIST)
    else if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_BACK ) then
       call srctool_misc_set_filename(str_filename, inst, src, SRCTOOL_FILETYPE_CODE_BACK_SAMPLELIST)
    else
       ierr = ERROR_LEVEL_SEVERE
       call wrserr (subroutinename, ierr, "Unknown event class code")
       return
    end if

    call wrschatty (subroutinename, "Writing samplelist file, filename= " // trim(str_filename))

    if ( len_trim(str_filename) .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Illegal filename for samplelist file")
      return
    end if

    !-----------------------------------
    ! get a fortran unit
    call FTGIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    ! set up the output file

    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT(unit,str_filename,FITSIO_BLOCKSIZE, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

! thsi was causing segmentation faults on OSX
!_!    ! append a new empty primary image hdu and write date keyword to the HDU
!_!    call FTIIMG(unit,blank_bitpix,blank_naxis,blank_naxes, status)
!_!    call FTPDAT(unit, status)
!_!    if (status .ne. 0) then
!_!      ierr = ERROR_LEVEL_SEVERE
!_!      call fitswrap_write_fits_error(subroutinename, status)
!_!      return
!_!    end if

    !--------------------------------------------------------------------------------------------
    ! now append the samplelist table
    tform(COLNUM_RA      ) = "1D"
    tform(COLNUM_DEC     ) = "1D"
    tform(COLNUM_I       ) = "1J"
    tform(COLNUM_J       ) = "1J"
    write(tform(COLNUM_EXPMAP), '(I0,A)') samplelist%num_eval_e, 'D'
    tform(COLNUM_MODEL   ) = "1D"
    tform(COLNUM_DIST    ) = "1D"

    ! first append a new BINTABLE HDU to open file at unit
    call FTIBIN(unit,samplelist%num_points,ncols,ttype(:),tform(:),tunit(:), &
        trim(extname),varidat,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    if ( samplelist%num_points .gt. 0 ) then
      ! TODO - write out data in chunks for efficiency
      call  FTPCLD(unit,COLNUM_RA,      1,felem,samplelist%num_points,samplelist%pra(:),          status)
      call  FTPCLD(unit,COLNUM_DEC,     1,felem,samplelist%num_points,samplelist%pdec(:),         status)
      call  FTPCLJ(unit,COLNUM_I,       1,felem,samplelist%num_points,samplelist%pi(:),           status)
      call  FTPCLJ(unit,COLNUM_J,       1,felem,samplelist%num_points,samplelist%pj(:),           status)
      do k=1,samplelist%num_points
        do e=1,samplelist%num_eval_e
          call  FTPCLD(unit,COLNUM_EXPMAP,  k,e,1,samplelist%pexpmap(k,e),  status)
        end do
      end do
      if ( allocated(samplelist%pmodel)) then
        call  FTPCLD(unit,COLNUM_MODEL,   1,felem,samplelist%num_points,samplelist%pmodel(:),       status)
      end if

      call  FTPCLD(unit,COLNUM_DIST,     1,felem,samplelist%num_points,samplelist%pdist(:),       status)
    end if
    !--------------------------------------------------------------------------------------------



    naxes(1) = samplelist%nx
    naxes(2) = samplelist%ny
    ! make some space for a temporary image
    allocate(pdata(samplelist%nx*samplelist%ny), &
         STAT = status)

    if ( check_allocate_status(subroutinename, status, ierr) ) return

    !--------------------------------------------------------------------------------------------
    ! for each energy level, write out an image of each samplelist%pexpmap(:,e) array
    do e=1,samplelist%num_eval_e
      ! append a new fits image hdu and write date keyword to the HDU
      call FTIIMG(unit,real64_bitpix,naxis,naxes, status)
      call FTPDAT(unit, status)
      if (status .ne. 0) then
         ierr = ERROR_LEVEL_SEVERE
         call fitswrap_write_fits_error(subroutinename, status)
         return
      end if
      write (str_buffer, '(A,I0)') "SAMPLEMAP", e
      call FTPKYS(unit,"EXTNAME",trim(str_buffer),"Name of the extension", status)
      call FTPKYD(unit,"SAMP_ENERGY",samplelist%peval_e(e),FITSIO_DECIMALS,"Energy for this sample point(keV)", status)

      ! write the WCS keywords for DS9
      call FTPKYS(unit,"CTYPE1","RA---TAN",                                   "WCS keyword", status)
      call FTPKYS(unit,"CTYPE2","DEC--TAN",                                   "WCS keyword", status)
      if ( samplelist%num_points .gt. 0 ) then
         call FTPKYD(unit,"CRPIX1",REAL(samplelist%pi(1),real64),FITSIO_DECIMALS,"WCS keyword", status)
         call FTPKYD(unit,"CRPIX2",REAL(samplelist%pj(1),real64),FITSIO_DECIMALS,"WCS keyword", status)
         call FTPKYD(unit,"CRVAL1",samplelist%pra(1),FITSIO_DECIMALS,            "WCS keyword", status)
         call FTPKYD(unit,"CRVAL2",samplelist%pdec(1),FITSIO_DECIMALS,           "WCS keyword", status)
      else
         call FTPKYD(unit,"CRPIX1",0.0d0,FITSIO_DECIMALS,"WCS keyword", status)
         call FTPKYD(unit,"CRPIX2",0.0d0,FITSIO_DECIMALS,"WCS keyword", status)
         call FTPKYD(unit,"CRVAL1",0.0d0,FITSIO_DECIMALS,"WCS keyword", status)
         call FTPKYD(unit,"CRVAL2",0.0d0,FITSIO_DECIMALS,"WCS keyword", status)
      end if
      call FTPKYD(unit,"CDELT1",samplelist%dx,FITSIO_DECIMALS,                "WCS keyword", status)
      call FTPKYD(unit,"CDELT2",samplelist%dy,FITSIO_DECIMALS,                "WCS keyword", status)
      call FTPKYD(unit,"PC1_1",1.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
      call FTPKYD(unit,"PC1_2",0.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
      call FTPKYD(unit,"PC2_1",0.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
      call FTPKYD(unit,"PC2_2",1.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)


      ! set the temp data array entries to the correct values from the samplelist
      pdata(:) = NULLVAL_REAL64
      do k=1,samplelist%num_points
        i = samplelist%pi(k)
        j = samplelist%pj(k)
        if ( (i .ge. 1) .and. (i .le. samplelist%nx) .and. &
             (j .ge. 1) .and. (j .le. samplelist%ny) ) then
          ! TODO replace this line with something that actually records the mean psf over the observation
          ! samplelist%pval(k,e) just contains the last result of get_badpix
          ! samplelist%pexpmap(k,e) just contains the cumulative exposure map
!          pdata(i + ((j-1) * samplelist%nx) ) = samplelist%pval(k,e)
          pdata(i + ((j-1) * samplelist%nx) ) = samplelist%pexpmap(k,e)
!          if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
!            write (*,'(A,x,I3,x,I6,x,I5,x,I5,x,I5,x,F10.6)'), "DEBUG_SAMPLELIST", &
!                e, k, i, j, (i + j * samplelist%nx ), samplelist%pval(k,e)
!          end if
        end if
      end do

      ! write the data to the HDU
!      call FTPPRD(unit,group,fpixel,samplelist%nx*samplelist%ny,pdata(:), status)
      call FTPPND(unit,group,fpixel,samplelist%nx*samplelist%ny,pdata(:),NULLVAL_REAL64, status)
      if (status .ne. 0) then
         ierr = ERROR_LEVEL_SEVERE
         call fitswrap_write_fits_error(subroutinename, status)
         return
      end if

    end do
    !--------------------------------------------------------------------------------------------

    ! now write the event based image
    ! append a new fits image hdu and write date keyword to the HDU
    call FTIIMG(unit,int32_bitpix,naxis,naxes, status)
    call FTPDAT(unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    call FTPKYS(unit,"EXTNAME","IMAGE","Name of the extension", status)

    ! write the WCS keywords for DS9
    call FTPKYS(unit,"CTYPE1","RA---TAN",                                   "WCS keyword", status)
    call FTPKYS(unit,"CTYPE2","DEC--TAN",                                   "WCS keyword", status)
    call FTPKYD(unit,"CRPIX1",REAL(samplelist%pi(1),real64),FITSIO_DECIMALS,"WCS keyword", status)
    call FTPKYD(unit,"CRPIX2",REAL(samplelist%pj(1),real64),FITSIO_DECIMALS,"WCS keyword", status)
    call FTPKYD(unit,"CRVAL1",samplelist%pra(1),            FITSIO_DECIMALS,"WCS keyword", status)
    call FTPKYD(unit,"CRVAL2",samplelist%pdec(1),           FITSIO_DECIMALS,"WCS keyword", status)
    call FTPKYD(unit,"CDELT1",samplelist%dx,                FITSIO_DECIMALS,"WCS keyword", status)
    call FTPKYD(unit,"CDELT2",samplelist%dy,                FITSIO_DECIMALS,"WCS keyword", status)
    call FTPKYD(unit,"PC1_1",1.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
    call FTPKYD(unit,"PC1_2",0.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
    call FTPKYD(unit,"PC2_1",0.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
    call FTPKYD(unit,"PC2_2",1.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)

    call FTPPRJ(unit,group,fpixel,samplelist%nx*samplelist%ny,samplelist%pimage(:), status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !--------------------------------------------------------------------------------------------


    !--------------------------------------------------------------------------------------------
    ! now write an image of the model
    ! append a new fits image hdu and write date keyword to the HDU
    if ( allocated(samplelist%pmodel)) then
       call FTIIMG(unit,real64_bitpix,naxis,naxes, status)
       call FTPDAT(unit, status)
       if (status .ne. 0) then
          ierr = ERROR_LEVEL_SEVERE
          call fitswrap_write_fits_error(subroutinename, status)
          return
       end if
       call FTPKYS(unit,"EXTNAME","MODEL","Name of the extension", status)

       ! write the WCS keywords for DS9
       call FTPKYS(unit,"CTYPE1","RA---TAN",                                   "WCS keyword", status)
       call FTPKYS(unit,"CTYPE2","DEC--TAN",                                   "WCS keyword", status)
       call FTPKYD(unit,"CRPIX1",REAL(samplelist%pi(1),real64),FITSIO_DECIMALS,"WCS keyword", status)
       call FTPKYD(unit,"CRPIX2",REAL(samplelist%pj(1),real64),FITSIO_DECIMALS,"WCS keyword", status)
       call FTPKYD(unit,"CRVAL1",samplelist%pra(1),            FITSIO_DECIMALS,"WCS keyword", status)
       call FTPKYD(unit,"CRVAL2",samplelist%pdec(1),           FITSIO_DECIMALS,"WCS keyword", status)
       call FTPKYD(unit,"CDELT1",samplelist%dx,                FITSIO_DECIMALS,"WCS keyword", status)
       call FTPKYD(unit,"CDELT2",samplelist%dy,                FITSIO_DECIMALS,"WCS keyword", status)
       call FTPKYD(unit,"PC1_1",1.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
       call FTPKYD(unit,"PC1_2",0.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
       call FTPKYD(unit,"PC2_1",0.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)
       call FTPKYD(unit,"PC2_2",1.0d0,FITSIO_DECIMALS,                         "WCS keyword", status)

       ! set the temp data array entries to the correct values from the samplelist
       pdata(:) = NULLVAL_REAL64
       do k=1,samplelist%num_points
          i = samplelist%pi(k)
          j = samplelist%pj(k)
          if ( (i .ge. 1) .and. (i .le. samplelist%nx) .and. &
               (j .ge. 1) .and. (j .le. samplelist%ny) ) then
             pdata(i + ((j-1) * samplelist%nx) ) = samplelist%pmodel(k)
          end if
       end do

       ! write the data to the HDU
       call FTPPND(unit,group,fpixel,samplelist%nx*samplelist%ny,pdata(:),NULLVAL_REAL64, status)
       if (status .ne. 0) then
          ierr = ERROR_LEVEL_SEVERE
          call fitswrap_write_fits_error(subroutinename, status)
          return
       end if
      !--------------------------------------------------------------------------------------------
    end if

    !-----------------------------------
    ! now close the fits file
    call FTCLOS(unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    if (allocated(pdata))  deallocate(pdata)

  end subroutine srctool_samplelist_write
  !-------------------------------------------------------------------------------

  !----------------------------------------------------------------------------------------
  !- write out samplelists as a set of images
  !- Separate files for src and back samplelists. Different energies in separate HDUs.
  subroutine srctool_samplelist_write_samplelists(instlist, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_samplelist_write_samplelists'
    type(instlist_type),  intent(in) :: instlist
    type(src_type), intent(in) :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    ierr = ERROR_LEVEL_OK

    do i=1,MAX_INST
       if ( instlist%inst(i)%to_consider ) then

          call srctool_samplelist_write(instlist%inst(i), src, instlist%inst(i)%samplelist_src, ierr)
          if ( src%do_back ) then
             call srctool_samplelist_write(instlist%inst(i), src, instlist%inst(i)%samplelist_back, ierr)
          end if
          if ( ierr .ne. ERROR_LEVEL_OK ) then
             call wrserr (subroutinename, ierr, "Problem writing samplelist file")
             return
          end if
       end if
    end do
  end subroutine srctool_samplelist_write_samplelists
  !----------------------------------------------------------------------------------------

  recursive subroutine srctool_samplelist_accumulate_expmap(samplelist, tstep)
    character(len=*), parameter :: subroutinename='srctool_samplelist_accumulate_expmap'
    type(samplelist_type), intent(inout), target :: samplelist
    type(tstep_type), intent(in) :: tstep
    integer(kind=int32)    :: e ! indices

    if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
       ! accumulate a vignetted exposure map at each energy
       do e=1,samplelist%num_eval_e
          samplelist%pexpmap(:,e) = samplelist%pexpmap(:,e) + &
               (tstep%dt_live * samplelist%peffmap(:) * samplelist%pvign(:,e))
       end do
    else if ( samplelist%evt_class .eq. EVT_CLASS_FLAG_BACK ) then
       ! accumulate an unvignetted exposure map at a single energy
       samplelist%pexpmap(:,1) = samplelist%pexpmap(:,1) + &
            (tstep%dt_live * samplelist%peffmap(:))
    end if

  end subroutine srctool_samplelist_accumulate_expmap


  recursive subroutine srctool_samplelist_compute_arf(samplelist, refarf, arf, ierr)
    character(len=*), parameter :: subroutinename='srctool_samplelist_compute_arf'
    type(samplelist_type), intent(in), target :: samplelist
    type(arf_type), intent(in)         :: refarf
    type(arf_type), intent(inout)      :: arf
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)    :: status
    type(linear_interp_real64_type) :: interp
    ierr=ERROR_LEVEL_OK
    status = 0

    !----------------------------------------------------------------------------------------
    ! now carry out the necessary interpolations for the output arf file
    if ( arf%to_consider ) then

       ! set the output arf resp values to zero:
       arf%presp(:) = 0.0d0

       call empty_linear_interp (interp)

       ! interpolate from the evaluation energy grid to the ARF energy grid
       call setup_linear_interp (interp, &
            samplelist%peval_e(:), &
            0.5d0*(arf%pelo(:) + arf%pehi(:)), status )
       if ( status .ne. 0 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Error setting up interpolator - arf")
          return
       end if

       ! interpolate from the evaluation points
       call evaluate_linear_interp(interp, samplelist%eval_values_arf(:), status)
       if ( status .ne. 0 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Error evaluating interpolator - arf")
          return
       end if

       ! now multiply the by the fractional area
       arf%presp(:) = interp%y_eval(:) * refarf%presp(:)

       ! total correction
       arf%pcorrcomb(:) = interp%y_eval(:)
       ! correction from PSF
       call evaluate_linear_interp(interp, samplelist%eval_sum_psf(:), status)
       if ( status .ne. 0 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Error evaluating interpolator - arf")
          return
       end if
       arf%pcorrpsf(:) = interp%y_eval(:)
       ! correction from vignetting
       call evaluate_linear_interp(interp, samplelist%eval_sum_vign(:), status)
       if ( status .ne. 0 ) then
          ierr = ERROR_LEVEL_SEVERE
          call wrserr (subroutinename, ierr, "Error evaluating interpolator - arf")
          return
       end if
       arf%pcorrvign(:) = interp%y_eval(:)

       ! divide by the total live exposure time
       if ( samplelist%t_tot_live .gt. 0.0d0 ) then
          arf%presp(:) = arf%presp(:) / samplelist%t_tot_live
          arf%backscale_src = arf%backscale_src / samplelist%t_tot_live
          arf%backscale_back = arf%backscale_back / samplelist%t_tot_live
          arf%pcorrcomb(:) = arf%pcorrcomb(:) / samplelist%t_tot_live
          arf%pcorrpsf(:) = arf%pcorrpsf(:) / samplelist%t_tot_live
          arf%pcorrvign(:) = arf%pcorrvign(:) / samplelist%t_tot_live
       else
          !        ierr = ERROR_LEVEL_WARNING
          call wrserr(subroutinename,ERROR_LEVEL_WARNING,"Total live exposure time was zero")
       end if
       ! set the exposure time values in the arf struct
       call fitswrap_set_key(arf%key_telapse,   samplelist%t_tot     )
       call fitswrap_set_key(arf%key_ontime,    samplelist%t_tot     )
       call fitswrap_set_key(arf%key_livetime,  samplelist%t_tot_live)
       call fitswrap_set_key(arf%key_exposure,  samplelist%t_tot_live)
    end if
    !----------------------------------------------------------------------------------------

  end subroutine srctool_samplelist_compute_arf

end module srctool_samplelist
