! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_spec

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-24
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling spectra
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         ! module with some srctool-specific constants and datatypes
  use srctool_misc              ! module with some misc routines
  use srctool_ebounds           ! module with some ebounds routines
  use srctool_gti               ! module with some gti routines
  use srctool_owner             ! handle owner flags

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_SPEC_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_spec
  private :: srctool_spec_empty_spec
  public  :: srctool_spec_empty_speclist
  private :: srctool_spec_setup_spec
  public  :: srctool_spec_setup_speclist
  private :: srctool_spec_accumulate_spec
  public  :: srctool_spec_accumulate_speclist
  public  :: srctool_spec_coadd_speclist
  private :: srctool_spec_write_spec_keys
  private :: srctool_spec_write_spec
  public  :: srctool_spec_write_speclist

contains
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !------------------------START of Spec file/structure handling-------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------
  subroutine srctool_spec_empty_spec(spec)
    type(spec_type), intent(inout)   :: spec
!    integer(kind=int32) :: i
    call fitswrap_init_hduinfo(spec%hduinfo)

    call fitswrap_init_key(spec%key_telapse,  "TELAPSE" , "[s] Total observation duration"                 )
    call fitswrap_init_key(spec%key_ontime,   "ONTIME"  , "[s] On-target exposure time (pre deadtime corr)")
    call fitswrap_init_key(spec%key_livetime, "LIVETIME", "[s] On-target active-detector exposure time"    )
    call fitswrap_init_key(spec%key_exposure, "EXPOSURE", "[s] Fully corrected effective exposure time"    )

    call fitswrap_init_key(spec%key_backfile, "BACKFILE", "[filename] Name of the corresponding background spec file")
    call fitswrap_init_key(spec%key_corrfile, "CORRFILE", "[filename] Name of the corresponding correction file")
    call fitswrap_init_key(spec%key_corrscal, "CORRSCAL", "[factor] The correction scaling factor")
    call fitswrap_init_key(spec%key_respfile, "RESPFILE", "[filename] Name of the corresponding response matrix")
    call fitswrap_init_key(spec%key_ancrfile, "ANCRFILE", "[filename] Name of the corresponding ancillary response file")
    call fitswrap_init_key(spec%key_poisserr, "POISSERR", "[bool] Are Poisson errors appropriate?")
    call fitswrap_init_key(spec%key_chantype, "CHANTYPE", "Have pulse height corrections been applied?")
    call fitswrap_init_key(spec%key_detchans, "DETCHANS", "Total number of detector channels")
    call fitswrap_init_key(spec%key_areascal, "AREASCAL", "[factor] Factor used when scaling the background")
    call fitswrap_init_key(spec%key_backscal, "BACKSCAL", "[deg^2] Factor used when scaling the background")
    call fitswrap_init_key(spec%key_grouping, "GROUPING", "Are grouping flags defined?")
    call fitswrap_init_key(spec%key_quality , "QUALITY" , "Are quality flags defined?")

!    call fitswrap_init_key(spec%key_ninst, "NINST", "Number of instruments that were merged")
!    do i=1,MAX_INST
!      call fitswrap_init_key(spec%pkey_instrum(i), "INSTRUM" // trim(i2s(i)), "Contributing instrument name")
!    end do

    call fitswrap_init_key(spec%key_geomregarea, "REGAREA", &
        "[deg^2] Geometric area of region")
    call fitswrap_init_key(spec%key_geomregmodarea, "RGMDAREA", &
        "[deg^2] Geom. area of region where src. model>0")

    call fitswrap_init_key(spec%key_ncounts,   "CTS",    "Number of counts - all patterns")
    call fitswrap_init_key(spec%key_ncounts_s, "CNTS_S", "Number of counts - singles")
    call fitswrap_init_key(spec%key_ncounts_d, "CNTS_D", "Number of counts - doubles")
    call fitswrap_init_key(spec%key_ncounts_t, "CNTS_T", "Number of counts - triples")
    call fitswrap_init_key(spec%key_ncounts_q, "CNTS_Q", "Number of counts - quadruples")

    call fitswrap_init_key(spec%key_flagsel , "FLAGSEL", "Flag selection mask which has been applied" )
    call fitswrap_init_key(spec%key_pat_sel , "PAT_SEL", "Pattern mask which has been applied")

    call fitswrap_init_key(spec%key_src_region,  "SRC_REG",  "Source extraction (ds9) region used")
    call fitswrap_init_key(spec%key_back_region, "BACK_REG", "Background extraction (ds9) region used")


    if ( allocated(spec%pchannel ))  deallocate ( spec%pchannel )
    if ( allocated(spec%pcounts ))   deallocate ( spec%pcounts )
!    if ( allocated(spec%pquality ))  deallocate ( spec%pquality )
 !   if ( allocated(spec%pgrouping )) deallocate ( spec%pgrouping )
    if ( allocated(spec%pcounts_s)) deallocate ( spec%pcounts_s )
    if ( allocated(spec%pcounts_d)) deallocate ( spec%pcounts_d )
    if ( allocated(spec%pcounts_t)) deallocate ( spec%pcounts_t )
    if ( allocated(spec%pcounts_q)) deallocate ( spec%pcounts_q )
    spec%num_rows = 0

  end subroutine srctool_spec_empty_spec
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_spec_empty_speclist(speclist)
    type(speclist_type), intent(inout)   :: speclist
    integer(kind=int32) :: i
    do i=0,MAX_INST
!      write (*,'(A,I0)') "Emptying src  spec ", i
      call srctool_spec_empty_spec(speclist%src_spec(i))
!      write (*,'(A,I0)') "Emptying back spec ", i
      call srctool_spec_empty_spec(speclist%back_spec(i))
      speclist%src_spec(i)%evt_class  = EVT_CLASS_FLAG_SRC
      speclist%back_spec(i)%evt_class = EVT_CLASS_FLAG_BACK
    end do
    speclist%num_spec = 0
    speclist%do_back = .true.
  end subroutine srctool_spec_empty_speclist
  !-------------------------------------------------------------------------------



  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for a spec file
  subroutine srctool_spec_setup_spec(spec, inst, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_spec_setup_spec'
    type(spec_type),      intent(inout) :: spec
    type(inst_type),     intent(in),target :: inst
    type(src_type),      intent(in)    :: src
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32)                :: status
    character(len=STR_MAX)             :: str_buffer, str_buffer2
    type(ebounds_type),   pointer      :: pebounds

    ierr = ERROR_LEVEL_OK
    status = 0

    !
    call srctool_misc_select_ebounds_pointer(inst,pebounds,ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Could not find suitable reference EBOUNDS for inst:" // trim(inst%name) )
      return
    end if

    ! base the spec file on the ebounds description
    if ( .not. allocated(spec%pchannel) ) then
      spec%num_rows = pebounds%num_rows
      allocate(spec%pchannel(spec%num_rows),&
          spec%pcounts(spec%num_rows),&
          spec%pcounts_s(spec%num_rows),&
          spec%pcounts_d(spec%num_rows),&
          spec%pcounts_t(spec%num_rows),&
          spec%pcounts_q(spec%num_rows),&
          STAT = status)

      if ( check_allocate_status(subroutinename, status, ierr) ) return

    end if

!    allocate(spec%pquality(spec%num_rows),&
!        spec%pgrouping(spec%num_rows),&
!        STAT = status)

    if ( spec%key_detchans%flag .ne. FITSWRAP_KEY_FLAG_SUPPLIED ) then
!      call fitswrap_set_key(spec%key_detchans, FITSWRAP_SET_KEY_BLANK, spec%num_rows, "Total number of detector channels")
      call fitswrap_set_key(spec%key_detchans, spec%num_rows)
    end if

    ! initialise the values in the arrays
    spec%pchannel(:)  = pebounds%pchannel(:)
    spec%pcounts(:)   = 0
    spec%pcounts_s(:) = 0
    spec%pcounts_d(:) = 0
    spec%pcounts_t(:) = 0
    spec%pcounts_q(:) = 0

!    spec%pquality  = SPEC_QUALITY_GOOD  !
!    spec%pgrouping = SPEC_GROUPING_INDEF

    if ( inst%index .eq. 0 ) then
      !TODO
!      ! we put some placeholder values here for now, but eventually derive these from TM1-7 values
!      call fitswrap_set_key(spec%hduinfo%key_telescop, FITSWRAP_SET_KEY_BLANK,  "eROSITA", "Telescope name")
!      call fitswrap_set_key(spec%hduinfo%key_instrume, FITSWRAP_SET_KEY_BLANK,      "TM0", "Instrument name")
!      call fitswrap_set_key(spec%hduinfo%key_filter,   FITSWRAP_SET_KEY_BLANK,  "UNKNOWN", "CCD filter used")
!      call fitswrap_set_key(spec%hduinfo%key_tstart,   FITSWRAP_SET_KEY_BLANK,      0.0d0, "[s]")
!      call fitswrap_set_key(spec%hduinfo%key_tstop,    FITSWRAP_SET_KEY_BLANK,     1.0d0, "[s]")
!      call fitswrap_set_key(spec%hduinfo%key_date_obs, FITSWRAP_SET_KEY_BLANK, "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")
!      call fitswrap_set_key(spec%hduinfo%key_date_end, FITSWRAP_SET_KEY_BLANK, "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")
!      call fitswrap_set_key(spec%key_detchans, "DETCHANS" ,  spec%num_rows, "")

    else
      ! copy the hduinfo from the first evtlist in the inst structure
      ! but first check that there is at least one valid evtlist
      if ( inst%evtlistlist%num_evtlists .le. 0  ) then
        ierr = ERROR_LEVEL_SEVERE
        call wrserr (subroutinename, ierr, "Instrument structure contains no evtlists")
        return
      end if
      spec%hduinfo = inst%evtlistlist%hduinfo
    end if

    if ( spec%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
      call srctool_misc_set_filename(spec%hduinfo%filename, inst, src, SRCTOOL_FILETYPE_CODE_SRC_SPEC )
      call fitswrap_set_key(spec%hduinfo%key_dstype, ESASS_DSTYPE_SRC_SPEC )
    else if ( spec%evt_class .eq. EVT_CLASS_FLAG_BACK ) then
      call srctool_misc_set_filename(spec%hduinfo%filename, inst, src, SRCTOOL_FILETYPE_CODE_BACK_SPEC )
      call fitswrap_set_key(spec%hduinfo%key_dstype, ESASS_DSTYPE_BACK_SPEC )
    else
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Unknown event class code")
      return
    end if

    call fitswrap_set_key(spec%hduinfo%key_dscode, &
        srctool_misc_dscode(inst%index,ESASS_FILTER_STR(inst%filter_code),0) )


    ! now set some more appropriate values for an OGIP compliant spectrum file
    call fitswrap_set_key(spec%hduinfo%key_extname,  "SPECTRUM")
    call fitswrap_set_key(spec%hduinfo%key_hduclass, "OGIP"    )
    call fitswrap_set_key(spec%hduinfo%key_hduclas1, "SPECTRUM")
    if ( spec%evt_class .eq. EVT_CLASS_FLAG_SRC ) then
      call fitswrap_set_key(spec%hduinfo%key_hduclas2, "TOTAL")
    else if ( spec%evt_class .eq. EVT_CLASS_FLAG_BACK ) then
      call fitswrap_set_key(spec%hduinfo%key_hduclas2, "BACK")
    else
      ! TODO add error
    end if

    call fitswrap_set_key(spec%hduinfo%key_hduclas3, "COUNT")
    call fitswrap_set_key(spec%hduinfo%key_hduvers,  "1.2.1")
!    call fitswrap_set_key(spec%hduinfo%key_creator,  CREATOR_STRING, "Software which created this file")

!   TODO: Check if these keywords should just be populated from the input evtlist
    call fitswrap_set_key(spec%hduinfo%key_object  , trim(src%name) )
    call fitswrap_set_key(spec%hduinfo%key_ra_obj  , src%coord%long)
    call fitswrap_set_key(spec%hduinfo%key_dec_obj , src%coord%lat)
    call fitswrap_set_key(spec%hduinfo%key_radecsys, COORDSYS_STR(src%coord%coordsys_code))
    call fitswrap_set_key(spec%hduinfo%key_equinox , 2000.0d0 )

    call fitswrap_set_key(spec%key_telapse,  0.0d0 )
    call fitswrap_set_key(spec%key_ontime,   0.0d0 )
    call fitswrap_set_key(spec%key_livetime, 0.0d0 )
    call fitswrap_set_key(spec%key_exposure, 0.0d0 )

    call fitswrap_set_key(spec%key_backfile, "NONE" )
    call fitswrap_set_key(spec%key_corrfile, "NONE" )

    !----------------------------------------------------------------------
    ! if SRCTOOL is used to also create RMF and ARF files, then wtrite these filenames into the relevant keyword
    ! if not then just write "NONE"
    if (IAND(src%todo_code,TODO_CODE_RMF)   .ne. TODO_CODE_NULL ) then
      call srctool_misc_set_filename(str_buffer, inst, src, SRCTOOL_FILETYPE_CODE_RMF )
      call pathtools_extract_tail(trim(str_buffer), str_buffer2)
!      if ( str_buffer2(1:1) == "!" ) then  ! remove leading '!' characters
!        str_buffer2(1:1) = "~"
!      endif
    else
      str_buffer2 = "NONE"
    endif
    call fitswrap_set_key(spec%key_respfile, trim(str_buffer2))

    if (IAND(src%todo_code,TODO_CODE_ARF)   .ne. TODO_CODE_NULL ) then
      call srctool_misc_set_filename(str_buffer, inst, src, SRCTOOL_FILETYPE_CODE_ARF )
      call pathtools_extract_tail(trim(str_buffer), str_buffer2)
!      if ( str_buffer2(1:1) == "!" ) then  ! remove leading '!' characters
!        str_buffer2(1:1) = "~"
!      endif
    else
      str_buffer2 = "NONE"
    endif
    call fitswrap_set_key(spec%key_ancrfile, trim(str_buffer2))
    !----------------------------------------------------------------------

    call fitswrap_set_key(spec%key_areascal,       1.0d0)
    call fitswrap_set_key(spec%key_backscal,       1.0d0)
    call fitswrap_set_key(spec%key_corrscal,       1.0d0)
    call fitswrap_set_key(spec%key_poisserr,      .true.)
    call fitswrap_set_key(spec%key_chantype,        "PI")
    call fitswrap_set_key(spec%key_grouping,           0)
    call fitswrap_set_key(spec%key_quality,            0)
    call fitswrap_set_key(spec%key_geomregarea,    0.0d0)
    call fitswrap_set_key(spec%key_geomregmodarea, 0.0d0)

    ! TODO - put the correct values in the following
!    call fitswrap_set_key(spec%hduinfo%key_date_end, FITSWRAP_SET_KEY_BLANK, "YYYY-MM-DDThh:mm:ss.s", "[YYYY-MM-DDThh:mm:ss.s]")
    call fitswrap_set_key(spec%hduinfo%key_date_end, trim(inst%evtlistlist%hduinfo%key_date_end%value) )

    call fitswrap_set_key(spec%key_ncounts,    0)
    call fitswrap_set_key(spec%key_ncounts_s,  0)
    call fitswrap_set_key(spec%key_ncounts_d,  0)
    call fitswrap_set_key(spec%key_ncounts_t,  0)
    call fitswrap_set_key(spec%key_ncounts_q,  0)

    call fitswrap_set_key(spec%key_src_region,  trim(adjustl(src%str_src_region)))
    call fitswrap_set_key(spec%key_back_region, trim(adjustl(src%str_back_region)))

    ! ownership
    call fitswrap_set_key(spec%hduinfo%key_owner, srctool_owner_getkeyword_owner(src%owner))

  end subroutine srctool_spec_setup_spec
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! set up keywords,arrays etc for a speclist
  subroutine srctool_spec_setup_speclist(speclist, instlist, src, ierr )
    character(len=*), parameter :: subroutinename='srctool_spec_setup_speclist'
    type(speclist_type),   intent(inout) :: speclist
    type(instlist_type),  intent(in)    :: instlist
    type(src_type),       intent(in)    :: src
    integer(kind=int32),  intent(out)   :: ierr
    integer(kind=int32)                :: i , nok
!    character(len=STR_MAX) :: str_buffer

    speclist%do_back = src%do_back

    ierr = ERROR_LEVEL_OK
    do i=0,MAX_INST
      if ( (i .eq. 0) .or. instlist%inst(i)%to_consider ) then
        call srctool_spec_setup_spec(speclist%src_spec(i),  instlist%inst(i), src, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error setting up source spec structure")
          return
        end if
        if ( speclist%do_back ) then
          call srctool_spec_setup_spec(speclist%back_spec(i), instlist%inst(i), src, ierr )
          if ( ierr .ne. ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Error setting up background spec structure")
            return
          end if
          call pathtools_extract_tail(trim(speclist%back_spec(i)%hduinfo%filename), &
              speclist%src_spec(i)%key_backfile%value)
        else
          call fitswrap_set_key(speclist%src_spec(i)%key_backfile, "NONE")
        end if
      end if
    end do

    ! copy the basic keywords from the first valid instrument for the merged instrument files
    do i=1,MAX_INST
      if (instlist%inst(i)%to_consider ) then
        speclist%src_spec(0)%hduinfo  = speclist%src_spec(i)%hduinfo
        if ( speclist%do_back ) then
          speclist%back_spec(0)%hduinfo = speclist%src_spec(i)%hduinfo
        endif
        exit
      end if
    end do


    ! now set up the hduinfo values in the "MERGED" spec files
    ! derive them from the first valid spec file - with INSTRUME and filename adapted accordingly
    call fitswrap_set_key(speclist%src_spec(0)%hduinfo%key_instrume, INSTRUME_IS_MERGED)
    call srctool_misc_set_filename(speclist%src_spec(0)%hduinfo%filename, &
          instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_SRC_SPEC )
    if ( speclist%do_back ) then
      call fitswrap_set_key(speclist%back_spec(0)%hduinfo%key_instrume, INSTRUME_IS_MERGED)
      call srctool_misc_set_filename(speclist%back_spec(0)%hduinfo%filename, &
          instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_BACK_SPEC )
    endif

    ! now deal with the NINST and INSTRUMn keywords
    nok = 0
    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        nok = nok + 1
!        write (str_buffer, '(A,I0)') "INSTRUM", nok
        call fitswrap_set_key(speclist%src_spec(0)%hduinfo%pkey_instrum(nok),  trim(instlist%inst(i)%name))
        if ( speclist%do_back ) then
          call fitswrap_set_key(speclist%back_spec(0)%hduinfo%pkey_instrum(nok), trim(instlist%inst(i)%name))
        endif
      end if
    end do
    call fitswrap_set_key(speclist%src_spec(0)%hduinfo%key_ninst, nok)
    if ( speclist%do_back ) then
      call fitswrap_set_key(speclist%back_spec(0)%hduinfo%key_ninst, nok)
    endif
!    ! now set up the hduinfo values in the merged instrument spec files
!    ! derive them from the first valid spec file - with INSTRUME and filename adapted accordingly
!    do i=1,MAX_INST
!      if ( (instlist%inst(i)%to_consider) ) then
!        ! first the src
!        speclist%src_spec(0)%hduinfo = speclist%src_spec(i)%hduinfo
!        call fitswrap_set_key(speclist%src_spec(0)%hduinfo%key_instrume, &
!            "INSTRUME",trim(instlist%inst(0)%name),"Instrument name")
!        call srctool_misc_set_filename(speclist%src_spec(0)%hduinfo%filename, &
!            instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_SRC_SPEC )
!        ! now the background
!        speclist%back_spec(0)%hduinfo = speclist%back_spec(i)%hduinfo
!        call fitswrap_set_key(speclist%back_spec(0)%hduinfo%key_instrume, &
!            "INSTRUME",trim(instlist%inst(0)%name),"Instrument name")
!        call srctool_misc_set_filename(speclist%back_spec(0)%hduinfo%filename, &
!            instlist%inst(0), src, SRCTOOL_FILETYPE_CODE_BACK_SPEC )
!        exit
!      end if
!    end do


  end subroutine srctool_spec_setup_speclist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! gather spectral info from an evtlist and add to a spec file
  ! assume that the spec file has been set up already
  subroutine srctool_spec_accumulate_spec(spec, evtlist, ebounds, ierr )
    character(len=*), parameter :: subroutinename='srctool_spec_accumulate_spec    '
    type(spec_type),      intent(inout) :: spec
    type(evtlist_type),  intent(in)    :: evtlist
    type(ebounds_type),  intent(in)    :: ebounds
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: i
    integer(kind=int32) :: a
    integer(kind=int32) :: status

    integer(kind=int32) :: ncounts, ncounts_s,ncounts_d,ncounts_t,ncounts_q
    integer(kind=int32) :: ncounts_out !counters
    ncounts = 0
    ncounts_s = 0
    ncounts_d = 0
    ncounts_t = 0
    ncounts_q = 0
    ncounts_out = 0
    ierr = ERROR_LEVEL_OK
    status = 0

    do i=1,evtlist%num_rows
      if ( IAND(evtlist%pclass_mask(i),spec%evt_class) .ne. 0 ) then
        !        a = evtlist%ppha(i)
        ! change to using the "ENERGY" column of the reconstructed events
        ! ENERGY is given in eV

!        ! convert to channel using CHANNEL = nint(ENERGY/5eV)
!        ! Check that this is correct handling, maybe should be floor(ENERGY/5eV)??
!        a = NINT(evtlist%penergy(i)*INV_ENERGY_PER_CHANNEL_IN_EV, int32)
!!        a = int(floor(evtlist%penergy(i)*INV_ENERGY_PER_CHANNEL_IN_EV))
!        if ( ebounds%is_valid ) then
!          ! left here so that compiler does not complain
!        endif

        ! Get the channel number using the EBOUNDS description via binary search
        call binary_search(ebounds%pemin(:), REAL(evtlist%penergy(i)*EV2KEV,real32), a, status)
        if ( status .ne. 0 ) then
          a = 0  ! make sure that this count does not get accumulated
          status = 0
        endif

        if ((a .ge. 1) .and.  &
            (a .le. spec%num_rows) ) then
          spec%pcounts(a) = spec%pcounts(a) + 1
          ncounts = ncounts + 1
!          write (*,'(A,A,I0,x,I0,x,F10.3)'), subroutinename, " DEBUG  in: ", i, a, evtlist%penergy(i)
          if (evtlist%ppat_typ(i)      .eq. SRCTOOL_PATT_CLASS_S ) then
            spec%pcounts_s(a) = spec%pcounts_s(a) + 1
            ncounts_s = ncounts_s + 1
          else if (evtlist%ppat_typ(i) .eq. SRCTOOL_PATT_CLASS_D ) then
            spec%pcounts_d(a) = spec%pcounts_d(a) + 1
            ncounts_d = ncounts_d + 1
          else if (evtlist%ppat_typ(i) .eq. SRCTOOL_PATT_CLASS_T ) then
            spec%pcounts_t(a) = spec%pcounts_t(a) + 1
            ncounts_t = ncounts_t + 1
          else if (evtlist%ppat_typ(i) .eq. SRCTOOL_PATT_CLASS_Q ) then
            spec%pcounts_q(a) = spec%pcounts_q(a) + 1
            ncounts_q = ncounts_q + 1
          else
            ! silently ignore
            write (*,'(A,1x,I8,1x,I4)') "odd_pattern_type",i, evtlist%ppat_typ(i)
          endif
        else
          ncounts_out = ncounts_out + 1
          ! silently ignore any out-of-bounds events
!          write (*,'(A,A,I0,x,I0,x,F10.3)'), subroutinename, " DEBUG out: ", i, a, evtlist%penergy(i)
        end if
     end if

    end do

    call fitswrap_set_key(spec%key_ncounts,   ncounts   )
    call fitswrap_set_key(spec%key_ncounts_s, ncounts_s )
    call fitswrap_set_key(spec%key_ncounts_d, ncounts_d )
    call fitswrap_set_key(spec%key_ncounts_t, ncounts_t )
    call fitswrap_set_key(spec%key_ncounts_q, ncounts_q )

    ! copy the flagsel and pat_sel keywords from the evtlist
    spec%key_flagsel = evtlist%key_flagsel
    spec%key_pat_sel = evtlist%key_pat_sel

    ! debug
!     write (str_buffer, '(A,A6,1x,A6,7(1x,I8))') &  !, 2(1x,F8.2))'),
!         "DEBUG SPEC_ACCUM: ", trim(EVT_CLASS_FLAG_STR(spec%evt_class)), &
!         trim(evtlist%hduinfo%key_instrume%value), &
!         evtlist%num_rows,  ncounts, ncounts_out, &
!         ncounts_s, ncounts_d, ncounts_t, ncounts_q
! !        spec%key_ontime%value, spec%key_exposure%value
!     call wrschatty (subroutinename, str_buffer)

  end subroutine srctool_spec_accumulate_spec
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! gather spec info from an evtlist and add to a spec file
  ! assume that the spec file has been setup
  subroutine srctool_spec_accumulate_speclist(speclist, instlist, ierr )
    character(len=*), parameter :: subroutinename='srctool_spec_accumulate_spec'
    type(speclist_type),    intent(inout) :: speclist
    type(instlist_type),   intent(in), target    :: instlist
    type(ebounds_type), pointer :: pebounds
    integer(kind=int32), intent(out)     :: ierr
    integer(kind=int32) :: i,j

    ierr = ERROR_LEVEL_OK
!     write (str_buffer, '(A,A6,1x,A6,7(1x,A8))') &
!         "DEBUG SPEC_ACCUM: ", "CLASS", &
!         "INSTR", "num_rows",  "Ncounts", "Ncts_out", &
!         "Nsingles",  "Ndoubles", "Ntriples", "Nquads"
! !        "ONTIME", "EXPOSURE"
!     call wrschatty (subroutinename, str_buffer)

    !-----------------------------------------------
    do i=1,MAX_INST
      call srctool_misc_select_ebounds_pointer(instlist%inst(i), pebounds, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Error selecting EBOUNDS description")
        return
      end if

      do j=1,instlist%inst(i)%evtlistlist%num_evtlists
        call srctool_spec_accumulate_spec(speclist%src_spec(i), instlist%inst(i)%evtlistlist%evtlist(j), pebounds, ierr )
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error accumulating SRC speclist")
          return
        end if
      end do
    end do
    !-----------------------------------------------

    !-----------------------------------------------
    if ( speclist%do_back ) then
      do i=1,MAX_INST
        call srctool_misc_select_ebounds_pointer(instlist%inst(i), pebounds, ierr)
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Error selecting EBOUNDS description")
          return
        end if
        do j=1,instlist%inst(i)%evtlistlist%num_evtlists
          call srctool_spec_accumulate_spec(speclist%back_spec(i), instlist%inst(i)%evtlistlist%evtlist(j), pebounds, ierr )
          if ( ierr .ne. ERROR_LEVEL_OK ) then
            call wrserr (subroutinename, ierr, "Error accumulating BACK speclist")
            return
          end if
        end do
      end do
    endif
    !-----------------------------------------------

!    ! now add up the counts to make the the merged instrument spec file
!   ! sum the specs over the seven telescope modules
!    call srctool_coadd_speclist(speclist, instlist )

  end subroutine srctool_spec_accumulate_speclist
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! coadd the spec data across all valid telescopes
  subroutine srctool_spec_coadd_speclist(speclist, instlist )
    character(len=*), parameter :: subroutinename='srctool_spec_coadd_speclist'
    type(instlist_type),  intent(in) :: instlist
    type(speclist_type),   intent(inout) :: speclist
    integer(kind=int32) :: i, num_to_consider
    real(kind=real64) :: backscal_src, backscal_back
    real(kind=real64) :: exposure_src, exposure_back
    real(kind=real64) :: regarea_src, regarea_back, regmodarea_src, regmodarea_back
    integer(kind=int32) :: ncounts_src, ncounts_s_src,ncounts_d_src,ncounts_t_src,ncounts_q_src
    integer(kind=int32) :: ncounts_back, ncounts_s_back,ncounts_d_back,ncounts_t_back,ncounts_q_back
    integer(kind=int32) :: first_valid_inst
    !-----------------------------------------------
    num_to_consider = 0
    speclist%src_spec(0)%pcounts(:)  = 0
    if ( speclist%do_back ) speclist%back_spec(0)%pcounts(:) = 0
    backscal_src = 0.0d0
    backscal_back = 0.0d0
    regarea_src = 0.0d0
    regarea_back = 0.0d0
    regmodarea_src = 0.0d0
    regmodarea_back = 0.0d0
    exposure_src = 0.0d0
    exposure_back = 0.0d0
    ncounts_src = 0
    ncounts_s_src = 0
    ncounts_d_src = 0
    ncounts_t_src = 0
    ncounts_q_src = 0
    ncounts_back = 0
    ncounts_s_back = 0
    ncounts_d_back = 0
    ncounts_t_back = 0
    ncounts_q_back = 0

    first_valid_inst = -1

    do i=1,MAX_INST
      if ( instlist%inst(i)%to_consider ) then
        num_to_consider = num_to_consider + 1
        if ( first_valid_inst .lt. 0 ) first_valid_inst = i
        speclist%src_spec(0)%pcounts(:)    = speclist%src_spec(0)%pcounts(:)   + speclist%src_spec(i)%pcounts(:)
        speclist%src_spec(0)%pcounts_s(:)  = speclist%src_spec(0)%pcounts_s(:) + speclist%src_spec(i)%pcounts_s(:)
        speclist%src_spec(0)%pcounts_d(:)  = speclist%src_spec(0)%pcounts_d(:) + speclist%src_spec(i)%pcounts_d(:)
        speclist%src_spec(0)%pcounts_t(:)  = speclist%src_spec(0)%pcounts_t(:) + speclist%src_spec(i)%pcounts_t(:)
        speclist%src_spec(0)%pcounts_q(:)  = speclist%src_spec(0)%pcounts_q(:) + speclist%src_spec(i)%pcounts_q(:)

        exposure_src  = exposure_src  + speclist%src_spec(i)%key_exposure%value
        backscal_src  = backscal_src  + &
            speclist%src_spec(i)%key_exposure%value * speclist%src_spec(i)%key_backscal%value
        regarea_src = regarea_src + &
            speclist%src_spec(i)%key_exposure%value * speclist%src_spec(i)%key_geomregarea%value
        regmodarea_src = regmodarea_src + &
            speclist%src_spec(i)%key_exposure%value * speclist%src_spec(i)%key_geomregmodarea%value

        ncounts_src    = ncounts_src    + speclist%src_spec(i)%key_ncounts%value
        ncounts_s_src  = ncounts_s_src  + speclist%src_spec(i)%key_ncounts_s%value
        ncounts_d_src  = ncounts_d_src  + speclist%src_spec(i)%key_ncounts_d%value
        ncounts_t_src  = ncounts_t_src  + speclist%src_spec(i)%key_ncounts_t%value
        ncounts_q_src  = ncounts_q_src  + speclist%src_spec(i)%key_ncounts_q%value

        if ( speclist%do_back ) then
          speclist%back_spec(0)%pcounts(:)   = speclist%back_spec(0)%pcounts(:)   + speclist%back_spec(i)%pcounts(:)
          speclist%back_spec(0)%pcounts_s(:) = speclist%back_spec(0)%pcounts_s(:) + speclist%back_spec(i)%pcounts_s(:)
          speclist%back_spec(0)%pcounts_d(:) = speclist%back_spec(0)%pcounts_d(:) + speclist%back_spec(i)%pcounts_d(:)
          speclist%back_spec(0)%pcounts_t(:) = speclist%back_spec(0)%pcounts_t(:) + speclist%back_spec(i)%pcounts_t(:)
          speclist%back_spec(0)%pcounts_q(:) = speclist%back_spec(0)%pcounts_q(:) + speclist%back_spec(i)%pcounts_q(:)

          exposure_back  = exposure_back + speclist%back_spec(i)%key_exposure%value
          backscal_back  = backscal_back + &
              speclist%back_spec(i)%key_exposure%value * speclist%back_spec(i)%key_backscal%value
          regarea_back = regarea_back + &
              speclist%back_spec(i)%key_exposure%value * speclist%back_spec(i)%key_geomregarea%value
          regmodarea_back = regmodarea_back + &
              speclist%back_spec(i)%key_exposure%value * speclist%back_spec(i)%key_geomregmodarea%value

          ncounts_back   = ncounts_back   + speclist%back_spec(i)%key_ncounts%value
          ncounts_s_back = ncounts_s_back + speclist%back_spec(i)%key_ncounts_s%value
          ncounts_d_back = ncounts_d_back + speclist%back_spec(i)%key_ncounts_d%value
          ncounts_t_back = ncounts_t_back + speclist%back_spec(i)%key_ncounts_t%value
          ncounts_q_back = ncounts_q_back + speclist%back_spec(i)%key_ncounts_q%value
       end if

      end if
    end do

    !form an exposure weighted average of the backscals

    if ( exposure_src .gt. 0.0d0 .and. num_to_consider .gt. 0 ) then
      backscal_src = backscal_src / exposure_src
      regarea_src = regarea_src / exposure_src
      regmodarea_src = regmodarea_src / exposure_src
      exposure_src = exposure_src / real(num_to_consider)
    end if
!    speclist%src_spec(0)%key_exposure%value = exposure_src
!    speclist%src_spec(0)%key_backscal%value = backscal_src
    call fitswrap_set_key(speclist%src_spec(0)%key_exposure, exposure_src)
    call fitswrap_set_key(speclist%src_spec(0)%key_backscal, backscal_src)
    call fitswrap_set_key(speclist%src_spec(0)%key_geomregarea, regarea_src)
    call fitswrap_set_key(speclist%src_spec(0)%key_geomregmodarea, regmodarea_src)

    call fitswrap_set_key(speclist%src_spec(0)%key_ncounts,   ncounts_src  )
    call fitswrap_set_key(speclist%src_spec(0)%key_ncounts_s, ncounts_s_src)
    call fitswrap_set_key(speclist%src_spec(0)%key_ncounts_d, ncounts_d_src)
    call fitswrap_set_key(speclist%src_spec(0)%key_ncounts_t, ncounts_t_src)
    call fitswrap_set_key(speclist%src_spec(0)%key_ncounts_q, ncounts_q_src)

    if ( speclist%do_back ) then
      if ( exposure_back .gt. 0.0d0 .and. num_to_consider .gt. 0 ) then
        backscal_back = backscal_back / exposure_back
        regarea_back = regarea_back / exposure_back
        regmodarea_back = regmodarea_back / exposure_back
        exposure_back = exposure_back / real(num_to_consider)
      end if
      !    speclist%back_spec(0)%key_exposure%value = exposure_back
      !    speclist%back_spec(0)%key_backscal%value = backscal_back
      call fitswrap_set_key(speclist%back_spec(0)%key_exposure, exposure_back)
      call fitswrap_set_key(speclist%back_spec(0)%key_backscal, backscal_back)
      call fitswrap_set_key(speclist%back_spec(0)%key_geomregarea, regarea_back)
      call fitswrap_set_key(speclist%back_spec(0)%key_geomregmodarea, regmodarea_back)

      call fitswrap_set_key(speclist%back_spec(0)%key_ncounts,   ncounts_back )
      call fitswrap_set_key(speclist%back_spec(0)%key_ncounts_s, ncounts_s_back)
      call fitswrap_set_key(speclist%back_spec(0)%key_ncounts_d, ncounts_d_back)
      call fitswrap_set_key(speclist%back_spec(0)%key_ncounts_t, ncounts_t_back)
      call fitswrap_set_key(speclist%back_spec(0)%key_ncounts_q, ncounts_q_back)
    endif

    ! copy the flagsel and pat_sel keywords from the first valid evtlist
    if ( first_valid_inst .gt. 0 ) then
      speclist%src_spec(0)%key_flagsel  = speclist%src_spec(first_valid_inst)%key_flagsel
      speclist%src_spec(0)%key_pat_sel  = speclist%src_spec(first_valid_inst)%key_pat_sel
      if ( speclist%do_back ) then
        speclist%back_spec(0)%key_flagsel = speclist%back_spec(first_valid_inst)%key_flagsel
        speclist%back_spec(0)%key_pat_sel = speclist%back_spec(first_valid_inst)%key_pat_sel
      end if
    end if


    ! set the FILTER keywords
    call fitswrap_set_key(speclist%src_spec(0)%hduinfo%key_filter, ESASS_FILTER_STR(instlist%inst(0)%filter_code) )
    if ( speclist%do_back ) then
      call fitswrap_set_key(speclist%back_spec(0)%hduinfo%key_filter, ESASS_FILTER_STR(instlist%inst(0)%filter_code) )
    end if



    ! debug
    ! write (str_buffer, '(A,A6,1x,A5,A,1x,A6,1x,I8)') &
    !     "DEBUG SPEC_ACCUM: ", trim(EVT_CLASS_FLAG_STR(EVT_CLASS_FLAG_SRC)), &
    !     INST_STEM, "0", "-", sum(speclist%src_spec(0)%pcounts(:))
    ! call wrschatty (subroutinename, str_buffer)
    ! if ( speclist%do_back ) then
    !   write (str_buffer, '(A,A6,1x,A5,A,1x,A6,1x,I8)') &
    !       "DEBUG SPEC_ACCUM: ", trim(EVT_CLASS_FLAG_STR(EVT_CLASS_FLAG_BACK)), &
    !       INST_STEM, "0", "-", sum(speclist%back_spec(0)%pcounts(:))
    !   call wrschatty (subroutinename, str_buffer)
    ! end if
  end subroutine srctool_spec_coadd_speclist
  !-------------------------------------------------------------------------------





  !-------------------------------------------------------------------------------
  ! write spec fits keys info to the chdu
  subroutine srctool_spec_write_spec_keys(unit, spec, ierr)
    character(len=*), parameter :: subroutinename='srctool_spec_write_spec_keys'
    integer(kind=int32), intent(in)  :: unit
    type(spec_type),      intent(in)  :: spec
    integer(kind=int32), intent(out)   :: ierr
!    integer(kind=int32)    :: i

!    type(fitskey_int32_type)  :: key_test

    ierr = ERROR_LEVEL_OK


    !-----------------------------------
    ! write the standard keywords to the binary table HDU:
    call fitswrap_write_hduinfo_quiet(unit, spec%hduinfo, ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      return
    end if

!    call fitswrap_init_key(key_test, "KEY_TEST")
!    call fitswrap_set_key_null(key_test)
!    call fitswrap_write_key (unit, key_test, ierr)

!    !-----------------------------------
!    ! write the compulsory keywords to the binary table HDU:
!    call fitswrap_write_key (unit, spec%hduinfo%key_telescop, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_instrume, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_filter, ierr)

    call fitswrap_write_key (unit, spec%key_telapse,  ierr)
    call fitswrap_write_key (unit, spec%key_ontime,   ierr)
    call fitswrap_write_key (unit, spec%key_livetime, ierr)
    call fitswrap_write_key (unit, spec%key_exposure, ierr)

    call fitswrap_write_key (unit, spec%key_backfile, ierr)
    call fitswrap_write_key (unit, spec%key_corrfile, ierr)
    call fitswrap_write_key (unit, spec%key_corrscal, ierr)
    call fitswrap_write_key (unit, spec%key_respfile, ierr)
    call fitswrap_write_key (unit, spec%key_ancrfile, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_hduclass, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_hduclas1, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_hduvers,  ierr)
    call fitswrap_write_key (unit, spec%key_poisserr, ierr)
    call fitswrap_write_key (unit, spec%key_chantype, ierr)
    call fitswrap_write_key (unit, spec%key_detchans, ierr)
    call fitswrap_write_key (unit, spec%key_areascal, ierr)
    call fitswrap_write_key (unit, spec%key_backscal, ierr)
    call fitswrap_write_key (unit, spec%key_geomregarea, ierr)
    call fitswrap_write_key (unit, spec%key_geomregmodarea, ierr)

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing fits keywords")
      return
    end if

    ! now the optional keys
!    call fitswrap_write_key (unit, spec%hduinfo%key_hduclas2, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_hduclas3, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_creator,  ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_object,   ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_ra_obj,   ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_dec_obj,  ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_equinox,  ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_radecsys, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_tstart,   ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_tstop,    ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_date_obs, ierr)
!    call fitswrap_write_key (unit, spec%hduinfo%key_date_end, ierr)
    call fitswrap_write_key (unit, spec%key_ncounts, ierr)
    call fitswrap_write_key (unit, spec%key_ncounts_s, ierr)
    call fitswrap_write_key (unit, spec%key_ncounts_d, ierr)
    call fitswrap_write_key (unit, spec%key_ncounts_t, ierr)
    call fitswrap_write_key (unit, spec%key_ncounts_q, ierr)
    call fitswrap_write_key_if_supplied (unit, spec%key_flagsel, ierr)
    call fitswrap_write_key_if_supplied (unit, spec%key_pat_sel, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing fits keywords")
      return
    end if

!    ! XFLTXXXX keywords
!    call fitswrap_write_key (unit, spec%hduinfo%key_xflt(i), ierr)

!    if ( trim(spec%hduinfo%key_instrume%value) .eq. "MERGED" ) then
!      call fitswrap_write_key (unit, spec%key_ninst,    ierr)
!      do i=1,spec%key_ninst%value
!        call fitswrap_write_key (unit, spec%pkey_instrum(i), ierr)
!      end do
!    end if


    call fitswrap_write_key (unit, spec%key_src_region,    ierr)
    call fitswrap_write_key (unit, spec%key_back_region,   ierr)

    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing fits keywords")
      return
    end if

    ! TODO - rest of keywords
    !-----------------------------------

  end subroutine srctool_spec_write_spec_keys

  !-------------------------------------------------------------------------------
  ! write a spec file to disk
  ! we expect that all relevant parameters are already set up
  subroutine srctool_spec_write_spec(spec, gtilist, ierr)
    character(len=*), parameter :: subroutinename='srctool_spec_write_spec'
    type(spec_type),     intent(inout) :: spec
    type(gtilist_type),  intent(in) :: gtilist
    integer(kind=int32), intent(out)   :: ierr

    integer(kind=int32), parameter :: blank_bitpix = 8
    integer(kind=int32), parameter :: blank_naxis = 0
    integer(kind=int32), parameter :: blank_naxes(1) = (/ 0 /)

!    integer(kind=int32), parameter :: ncols=6
!    character(len=*),    parameter :: ttype(ncols) = (/ "CHANNEL  ","COUNTS   ","COUNTS_P1","COUNTS_P2","COUNTS_P3","COUNTS_P4" /)
!    character(len=*),    parameter :: tunit(ncols) = (/ "     "    ,"count"    ,"count"    ,"count"    ,"count"    ,"count"    /)
!    character(len=*),    parameter :: tform(ncols) = (/ "J"        , "J"       , "J"       , "J"       , "J"       , "J"       /)
!    integer(kind=int32), parameter :: COLNUM_CHANNEL  =  1
!    integer(kind=int32), parameter :: COLNUM_COUNTS   =  2
!    integer(kind=int32), parameter :: COLNUM_COUNTS_P1 =  3
!    integer(kind=int32), parameter :: COLNUM_COUNTS_P2 =  4
!    integer(kind=int32), parameter :: COLNUM_COUNTS_P3 =  5
!    integer(kind=int32), parameter :: COLNUM_COUNTS_P4 =  6

    ! modify so that only write out COUNTS_Pn columns that satisfy the pat_sel keyword
    integer(kind=int32), parameter :: max_cols=6
    integer(kind=int32) :: ncols
    character(len=FITSIO_KEY_MAXLEN) :: ttype(max_cols)
    character(len=*),    parameter :: tunit(max_cols) = (/ "     "    ,"count"    ,"count"    ,"count"    ,"count"    ,"count"    /)
    character(len=*),    parameter :: tform(max_cols) = (/ "J"        , "J"       , "J"       , "J"       , "J"       , "J"       /)
    integer(kind=int32), parameter :: COLNUM_CHANNEL  =  1
    integer(kind=int32), parameter :: COLNUM_COUNTS   =  2
    integer(kind=int32) :: COLNUM_COUNTS_P(SRCTOOL_PATT_CLASS_S:SRCTOOL_PATT_CLASS_Q)


!    character(len=*),    parameter :: extname = "SPECTRUM"
    integer(kind=int32), parameter :: varidat=0
    integer(kind=int64), parameter :: frow=1, felem=1

    character(len=LSTR_MAX) :: str_buffer
    integer(kind=int32)    :: unit
    integer(kind=int32)    :: status
    integer(kind=int32)    :: i
    integer(kind=int16)    :: pat_mask_int16

    ierr = ERROR_LEVEL_OK
    status = 0
    unit = 0
    str_buffer = ""

    ! set up the columns
    ttype(COLNUM_CHANNEL) = "CHANNEL"
    ttype(COLNUM_COUNTS)  = "COUNTS"

    ncols = 2
    do i=SRCTOOL_PATT_CLASS_S,SRCTOOL_PATT_CLASS_Q
      pat_mask_int16 = INT(2**(i-1),int16)
      if ( IAND(spec%key_pat_sel%value,pat_mask_int16) .ne. 0 ) then
        ncols = ncols + 1
        COLNUM_COUNTS_P(i) = ncols
        write(str_buffer,'(A,I0)')  "COUNTS_P", i
        ttype(COLNUM_COUNTS_P(i)) = TRIM(str_buffer)
      endif
    end do

    if ( len_trim(spec%hduinfo%filename) .le. 0 ) then
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Illegal filename for spec file")
      return
    end if

    write (str_buffer,'(A,A)')  "Writing spec file, filename= ", trim(spec%hduinfo%filename)
    call wrschatty (subroutinename, str_buffer)


    !-----------------------------------
    ! get a fortran unit
    call FTGIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------


    !-----------------------------------
    ! try to open the fits file for writing
    call FTINIT(unit,spec%hduinfo%filename,FITSIO_BLOCKSIZE, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if
    !-----------------------------------

    ! append a new empty primary image hdu and write date keyword to the HDU
    call FTIIMG(unit,blank_bitpix,blank_naxis,blank_naxes, status)
    call FTPDAT(unit, status)
    if (status .ne. 0) then
      ierr = ERROR_LEVEL_SEVERE
      call fitswrap_write_fits_error(subroutinename, status)
      return
    end if

    !-----------------------------------
    ! append a new binary table HDU
    !FTIBIN(unit,nrows,tfields,ttype,tform,tunit,extname,varidat > status)
    call FTIBIN(unit,spec%num_rows,ncols,ttype,tform,tunit,trim(spec%hduinfo%key_extname%value),varidat,status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    ! update TSTART/TSTOP using GTI
    if(gtilist%is_valid .and. gtilist%num_rows >= 1) then
       call fitswrap_set_key(spec%hduinfo%key_tstart, gtilist%pstart(1))
       call fitswrap_set_key(spec%hduinfo%key_tstop, gtilist%pstop(gtilist%num_rows))
    end if

    !-----------------------------------
    call srctool_spec_write_spec_keys(unit, spec, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem writing spec fits keys")
      return
    end if
    !-----------------------------------

    !-----------------------------------
    ! now write the column data to the file
    ! format: FTPCL[SLBIJKEDCM](unit,colnum,frow,felem,nelements,values, > status)
    ! TODO - write out data in chunks

    ! write CHANNEL column
    call  FTPCLJ(unit,COLNUM_CHANNEL,frow,felem,spec%num_rows,spec%pchannel,status)
    ! write COUNTS column
    call  FTPCLJ(unit,COLNUM_COUNTS ,frow,felem,spec%num_rows,spec%pcounts,status)
!    ! write QUALITY column
!    call  FTPCLI(unit,3,frow,felem,spec%num_rows,spec%pquality,status)
!    ! write GROUPING column
!    call  FTPCLI(unit,4,frow,felem,spec%num_rows,spec%pgrouping,status)

    if ( IAND(spec%key_pat_sel%value,SRCTOOL_PATT_MASK_S) .ne. 0 ) then
      call  FTPCLJ(unit,COLNUM_COUNTS_P(SRCTOOL_PATT_CLASS_S),frow,felem,spec%num_rows,spec%pcounts_s,status)
    endif
    if ( IAND(spec%key_pat_sel%value,SRCTOOL_PATT_MASK_D) .ne. 0 ) then
      call  FTPCLJ(unit,COLNUM_COUNTS_P(SRCTOOL_PATT_CLASS_D),frow,felem,spec%num_rows,spec%pcounts_d,status)
    endif
    if ( IAND(spec%key_pat_sel%value,SRCTOOL_PATT_MASK_T) .ne. 0 ) then
      call  FTPCLJ(unit,COLNUM_COUNTS_P(SRCTOOL_PATT_CLASS_T),frow,felem,spec%num_rows,spec%pcounts_t,status)
    endif
    if ( IAND(spec%key_pat_sel%value,SRCTOOL_PATT_MASK_Q) .ne. 0 ) then
      call  FTPCLJ(unit,COLNUM_COUNTS_P(SRCTOOL_PATT_CLASS_Q),frow,felem,spec%num_rows,spec%pcounts_q,status)
    endif

    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if

    !-----------------------------------




    !-----------------------------------
    ! append the gtilist as a new hdu
    if ( gtilist%is_valid ) then
      call srctool_gti_append_gtilist_hdu(unit, gtilist, ierr)
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem appending GTI extension")
        return
      end if
    end if
    !-----------------------------------


    !-----------------------------------
    ! now close the fits file
    call FTCLOS(unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------

    !-----------------------------------
    ! deallocate the fortran unit
    call FTFIOU( unit, status)
    if (status .ne. 0) then
       ierr = ERROR_LEVEL_SEVERE
       call fitswrap_write_fits_error(subroutinename, status)
       return
    end if
    !-----------------------------------
  end subroutine srctool_spec_write_spec
  !-------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  ! write a set of spec files to disk
  ! we expect that all relevant parameters are already set up
  subroutine srctool_spec_write_speclist(speclist, instlist, ierr)
    character(len=*), parameter :: subroutinename='srctool_spec_write_speclist'
    type(speclist_type),  intent(inout) :: speclist
    type(instlist_type),  intent(in) :: instlist
    integer(kind=int32), intent(out)   :: ierr

!    type(gtilist_type) :: dummy_gtilist

    integer(kind=int32) :: i

!    dummy_gtilist%is_valid = .false.


    ierr = ERROR_LEVEL_OK

    do i=0,MAX_INST
      if ( i .eq. 0 ) then
        call srctool_spec_write_spec(speclist%src_spec(i), instlist%inst(i)%src_gtilist, ierr )
      else if (instlist%inst(i)%to_consider ) then
        call srctool_spec_write_spec(speclist%src_spec(i), instlist%inst(i)%src_gtilist, ierr )
      end if
      if ( ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem writing source spectral file(s)")
        return
      end if
    end do

    if ( speclist%do_back ) then
      do i=0,MAX_INST
        if ( i .eq. 0 ) then
          call srctool_spec_write_spec(speclist%back_spec(i), instlist%inst(i)%src_gtilist, ierr )
        else if (instlist%inst(i)%to_consider ) then
          call srctool_spec_write_spec(speclist%back_spec(i), instlist%inst(i)%src_gtilist, ierr )
        end if
        if ( ierr .ne. ERROR_LEVEL_OK ) then
          call wrserr (subroutinename, ierr, "Problem writing background spectral file(s)")
          return
        end if
      end do
    end if

  end subroutine srctool_spec_write_speclist
  !-------------------------------------------------------------------------------


end module srctool_spec
