! Copyright (C) 2019-2021:
!   Max Planck Institut fuer extraterrestrische Physik, Leibniz Institute for Astrophysics Potsdam (AIP)
!
! This file is part of eRosita's ESASS (https://erosita.mpe.mpg.de/). It is free software; you can
! redistribute it and/or modify it under the terms of the GNU General
! Public License as published by the Free Software Foundation; either
! version 2 of the License, or (at your option) any later version.
!
! This program is distributed in the hope that it will be useful,
! but WITHOUT ANY WARRANTY; without even the implied warranty of
! MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
! General Public License for more details.
!
! You should have received a copy of the GNU General Public License
! along with this library; if not, write to the Free Software
! Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307,
! USA, or see https://www.gnu.org/licenses/.
!

module srctool_vign

  !-----------------------------------------------------------------------------
  ! Author: T. Dwelly (dwelly@mpe.mpg.de)
  !
  ! First version: 2014-07-24
  ! Current version: $Revision$
  ! CVS Date: $Date$
  !-----------------------------------------------------------------------------
  !
  !-----------------------------------------------------------------------------
  ! Summary:
  ! module of routines for handling vignetting
  !
  !-----------------------------------------------------------------------------


  use iso_fortran_env ! a more standard replacement for numkind
  use param          ! standard eROSITA module handling parameter/logging/error interface
  use ero_general    ! standard eROSITA module handling flag definitions etc

  use srctool_generic           ! generic module with some universal constants
  use srctool_generic_fitswrap  ! generic fits file handling routines
  use srctool_datatypes         ! module with some srctool-specific constants and datatypes
  use srctool_misc              ! module with some misc routines

  implicit none     ! insist on explicit typing
  !some private constants
  character(len=*), private, parameter :: modulename='SRCTOOL_VIGN_MOD'

  public

  ! subroutine declarations and interfaces for module: srctool_vign
!unused!   private :: srctool_vign_setup_dummy_vign
  private :: srctool_vign_report_vign_info
!unused!   private :: srctool_vign_calc_vign_values

  public ::  srctool_vign_calc_vign_values_old
  public ::  srctool_vign_empty_vign
  public ::  srctool_vign_read_vign_from_chdu


contains


  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------Start of VIGNETTING file handling---------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  subroutine srctool_vign_empty_vign(vign)
    type(vign_type), intent(inout)      :: vign
    if ( allocated(vign%pelo))   deallocate (vign%pelo)
    if ( allocated(vign%pehi))   deallocate (vign%pehi)
    if ( allocated(vign%ptheta)) deallocate (vign%ptheta)
    if ( allocated(vign%pphi))   deallocate (vign%pphi)
    if ( allocated(vign%pvign))  deallocate (vign%pvign)
    if ( allocated(vign%pe))     deallocate (vign%pe)
    vign%class     = SRCTOOL_VIGN_CLASS_NULL
    vign%num_e     = 0
    vign%num_theta = 0
    vign%num_phi   = 0
    vign%is_valid = .false.
    call fitswrap_init_hduinfo(vign%hduinfo)


  end subroutine srctool_vign_empty_vign
  !---------------------------------------------------------------------------------------------

!unused!   !-------------------------------------------------------------------------------
!unused!   subroutine srctool_vign_setup_dummy_vign(vign)
!unused!     type(vign_type), intent(inout)      :: vign
!unused!
!unused!     vign%is_valid = .true.
!unused!     vign%class = SRCTOOL_VIGN_CLASS_NONE
!unused!     vign%num_e     = 0
!unused!     vign%num_theta = 0
!unused!     vign%num_phi   = 0
!unused!
!unused!   end subroutine srctool_vign_setup_dummy_vign
!unused!   !---------------------------------------------------------------------------------------------

  !-------------------------------------------------------------------------------
  !read the vign description from the current hdu of the fits file identified by unit
  subroutine srctool_vign_read_vign_from_chdu(unit, vign, ierr)
    character(len=*), parameter :: subroutinename='srctool_vign_read_vign_from_chdu'

    integer(kind=int32), intent(in)   :: unit
    type(vign_type), intent(inout)    :: vign
    integer(kind=int32), intent(out)  :: ierr

    integer(kind=int32) :: status
    character(len=STR_MAX) :: str_buffer
    integer(kind=int32) :: i

    ierr = ERROR_LEVEL_OK
    status = 0
!    call wrschatty (subroutinename, "Reading in VIGNET extension from CHDU")

    ! get the hdu info/keys
    call fitswrap_read_hduinfo(unit, vign%hduinfo, ierr)
    if ( ierr .ne. ERROR_LEVEL_OK ) then
      call wrserr (subroutinename, ierr, "Problem reading HDU info")
      return
    end if
    !- check the following keader keywords are present - TODO
    !- HDUCLASS       'OGIP'
    !- HDUCLAS1       'RESPONSE'
    !- HDUCLAS2       'VIGNET'

    ! assume that we have an [r,phi] grid
    vign%class = SRCTOOL_VIGN_CLASS_GRID_RPHI

    !get the number of rows in the extension
    call FTGNRW(unit, vign%num_rows, status)
    if (status .ne. 0) then
      call fitswrap_write_fits_error(subroutinename, status)
      ierr = ERROR_LEVEL_SEVERE
      return
    end if


    ! vignetting information is stored in vector columns
    ! so we need to get the dimensions of each column


    !check that have at least some entries and then read in the column data
    if ( vign%num_rows .gt. 0 ) then
      write (str_buffer, '(A,I0,A,A,A,A,A)') &
          "Reading in: ", vign%num_rows, " rows of VIGNET info from extname='", &
          trim(vign%hduinfo%key_extname%value), "' instrument='", &
          trim(vign%hduinfo%key_instrume%value), "'"
      call wrschatty (subroutinename, str_buffer)
      call fitswrap_proc_vec_col(unit, "ENERG_LO", vign%num_rows, vign%num_e,     vign%pelo,   ierr)
      call fitswrap_proc_vec_col(unit, "ENERG_HI", vign%num_rows, vign%num_e,     vign%pehi,   ierr)
      call fitswrap_proc_vec_col(unit, "THETA",    vign%num_rows, vign%num_theta, vign%ptheta, ierr)
      call fitswrap_proc_vec_col(unit, "PHI",      vign%num_rows, vign%num_phi,   vign%pphi,   ierr)
      vign%num_vign = vign%num_rows * vign%num_e * vign%num_theta * vign%num_phi
      call fitswrap_proc_col    (unit, "VIGNET",   vign%num_vign, vign%pvign,  ierr)

      if (ierr .ne. ERROR_LEVEL_OK ) then
        call wrserr (subroutinename, ierr, "Problem reading column(s) from HDU")
        return
      end if

    end if

    ! now allocate space for, and calculate the reference energy array
    allocate(vign%pe(vign%num_e), STAT = status)
    if ( check_allocate_status(subroutinename, status, ierr) ) return

    do i=1,vign%num_e
      vign%pe(i) = 0.5e0 * (vign%pelo(i) + vign%pehi(i)) ! arithmetic average
!      vign%pe(i) = sqrt(vign%pelo(i)*vign%pehi(i)) ! geometric average
    end do

    !----------------------------------------------
    ! now report the scope of the vignetting data
    call srctool_vign_report_vign_info(vign)
    !----------------------------------------------

    vign%is_valid = .true.

  end subroutine srctool_vign_read_vign_from_chdu
  !---------------------------------------------------------------------------------------------

  !---------------------------------------------------------------------------------------------
  subroutine srctool_vign_report_vign_info(vign)
    character(len=*), parameter :: subroutinename='srctool_vign_report_vign_info'
    type(vign_type), intent(in)    :: vign

    integer(kind=int32) :: e,t,p,i
    character(len=STR_MAX) :: str_buffer
    character(len=STR_MAX) :: buf1
    character(len=STR_MAX) :: buf2
    character(len=STR_MAX) :: buf3

    write (str_buffer, '(A12,1x,A6,1x,A5,1x,A8,1x,A8,1x,A8,1x,A8,1x,A8,1x,A8)') &
        "#VignInfo", "Inst", "Index", "Elo(keV)","Ehi(keV)", "E0(keV)", "Theta", "Phi", "Vignet"
    call wrschatty(subroutinename, str_buffer)
    do e=1,vign%num_e
      write(buf1,'(F8.2,1x,F8.2,1x,F8.2)') vign%pelo(e), vign%pehi(e), vign%pe(e)
      do t=1,vign%num_theta
        write(buf2,'(F8.3)') vign%ptheta(t)
        do p=1,vign%num_phi
!          i = ((e-1)*vign%num_theta + (t-1))*vign%num_phi + p
          i = ((p-1)*vign%num_theta + (t-1))*vign%num_e + e
          write(buf3,'(F8.3)') vign%pphi(p)
          write (str_buffer, '(A12,1x,A6,1x,I5,1x,A26,1x,A8,1x,A8,1x,F8.3)') &
              "VignInfo", trim(vign%hduinfo%key_instrume%value), i, &
              buf1, buf2, buf3, vign%pvign(i)
          call wrschatty(subroutinename, str_buffer)
          write(buf2,'(A8)') "  .. "
        end do
        write(buf1,'(A8,1x,A8,1x,A8)') "  .. ", "  .. ", "  .. "
      end do
    end do

  end subroutine srctool_vign_report_vign_info
  !---------------------------------------------------------------------------------------------


!unused!   !---------------------------------------------------------------------------------------------
!unused!   ! calculates the vignetting at location rawx, rawy for the energies specified in "energies"
!unused!   ! carries out full trilinear interpolation over theta, phi, energy
!unused!   recursive subroutine srctool_vign_calc_vign_values(inst, rawx, rawy, energies, vign_values, ierr)
!unused!     character(len=*), parameter :: subroutinename='srctool_vign_calc_vign_values'
!unused!     type(inst_type), intent(in), target :: inst
!unused!     real(kind=real64), intent(in)  :: rawx, rawy
!unused!     real(kind=real64), dimension(:), intent(in)  :: energies
!unused!     real(kind=real64), dimension(:), intent(out)  :: vign_values
!unused!     integer(kind=int32), intent(out)   :: ierr
!unused!     integer(kind=int32) :: rawx_int, rawy_int, e
!unused!     integer(kind=int32), dimension(1) :: e1,e2,t1,t2,p1,p2
!unused!     integer(kind=int32) :: num_energies
!unused!     real(kind=real64), dimension(1) :: offaxis, phi
!unused!     real(kind=real64), dimension(3)     :: v1, v2, v_eval
!unused!     real(kind=real64), dimension(2,2,2) :: f
!unused!
!unused!
!unused!     type(vign_type), pointer :: pvign
!unused!     type(instpar_type), pointer :: pinstpar
!unused!     ierr=ERROR_LEVEL_OK
!unused!
!unused!     num_energies=size(energies)
!unused!     if ( num_energies .gt. size(vign_values)) then
!unused!       ierr=ERROR_LEVEL_SEVERE
!unused!       call wrserr(subroutinename,ierr,&
!unused!           "The number of requested energies exceeds the space available in the vign_values array")
!unused!       return
!unused!     end if
!unused!
!unused!     nullify (pinstpar)
!unused!     call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
!unused!     if (ierr .ne. ERROR_LEVEL_OK) then
!unused!       call wrserr (subroutinename, ierr, "Could not find suitable instpar")
!unused!       return
!unused!     end if
!unused!
!unused!     nullify (pvign)
!unused!     call srctool_misc_select_vign_pointer(inst,pvign,ierr)
!unused!     if ( ierr .ne.  ERROR_LEVEL_OK ) then
!unused!       return
!unused!     end if
!unused!
!unused!     rawx_int = int(rawx+0.5d0, int32)
!unused!     rawy_int = int(rawy+0.5d0, int32)
!unused!     if ((rawx_int .lt. 1) .or. (rawx_int .gt. pinstpar%naxis1) .or. &
!unused!         (rawy_int .lt. 1) .or. (rawy_int .gt. pinstpar%naxis2)) then
!unused!       vign_values(:) = 0.0d0  ! return zero for out of field positions
!unused!     else if ( pvign%class .eq. SRCTOOL_VIGN_CLASS_NONE ) then
!unused!       vign_values(:) = 1.0d0  ! return 1 for null vignetting
!unused!     !------------------------------------------------------------------
!unused!     else if ( pvign%class .eq. SRCTOOL_VIGN_CLASS_GRID_RPHI ) then
!unused!       ! calc off-axis angle in arcmin
!unused!       offaxis(1) = DEG2ARCMIN * sqrt( &
!unused!           ((rawx - pinstpar%oa_rawx)*pinstpar%pixscale_x)**2 + &
!unused!           ((rawy - pinstpar%oa_rawy)*pinstpar%pixscale_y)**2)
!unused!       ! and azimuthal angle in degrees
!unused!       phi(1) = RAD2DEG*atan2( &
!unused!           (rawy-pinstpar%oa_rawy)*pinstpar%pixscale_y, &
!unused!           (rawx-pinstpar%oa_rawx)*pinstpar%pixscale_x)
!unused!
!unused!       ! get the theta indices which bound the offaxis value
!unused!       call get_interp_indices (REAL(pvign%ptheta(:),real64), offaxis(:), t1(:), t2(:))
!unused!
!unused!       ! get the phi   indices which bound the phi value
!unused!       call get_interp_indices (REAL(pvign%pphi(:),real64), phi(:), p1(:), p2(:))
!unused!
!unused!       v1(2) =  pvign%ptheta(t1(1))
!unused!       v2(2) =  pvign%ptheta(t2(1))
!unused!       v1(3) =  pvign%pphi(p1(1))
!unused!       v2(3) =  pvign%pphi(p2(1))
!unused!
!unused!       v_eval(2) =  offaxis(1)
!unused!       v_eval(3) =  phi(1)
!unused!       do e=1,num_energies
!unused!         ! get the energy indices which bound the energy value
!unused!         call get_interp_indices (REAL(pvign%pe(:),real64), energies(e:e), e1(:), e2(:))
!unused!         v_eval(1) = energies(e)
!unused!         v1(1)     = pvign%pe(e1(1))
!unused!         v2(1)     = pvign%pe(e2(1))
!unused!         f(1,1,1)  = pvign%pvign(e1(1) + pvign%num_e*((t1(1)-1) + (p1(1)-1)*pvign%num_theta))
!unused!         f(1,1,2)  = pvign%pvign(e1(1) + pvign%num_e*((t1(1)-1) + (p2(1)-1)*pvign%num_theta))
!unused!         f(1,2,1)  = pvign%pvign(e1(1) + pvign%num_e*((t2(1)-1) + (p1(1)-1)*pvign%num_theta))
!unused!         f(1,2,2)  = pvign%pvign(e1(1) + pvign%num_e*((t2(1)-1) + (p2(1)-1)*pvign%num_theta))
!unused!         f(2,1,1)  = pvign%pvign(e2(1) + pvign%num_e*((t1(1)-1) + (p1(1)-1)*pvign%num_theta))
!unused!         f(2,1,2)  = pvign%pvign(e2(1) + pvign%num_e*((t1(1)-1) + (p2(1)-1)*pvign%num_theta))
!unused!         f(2,2,1)  = pvign%pvign(e2(1) + pvign%num_e*((t2(1)-1) + (p1(1)-1)*pvign%num_theta))
!unused!         f(2,2,2)  = pvign%pvign(e2(1) + pvign%num_e*((t2(1)-1) + (p2(1)-1)*pvign%num_theta))
!unused!
!unused!         call trilinear_interp ( v1(:), v2(:), f(:,:,:), v_eval(:), vign_values(e) )
!unused!
!unused!       end do
!unused!     else
!unused!       ! unknown Vignetting class
!unused!       ierr = ERROR_LEVEL_SEVERE
!unused!       call wrserr (subroutinename, ierr, "Error, unknown vignetting description type")
!unused!       return
!unused!     end if
!unused!
!unused!
!unused!   end subroutine srctool_vign_calc_vign_values
!unused!   !---------------------------------------------------------------------------------------------


  !---------------------------------------------------------------------------------------------
  ! the following is partly adapted from Hermann's make_vignmaps() subroutine
  ! calculates the vignetting at location rawx, rawy for the energies specified in "energies"
  ! assumes that vigneting is a step function in energy over a set of finite energy bands
  recursive subroutine srctool_vign_calc_vign_values_old(inst, rawx, rawy, energies, vign_values, ierr)
    character(len=*), parameter :: subroutinename='srctool_vign_calc_vign_values_old'
    type(inst_type), intent(in), target :: inst
    real(kind=real64), intent(in)  :: rawx, rawy
    real(kind=real64), dimension(:), intent(in)  :: energies
    real(kind=real64), dimension(:), intent(out)  :: vign_values
    integer(kind=int32), intent(out)   :: ierr
    integer(kind=int32) :: rawx_int, rawy_int, i1,i2, e,ee, t1,t2, p1,p2
    integer(kind=int32) :: num_energies
    real(kind=real64) :: offaxis, phi, energy
    type(vign_type), pointer :: pvign
    type(instpar_type), pointer :: pinstpar
    ierr=ERROR_LEVEL_OK

    num_energies=size(energies)
    if ( num_energies .gt. size(vign_values)) then
      ierr=ERROR_LEVEL_SEVERE
      call wrserr(subroutinename,ierr,&
          "The number of requested energies exceeds the space available in the vign_values array")
      return
    end if

    nullify (pinstpar)
    call srctool_misc_select_instpar_pointer(inst,pinstpar,ierr)
    if (ierr .ne. ERROR_LEVEL_OK) then
      call wrserr (subroutinename, ierr, "Could not find suitable instpar")
      return
    end if

    nullify (pvign)
    call srctool_misc_select_vign_pointer(inst,pvign,ierr)
    if ( ierr .ne.  ERROR_LEVEL_OK ) then
      return
    end if
!
!    if ( inst%vign%is_valid ) then
!      pvign => inst%vign
!    else if ( inst%pinst0%vign%is_valid ) then
!      pvign => inst%pinst0%vign
!    else
!      ierr=ERROR_LEVEL_SEVERE
!      call wrserr(subroutinename,ierr,"No valid vignetting description was found")
!      return
!    end if

!    rawx_int = nint(rawx,int32)
!    rawy_int = nint(rawy,int32)
    rawx_int = int(rawx+0.5d0)
    rawy_int = int(rawy+0.5d0)
    if ((rawx_int .lt. 1) .or. (rawx_int .gt. pinstpar%naxis1) .or. &
        (rawy_int .lt. 1) .or. (rawy_int .gt. pinstpar%naxis2)) then
      vign_values(:) = 0.0d0  ! return zero for out of field positions
    else if ( pvign%class .eq. SRCTOOL_VIGN_CLASS_NONE ) then
      vign_values(:) = 1.0d0  ! return 1 for null vignetting
    else if ( pvign%class .eq. SRCTOOL_VIGN_CLASS_GRID_RPHI ) then
      ! get the theta indices which bound the offaxis value
      ! calc off-axis angle in arcmin
      offaxis = DEG2ARCMIN * sqrt( &
          ((rawx - pinstpar%oa_rawx)*pinstpar%pixscale_x)**2 + &
          ((rawy - pinstpar%oa_rawy)*pinstpar%pixscale_y)**2)
      if ( pvign%num_theta .eq. 1 ) then
        t1 = 1
        t2 = t1
      else if ( offaxis .lt. pvign%ptheta(1) ) then
        t1 = 1
        t2 = t1
      else if ( offaxis .gt. pvign%ptheta(pvign%num_theta) ) then
        t1 = pvign%num_theta
        t2 = t1
      else
        t2 = 2
        do while ( t2 .lt. pvign%num_theta .and. offaxis .ge. pvign%ptheta(t2))
          t2 = t2 + 1
        end do
        t1 = t2 - 1
      end if


      ! get the phi indices which bound the phi value
      ! calc phi - assume it's in degrees in the calibration file ! TODO allow radians
      if ( pvign%num_phi .eq. 1 ) then
        p1 = 1
        p2 = p1
      else
        phi = RAD2DEG*atan2((rawy-pinstpar%oa_rawy)*pinstpar%pixscale_y, (rawx-pinstpar%oa_rawx)*pinstpar%pixscale_x)
        if ( phi .lt. pvign%pphi(1) ) then
          p1 = 1
          p2 = p1
        else if ( phi .gt. pvign%pphi(pvign%num_phi) ) then
          p1 = pvign%num_phi
          p2 = p1
        else
          p2 = 2
          do while ( (p2 .lt. pvign%num_phi) .and. (phi .ge. pvign%pphi(p2)))
            p2 = p2 + 1
          end do
          p1 = p2 - 1
        end if
      end if

      do ee=1,num_energies
        energy = energies(ee)

        ! get the correct energy band
        if ( energy .lt. pvign%pelo(1) ) then
          e = 1
        else if ( energy .gt. pvign%pehi(pvign%num_e) ) then
          e =  pvign%num_e
        else
          e = 1
          do while ( (e .lt. pvign%num_e) .and. (energy .ge. pvign%pehi(e)))
            e = e + 1
          end do
        end if

        ! interpolate in theta direction
        ! need to add in a proper interpolation code here
        ! should probably use a bicubic spline
        ! TODO - include phi interpolation
        i1 = ((p1-1)*pvign%num_theta + (t1-1))*pvign%num_e + e
        i2 = ((p1-1)*pvign%num_theta + (t2-1))*pvign%num_e + e
        if ( t1 .ne. t2 ) then
          vign_values(ee) = pvign%pvign(i1) + &
              (pvign%pvign(i2) - pvign%pvign(i1))*(offaxis - pvign%ptheta(t1))&
              /(pvign%ptheta(t2) - pvign%ptheta(t1))
        else
          vign_values(ee) = pvign%pvign(i1)
        end if
      end do
    else
      ! unknown Vignetting class
      ierr = ERROR_LEVEL_SEVERE
      call wrserr (subroutinename, ierr, "Error, unknown vignetting description type")
      return
    end if


  end subroutine srctool_vign_calc_vign_values_old
  !---------------------------------------------------------------------------------------------




  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------END of VIGNETTING file handling  ---------------------------
  !-------------------------------------------------------------------------------------
  !-------------------------------------------------------------------------------------


end module srctool_vign
